(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    735864,      28748]*)
(*NotebookOutlinePosition[    736668,      28776]*)
(*  CellTagsIndexPosition[    736624,      28772]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["A simple FEM Code", "Title",
  TextAlignment->Center],

Cell["\<\
Version 0.1, April 2000
Andreas Stahel, HTA Biel, Switzerland
Andreas.Stahel@hta-bi.bfh.ch\
\>", "Text",
  TextAlignment->Center],

Cell[CellGroupData[{

Cell["Description", "Section"],

Cell[TextData[{
  "This package allows to solve second order elliptic differential equations \
in two variables:\n\n     div(a*grad u) - b*u = f         in domain\n         \
                             u =  gD     Dirichlet boundary conditions on \
first part of boundary\n                         a*du/dn =  gN      Neumann \
condition on the other part of the boundary\n                       \nIf the \
functions a, b f, gD and gN are given, then a numerical approximation is \
computed, using the method of finite elements.\n\nThe mesh on the domain has \
to be generated by the program ",
  StyleBox["EasyMesh",
    FontWeight->"Bold"],
  ".  Within ",
  StyleBox["EasyMesh",
    FontWeight->"Bold"],
  " the boundary points have to be marked, such that the code knows whether \
to apply Dirichlet or Neumann boundary conditions. The codes are\n      type \
1  = Dirichlet boundary\n      type 2  = Neumann boundary\nThe files ",
  StyleBox["EasyMeshIntro.A4.ps",
    FontWeight->"Bold"],
  " and ",
  StyleBox["EasyMeshIntro.letter.ps ",
    FontWeight->"Bold"],
  "may help to get started. But as usual the standard documentation is best, \
see at http://www-dinma.univ.trieste.it/~nirftc/research/easymesh/ \nIf you \
plan to use this package you will have to install ",
  StyleBox["EasyMesh",
    FontWeight->"Bold"],
  ". The enclosed examples can be run without installing the mesh generator.\n\
\nThe command given in this package should\n-  read the information on the \
mesh.\n-  create the finite element model, set up the system of linear \
equations and solve it.\n-  allow to visualize the solutions.\nConsider the \
enclosed example.\n\nThe notebook ",
  StyleBox["FEM2DExamples.nb ",
    FontWeight->"Bold"],
  "contains a few examples.\nThe notebook ",
  StyleBox["boundarygenerator.nb",
    FontWeight->"Bold"],
  " ( written by Pierre Andre Chevalier) does allow to generate the input \
file for ",
  StyleBox["EasyMesh",
    FontWeight->"Bold"],
  " within a ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " notebook.\nThe subdirectory ./EasyMesh contains some input files for the \
examples.\n\n\nThis code was written as support material for a class on ",
  StyleBox["Calculus of Variations and Finite Elements",
    FontSlant->"Italic"],
  ". It should only be used for educational purposes.  It was developed and \
tested on a Linux system and thus will \nbehave the same on all Unix systems. \
I will try to help to get it up and running. On MS basesd systems you are on \
your own.\n- If you like it, let me know\n- If you have suggestions on how to \
impove the package, let me know. \n   There is some obvious space for \
improvement.\n- If you think it is a piece of .... , delete it and dont tell \
anybody \n\nEnjoy this day\n\nAndreas"
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["A First Example", "Section"],

Cell[CellGroupData[{

Cell["Reading the Package and display the built in documentation", 
  "Subsection"],

Cell[BoxData[
    \(<< BVP2.m\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?ReadMesh\)\)], "Input"],

Cell[BoxData[
    \("{nodes,elements,segments}=ReadMesh[filename] Read all information for \
a mesh generated by EasyMesh filename is a string with the name of the mesh \
files without extension,  e.g. EasyMesh/test2"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?MeshGraphics\)\)], "Input"],

Cell[BoxData[
    \("gmesh=MeshGraphics[nodes,elements] generates a graphics object to \
visualize the mesh by Show[gmesh]"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?FEMSolve\)\)], "Input"],

Cell[BoxData[
    \("points=FEMSolve[a, b, f, gDirichlet, \
gNeumann,nodes,elements,segments]; Solve a second order boundary value \
problem"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?FEMSurface\)\)], "Input"],

Cell[BoxData[
    \("g=FEMSurface[points,elements] generates a 3D graphics object to be \
displayed with Show[g]"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?FEMSurfaceNoMesh\)\)], "Input"],

Cell[BoxData[
    \("g=FEMSurfaceNoMesh[points,elements] generates a 3D graphics object to \
be displayed with Show[g]\n the mesh is not shown"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?LevelGraphics2D\)\)], "Input"],

Cell[BoxData[
    \("glevel=LevelGraphics2D[points,elements,levels] generates a graphics \
object to visualize the mesh by Show[glevel]"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?LevelGraphics3D\)\)], "Input"],

Cell[BoxData[
    \("glevel=LevelGraphics3D[points,elements,levels] generates a graphics \
object to visualize the mesh by Show[glevel]"\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Visualization of Mesh for Control", "Subsection"],

Cell[BoxData[
    \(\({nodes, elements, segments} = ReadMesh["\<./example2\>"]; \)\)], 
  "Input"],

Cell[BoxData[
    \(\(gmesh = MeshGraphics[nodes, elements]; \)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[gmesh\ , \nAspectRatio -> Automatic, \ \nAxes -> Automatic]; 
    \)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .6 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0.0142857 0.190476 [
[.21429 .00179 -3.6 -10.8 ]
[.21429 .00179 3.6 0 ]
[.40476 .00179 -3.6 -10.8 ]
[.40476 .00179 3.6 0 ]
[.59524 .00179 -3.6 -10.8 ]
[.59524 .00179 3.6 0 ]
[.78571 .00179 -3.6 -10.8 ]
[.78571 .00179 3.6 0 ]
[.97619 .00179 -3.6 -10.8 ]
[.97619 .00179 3.6 0 ]
[.01131 .10952 -21.6 -5.4 ]
[.01131 .10952 0 5.4 ]
[.01131 .20476 -7.2 -5.4 ]
[.01131 .20476 0 5.4 ]
[.01131 .3 -21.6 -5.4 ]
[.01131 .3 0 5.4 ]
[.01131 .39524 -7.2 -5.4 ]
[.01131 .39524 0 5.4 ]
[.01131 .49048 -21.6 -5.4 ]
[.01131 .49048 0 5.4 ]
[.01131 .58571 -7.2 -5.4 ]
[.01131 .58571 0 5.4 ]
[ 0 0 0 0 ]
[ 1 .6 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01429 m
.21429 .02054 L
s
[(1)] .21429 .00179 0 1 Mshowa
.40476 .01429 m
.40476 .02054 L
s
[(2)] .40476 .00179 0 1 Mshowa
.59524 .01429 m
.59524 .02054 L
s
[(3)] .59524 .00179 0 1 Mshowa
.78571 .01429 m
.78571 .02054 L
s
[(4)] .78571 .00179 0 1 Mshowa
.97619 .01429 m
.97619 .02054 L
s
[(5)] .97619 .00179 0 1 Mshowa
.125 Mabswid
.0619 .01429 m
.0619 .01804 L
s
.1 .01429 m
.1 .01804 L
s
.1381 .01429 m
.1381 .01804 L
s
.17619 .01429 m
.17619 .01804 L
s
.25238 .01429 m
.25238 .01804 L
s
.29048 .01429 m
.29048 .01804 L
s
.32857 .01429 m
.32857 .01804 L
s
.36667 .01429 m
.36667 .01804 L
s
.44286 .01429 m
.44286 .01804 L
s
.48095 .01429 m
.48095 .01804 L
s
.51905 .01429 m
.51905 .01804 L
s
.55714 .01429 m
.55714 .01804 L
s
.63333 .01429 m
.63333 .01804 L
s
.67143 .01429 m
.67143 .01804 L
s
.70952 .01429 m
.70952 .01804 L
s
.74762 .01429 m
.74762 .01804 L
s
.82381 .01429 m
.82381 .01804 L
s
.8619 .01429 m
.8619 .01804 L
s
.9 .01429 m
.9 .01804 L
s
.9381 .01429 m
.9381 .01804 L
s
.25 Mabswid
0 .01429 m
1 .01429 L
s
.02381 .10952 m
.03006 .10952 L
s
[(0.5)] .01131 .10952 1 0 Mshowa
.02381 .20476 m
.03006 .20476 L
s
[(1)] .01131 .20476 1 0 Mshowa
.02381 .3 m
.03006 .3 L
s
[(1.5)] .01131 .3 1 0 Mshowa
.02381 .39524 m
.03006 .39524 L
s
[(2)] .01131 .39524 1 0 Mshowa
.02381 .49048 m
.03006 .49048 L
s
[(2.5)] .01131 .49048 1 0 Mshowa
.02381 .58571 m
.03006 .58571 L
s
[(3)] .01131 .58571 1 0 Mshowa
.125 Mabswid
.02381 .03333 m
.02756 .03333 L
s
.02381 .05238 m
.02756 .05238 L
s
.02381 .07143 m
.02756 .07143 L
s
.02381 .09048 m
.02756 .09048 L
s
.02381 .12857 m
.02756 .12857 L
s
.02381 .14762 m
.02756 .14762 L
s
.02381 .16667 m
.02756 .16667 L
s
.02381 .18571 m
.02756 .18571 L
s
.02381 .22381 m
.02756 .22381 L
s
.02381 .24286 m
.02756 .24286 L
s
.02381 .2619 m
.02756 .2619 L
s
.02381 .28095 m
.02756 .28095 L
s
.02381 .31905 m
.02756 .31905 L
s
.02381 .3381 m
.02756 .3381 L
s
.02381 .35714 m
.02756 .35714 L
s
.02381 .37619 m
.02756 .37619 L
s
.02381 .41429 m
.02756 .41429 L
s
.02381 .43333 m
.02756 .43333 L
s
.02381 .45238 m
.02756 .45238 L
s
.02381 .47143 m
.02756 .47143 L
s
.02381 .50952 m
.02756 .50952 L
s
.02381 .52857 m
.02756 .52857 L
s
.02381 .54762 m
.02756 .54762 L
s
.02381 .56667 m
.02756 .56667 L
s
.25 Mabswid
.02381 0 m
.02381 .6 L
s
0 0 m
1 0 L
1 .6 L
0 .6 L
closepath
clip
newpath
.5 Mabswid
.02381 .01429 m
.04865 .03912 L
.02381 .05238 L
.02381 .01429 L
s
.02381 .01429 m
.0619 .01429 L
.04865 .03912 L
.02381 .01429 L
s
.05439 .06986 m
.02381 .05238 L
.04865 .03912 L
.05439 .06986 L
s
.04865 .03912 m
.0619 .01429 L
.07937 .04492 L
.04865 .03912 L
s
.05439 .06986 m
.04865 .03912 L
.07937 .04492 L
.05439 .06986 L
s
.02381 .05238 m
.05439 .06986 L
.02381 .09048 L
.02381 .05238 L
s
.0619 .01429 m
.1 .01429 L
.07937 .04492 L
.0619 .01429 L
s
.09395 .08469 m
.05439 .06986 L
.07937 .04492 L
.09395 .08469 L
s
.05686 .108 m
.02381 .09048 L
.05439 .06986 L
.05686 .108 L
s
.05439 .06986 m
.09395 .08469 L
.05686 .108 L
.05439 .06986 L
s
.07937 .04492 m
.1 .01429 L
.11744 .04771 L
.07937 .04492 L
s
.09395 .08469 m
.07937 .04492 L
.11744 .04771 L
.09395 .08469 L
s
.05686 .108 m
.02381 .12857 L
.02381 .09048 L
.05686 .108 L
s
.1 .01429 m
.1381 .01429 L
.11744 .04771 L
.1 .01429 L
s
.08925 .1281 m
.05686 .108 L
.09395 .08469 L
.08925 .1281 L
s
.11744 .04771 m
.13685 .08069 L
.09395 .08469 L
.11744 .04771 L
s
.02381 .12857 m
.05686 .108 L
.0561 .14681 L
.02381 .12857 L
s
.12382 .11463 m
.08925 .1281 L
.09395 .08469 L
.12382 .11463 L
s
.08925 .1281 m
.0561 .14681 L
.05686 .108 L
.08925 .1281 L
s
.12382 .11463 m
.09395 .08469 L
.13685 .08069 L
.12382 .11463 L
s
.15658 .048 m
.11744 .04771 L
.1381 .01429 L
.15658 .048 L
s
.11744 .04771 m
.15658 .048 L
.13685 .08069 L
.11744 .04771 L
s
.0561 .14681 m
.02381 .16667 L
.02381 .12857 L
.0561 .14681 L
s
.08699 .16547 m
.0561 .14681 L
.08925 .1281 L
.08699 .16547 L
s
.15658 .048 m
.1381 .01429 L
.17619 .01429 L
.15658 .048 L
s
.11811 .14992 m
.08925 .1281 L
.12382 .11463 L
.11811 .14992 L
s
.12382 .11463 m
.13685 .08069 L
.15475 .10763 L
.12382 .11463 L
s
.11811 .14992 m
.08699 .16547 L
.08925 .1281 L
.11811 .14992 L
s
.15658 .048 m
.17514 .08259 L
.13685 .08069 L
.15658 .048 L
s
.05608 .18448 m
.02381 .16667 L
.0561 .14681 L
.05608 .18448 L
s
.08699 .16547 m
.05608 .18448 L
.0561 .14681 L
.08699 .16547 L
s
.13685 .08069 m
.17514 .08259 L
.15475 .10763 L
.13685 .08069 L
s
.15658 .048 m
.17619 .01429 L
.19615 .0491 L
.15658 .048 L
s
.11811 .14992 m
.12382 .11463 L
.15058 .13798 L
.11811 .14992 L
s
.12382 .11463 m
.15475 .10763 L
.15058 .13798 L
.12382 .11463 L
s
.15658 .048 m
.19615 .0491 L
.17514 .08259 L
.15658 .048 L
s
.11482 .18314 m
.08699 .16547 L
.11811 .14992 L
.11482 .18314 L
s
.05608 .18448 m
.02381 .20476 L
.02381 .16667 L
.05608 .18448 L
s
.08792 .20098 m
.05608 .18448 L
.08699 .16547 L
.08792 .20098 L
s
.19615 .0491 m
.17619 .01429 L
.21429 .01429 L
.19615 .0491 L
s
.08792 .20098 m
.08699 .16547 L
.11482 .18314 L
.08792 .20098 L
s
.15475 .10763 m
.17514 .08259 L
.18799 .12338 L
.15475 .10763 L
s
.11811 .14992 m
.15058 .13798 L
.14374 .17113 L
.11811 .14992 L
s
.15475 .10763 m
.18799 .12338 L
.15058 .13798 L
.15475 .10763 L
s
.11482 .18314 m
.11811 .14992 L
.14374 .17113 L
.11482 .18314 L
s
.19615 .0491 m
.21924 .08777 L
.17514 .08259 L
.19615 .0491 L
s
.05608 .18448 m
.05803 .22244 L
.02381 .20476 L
.05608 .18448 L
s
.05803 .22244 m
.05608 .18448 L
.08792 .20098 L
.05803 .22244 L
s
.18799 .12338 m
.17514 .08259 L
.21924 .08777 L
.18799 .12338 L
s
.21429 .01429 m
.2359 .0482 L
.19615 .0491 L
.21429 .01429 L
s
.11482 .18314 m
.11445 .21088 L
.08792 .20098 L
.11482 .18314 L
s
.19615 .0491 m
.2359 .0482 L
.21924 .08777 L
.19615 .0491 L
s
.18799 .12338 m
.17583 .1623 L
.15058 .13798 L
.18799 .12338 L
s
.15058 .13798 m
.17583 .1623 L
.14374 .17113 L
.15058 .13798 L
s
.13815 .20083 m
.11482 .18314 L
.14374 .17113 L
.13815 .20083 L
s
.05803 .22244 m
.02381 .24286 L
.02381 .20476 L
.05803 .22244 L
s
.11482 .18314 m
.13815 .20083 L
.11445 .21088 L
.11482 .18314 L
s
.05803 .22244 m
.08792 .20098 L
.09757 .23962 L
.05803 .22244 L
s
.18799 .12338 m
.21924 .08777 L
.23014 .12786 L
.18799 .12338 L
s
.11445 .21088 m
.09757 .23962 L
.08792 .20098 L
.11445 .21088 L
s
.21429 .01429 m
.25238 .01429 L
.2359 .0482 L
.21429 .01429 L
s
.2359 .0482 m
.25936 .07807 L
.21924 .08777 L
.2359 .0482 L
s
.21129 .15871 m
.17583 .1623 L
.18799 .12338 L
.21129 .15871 L
s
.14374 .17113 m
.17583 .1623 L
.16546 .19317 L
.14374 .17113 L
s
.14374 .17113 m
.16546 .19317 L
.13815 .20083 L
.14374 .17113 L
s
.21129 .15871 m
.18799 .12338 L
.23014 .12786 L
.21129 .15871 L
s
.05932 .26193 m
.02381 .24286 L
.05803 .22244 L
.05932 .26193 L
s
.05932 .26193 m
.05803 .22244 L
.09757 .23962 L
.05932 .26193 L
s
.25079 .10606 m
.23014 .12786 L
.21924 .08777 L
.25079 .10606 L
s
.11445 .21088 m
.13815 .20083 L
.13417 .22972 L
.11445 .21088 L
s
.25079 .10606 m
.21924 .08777 L
.25936 .07807 L
.25079 .10606 L
s
.13417 .22972 m
.09757 .23962 L
.11445 .21088 L
.13417 .22972 L
s
.2359 .0482 m
.25238 .01429 L
.27422 .04598 L
.2359 .0482 L
s
.2359 .0482 m
.27422 .04598 L
.25936 .07807 L
.2359 .0482 L
s
.17583 .1623 m
.21129 .15871 L
.19657 .18999 L
.17583 .1623 L
s
.17583 .1623 m
.19657 .18999 L
.16546 .19317 L
.17583 .1623 L
s
.16546 .19317 m
.15666 .21681 L
.13815 .20083 L
.16546 .19317 L
s
.15666 .21681 m
.13417 .22972 L
.13815 .20083 L
.15666 .21681 L
s
.05932 .26193 m
.02381 .28095 L
.02381 .24286 L
.05932 .26193 L
s
.23014 .12786 m
.24736 .15986 L
.21129 .15871 L
.23014 .12786 L
s
.05932 .26193 m
.09757 .23962 L
.0947 .2835 L
.05932 .26193 L
s
.13417 .22972 m
.13516 .26764 L
.09757 .23962 L
.13417 .22972 L
s
.25079 .10606 m
.26495 .13213 L
.23014 .12786 L
.25079 .10606 L
s
.25079 .10606 m
.25936 .07807 L
.28075 .10498 L
.25079 .10606 L
s
.27422 .04598 m
.25238 .01429 L
.29048 .01429 L
.27422 .04598 L
s
.13516 .26764 m
.0947 .2835 L
.09757 .23962 L
.13516 .26764 L
s
.26495 .13213 m
.24736 .15986 L
.23014 .12786 L
.26495 .13213 L
s
.27422 .04598 m
.29551 .07583 L
.25936 .07807 L
.27422 .04598 L
s
.21129 .15871 m
.23059 .18939 L
.19657 .18999 L
.21129 .15871 L
s
.18224 .21784 m
.16546 .19317 L
.19657 .18999 L
.18224 .21784 L
s
.18224 .21784 m
.15666 .21681 L
.16546 .19317 L
.18224 .21784 L
s
.21129 .15871 m
.24736 .15986 L
.23059 .18939 L
.21129 .15871 L
s
.29551 .07583 m
.28075 .10498 L
.25936 .07807 L
.29551 .07583 L
s
.05932 .26193 m
.05848 .3021 L
.02381 .28095 L
.05932 .26193 L
s
.25079 .10606 m
.28075 .10498 L
.26495 .13213 L
.25079 .10606 L
s
.05848 .3021 m
.05932 .26193 L
.0947 .2835 L
.05848 .3021 L
s
.15666 .21681 m
.16356 .24199 L
.13417 .22972 L
.15666 .21681 L
s
.13417 .22972 m
.16356 .24199 L
.13516 .26764 L
.13417 .22972 L
s
.29048 .01429 m
.31163 .0453 L
.27422 .04598 L
.29048 .01429 L
s
.29551 .07583 m
.27422 .04598 L
.31163 .0453 L
.29551 .07583 L
s
.28227 .16135 m
.24736 .15986 L
.26495 .13213 L
.28227 .16135 L
s
.18224 .21784 m
.16356 .24199 L
.15666 .21681 L
.18224 .21784 L
s
.1236 .30255 m
.0947 .2835 L
.13516 .26764 L
.1236 .30255 L
s
.21443 .21844 m
.19657 .18999 L
.23059 .18939 L
.21443 .21844 L
s
.18224 .21784 m
.19657 .18999 L
.21443 .21844 L
.18224 .21784 L
s
.24736 .15986 m
.26547 .19011 L
.23059 .18939 L
.24736 .15986 L
s
.29884 .13318 m
.26495 .13213 L
.28075 .10498 L
.29884 .13318 L
s
.02381 .31905 m
.02381 .28095 L
.05848 .3021 L
.02381 .31905 L
s
.24736 .15986 m
.28227 .16135 L
.26547 .19011 L
.24736 .15986 L
s
.05848 .3021 m
.0947 .2835 L
.09499 .32697 L
.05848 .3021 L
s
.29551 .07583 m
.31506 .10492 L
.28075 .10498 L
.29551 .07583 L
s
.29884 .13318 m
.28227 .16135 L
.26495 .13213 L
.29884 .13318 L
s
.1236 .30255 m
.09499 .32697 L
.0947 .2835 L
.1236 .30255 L
s
.13516 .26764 m
.16356 .24199 L
.1766 .27314 L
.13516 .26764 L
s
.15686 .30232 m
.1236 .30255 L
.13516 .26764 L
.15686 .30232 L
s
.29048 .01429 m
.32857 .01429 L
.31163 .0453 L
.29048 .01429 L
s
.31506 .10492 m
.29884 .13318 L
.28075 .10498 L
.31506 .10492 L
s
.29551 .07583 m
.31163 .0453 L
.33165 .07581 L
.29551 .07583 L
s
.18224 .21784 m
.19694 .24597 L
.16356 .24199 L
.18224 .21784 L
s
.1766 .27314 m
.15686 .30232 L
.13516 .26764 L
.1766 .27314 L
s
.21443 .21844 m
.23059 .18939 L
.24874 .21913 L
.21443 .21844 L
s
.18224 .21784 m
.21443 .21844 L
.19694 .24597 L
.18224 .21784 L
s
.26547 .19011 m
.24874 .21913 L
.23059 .18939 L
.26547 .19011 L
s
.31506 .10492 m
.29551 .07583 L
.33165 .07581 L
.31506 .10492 L
s
.05483 .34017 m
.02381 .31905 L
.05848 .3021 L
.05483 .34017 L
s
.09499 .32697 m
.05483 .34017 L
.05848 .3021 L
.09499 .32697 L
s
.1766 .27314 m
.16356 .24199 L
.19694 .24597 L
.1766 .27314 L
s
.26547 .19011 m
.28227 .16135 L
.30033 .19093 L
.26547 .19011 L
s
.29884 .13318 m
.31674 .1621 L
.28227 .16135 L
.29884 .13318 L
s
.1236 .30255 m
.13727 .33213 L
.09499 .32697 L
.1236 .30255 L
s
.31163 .0453 m
.32857 .01429 L
.34892 .04558 L
.31163 .0453 L
s
.15686 .30232 m
.13727 .33213 L
.1236 .30255 L
.15686 .30232 L
s
.33165 .07581 m
.31163 .0453 L
.34892 .04558 L
.33165 .07581 L
s
.333 .13384 m
.29884 .13318 L
.31506 .10492 L
.333 .13384 L
s
.30033 .19093 m
.28227 .16135 L
.31674 .1621 L
.30033 .19093 L
s
.23157 .24768 m
.19694 .24597 L
.21443 .21844 L
.23157 .24768 L
s
.23157 .24768 m
.21443 .21844 L
.24874 .21913 L
.23157 .24768 L
s
.26547 .19011 m
.28361 .21997 L
.24874 .21913 L
.26547 .19011 L
s
.19428 .30434 m
.15686 .30232 L
.1766 .27314 L
.19428 .30434 L
s
.26547 .19011 m
.30033 .19093 L
.28361 .21997 L
.26547 .19011 L
s
.333 .13384 m
.31674 .1621 L
.29884 .13318 L
.333 .13384 L
s
.31506 .10492 m
.33165 .07581 L
.3499 .10592 L
.31506 .10492 L
s
.02381 .35714 m
.02381 .31905 L
.05483 .34017 L
.02381 .35714 L
s
.09499 .32697 m
.0796 .36521 L
.05483 .34017 L
.09499 .32697 L
s
.1766 .27314 m
.19694 .24597 L
.2133 .2758 L
.1766 .27314 L
s
.333 .13384 m
.31506 .10492 L
.3499 .10592 L
.333 .13384 L
s
.13727 .33213 m
.11761 .36313 L
.09499 .32697 L
.13727 .33213 L
s
.15686 .30232 m
.17523 .33374 L
.13727 .33213 L
.15686 .30232 L
s
.2133 .2758 m
.19428 .30434 L
.1766 .27314 L
.2133 .2758 L
s
.34892 .04558 m
.32857 .01429 L
.36667 .01429 L
.34892 .04558 L
s
.0796 .36521 m
.09499 .32697 L
.11761 .36313 L
.0796 .36521 L
s
.2133 .2758 m
.19694 .24597 L
.23157 .24768 L
.2133 .2758 L
s
.33165 .07581 m
.34892 .04558 L
.36795 .07719 L
.33165 .07581 L
s
.19428 .30434 m
.17523 .33374 L
.15686 .30232 L
.19428 .30434 L
s
.30033 .19093 m
.31674 .1621 L
.33501 .1912 L
.30033 .19093 L
s
.3499 .10592 m
.33165 .07581 L
.36795 .07719 L
.3499 .10592 L
s
.24874 .21913 m
.26658 .24873 L
.23157 .24768 L
.24874 .21913 L
s
.28361 .21997 m
.26658 .24873 L
.24874 .21913 L
.28361 .21997 L
s
.333 .13384 m
.35055 .16217 L
.31674 .1621 L
.333 .13384 L
s
.30033 .19093 m
.31864 .22077 L
.28361 .21997 L
.30033 .19093 L
s
.04875 .37059 m
.02381 .35714 L
.05483 .34017 L
.04875 .37059 L
s
.0796 .36521 m
.04875 .37059 L
.05483 .34017 L
.0796 .36521 L
s
.36598 .13469 m
.333 .13384 L
.3499 .10592 L
.36598 .13469 L
s
.33501 .1912 m
.31674 .1621 L
.35055 .16217 L
.33501 .1912 L
s
.30033 .19093 m
.33501 .1912 L
.31864 .22077 L
.30033 .19093 L
s
.13727 .33213 m
.15638 .36402 L
.11761 .36313 L
.13727 .33213 L
s
.17523 .33374 m
.15638 .36402 L
.13727 .33213 L
.17523 .33374 L
s
.2133 .2758 m
.23157 .24768 L
.24914 .27715 L
.2133 .2758 L
s
.36667 .01429 m
.3864 .04652 L
.34892 .04558 L
.36667 .01429 L
s
.36795 .07719 m
.34892 .04558 L
.3864 .04652 L
.36795 .07719 L
s
.23121 .30571 m
.19428 .30434 L
.2133 .2758 L
.23121 .30571 L
s
.36598 .13469 m
.35055 .16217 L
.333 .13384 L
.36598 .13469 L
s
.24914 .27715 m
.23157 .24768 L
.26658 .24873 L
.24914 .27715 L
s
.19428 .30434 m
.21291 .33484 L
.17523 .33374 L
.19428 .30434 L
s
.11761 .36313 m
.1 .39524 L
.0796 .36521 L
.11761 .36313 L
s
.28361 .21997 m
.30148 .24978 L
.26658 .24873 L
.28361 .21997 L
s
.3499 .10592 m
.36795 .07719 L
.38448 .10892 L
.3499 .10592 L
s
.31864 .22077 m
.30148 .24978 L
.28361 .21997 L
.31864 .22077 L
s
.36598 .13469 m
.3499 .10592 L
.38448 .10892 L
.36598 .13469 L
s
.23121 .30571 m
.21291 .33484 L
.19428 .30434 L
.23121 .30571 L
s
.04875 .37059 m
.02381 .39524 L
.02381 .35714 L
.04875 .37059 L
s
.24914 .27715 m
.23121 .30571 L
.2133 .2758 L
.24914 .27715 L
s
.0796 .36521 m
.0619 .39524 L
.04875 .37059 L
.0796 .36521 L
s
.1381 .39524 m
.11761 .36313 L
.15638 .36402 L
.1381 .39524 L
s
.33501 .1912 m
.35055 .16217 L
.3691 .19004 L
.33501 .1912 L
s
.0796 .36521 m
.1 .39524 L
.0619 .39524 L
.0796 .36521 L
s
.17523 .33374 m
.1945 .36469 L
.15638 .36402 L
.17523 .33374 L
s
.33501 .1912 m
.35409 .22096 L
.31864 .22077 L
.33501 .1912 L
s
.1381 .39524 m
.1 .39524 L
.11761 .36313 L
.1381 .39524 L
s
.21291 .33484 m
.1945 .36469 L
.17523 .33374 L
.21291 .33484 L
s
.36667 .01429 m
.40476 .01429 L
.3864 .04652 L
.36667 .01429 L
s
.36795 .07719 m
.3864 .04652 L
.40509 .08022 L
.36795 .07719 L
s
.36598 .13469 m
.38259 .16096 L
.35055 .16217 L
.36598 .13469 L
s
.24914 .27715 m
.26658 .24873 L
.28426 .27799 L
.24914 .27715 L
s
.04875 .37059 m
.0619 .39524 L
.02381 .39524 L
.04875 .37059 L
s
.38448 .10892 m
.36795 .07719 L
.40509 .08022 L
.38448 .10892 L
s
.33501 .1912 m
.3691 .19004 L
.35409 .22096 L
.33501 .1912 L
s
.28426 .27799 m
.26658 .24873 L
.30148 .24978 L
.28426 .27799 L
s
.31864 .22077 m
.33669 .2515 L
.30148 .24978 L
.31864 .22077 L
s
.3691 .19004 m
.35055 .16217 L
.38259 .16096 L
.3691 .19004 L
s
.26732 .30629 m
.23121 .30571 L
.24914 .27715 L
.26732 .30629 L
s
.36598 .13469 m
.38448 .10892 L
.39401 .13608 L
.36598 .13469 L
s
.35409 .22096 m
.33669 .2515 L
.31864 .22077 L
.35409 .22096 L
s
.23121 .30571 m
.2501 .33534 L
.21291 .33484 L
.23121 .30571 L
s
.15638 .36402 m
.17619 .39524 L
.1381 .39524 L
.15638 .36402 L
s
.1945 .36469 m
.17619 .39524 L
.15638 .36402 L
.1945 .36469 L
s
.36598 .13469 m
.39401 .13608 L
.38259 .16096 L
.36598 .13469 L
s
.26732 .30629 m
.24914 .27715 L
.28426 .27799 L
.26732 .30629 L
s
.40476 .01429 m
.42432 .04748 L
.3864 .04652 L
.40476 .01429 L
s
.23234 .36506 m
.1945 .36469 L
.21291 .33484 L
.23234 .36506 L
s
.40509 .08022 m
.3864 .04652 L
.42432 .04748 L
.40509 .08022 L
s
.26732 .30629 m
.2501 .33534 L
.23121 .30571 L
.26732 .30629 L
s
.38448 .10892 m
.40509 .08022 L
.42248 .11923 L
.38448 .10892 L
s
.23234 .36506 m
.21291 .33484 L
.2501 .33534 L
.23234 .36506 L
s
.28426 .27799 m
.30148 .24978 L
.31821 .27928 L
.28426 .27799 L
s
.31821 .27928 m
.30148 .24978 L
.33669 .2515 L
.31821 .27928 L
s
.39019 .21859 m
.35409 .22096 L
.3691 .19004 L
.39019 .21859 L
s
.42248 .11923 m
.39401 .13608 L
.38448 .10892 L
.42248 .11923 L
s
.4014 .18606 m
.3691 .19004 L
.38259 .16096 L
.4014 .18606 L
s
.35409 .22096 m
.37426 .25336 L
.33669 .2515 L
.35409 .22096 L
s
.1945 .36469 m
.21429 .39524 L
.17619 .39524 L
.1945 .36469 L
s
.23234 .36506 m
.21429 .39524 L
.1945 .36469 L
.23234 .36506 L
s
.39019 .21859 m
.3691 .19004 L
.4014 .18606 L
.39019 .21859 L
s
.30182 .30616 m
.26732 .30629 L
.28426 .27799 L
.30182 .30616 L
s
.39401 .13608 m
.41339 .15637 L
.38259 .16096 L
.39401 .13608 L
s
.39019 .21859 m
.37426 .25336 L
.35409 .22096 L
.39019 .21859 L
s
.42432 .04748 m
.40476 .01429 L
.44286 .01429 L
.42432 .04748 L
s
.40509 .08022 m
.42432 .04748 L
.44374 .08186 L
.40509 .08022 L
s
.44374 .08186 m
.42248 .11923 L
.40509 .08022 L
.44374 .08186 L
s
.30182 .30616 m
.28426 .27799 L
.31821 .27928 L
.30182 .30616 L
s
.28655 .33499 m
.2501 .33534 L
.26732 .30629 L
.28655 .33499 L
s
.4014 .18606 m
.38259 .16096 L
.41339 .15637 L
.4014 .18606 L
s
.42248 .11923 m
.41339 .15637 L
.39401 .13608 L
.42248 .11923 L
s
.31821 .27928 m
.33669 .2515 L
.35256 .2845 L
.31821 .27928 L
s
.26989 .36501 m
.23234 .36506 L
.2501 .33534 L
.26989 .36501 L
s
.35256 .2845 m
.33669 .2515 L
.37426 .25336 L
.35256 .2845 L
s
.30182 .30616 m
.28655 .33499 L
.26732 .30629 L
.30182 .30616 L
s
.25238 .39524 m
.21429 .39524 L
.23234 .36506 L
.25238 .39524 L
s
.26989 .36501 m
.2501 .33534 L
.28655 .33499 L
.26989 .36501 L
s
.4014 .18606 m
.42393 .20971 L
.39019 .21859 L
.4014 .18606 L
s
.41787 .25107 m
.37426 .25336 L
.39019 .21859 L
.41787 .25107 L
s
.46249 .04772 m
.42432 .04748 L
.44286 .01429 L
.46249 .04772 L
s
.44374 .08186 m
.42432 .04748 L
.46249 .04772 L
.44374 .08186 L
s
.25238 .39524 m
.23234 .36506 L
.26989 .36501 L
.25238 .39524 L
s
.45036 .15531 m
.41339 .15637 L
.42248 .11923 L
.45036 .15531 L
s
.33125 .30539 m
.30182 .30616 L
.31821 .27928 L
.33125 .30539 L
s
.44374 .08186 m
.46495 .11571 L
.42248 .11923 L
.44374 .08186 L
s
.42851 .18025 m
.4014 .18606 L
.41339 .15637 L
.42851 .18025 L
s
.33125 .30539 m
.31821 .27928 L
.35256 .2845 L
.33125 .30539 L
s
.41787 .25107 m
.39019 .21859 L
.42393 .20971 L
.41787 .25107 L
s
.45036 .15531 m
.42248 .11923 L
.46495 .11571 L
.45036 .15531 L
s
.35256 .2845 m
.37426 .25336 L
.3935 .29327 L
.35256 .2845 L
s
.42851 .18025 m
.42393 .20971 L
.4014 .18606 L
.42851 .18025 L
s
.32216 .3328 m
.28655 .33499 L
.30182 .30616 L
.32216 .3328 L
s
.3935 .29327 m
.37426 .25336 L
.41787 .25107 L
.3935 .29327 L
s
.42851 .18025 m
.41339 .15637 L
.45036 .15531 L
.42851 .18025 L
s
.30715 .3642 m
.26989 .36501 L
.28655 .33499 L
.30715 .3642 L
s
.33125 .30539 m
.32216 .3328 L
.30182 .30616 L
.33125 .30539 L
s
.46249 .04772 m
.44286 .01429 L
.48095 .01429 L
.46249 .04772 L
s
.26989 .36501 m
.29048 .39524 L
.25238 .39524 L
.26989 .36501 L
s
.44374 .08186 m
.46249 .04772 L
.48226 .08091 L
.44374 .08186 L
s
.48226 .08091 m
.46495 .11571 L
.44374 .08186 L
.48226 .08091 L
s
.30715 .3642 m
.28655 .33499 L
.32216 .3328 L
.30715 .3642 L
s
.42393 .20971 m
.45233 .22841 L
.41787 .25107 L
.42393 .20971 L
s
.36048 .32405 m
.33125 .30539 L
.35256 .2845 L
.36048 .32405 L
s
.30715 .3642 m
.29048 .39524 L
.26989 .36501 L
.30715 .3642 L
s
.35256 .2845 m
.3935 .29327 L
.36048 .32405 L
.35256 .2845 L
s
.42851 .18025 m
.45297 .19356 L
.42393 .20971 L
.42851 .18025 L
s
.42851 .18025 m
.45036 .15531 L
.45297 .19356 L
.42851 .18025 L
s
.48799 .14469 m
.45036 .15531 L
.46495 .11571 L
.48799 .14469 L
s
.3935 .29327 m
.41787 .25107 L
.43415 .29092 L
.3935 .29327 L
s
.36048 .32405 m
.32216 .3328 L
.33125 .30539 L
.36048 .32405 L
s
.42393 .20971 m
.45297 .19356 L
.45233 .22841 L
.42393 .20971 L
s
.46249 .04772 m
.48095 .01429 L
.50046 .04762 L
.46249 .04772 L
s
.48226 .08091 m
.46249 .04772 L
.50046 .04762 L
.48226 .08091 L
s
.41787 .25107 m
.45233 .22841 L
.45656 .26308 L
.41787 .25107 L
s
.34468 .36196 m
.30715 .3642 L
.32216 .3328 L
.34468 .36196 L
s
.43415 .29092 m
.41787 .25107 L
.45656 .26308 L
.43415 .29092 L
s
.48226 .08091 m
.50216 .11242 L
.46495 .11571 L
.48226 .08091 L
s
.45297 .19356 m
.45036 .15531 L
.48145 .17728 L
.45297 .19356 L
s
.48799 .14469 m
.48145 .17728 L
.45036 .15531 L
.48799 .14469 L
s
.34468 .36196 m
.32216 .3328 L
.36048 .32405 L
.34468 .36196 L
s
.48799 .14469 m
.46495 .11571 L
.50216 .11242 L
.48799 .14469 L
s
.32857 .39524 m
.29048 .39524 L
.30715 .3642 L
.32857 .39524 L
s
.36048 .32405 m
.3935 .29327 L
.39567 .3302 L
.36048 .32405 L
s
.41896 .31871 m
.3935 .29327 L
.43415 .29092 L
.41896 .31871 L
s
.32857 .39524 m
.30715 .3642 L
.34468 .36196 L
.32857 .39524 L
s
.50046 .04762 m
.48095 .01429 L
.51905 .01429 L
.50046 .04762 L
s
.45233 .22841 m
.45297 .19356 L
.48094 .21027 L
.45233 .22841 L
s
.41896 .31871 m
.39567 .3302 L
.3935 .29327 L
.41896 .31871 L
s
.48226 .08091 m
.50046 .04762 L
.51931 .08043 L
.48226 .08091 L
s
.4818 .24266 m
.45656 .26308 L
.45233 .22841 L
.4818 .24266 L
s
.51931 .08043 m
.50216 .11242 L
.48226 .08091 L
.51931 .08043 L
s
.38252 .36024 m
.34468 .36196 L
.36048 .32405 L
.38252 .36024 L
s
.45297 .19356 m
.48145 .17728 L
.48094 .21027 L
.45297 .19356 L
s
.43415 .29092 m
.45656 .26308 L
.46917 .29498 L
.43415 .29092 L
s
.38252 .36024 m
.36048 .32405 L
.39567 .3302 L
.38252 .36024 L
s
.45233 .22841 m
.48094 .21027 L
.4818 .24266 L
.45233 .22841 L
s
.48799 .14469 m
.50216 .11242 L
.52109 .14153 L
.48799 .14469 L
s
.48145 .17728 m
.48799 .14469 L
.50745 .16615 L
.48145 .17728 L
s
.44888 .32441 m
.41896 .31871 L
.43415 .29092 L
.44888 .32441 L
s
.34468 .36196 m
.36667 .39524 L
.32857 .39524 L
.34468 .36196 L
s
.51905 .01429 m
.53851 .04835 L
.50046 .04762 L
.51905 .01429 L
s
.50745 .16615 m
.48799 .14469 L
.52109 .14153 L
.50745 .16615 L
s
.51931 .08043 m
.50046 .04762 L
.53851 .04835 L
.51931 .08043 L
s
.43415 .29092 m
.46917 .29498 L
.44888 .32441 L
.43415 .29092 L
s
.38252 .36024 m
.36667 .39524 L
.34468 .36196 L
.38252 .36024 L
s
.45656 .26308 m
.4818 .24266 L
.48439 .27045 L
.45656 .26308 L
s
.45656 .26308 m
.48439 .27045 L
.46917 .29498 L
.45656 .26308 L
s
.53695 .11164 m
.50216 .11242 L
.51931 .08043 L
.53695 .11164 L
s
.41896 .31871 m
.42248 .35482 L
.39567 .3302 L
.41896 .31871 L
s
.48094 .21027 m
.48145 .17728 L
.50869 .19382 L
.48094 .21027 L
s
.53695 .11164 m
.52109 .14153 L
.50216 .11242 L
.53695 .11164 L
s
.42248 .35482 m
.38252 .36024 L
.39567 .3302 L
.42248 .35482 L
s
.48145 .17728 m
.50745 .16615 L
.50869 .19382 L
.48145 .17728 L
s
.50846 .22596 m
.4818 .24266 L
.48094 .21027 L
.50846 .22596 L
s
.41896 .31871 m
.44888 .32441 L
.42248 .35482 L
.41896 .31871 L
s
.51905 .01429 m
.55714 .01429 L
.53851 .04835 L
.51905 .01429 L
s
.53851 .04835 m
.55535 .08205 L
.51931 .08043 L
.53851 .04835 L
s
.48094 .21027 m
.50869 .19382 L
.50846 .22596 L
.48094 .21027 L
s
.40476 .39524 m
.36667 .39524 L
.38252 .36024 L
.40476 .39524 L
s
.53695 .11164 m
.51931 .08043 L
.55535 .08205 L
.53695 .11164 L
s
.46917 .29498 m
.4851 .32733 L
.44888 .32441 L
.46917 .29498 L
s
.50745 .16615 m
.52109 .14153 L
.53779 .1737 L
.50745 .16615 L
s
.40476 .39524 m
.38252 .36024 L
.42248 .35482 L
.40476 .39524 L
s
.50853 .25783 m
.48439 .27045 L
.4818 .24266 L
.50853 .25783 L
s
.48439 .27045 m
.5064 .2937 L
.46917 .29498 L
.48439 .27045 L
s
.50853 .25783 m
.4818 .24266 L
.50846 .22596 L
.50853 .25783 L
s
.53695 .11164 m
.55387 .14105 L
.52109 .14153 L
.53695 .11164 L
s
.50745 .16615 m
.53779 .1737 L
.50869 .19382 L
.50745 .16615 L
s
.46361 .35975 m
.42248 .35482 L
.44888 .32441 L
.46361 .35975 L
s
.46917 .29498 m
.5064 .2937 L
.4851 .32733 L
.46917 .29498 L
s
.52109 .14153 m
.55387 .14105 L
.53779 .1737 L
.52109 .14153 L
s
.53851 .04835 m
.55714 .01429 L
.57973 .05175 L
.53851 .04835 L
s
.46361 .35975 m
.44888 .32441 L
.4851 .32733 L
.46361 .35975 L
s
.53851 .04835 m
.57973 .05175 L
.55535 .08205 L
.53851 .04835 L
s
.50853 .25783 m
.5064 .2937 L
.48439 .27045 L
.50853 .25783 L
s
.56975 .11291 m
.53695 .11164 L
.55535 .08205 L
.56975 .11291 L
s
.50869 .19382 m
.53602 .20978 L
.50846 .22596 L
.50869 .19382 L
s
.42248 .35482 m
.44286 .39524 L
.40476 .39524 L
.42248 .35482 L
s
.56975 .11291 m
.55387 .14105 L
.53695 .11164 L
.56975 .11291 L
s
.50869 .19382 m
.53779 .1737 L
.53602 .20978 L
.50869 .19382 L
s
.46361 .35975 m
.44286 .39524 L
.42248 .35482 L
.46361 .35975 L
s
.50853 .25783 m
.50846 .22596 L
.53509 .24155 L
.50853 .25783 L
s
.57973 .05175 m
.55714 .01429 L
.59524 .01429 L
.57973 .05175 L
s
.53509 .24155 m
.50846 .22596 L
.53602 .20978 L
.53509 .24155 L
s
.5064 .2937 m
.52143 .32959 L
.4851 .32733 L
.5064 .2937 L
s
.57973 .05175 m
.58785 .08777 L
.55535 .08205 L
.57973 .05175 L
s
.56975 .11291 m
.55535 .08205 L
.58785 .08777 L
.56975 .11291 L
s
.53779 .1737 m
.55387 .14105 L
.57256 .16715 L
.53779 .1737 L
s
.50853 .25783 m
.53566 .27268 L
.5064 .2937 L
.50853 .25783 L
s
.50152 .36148 m
.46361 .35975 L
.4851 .32733 L
.50152 .36148 L
s
.53602 .20978 m
.53779 .1737 L
.56472 .19687 L
.53602 .20978 L
s
.50853 .25783 m
.53509 .24155 L
.53566 .27268 L
.50853 .25783 L
s
.50152 .36148 m
.4851 .32733 L
.52143 .32959 L
.50152 .36148 L
s
.5849 .1399 m
.55387 .14105 L
.56975 .11291 L
.5849 .1399 L
s
.57256 .16715 m
.56472 .19687 L
.53779 .1737 L
.57256 .16715 L
s
.46361 .35975 m
.48095 .39524 L
.44286 .39524 L
.46361 .35975 L
s
.5064 .2937 m
.5412 .30289 L
.52143 .32959 L
.5064 .2937 L
s
.55387 .14105 m
.5849 .1399 L
.57256 .16715 L
.55387 .14105 L
s
.59524 .01429 m
.61728 .04547 L
.57973 .05175 L
.59524 .01429 L
s
.53566 .27268 m
.5412 .30289 L
.5064 .2937 L
.53566 .27268 L
s
.46361 .35975 m
.50152 .36148 L
.48095 .39524 L
.46361 .35975 L
s
.53509 .24155 m
.53602 .20978 L
.56156 .22699 L
.53509 .24155 L
s
.57973 .05175 m
.60678 .07048 L
.58785 .08777 L
.57973 .05175 L
s
.56975 .11291 m
.58785 .08777 L
.59952 .11562 L
.56975 .11291 L
s
.53602 .20978 m
.56472 .19687 L
.56156 .22699 L
.53602 .20978 L
s
.5849 .1399 m
.56975 .11291 L
.59952 .11562 L
.5849 .1399 L
s
.56082 .25611 m
.53566 .27268 L
.53509 .24155 L
.56082 .25611 L
s
.57973 .05175 m
.61728 .04547 L
.60678 .07048 L
.57973 .05175 L
s
.53918 .36162 m
.50152 .36148 L
.52143 .32959 L
.53918 .36162 L
s
.56082 .25611 m
.53509 .24155 L
.56156 .22699 L
.56082 .25611 L
s
.59277 .18856 m
.56472 .19687 L
.57256 .16715 L
.59277 .18856 L
s
.59524 .01429 m
.63333 .01429 L
.61728 .04547 L
.59524 .01429 L
s
.57256 .16715 m
.5849 .1399 L
.60114 .16317 L
.57256 .16715 L
s
.5412 .30289 m
.55556 .33022 L
.52143 .32959 L
.5412 .30289 L
s
.50152 .36148 m
.51905 .39524 L
.48095 .39524 L
.50152 .36148 L
s
.56248 .28386 m
.5412 .30289 L
.53566 .27268 L
.56248 .28386 L
s
.53918 .36162 m
.52143 .32959 L
.55556 .33022 L
.53918 .36162 L
s
.61594 .09441 m
.58785 .08777 L
.60678 .07048 L
.61594 .09441 L
s
.59277 .18856 m
.57256 .16715 L
.60114 .16317 L
.59277 .18856 L
s
.59952 .11562 m
.58785 .08777 L
.61594 .09441 L
.59952 .11562 L
s
.56248 .28386 m
.53566 .27268 L
.56082 .25611 L
.56248 .28386 L
s
.53918 .36162 m
.51905 .39524 L
.50152 .36148 L
.53918 .36162 L
s
.56156 .22699 m
.56472 .19687 L
.58766 .21532 L
.56156 .22699 L
s
.61238 .14 m
.5849 .1399 L
.59952 .11562 L
.61238 .14 L
s
.59277 .18856 m
.58766 .21532 L
.56472 .19687 L
.59277 .18856 L
s
.60678 .07048 m
.61728 .04547 L
.63307 .07349 L
.60678 .07048 L
s
.5849 .1399 m
.61238 .14 L
.60114 .16317 L
.5849 .1399 L
s
.56082 .25611 m
.56156 .22699 L
.58536 .2424 L
.56082 .25611 L
s
.61728 .04547 m
.63333 .01429 L
.65547 .04888 L
.61728 .04547 L
s
.56609 .30692 m
.55556 .33022 L
.5412 .30289 L
.56609 .30692 L
s
.56609 .30692 m
.5412 .30289 L
.56248 .28386 L
.56609 .30692 L
s
.61594 .09441 m
.60678 .07048 L
.63307 .07349 L
.61594 .09441 L
s
.58536 .2424 m
.56156 .22699 L
.58766 .21532 L
.58536 .2424 L
s
.63307 .07349 m
.61728 .04547 L
.65547 .04888 L
.63307 .07349 L
s
.61796 .18276 m
.59277 .18856 L
.60114 .16317 L
.61796 .18276 L
s
.58049 .35774 m
.53918 .36162 L
.55556 .33022 L
.58049 .35774 L
s
.55714 .39524 m
.51905 .39524 L
.53918 .36162 L
.55714 .39524 L
s
.59952 .11562 m
.61594 .09441 L
.62611 .11936 L
.59952 .11562 L
s
.56248 .28386 m
.56082 .25611 L
.58522 .26912 L
.56248 .28386 L
s
.61238 .14 m
.59952 .11562 L
.62611 .11936 L
.61238 .14 L
s
.59277 .18856 m
.61246 .2063 L
.58766 .21532 L
.59277 .18856 L
s
.58536 .2424 m
.58522 .26912 L
.56082 .25611 L
.58536 .2424 L
s
.55714 .39524 m
.53918 .36162 L
.58049 .35774 L
.55714 .39524 L
s
.61238 .14 m
.62595 .16097 L
.60114 .16317 L
.61238 .14 L
s
.61796 .18276 m
.61246 .2063 L
.59277 .18856 L
.61796 .18276 L
s
.65547 .04888 m
.63333 .01429 L
.67143 .01429 L
.65547 .04888 L
s
.60114 .16317 m
.62595 .16097 L
.61796 .18276 L
.60114 .16317 L
s
.64218 .10035 m
.61594 .09441 L
.63307 .07349 L
.64218 .10035 L
s
.5851 .32212 m
.55556 .33022 L
.56609 .30692 L
.5851 .32212 L
s
.58049 .35774 m
.55556 .33022 L
.5851 .32212 L
.58049 .35774 L
s
.58536 .2424 m
.58766 .21532 L
.6092 .23091 L
.58536 .2424 L
s
.56609 .30692 m
.56248 .28386 L
.58626 .2951 L
.56609 .30692 L
s
.64218 .10035 m
.62611 .11936 L
.61594 .09441 L
.64218 .10035 L
s
.63307 .07349 m
.65547 .04888 L
.6606 .08192 L
.63307 .07349 L
s
.58626 .2951 m
.56248 .28386 L
.58522 .26912 L
.58626 .2951 L
s
.6092 .23091 m
.58766 .21532 L
.61246 .2063 L
.6092 .23091 L
s
.62611 .11936 m
.63624 .14147 L
.61238 .14 L
.62611 .11936 L
s
.62595 .16097 m
.61238 .14 L
.63624 .14147 L
.62595 .16097 L
s
.58536 .2424 m
.60802 .25592 L
.58522 .26912 L
.58536 .2424 L
s
.64218 .10035 m
.63307 .07349 L
.6606 .08192 L
.64218 .10035 L
s
.59524 .39524 m
.55714 .39524 L
.58049 .35774 L
.59524 .39524 L
s
.5851 .32212 m
.56609 .30692 L
.58626 .2951 L
.5851 .32212 L
s
.63511 .19904 m
.61246 .2063 L
.61796 .18276 L
.63511 .19904 L
s
.6092 .23091 m
.60802 .25592 L
.58536 .2424 L
.6092 .23091 L
s
.67143 .01429 m
.6921 .04472 L
.65547 .04888 L
.67143 .01429 L
s
.62595 .16097 m
.63998 .17869 L
.61796 .18276 L
.62595 .16097 L
s
.58049 .35774 m
.5851 .32212 L
.61201 .33519 L
.58049 .35774 L
s
.65547 .04888 m
.68019 .06839 L
.6606 .08192 L
.65547 .04888 L
s
.65 .12486 m
.62611 .11936 L
.64218 .10035 L
.65 .12486 L
s
.61796 .18276 m
.63998 .17869 L
.63511 .19904 L
.61796 .18276 L
s
.65547 .04888 m
.6921 .04472 L
.68019 .06839 L
.65547 .04888 L
s
.60862 .28112 m
.58626 .2951 L
.58522 .26912 L
.60862 .28112 L
s
.65 .12486 m
.63624 .14147 L
.62611 .11936 L
.65 .12486 L
s
.6092 .23091 m
.61246 .2063 L
.632 .22128 L
.6092 .23091 L
s
.60862 .28112 m
.58522 .26912 L
.60802 .25592 L
.60862 .28112 L
s
.619 .36419 m
.59524 .39524 L
.58049 .35774 L
.619 .36419 L
s
.62595 .16097 m
.63624 .14147 L
.64748 .16 L
.62595 .16097 L
s
.619 .36419 m
.58049 .35774 L
.61201 .33519 L
.619 .36419 L
s
.632 .22128 m
.61246 .2063 L
.63511 .19904 L
.632 .22128 L
s
.61002 .30708 m
.5851 .32212 L
.58626 .2951 L
.61002 .30708 L
s
.64218 .10035 m
.6606 .08192 L
.66689 .10871 L
.64218 .10035 L
s
.64748 .16 m
.63998 .17869 L
.62595 .16097 L
.64748 .16 L
s
.67143 .01429 m
.70952 .01429 L
.6921 .04472 L
.67143 .01429 L
s
.6092 .23091 m
.62982 .2443 L
.60802 .25592 L
.6092 .23091 L
s
.61002 .30708 m
.61201 .33519 L
.5851 .32212 L
.61002 .30708 L
s
.64218 .10035 m
.66689 .10871 L
.65 .12486 L
.64218 .10035 L
s
.60862 .28112 m
.61002 .30708 L
.58626 .2951 L
.60862 .28112 L
s
.6092 .23091 m
.632 .22128 L
.62982 .2443 L
.6092 .23091 L
s
.68019 .06839 m
.6853 .09227 L
.6606 .08192 L
.68019 .06839 L
s
.65272 .1926 m
.63511 .19904 L
.63998 .17869 L
.65272 .1926 L
s
.65 .12486 m
.65489 .14401 L
.63624 .14147 L
.65 .12486 L
s
.6606 .08192 m
.6853 .09227 L
.66689 .10871 L
.6606 .08192 L
s
.60802 .25592 m
.62999 .26763 L
.60862 .28112 L
.60802 .25592 L
s
.65489 .14401 m
.64748 .16 L
.63624 .14147 L
.65489 .14401 L
s
.68019 .06839 m
.6921 .04472 L
.70707 .07462 L
.68019 .06839 L
s
.632 .22128 m
.63511 .19904 L
.65521 .21264 L
.632 .22128 L
s
.63714 .38872 m
.59524 .39524 L
.619 .36419 L
.63714 .38872 L
s
.62982 .2443 m
.62999 .26763 L
.60802 .25592 L
.62982 .2443 L
s
.64748 .16 m
.66008 .17685 L
.63998 .17869 L
.64748 .16 L
s
.619 .36419 m
.61201 .33519 L
.6412 .34271 L
.619 .36419 L
s
.63511 .19904 m
.65272 .1926 L
.65521 .21264 L
.63511 .19904 L
s
.70952 .01429 m
.72859 .0482 L
.6921 .04472 L
.70952 .01429 L
s
.68019 .06839 m
.70707 .07462 L
.6853 .09227 L
.68019 .06839 L
s
.61201 .33519 m
.61002 .30708 L
.63565 .31687 L
.61201 .33519 L
s
.63998 .17869 m
.66008 .17685 L
.65272 .1926 L
.63998 .17869 L
s
.61002 .30708 m
.60862 .28112 L
.63208 .29168 L
.61002 .30708 L
s
.70707 .07462 m
.6921 .04472 L
.72859 .0482 L
.70707 .07462 L
s
.65 .12486 m
.66689 .10871 L
.67303 .13576 L
.65 .12486 L
s
.63714 .38872 m
.62985 .41997 L
.59524 .39524 L
.63714 .38872 L
s
.65 .12486 m
.67303 .13576 L
.65489 .14401 L
.65 .12486 L
s
.60862 .28112 m
.62999 .26763 L
.63208 .29168 L
.60862 .28112 L
s
.632 .22128 m
.65006 .23457 L
.62982 .2443 L
.632 .22128 L
s
.61201 .33519 m
.63565 .31687 L
.6412 .34271 L
.61201 .33519 L
s
.65489 .14401 m
.66704 .15823 L
.64748 .16 L
.65489 .14401 L
s
.632 .22128 m
.65521 .21264 L
.65006 .23457 L
.632 .22128 L
s
.6853 .09227 m
.68959 .11762 L
.66689 .10871 L
.6853 .09227 L
s
.59524 .43333 m
.59524 .39524 L
.62985 .41997 L
.59524 .43333 L
s
.61002 .30708 m
.63208 .29168 L
.63565 .31687 L
.61002 .30708 L
s
.66704 .15823 m
.66008 .17685 L
.64748 .16 L
.66704 .15823 L
s
.64729 .36512 m
.63714 .38872 L
.619 .36419 L
.64729 .36512 L
s
.64729 .36512 m
.619 .36419 L
.6412 .34271 L
.64729 .36512 L
s
.62982 .2443 m
.64959 .25552 L
.62999 .26763 L
.62982 .2443 L
s
.70707 .07462 m
.70685 .10289 L
.6853 .09227 L
.70707 .07462 L
s
.66689 .10871 m
.68959 .11762 L
.67303 .13576 L
.66689 .10871 L
s
.67283 .1958 m
.65521 .21264 L
.65272 .1926 L
.67283 .1958 L
s
.65489 .14401 m
.67303 .13576 L
.66704 .15823 L
.65489 .14401 L
s
.70952 .01429 m
.74762 .01429 L
.72859 .0482 L
.70952 .01429 L
s
.67283 .1958 m
.65272 .1926 L
.66008 .17685 L
.67283 .1958 L
s
.65006 .23457 m
.64959 .25552 L
.62982 .2443 L
.65006 .23457 L
s
.70707 .07462 m
.72859 .0482 L
.73036 .07367 L
.70707 .07462 L
s
.70685 .10289 m
.68959 .11762 L
.6853 .09227 L
.70685 .10289 L
s
.63208 .29168 m
.62999 .26763 L
.65159 .2772 L
.63208 .29168 L
s
.6412 .34271 m
.63565 .31687 L
.66167 .32418 L
.6412 .34271 L
s
.65521 .21264 m
.66635 .22864 L
.65006 .23457 L
.65521 .21264 L
s
.62999 .26763 m
.64959 .25552 L
.65159 .2772 L
.62999 .26763 L
s
.66306 .40801 m
.62985 .41997 L
.63714 .38872 L
.66306 .40801 L
s
.63565 .31687 m
.63208 .29168 L
.6561 .30003 L
.63565 .31687 L
s
.66704 .15823 m
.67983 .17582 L
.66008 .17685 L
.66704 .15823 L
s
.70707 .07462 m
.72545 .09278 L
.70685 .10289 L
.70707 .07462 L
s
.64729 .36512 m
.66764 .37931 L
.63714 .38872 L
.64729 .36512 L
s
.67983 .17582 m
.67283 .1958 L
.66008 .17685 L
.67983 .17582 L
s
.70707 .07462 m
.73036 .07367 L
.72545 .09278 L
.70707 .07462 L
s
.64729 .36512 m
.6412 .34271 L
.67073 .3502 L
.64729 .36512 L
s
.67283 .1958 m
.67708 .21655 L
.65521 .21264 L
.67283 .1958 L
s
.66704 .15823 m
.67303 .13576 L
.68627 .15709 L
.66704 .15823 L
s
.69576 .14018 m
.67303 .13576 L
.68959 .11762 L
.69576 .14018 L
s
.65159 .2772 m
.6561 .30003 L
.63208 .29168 L
.65159 .2772 L
s
.6561 .30003 m
.66167 .32418 L
.63565 .31687 L
.6561 .30003 L
s
.65521 .21264 m
.67708 .21655 L
.66635 .22864 L
.65521 .21264 L
s
.66306 .40801 m
.63714 .38872 L
.66764 .37931 L
.66306 .40801 L
s
.67073 .3502 m
.6412 .34271 L
.66167 .32418 L
.67073 .3502 L
s
.59524 .43333 m
.62985 .41997 L
.62775 .45471 L
.59524 .43333 L
s
.74762 .01429 m
.76651 .04916 L
.72859 .0482 L
.74762 .01429 L
s
.73036 .07367 m
.72859 .0482 L
.74648 .06705 L
.73036 .07367 L
s
.65006 .23457 m
.6672 .24514 L
.64959 .25552 L
.65006 .23457 L
s
.70685 .10289 m
.70905 .1263 L
.68959 .11762 L
.70685 .10289 L
s
.66704 .15823 m
.68627 .15709 L
.67983 .17582 L
.66704 .15823 L
s
.66635 .22864 m
.6672 .24514 L
.65006 .23457 L
.66635 .22864 L
s
.69576 .14018 m
.68627 .15709 L
.67303 .13576 L
.69576 .14018 L
s
.66306 .40801 m
.66009 .43947 L
.62985 .41997 L
.66306 .40801 L
s
.64729 .36512 m
.67073 .3502 L
.66764 .37931 L
.64729 .36512 L
s
.74648 .06705 m
.72859 .0482 L
.76651 .04916 L
.74648 .06705 L
s
.65159 .2772 m
.64959 .25552 L
.66861 .26438 L
.65159 .2772 L
s
.62775 .45471 m
.62985 .41997 L
.66009 .43947 L
.62775 .45471 L
s
.70905 .1263 m
.69576 .14018 L
.68959 .11762 L
.70905 .1263 L
s
.67983 .17582 m
.69405 .19141 L
.67283 .1958 L
.67983 .17582 L
s
.70685 .10289 m
.72545 .09278 L
.72407 .11422 L
.70685 .10289 L
s
.67261 .28398 m
.6561 .30003 L
.65159 .2772 L
.67261 .28398 L
s
.62775 .45471 m
.59524 .47143 L
.59524 .43333 L
.62775 .45471 L
s
.6561 .30003 m
.68217 .3061 L
.66167 .32418 L
.6561 .30003 L
s
.66861 .26438 m
.64959 .25552 L
.6672 .24514 L
.66861 .26438 L
s
.74375 .08613 m
.72545 .09278 L
.73036 .07367 L
.74375 .08613 L
s
.69048 .20476 m
.67708 .21655 L
.67283 .1958 L
.69048 .20476 L
s
.70685 .10289 m
.72407 .11422 L
.70905 .1263 L
.70685 .10289 L
s
.67073 .3502 m
.66167 .32418 L
.6832 .32883 L
.67073 .3502 L
s
.74375 .08613 m
.73036 .07367 L
.74648 .06705 L
.74375 .08613 L
s
.66306 .40801 m
.66764 .37931 L
.69215 .39533 L
.66306 .40801 L
s
.66635 .22864 m
.67708 .21655 L
.68202 .23433 L
.66635 .22864 L
s
.66861 .26438 m
.67261 .28398 L
.65159 .2772 L
.66861 .26438 L
s
.74762 .01429 m
.78571 .01429 L
.76651 .04916 L
.74762 .01429 L
s
.67283 .1958 m
.69405 .19141 L
.69048 .20476 L
.67283 .1958 L
s
.67983 .17582 m
.68627 .15709 L
.69831 .17554 L
.67983 .17582 L
s
.66635 .22864 m
.68202 .23433 L
.6672 .24514 L
.66635 .22864 L
s
.66764 .37931 m
.67073 .3502 L
.69449 .36777 L
.66764 .37931 L
s
.67261 .28398 m
.68217 .3061 L
.6561 .30003 L
.67261 .28398 L
s
.6832 .32883 m
.66167 .32418 L
.68217 .3061 L
.6832 .32883 L
s
.69046 .42468 m
.66009 .43947 L
.66306 .40801 L
.69046 .42468 L
s
.70324 .15714 m
.68627 .15709 L
.69576 .14018 L
.70324 .15714 L
s
.67983 .17582 m
.69831 .17554 L
.69405 .19141 L
.67983 .17582 L
s
.72545 .09278 m
.74001 .10559 L
.72407 .11422 L
.72545 .09278 L
s
.70905 .1263 m
.71301 .14737 L
.69576 .14018 L
.70905 .1263 L
s
.69046 .42468 m
.66306 .40801 L
.69215 .39533 L
.69046 .42468 L
s
.74648 .06705 m
.76651 .04916 L
.76278 .07845 L
.74648 .06705 L
s
.74375 .08613 m
.74001 .10559 L
.72545 .09278 L
.74375 .08613 L
s
.69215 .39533 m
.66764 .37931 L
.69449 .36777 L
.69215 .39533 L
s
.62775 .45471 m
.66009 .43947 L
.65874 .47311 L
.62775 .45471 L
s
.68628 .2535 m
.66861 .26438 L
.6672 .24514 L
.68628 .2535 L
s
.70324 .15714 m
.69831 .17554 L
.68627 .15709 L
.70324 .15714 L
s
.69048 .20476 m
.69541 .22316 L
.67708 .21655 L
.69048 .20476 L
s
.6268 .49081 m
.59524 .47143 L
.62775 .45471 L
.6268 .49081 L
s
.74648 .06705 m
.76278 .07845 L
.74375 .08613 L
.74648 .06705 L
s
.69848 .34285 m
.67073 .3502 L
.6832 .32883 L
.69848 .34285 L
s
.68202 .23433 m
.67708 .21655 L
.69541 .22316 L
.68202 .23433 L
s
.70905 .1263 m
.72407 .11422 L
.72463 .13575 L
.70905 .1263 L
s
.69576 .14018 m
.71301 .14737 L
.70324 .15714 L
.69576 .14018 L
s
.69848 .34285 m
.69449 .36777 L
.67073 .3502 L
.69848 .34285 L
s
.68202 .23433 m
.68628 .2535 L
.6672 .24514 L
.68202 .23433 L
s
.68391 .27104 m
.67261 .28398 L
.66861 .26438 L
.68391 .27104 L
s
.78571 .01429 m
.80737 .0482 L
.76651 .04916 L
.78571 .01429 L
s
.6268 .49081 m
.62775 .45471 L
.65874 .47311 L
.6268 .49081 L
s
.67261 .28398 m
.69257 .28515 L
.68217 .3061 L
.67261 .28398 L
s
.70905 .1263 m
.72463 .13575 L
.71301 .14737 L
.70905 .1263 L
s
.68963 .45599 m
.66009 .43947 L
.69046 .42468 L
.68963 .45599 L
s
.6832 .32883 m
.68217 .3061 L
.70219 .32048 L
.6832 .32883 L
s
.76278 .07845 m
.76651 .04916 L
.78676 .07392 L
.76278 .07845 L
s
.68628 .2535 m
.68391 .27104 L
.66861 .26438 L
.68628 .2535 L
s
.68963 .45599 m
.65874 .47311 L
.66009 .43947 L
.68963 .45599 L
s
.74001 .10559 m
.7381 .12229 L
.72407 .11422 L
.74001 .10559 L
s
.78676 .07392 m
.76651 .04916 L
.80737 .0482 L
.78676 .07392 L
s
.75691 .0997 m
.74001 .10559 L
.74375 .08613 L
.75691 .0997 L
s
.71808 .38192 m
.69215 .39533 L
.69449 .36777 L
.71808 .38192 L
s
.68391 .27104 m
.69257 .28515 L
.67261 .28398 L
.68391 .27104 L
s
.76278 .07845 m
.75691 .0997 L
.74375 .08613 L
.76278 .07845 L
s
.72407 .11422 m
.7381 .12229 L
.72463 .13575 L
.72407 .11422 L
s
.69215 .39533 m
.71801 .40959 L
.69046 .42468 L
.69215 .39533 L
s
.69848 .34285 m
.6832 .32883 L
.70219 .32048 L
.69848 .34285 L
s
.6268 .49081 m
.59524 .50952 L
.59524 .47143 L
.6268 .49081 L
s
.68202 .23433 m
.69966 .23903 L
.68628 .2535 L
.68202 .23433 L
s
.68202 .23433 m
.69541 .22316 L
.69966 .23903 L
.68202 .23433 L
s
.68217 .3061 m
.69257 .28515 L
.70588 .29984 L
.68217 .3061 L
s
.70219 .32048 m
.68217 .3061 L
.70588 .29984 L
.70219 .32048 L
s
.69848 .34285 m
.71896 .35633 L
.69449 .36777 L
.69848 .34285 L
s
.78571 .01429 m
.82381 .01429 L
.80737 .0482 L
.78571 .01429 L
s
.71808 .38192 m
.71801 .40959 L
.69215 .39533 L
.71808 .38192 L
s
.65789 .50865 m
.6268 .49081 L
.65874 .47311 L
.65789 .50865 L
s
.68963 .45599 m
.69046 .42468 L
.71849 .43912 L
.68963 .45599 L
s
.71896 .35633 m
.71808 .38192 L
.69449 .36777 L
.71896 .35633 L
s
.68628 .2535 m
.69904 .26826 L
.68391 .27104 L
.68628 .2535 L
s
.69046 .42468 m
.71801 .40959 L
.71849 .43912 L
.69046 .42468 L
s
.74001 .10559 m
.75144 .11871 L
.7381 .12229 L
.74001 .10559 L
s
.75691 .0997 m
.75144 .11871 L
.74001 .10559 L
.75691 .0997 L
s
.68936 .48892 m
.65874 .47311 L
.68963 .45599 L
.68936 .48892 L
s
.77591 .09511 m
.76278 .07845 L
.78676 .07392 L
.77591 .09511 L
s
.68391 .27104 m
.69904 .26826 L
.69257 .28515 L
.68391 .27104 L
s
.76278 .07845 m
.77591 .09511 L
.75691 .0997 L
.76278 .07845 L
s
.69848 .34285 m
.70219 .32048 L
.72048 .33302 L
.69848 .34285 L
s
.70324 .25238 m
.68628 .2535 L
.69966 .23903 L
.70324 .25238 L
s
.68936 .48892 m
.65789 .50865 L
.65874 .47311 L
.68936 .48892 L
s
.81203 .07801 m
.78676 .07392 L
.80737 .0482 L
.81203 .07801 L
s
.62552 .52746 m
.59524 .50952 L
.6268 .49081 L
.62552 .52746 L
s
.72048 .33302 m
.71896 .35633 L
.69848 .34285 L
.72048 .33302 L
s
.69904 .26826 m
.68628 .2535 L
.70324 .25238 L
.69904 .26826 L
s
.65789 .50865 m
.62552 .52746 L
.6268 .49081 L
.65789 .50865 L
s
.7112 .28168 m
.70588 .29984 L
.69257 .28515 L
.7112 .28168 L
s
.80737 .0482 m
.82381 .01429 L
.84455 .04265 L
.80737 .0482 L
s
.72221 .312 m
.70219 .32048 L
.70588 .29984 L
.72221 .312 L
s
.69904 .26826 m
.7112 .28168 L
.69257 .28515 L
.69904 .26826 L
s
.71808 .38192 m
.74274 .39395 L
.71801 .40959 L
.71808 .38192 L
s
.68963 .45599 m
.71849 .43912 L
.71924 .47042 L
.68963 .45599 L
s
.71924 .47042 m
.68936 .48892 L
.68963 .45599 L
.71924 .47042 L
s
.71896 .35633 m
.7412 .36795 L
.71808 .38192 L
.71896 .35633 L
s
.72048 .33302 m
.70219 .32048 L
.72221 .312 L
.72048 .33302 L
s
.80737 .0482 m
.83206 .06535 L
.81203 .07801 L
.80737 .0482 L
s
.78676 .07392 m
.79578 .09483 L
.77591 .09511 L
.78676 .07392 L
s
.75691 .0997 m
.76732 .11445 L
.75144 .11871 L
.75691 .0997 L
s
.80737 .0482 m
.84455 .04265 L
.83206 .06535 L
.80737 .0482 L
s
.71849 .43912 m
.71801 .40959 L
.74486 .42192 L
.71849 .43912 L
s
.75691 .0997 m
.77591 .09511 L
.76732 .11445 L
.75691 .0997 L
s
.81203 .07801 m
.79578 .09483 L
.78676 .07392 L
.81203 .07801 L
s
.7412 .36795 m
.74274 .39395 L
.71808 .38192 L
.7412 .36795 L
s
.59524 .54762 m
.59524 .50952 L
.62552 .52746 L
.59524 .54762 L
s
.68976 .52259 m
.65789 .50865 L
.68936 .48892 L
.68976 .52259 L
s
.74486 .42192 m
.71801 .40959 L
.74274 .39395 L
.74486 .42192 L
s
.72048 .33302 m
.7404 .34398 L
.71896 .35633 L
.72048 .33302 L
s
.7112 .28168 m
.72431 .29352 L
.70588 .29984 L
.7112 .28168 L
s
.72431 .29352 m
.72221 .312 L
.70588 .29984 L
.72431 .29352 L
s
.69904 .26826 m
.70324 .25238 L
.71671 .26585 L
.69904 .26826 L
s
.65789 .50865 m
.65653 .54914 L
.62552 .52746 L
.65789 .50865 L
s
.71896 .35633 m
.7404 .34398 L
.7412 .36795 L
.71896 .35633 L
s
.84455 .04265 m
.82381 .01429 L
.8619 .01429 L
.84455 .04265 L
s
.69904 .26826 m
.71671 .26585 L
.7112 .28168 L
.69904 .26826 L
s
.71924 .47042 m
.71849 .43912 L
.74722 .45179 L
.71924 .47042 L
s
.71977 .50293 m
.68936 .48892 L
.71924 .47042 L
.71977 .50293 L
s
.68976 .52259 m
.65653 .54914 L
.65789 .50865 L
.68976 .52259 L
s
.74722 .45179 m
.71849 .43912 L
.74486 .42192 L
.74722 .45179 L
s
.72048 .33302 m
.72221 .312 L
.7401 .32216 L
.72048 .33302 L
s
.79578 .09483 m
.78571 .10952 L
.77591 .09511 L
.79578 .09483 L
s
.77591 .09511 m
.78571 .10952 L
.76732 .11445 L
.77591 .09511 L
s
.71977 .50293 m
.68976 .52259 L
.68936 .48892 L
.71977 .50293 L
s
.83206 .06535 m
.83499 .08659 L
.81203 .07801 L
.83206 .06535 L
s
.79578 .09483 m
.81203 .07801 L
.81513 .09944 L
.79578 .09483 L
s
.7401 .32216 m
.7404 .34398 L
.72048 .33302 L
.7401 .32216 L
s
.83206 .06535 m
.84455 .04265 L
.86129 .07152 L
.83206 .06535 L
s
.7412 .36795 m
.76478 .37774 L
.74274 .39395 L
.7412 .36795 L
s
.62552 .52746 m
.62119 .55912 L
.59524 .54762 L
.62552 .52746 L
s
.74486 .42192 m
.74274 .39395 L
.76857 .40416 L
.74486 .42192 L
s
.65653 .54914 m
.62119 .55912 L
.62552 .52746 L
.65653 .54914 L
s
.72833 .27747 m
.72431 .29352 L
.7112 .28168 L
.72833 .27747 L
s
.71671 .26585 m
.72833 .27747 L
.7112 .28168 L
.71671 .26585 L
s
.83499 .08659 m
.81513 .09944 L
.81203 .07801 L
.83499 .08659 L
s
.72221 .312 m
.72431 .29352 L
.73986 .30288 L
.72221 .312 L
s
.84455 .04265 m
.8619 .01429 L
.88079 .04246 L
.84455 .04265 L
s
.76169 .3534 m
.7412 .36795 L
.7404 .34398 L
.76169 .3534 L
s
.7401 .32216 m
.72221 .312 L
.73986 .30288 L
.7401 .32216 L
s
.83206 .06535 m
.86129 .07152 L
.83499 .08659 L
.83206 .06535 L
s
.71924 .47042 m
.74722 .45179 L
.74955 .48382 L
.71924 .47042 L
s
.86129 .07152 m
.84455 .04265 L
.88079 .04246 L
.86129 .07152 L
s
.74274 .39395 m
.76478 .37774 L
.76857 .40416 L
.74274 .39395 L
s
.74955 .48382 m
.71977 .50293 L
.71924 .47042 L
.74955 .48382 L
s
.68976 .52259 m
.69628 .55494 L
.65653 .54914 L
.68976 .52259 L
s
.76169 .3534 m
.76478 .37774 L
.7412 .36795 L
.76169 .3534 L
s
.74722 .45179 m
.74486 .42192 L
.77272 .43259 L
.74722 .45179 L
s
.79578 .09483 m
.79906 .1131 L
.78571 .10952 L
.79578 .09483 L
s
.81513 .09944 m
.79906 .1131 L
.79578 .09483 L
.81513 .09944 L
s
.71977 .50293 m
.71801 .53129 L
.68976 .52259 L
.71977 .50293 L
s
.7401 .32216 m
.75964 .33083 L
.7404 .34398 L
.7401 .32216 L
s
.76857 .40416 m
.77272 .43259 L
.74486 .42192 L
.76857 .40416 L
s
.62119 .55912 m
.59524 .58571 L
.59524 .54762 L
.62119 .55912 L
s
.72833 .27747 m
.7381 .28724 L
.72431 .29352 L
.72833 .27747 L
s
.7404 .34398 m
.75964 .33083 L
.76169 .3534 L
.7404 .34398 L
s
.68976 .52259 m
.71801 .53129 L
.69628 .55494 L
.68976 .52259 L
s
.72431 .29352 m
.7381 .28724 L
.73986 .30288 L
.72431 .29352 L
s
.65653 .54914 m
.63333 .58571 L
.62119 .55912 L
.65653 .54914 L
s
.83499 .08659 m
.83392 .1069 L
.81513 .09944 L
.83499 .08659 L
s
.83499 .08659 m
.86129 .07152 L
.85538 .09861 L
.83499 .08659 L
s
.88079 .04246 m
.8619 .01429 L
.9 .01429 L
.88079 .04246 L
s
.69628 .55494 m
.67143 .58571 L
.65653 .54914 L
.69628 .55494 L
s
.7401 .32216 m
.73986 .30288 L
.75777 .31015 L
.7401 .32216 L
s
.74955 .48382 m
.74722 .45179 L
.77692 .46289 L
.74955 .48382 L
s
.77692 .46289 m
.74722 .45179 L
.77272 .43259 L
.77692 .46289 L
s
.74955 .48382 m
.75165 .52095 L
.71977 .50293 L
.74955 .48382 L
s
.65653 .54914 m
.67143 .58571 L
.63333 .58571 L
.65653 .54914 L
s
.76857 .40416 m
.76478 .37774 L
.7897 .3858 L
.76857 .40416 L
s
.7401 .32216 m
.75777 .31015 L
.75964 .33083 L
.7401 .32216 L
s
.71977 .50293 m
.75165 .52095 L
.71801 .53129 L
.71977 .50293 L
s
.86129 .07152 m
.88079 .04246 L
.89902 .07024 L
.86129 .07152 L
s
.62119 .55912 m
.63333 .58571 L
.59524 .58571 L
.62119 .55912 L
s
.83392 .1069 m
.83499 .08659 L
.85538 .09861 L
.83392 .1069 L
s
.76478 .37774 m
.76169 .3534 L
.78423 .36106 L
.76478 .37774 L
s
.81513 .09944 m
.81493 .11735 L
.79906 .1131 L
.81513 .09944 L
s
.76857 .40416 m
.79562 .4128 L
.77272 .43259 L
.76857 .40416 L
s
.81513 .09944 m
.83392 .1069 L
.81493 .11735 L
.81513 .09944 L
s
.86129 .07152 m
.88196 .09513 L
.85538 .09861 L
.86129 .07152 L
s
.73986 .30288 m
.7381 .28724 L
.75649 .29217 L
.73986 .30288 L
s
.7897 .3858 m
.76478 .37774 L
.78423 .36106 L
.7897 .3858 L
s
.76169 .3534 m
.75964 .33083 L
.77967 .33883 L
.76169 .3534 L
s
.73209 .55547 m
.69628 .55494 L
.71801 .53129 L
.73209 .55547 L
s
.86129 .07152 m
.89902 .07024 L
.88196 .09513 L
.86129 .07152 L
s
.79562 .4128 m
.76857 .40416 L
.7897 .3858 L
.79562 .4128 L
s
.75777 .31015 m
.73986 .30288 L
.75649 .29217 L
.75777 .31015 L
s
.88079 .04246 m
.9 .01429 L
.91783 .042 L
.88079 .04246 L
s
.70952 .58571 m
.67143 .58571 L
.69628 .55494 L
.70952 .58571 L
s
.89902 .07024 m
.88079 .04246 L
.91783 .042 L
.89902 .07024 L
s
.77967 .33883 m
.78423 .36106 L
.76169 .3534 L
.77967 .33883 L
s
.77692 .46289 m
.78123 .49434 L
.74955 .48382 L
.77692 .46289 L
s
.78123 .49434 m
.75165 .52095 L
.74955 .48382 L
.78123 .49434 L
s
.73209 .55547 m
.71801 .53129 L
.75165 .52095 L
.73209 .55547 L
s
.77692 .46289 m
.77272 .43259 L
.80189 .44196 L
.77692 .46289 L
s
.75777 .31015 m
.77813 .31648 L
.75964 .33083 L
.75777 .31015 L
s
.85538 .09861 m
.84966 .11735 L
.83392 .1069 L
.85538 .09861 L
s
.77272 .43259 m
.79562 .4128 L
.80189 .44196 L
.77272 .43259 L
s
.70952 .58571 m
.69628 .55494 L
.73209 .55547 L
.70952 .58571 L
s
.77967 .33883 m
.75964 .33083 L
.77813 .31648 L
.77967 .33883 L
s
.83392 .1069 m
.83333 .12229 L
.81493 .11735 L
.83392 .1069 L
s
.88196 .09513 m
.87019 .11455 L
.85538 .09861 L
.88196 .09513 L
s
.7897 .3858 m
.78423 .36106 L
.80859 .36669 L
.7897 .3858 L
s
.75777 .31015 m
.75649 .29217 L
.77237 .29642 L
.75777 .31015 L
s
.84966 .11735 m
.83333 .12229 L
.83392 .1069 L
.84966 .11735 L
s
.91783 .042 m
.9 .01429 L
.9381 .01429 L
.91783 .042 L
s
.81585 .39241 m
.79562 .4128 L
.7897 .3858 L
.81585 .39241 L
s
.85538 .09861 m
.87019 .11455 L
.84966 .11735 L
.85538 .09861 L
s
.89902 .07024 m
.91254 .09874 L
.88196 .09513 L
.89902 .07024 L
s
.75777 .31015 m
.77237 .29642 L
.77813 .31648 L
.75777 .31015 L
s
.75165 .52095 m
.76916 .55456 L
.73209 .55547 L
.75165 .52095 L
s
.77967 .33883 m
.80074 .34402 L
.78423 .36106 L
.77967 .33883 L
s
.77692 .46289 m
.80771 .47308 L
.78123 .49434 L
.77692 .46289 L
s
.80189 .44196 m
.80771 .47308 L
.77692 .46289 L
.80189 .44196 L
s
.89902 .07024 m
.91783 .042 L
.93938 .0719 L
.89902 .07024 L
s
.78123 .49434 m
.79092 .52491 L
.75165 .52095 L
.78123 .49434 L
s
.7897 .3858 m
.80859 .36669 L
.81585 .39241 L
.7897 .3858 L
s
.80074 .34402 m
.80859 .36669 L
.78423 .36106 L
.80074 .34402 L
s
.79562 .4128 m
.82412 .41988 L
.80189 .44196 L
.79562 .4128 L
s
.74762 .58571 m
.70952 .58571 L
.73209 .55547 L
.74762 .58571 L
s
.89902 .07024 m
.93938 .0719 L
.91254 .09874 L
.89902 .07024 L
s
.79092 .52491 m
.76916 .55456 L
.75165 .52095 L
.79092 .52491 L
s
.77813 .31648 m
.79359 .32736 L
.77967 .33883 L
.77813 .31648 L
s
.82412 .41988 m
.79562 .4128 L
.81585 .39241 L
.82412 .41988 L
s
.89287 .11753 m
.87019 .11455 L
.88196 .09513 L
.89287 .11753 L
s
.89287 .11753 m
.88196 .09513 L
.91254 .09874 L
.89287 .11753 L
s
.76916 .55456 m
.74762 .58571 L
.73209 .55547 L
.76916 .55456 L
s
.91783 .042 m
.9381 .01429 L
.94955 .03896 L
.91783 .042 L
s
.79359 .32736 m
.80074 .34402 L
.77967 .33883 L
.79359 .32736 L
s
.84966 .11735 m
.8431 .13206 L
.83333 .12229 L
.84966 .11735 L
s
.86182 .12994 m
.84966 .11735 L
.87019 .11455 L
.86182 .12994 L
s
.91783 .042 m
.94955 .03896 L
.93938 .0719 L
.91783 .042 L
s
.77813 .31648 m
.77237 .29642 L
.78571 .3 L
.77813 .31648 L
s
.80983 .50088 m
.78123 .49434 L
.80771 .47308 L
.80983 .50088 L
s
.80189 .44196 m
.83325 .45055 L
.80771 .47308 L
.80189 .44196 L
s
.80983 .50088 m
.79092 .52491 L
.78123 .49434 L
.80983 .50088 L
s
.83369 .37223 m
.81585 .39241 L
.80859 .36669 L
.83369 .37223 L
s
.82412 .41988 m
.83325 .45055 L
.80189 .44196 L
.82412 .41988 L
s
.80074 .34402 m
.82704 .34477 L
.80859 .36669 L
.80074 .34402 L
s
.86182 .12994 m
.8431 .13206 L
.84966 .11735 L
.86182 .12994 L
s
.77813 .31648 m
.79821 .31194 L
.79359 .32736 L
.77813 .31648 L
s
.79821 .31194 m
.77813 .31648 L
.78571 .3 L
.79821 .31194 L
s
.91254 .09874 m
.93938 .0719 L
.94405 .1107 L
.91254 .09874 L
s
.87896 .12912 m
.87019 .11455 L
.89287 .11753 L
.87896 .12912 L
s
.84281 .39707 m
.82412 .41988 L
.81585 .39241 L
.84281 .39707 L
s
.87896 .12912 m
.86182 .12994 L
.87019 .11455 L
.87896 .12912 L
s
.79092 .52491 m
.80676 .55494 L
.76916 .55456 L
.79092 .52491 L
s
.82704 .34477 m
.83369 .37223 L
.80859 .36669 L
.82704 .34477 L
s
.94955 .03896 m
.9381 .01429 L
.97619 .01429 L
.94955 .03896 L
s
.78571 .58571 m
.74762 .58571 L
.76916 .55456 L
.78571 .58571 L
s
.91254 .09874 m
.91739 .12722 L
.89287 .11753 L
.91254 .09874 L
s
.84281 .39707 m
.81585 .39241 L
.83369 .37223 L
.84281 .39707 L
s
.79359 .32736 m
.81129 .32572 L
.80074 .34402 L
.79359 .32736 L
s
.93938 .0719 m
.94955 .03896 L
.97619 .05238 L
.93938 .0719 L
s
.91254 .09874 m
.94405 .1107 L
.91739 .12722 L
.91254 .09874 L
s
.80074 .34402 m
.81129 .32572 L
.82704 .34477 L
.80074 .34402 L
s
.84243 .4881 m
.80983 .50088 L
.80771 .47308 L
.84243 .4881 L
s
.84243 .4881 m
.80771 .47308 L
.83325 .45055 L
.84243 .4881 L
s
.78571 .58571 m
.76916 .55456 L
.80676 .55494 L
.78571 .58571 L
s
.80983 .50088 m
.8265 .52432 L
.79092 .52491 L
.80983 .50088 L
s
.85465 .42434 m
.83325 .45055 L
.82412 .41988 L
.85465 .42434 L
s
.79359 .32736 m
.79821 .31194 L
.81129 .32572 L
.79359 .32736 L
s
.86182 .12994 m
.85472 .14367 L
.8431 .13206 L
.86182 .12994 L
s
.93938 .0719 m
.97619 .09048 L
.94405 .1107 L
.93938 .0719 L
s
.79092 .52491 m
.8265 .52432 L
.80676 .55494 L
.79092 .52491 L
s
.85465 .42434 m
.82412 .41988 L
.84281 .39707 L
.85465 .42434 L
s
.94955 .03896 m
.97619 .01429 L
.97619 .05238 L
.94955 .03896 L
s
.87896 .12912 m
.89287 .11753 L
.89546 .14068 L
.87896 .12912 L
s
.97619 .05238 m
.97619 .09048 L
.93938 .0719 L
.97619 .05238 L
s
.79821 .31194 m
.78571 .3 L
.80411 .29507 L
.79821 .31194 L
s
.87896 .12912 m
.87393 .14309 L
.86182 .12994 L
.87896 .12912 L
s
.89287 .11753 m
.91739 .12722 L
.89546 .14068 L
.89287 .11753 L
s
.82704 .34477 m
.85224 .35517 L
.83369 .37223 L
.82704 .34477 L
s
.84243 .4881 m
.8265 .52432 L
.80983 .50088 L
.84243 .4881 L
s
.83369 .37223 m
.85523 .3773 L
.84281 .39707 L
.83369 .37223 L
s
.86182 .12994 m
.87393 .14309 L
.85472 .14367 L
.86182 .12994 L
s
.82704 .34477 m
.81129 .32572 L
.83167 .31932 L
.82704 .34477 L
s
.84243 .4881 m
.83325 .45055 L
.86819 .45577 L
.84243 .4881 L
s
.81129 .32572 m
.79821 .31194 L
.81646 .30716 L
.81129 .32572 L
s
.85465 .42434 m
.86819 .45577 L
.83325 .45055 L
.85465 .42434 L
s
.91739 .12722 m
.94405 .1107 L
.94219 .14739 L
.91739 .12722 L
s
.80676 .55494 m
.82381 .58571 L
.78571 .58571 L
.80676 .55494 L
s
.87896 .12912 m
.89546 .14068 L
.87393 .14309 L
.87896 .12912 L
s
.83369 .37223 m
.85224 .35517 L
.85523 .3773 L
.83369 .37223 L
s
.79821 .31194 m
.80411 .29507 L
.81646 .30716 L
.79821 .31194 L
s
.94405 .1107 m
.97619 .09048 L
.97619 .12857 L
.94405 .1107 L
s
.84437 .55473 m
.80676 .55494 L
.8265 .52432 L
.84437 .55473 L
s
.87285 .396 m
.85465 .42434 L
.84281 .39707 L
.87285 .396 L
s
.85224 .35517 m
.82704 .34477 L
.85079 .33055 L
.85224 .35517 L
s
.82704 .34477 m
.83167 .31932 L
.85079 .33055 L
.82704 .34477 L
s
.94405 .1107 m
.97619 .12857 L
.94219 .14739 L
.94405 .1107 L
s
.87285 .396 m
.84281 .39707 L
.85523 .3773 L
.87285 .396 L
s
.81646 .30716 m
.83167 .31932 L
.81129 .32572 L
.81646 .30716 L
s
.84243 .4881 m
.86462 .52269 L
.8265 .52432 L
.84243 .4881 L
s
.91739 .12722 m
.91652 .14833 L
.89546 .14068 L
.91739 .12722 L
s
.84437 .55473 m
.82381 .58571 L
.80676 .55494 L
.84437 .55473 L
s
.91739 .12722 m
.94219 .14739 L
.91652 .14833 L
.91739 .12722 L
s
.87393 .14309 m
.86819 .15714 L
.85472 .14367 L
.87393 .14309 L
s
.88665 .48944 m
.84243 .4881 L
.86819 .45577 L
.88665 .48944 L
s
.89546 .14068 m
.88399 .15806 L
.87393 .14309 L
.89546 .14068 L
s
.86462 .52269 m
.84437 .55473 L
.8265 .52432 L
.86462 .52269 L
s
.88629 .4263 m
.86819 .45577 L
.85465 .42434 L
.88629 .4263 L
s
.88665 .48944 m
.86462 .52269 L
.84243 .4881 L
.88665 .48944 L
s
.85523 .3773 m
.85224 .35517 L
.87418 .36587 L
.85523 .3773 L
s
.87285 .396 m
.88629 .4263 L
.85465 .42434 L
.87285 .396 L
s
.80411 .29507 m
.81999 .29082 L
.81646 .30716 L
.80411 .29507 L
s
.81646 .30716 m
.8335 .30025 L
.83167 .31932 L
.81646 .30716 L
s
.88399 .15806 m
.86819 .15714 L
.87393 .14309 L
.88399 .15806 L
s
.87418 .36587 m
.87285 .396 L
.85523 .3773 L
.87418 .36587 L
s
.90369 .16146 m
.89546 .14068 L
.91652 .14833 L
.90369 .16146 L
s
.87205 .34037 m
.85224 .35517 L
.85079 .33055 L
.87205 .34037 L
s
.90369 .16146 m
.88399 .15806 L
.89546 .14068 L
.90369 .16146 L
s
.85072 .30771 m
.85079 .33055 L
.83167 .31932 L
.85072 .30771 L
s
.94219 .14739 m
.97619 .12857 L
.97619 .16667 L
.94219 .14739 L
s
.8619 .58571 m
.82381 .58571 L
.84437 .55473 L
.8619 .58571 L
s
.87205 .34037 m
.87418 .36587 L
.85224 .35517 L
.87205 .34037 L
s
.81999 .29082 m
.8335 .30025 L
.81646 .30716 L
.81999 .29082 L
s
.88665 .48944 m
.86819 .45577 L
.90576 .45536 L
.88665 .48944 L
s
.8335 .30025 m
.85072 .30771 L
.83167 .31932 L
.8335 .30025 L
s
.91652 .14833 m
.94219 .14739 L
.92319 .1648 L
.91652 .14833 L
s
.88629 .4263 m
.90576 .45536 L
.86819 .45577 L
.88629 .4263 L
s
.86462 .52269 m
.88267 .55512 L
.84437 .55473 L
.86462 .52269 L
s
.88665 .48944 m
.90521 .52443 L
.86462 .52269 L
.88665 .48944 L
s
.89906 .40294 m
.88629 .4263 L
.87285 .396 L
.89906 .40294 L
s
.8973 .37748 m
.87285 .396 L
.87418 .36587 L
.8973 .37748 L
s
.90369 .16146 m
.91652 .14833 L
.92319 .1648 L
.90369 .16146 L
s
.88399 .15806 m
.87177 .17049 L
.86819 .15714 L
.88399 .15806 L
s
.8619 .58571 m
.84437 .55473 L
.88267 .55512 L
.8619 .58571 L
s
.94219 .14739 m
.97619 .16667 L
.94117 .18532 L
.94219 .14739 L
s
.92319 .1648 m
.94219 .14739 L
.94117 .18532 L
.92319 .1648 L
s
.87205 .34037 m
.85079 .33055 L
.8706 .31618 L
.87205 .34037 L
s
.85072 .30771 m
.8706 .31618 L
.85079 .33055 L
.85072 .30771 L
s
.88267 .55512 m
.86462 .52269 L
.90521 .52443 L
.88267 .55512 L
s
.8973 .37748 m
.89906 .40294 L
.87285 .396 L
.8973 .37748 L
s
.89048 .17543 m
.88399 .15806 L
.90369 .16146 L
.89048 .17543 L
s
.81999 .29082 m
.83333 .28724 L
.8335 .30025 L
.81999 .29082 L
s
.89489 .3486 m
.87418 .36587 L
.87205 .34037 L
.89489 .3486 L
s
.88665 .48944 m
.90576 .45536 L
.93374 .48996 L
.88665 .48944 L
s
.85072 .30771 m
.8335 .30025 L
.84868 .28941 L
.85072 .30771 L
s
.88665 .48944 m
.93374 .48996 L
.90521 .52443 L
.88665 .48944 L
s
.89048 .17543 m
.87177 .17049 L
.88399 .15806 L
.89048 .17543 L
s
.88629 .4263 m
.91707 .42258 L
.90576 .45536 L
.88629 .4263 L
s
.89489 .3486 m
.8973 .37748 L
.87418 .36587 L
.89489 .3486 L
s
.89906 .40294 m
.91707 .42258 L
.88629 .4263 L
.89906 .40294 L
s
.92319 .1648 m
.91177 .18128 L
.90369 .16146 L
.92319 .1648 L
s
.8335 .30025 m
.83333 .28724 L
.84868 .28941 L
.8335 .30025 L
s
.9 .58571 m
.8619 .58571 L
.88267 .55512 L
.9 .58571 L
s
.91177 .18128 m
.89048 .17543 L
.90369 .16146 L
.91177 .18128 L
s
.88903 .32592 m
.87205 .34037 L
.8706 .31618 L
.88903 .32592 L
s
.94117 .18532 m
.97619 .16667 L
.97619 .20476 L
.94117 .18532 L
s
.92319 .1648 m
.94117 .18532 L
.91177 .18128 L
.92319 .1648 L
s
.85072 .30771 m
.8687 .29075 L
.8706 .31618 L
.85072 .30771 L
s
.89489 .3486 m
.87205 .34037 L
.88903 .32592 L
.89489 .3486 L
s
.90521 .52443 m
.92007 .55741 L
.88267 .55512 L
.90521 .52443 L
s
.85072 .30771 m
.84868 .28941 L
.8687 .29075 L
.85072 .30771 L
s
.8973 .37748 m
.92161 .39161 L
.89906 .40294 L
.8973 .37748 L
s
.90576 .45536 m
.94265 .44737 L
.93374 .48996 L
.90576 .45536 L
s
.91707 .42258 m
.94265 .44737 L
.90576 .45536 L
.91707 .42258 L
s
.89048 .17543 m
.87602 .18636 L
.87177 .17049 L
.89048 .17543 L
s
.92007 .55741 m
.9 .58571 L
.88267 .55512 L
.92007 .55741 L
s
.93374 .48996 m
.94369 .53179 L
.90521 .52443 L
.93374 .48996 L
s
.89906 .40294 m
.92161 .39161 L
.91707 .42258 L
.89906 .40294 L
s
.89489 .3486 m
.92084 .35924 L
.8973 .37748 L
.89489 .3486 L
s
.90521 .52443 m
.94369 .53179 L
.92007 .55741 L
.90521 .52443 L
s
.91177 .18128 m
.94117 .18532 L
.9163 .20567 L
.91177 .18128 L
s
.84868 .28941 m
.83333 .28724 L
.8468 .27377 L
.84868 .28941 L
s
.8973 .37748 m
.92084 .35924 L
.92161 .39161 L
.8973 .37748 L
s
.91177 .18128 m
.89521 .19474 L
.89048 .17543 L
.91177 .18128 L
s
.8706 .31618 m
.89166 .30196 L
.88903 .32592 L
.8706 .31618 L
s
.94117 .18532 m
.97619 .20476 L
.94188 .2265 L
.94117 .18532 L
s
.8706 .31618 m
.8687 .29075 L
.89166 .30196 L
.8706 .31618 L
s
.91536 .32253 m
.89489 .3486 L
.88903 .32592 L
.91536 .32253 L
s
.89048 .17543 m
.89521 .19474 L
.87602 .18636 L
.89048 .17543 L
s
.9163 .20567 m
.94117 .18532 L
.94188 .2265 L
.9163 .20567 L
s
.85876 .27687 m
.8687 .29075 L
.84868 .28941 L
.85876 .27687 L
s
.97619 .47143 m
.93374 .48996 L
.94265 .44737 L
.97619 .47143 L
s
.91536 .32253 m
.92084 .35924 L
.89489 .3486 L
.91536 .32253 L
s
.94699 .41105 m
.91707 .42258 L
.92161 .39161 L
.94699 .41105 L
s
.89521 .19474 m
.91177 .18128 L
.9163 .20567 L
.89521 .19474 L
s
.91707 .42258 m
.94699 .41105 L
.94265 .44737 L
.91707 .42258 L
s
.93374 .48996 m
.97619 .50952 L
.94369 .53179 L
.93374 .48996 L
s
.85876 .27687 m
.84868 .28941 L
.8468 .27377 L
.85876 .27687 L
s
.9381 .58571 m
.9 .58571 L
.92007 .55741 L
.9381 .58571 L
s
.89166 .30196 m
.91536 .32253 L
.88903 .32592 L
.89166 .30196 L
s
.97619 .47143 m
.97619 .50952 L
.93374 .48996 L
.97619 .47143 L
s
.94369 .53179 m
.95084 .56166 L
.92007 .55741 L
.94369 .53179 L
s
.92084 .35924 m
.94821 .37563 L
.92161 .39161 L
.92084 .35924 L
s
.94188 .2265 m
.97619 .20476 L
.97619 .24286 L
.94188 .2265 L
s
.8687 .29075 m
.88892 .27498 L
.89166 .30196 L
.8687 .29075 L
s
.89521 .19474 m
.88095 .20476 L
.87602 .18636 L
.89521 .19474 L
s
.92007 .55741 m
.95084 .56166 L
.9381 .58571 L
.92007 .55741 L
s
.85876 .27687 m
.87096 .26835 L
.8687 .29075 L
.85876 .27687 L
s
.92161 .39161 m
.94821 .37563 L
.94699 .41105 L
.92161 .39161 L
s
.9163 .20567 m
.89542 .2148 L
.89521 .19474 L
.9163 .20567 L
s
.94265 .44737 m
.97619 .43333 L
.97619 .47143 L
.94265 .44737 L
s
.97619 .43333 m
.94265 .44737 L
.94699 .41105 L
.97619 .43333 L
s
.91186 .2312 m
.9163 .20567 L
.94188 .2265 L
.91186 .2312 L
s
.88892 .27498 m
.8687 .29075 L
.87096 .26835 L
.88892 .27498 L
s
.91536 .32253 m
.94726 .33916 L
.92084 .35924 L
.91536 .32253 L
s
.94369 .53179 m
.97619 .50952 L
.97619 .54762 L
.94369 .53179 L
s
.85876 .27687 m
.8468 .27377 L
.85842 .26215 L
.85876 .27687 L
s
.94369 .53179 m
.97619 .54762 L
.95084 .56166 L
.94369 .53179 L
s
.92084 .35924 m
.94726 .33916 L
.94821 .37563 L
.92084 .35924 L
s
.91667 .28152 m
.91536 .32253 L
.89166 .30196 L
.91667 .28152 L
s
.89542 .2148 m
.88095 .20476 L
.89521 .19474 L
.89542 .2148 L
s
.91186 .2312 m
.89542 .2148 L
.9163 .20567 L
.91186 .2312 L
s
.91667 .28152 m
.89166 .30196 L
.88892 .27498 L
.91667 .28152 L
s
.94188 .2265 m
.97619 .24286 L
.94698 .26409 L
.94188 .2265 L
s
.85876 .27687 m
.85842 .26215 L
.87096 .26835 L
.85876 .27687 L
s
.97619 .39524 m
.94699 .41105 L
.94821 .37563 L
.97619 .39524 L
s
.95084 .56166 m
.97619 .58571 L
.9381 .58571 L
.95084 .56166 L
s
.94188 .2265 m
.92407 .25154 L
.91186 .2312 L
.94188 .2265 L
s
.97619 .39524 m
.97619 .43333 L
.94699 .41105 L
.97619 .39524 L
s
.91536 .32253 m
.9465 .30122 L
.94726 .33916 L
.91536 .32253 L
s
.94188 .2265 m
.94698 .26409 L
.92407 .25154 L
.94188 .2265 L
s
.91667 .28152 m
.9465 .30122 L
.91536 .32253 L
.91667 .28152 L
s
.95084 .56166 m
.97619 .54762 L
.97619 .58571 L
.95084 .56166 L
s
.88892 .27498 m
.87096 .26835 L
.88243 .25305 L
.88892 .27498 L
s
.89542 .2148 m
.87738 .21811 L
.88095 .20476 L
.89542 .2148 L
s
.94726 .33916 m
.97619 .35714 L
.94821 .37563 L
.94726 .33916 L
s
.89542 .2148 m
.91186 .2312 L
.8905 .23426 L
.89542 .2148 L
s
.97619 .35714 m
.97619 .39524 L
.94821 .37563 L
.97619 .35714 L
s
.9025 .25441 m
.91667 .28152 L
.88892 .27498 L
.9025 .25441 L
s
.94698 .26409 m
.97619 .24286 L
.97619 .28095 L
.94698 .26409 L
s
.87096 .26835 m
.85842 .26215 L
.86819 .25238 L
.87096 .26835 L
s
.92407 .25154 m
.94698 .26409 L
.91667 .28152 L
.92407 .25154 L
s
.91667 .28152 m
.94698 .26409 L
.9465 .30122 L
.91667 .28152 L
s
.88243 .25305 m
.9025 .25441 L
.88892 .27498 L
.88243 .25305 L
s
.8905 .23426 m
.87738 .21811 L
.89542 .2148 L
.8905 .23426 L
s
.92407 .25154 m
.9025 .25441 L
.91186 .2312 L
.92407 .25154 L
s
.87096 .26835 m
.86819 .25238 L
.88243 .25305 L
.87096 .26835 L
s
.92407 .25154 m
.91667 .28152 L
.9025 .25441 L
.92407 .25154 L
s
.9465 .30122 m
.97619 .31905 L
.94726 .33916 L
.9465 .30122 L
s
.9025 .25441 m
.8905 .23426 L
.91186 .2312 L
.9025 .25441 L
s
.94698 .26409 m
.97619 .28095 L
.9465 .30122 L
.94698 .26409 L
s
.94726 .33916 m
.97619 .31905 L
.97619 .35714 L
.94726 .33916 L
s
.9025 .25441 m
.88243 .25305 L
.8905 .23426 L
.9025 .25441 L
s
.9465 .30122 m
.97619 .28095 L
.97619 .31905 L
.9465 .30122 L
s
.8905 .23426 m
.88243 .25305 L
.87312 .23398 L
.8905 .23426 L
s
.8905 .23426 m
.87312 .23398 L
.87738 .21811 L
.8905 .23426 L
s
.88243 .25305 m
.86819 .25238 L
.87312 .23398 L
.88243 .25305 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 172.75},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00010`P10`4000010?ooo`000<c<c;no_`00
E@00ZW=cLiVIVNfXL<V>GoOI`7eI>ooooiVIVH:2P]7AdTi>C_ooohVHZWB1T<[AfDI=E_ooofQ_PUQN
K[ZmaS4e?_oooiBDU:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJF
UWemOFATI4];Bc8b<P0T000TE@0TZP0To`190019E@19ZP19o`1]001]E@1]ZP1]o`2B002BE@2BZP2B
o`2f002fE@2fZP2fo`3K003KE@3KZP3Ko`3oE@3oZUD005D0EED0ZUD0oeDT05DTEEDTZUDToeE905E9
EEE9ZUE9oeE]05E]EEE]ZUE]oeFB05FBEEFBZUFBoeFf05FfEEFfZUFfoeGK05GKEEGKZUGKoeGo05Go
EEGoZUGoojX00:X0EJX0ZZX0ojXT0:XTEJXTZZXTojY90:Y9EJY9ZZY9ojY]0:Y]EJY]ZZY]ojZB0:ZB
EJZBZZZBojZf0:ZfEJZfZZZfoj[K0:[KEJ[KZZ[Koj[o0:[oEJ[oZZ[oool0EOl0Z_lT0?lTEOlTZ_lT
oom90?m9EOm9Z_m9oom]0?m]EOm]Z_m]oonB0?nBEOnBZ_nBoonf0?nfEOnfZ_nfoooK0?oKEOoKZ_oK
ooooEOooZ/d0003=0<g=0000cLd0c@3=cNGUiDe=C@00:P00O`00e00Z000Z:P0ZE@0ZO`0ZZP0Ze01E
001E:P1EE@1EO`1EZP1Ee01o001o:P1oE@1oO`1oZP1oe02Z002Z:P2ZE@2ZO`2ZZP2Ze03D003D:P3D
E@3DO`3DZP3De2X002X0:RX0EBX0ObX0ZRX0e2XZ02XZ:RXZEBXZObXZZRXZe2YE02YE:RYEEBYEObYE
ZRYEe2Yo02Yo:RYoEBYoObYoZRYoe2ZZ02ZZ:RZZEBZZObZZZRZZe2[D02[D:R[DEB[DOb[DZR[De5D0
:UD0OeD0e5DZ05DZ:UDZEEDZOeDZZUDZe5EE05EE:UEEEOl0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@0006l01`5400P1A`040DX0105600@12`00078000<10000AP000`40000300030@0004@0
00<100000P000`40001900030@0004<00P4400030@0000P0001b00030@0004L000<10000C`000`40
001800030@0004X000<100001`00078000<10000B0000`40001>00030@0004<01`5;00030@0000L0
001b00030@0004T000<10000C@000`40001300030@00008000<10000BP000`4000070000LP000`40
001:00030@0004/000<10000A@001@40000104`000<100001`00078000<10000B`000`40001700<1
B0001@40000104H00P4300030@0000P0001b00030@0004`000<10000B@000`40001600040@000DH0
00<100400P4;0000LP000`40001600030@0000<000<10000BP000`40001500040@000DH000<10000
3@00078000<10000AP000`40000300030@0004X000<10000AP000`400@1600030@0000d0000V0003
0@0004H00P400`01001800050@000040A`000`40000200030@0004L000<10040AP000`40000=0000
9P000`4000180081BP030DT0105;0081AP060@X0000V00030@000?l0R00002H000<10000o`280000
9P000`40003o08P0000V00030@000?l0R00001d0o`6B0@80000V00813@030@`00`4<00<130030@d0
0`4<00<13@020@d00P4=00<130030@`00`4<00<13@030@`00`4=00813@020@d00`4<00<130030@d0
0P4=00<130030@`00`4<00<13@020@/0000V00030@0100`000<1004030000`400@0<00030@0100`0
00<100403@030@`000<1004030000`400@0<00030@0100d000<1004030030@`000<1004030000`40
0@0=00030@0100`000<1004030030@`000<100403@000`400@0<00030@0100`000<1004030030@d0
00<1004030000`400@0<00030@0100`000<100402`020@03004000X0000V00040@000@X000D10000
0@0:00050@0000402P001@40000100X000D100000@0;00050@0000402P001@40000100/000@10001
2`001040004;00050@0000402P001@40000100X000D100000@0:00050@0000402`001@40000100X0
00D100000@0;00040@000@/000@100012`001@40000100X000D100000@0:00050@0000402`001040
004;00050@0000402P001@40000100X000D100000@0:00050@0000402@001@40000100/0000V0005
0@0000402@001@40000100T000<100000P000`40000800050@0000402P001@40000100/000D10000
0@0:00030@00008000<100001`000`40000200030@0000L000<100000P000`40000700030@000080
00<100001`000`40000200030@0000P000D100000@0:00050@0000402`001@40000100X000D10000
0@0:00030@00008000<100001`000`40000200030@0000P000<100000P000`40000700050@000040
2P001@40000100X000<100000P000`40000800030@00008000<100001P000`40000300030@0000H0
00<100000`000`40000600030@00008000<100001P000`40000200030@0000T0000V00030@000080
00<100001@000`40000300030@0000H000<100000`000`40000600030@0000<000<100001P000`40
000300030@0000L000<100000`000`40000600030@0000@000<100001P000`40000300030@0000H0
00<100000P000`40000700030@0000<000<100001P000`40000300030@0000H000<100000`000`40
000600030@0000<000<100001`000`40000300030@0000H000<100000`000`40000700030@0000<0
00<100001P000`40000200030@0000L000<100000`000`40000600030@0000<000<100001P000`40
000300030@0000L000<100000P000`40000700030@0000@000<100001@000`40000300030@0000D0
00<1000010000`40000500030@0000@000<1000010000`40000300030@0000T0000V00030@0000<0
00<100000`000`40000400030@0000D000<1000010000`40000500030@0000@000<100001@000`40
000400030@0000L000<1000010000`40000500030@0000@000<100001@000`40000400030@0000D0
00<1000010000`40000500030@0000@000<100001@000`40000500030@0000D000<1000010000`40
000500030@0000@000<100001@000`40000400030@0000D000<100001@000`40000600030@0000<0
00<100001@000`40000400030@0000H000<1000010000`40000500030@0000<000<100001@000`40
000400030@0000H000<1000010000`40000600030@0000@000<1000010000`40000500030@0000@0
00<100001@000`40000400030@0000@000<100000`000`40000400030@0000T0000V00<110000`40
000200030@0000D000<100000`000`40000600030@0000@000<100001@000`40000400030@0000D0
00<100001@000`40000500030@0000@000<100001P000`40000400030@0000D000<1000010000`40
000500030@0000<000<100001P000`40000300030@0000H000<1000010000`40000500030@0000@0
00<100001@000`40000500030@0000D000<1000010000`40000500030@0000D000<100001@000`40
000400030@0000D000<1000010000`40000600030@0000<000<100001@000`40000400030@0000D0
00<100001@000`40000500030@0000@000<100001P000`40000300030@0000D000<100000`000`40
000600030@0000<000<100001P000`4000020@L000<100002@0002H000<100001@00104000490003
0@00008000<100001P000`40000300030@0000H000<100000`000`40000700030@0000<000<10000
1`000`40000300030@0000L000<100000P000`40000700030@00008000<100001P000`4000030003
0@0000L000<100000P000`40000700030@0000<000<100001P000`40000300030@0000H000<10000
0`000`40000600030@0000<000<100001`000`40000400030@0000H000<100000P000`4000070003
0@0000<000<100001P000`40000200030@0000H000<100000`000`40000700030@0000<000<10000
1`000`40000300030@0000L000D100000@0900050@0000402P001@40000100T000<100402@000`40
000900009P000`40000600030@0100T000D100000@0:00050@0000402P001@40000100X000<10000
0`000`40000800050@0000402`001@40000100X000D100000@0900050@0000402P001@40000100/0
00D100000@0:00050@0000402P001@40000100X000D100000@0:00030@0000<000<1000020001@40
000100T000<100000P000`40000800050@0000402@001@40000100X000<100000`000`4000070003
0@00008000<100002@000`400@0;00030@0100`000<1004020040@X000<100002@0002H000<10000
1P050@P000<100402`001@40000100X000D100000@0;00050@0000402`001@40000100/000@10001
2`001040004;00030@0100`000<1004030001@40000100X000D100000@0:00050@0000402P001@40
000100/000<100000P000`40000800040@000@/000D100000@0;00030@0100X000D100000@0;0005
0@0000402`001@40000100/000<100402`000`400@0<00030@0100801P4300<120000`4000090000
9P000`400004008100<00@000P060@8000<1004030000`400@0<00030@0100`000D100000@0<0003
0@0100d000<1004030000`400@0;00030@0100`000<100403@000`400@0<00030@0100`000<10040
30000`400@0=00050@0000402`000`400@0<00030@0100`000<100402`000`400@0<00050@000040
30000`400@0:02@12000104000020@H000<100002@0002H000<100000P020@<000<100001`0;0@H0
00<1004030000`400@0=00030@0100d000<100401P0l0@L000<1004030000`400@0<00030@0100`0
00<100403@001040004;00X11P000`400@0600h11P000`400@0=00030@0100d000<100400P080@80
00<100402`000`400@0<00030@0100X000<100000`020@@000<100002@0002H000<100000P450003
0@0000T00P4701h11`000`400@06010120020@`000<1004030020@`000<100401P0j0@/000<10040
1P050@800P4600l11P020@H03`4700030@0100H0304:00030@0100/000<1004030000`400@0:0003
0@0000D00P4200030@0000T0000V00<120000`40000700040@000@`000<1004030000`400@060101
20020@d000<1004030000`400@0;00030@0100`000<100403@020@d000<1004030000`400@0<0081
10080@<00P4200D11P000`400@0<00<130000`400@0;00<11P0@0@L0104:00050@0000402@001@40
000100X000D100000@0900030@0000L00`4;00009P030@P000<100001P001@40000100`000<10040
30000`400@0=00030@0100d000@1000130001040004:00050@0000402P001040004:00050@000040
30000`400@0<00030@0100`000<1004030000`400@0;00L12`001040004:00040@0100812@001040
004;008100<00@002`030@`000<100400P000`40000700030@00008000<100001@000`4000030003
0@0000H000<100000`000`40000500030@0000X000<100002@0002H00`4800030@0000D000<10000
0P000`40000900050@0000402P001@40000100`000<100403@001040004;00030@00008000<10000
1`000`40000200030@0000H000D100000@0:00050@0000402`001040004;00050@0000402P001@40
000100X000@100013P040@X000D100000@0900040@000@<000<100001P001@40000100T000<10040
0P000`40000800050@0100402`000`400@0300030@0000D000<100000`000`40000500030@0000@0
00<100001@000`40000400030@0000@000<1000020030@/0000V00030@0000811P000`4000030081
1P000`40000700030@00008000<1000020001@40000100/000D100000@0<00050@0000402P000`40
000200030@0000H000<100000`000`40000600030@00008000<100001P000`40000300030@0000P0
00D100000@0:00050@0000402P001@40000100X000D100000@0<00030@01008000<100001`000`40
000200030@0000D000<100000P001@40000100H000<100000P000`40000600040@000@<000<10000
1@020@030040008000<100001`001@40000100<00P4500030@0000@000<100000`000`4000050003
0@0000@000<100001@000`40000400030@0000H00P4200030@0000T0000V00030@00008000<10000
0`000`40000200030@0000H000<100001`000`40000300030@0000H000<100000`000`4000080003
0@00008000<1000020000`40000300030@0000H000<1000010000`40000500030@0000@000<10000
10000`40000400030@0000D000<100000`000`40000700030@00008000<100001`000`4000030003
0@0000H000<100000`000`40000600030@00008000<100002P001040004200811`000`4000020003
0@0000D000<100000P000`40000200030@0000<000<100000`000`40000400050@00004010000`40
000300050@00004010000`40000500030@00008000<100000`001@40000100L000<100000`000`40
000600030@0000<000<100001P000`40000200030@0000D00P4400030@0000T0000V00030@0000<0
0P4300050@0000402@000`40000600030@0000D000<100001@000`40000300030@0000L000<10000
0`000`40000800030@0000@000<100001@000`40000500030@0000<000<100001P000`4000030003
0@0000@000<1000010000`40000500030@0000D000<1000010000`40000600030@0000@000<10000
1@000`40000400030@0000D000<100000`000`40000800050@00004010020@@000<1000010000`40
000300030@0000<000<100000`020@8000<1000010000`40000400030@00008000<100000P001@40
000100@000<100000P000`40000400030@0000<000<1000010020@03004000L000D100000@0:0005
0@0000402P001@40000100D00P4600030@0000T0000V00030@0000D00P4200030@0100X000<10000
1@000`40000600030@0000@000<100001@000`40000500030@0000D000<100001P000`4000060003
0@0000<000<100001P000`40000200030@0000L000<100000P000`40000600030@0000<000<10000
1P000`40000400030@0000@000<100001@000`40000500030@0000@000<100001@000`4000040003
0@0000@000<100001`000`40000200030@0000@000D100000@0600030@00008000<1000010000`40
000500030@0100L000<100000P000`40000300030@0000<000<100401@000`40000300030@000080
00<1000010000`40000600<120000`400@0<00030@0100`000@100010`020@P000<100002@0002H0
00<100001`040@`000<1000010000`40000700030@0000<000<100001P000`40000400030@0000H0
00<100001@000`40000700030@00008000<100001`001@40000100X000D100000@0900050@000040
2@000`40000300030@0000H000<100000`000`40000700030@0000<000<100001P000`4000030003
0@0000D000<100001@000`40000300030@0000D00`4800040@000@L000<100001P020@L000D10000
0@0600030@0000<00P4600030@0000@000D100000@0600030@0000@00`4200H10P000`400@0<0003
0@0100`000D1000100020@X000<100002@0002H00`49008130000`40000300030@0000T000D10000
0@0900030@0000<000<100001`000`40000500030@0000L000D100000@0;00030@0100`000<10040
2P001@40000100X000D100000@0900030@0000<000<1000020001@40000100X000D100000@080003
0@0000@000<1000010000`40000700D110000`400@0800030@0000@00`400`010@020@@000<10040
1`000`400@030@0400410@@000<100001@000`400@0700030@0000800P4200030@0000H06P4700@1
30000`40000900009P000`40000700<100<00@402P000`40000200030@0000X000D100000@0:0005
0@0000402`000`40000300030@0000T000@100012`000`400@0<00030@0100/000<100402`001@40
000100X000D100000@0;00050@0000402P001@40000100T000<100000P000`40000600030@0000D0
0P4400L120000`40000300040@000@D01@400`01000200H11@001040004200030@0000D01@470004
0@000@@000<100001`040@`00P4700X13@000`40000900009P000`40000600040@000@<00`480005
0@0000403@000`400@0;00050@00004030000`40000200030@0000X00P4601d11P000`400@0<0003
0@0100/000D100000@0<00030@0100`000<100402P000`40000200030@0000H000<1000010001040
0047008100@00@411P001@40000100@000<100001P060@@00P4500040@0000<10P050@0300410080
1@4200<11@000`40000600030@01008000<100002@000`400@0=008100<00@402`000`4000090000
9P000`40000400810`000`40000400<11@001040004>00030@0100`000<100403@000`4000020003
0@0000T02@47008130000`400@0500l11P000`400@0<00030@0100d000<1004030000`400@0;0004
0@000@T000<100000P020@<000<100001000104000430081100010400@4500030@0000H00`470003
0@0100<000<100000`040@H000<100401P040@L000<100000`020@8000@100012P001040004<0003
0@0100<00P4900030@0000T0000V00030@0000<000<100000`000`40000700810`000`400@0<00`1
1P000`400@0>00040@000@P0104200812`000`400@0<00030@0100/000<100401P0?0@H000<10040
1P0N0@H000<100402P001@40000100D000<1000010001040004500<100<00@401`000`4000040081
00@00@011@00104000020@03004000800P400`010@0600040@000@L00P400`010@0500040@000081
10001@40000100P000<100000P000`40000800040@000@D00P4700030@0000T0000V00040@000081
1P000`40000900<100<00@400`090@@00P4600`130000`400@0400@11@000`400@0;00040@000@X0
00D100000@0:00030@0100`00P4700l120020@`000<100401@0;0@T000<100401P000`4000030005
0@00004020030@L000<100000`001@400001008000<100000P000`400003008100<00@400`020@H0
00@100011P000`400@0300810`001040004600030@00008000<100001@000`40000200030@0000L0
00<100000P000`40000400811@000`40000900009P001040004800030@0000`01P4<00030@0100`0
0P4400T10`020@03004100812@001040004900030@00008000<100001`001@40000100T000D10000
0@0;00030@0100`000<100403@000`400@0<00030@0100/00P4300H10`020@P000D100000@050003
0@0000D00P400`010@0500030@00008000<100000P001@40000100<000<100000`030@@000@10001
10000`40000200030@00008000@100011@020@0400410@H000<1000010000`40000300030@0000@0
00<1000010020@D000<100001P020@<000<100002@0002H00`4900030@0000/0104?00040@000@X0
00@1000130060@d000D100000@0800030@00008000<100001P000`40000300030@0000D000<10000
0P000`40000800040@000@`000<100403@001040004:00050@0000402P000`400@0800D120001040
004600030@0000@000@100010P030@<000@100411@000`40000200040@000@800`400`010@040005
0@0000400`000`40000200050@0000400`000`400005008120000`40000500050@0000401`000`40
000300030@0000D000<1000020020@03004000X0000V00<12P000`40000800810P000`400@0=0003
0@00008000<100001`001040004?00@12`000`40000300030@0000@000<1000010000`4000050003
0@0000<000<100001@000`40000300030@0000H000<100000P000`40000800050@0000402`001@40
000100X000D100000@0900040@000@X00P400`010@020@@000@100011`000`40000200050@000040
1@020@03004100L000<100000`020@03004100@000<100400`000`40000200030@0100D000@10001
1@000`40000300D11P000`40000500040@000@T000D100000@0800030@0000X00P4;00009P030@X0
00<100001`000`40000200040@000@`000<100000P000`40000600050@0000403P000`400@020081
2@000`40000300030@0000@000<100001@000`40000300030@0000D000<100000`000`4000040003
0@0000D000<100000`000`40000800030@00008000<1000020000`40000200030@0000H000<10000
0`000`40000600050@0000402@000`400@0400@100<00@4020001@40000100@000<100001@030@L0
00<1000010020@D000@100010`000`40000200030@0100H00P4600040@0000810P000`4000020@@0
00<100001P000`400@0900040@000@T000<100002@030@/0000V00030@00008120000`4000050081
1@001@40000100X000<1000010000`40000400030@0000<000<100002P001@40000100<000<10000
1@000`40000500030@00008000<100001P000`40000200030@0000L000<100000P000`4000050003
0@0000@000<1000010000`40000600030@0000<000<100001`000`40000400030@0000D000<10000
0`000`40000500030@00008000<100001P001040004800@11`001040004500030@0000D00P400`01
0@0500030@000080104500070@0000400@0500/11P001040004400030@0000800P400`01000400/1
1P020@X000<100001`020@8000<100002@0002H000<100000P020@H000<100000`020@H000<10000
0`000`40000700030@0000@000<100000`000`40000400030@0000T000<100000P000`4000020081
1@000`40000600050@0000402@001@40000100T000D100000@0900030@00008000<100001@000`40
000600030@0000@000<100001P000`40000400030@0000@000<100001@000`40000400030@0000<0
00<1000010000`40000200030@0000D00P4200H10P020@L000<100000`000`400@0300810`000`40
00020@8000<1004010000`400002008100<00@400P400`010@05008100<00@400P4200<11@000`40
000400L110020@D0204:00030@0000H000<100000P000`40000900001@030@P00P4800@120000`40
0004008110000`40000200030@0000H000<1000010000`40000600030@0000D000D100000@070003
0@0000P000<100000`000`40000400810`000`40000700030@0100/000<100402`000`400@0:0005
0@00004020000`40000500030@0000H000<1000010000`40000600030@0000<000<100001@000`40
000300030@0000@000<1000010000`40000200030@0000D000<100401`040@L000@100000P430003
0@0000800P4010010@4400030@0000810P000`40000200<110000`400@0500030@0100@010460003
0@0000800P400`010@0800030@0100T00P400`010@0800030@0000@00P4500030@0000T000040005
0@0000401`020@H00P4400030@0000D000<100001P020@8000<100000P4900030@0000D000<10000
10000`40000700040@000@L000<100001`000`40000400030@0000H00`4:00030@0100/000<10040
2`000`400@0;00030@0100X000<1000010000`40000700030@0000<000<100001`001@40000100T0
00D100000@0800030@00008000<1000010000`40000300040@000@P00P4010010@4400040@000@D0
00<1000010020@H000<100000P020@030041008100<00@4010001040004300040@000@D00P400`01
0@0400030@0000810`000`400@0600040@000@P000<100400`030@H000<100000P000`4000050003
0@0000T0000300030@0000<000<100004`000`40000400030@0000P01@4;00030@0000H000<10000
0`000`40000700030@0100T000<100001P000`40000500030@0000L0;`4600030@0100X000<10000
0`000`40000800030@00008000<1000020001@40000100T000D100000@0800050@0000401`000`40
000200030@00008000<1000010001040004300@100<00@401P000`40000400@110000`4000030081
10000`400@0400070@0000400@0300040@00008100<00@000P030@0400410@@000D100000@050005
0@0000401P001040004600<10`000`4000020@P000<100002@0000<000<100000`000`40000C0003
0@0000@000<100002P000`40000;00030@0000L000D100000@0900@12@000`40000500030@0000H0
00<100001P000`400@0900030@0100/000<100402`020@H02@4;00030@00008000<100002@001@40
000100/000<100402`000`400@0:00030@0100T000@100011@000`40000300050@0000401`030@H0
00<100000`000`400@0200050@0000400`040@@000@100010P000`40000200030@0100<000@10001
0`000`40000400811P001@40000100@000<100000P000`40000300030@00008000<100001P020@04
00410@X000<100002@0000<000<100000`000`40000C00030@0000@0104700D12`000`4000080004
0@000@D0104200040@010@L000<1000010000`40000700030@0000D000@100012@000`400@0;0003
0@0100/000<100402`020@0400410@P000D100000@0=00030@0100`000<100402`000`400@0:0003
0@0100T000@100011@000`40000300050@0000401`020@03004100@000@100000P4200050@000040
0P001040004200030@0100<000L100000@0100@01043008110000`400002008100@00@4110000`40
000200040@000@H000D100000@0500030@0000P00P4<00030@0000T0000300030@0000<000<10000
4`000`40000400030@0000H00P4200030@0000812@000`40000900810P040@H000@100000P460003
0@00008000<1000020000`40000400030@00008000<100001@001@40000100T000D100000@090004
0@000@/000<100400`030@H000@100013@000`400@0602P11P020@L000D100000@0400030@0000@0
00<100400`020@<000<1004010000`40000200D10`001040004200040@00008100<00@400P4400D1
10000`4000020@<000@100010`000`40000300030@0100H000@100011P000`40000700D12P000`40
000900000`000`40000300030@0000d00P4300030@0000D000<100001@000`40000200030@000080
00<100001P000`40000900@12P000`400003008110001@40000100`000<100000P000`4000030003
0@0000D000<100000P000`40000500030@0000<000<100001@000`40000200030@0000L000@10001
1P030@<000<100402`0<0@L00P4;00030@0100X000<1004010090@H000@100011P000`4000020005
0@00004010020@03004100H000<100000`000`40000200030@0000800P4200D13@030@0400410@D0
00@100000`4300T110020@L000<100001P0010400042008120000`40000900000`000`4000030003
0@0000d000<100400P4800030@0000<00P4500030@0000<00P4500030@0000L00`400`010@0:0003
0@0000H00P4200040@000@d000<100000P000`40000400030@0000<000<100000`000`4000050003
0@0000<000<100001@000`40000200030@0000H000D100000@0900<100<00@400`080@@00P4<0003
0@0100/000<100402P000`400@0900810P050@800P4700050@00004010000`40000400<11P001040
00050@8000<100000P040AD00`4500030@0000<01@4400810`060@L000<1000010020@<000<10000
0P020@H000<100002@0000<000<100000`000`40000=00030@0000X000<100000P000`4000050003
0@0000D000<100000P000`40000400<1100010400@4800030@0000P00P400`010@0>00050@000040
1`000`40000300030@0000@000<100000`000`40000500030@0000<000<1000010000`4000040003
0@00008000<100002P060@/000<1004030001040004900050@00004020001@40000100P000<10040
1P050@H000@100011@000`400004008100<00@4010001040004200030@0000<100@00@4170040@03
004000@100<00@401@000`400@06008100<00@401@000`40000200811@000`400004008110000`40
000900000`000`40000300030@0000d000<100002P000`4000020@P000<100001P020@8000<10000
0P020@L000D100000@0700030@0000X01P4;00040@000@T000D100000@0700030@0000<000<10000
1@000`40000300030@0000@000<1000010000`40000300030@0000L00P400`010@0>00040@000@X0
00D100000@0900030@00008000<100001@001@40000100L000@100011`020@801@4200811P000`40
000300030@0100<00`4010010@4300030@0000800P4S00<110001040004400040@000@@000<10040
0`030@<000@100011`000`40000600810P000`400009000010001@400001010000<100002P030@X0
00<1000020030@0400410@P000<100000`000`40000400030@0000P00P400`010@0300P110020@X0
00D100000@0800050@0000402@001@40000100P000<100000P000`40000400030@0000D00P420004
0@000@d000D100000@0900030@00008000<100001@000`40000300030@0000@000<100000`000`40
000400050@0000401P000`400@0600@11P000`40000200050@0000401@020@D000D100000@0W0003
0@00008000D100000@0200030@00008000@100010P000`400004008100@00@4120000`40000800<1
2`0000D00`4A00H11`030@X000<100002P020@/000<1000010000`40000300030@0000D00`420003
0@0100/0204900030@0100X000<100402`000`400@0900050@0000401`000`400003008110001040
004<00030@0000<000<100001@000`40000300030@0000D000<1000010000`40000300030@0000<0
00<100000`000`40000200030@0000<000@100011`020@03004100@000D100000@0400030@0000<0
104300040@000BT000@100011@000`400@0400030@0100<000<100001P020@X000<100002P000`40
000900009P030@X000<1000020040@/000<100001@000`40000200030@0000<00P4500040@000@d0
00<100400P070@8000<100402P000`400@0;00030@0100X000<100402000104000020@H000D10000
0@0;00030@0000<000<100001@000`40000400030@0000<000<100001@000`40000200030@0000D0
00D100000@0600030@00008000D100000@0500030@0100<00P4300030@0100D000<100000P000`40
0@0200060@000@01:`000`400@0200`11@000`40000400@120000`40000900<12`0002H000<10000
0P4800030@0000H00P4200040@010@T000<100001P020@03004000800`4700050@00004030000`40
0@0902D11@000`400@0900<120000`40000200030@0000L000<100001@000`40000300030@0000H0
00D100000@0900050@0000401`001@40000100H000D100000@0400030@00008000D100000@040081
00<00@401P001@400001008000@10000104]00@10`000`400@0500<10`000`40000200810P001040
0@4600030@0000L00P4200030@0000T0000V00030@00008000<100001@000`40000500030@000080
00D100000@0800030@0000P00P400`010@0900030@00008000<100002@001@40000100X000<10040
2P000`400@0<00811@0B0@T000<1000010000`40000600030@0000H000<100000P000`4000060005
0@0000402P000`400@0900030@0100P000<100401P001@40000100<000<1000010040@H000<10040
0`000`40000300030@0002d000<100000`001040004300050@0000400P000`4000020@@000D10000
0@0500030@0000D00P4400030@0000T0000V00030@0000<00P4500030@0000<00P4500030@000080
00<100001@000`40000800<12`000`40000300030@0000L000<100000P000`40000800030@0100X0
00<100402`000`400@0:00030@0100T00P4800030@0000@000<100001P000`40000700040@000@X0
00<100402`000`400@0900030@0100P000<100401P001040004400030@0000<010400`010@020@80
0P4500050@000040<@020@8000<100000P001040004400D11P000`40000200030@00008000<10000
10000`40000400030@0000T0000V00030@0000D000<100000P000`40000200030@0000D000<10000
0`020@D000<100001P020@03004100/000<100000`000`40000700030@0000<000<100001P001@40
000100P000D100000@0:00040@000@T000<1004020001040004700030@0000D000<1000010000`40
000800040@000@X000<100401@0S0@@00P4500040@0000810P020@D0104500040@000C@000<10040
1@020@<01P4700030@0000<000D100000@0400811`000`40000900009P000`40000600810P000`40
00020@P000<100001@001@40000100H00`4300030@0100X000<1000010000`40000500030@0000@0
00<100001@000`40000300030@0000D000<100000P000`40000600050@00004020001@40000100L0
00D100000@0500030@0000L000<100000`000`40000900811P0=0@H000<100402P020@P000<10040
0`080@@00`4300040@000@H000@100410`000`400@0f00`10`020@0400410@@000<1000010020@04
004000812@000`40000900009P000`40000800D12P000`40000600040@000@<00`4500050@000040
20000`40000600030@0000<000<100001P000`40000400030@0000<000<1000010000`4000030003
0@0000H000<100000P000`40000500030@00008000<100000`000`40000300030@00008000<10000
20000`40000200030@0000T0204500030@0100/000<100402@000`400@0800030@0100H000<10040
0P050@D000D100000@0400030@0100800P400`010@0h00030@0000@00P4500030@0100<00`400`01
000700@12`000`40000900009P030@X000<100002P000`40000700<100<00@4020000`4000020003
0@0000D000<100001P000`40000300030@0000H000<100000`000`40000500030@0000<000<10000
10000`40000400030@0000<000<1000010000`40000300030@0000<000<1000010001@40000100/0
00@1000120040@03004100X000<100402P001@40000100P000@100011P001@40000100D000<10040
1P020@@000D100000@0400030@0100@00P4j00030@00008000<1004010001040004600`13@000`40
000900009P000`40000800D12P000`40000800<12P000`40000300030@0000@000<100001`001@40
000100X000<100000P000`40000500030@00008000<100001P000`40000300030@0000@000<10000
0`000`40000400030@00008000<100001@001040004;00040@000@@0104500030@0100P000D10000
0@0900030@00008000<1000010000`40000200030@0000<000D100000@0400050@00004010001040
004300030@00008000@100010`001@40000103/000<100000P001040004300050@0000401@020@P0
104;00030@0000T0000V00030@0000H00P4200030@00008120000`400006008100<00@402P000`40
000400030@00008000<100002@000`400@0;00050@0000402@001@40000100P000<100000P000`40
000600050@00004020001040004900813@020@03004100812@001040004700050@00004020000`40
000300030@0000@000<100000P000`40000200030@0000<000D100000@0400050@0000400`000`40
0@0600030@0100<000@10001?@001040004400030@0100@000<100000P001040004600030@010080
0P4900030@0000T0000V00030@0000D000<100000P000`40000200811P000`40000400810`001040
0@4700030@0000H000D100000@0;00030@0100`000<100402`000`400@0:00050@00004020001@40
000100T000<100401`0D0@`000<100000P000`40000300030@0000<000<100001@000`4000040003
0@00008000<1000010001@40000100H000<100401P000`400@0500811P020@D00P4o00030@0100@0
00<100401@001@400001008000<100000P020@8000<100000P020@L000<100002@0002H000<10000
0`020@D000<1000010020@@000<100000P020@D000D100000@0600030@0000L000@100012@0;0@H0
00<100402`000`400@0;00030@0100X000<100402P020@801P43008130040@/000<100000`000`40
000200030@0000<000<1000010000`40000500030@00008000<100001@000`400@0700030@0100H0
00<100400P0C0D402@4500040@000@<000D100000@0400030@0000@000<1000010000`4000090000
9P000`40000200030@0000D000<100001P020@8000<100000P4700030@00008000<100000`000`40
000800810`070@<00P4601`11P000`400@0:00030@0100D02@4800030@0100/000<100400P000`40
000800030@0000@000@100011`000`40000300030@0000H000@1000120000`400@03018110020@H0
0P4400030@0003l000<100000`020@03004100810P020@@000@100011@000`400005008110000`40
000900009P000`4000020@P000<1000020050@T000<100000`000`40000200030@0000L01P490003
0@0100`000<100402`000`400@0501/11P020@T000@100012@0010400043008120000`4000050003
0@0100P000D100000@0:00030@0100L01`4500811`020@D000<100401@000`400@0300030@000440
00D100000@0200030@0000801P4300811P000`40000700810P000`40000900009P030@X000<10000
2P000`40000900030@0000@00P400`01000500@10P020@/000@100012`000`400@0;00030@0100/0
00<100402P000`400@0:00030@0100P000D100000@0800050@0000401@000`40000500030@0000D0
0`4800050@0000402P020@801P4200811`001040004500030@0100D000@100011000104000420003
0@00044000@100010`000`40000300810`040@L000<100002@030@/0000V00<12P000`40000800D1
20000`40000700810P040@D000<100402P001@40000100X000D100000@0900050@0000402P000`40
0@0:00030@0100X000@100011`000`40000200030@0000D000D100000@0600030@0000@000<10000
0`020@03004000@11@000`400@0900H11`001040004600050@00004010001040004400050@000040
0P001@400001008000<10000@P020@@000@100000P4200030@00008000<100401P000`40000:0081
2`0002H00`4:00030@0000H00P4200040@000@L000<100001`040@T000@100012@000`4000020003
0@0000L000D100000@0900050@0000402@001@40000100P000D100000@0800030@00008000<10000
10000`40000300030@0000<000<100000P000`40000500810`000`40000200810`000`40000300@1
00@00@010P070@<00P4800050@0000401@001@40000100<000<100000P001040004500030@0100@0
00<10040A0020@@000@1000110000`40000200040@000@D000<1000020020@03004000X0000V0003
0@00008120000`40000500030@00008000@100000P4500030@0000D00P400`010@0:00050@000040
1`000`40000400030@0000D000<100000`000`40000500030@0000<000<100001@000`4000030003
0@0000D000<100000P000`40000500030@00008000<100000`000`40000400030@00008000<10000
0`000`40000700040@000@800P4500030@0000L01P4:00030@0100L000D100000@0400030@0000<0
00@100011@000`400@0500811@020DD000<100400`400`010@0600040@000@@000<100000P000`40
000600810`000`40000900009P000`40000200811P000`40000300811@000`40000300810`000`40
000300810`000`400@0800030@0000<000<1000010000`40000400030@0000D000<100000`000`40
000500030@0000<000<100001@000`40000300030@0000@000<1000010000`40000400030@0000<0
00<100000P000`40000500040@000@L000<100001`050@L000<100001`040@/000D100000@060003
0@00008000D100000@0600811P020@<0305600030@000080104500030@0100H000D100000@060081
1@000`40000900009P000`400004008110000`40000200030@0000D000<100001@00104000430081
1@001040004700030@0000<000<100000`000`40000600030@0000<000<100001@000`4000030003
0@0000D000<100000`000`40000500030@0000<000<1000010000`40000300030@0000D000@10001
20000`400@0800030@0000P00P4900030@0000H000H10001004:00030@00008000<100000P000`40
000400040@000@D03P4400811@000`40001400030@0000800P4200@10P020@L000@1000110020@L0
00<100002@0002H000<100001P020@8000<100000P4800030@0000H00`400`010@0600030@000080
00<1000010000`40000400030@00008000<100001`001@40000100T000D100000@0900030@000080
00<100001@000`40000200030@0000H000<100000P000`40000600030@0100L00`4900030@0000H0
0P4010010@4700030@0000@00P4300040@000@T000<100000`001@40000100L00P4200@10P020@H0
0P4500040@000@@000<10000A0000`4000020@03004000D01@4700030@0100800P4900030@0000T0
000V00030@0000P01@4:00030@0000P00P4800030@0000<000<100000P000`40000600040@000@X0
00D100000@0900050@0000402@001@40000100T000D100000@0900040@000@T00P4200H10P030@L0
00<1000010020@<000<100000P4500030@0000<000<100000`001@40000100P000<1000010001040
004600D11P000`400@0400030@0100D000D100000@0300030@0004D00P4400030@0000@00P4200H1
0P040@/000<100002@0002H00`4:00812`000`40000600@120000`40000400050@0000402@000`40
0@0;00030@0100/000<100402`000`400@0;00030@0100/000<100401`060@P000<100000P450003
0@0000800P4500030@00008000<100000P00104000020@H000<100000P020@D000<100001P020@80
1@4200811P001040004400040@000@@000D100000@0200030@0004H000<100000`000`4000030003
0@0100P0104=00030@0000T0000V00030@0000P00`400`010@0900030@0000D000D1000100020@H0
00<100001@0010400049008130000`400@0;00030@0100/000<100402`000`400@0;00810P060@<0
0P4900030@0000800P4300030@0000811`000`40000300810P001040004800030@0000@000<10000
0P000`40000600@11`000`400@0500050@0000400P000`40000200040@000@D000<10040B0030@<0
00@100000P4300030@0000L0104;00030@0000T0000V00030@0000H00P4200040@0000811`000`40
000300810`001@40000100D000<100001P020@D0C`4800040@000@L000<100001@020@0400410@T0
00<100001@050@T000<100001@001@40000100D00`400`010@0700050@00004010000`4000020003
0@0100D000<100401P020DP000<100000P401@01000100D000<100001@020@04004000812@000`40
000900009P000`40000500030@00008000<100000`020@D000@100000P4500030@00008000<10000
0P000`40000600L11P020@/000<100402`000`400@0;00030@0100/000<1004030020@T000D10000
0@0600030@0000L00P4;00030@0000L000<1000020000`40000700040@000@<00P4400030@0100H0
00<100000P001@40000100@01@43008110050DP000<100000P040@H000<1000010001@40000100@0
0P4700030@0000T0000V00030@0000<00P4500030@0000D00P4300040@000@L000<100000`020@03
004000@010400`010@0900030@0100/000<100402`000`400@0;00030@0100/000<100402`000`40
0@0900030@00008000<100000`000`40000600D120000`40000600D120000`400008008100@00@41
1@001@40000100D000<100000`000`400@03008100<00@000P090@@000<10000AP000`4000040081
1P000`40000300030@00008000<1000010000`40000400030@0000T0000H00L11`000`4000020003
0@0000D000<100001`020@0400410@P000<100001@020@030041008110000`400@0900040@000@T0
00D100000@0900050@0000402@001@40000100T000D100000@0:00040@000@L000<100000`000`40
000300030@0000@00P4200040@000@L000<100001@001040004200030@0000D000<1000020030@P0
00<100000P020@<000<1000010020@03004100<000<1000010040@D000<10000B000104000030@03
004100@000<100000P000`40000300030@0000D00P4400030@0000T0000K00030@0000P000<10000
0P4800030@0000T00`4900030@0000D00`4800040@000@L000<100000P000`40000600050@000040
2@001@40000100T000D100000@0900030@00008000<100001P000`40000200030@0000@000<10000
10000`40000200030@0000800P4400040@0000811@000`40000300810`00104000020@D000<10000
1P020@03004100P000<1000010000`400@0600@11@000`40000300030@0100800P4300030@0004P0
00@100010P00104000020@<000@100411P000`40000700810P000`40000900006`000`40000800<1
2P000`40000800<100@00@411P000`400003008100<00@401`000`40000200030@0000@000<10000
0`000`40000400030@0000<000<100001@000`40000300030@0000D000<100000`000`4000050003
0@0000<000<100001P000`40000300030@0000<000<100001@001@40000100<000<1000010000`40
000300810`00104000020@D000<100000`020@<000<100000`030@<000@100411P000`4000050081
0P040@800`4400030@00008000D100000@0300<1B`030@<000<100000`020@03004100P000<10000
2@030@/0000K00030@0000P0104:00030@0000H000@1000110020@@000@100000P4300030@0100H0
00<100000P000`40000300030@0000@000<1000010000`40000300030@0000D000<100000`000`40
000500030@0000<000<100001@000`40000400030@0000@000<1000010000`40000300030@0000H0
00<100400P020@L000<100001@001040004200030@0000D000<100001@001040004300811P001@40
000100D000<100001@040@H000@1000110001040004400030@0000<000<10000B@000`4000030003
0@0000D00P4900030@0000/000<100002@0001/000<1000020030@/000<100001@000`4000020003
0@0000<00`400`0100020@D000@100011@000`40000300030@00008000<100001@000`4000020003
0@0000D000<100000`000`40000500030@0000<000<100001@000`40000300030@0000H000<10000
0`000`40000500050@0000402P040@T000<100001P050@P000<100001P030@03004100P000<10000
0P000`40000200030@0000<00P400`010@0700040@0000810P000`400@0500030@0000800P5;00<1
0`000`400003008100<00@4020000`40000900<12`0001/000<1000020000`4000020@T000<10000
0`020@D000<100001P030@H000<100000P001@40000100L000@100012@001@40000100P000D10000
0@0900050@0000402@001@40000100T000<100000P000`40000700040@000@T00`4;00030@0000P0
00<1000020000`40000800812P000`400003008100<00@000P030@<000@100411@000`40000300@1
1P001@400001008000<10000B0001040004200040@0000810`000`400@0700030@0000P000@10001
2`0001/000<1000020000`40000200811`000`40000200030@0000D000<100001P000`400@060003
0@0000<000@1000120000`400@0:00030@0100T000D100000@0900050@0000402@001@40000100X0
00D100000@0:00030@0100D0104200<12@000`40000600D120000`40000600D12@000`4000050081
00<00@401P001@40000100<000<100001@040@H000<100400`000`40001700030@0000800`400`01
0@0400050@0000401P000`40000600810`000`40000900006`000`40000800030@0000@00P450005
0@00004020000`40000600040@000@D000<1000010020@T00P4;00030@0100X000<100402`000`40
0@0;00030@0100`000<100402`020@80104600030@0000811`000`40000500040@000@800P460003
0@0000@00P4200040@000@L000<100001P030@P000<100000P001040004400<100@00@4010420081
10000`40001700040@0000D11P000`40000200030@0000<000<1000010020@D000<100002@0001/0
00<1000020000`40000600810`0010400@4:00030@0000@000<100000P000`40000200030@0000D0
7`4500030@0100/000<100402`000`400@0<00030@0100D02P4:00030@00008000<1000010000`40
000300810`000`400002008110000`40000200811000104000020@D000<1000010020@0400410@L0
00<100000`030@0400410@@000<1004010040@@000<10000A`040@<00P4600030@0000<00P430003
0@0000<000<100001@000`40000900006`000`40000800030@0000P00P400`010@0<00030@0000@0
00<100000P000`40000200030@0000<00`4900030@0100X00P4603P11`020@/000<100000`020@@0
00@100000P4500030@0000@00P4200030@0000811P000`40000300810`000`40000200810`001040
004500030@0000D00`4600040@000@D00P400`010@020@03004004L00P4500030@0100D000<10000
1@0010400043008120000`400009000060020@03004000T00`4:00<130000`40000300030@0000@0
00D100000@0300810`000`40000500040@000@X000<100402P000`400@0;00030@0100/000<10040
30000`400@0;00040@000@T000<100001P020@8000@100011`000`40000600D120000`4000050004
0@000@800P4500040@0000810`000`400003008100<00@401P001@40000100<000@1000110020DL0
00<100401000104000020@<000<100001P000`400@0200030@0000P000<100002@0001X00P4:0003
0@0000T00`400`010@0:00030@0000<000<100001@000`400@0200811@000`40000500050@000040
20001040004:00030@0100/000<100402`000`400@0<00030@0100/000D100000@0800030@0000P0
1@4800030@0000P000<1000020000`40000600D11`000`40000300040@000@<00P4300030@0100D0
00<100000P000`400@0300030@0000800P5700040@000@<000<100000`000`400@0:00@12`000`40
000900009P000`40000700810P00104000020@T000D100000@0900@11`000`40000400030@000080
00<100001@000`40000200030@0000H000D100000@0900050@0000402@001@40000100X000D10000
0@0900030@00008000<100001P000`40000:00030@0000T000<100001@050@P000<1000020000`40
000700030@0000@00`400`010@0500030@0000810`000`400002008110001@400001008000<10000
A0001@400001008000<1000010020@D0204<00030@0000T0000V00030@0000D00P4400030@0000<0
0P4700040@000@X00`4800030@0000<000<1000010000`40000400030@00008000<100001@000`40
000300030@0000D000<100000P000`40000700050@0000402P001@40000100T000<100000`000`40
000500030@0000P01@4900030@0000@000@100010P000`40000500030@0000H01@4600030@0000H0
0`4700030@00008000<100400`020@0400410@8000@100010`000`40001400030@00008000<10040
1@080@D00P400`010@0:00030@0000T0000V00030@0000@000<1000010000`40000500<110001040
004600@100<00@000P4700050@0000401`000`40000300030@0000@000<1000010000`4000030003
0@0000D000<100000`000`40000500030@0000<000<100001P000`40000300030@0000D000<10000
1@000`40000400030@0000L000@100010P020@L000<100000P020@<000@100000P4500030@0000@0
0P4200040@000@D000<1000010020@03004100L000<100000`020@03004100<000<100000P400`01
0@0400030@0004<000<100000`020@80104200812@020@<00P4800030@0000T0000V00030@000080
0P4700030@0000P00P4200030@0100@00`4500040@0000811@001040004900030@00008000<10000
10000`40000300030@0000D000<100000`000`40000400030@0000D000<1000010000`4000050003
0@0000@000<1000010000`40000600030@0000<000<100001@020@<000<100000P020@D000D10000
0@0500030@0000<00P4300030@0000800P4400040@0000810`000`40000200810`0010400@440003
0@0000@00`4500030@0000800`4400030@0004<02P4500030@0100P000<100401@030@D000<10000
2@0002H000<100000P4900030@0000X00P400`010@040@P000<100000`030@8000@100012@001@40
000100P000<100000P000`40000500030@0000<000<100001@000`40000300030@0000D000<10000
10000`40000500030@0000@000<100001`000`40000200030@0000<00P4600030@0000<00P430004
0@000@H000<100001@001040004200811P000`40000300040@000@800P4500050@0000400`000`40
0002008100@00@411000104000030@030041008100<00@00@`020@D00P4600040@000@H000D10000
0@0700810`000`40000900009P030@/000<1000030030@`000<100001P020@03004100/000<10040
2@001@40000100T000D100000@0900050@0000402@000`40000300030@0000H000<100000P000`40
000900050@00004010000`40000600030@0000D00P4010010@4700030@0000H01@4800030@0000@0
1@4700030@00008000@100010P020@<000<100000P4200040@000@<000<100000P020D<000<10040
10000`400@0500050@0000401P001@40000100T00P400`01000:00009P030@/000<100002`050@/0
00<1000020080@H000<100402P000`400@0;00030@0100X000D100000@0:00050@0000402P001@40
000100/000@100010`020@T000<100001`020@T000<1000020000`40000700030@0000H00P490003
0@0000<01@4500030@0000801@4400030@0000800P5300040@000@<000@1000110000`4000020003
0@00008000<100000P000`40000900812`0002H00`4;00030@0000T00P4200040@000@X000<10000
20020@H03P4500030@0100/000<100402`000`400@0;00050@0000402P001040004=00030@010080
00<100002@000`40000700@11`000`40000600D11`000`400004008100@00@411P000`4000050081
1`000`40000400811@001`4000010040@`001040004200050@00004010000`40000200050@000040
1@000`40000800<12`0002H000<100000P4900030@0000L00P4400050@0000402@000`4000070004
0@000@/00P4601/11P000`400@0<00030@0100`000<100403P040@`000<100001@020@0400400081
1@000`40000400810P000`4000020@D000<100000P020@<000<100000P4400030@0000<00P401001
0@4500030@0000800`400`010@0300040@000@<000<10000@0001`40000100401000104000460003
0@0100H000<100001`001040004;00009P000`40000200811`000`40000500811P000`4000020081
20000`40000500050@0000402P001040004:00030@0100/000<100401@0?0@H000<1004030020@P0
2P4=00030@0000@000D100000@0400030@00008000<100000P020@@000<100000P020@<000<10000
0P4500030@0000800P4200040@0000810`000`4000020@<000<100000P4200040@0000<100<00@40
10000`40000o00030@0000800P4600030@0100H000<100401P000`40000600050@0000402`0002H0
00<1000010000`40000400030@0000@000<100001P000`40000400030@0000D000<1000010000`40
000300030@0000L000@100012P000`400@0;00030@0100/000<100401P0O0@L00P4010010@4:0003
0@0000<000<100000P000`40000300810P000`4000020@H000<1000010020@0400410@L000<10000
10030@03004100D000<100000P020@0400410@@000<1000010080D402P43008110050@L000<10000
10020@@000<100002@0002H000<100001@020@@000<100000P020@T000<100001@000`4000040003
0@0000800P4600030@0000H000<100000P000`40000600050@0000402@001@40000100X000<10040
30020@d00P4?00030@0100<00`4800050@0000401@000`40000500D120000`40000600812@000`40
000600811`000`40000400811P000`40000300@110020Cl00P4400810`090@@00`4500030@0000<0
00<1000010000`40000900009P000`40000700810P000`4000020@/000<100001P020@@000D10000
0@0800030@0000D000<100000`000`40000500030@00008000<100001`000`40000200030@0000H0
00D100000@0;00030@0100`000<100403@001040004600@1100010400@4700030@0000H000<10000
20000`40000500D11`000`40000400D11P000`40000300D11000104000020@8000<100400`000`40
0@0m00030@0100<000@100011@000`400@0700030@00008110001@40000100L000<100002@0002H0
00<100002@050@d000<1000020001@400001008000<1000020000`40000300030@0000D000<10000
10000`40000300030@0000D000<100000`000`40000600050@0000402P001040004;00040@000@d0
00D100000@0900<100<00@402@000`40000400D120000`40000400040@000@800P4500030@0000<0
00@100010P020@@000@100000P4200030@0000810P001040004400030@0100<000<100000P4j0004
0@000@<000@100011@000`400@0600030@0000<00P4200040@010@P000<100002@0002H00`4:0081
3P000`40000:00060@000@012`000`40000200030@0000H000<100000`000`40000400030@0000D0
00<1000010000`40000400030@0000<000<100001P000`40000200030@0000L000<100000P000`40
000900030@00008000<100002P030@T000<100000P020@8000<100000P4600030@0000800P430003
0@00008000<100000P00104000020@<000<100000P020@8000@1000110000`40000200D110001`40
000100401@000`40000f00810`000`400@0300030@00008000D100000@0500030@0000D0104:0003
0@0000T0000V00030@0000P00P4010010@4<00030@0000/0104=00040@000@X000<100000P000`40
000500030@0000<000<100001P000`40000300030@0000<000<100001@000`40000300030@0000L0
00<100000P000`40000800030@0000@000<100002@020@03004100811@000`4000020@@000<10000
0P020@@000D100000@0500030@0000<00P4200040@000@D000<1000010050@D000<1000010020@D0
00<100000P020@<0104f008100@00@400P4500050@0000400`000`40000300030@0000L000<10000
2@000`40000900009P000`40000600810`000`4000020@X000<1000030050@/000@100012P001@40
000100T000D100000@0900030@00008000<100001@000`40000400030@0000@000<100001@000`40
000400030@0000L000<1000010000`40000800040@000@@0104010010@4600030@0000@00P420004
0@000@H000<100001@050@H000<100001P020@L000<100000P020@030041008100@00@001`430081
=0000`400@0300<11@001040004500030@00008000<100001@050@T000<100002@0002H000<10000
10020@D000<100000P020@P000<100002@030@03004100802044008130000`400@0:00050@000040
2P001@40000100P000<100000P000`40000500030@0000@000<100001@000`40000600030@0000D0
00<10000200010400048008120000`40000600D11`000`40000700811`000`40000400<100<00@40
1@000`4000020@03004000@01043008110001040004b00040@000@<00P400`010@0300030@0100H0
00D100000@0600810P000`4000020@L000<100002@0002H000<100000P020@L000<1000010020@H0
00<100001`020@<000<100402P0<0@H000<100402`000`400@0<00030@0100T000D100000@090003
0@00008000<100001P000`40000600030@0000H000<100001P000`40000200030@0000D000<10040
1`000`40000800812@000`40000400<100<00@401@000`40000200810P00104000020@<000<10040
10000`400002008110001040004300040@000C400P4300050@0100400`030@03004100L000@10001
10020@@000<100000P020@D000<100002@0002H000<100000P4900030@0000H00P4400030@0000@0
0`4500040@000@`000<100401P0>0@H000<1004030000`400@0:00030@0100X000<100000P000`40
000700030@0000@000<100001`000`40000500030@0000<000<100001@001040004600030@0000H0
0`400`010@0700030@0000<000@100010`020@<000<100000P4400030@0000<00P400`010@050005
0@0000400P001@400001008000D100000@0300030@0002d00P4010010@020@8000<1000010030@L0
00@100010`000`40000400030@0000@00P4300030@0000T0000V00<12`000`40000800810P001040
00030@P000D100000@0;00030@0100`000<100401@0M0@D000<100402`001040004:00030@0000@0
00<1000020000`40000300030@0000@000<1000010000`40000200030@0000<000<100001@001040
004300811@00104000020@<000<100000`020@0400410@H000<100001@030@D000D100000@020004
0@000@@000@100010`020Bd000<100400`020@<000<100000`030@0300410081100010400@020@L0
00<100001P020@03004000X0000V00<12`000`40000:00<100<00@402P000`40000200030@0000P0
00D100000@0;00030@0100/000<1004030000`400@0500X12@000`400@0<00030@00008000<10000
2@000`40000200030@0000H000<100000`000`40000300030@00008000<1000010001@40000100D0
0P4300040@000@D000<100001@020@P000<1000010030@03004100<000@1000110000`400@050081
00D00@4100020B/000@100010`000`400@0200030@00008000@100011@080@T000<1000020020@/0
000V00<12`000`40000<008130000`40000300030@0000L000D100000@0:00050@0000402P000`40
0@0<00030@0100X00P4300L10`020@`000D100000@0=00040@000@T000<100000P000`4000050005
0@00004010020@D000<1000010020@0400410@H000<1000010050@L000<100000P001040004300<1
00<00@401@020@<01P4400030@0102T000<100000P000`400@0300030@0100<000D100000@090081
2`000`40000700<12`0002H000@100012P000`40000:00D12`000`40000300030@0000H000<10000
0`000`40000700050@0000402@001@40000100X000D100000@0900030@0100T01P4<00040@000@d0
00@100012@001@40000100T000@100010`000`40000500030@0000H00`4700030@0000<000@10001
0P020@D000D100000@0300030@0000@03@44008110001040004W00030@0000<000<100400`001040
0@020@@000<100001`040@T000<100001P001040004;00009P00104000020@P000<1000020020@80
00@100012@000`40000500030@0000D000<1000010000`40000500030@0000<000<100001@000`40
000300030@0000L000<100000P000`40000500040@000@/00P4200D11`000`400@0?00812`001040
004:00030@01008000<100001P000`400006008100<00@401@00104000020@<000<100000P020@<0
00@1004110000`40000400811`020@D000<100000P400`01000200<18P060@<00P4500<11P000`40
0005008100@00@000P4700030@0000@00P4300030@0000T0000V00030@0000<000<100001@000`40
0006008110001@40000100P000<100001P000`40000300030@0000D000<100001@000`4000030003
0@0000D000<100000`000`40000600030@0000<000<100001@001@40000100T000<100401`060@80
0P4700`12@000`400@0<00@12@000`400004008100<00@000P020@<000@100011@000`4000040081
00<00@401`000`40000200040@000@D000@1000110000`400002008100H00@400@040AX010420004
0@0000`120000`40000300050@00004010000`40000400030@0000<000<100000`000`4000090000
0`070@H00P4800@120000`40000400811@000`40000400811P000`40000200811@000`4000070003
0@00008000<100001`000`40000300030@0000D000<100000`000`40000500030@0000D000<10000
10000`40000300030@00008000<100001P001@40000100`02`4700810`070@<00P4600T12P000`40
000300050@0000401@020@0400410@H000<100001P030@L000<100000P4300030@0000<000@10001
0`000`40000300<10`000`40000300@14`030@03004100<000<100000P030@H0104600030@000080
00<100000P000`40000300811000104000020@H000<100002@0000H000<100001`020@H00P440003
0@0000D000<100001P020@<000<100000P020@P000<1000010000`40000200030@0000P000D10000
0@0:00050@0000402@000`40000200030@0000H000<100000`000`40000500030@00008000<10000
10000`400003008110000`40000;00<13P000`400@0900/11P020@0400410@L000<100000P000`40
000200030@0000D00P4900030@0000@00`4010010@4400030@0100H000D100000@0400040@000@80
0`4010010@4200030@0000<00`4010010@4<00@110001P400@400@800P4200030@0000<000<10040
0P020@@000D100000@0500030@0000D00P4200040@000@P000<100002@0000H000<100005P000`40
000400030@0000P000@100010P020@X000<100001@001@40000100/000<100402`001@40000100T0
00D100000@0900030@00008000<100001`001@40000100H000<100000P000`40000400030@0000X0
00D100400@0<00040@000@/00P4=00030@0100<0104400040@010@H000<100000`050@L000<10000
0`001040004400<100<00@401`001@40000100D00P400`010@0400040@000@8000<100000P420003
0@000080104400@10P020@@000@10000104400040@0000810P000`40000200810P0010400@460003
0@0000L01@4900030@0000T0000600030@0001H000<1000010040@P01@4<00030@0000H00`4<0003
0@0100`000<100402`000`400@0;00050@0000402P000`400@0800040@000@L000<100002@001040
004200030@0000P000D100000@0;00030@0100/000@100011`030@03004100P000@100000P420003
0@0000811@00104000020@<000<100001@040@L000<100401@040@H000@100001@4400030@000080
0P4200@11@001040004200030@0000800`4500040@000@@000<1000010040@P000<100002@000`40
000900030@0000T0000600030@0001H000<1000010000`40000:00813P000`40000801H11P000`40
0@0;00030@0100`000<100402`000`400@0800030@0100P000<1000020001@40000100<000<10000
1`000`40000200030@0000L000@100012P000`40000200030@0000L0104700040@000@@000<10000
0P020@<000@100011@000`40000500810P060@0500400@00104200<11@000`40000300811@000`40
00020@03004000<00P4500070@0000400@02008100<00@000P4300030@0100D000<100001P000`40
000700030@0000L01@4900030@0000T0000600030@0001H000<1000010000`400008008100@00@41
30000`400007008130000`400@0601d11P000`400@0501<12@000`40000700030@00008000<10000
0P000`40000500030@0000<000<100001P000`40000200030@0000L000<100000P000`4000070004
0@0100D10P030@D000<1000010020@0400410@H000<1000010001040004700D11P00104000430003
0@0000800`400`010@0300030@0100@000@100010P001@40000100D0104300040@00008100<00@40
1P000`40000400@11`000`40000700040@000@8000<100001P000`40000900001P000`40000@0081
0`000`40000500030@0000H00P4300040@000@/000<100001P001040004;00030@0100`000<10040
2`000`400@0600h11P000`400@0900<120000`40000500030@0000<000<100000`000`4000040003
0@0000<000<100001@000`40000300030@0000H000<100000`000`40000700030@0100H00`470003
0@0000H00`4800040@0000810`000`40000600<11`001@400001008000<100000P4200040@000081
00<00@401@001040004300040@000@@00`4500030@0000<00`4600030@0000800P4200030@0100H0
00<100001@020@<000@100000P4600030@0000T0000600030@00010000<100400P4800030@0000@0
0P4500040@0000812@000`40000500050@0000402P001@40000100/000<100402`000`400@0<0003
0@0100/000<1004020001@40000100L000<1000010000`40000400030@0000@000<100000P000`40
000500030@0000@000<1000010000`40000500030@0000@000<100001@001@40000100L00P460003
0@0000D00`400`010@0600040@000@D000<100001@0010400@020@D000<100000P050@@000<10000
0`030@D000<100401@020@03004100810P020@@000<100000P030@03004100@000<100000P440004
0@000@D000<1000010000`40000300030@0000<000<100000`000`40000900001P000`40000@0003
0@0000X000<100000P020@L000<100000`000`40000600030@0000@000<100000`000`4000060003
0@00008000<1000020001@40000100T000D100000@0;00030@0100X000D100000@0700030@000080
0P4500030@0000<000<100001P000`40000400040@000@P000<100000`000`40000500030@0000@0
00<100001@000`40000400030@00008000<1000010000`4000020@@000<100000`020@8000@10000
0`4300030@0100H000<1000010001040004300050@0000401P020@H000<100000P030@03004100<0
00<1004010040@D000@100010`001@400001008000@100000`4010010@4600040@0000810`000`40
000300030@0000@000<1000010020@<000<100002@0000H000<1000040000`40000:00030@000081
2@000`40000400811P000`40000200811P000`40000600030@0000<000<100001P000`4000020003
0@0000L000<100000P000`40000700050@0000402@000`40000200030@0000@000<1000010000`40
000200030@00008000<100001`000`40000500030@0100T000D100000@0900030@00008000<10000
1`000`40000300030@0000<000<100000`000`40000200050@00004010000`40000200030@0000@0
0P400`010@0800030@00008000D100000@0400040@000@<00`4010010@4400050@0000400P001040
00030@03004000@10P020@D000@100000P401@01000100<000<1000010020@L000<1000010001040
004300811`000`40000600040@000@/00003008100<00@004@000`40000:00<12`000`4000060003
0@0000<000D100000@0900030@0000@000<1000010000`40000500030@0000@000<100001@000`40
000300030@0000H000<100000`000`40000500030@0000@000<100000`000`40000500810P001@40
000100X000<100001P020@T000@100012P000`40000200030@0000L000<100000P000`4000040003
0@00008000<1000010001040004200811@000`40000600<120001@40000100@000<100000`030@03
004100@000<100000P4200040@000@<000<1000010030@H000<1004010000`40000300@110000`40
000300D11@000`40000500030@01008000<100001`000`40000700<12`0000D00P4B00H11`030@/0
00<100001`020@<000@100012P000`40000300030@0000H000<1000010000`40000400030@0000D0
00<1000010000`40000500030@0000<000<100001@000`40000400030@00008000<1000020001P40
00400@/000<1000010030@03004100811@001040004;00040@000@X000<100000P000`4000050005
0@0000401`050@L000<100001P020@030041008110001040004500030@0000@00`4600030@000080
1@4400030@0000<01@4500030@0000810P000`40000200<11P000`40000200040@000@800P430003
0@0000H0104:00030@0000T000<100002@0002H00`4:00030@0000X000H10001004<00030@000080
00<100001`000`40000200030@0000H000<100000`000`40000600030@0000<000<100001@000`40
000300030@0000H000D100000@0;00@130000`40000200810P000`40000300@10P020@`000<10040
30001040004800040@000@P0104900030@0000<00P400`01000400@100<00@401P000`4000020081
00@00@411@000`40000300<11P001@400001008000<100000P4200030@0000<00P400`010@020@80
0`4400050@0000400`000`400002008100<00@000`060@`000<100001`030@/0000V00030@000081
20000`40000;00@13@001@40000100X000D100000@0:00050@0000402@000`40000300030@0000D0
00<100000P000`40000800040@000@/01P4:00030@00008110000`40000700X11`020@`000<10040
2P000`400@0400@10P0010400@4700030@00008000<100000P000`40000600<11P000`4000020@<0
00@100010`000`400002008100<00@000P4400040@000@<000<100000P030@<01@4600030@000081
0P0010400@4400030@0000@01P4400@12P000`40000600040@000@/0000V00030@00008000<10000
1@000`40000<00H12`000`400@0<00030@0100/000D100000@0:00050@0000402@001@40000100/0
00<100401P050@800P4200L10`030@L000<100001`020@L03`4700812P020@03004100811P001040
00020@D000D100000@0500030@0000H00P400`010@020@800`4400030@00008000@100010P020@<0
00@100000`4010010@4400030@0000@0104200030@0100D000<100000`040@H000<1000010020@L0
00<100400P020@P000<1000010020@<000<100002@0002H000<100000`020@D000<100002@030@03
004100<0204300030@0100`000<1004030000`400@0<00030@0100/000<1004030020@801@460003
0@0100T0104900030@0000H000@1000130020@L05P4:00030@0000<00P4300040@010@H000<10000
10020@8000<100000P030@H000<100000`050@D000<1000010020@L000<100000P020@D000<10040
1@000`40000500811`000`40000200030@0100L000<1004010020@H000<100000`000`4000030003
0@0000T0000V00030@0000D000<100000P000`40000500@10`000`400@0;01/11P000`400@0<0003
0@0100/000<100401P0:0@/000@100012`020@P000<100001P001040004<00030@0100`000<10040
2`020@T000<100001P020@03004100P000<100000`000`40000200030@0000<000<100401@000`40
000400811`000`40000300D11@000`4000020@8000D100000@0300030@00008000<100000`030@03
004100D000<100000P4300030@0000<000D100000@0500811000104000020@H000<100002@0002H0
00<100001P020@8000<100000P030@L000@100013@000`400@0<00030@0100H0;04700812`000`40
000200030@0000L000@100011`000`40000500030@00008000<1000020001040004<00030@0100X0
00@1000120000`40000800<120000`40000200030@0000<000<100000`001040004400030@000080
0P4010010@4500030@00008000@100010P020@<000<1004010001@40000100@000D100000@040004
0@000@<00P4300030@0100D000<100000P000`40000200030@0000D00P4200040@000@P000<10000
2@0002H00`4800<100@00@412@000`40000200030@0000X000<1004030000`400@0<00030@0100`0
00<1004030020@`000<100402`000`40000200030@0000L000D100000@0600030@0000@000<10000
0`000`40000700030@00008000<1000020001@40000100T000D100000@0700030@0000L00`400`01
0@020@D000@100411P000`40000200030@0000800P400`01000200810`001040004400050@000040
0`000`400002008100<00@401@001040004600040@000@800P4300030@0000<00P400`010@060005
0@0000401@000`40000700D12@000`40000900009P000`40000:008130000`40000200030@0000T0
00D100000@0:00050@0000402`000`400@0<00030@0100/000<1004030001040004:00030@0000<0
00<100001@000`40000300030@0000<000<100000`000`40000500030@0000H000<100000P000`40
000800050@00004020000`40000300030@0000@000<100001P001040004500@100<00@402@001@40
000100H00P400`010@0500040@0000810P0010400@4400030@0000@00`4500030@0100L000<10040
0P000`40000300030@0000D00`4600040@000@H000<100002@000`40000900030@0000T0000V0003
0@0000P0104<00030@0000<000<100001`000`40000200030@0000P000D100000@0:00050@000040
2P001@40000100X000@100012P000`40000200030@0000H000<1000010000`40000500030@0000@0
0P4300030@00008000<100001P000`40000500030@0000@000<100001P000`40000300030@0000D0
00<1000010000`40000300030@0000@00P4400030@0000H00`4900040@000@L00`4700030@0000<0
104700030@0000800`400`010@020@800P4800@11P000`40000400<100@00@4110000`400@060003
0@0000L01@4900030@0000T0000V00030@0000H00P4200030@0100X000<1000010000`4000070003
0@0000<000<100001P000`40000300030@0000L000<100000P000`40000700030@00008000<10000
1P000`40000200030@0000L000<100000`000`40000500030@0000D000<1000010000`4000060004
0@000@<000<100001`000`40000400030@0000D000<100001@000`40000400030@0000@000<10000
1@000`40000300030@0000<000<1000010000`400006008100<00@400P4500030@0100@010401001
0@4600030@0000D00P4700050@0000400P000`40000300l120000`40000200040@000@@00`400`01
0@0700030@0000H000@100010P020@L000<100002@0002H000<1000010020@<000D100000@090003
0@0000D000<100001@000`40000400030@0000D000<100001@000`40000500030@0000<000<10000
1P000`40000300030@0000H000<100000`000`40000500030@0000@000<1000010000`4000070003
0@00008000<1000020001040004200030@0000P000<100000P000`40000700030@0000@000<10000
1@000`40000300030@0000H000<100000P000`40000200030@0000D000<100001@001040004400@1
0P060@D000@100011@000`40000300<100@00@4110001040004400030@0000<00P49008100<00@40
1P001@40000100<000<100001@020@P000<100001@001@40000100@000<1000010000`4000090000
9P000`40000200811@000`40000200030@0000D000<100001`000`40000300030@0000H000<10000
10000`40000500030@0000D000<1000010000`40000500030@0000@000<1000010000`4000040003
0@0000D000<100001@000`40000300030@0000L000<100000P000`40000900050@0100402`000`40
000200030@0000L000<100000`000`40000600030@00008000<10000200010400043008120000`40
000400050@00004020040@T000@100000P4300040@0000810P000`40000200<100@00@411@000`40
000200040@000@L000@100010P030@<000@1004110000`40000500@11P000`400003008110000`40
0003008110000`40000900009P000`4000020@L000<100000`000`40000400030@0000L000<10000
0`000`40000600030@0000<000<100001`000`40000300030@0000H000<100000`000`4000060003
0@0000<000<100001@000`40000300030@0000L000<100000P000`40000800040@000@d00`4=0004
0@000@/000D100000@0:00050@0000402`000`400@0200030@0000T000<100000P000`4000020003
0@0000H01@4800030@0000<000@100010P000`40000200030@0000D00`4600050@0000400`000`40
000500040@000@D00P400`010@0600030@0000@000<100400P020@D000D100000@0600030@0000D0
0P4200030@0000T0000V00<12@000`40000400030@0000<000<1000020001@40000100X000D10000
0@0;00050@0000402@000`40000200030@0000L000<100000P000`40000700030@00008000<10000
20001040004<00030@0100X0204;00030@0100`000D100000@0:00040@000@d0104<00030@000080
00<100000`000`40000400040@000@8000<100001@000`40000400D11@000`40000400<100@00@41
0`001040004500030@0000<000D100000@0700<11`00104000020@8000<100000P020@<000@10001
1`000`40000700<12`0002H00`4800030@0000H000D100000@0<00030@0100/000D100000@0;0005
0@0000402P001@40000100X000D100000@0:00040@000@/000@1000130020@<0204400040@0000P1
10020@d000<1004030000`400@0:00L130001@40000100H000<100000P020@<000D100000@050003
0@0000H00`4700030@00008000@1000110030@03004100H000<100000P000`40000200030@0000D0
0P400`010@0500040@000@@000<1000010020@0400410@P000<100002@000`40000900009P030@P0
00<100001`001040004<00030@0100`000<100403@000`400@0<00030@0100`000<100402`001040
004<00811`0;0@/000<100402P0<0@L000<1004030020@<0204300810P050@L000@100011`001@40
000100D000<100000P020@<000<1000010020@03004000811@001@40000100<000<100001@030@L0
00@100011P000`40000300030@0100<00P4300030@0100D000<100001P020@X000<100001`030@/0
000V00030@0000811P000`40000800811`0?0@H000<100403@000`400@0<00030@0100`000<10040
30020@L03`4700813@000`400@0=00811`0K0@/000<100401P060@800P4900<11P000`4000040004
0@000@<00`4300040@0000810`0010400@4400030@0000D00P4010010@4400040@000@H000<10000
0P0010400045008100<00@401P000`40000600@120000`40000500810P000`40000900009P000`40
0002008110000`40000600/11P000`400@0604`11P000`400@0<00030@0100`000D100000@0<0003
0@0100d000<1004030000`400@0<00040@000@`03P4800030@0000D00`400`010@0600030@0000<0
0P400`010@0600030@0000@000<1004010040@03004100L000D100000@0300030@0000D00P480003
0@0000<00P40100100020@H000<100000`020@@000<100002@0002H000<1000010020@0300400080
1P4200030@0100`000<1004030000`400@0=00030@0100`000<1004030000`400@0<00030@0100`0
00<1004030001040004;00050@0000402`001040004=00030@0100`000<1004030001@40000100`0
00<100402@020@H000<100001`030@P000<100001@030@L000@100000P4300030@0000D010460004
0@000@D000<100000`050@H000<100000P001@40000100@00P4400030@00008000<1000010000`40
000900009P000`40000600D120000`400@0;00050@0000402`000`400@0=00030@0100`000<10040
30000`400@0<00030@0100/000D100000@0:00030@00008000<100001`000`40000300030@0000P0
00D100000@0;00050@0000402P001@40000100X000<100000P000`40000900040@000@P000<10000
0P4400030@0000D00P4010010@4700030@0000@00`400`010@0500040@000@D000<100001@020@80
1@4200811P00104000020@8000<100000P4400050@00004010000`40000400810P000`4000020@L0
00<100002@0002H000<100001P000`400@0900050@0000402P001@40000100X000D100000@0;0005
0@0000402P001@40000100X000D100000@0:00050@0000402P000`40000200030@0000L000<10000
0P000`40000700030@0000<000<100001`000`40000200030@0000P000<100000P000`4000080005
0@0000402@000`40000300030@0000T000D100000@0700030@00008000D100000@0500810`001040
004500030@0000<00P4200040@0000810`000`400@0600030@0000@000<100401`040@H000@10001
10000`40000200810P0010400@4500030@0000H01@4900030@0000T0000V00<11@00104000490003
0@00008000<100001P000`40000300030@0000L000<100000P000`40000700030@0000<000<10000
1`000`40000200030@0000L000<100000P000`40000700030@00008000<100001P000`4000030003
0@0000H000<1000010000`40000600030@0000@000<100001P000`40000300030@0000L000<10000
0`000`40000600030@0000<000<100001P000`40000400030@0000L000<100000`000`4000030003
0@0000@000@100010`020@D000@100000P4300030@00008000<100000P000`400003008100<00@40
1`000`40000200810`000`400005008100<00@400P4200<11@000`40000400@11`000`4000080003
0@0000T000<100002@0002H000<1000010000`40000200030@0000D000<100000`000`4000060003
0@0000@000<100001@000`40000300030@0000L000<100000`000`40000600030@0000<000<10000
1P000`40000300030@0000H000<100000`000`40000600030@0000@000<100001@000`4000040003
0@0000D000<100001@000`40000500030@0000@000<100001P000`40000400030@0000D000<10000
10000`40000500030@0000D000<100001P000`40000400030@0000<000<100001@030@03004100L0
00<100000`001040004200811@000`40000500<11`001@40000100D000<1000010001040004400<1
1`000`40000600030@0000L000<100001@050@T000<100002@0002H000<100000`000`4000030003
0@0000D000<1000010000`40000400030@0000D000<100001@000`40000400030@0000D000<10000
1@000`40000500030@0000@000<100001@000`40000400030@0000D000<1000010000`4000040003
0@0000H000<1000010000`40000500030@0000@000<100001P000`40000300030@0000H000<10000
10000`40000600030@0000@000<100001@000`40000400030@0000H000<1000010000`4000060003
0@00008000<100001P030@T000<1000010050@L000<100001@020@0400410@D000<100401P000`40
000300050@0000401@000`400@0600030@0000@01@4700030@0000@000@100010P000`4000060003
0@0000T0000V00030@00008000<100001@000`40000300030@0000H000<100000`000`4000060003
0@0000<000<100001P000`40000300030@0000L000<100000`000`40000600030@0000<000<10000
1P000`40000300030@0000H000<100000`000`40000700030@00008000<100001`000`4000020003
0@0000L000<100000`000`40000600030@0000<000<100001`000`40000300030@0000H000<10000
0`000`40000700030@0000<000<1000020001040004600<100@00@411`000`40000600812@000`40
0003008100<00@000`040@03004100L000<100000P000`40000200030@0000<000@1000110000`40
000400040@000@800P4500030@0000800P4300040@0000811P000`40000900009P001@40000100T0
00D100000@0900030@00008000<1000020001@40000100T000<100000`000`40000700030@000080
00<100001`000`40000200030@0000L000<100000P000`40000700030@00008000<1000020000`40
000200030@0000L000<100000P000`40000800050@0000402P000`40000200030@0000P000D10000
0@0:00050@0000402P000`40000200030@0000X000<100400`030@@000<100000P4500030@0000@0
0P4010010@4700030@00008000D100000@0800<11`001@40000100D000<100000P000`4000020005
0@00004010020@<000<100000P000`40000200050@0000401@000`40000300030@0000<000<10000
2@0001L0204700040@000@X000D100000@0:00050@0000402P001@40000100X000D100000@0;0005
0@0000402P001@40000100X000D100000@0:00050@0000402`001040004;00040@000@/000D10000
0@0:00050@0000402`001@40000100X000D100000@0;00050@00004030020@0400410@H000<10000
0`000`40000200030@0000800P4300030@0000811@001@40000100@000<100001P020@0400410@D0
00<100401`001@40000100D000@100010`000`40000300030@0000<00P4200040@000@H000<10000
10020@<000<100002@0001P000<100000`000`40000500030@0100`000<1004030000`400@0<0003
0@0100`000<100403@000`400@0<00030@0100`000<1004030000`400@0=00030@0100`000<10040
30000`400@0<00030@0100d000<1004030000`400@0<00040@000@X01P4900030@0000@00P420003
0@0000811@000`40000200810`0010400@4500030@0000D000<1004010040@03004100P000@10001
1`000`400@0200030@0000@000<100001@050@L000<100001P001040004;00006@000`40000:0081
3@000`400@0<00030@0100`000<1004030000`400@0=00030@0100`000<1004030000`400@0<0003
0@0100d00P4=00813@000`400@0<00030@0100d000<1004030000`400@0=00810`080@@00`480003
0@0000H01@4700030@0000@00P400`010@0800030@0000<000D100000@0700@11`000`400@0900@1
20000`40000500812@000`40000700<12`0001X000<100002@3Z0@`000@100011`000`4000070081
2@000`40000600<12000104000020@@000<100001@020@801P42008110080@T000<100000`020@04
00410@L000<1000020020@/0000K00030@0000P000<10000hP030@d000<100000P000`4000040003
0@0000D00P4010010@4700030@0000D00`400`010@0600040@000@H000<1000010001040004700P1
1@020@03004100811@00104000020@<000@100011P000`400006008100<00@002P0001`000<10000
1`000`40003R00040@000@`000<100000`020@@000<100000P030@<000<100000P4500030@0000<0
0P4200040@0000<10`000`400@0700030@0000<000D100000@0800812P000`400@0400@100<00@00
0P4500050@0000401@000`40000400810`000`40000900007@000`40000600030@000>8000D10000
0@0;00030@0000D000@100010`020@H000<100000P000`40000200030@00008000<100000P000`40
0004008100<00@4020000`40000200030@00008000<100001@001040004800040@000@P00`460003
0@0000<00P4300030@0000800P4500030@0000T0000N00030@0000D000<10000hP000`4000020081
2@000`40000600<100<00@4020000`40000300810P000`4000020@D000<100001P030@P000D10000
0@0600030@0000@000@1000120001@40000100P00P4600030@0000D000@100010P020@L000<10000
2@0001P000<100000`000`40000500030@000>8000<1000010000`40000500030@0000P00`4:0003
0@0000D01@4700030@0000H00P400`010@020@D000<100401`000`40000300030@00008000<10000
10000`40000200030@0000D000@100011@000`40000600D12@000`400009000060000`4000030003
0@0000D000<10000hP000`40000500811@000`400006008100@00@412@000`40000700812@000`40
0003008100<00@0010040@03004100T000D100000@0500030@0000@000<100000`000`4000040005
0@00004010000`40000700812`000`40000900006@001@40000100P000<10000hP000`4000070003
0@00008000<100000`030@<000@100011`000`40000600<100<00@401`000`40000200030@000080
00<100001P030@T000@100011`000`40000200030@0000@000<100000`000`40000200030@000080
00<100001@020@03004100X000<100002@0001X00`4900<1hP000`40000800050@0000400P030@H0
00@100000P4500030@0000@00P4200040@0000811@001@40000100D000<100001P020@0300410081
1@001040004700050@00004020000`40000200030@0000<000D100000@0400<10`0010400@480003
0@0000T0000V00030@000>8000<100002@030@03004100T000<100000`020@<000<100000P020@@0
00<100000`020@<000@100411P000`40000500030@0100D0104200812@001040004800050@000040
1`000`400@0300811P001@40000100L000<100002@0002H000<10000hP000`40000:00<12`000`40
000500040@000@800P4600030@0000D00P400`010@0800030@0000@000D100000@0800D120000`40
0@0:00040@000@P00P400`010@0700030@0000<000<1000010000`40000900009P000`40003R0003
0@0000L00`4010010@4:00030@0000H01@4800030@0000L00`4800030@0000800P4400030@0000L0
0P4200H10`020@X000<100401`050@T000<1000010020@@000<100002@0002H000<10000hP000`40
000400<110001040004800030@0000P00P4:00030@0000H00`4010010@4500050@0000401P000`40
000600030@0100P02P46008100<00@40104300812@000`40000600050@0000402`0002H000<10000
hP000`40000200811`00104000020@H000<100001P020@0400410@P000<100001@00104000440081
10000`400@0700030@0000D000D100000@0900811P090@P000@1000120000`40000700040@000@/0
000V00030@000>8000<100400P4900030@0000<000<100000`000`40000400810`000`4000020@H0
00<100000`020@<000<1000010030@03004100T000<100000`000`40000200030@0000H000@10001
2P000`400@0900050@0000401`000`40000800<12`0002H00`7R008130000`40000400810`000`40
000200811@000`40000200811000104000020@D000<100001`030@T000<100000P000`4000030003
0@0000D000D100000@0:00030@0100P000<100000`000`40000400030@0000T00P4;00009P000`40
003R00<12`000`40000600040@000@800P4700030@0000@00P4200040@000@L000<100001`020@04
00410@H000D100000@0600030@0000@000<100000`000`40000600050@0000401`000`4000040003
0@0000<000<100001`020@03004000X0000V00030@000>8000@100012P000`40000700D12@000`40
000600D120000`400005008100<00@000`030@<000@1000120000`40000300030@0000<000<10000
1@000`40000300030@0000@000<100001@000`40000200030@0000D00P4300030@0000T0000V0003
0@000>8000@100000P4700030@0000T00P4;00030@0000P00P4:00030@0000<000D100000@0700<1
00<00@402@000`40000200030@0000@000<1000010000`40000400030@0000<000<100001`001040
004500<11@000`40000900009P000`40003R00030@0000<000<1000010000`400007008100@00@41
2@000`40000600<100<00@4020000`40000200030@0000<000<100001`030@T000D100000@080003
0@0000<000<100001@000`40000200030@0000P000<100400`020@P000<100002@0002H000<10000
hP000`40000400030@0000<000<100001@020@<000<100000P4700030@0000@00P4200040@000081
1P001@40000100H000<100001`020@03004100811@001040004900030@00008000<100001P000`40
000200030@0000T00P400`010@0:00030@0000T0000V00030@000>8000<100001@020@<000<10000
0`020@D000<100000P020@D000<100000P020@@000<100000`030@<000@100411`000`4000050081
00<00@0010040@800P4;00040@000@X000@100012`040@`000<100002@0002H000<10000hP000`40
000700040@000@<00P4700030@0000@00P4300030@0000811P000`400006008100<00@402@000`40
000400050@0000402@060@T000<1004030000`400@0600D10P030@/000<100002@0002H00`7R0003
0@0000P00`400`010@0900030@0000H00P4010010@4800030@0000P00`4900030@0000<000<10000
0`000`40000700810`070@<00P4<00810P050@L000@100012P000`40000900009P000`40003R0003
0@0000X00P4;00030@0000P00P4:00030@0000L00`4010010@4600030@00008000<1000010000`40
000600040@000@T05@4;00030@00008000<100001`000`40000900009P000`40003R00030@0000P0
1@4:00030@0000L01@4900030@0000D000@1000110030@@000@100411`000`40000500050@000040
2`000`400@0=008130000`40000300811`000`40000900009P000`40003R00030@0000H00P420003
0@00008120000`40000500810P000`4000020@L000<100000`020@<000<100001@030@03004100T0
00<100000`020@@000<1000020001040004<00040@000@/000<100001@000`40000400030@0000T0
000V00030@000>8000<1000010020@@000<100000P020@H000<100000`020@@000<100000P020@D0
00@100000P4500030@0000P00`4900030@00008000<1000010000`40000800050@0000402`001@40
000100X000<100001P020@@000<100002@0002H000<10000hP000`40000200811P000`4000040081
10000`40000200030@0000@000<1000010020@<000@100011`000`400008008100@00@411P001@40
000100P000<100001P000`40000200030@0000P000<100000P000`40000800030@0000P000D10000
0@0;00009P000`40003R00030@00008120000`40000600810P000`4000020@L000<100001P020@04
00410@P000<100001P020@03004000<00`4400030@0100T000<100001@000`40000400030@0000L0
00<100000`000`40000700030@0000T000@100012`0002H000<10000hP030@X000<1000020050@T0
00<1000020020@/000<1000010001@40000100L00`400`010@0:00030@0000@000<100001@000`40
000600030@0000D000<100001@000`40000;00<12`0002H00`7R00030@0000X000<100002P000`40
000900030@0000L01@4900030@0000<000<100000`000`40000700<12P000`40000400030@0000H0
00<100001@000`40000600030@0000@000<1000030020@/0000V00030@000>800`4:00030@0000P0
1@4900030@0000H000@100010P020@L000<100000P000`40000400030@0000L00P400`010@090003
0@00008000<1000020000`40000300030@0000P000<100000`000`40000:008100<00@002P0002H0
00<10000hP000`4000020@P000<100001P020@8000<100000P4700030@0000@00P4300030@000080
0P4500050@0000401`000`40000600030@0100<00P4700050@0000402`000`40000300030@0000T0
00<100000P000`40000700<10`000`40000900009P000`40003R00030@0000800P4600030@0000@0
0P4400030@00008000<1000010000`40000300030@0000<000<1000010020@<000@1004120000`40
000500040@000@D00`4400040@000@d000D100000@0=00050@0000401`020@H000<100002@000080
2046008120040@P000<10000hP000`400004008110000`40000200811P000`400003008110001040
00020@H000<100001P020@03004100X000<100000`020@<000<100001P020@8000@100013@001@40
000100d000D100000@05008120000`40000900000`000`40000300030@0000@00P46008110000`40
000500030@000>8000<100001P020@8000<100000P4800030@0000D00P4200040@000@P000<10000
20030@X000<100000P000`40000300030@0000P00P400`010@0?00030@0100l000<100400`030@X0
00<100002@0000@000<1000060000`40000400030@000>8000<1000020050@X000<100001`050@T0
00<100001`030@0400410@P000@100011`000`40000900X120000`400@0@008100<00@403@000`40
000900001@000`40000G00030@0000@0107Q00030@0000X000<100002P000`40000900812P000`40
000500810P000`40000200@110000`400@0800030@0000L00P400`010@0701h13`000`4000090000
1P000`40000F00030@0000@000<10000hP000`40000800D12@000`40000800<100<00@4020000`40
0003008110000`40000600<100<00@402@000`40000400<10P000`400@0?00814@020@03004100d0
00<100002@0000L000<100005@000`40000400030@000>8000<100001P020@8000@1000120000`40
000600810P00104000020@H000<100000P000`40000400030@0000T0104800040@0000<11@001040
004>00030@0100l000<100400`020@/000<100002@0000P000<100003P020@<000<100001@000`40
003R00030@0000D000<100000P00104000020@H000<100001@000`40000200030@0000<00`430003
0@0000811`000`40000900810P060@8000@100411`000`40000200030@0000X000@100013P001@40
000100@00P4900030@0000T0000900030@0000d000<100400P4800030@000>8000<100000`020@D0
00<100000`000`40000300030@0000<00P4500030@0000H00P4010010@4900030@0000L00P400`01
000700@12@000`40000200030@0000T000<100000P000`40000;00050@0000401P020@L000<10000
2@0000<000<100000`000`40000=00030@0000X000<10000hP000`40000200030@0000D000<10000
10020@<000<100000P000`40000500030@0000P00P4;00030@0000H000<100000P000`4000060003
0@0100T000<100000`000`40000800030@00008000<100002P000`40000200030@0000H00P450003
0@0000T0000300030@0000<000<100003@000`40000:00030@000>8000<100000P4800030@0000H0
00@100010P020@P000<100001`050@X000<1000010000`40000300030@0000D000@1000120000`40
000400030@0000L000<1000010000`40000800030@0000<000<1000020020@<000<100002@0000@0
00D100000@0@00030@0000X000<10000hP030@X000<100001`050@X000<100001P00104000420081
20000`40000300030@0000@000<100001@001@40000100L000<100001@000`40000500030@0000D0
00<100001`000`40000500030@0000T00P400`01000:00001@030A401P4700<1hP000`4000090003
0@0000X00P4;00030@0000@00P4300030@0000800P4600030@00008000<100001@000`4000040003
0@0000<000<100000`000`40000700030@0000<000<100001`000`40000500030@0000H000<10000
2`020@/0000V00030@000>800`4900030@0000P00`400`010@0900030@0000<000<100000`000`40
0004008110000`4000020@T000<100000P000`40000400030@0000<000<100001`000`4000030003
0@0000L000<1000010000`40000700030@0000X00`4;00009P000`40003R00030@0000811`000`40
000600810P00104000030@H000@100000P4600030@0000H00P4200030@0100/000D100000@080003
0@00008000<1000020001@40000100/000<100000P000`40000800030@0000T000@100012`0002H0
00<10000hP000`40000200030@0000@000<1000010020@@000<1000010030@<000@1000120000`40
000800@130001040004:00040@000@`000<1004030000`40000200030@0000T000<100001P020@<0
00<100002@0002H000<10000hP000`400003008110000`40000200811P000`400007008100@00@41
20000`40000:00D12P001040004;00030@0100`000<100403@001040004<00030@0000D000<10000
0`000`40000900009P000`40003R00030@0000D00P4200030@00008120000`40000900<12@000`40
000700D10`080@<00P4<00811`0?0@L000<100403@000`40000400030@0000@000<100002@0002H0
00<10000hP000`40000700D12P000`400009008100@00@411P000`40000400<10`000`400@0:01X1
1P000`400@0600/130000`40000200811`000`40000900009P000`40003R00030@0000T000<10000
2P000`40000800030@0100@00P4400030@0000800P4500040@000@`000<1004030000`400@0<0003
0@0100d00P4200D120001040004900030@0000T0000V00<1hP000`40000700D12P000`4000060081
0`000`40000300<100<00@000`4700050@0000402`000`400@0<00030@0100/000D100000@0<0003
0@0100H01P4200<12P000`40000900009P000`40003R00030@0000H000@100010P000`4000070003
0@0000D000<100000`000`40000600<12@000`40000200030@0000P000D100000@0:00050@000040
2P001@40000100/000@1000130040@/000<100002@0002H000<10000hP000`40000400810`001040
00020@L000<1000010000`40000400030@0000H000<100402@000`40000300030@0000H000<10000
0P000`40000700030@00008000<100001`000`40000300030@0000L000<100000P000`40000:0081
00<00@402@000`40000900009P000`40003R00030@0000<000<100000P000`40000400030@0000@0
00<100000`000`40000500030@0000D000@1000120000`40000400030@0000H000<100000`000`40
000600030@0000<000<100001@000`40000400030@0000H000<100000`000`40000900030@0100<0
0P4700030@0000T0000V00030@000>8000@100000P4500030@0000D000<100000`000`4000020003
0@0000L000<100000`000`40000200030@0000@000<100001P000`40000400030@0000@000<10000
1@000`40000400030@0000D000<100001@000`40000500030@0000@000<100001`00104000450081
1@000`40000900009P000`40003R00040@000@L000<100001P020@<000<100000P4:00030@000080
00<1000010000`40000300030@0000H000<100000`000`40000600030@0000<000<100001P000`40
000300030@0000H000<1000010000`40000500030@0000H000<100000P000`40000400810`000`40
000900009P000`40003R00<120000`40000800060@000@0130000`40000200030@0000@000<10000
0P000`40000800030@00008000<100001`000`40000200030@0000L000D100000@0:00030@000080
00<100001`000`40000400030@0000<000<100001P020@03004000X0000V00<1hP000`4000080003
0@0000T0104=00050@00004020001@40000100X000D100000@0:00050@0000402P001@40000100X0
00D100000@0:00030@0000<000<1000010000`40000800812`0002H000<10000hP030@P000<10000
2P050@`000<100402P000`400@0<00030@0100`000<1004030000`400@0<00040@000@/000D10000
0@0700030@0000L00`4;00009P000`40003R00030@0000811@000`40000700H10`080@@00P4;0003
0@0100`000<1004030000`400@0<00030@0100`000<1004030001040004800030@0000D00P420003
0@0000T0000V00030@000>8000<100000P020@<000<100000`040@@000<100402P1F0@L00P490003
0@0000<00P4400030@0000T0000V00030@000>8000<1000010020@03004000@11`001040004=0081
30020@d00P4=00030@0100`000<1004030000`400@0600X12@000`4000020@H000<100002@0002H0
00<10000hP000`40000600<12P000`40000200030@0000T000@100012`000`400@0<00030@0100`0
00<1004030000`400@0<00030@0100`000@100401P4300<120000`40000900009P000`40003R0003
0@0000H00P4:00030@0000<000<100002@001040004:00040@000@/000@100012`001@40000100X0
00D100000@0:00050@0000402`000`400@0700@12P000`40000900009P000`40003R00030@0000D0
00@100012@000`40000400030@0000L000<100000P000`40000600030@00008000<100001`000`40
000200030@0000L000<100000P000`40000700030@00008000<100001`000`40000200030@0000P0
00D100000@0800030@0100T000<100002@0002H00`7R00030@0000@000D100000@0800030@0000D0
00<100001P000`40000300030@0000D000<100000`000`40000600030@0000<000<100001P000`40
000400030@0000H000<100000`000`40000600030@0000<000<100001P000`40000300030@0000D0
00<100000P4700030@0000T0000V00030@000>8000<100000`000`40000300030@0000@000<10000
1P000`40000500030@0000@000<100001@000`40000400030@0000D000<1000010000`4000050003
0@0000@000<100001@000`40000400030@0000D000<1000010000`40000500030@0000@000<10000
10000`40000300030@0000@000<100002@0002H000<10000hP000`40000200030@0000@000<10000
10000`40000700030@0000<000<100001P000`40000300030@0000D000<1000010000`4000050003
0@0000@000<100001P000`40000300030@0000H000<100000`000`40000600030@0000@000<10000
1@000`40000300030@0000@000<100000`000`40000900009P000`40003R00050@00004020000`40
000200030@0000P000<100000P000`40000700030@00008000<100001`000`40000200030@0000L0
00<100000P000`40000800030@00008000<100001`000`40000200030@0000L000<100000P000`40
000600030@00008000<100001P000`40000200030@0000T0000I00@12@000`40003R00040@000@T0
00D100000@0<00050@0000402P001040004:00050@0000402P001@40000100/000D100000@0:0005
0@0000402P001@40000100X000D100000@0900050@0000402`0001P000<100000P000`4000060003
0@000>8000<100402`000`400@0=00040@000@/000@100012`001040004;00040@000@`000<10040
30000`400@0<00030@0100`000<100402`020@03004000X0000N00030@0000D000<10000hP020@`0
00<100403P020@d00P4<00030@0100`000<100403@000`400@0<00030@0100`000<1004030000`40
0@0=00812`0001h000<100001@040N40VP4;00007P000`40000500030@000?l0R00001d000<10000
1P000`40003o08P0000J00<12@000`40003o08P0000M00030@0000H000<10000o`2800007P000`40
000500030@000?l0R00001h000<100001@000`40003o08P0000H00030@00008000<10000o`2A0000
6@040Ol0U0000001\
\>"],
  ImageRangeCache->{{{0, 431}, {258.125, 0}} -> {-0.508062, -0.278011, 
  0.0196661, 0.0196661}}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Computations", "Subsection"],

Cell[BoxData[{
    \(\(gD[x_, y_] = 
      Which[3.5 \[LessEqual] x \[LessEqual] 4.5 && 
          0.5 \[LessEqual] y \[LessEqual] 1.5, 100, True, 0]; \)\), 
    \(\(a = Function[{x, y}, 1]; \)\), 
    \(\(b = Function[{x, y}, 0]; \)\), 
    \(\(f = Function[{x, y}, 0]; \)\), 
    \(\(gDirichlet = Function[{x, y}, gD[x, y]]; \)\), 
    \(\(gNeumann = Function[{x, y}, 1]; \)\), 
    \(\(points = 
      FEMSolve[a, b, f, gDirichlet, gNeumann, nodes, elements, segments]; 
    \)\)}], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["A 3D Plot ", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(g = FEMSurface[points, elements]; \)\), 
    \(\(solution = 
      Show[g, \n\t\tAxes \[Rule] True, 
        AxesLabel \[Rule] {"\<x\>", "\<y\>", "\<u\>"}, AspectRatio \[Rule] 1, 
        PlotRange \[Rule] All, ViewPoint \[Rule] {1.582, \(-2.9\), 0.647}]; 
    \)\)}], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12  scalefont  setfont
% Scaling calculations
0 1.007 -1.29925 3.74173 [
[.02012 .31927 -3.48107 -10.8 ]
[.02012 .31927 3.71893 0 ]
[.29685 .18904 -3.54904 -10.8 ]
[.29685 .18904 3.65096 0 ]
[.60677 .04321 -3.62554 -10.8 ]
[.60677 .04321 3.57446 0 ]
[.37156 -0.0797 -5.54294 -11.375 ]
[.37156 -0.0797 5.64456 0 ]
[.00732 .7952 -3.72102 0 ]
[.00732 .7952 3.47898 10.8 ]
[.10546 .88495 -3.69692 0 ]
[.10546 .88495 3.50308 10.8 ]
[.19484 .96668 -3.67505 0 ]
[.19484 .96668 3.52495 10.8 ]
[.27656 1.04142 -3.6551 0 ]
[.27656 1.04142 3.5449 10.8 ]
[.14967 1.16054 -5.79094 0 ]
[.14967 1.16054 5.52156 11.375 ]
[-0.00919 .38824 -7.2 -4.71937 ]
[-0.00919 .38824 0 6.08063 ]
[-0.01029 .50705 -21.6 -4.75471 ]
[-0.01029 .50705 0 6.04529 ]
[-0.0114 .62638 -21.6 -4.79022 ]
[-0.0114 .62638 0 6.00978 ]
[-0.01252 .74624 -21.6 -4.82591 ]
[-0.01252 .74624 0 5.97409 ]
[-0.07332 .58937 -11.1875 -5.02518 ]
[-0.07332 .58937 0 6.34982 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.00338 .37369 m
.79826 0 L
s
.01971 .36602 m
.0195 .38939 L
s
[(0)] .02012 .31927 -0.03303 1 Mshowa
.29667 .23581 m
.29658 .25919 L
s
[(2)] .29685 .18904 -0.01416 1 Mshowa
.60686 .08998 m
.6069 .11337 L
s
[(4)] .60677 .04321 .00709 1 Mshowa
.125 Mabswid
.08613 .33479 m
.08602 .34881 L
s
.15438 .3027 m
.15429 .31673 L
s
.22453 .26972 m
.22446 .28375 L
s
.37088 .20092 m
.37084 .21495 L
s
.44725 .16502 m
.44723 .17905 L
s
.52587 .12806 m
.52588 .14209 L
s
.69031 .05075 m
.69036 .06478 L
s
.77634 .01031 m
.77641 .02434 L
s
gsave
.37156 -0.0797 -66.5429 -15.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(x) show
70.187500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
0 .74137 m
.2826 1 L
s
.00774 .74845 m
.00796 .72508 L
s
[(0)] .00732 .7952 .03362 -1 Mshowa
.1058 .8382 m
.10597 .81482 L
s
[(1)] .10546 .88495 .02692 -1 Mshowa
.1951 .91992 m
.19523 .89654 L
s
[(2)] .19484 .96668 .02085 -1 Mshowa
.27675 .99465 m
.27685 .97127 L
s
[(3)] .27656 1.04142 .01531 -1 Mshowa
.125 Mabswid
.02812 .7671 m
.02824 .75308 L
s
.0481 .78539 m
.04822 .77136 L
s
.0677 .80333 m
.06781 .7893 L
s
.08693 .82093 m
.08704 .8069 L
s
.12432 .85515 m
.12442 .84112 L
s
.1425 .87179 m
.1426 .85776 L
s
.16035 .88812 m
.16044 .8741 L
s
.17788 .90417 m
.17797 .89014 L
s
.21201 .9354 m
.21208 .92137 L
s
.22862 .9506 m
.22869 .93657 L
s
.24494 .96554 m
.24501 .95151 L
s
.26099 .98022 m
.26105 .96619 L
s
gsave
.14967 1.16054 -66.7909 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.125000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(y) show
70.312500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
.00338 .37369 m
0 .74137 L
s
.0033 .38239 m
.00954 .37946 L
s
[(0)] -0.00919 .38824 1 -0.12604 Mshowa
.0022 .5015 m
.00845 .49872 L
s
[(0.2)] -0.01029 .50705 1 -0.1195 Mshowa
.0011 .62113 m
.00736 .61851 L
s
[(0.4)] -0.0114 .62638 1 -0.11292 Mshowa
0 .74129 m
.00626 .73882 L
s
[(0.6)] -0.01252 .74624 1 -0.10631 Mshowa
.125 Mabswid
.00303 .41212 m
.00677 .41038 L
s
.00275 .44188 m
.0065 .44017 L
s
.00248 .47167 m
.00623 .46999 L
s
.00193 .53136 m
.00568 .52972 L
s
.00166 .56125 m
.00541 .55963 L
s
.00138 .59118 m
.00513 .58958 L
s
.00083 .65112 m
.00458 .64957 L
s
.00055 .68115 m
.00431 .67962 L
s
.00028 .7112 m
.00403 .7097 L
s
gsave
-0.07332 .58937 -72.1875 -9.02518 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(u) show
70.187500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
.00338 .37369 m
0 .74137 L
s
0 .74137 m
.2826 1 L
s
.2826 1 m
.28388 .67992 L
s
.28388 .67992 m
.00338 .37369 L
s
.79826 0 m
.99671 .40534 L
s
.99671 .40534 m
1 .76812 L
s
1 .76812 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.00338 .37369 m
0 .74137 L
s
0 .74137 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.79826 0 m
.00338 .37369 L
s
.28388 .67992 m
.99671 .40534 L
s
.99671 .40534 m
1 .76812 L
s
1 .76812 m
.2826 1 L
s
.2826 1 m
.28388 .67992 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.5 Mabswid
.21767 .56799 m .23719 .57651 L .21234 .58667 L closepath p .41 .805 .986 r
F P
0 g
s
.21767 .56799 m .21234 .58667 L .19533 .56796 L closepath p .41 .805 .986 r
F P
0 g
s
.2356 .55712 m .26229 .56624 L .23719 .57651 L closepath p .409 .805 .985 r
F P
0 g
s
.2356 .55712 m .23719 .57651 L .21767 .56799 L closepath p .407 .804 .985 r
F P
0 g
s
.21767 .56799 m .19533 .56796 L .20803 .55132 L closepath p .407 .804 .985 r
F P
0 g
s
.19533 .56796 m .17801 .54892 L .20803 .55132 L closepath p .407 .804 .986 r
F P
0 g
s
.2356 .55712 m .21767 .56799 L .20803 .55132 L closepath p .408 .804 .986 r
F P
0 g
s
.28763 .55588 m .26229 .56624 L .25995 .5459 L closepath p .406 .805 .984 r
F P
0 g
s
.25995 .5459 m .26229 .56624 L .2356 .55712 L closepath p .404 .804 .984 r
F P
0 g
s
.2356 .55712 m .22875 .53405 L .25995 .5459 L closepath p .405 .804 .985 r
F P
0 g
s
.22875 .53405 m .2356 .55712 L .20803 .55132 L closepath p .404 .803 .985 r
F P
0 g
s
.28638 .53588 m .31322 .54541 L .28763 .55588 L closepath p .403 .805 .983 r
F P
0 g
s
.28763 .55588 m .25995 .5459 L .28638 .53588 L closepath p .401 .803 .983 r
F P
0 g
s
.22875 .53405 m .20803 .55132 L .19313 .53118 L closepath p .405 .803 .985 r
F P
0 g
s
.20803 .55132 m .17801 .54892 L .19313 .53118 L closepath p .405 .803 .985 r
F P
0 g
s
.17801 .54892 m .16038 .52953 L .19313 .53118 L closepath p .405 .802 .985 r
F P
0 g
s
.31252 .52591 m .33907 .53484 L .31322 .54541 L closepath p .399 .805 .981 r
F P
0 g
s
.31252 .52591 m .31322 .54541 L .28638 .53588 L closepath p .396 .803 .981 r
F P
0 g
s
.25944 .52536 m .28638 .53588 L .25995 .5459 L closepath p .402 .803 .983 r
F P
0 g
s
.25944 .52536 m .25995 .5459 L .22875 .53405 L closepath p .4 .802 .983 r
F P
0 g
s
.2859 .51607 m .31252 .52591 L .28638 .53588 L closepath p .397 .803 .982 r
F P
0 g
s
.2859 .51607 m .28638 .53588 L .25944 .52536 L closepath p .396 .801 .982 r
F P
0 g
s
.36518 .52417 m .33907 .53484 L .3386 .51588 L closepath p .393 .804 .979 r
F P
0 g
s
.3386 .51588 m .33907 .53484 L .31252 .52591 L closepath p .39 .801 .979 r
F P
0 g
s
.23694 .51413 m .25944 .52536 L .22875 .53405 L closepath p .401 .802 .984 r
F P
0 g
s
.23694 .51413 m .22875 .53405 L .20878 .51206 L closepath p .4 .801 .984 r
F P
0 g
s
.19313 .53118 m .20878 .51206 L .22875 .53405 L closepath p .402 .802 .984 r
F P
0 g
s
.19313 .53118 m .17507 .51029 L .20878 .51206 L closepath p .402 .802 .985 r
F P
0 g
s
.17507 .51029 m .19313 .53118 L .16038 .52953 L closepath p .403 .802 .985 r
F P
0 g
s
.17507 .51029 m .16038 .52953 L .14242 .50977 L closepath p .403 .802 .985 r
F P
0 g
s
.69079 .50771 m .7088 .51518 L .68086 .52594 L closepath p .37 .8 .968 r
F P
0 g
s
.69079 .50771 m .68086 .52594 L .66769 .50613 L closepath p .367 .782 .979 r
F P
0 g
s
.3386 .51588 m .31252 .52591 L .31218 .50669 L closepath p .392 .802 .98 r
F P
0 g
s
.31218 .50669 m .31252 .52591 L .2859 .51607 L closepath p .389 .8 .98 r
F P
0 g
s
.36456 .5058 m .39155 .51338 L .36518 .52417 L closepath p .386 .803 .976 r
F P
0 g
s
.36518 .52417 m .3386 .51588 L .36456 .5058 L closepath p .381 .799 .975 r
F P
0 g
s
.25941 .50524 m .2859 .51607 L .25944 .52536 L closepath p .397 .801 .982 r
F P
0 g
s
.25941 .50524 m .25944 .52536 L .23694 .51413 L closepath p .396 .8 .982 r
F P
0 g
s
.71364 .4986 m .73701 .5043 L .7088 .51518 L closepath p .306 .78 .929 r
F P
0 g
s
.71364 .4986 m .7088 .51518 L .69079 .50771 L closepath p .287 .759 .937 r
F P
0 g
s
.36456 .5058 m .3386 .51588 L .33812 .4973 L closepath p .384 .8 .977 r
F P
0 g
s
.3386 .51588 m .31218 .50669 L .33812 .4973 L closepath p .381 .797 .977 r
F P
0 g
s
.28593 .4965 m .31218 .50669 L .2859 .51607 L closepath p .392 .8 .981 r
F P
0 g
s
.28593 .4965 m .2859 .51607 L .25941 .50524 L closepath p .39 .799 .981 r
F P
0 g
s
.41818 .50249 m .39155 .51338 L .39025 .49561 L closepath p .375 .801 .97 r
F P
0 g
s
.39025 .49561 m .39155 .51338 L .36456 .5058 L closepath p .369 .796 .97 r
F P
0 g
s
.25941 .50524 m .23694 .51413 L .22888 .49313 L closepath p .398 .801 .983 r
F P
0 g
s
.23694 .51413 m .20878 .51206 L .22888 .49313 L closepath p .398 .801 .983 r
F P
0 g
s
.22888 .49313 m .20878 .51206 L .19035 .48837 L closepath p .399 .8 .984 r
F P
0 g
s
.17507 .51029 m .19035 .48837 L .20878 .51206 L closepath p .401 .801 .984 r
F P
0 g
s
.17507 .51029 m .14242 .50977 L .15554 .48985 L closepath p .402 .801 .985 r
F P
0 g
s
.15554 .48985 m .14242 .50977 L .12412 .48965 L closepath p .402 .801 .985 r
F P
0 g
s
.17507 .51029 m .15554 .48985 L .19035 .48837 L closepath p .401 .801 .984 r
F P
0 g
s
.68306 .49472 m .69079 .50771 L .66769 .50613 L closepath p .285 .753 .942 r
F P
0 g
s
.66769 .50613 m .65427 .48594 L .68306 .49472 L closepath p .304 .747 .964 r
F P
0 g
s
.71364 .4986 m .69079 .50771 L .68306 .49472 L closepath p .298 .762 .946 r
F P
0 g
s
.76552 .49332 m .73701 .5043 L .74542 .49262 L closepath p .239 .748 .875 r
F P
0 g
s
.74542 .49262 m .73701 .5043 L .71364 .4986 L closepath p .234 .739 .894 r
F P
0 g
s
.31207 .48777 m .33812 .4973 L .31218 .50669 L closepath p .384 .798 .978 r
F P
0 g
s
.31207 .48777 m .31218 .50669 L .28593 .4965 L closepath p .383 .797 .978 r
F P
0 g
s
.39025 .49561 m .36456 .5058 L .36342 .48797 L closepath p .373 .797 .972 r
F P
0 g
s
.36456 .5058 m .33812 .4973 L .36342 .48797 L closepath p .369 .794 .972 r
F P
0 g
s
.28593 .4965 m .25941 .50524 L .25976 .48516 L closepath p .393 .799 .982 r
F P
0 g
s
.25976 .48516 m .25941 .50524 L .22888 .49313 L closepath p .392 .799 .982 r
F P
0 g
s
.41582 .48507 m .44509 .49149 L .41818 .50249 L closepath p .36 .798 .962 r
F P
0 g
s
.41818 .50249 m .39025 .49561 L .41582 .48507 L closepath p .35 .791 .961 r
F P
0 g
s
.79431 .48222 m .76552 .49332 L .77274 .4844 L closepath p .197 .724 .837 r
F P
0 g
s
.76552 .49332 m .74542 .49262 L .77274 .4844 L closepath p .194 .72 .849 r
F P
0 g
s
.36342 .48797 m .33812 .4973 L .33753 .47919 L closepath p .374 .795 .975 r
F P
0 g
s
.33753 .47919 m .33812 .4973 L .31207 .48777 L closepath p .372 .793 .974 r
F P
0 g
s
.31207 .48777 m .28593 .4965 L .28634 .47719 L closepath p .386 .797 .98 r
F P
0 g
s
.28634 .47719 m .28593 .4965 L .25976 .48516 L closepath p .386 .796 .98 r
F P
0 g
s
.41582 .48507 m .39025 .49561 L .3876 .47846 L closepath p .357 .792 .966 r
F P
0 g
s
.39025 .49561 m .36342 .48797 L .3876 .47846 L closepath p .352 .789 .965 r
F P
0 g
s
.47227 .48037 m .44509 .49149 L .4421 .47533 L closepath p .336 .791 .948 r
F P
0 g
s
.4421 .47533 m .44509 .49149 L .41582 .48507 L closepath p .321 .78 .946 r
F P
0 g
s
.22888 .49313 m .23653 .47216 L .25976 .48516 L closepath p .395 .799 .983 r
F P
0 g
s
.21064 .47333 m .23653 .47216 L .22888 .49313 L closepath p .395 .798 .983 r
F P
0 g
s
.21064 .47333 m .22888 .49313 L .19035 .48837 L closepath p .398 .8 .984 r
F P
0 g
s
.13594 .46996 m .15554 .48985 L .12412 .48965 L closepath p .402 .801 .985 r
F P
0 g
s
.13594 .46996 m .12412 .48965 L .10548 .46916 L closepath p .402 .801 .985 r
F P
0 g
s
.15554 .48985 m .16548 .47023 L .19035 .48837 L closepath p .401 .801 .985 r
F P
0 g
s
.15554 .48985 m .13594 .46996 L .16548 .47023 L closepath p .401 .801 .985 r
F P
0 g
s
.72964 .48925 m .74542 .49262 L .71364 .4986 L closepath p .173 .705 .844 r
F P
0 g
s
.72964 .48925 m .71364 .4986 L .70078 .48873 L closepath p .19 .709 .877 r
F P
0 g
s
.70078 .48873 m .71364 .4986 L .68306 .49472 L closepath p .168 .694 .868 r
F P
0 g
s
.21064 .47333 m .19035 .48837 L .18827 .46843 L closepath p .399 .8 .984 r
F P
0 g
s
.18827 .46843 m .19035 .48837 L .16548 .47023 L closepath p .399 .8 .984 r
F P
0 g
s
.70078 .48873 m .68306 .49472 L .67115 .48139 L closepath p .199 .703 .905 r
F P
0 g
s
.68306 .49472 m .65427 .48594 L .67115 .48139 L closepath p .175 .688 .896 r
F P
0 g
s
.67115 .48139 m .65427 .48594 L .6406 .46537 L closepath p .212 .691 .934 r
F P
0 g
s
.33753 .47919 m .31207 .48777 L .31196 .4691 L closepath p .377 .794 .977 r
F P
0 g
s
.31196 .4691 m .31207 .48777 L .28634 .47719 L closepath p .376 .793 .977 r
F P
0 g
s
.3876 .47846 m .36342 .48797 L .36179 .47112 L closepath p .36 .79 .969 r
F P
0 g
s
.36179 .47112 m .36342 .48797 L .33753 .47919 L closepath p .358 .788 .969 r
F P
0 g
s
.8234 .47101 m .79431 .48222 L .80081 .47447 L closepath p .176 .711 .816 r
F P
0 g
s
.80081 .47447 m .79431 .48222 L .77274 .4844 L closepath p .174 .709 .822 r
F P
0 g
s
.25976 .48516 m .26146 .4657 L .28634 .47719 L closepath p .389 .797 .981 r
F P
0 g
s
.25976 .48516 m .23653 .47216 L .26146 .4657 L closepath p .389 .797 .981 r
F P
0 g
s
.46871 .46489 m .49974 .46914 L .47227 .48037 L closepath p .301 .779 .923 r
F P
0 g
s
.47227 .48037 m .4421 .47533 L .46871 .46489 L closepath p .278 .762 .921 r
F P
0 g
s
.77274 .4844 m .74542 .49262 L .75404 .48596 L closepath p .162 .7 .818 r
F P
0 g
s
.72964 .48925 m .75404 .48596 L .74542 .49262 L closepath p .164 .699 .839 r
F P
0 g
s
.4421 .47533 m .41582 .48507 L .41123 .46652 L closepath p .333 .784 .954 r
F P
0 g
s
.41582 .48507 m .3876 .47846 L .41123 .46652 L closepath p .326 .779 .953 r
F P
0 g
s
.36179 .47112 m .33753 .47919 L .33707 .46141 L closepath p .365 .79 .973 r
F P
0 g
s
.33753 .47919 m .31196 .4691 L .33707 .46141 L closepath p .364 .789 .973 r
F P
0 g
s
.28634 .47719 m .28641 .45801 L .31196 .4691 L closepath p .382 .794 .979 r
F P
0 g
s
.28634 .47719 m .26146 .4657 L .28641 .45801 L closepath p .382 .794 .979 r
F P
0 g
s
.82998 .46291 m .8528 .45968 L .8234 .47101 L closepath p .183 .715 .823 r
F P
0 g
s
.8234 .47101 m .80081 .47447 L .82998 .46291 L closepath p .184 .716 .821 r
F P
0 g
s
.41123 .46652 m .3876 .47846 L .38238 .46443 L closepath p .338 .781 .962 r
F P
0 g
s
.38238 .46443 m .3876 .47846 L .36179 .47112 L closepath p .34 .782 .962 r
F P
0 g
s
.22006 .46039 m .23653 .47216 L .21064 .47333 L closepath p .396 .799 .983 r
F P
0 g
s
.22006 .46039 m .21064 .47333 L .19977 .45655 L closepath p .396 .798 .984 r
F P
0 g
s
.18827 .46843 m .19977 .45655 L .21064 .47333 L closepath p .398 .8 .984 r
F P
0 g
s
.77641 .47865 m .80081 .47447 L .77274 .4844 L closepath p .145 .689 .792 r
F P
0 g
s
.77274 .4844 m .75404 .48596 L .77641 .47865 L closepath p .134 .68 .799 r
F P
0 g
s
.16548 .47023 m .13594 .46996 L .14768 .45096 L closepath p .402 .801 .985 r
F P
0 g
s
.14768 .45096 m .13594 .46996 L .11746 .44923 L closepath p .402 .801 .985 r
F P
0 g
s
.13594 .46996 m .10548 .46916 L .11746 .44923 L closepath p .403 .801 .985 r
F P
0 g
s
.11746 .44923 m .10548 .46916 L .08649 .44827 L closepath p .403 .802 .985 r
F P
0 g
s
.50071 .45832 m .52748 .4578 L .49974 .46914 L closepath p .24 .75 .873 r
F P
0 g
s
.50071 .45832 m .49974 .46914 L .46871 .46489 L closepath p .213 .728 .877 r
F P
0 g
s
.17528 .45309 m .18827 .46843 L .16548 .47023 L closepath p .401 .8 .985 r
F P
0 g
s
.16548 .47023 m .14768 .45096 L .17528 .45309 L closepath p .401 .8 .985 r
F P
0 g
s
.23826 .45427 m .26146 .4657 L .23653 .47216 L closepath p .392 .797 .983 r
F P
0 g
s
.23826 .45427 m .23653 .47216 L .22006 .46039 L closepath p .392 .797 .983 r
F P
0 g
s
.46871 .46489 m .4421 .47533 L .43915 .46232 L closepath p .298 .769 .936 r
F P
0 g
s
.4421 .47533 m .41123 .46652 L .43915 .46232 L closepath p .287 .761 .934 r
F P
0 g
s
.17528 .45309 m .19977 .45655 L .18827 .46843 L closepath p .399 .799 .984 r
F P
0 g
s
.72964 .48925 m .70078 .48873 L .71784 .48671 L closepath p .054 .62 .764 r
F P
0 g
s
.71784 .48671 m .70078 .48873 L .68902 .48284 L closepath p .078 .635 .803 r
F P
0 g
s
.70078 .48873 m .67115 .48139 L .68902 .48284 L closepath p .027 .595 .775 r
F P
0 g
s
.67115 .48139 m .65898 .47062 L .68902 .48284 L closepath p .066 .616 .83 r
F P
0 g
s
.67115 .48139 m .6406 .46537 L .65898 .47062 L closepath p .01 .574 .799 r
F P
0 g
s
.65898 .47062 m .6406 .46537 L .62666 .44441 L closepath p .081 .602 .878 r
F P
0 g
s
.74589 .48497 m .75404 .48596 L .72964 .48925 L closepath p .081 .641 .763 r
F P
0 g
s
.74589 .48497 m .72964 .48925 L .71784 .48671 L closepath p .094 .65 .786 r
F P
0 g
s
.38238 .46443 m .36179 .47112 L .36178 .45485 L closepath p .35 .784 .967 r
F P
0 g
s
.36179 .47112 m .33707 .46141 L .36178 .45485 L closepath p .348 .783 .967 r
F P
0 g
s
.31196 .4691 m .31155 .45034 L .33707 .46141 L closepath p .372 .79 .976 r
F P
0 g
s
.31196 .4691 m .28641 .45801 L .31155 .45034 L closepath p .372 .79 .976 r
F P
0 g
s
.28641 .45801 m .26146 .4657 L .26102 .44641 L closepath p .387 .795 .981 r
F P
0 g
s
.23826 .45427 m .26102 .44641 L .26146 .4657 L closepath p .387 .795 .981 r
F P
0 g
s
.74589 .48497 m .77641 .47865 L .75404 .48596 L closepath p .121 .671 .782 r
F P
0 g
s
.88249 .44824 m .8528 .45968 L .85927 .44978 L closepath p .216 .736 .854 r
F P
0 g
s
.85927 .44978 m .8528 .45968 L .82998 .46291 L closepath p .22 .74 .847 r
F P
0 g
s
.8081 .46744 m .82998 .46291 L .80081 .47447 L closepath p .159 .7 .797 r
F P
0 g
s
.8081 .46744 m .80081 .47447 L .77641 .47865 L closepath p .158 .699 .801 r
F P
0 g
s
.52935 .45294 m .50071 .45832 L .50351 .45184 L closepath p .139 .682 .818 r
F P
0 g
s
.52935 .45294 m .50351 .45184 L .5197 .45284 L p .02 0 0 r
F P
0 g
s
.53908 .45305 m .52748 .4578 L .52935 .45294 L p .146 .69 .781 r
F P
0 g
s
.50071 .45832 m .52935 .45294 L .52748 .4578 L closepath p .113 .664 .791 r
F P
0 g
s
.5197 .45284 m .53151 .45356 L .52935 .45294 L p .02 0 0 r
F P
0 g
s
.52935 .45294 m .55551 .44633 L .53908 .45305 L p .146 .69 .781 r
F P
0 g
s
.23826 .45427 m .22006 .46039 L .21503 .445 L closepath p .395 .798 .984 r
F P
0 g
s
.21503 .445 m .22006 .46039 L .19977 .45655 L closepath p .396 .799 .984 r
F P
0 g
s
.41123 .46652 m .42207 .44857 L .43915 .46232 L closepath p .298 .765 .941 r
F P
0 g
s
.38869 .45096 m .42207 .44857 L .41123 .46652 L closepath p .297 .76 .946 r
F P
0 g
s
.41123 .46652 m .38238 .46443 L .38869 .45096 L closepath p .329 .777 .957 r
F P
0 g
s
.18946 .43845 m .21503 .445 L .19977 .45655 L closepath p .397 .798 .984 r
F P
0 g
s
.19977 .45655 m .17528 .45309 L .18946 .43845 L closepath p .4 .8 .985 r
F P
0 g
s
.38238 .46443 m .36178 .45485 L .38869 .45096 L closepath p .327 .773 .959 r
F P
0 g
s
.50071 .45832 m .46871 .46489 L .47556 .45323 L closepath p .224 .734 .886 r
F P
0 g
s
.4513 .4549 m .47556 .45323 L .46871 .46489 L closepath p .217 .724 .896 r
F P
0 g
s
.4513 .4549 m .46871 .46489 L .43915 .46232 L closepath p .252 .746 .911 r
F P
0 g
s
.33707 .46141 m .33684 .44337 L .36178 .45485 L closepath p .358 .785 .972 r
F P
0 g
s
.33707 .46141 m .31155 .45034 L .33684 .44337 L closepath p .358 .785 .972 r
F P
0 g
s
.28551 .43846 m .31155 .45034 L .28641 .45801 L closepath p .379 .791 .979 r
F P
0 g
s
.28641 .45801 m .26102 .44641 L .28551 .43846 L closepath p .38 .792 .979 r
F P
0 g
s
.17528 .45309 m .16146 .4335 L .18946 .43845 L closepath p .4 .8 .985 r
F P
0 g
s
.17528 .45309 m .14768 .45096 L .16146 .4335 L closepath p .403 .801 .985 r
F P
0 g
s
.23826 .45427 m .23525 .43504 L .26102 .44641 L closepath p .391 .796 .983 r
F P
0 g
s
.23826 .45427 m .21503 .445 L .23525 .43504 L closepath p .393 .797 .983 r
F P
0 g
s
.16146 .4335 m .14768 .45096 L .13083 .42926 L closepath p .403 .801 .985 r
F P
0 g
s
.14768 .45096 m .11746 .44923 L .13083 .42926 L closepath p .404 .802 .986 r
F P
0 g
s
.13083 .42926 m .11746 .44923 L .09856 .4272 L closepath p .404 .802 .985 r
F P
0 g
s
.08649 .44827 m .09856 .4272 L .11746 .44923 L closepath p .405 .802 .986 r
F P
0 g
s
.09856 .4272 m .08649 .44827 L .06713 .42698 L closepath p .405 .802 .985 r
F P
0 g
s
.9125 .43667 m .88249 .44824 L .88959 .43516 L closepath p .269 .764 .9 r
F P
0 g
s
.88249 .44824 m .85927 .44978 L .88959 .43516 L closepath p .275 .771 .889 r
F P
0 g
s
.85927 .44978 m .82998 .46291 L .83573 .45437 L closepath p .2 .728 .828 r
F P
0 g
s
.8081 .46744 m .83573 .45437 L .82998 .46291 L closepath p .202 .729 .825 r
F P
0 g
s
.4513 .4549 m .43915 .46232 L .42207 .44857 L closepath p .269 .749 .931 r
F P
0 g
s
.58384 .43475 m .55551 .44633 L .55746 .45008 L closepath p 0 0 0 r
F P
0 g
s
.55746 .45008 m .55551 .44633 L .52935 .45294 L closepath p .046 0 0 r
F P
0 g
s
.71784 .48671 m .68902 .48284 L .70662 .48767 L closepath p .088 0 0 r
F P
0 g
s
.70662 .48767 m .68902 .48284 L .67812 .48168 L closepath p .068 0 0 r
F P
0 g
s
.65898 .47062 m .67812 .48168 L .68902 .48284 L closepath p .159 0 0 r
F P
0 g
s
.65898 .47062 m .64796 .46566 L .67812 .48168 L closepath p .122 0 0 r
F P
0 g
s
.62666 .44441 m .64796 .46566 L .65898 .47062 L closepath p .246 0 0 r
F P
0 g
s
.74589 .48497 m .71784 .48671 L .73444 .4871 L closepath p .025 0 0 r
F P
0 g
s
.73444 .4871 m .71784 .48671 L .70662 .48767 L closepath p .015 0 0 r
F P
0 g
s
.62666 .44441 m .61246 .42304 L .64796 .46566 L closepath p 0 .451 .766 r
F P
0 g
s
.79061 .47197 m .8081 .46744 L .77641 .47865 L closepath p .103 .658 .744 r
F P
0 g
s
.79061 .47197 m .77641 .47865 L .76245 .48143 L closepath p .104 .659 .745 r
F P
0 g
s
.76245 .48143 m .77641 .47865 L .74908 .48431 L p .042 .611 .709 r
F P
0 g
s
.76245 .48143 m .74981 .4881 L .7762 .47847 L p 0 0 0 r
F P
0 g
s
.7762 .47847 m .77681 .47825 L .76245 .48143 L p 0 0 0 r
F P
0 g
s
.74908 .48431 m .74589 .48497 L .76245 .48143 L p .042 .611 .709 r
F P
0 g
s
.76245 .48143 m .74589 .48497 L .73444 .4871 L closepath p .043 .612 .711 r
F P
0 g
s
.50071 .45832 m .47556 .45323 L .50351 .45184 L closepath p .121 .666 .824 r
F P
0 g
s
.36178 .45485 m .36288 .43762 L .38869 .45096 L closepath p .34 .777 .966 r
F P
0 g
s
.36178 .45485 m .33684 .44337 L .36288 .43762 L closepath p .339 .776 .966 r
F P
0 g
s
.31069 .43105 m .33684 .44337 L .31155 .45034 L closepath p .368 .786 .977 r
F P
0 g
s
.31069 .43105 m .31155 .45034 L .28551 .43846 L closepath p .37 .788 .977 r
F P
0 g
s
.20758 .42467 m .23525 .43504 L .21503 .445 L closepath p .395 .797 .984 r
F P
0 g
s
.18946 .43845 m .20758 .42467 L .21503 .445 L closepath p .398 .799 .985 r
F P
0 g
s
.26102 .44641 m .25893 .42606 L .28551 .43846 L closepath p .386 .793 .982 r
F P
0 g
s
.26102 .44641 m .23525 .43504 L .25893 .42606 L closepath p .388 .795 .982 r
F P
0 g
s
.94282 .42499 m .9125 .43667 L .92005 .42062 L closepath p .329 .788 .944 r
F P
0 g
s
.92005 .42062 m .9125 .43667 L .88959 .43516 L closepath p .335 .799 .932 r
F P
0 g
s
.8653 .43729 m .88959 .43516 L .85927 .44978 L closepath p .263 .766 .879 r
F P
0 g
s
.8653 .43729 m .85927 .44978 L .83573 .45437 L closepath p .269 .771 .871 r
F P
0 g
s
.17817 .41749 m .20758 .42467 L .18946 .43845 L closepath p .399 .798 .985 r
F P
0 g
s
.16146 .4335 m .17817 .41749 L .18946 .43845 L closepath p .403 .8 .986 r
F P
0 g
s
.61246 .42304 m .58384 .43475 L .5871 .45572 L closepath p .216 0 0 r
F P
0 g
s
.58384 .43475 m .55746 .45008 L .5871 .45572 L closepath p .287 0 0 r
F P
0 g
s
.81514 .46116 m .83573 .45437 L .8081 .46744 L closepath p .162 .702 .784 r
F P
0 g
s
.81514 .46116 m .8081 .46744 L .79061 .47197 L closepath p .162 .702 .783 r
F P
0 g
s
.38869 .45096 m .39092 .43294 L .42207 .44857 L closepath p .314 .766 .956 r
F P
0 g
s
.38869 .45096 m .36288 .43762 L .39092 .43294 L closepath p .312 .764 .956 r
F P
0 g
s
.47556 .45323 m .4513 .4549 L .45083 .44403 L closepath p .236 .733 .91 r
F P
0 g
s
.4513 .4549 m .42207 .44857 L .45083 .44403 L closepath p .229 .728 .909 r
F P
0 g
s
.55746 .45008 m .52935 .45294 L .53151 .45356 L closepath p 0 0 0 r
F P
0 g
s
.16146 .4335 m .14804 .41148 L .17817 .41749 L closepath p .403 .8 .986 r
F P
0 g
s
.16146 .4335 m .13083 .42926 L .14804 .41148 L closepath p .405 .802 .986 r
F P
0 g
s
.3362 .42444 m .36288 .43762 L .33684 .44337 L closepath p .353 .779 .973 r
F P
0 g
s
.3362 .42444 m .33684 .44337 L .31069 .43105 L closepath p .355 .781 .973 r
F P
0 g
s
.28406 .41811 m .31069 .43105 L .28551 .43846 L closepath p .378 .789 .98 r
F P
0 g
s
.28406 .41811 m .28551 .43846 L .25893 .42606 L closepath p .381 .791 .981 r
F P
0 g
s
.23082 .4131 m .25893 .42606 L .23525 .43504 L closepath p .392 .795 .984 r
F P
0 g
s
.20758 .42467 m .23082 .4131 L .23525 .43504 L closepath p .396 .797 .984 r
F P
0 g
s
.14804 .41148 m .13083 .42926 L .11228 .40274 L closepath p .405 .802 .986 r
F P
0 g
s
.13083 .42926 m .09856 .4272 L .11228 .40274 L closepath p .407 .802 .986 r
F P
0 g
s
.47843 .44485 m .50351 .45184 L .47556 .45323 L closepath p .164 .692 .862 r
F P
0 g
s
.45083 .44403 m .47843 .44485 L .47556 .45323 L closepath p .154 .683 .865 r
F P
0 g
s
.09856 .4272 m .06713 .42698 L .07742 .40603 L closepath p .407 .803 .986 r
F P
0 g
s
.0474 .40528 m .07742 .40603 L .06713 .42698 L closepath p .407 .804 .986 r
F P
0 g
s
.42207 .44857 m .42192 .42805 L .45083 .44403 L closepath p .266 .742 .936 r
F P
0 g
s
.42207 .44857 m .39092 .43294 L .42192 .42805 L closepath p .265 .742 .936 r
F P
0 g
s
.07742 .40603 m .11228 .40274 L .09856 .4272 L closepath p .407 .803 .986 r
F P
0 g
s
.70662 .48767 m .67812 .48168 L .69623 .49363 L closepath p .245 0 0 r
F P
0 g
s
.69623 .49363 m .67812 .48168 L .66908 .48765 L closepath p .24 0 0 r
F P
0 g
s
.66908 .48765 m .67812 .48168 L .64796 .46566 L closepath p .354 0 0 r
F P
0 g
s
.70662 .48767 m .72306 .49351 L .73444 .4871 L closepath p .154 0 0 r
F P
0 g
s
.70662 .48767 m .69623 .49363 L .72306 .49351 L closepath p .15 0 0 r
F P
0 g
s
.94601 .40968 m .97346 .41318 L .94282 .42499 L closepath p .377 .801 .972 r
F P
0 g
s
.92005 .42062 m .94601 .40968 L .94282 .42499 L closepath p .392 .816 .968 r
F P
0 g
s
.73444 .4871 m .74981 .4881 L .76245 .48143 L closepath p .066 0 0 r
F P
0 g
s
.73444 .4871 m .72306 .49351 L .74981 .4881 L closepath p .066 0 0 r
F P
0 g
s
.66908 .48765 m .64796 .46566 L .64205 .4761 L closepath p .371 0 0 r
F P
0 g
s
.64205 .4761 m .64796 .46566 L .61246 .42304 L closepath p .543 0 0 r
F P
0 g
s
.8982 .41726 m .92005 .42062 L .88959 .43516 L closepath p .336 .8 .933 r
F P
0 g
s
.88959 .43516 m .8653 .43729 L .8982 .41726 L closepath p .341 .809 .919 r
F P
0 g
s
.83677 .44623 m .8653 .43729 L .83573 .45437 L closepath p .227 .746 .83 r
F P
0 g
s
.83677 .44623 m .83573 .45437 L .81514 .46116 L closepath p .227 .746 .83 r
F P
0 g
s
.77681 .47825 m .79061 .47197 L .76245 .48143 L closepath p 0 0 0 r
F P
0 g
s
.64205 .4761 m .61246 .42304 L .61895 .47259 L closepath p .618 .126 .061 r
F P
0 g
s
.61895 .47259 m .61246 .42304 L .5871 .45572 L closepath p .554 .092 .111 r
F P
0 g
s
.36199 .41863 m .39092 .43294 L .36288 .43762 L closepath p .332 .769 .967 r
F P
0 g
s
.36199 .41863 m .36288 .43762 L .3362 .42444 L closepath p .335 .772 .967 r
F P
0 g
s
.23082 .4131 m .20758 .42467 L .19734 .39831 L closepath p .398 .797 .985 r
F P
0 g
s
.19734 .39831 m .20758 .42467 L .17817 .41749 L closepath p .402 .799 .986 r
F P
0 g
s
.5871 .45572 m .55746 .45008 L .55748 .45956 L closepath p .186 0 0 r
F P
0 g
s
.55746 .45008 m .53151 .45356 L .55748 .45956 L closepath p .201 0 0 r
F P
0 g
s
.80293 .46654 m .79061 .47197 L .80443 .46462 L p .115 .667 .736 r
F P
0 g
s
.77681 .47825 m .80443 .46462 L .79061 .47197 L closepath p .106 .659 .715 r
F P
0 g
s
.78953 .47412 m .80443 .46462 L .77681 .47825 L closepath p 0 0 0 r
F P
0 g
s
.80443 .46462 m .78953 .47412 L .80605 .46255 L p .153 .692 .718 r
F P
0 g
s
.80605 .46255 m .81732 .45464 L .80443 .46462 L p .153 .692 .718 r
F P
0 g
s
.80443 .46462 m .81514 .46116 L .80293 .46654 L p .115 .667 .736 r
F P
0 g
s
.30947 .41094 m .3362 .42444 L .31069 .43105 L closepath p .367 .783 .978 r
F P
0 g
s
.28406 .41811 m .30947 .41094 L .31069 .43105 L closepath p .37 .786 .978 r
F P
0 g
s
.257 .40453 m .28406 .41811 L .25893 .42606 L closepath p .387 .792 .983 r
F P
0 g
s
.23082 .4131 m .257 .40453 L .25893 .42606 L closepath p .391 .794 .983 r
F P
0 g
s
.50572 .44816 m .49252 .43899 L .50459 .44475 L p 0 .549 .762 r
F P
0 g
s
.50572 .44816 m .53151 .45356 L .50702 .45206 L p .051 .616 .776 r
F P
0 g
s
.50572 .44816 m .53571 .45753 L .53151 .45356 L closepath p .115 0 0 r
F P
0 g
s
.50459 .44475 m .51924 .45173 L .50572 .44816 L p 0 .549 .762 r
F P
0 g
s
.50702 .45206 m .50351 .45184 L .50572 .44816 L p .051 .616 .776 r
F P
0 g
s
.47843 .44485 m .50572 .44816 L .50351 .45184 L closepath p .04 .606 .779 r
F P
0 g
s
.83677 .44623 m .81514 .46116 L .80443 .46462 L closepath p .184 .717 .774 r
F P
0 g
s
.16618 .39833 m .19734 .39831 L .17817 .41749 L closepath p .403 .799 .986 r
F P
0 g
s
.14804 .41148 m .16618 .39833 L .17817 .41749 L closepath p .406 .801 .987 r
F P
0 g
s
.9314 .3994 m .94601 .40968 L .92005 .42062 L closepath p .4 .818 .972 r
F P
0 g
s
.8982 .41726 m .9314 .3994 L .92005 .42062 L closepath p .4 .831 .957 r
F P
0 g
s
.94601 .40968 m .96244 .39121 L .97346 .41318 L closepath p .44 .821 .99 r
F P
0 g
s
.45083 .44403 m .45571 .43335 L .47843 .44485 L closepath p .2 .708 .901 r
F P
0 g
s
.45083 .44403 m .42192 .42805 L .45571 .43335 L closepath p .195 .702 .903 r
F P
0 g
s
.87249 .42034 m .8982 .41726 L .8653 .43729 L closepath p .322 .801 .905 r
F P
0 g
s
.87249 .42034 m .8653 .43729 L .83677 .44623 L closepath p .329 .809 .892 r
F P
0 g
s
.42192 .42805 m .39092 .43294 L .38738 .41303 L closepath p .299 .752 .957 r
F P
0 g
s
.38738 .41303 m .39092 .43294 L .36199 .41863 L closepath p .309 .759 .958 r
F P
0 g
s
.14804 .41148 m .14017 .38741 L .16618 .39833 L closepath p .406 .801 .987 r
F P
0 g
s
.14804 .41148 m .11228 .40274 L .14017 .38741 L closepath p .409 .803 .987 r
F P
0 g
s
.33472 .40448 m .36199 .41863 L .3362 .42444 L closepath p .352 .775 .975 r
F P
0 g
s
.30947 .41094 m .33472 .40448 L .3362 .42444 L closepath p .357 .779 .975 r
F P
0 g
s
.28406 .41811 m .28276 .39705 L .30947 .41094 L closepath p .38 .787 .982 r
F P
0 g
s
.257 .40453 m .28276 .39705 L .28406 .41811 L closepath p .384 .79 .982 r
F P
0 g
s
.22946 .38942 m .257 .40453 L .23082 .4131 L closepath p .396 .795 .985 r
F P
0 g
s
.23082 .4131 m .19734 .39831 L .22946 .38942 L closepath p .4 .797 .986 r
F P
0 g
s
.9314 .3994 m .96244 .39121 L .94601 .40968 L closepath p .423 .829 .977 r
F P
0 g
s
.69623 .49363 m .71234 .50487 L .72306 .49351 L closepath p .283 0 0 r
F P
0 g
s
.68706 .50544 m .71234 .50487 L .69623 .49363 L closepath p .284 0 0 r
F P
0 g
s
.69623 .49363 m .66908 .48765 L .68706 .50544 L closepath p .374 0 0 r
F P
0 g
s
.53571 .45753 m .55748 .45956 L .53151 .45356 L closepath p .104 0 0 r
F P
0 g
s
.74981 .4881 m .72306 .49351 L .7377 .49928 L closepath p .186 0 0 r
F P
0 g
s
.72306 .49351 m .71234 .50487 L .7377 .49928 L closepath p .188 0 0 r
F P
0 g
s
.66908 .48765 m .66216 .50131 L .68706 .50544 L closepath p .38 0 0 r
F P
0 g
s
.66908 .48765 m .64205 .4761 L .66216 .50131 L closepath p .459 0 0 r
F P
0 g
s
.11228 .40274 m .07742 .40603 L .08177 .38366 L closepath p .41 .803 .987 r
F P
0 g
s
.07742 .40603 m .05749 .38985 L .08177 .38366 L closepath p .411 .804 .987 r
F P
0 g
s
.07742 .40603 m .0474 .40528 L .05749 .38985 L closepath p .41 .804 .987 r
F P
0 g
s
.02728 .38315 m .05749 .38985 L .0474 .40528 L closepath p .41 .805 .986 r
F P
0 g
s
.77681 .47825 m .74981 .4881 L .76331 .48894 L closepath p .079 0 0 r
F P
0 g
s
.74981 .4881 m .7377 .49928 L .76331 .48894 L closepath p .084 0 0 r
F P
0 g
s
.61895 .47259 m .5871 .45572 L .60254 .48431 L closepath p .444 0 0 r
F P
0 g
s
.5871 .45572 m .5768 .47319 L .60254 .48431 L closepath p .442 0 0 r
F P
0 g
s
.5871 .45572 m .55748 .45956 L .5768 .47319 L closepath p .358 0 0 r
F P
0 g
s
.4796 .43859 m .50572 .44816 L .47843 .44485 L closepath p .113 .652 .851 r
F P
0 g
s
.45571 .43335 m .4796 .43859 L .47843 .44485 L closepath p .112 .651 .851 r
F P
0 g
s
.64113 .49466 m .66216 .50131 L .64205 .4761 L closepath p .478 0 0 r
F P
0 g
s
.64113 .49466 m .64205 .4761 L .61895 .47259 L closepath p .501 0 0 r
F P
0 g
s
.77681 .47825 m .76331 .48894 L .78953 .47412 L closepath p 0 0 0 r
F P
0 g
s
.10977 .37338 m .14017 .38741 L .11228 .40274 L closepath p .41 .803 .987 r
F P
0 g
s
.11228 .40274 m .08177 .38366 L .10977 .37338 L closepath p .412 .804 .987 r
F P
0 g
s
.91068 .38831 m .9314 .3994 L .8982 .41726 L closepath p .42 .836 .967 r
F P
0 g
s
.87249 .42034 m .91068 .38831 L .8982 .41726 L closepath p .425 .851 .949 r
F P
0 g
s
.81616 .45795 m .80443 .46462 L .81732 .45464 L p .168 .706 .759 r
F P
0 g
s
.7999 .46898 m .81732 .45464 L .78953 .47412 L closepath p 0 0 0 r
F P
0 g
s
.81732 .45464 m .7999 .46898 L .81848 .45134 L p .222 .74 .736 r
F P
0 g
s
.81848 .45134 m .8259 .44429 L .81732 .45464 L p .222 .74 .736 r
F P
0 g
s
.81732 .45464 m .83677 .44623 L .81616 .45795 L p .168 .706 .759 r
F P
0 g
s
.38738 .41303 m .36199 .41863 L .35915 .39858 L closepath p .331 .763 .97 r
F P
0 g
s
.33472 .40448 m .35915 .39858 L .36199 .41863 L closepath p .34 .77 .971 r
F P
0 g
s
.30947 .41094 m .30809 .39002 L .33472 .40448 L closepath p .369 .781 .98 r
F P
0 g
s
.30947 .41094 m .28276 .39705 L .30809 .39002 L closepath p .375 .785 .981 r
F P
0 g
s
.87249 .42034 m .83677 .44623 L .84699 .43033 L closepath p .281 .782 .849 r
F P
0 g
s
.83677 .44623 m .81732 .45464 L .84699 .43033 L closepath p .282 .784 .824 r
F P
0 g
s
.28276 .39705 m .257 .40453 L .2565 .38298 L closepath p .391 .791 .985 r
F P
0 g
s
.2565 .38298 m .257 .40453 L .22946 .38942 L closepath p .395 .794 .985 r
F P
0 g
s
.16618 .39833 m .16813 .37703 L .19734 .39831 L closepath p .407 .8 .987 r
F P
0 g
s
.14017 .38741 m .16813 .37703 L .16618 .39833 L closepath p .409 .802 .987 r
F P
0 g
s
.5768 .47319 m .55748 .45956 L .55626 .46687 L closepath p .274 0 0 r
F P
0 g
s
.55626 .46687 m .55748 .45956 L .53571 .45753 L closepath p .265 0 0 r
F P
0 g
s
.64113 .49466 m .61895 .47259 L .62799 .50241 L closepath p .522 0 0 r
F P
0 g
s
.61895 .47259 m .60254 .48431 L .62799 .50241 L closepath p .52 0 0 r
F P
0 g
s
.42192 .42805 m .43777 .41696 L .45571 .43335 L closepath p .238 .722 .93 r
F P
0 g
s
.40845 .40609 m .43777 .41696 L .42192 .42805 L closepath p .243 .719 .938 r
F P
0 g
s
.42192 .42805 m .38738 .41303 L .40845 .40609 L closepath p .283 .744 .95 r
F P
0 g
s
.19734 .39831 m .20214 .36725 L .22946 .38942 L closepath p .405 .798 .987 r
F P
0 g
s
.19734 .39831 m .16813 .37703 L .20214 .36725 L closepath p .407 .8 .987 r
F P
0 g
s
.9314 .3994 m .9512 .36881 L .96244 .39121 L closepath p .476 .839 .993 r
F P
0 g
s
.91068 .38831 m .9512 .36881 L .9314 .3994 L closepath p .462 .855 .976 r
F P
0 g
s
.51924 .45173 m .50595 .44169 L .51975 .45127 L p 0 .478 .716 r
F P
0 g
s
.51924 .45173 m .53571 .45753 L .51871 .45222 L p 0 .539 .721 r
F P
0 g
s
.54426 .46528 m .53571 .45753 L .51924 .45173 L closepath p .183 0 0 r
F P
0 g
s
.53171 .45958 m .54426 .46528 L .5292 .45712 L p .131 0 0 r
F P
0 g
s
.5292 .45712 m .51924 .45173 L .53171 .45958 L p .131 0 0 r
F P
0 g
s
.51975 .45127 m .53171 .45958 L .51924 .45173 L p 0 .478 .716 r
F P
0 g
s
.53171 .45958 m .50595 .44169 L .52025 .4488 L closepath p 0 .52 .767 r
F P
0 g
s
.53171 .45958 m .52025 .4488 L .53419 .46201 L p .177 0 0 r
F P
0 g
s
.53419 .46201 m .54508 .47232 L .53171 .45958 L p .177 0 0 r
F P
0 g
s
.51871 .45222 m .50572 .44816 L .51924 .45173 L p 0 .539 .721 r
F P
0 g
s
.49252 .43899 m .50572 .44816 L .4796 .43859 L closepath p .055 .613 .809 r
F P
0 g
s
.05749 .38985 m .05358 .3709 L .08177 .38366 L closepath p .411 .805 .987 r
F P
0 g
s
.02728 .38315 m .05358 .3709 L .05749 .38985 L closepath p .413 .806 .987 r
F P
0 g
s
.45571 .43335 m .46597 .42709 L .4796 .43859 L closepath p .163 .68 .89 r
F P
0 g
s
.46597 .42709 m .45571 .43335 L .43777 .41696 L closepath p .166 .677 .898 r
F P
0 g
s
.7377 .49928 m .71234 .50487 L .72629 .51551 L closepath p .301 0 0 r
F P
0 g
s
.70258 .52154 m .72629 .51551 L .71234 .50487 L closepath p .303 0 0 r
F P
0 g
s
.70258 .52154 m .71234 .50487 L .68706 .50544 L closepath p .39 0 0 r
F P
0 g
s
.70258 .52154 m .68706 .50544 L .67892 .52312 L closepath p .392 0 0 r
F P
0 g
s
.68706 .50544 m .66216 .50131 L .67892 .52312 L closepath p .462 0 0 r
F P
0 g
s
.7503 .50473 m .76331 .48894 L .7377 .49928 L closepath p .193 0 0 r
F P
0 g
s
.7377 .49928 m .72629 .51551 L .7503 .50473 L closepath p .197 0 0 r
F P
0 g
s
.33472 .40448 m .33286 .38373 L .35915 .39858 L closepath p .356 .772 .978 r
F P
0 g
s
.33472 .40448 m .30809 .39002 L .33286 .38373 L closepath p .364 .779 .979 r
F P
0 g
s
.66216 .50131 m .65383 .52076 L .67892 .52312 L closepath p .463 0 0 r
F P
0 g
s
.64113 .49466 m .65383 .52076 L .66216 .50131 L closepath p .509 0 0 r
F P
0 g
s
.28155 .37537 m .30809 .39002 L .28276 .39705 L closepath p .384 .786 .984 r
F P
0 g
s
.28155 .37537 m .28276 .39705 L .2565 .38298 L closepath p .39 .791 .985 r
F P
0 g
s
.13753 .36119 m .16813 .37703 L .14017 .38741 L closepath p .411 .802 .988 r
F P
0 g
s
.10977 .37338 m .13753 .36119 L .14017 .38741 L closepath p .413 .804 .988 r
F P
0 g
s
.64113 .49466 m .62799 .50241 L .65383 .52076 L closepath p .508 0 0 r
F P
0 g
s
.76331 .48894 m .77476 .48916 L .78953 .47412 L closepath p .065 0 0 r
F P
0 g
s
.7503 .50473 m .77476 .48916 L .76331 .48894 L closepath p .072 0 0 r
F P
0 g
s
.38123 .39275 m .40845 .40609 L .38738 .41303 L closepath p .302 .746 .963 r
F P
0 g
s
.38738 .41303 m .35915 .39858 L .38123 .39275 L closepath p .321 .759 .967 r
F P
0 g
s
.87249 .42034 m .87719 .40019 L .91068 .38831 L closepath p .398 .843 .93 r
F P
0 g
s
.87249 .42034 m .84699 .43033 L .87719 .40019 L closepath p .413 .855 .918 r
F P
0 g
s
.55626 .46687 m .53571 .45753 L .54426 .46528 L closepath p .188 0 0 r
F P
0 g
s
.49252 .43899 m .4796 .43859 L .46597 .42709 L closepath p .084 .626 .849 r
F P
0 g
s
.23377 .37002 m .2565 .38298 L .22946 .38942 L closepath p .401 .795 .987 r
F P
0 g
s
.23377 .37002 m .22946 .38942 L .20214 .36725 L closepath p .403 .797 .987 r
F P
0 g
s
.08177 .38366 m .08016 .35851 L .10977 .37338 L closepath p .413 .804 .988 r
F P
0 g
s
.05358 .3709 m .08016 .35851 L .08177 .38366 L closepath p .414 .806 .988 r
F P
0 g
s
.77476 .48916 m .7999 .46898 L .78953 .47412 L closepath p 0 0 0 r
F P
0 g
s
.5901 .49376 m .60254 .48431 L .5768 .47319 L closepath p .424 0 0 r
F P
0 g
s
.5901 .49376 m .5768 .47319 L .56703 .47918 L closepath p .408 0 0 r
F P
0 g
s
.5768 .47319 m .55626 .46687 L .56703 .47918 L closepath p .33 0 0 r
F P
0 g
s
.60254 .48431 m .61381 .5114 L .62799 .50241 L closepath p .498 0 0 r
F P
0 g
s
.60254 .48431 m .5901 .49376 L .61381 .5114 L closepath p .499 0 0 r
F P
0 g
s
.55626 .46687 m .54426 .46528 L .56703 .47918 L closepath p .304 0 0 r
F P
0 g
s
.8259 .44429 m .84699 .43033 L .81732 .45464 L closepath p .223 .743 .766 r
F P
0 g
s
.30685 .36821 m .33286 .38373 L .30809 .39002 L closepath p .376 .78 .984 r
F P
0 g
s
.30685 .36821 m .30809 .39002 L .28155 .37537 L closepath p .384 .786 .984 r
F P
0 g
s
.38123 .39275 m .35915 .39858 L .35596 .37814 L closepath p .339 .761 .975 r
F P
0 g
s
.35915 .39858 m .33286 .38373 L .35596 .37814 L closepath p .352 .771 .977 r
F P
0 g
s
.28155 .37537 m .2565 .38298 L .25495 .36125 L closepath p .396 .791 .987 r
F P
0 g
s
.23377 .37002 m .25495 .36125 L .2565 .38298 L closepath p .402 .795 .987 r
F P
0 g
s
.1663 .34938 m .20214 .36725 L .16813 .37703 L closepath p .411 .801 .989 r
F P
0 g
s
.13753 .36119 m .1663 .34938 L .16813 .37703 L closepath p .415 .803 .989 r
F P
0 g
s
.65383 .52076 m .62799 .50241 L .63696 .53085 L closepath p .53 .006 0 r
F P
0 g
s
.62799 .50241 m .61381 .5114 L .63696 .53085 L closepath p .538 .002 0 r
F P
0 g
s
.13753 .36119 m .10977 .37338 L .10704 .34599 L closepath p .414 .804 .988 r
F P
0 g
s
.08016 .35851 m .10704 .34599 L .10977 .37338 L closepath p .416 .806 .988 r
F P
0 g
s
.71589 .53695 m .72629 .51551 L .70258 .52154 L closepath p .399 0 0 r
F P
0 g
s
.69408 .54332 m .71589 .53695 L .70258 .52154 L closepath p .4 0 0 r
F P
0 g
s
.69408 .54332 m .70258 .52154 L .67892 .52312 L closepath p .471 0 0 r
F P
0 g
s
.93974 .34596 m .9512 .36881 L .91068 .38831 L closepath p .526 .863 .994 r
F P
0 g
s
.93974 .34596 m .91068 .38831 L .90477 .36433 L closepath p .5 .877 .974 r
F P
0 g
s
.87719 .40019 m .90477 .36433 L .91068 .38831 L closepath p .514 .885 .973 r
F P
0 g
s
.72629 .51551 m .73812 .52587 L .7503 .50473 L closepath p .304 0 0 r
F P
0 g
s
.71589 .53695 m .73812 .52587 L .72629 .51551 L closepath p .305 0 0 r
F P
0 g
s
.67294 .54559 m .69408 .54332 L .67892 .52312 L closepath p .471 0 0 r
F P
0 g
s
.67294 .54559 m .67892 .52312 L .65383 .52076 L closepath p .523 .012 0 r
F P
0 g
s
.43777 .41696 m .45149 .40998 L .46597 .42709 L closepath p .208 .698 .924 r
F P
0 g
s
.43777 .41696 m .42311 .39679 L .45149 .40998 L closepath p .214 .696 .931 r
F P
0 g
s
.40845 .40609 m .42311 .39679 L .43777 .41696 L closepath p .269 .73 .95 r
F P
0 g
s
.85212 .41518 m .87719 .40019 L .84699 .43033 L closepath p .361 .831 .873 r
F P
0 g
s
.85212 .41518 m .84699 .43033 L .8259 .44429 L closepath p .372 .839 .856 r
F P
0 g
s
.39887 .38762 m .42311 .39679 L .40845 .40609 L closepath p .275 .728 .956 r
F P
0 g
s
.39887 .38762 m .40845 .40609 L .38123 .39275 L closepath p .308 .748 .965 r
F P
0 g
s
.49252 .43899 m .46597 .42709 L .47907 .42488 L closepath p .12 .647 .874 r
F P
0 g
s
.47907 .42488 m .46597 .42709 L .45149 .40998 L closepath p .131 .648 .89 r
F P
0 g
s
.49252 .43899 m .50595 .44169 L .51924 .45173 L closepath p .007 .573 .791 r
F P
0 g
s
.49252 .43899 m .47907 .42488 L .50595 .44169 L closepath p .03 .581 .827 r
F P
0 g
s
.77476 .48916 m .7503 .50473 L .76082 .50982 L closepath p .181 0 0 r
F P
0 g
s
.7503 .50473 m .73812 .52587 L .76082 .50982 L closepath p .184 0 0 r
F P
0 g
s
.67294 .54559 m .65383 .52076 L .65567 .54658 L closepath p .524 .015 0 r
F P
0 g
s
.65383 .52076 m .63696 .53085 L .65567 .54658 L closepath p .552 .036 0 r
F P
0 g
s
.7999 .46898 m .77476 .48916 L .78404 .48865 L closepath p .028 0 0 r
F P
0 g
s
.78404 .48865 m .77476 .48916 L .76082 .50982 L closepath p .033 0 0 r
F P
0 g
s
.56703 .47918 m .54426 .46528 L .55583 .47824 L closepath p .278 0 0 r
F P
0 g
s
.54426 .46528 m .53171 .45958 L .55583 .47824 L closepath p .256 0 0 r
F P
0 g
s
.33159 .36247 m .35596 .37814 L .33286 .38373 L closepath p .366 .772 .983 r
F P
0 g
s
.33159 .36247 m .33286 .38373 L .30685 .36821 L closepath p .375 .78 .984 r
F P
0 g
s
.30685 .36821 m .28155 .37537 L .28002 .35242 L closepath p .392 .787 .987 r
F P
0 g
s
.28002 .35242 m .28155 .37537 L .25495 .36125 L closepath p .399 .792 .988 r
F P
0 g
s
.23377 .37002 m .22634 .34917 L .25495 .36125 L closepath p .404 .795 .988 r
F P
0 g
s
.23377 .37002 m .20214 .36725 L .22634 .34917 L closepath p .411 .799 .989 r
F P
0 g
s
.19451 .33625 m .22634 .34917 L .20214 .36725 L closepath p .411 .799 .989 r
F P
0 g
s
.1663 .34938 m .19451 .33625 L .20214 .36725 L closepath p .417 .803 .99 r
F P
0 g
s
.57879 .49756 m .5901 .49376 L .56703 .47918 L closepath p .402 0 0 r
F P
0 g
s
.57879 .49756 m .56703 .47918 L .55583 .47824 L closepath p .382 0 0 r
F P
0 g
s
.5901 .49376 m .60126 .51805 L .61381 .5114 L closepath p .492 0 0 r
F P
0 g
s
.5901 .49376 m .57879 .49756 L .60126 .51805 L closepath p .487 0 0 r
F P
0 g
s
.39887 .38762 m .38123 .39275 L .37684 .37184 L closepath p .322 .749 .973 r
F P
0 g
s
.38123 .39275 m .35596 .37814 L .37684 .37184 L closepath p .34 .762 .976 r
F P
0 g
s
.80733 .46347 m .7916 .4866 L .80684 .46528 L p 0 0 0 r
F P
0 g
s
.80733 .46347 m .8259 .44429 L .80789 .46139 L p 0 0 0 r
F P
0 g
s
.83132 .43297 m .8259 .44429 L .80733 .46347 L closepath p .312 .799 .765 r
F P
0 g
s
.80684 .46528 m .81078 .45976 L .80733 .46347 L p 0 0 0 r
F P
0 g
s
.80789 .46139 m .7999 .46898 L .80733 .46347 L p 0 0 0 r
F P
0 g
s
.80733 .46347 m .7999 .46898 L .78404 .48865 L closepath p 0 0 0 r
F P
0 g
s
.13753 .36119 m .1342 .33333 L .1663 .34938 L closepath p .416 .803 .989 r
F P
0 g
s
.13753 .36119 m .10704 .34599 L .1342 .33333 L closepath p .419 .806 .989 r
F P
0 g
s
.62336 .54031 m .63696 .53085 L .61381 .5114 L closepath p .548 .02 0 r
F P
0 g
s
.61381 .5114 m .60126 .51805 L .62336 .54031 L closepath p .556 .011 0 r
F P
0 g
s
.876 .38331 m .90477 .36433 L .87719 .40019 L closepath p .469 .881 .938 r
F P
0 g
s
.85212 .41518 m .876 .38331 L .87719 .40019 L closepath p .486 .892 .934 r
F P
0 g
s
.83132 .43297 m .85212 .41518 L .8259 .44429 L closepath p .303 .796 .788 r
F P
0 g
s
.25165 .33725 m .28002 .35242 L .25495 .36125 L closepath p .405 .792 .989 r
F P
0 g
s
.25165 .33725 m .25495 .36125 L .22634 .34917 L closepath p .412 .797 .99 r
F P
0 g
s
.69408 .54332 m .70669 .56357 L .71589 .53695 L closepath p .478 0 .035 r
F P
0 g
s
.68681 .56964 m .70669 .56357 L .69408 .54332 L closepath p .478 0 .032 r
F P
0 g
s
.68681 .56964 m .69408 .54332 L .67294 .54559 L closepath p .536 .061 .073 r
F P
0 g
s
.65567 .54658 m .63696 .53085 L .6463 .56541 L closepath p .576 .083 .048 r
F P
0 g
s
.62336 .54031 m .6463 .56541 L .63696 .53085 L closepath p .591 .079 .001 r
F P
0 g
s
.33159 .36247 m .30685 .36821 L .30622 .34494 L closepath p .387 .781 .988 r
F P
0 g
s
.30685 .36821 m .28002 .35242 L .30622 .34494 L closepath p .396 .788 .988 r
F P
0 g
s
.72709 .55237 m .73812 .52587 L .71589 .53695 L closepath p .398 0 0 r
F P
0 g
s
.72709 .55237 m .71589 .53695 L .70669 .56357 L closepath p .399 0 0 r
F P
0 g
s
.3531 .35911 m .37684 .37184 L .35596 .37814 L closepath p .352 .762 .981 r
F P
0 g
s
.33159 .36247 m .3531 .35911 L .35596 .37814 L closepath p .367 .773 .983 r
F P
0 g
s
.76082 .50982 m .73812 .52587 L .74787 .53616 L closepath p .291 0 0 r
F P
0 g
s
.73812 .52587 m .72709 .55237 L .74787 .53616 L closepath p .292 0 0 r
F P
0 g
s
.67294 .54559 m .66762 .57132 L .68681 .56964 L closepath p .536 .055 .061 r
F P
0 g
s
.67294 .54559 m .65567 .54658 L .66762 .57132 L closepath p .575 .098 .084 r
F P
0 g
s
.76082 .50982 m .76943 .51459 L .78404 .48865 L closepath p .147 0 0 r
F P
0 g
s
.76082 .50982 m .74787 .53616 L .76943 .51459 L closepath p .148 0 0 r
F P
0 g
s
.39887 .38762 m .40422 .36961 L .42311 .39679 L closepath p .315 .742 .972 r
F P
0 g
s
.39887 .38762 m .37684 .37184 L .40422 .36961 L closepath p .319 .748 .972 r
F P
0 g
s
.45149 .40998 m .46552 .4054 L .47907 .42488 L closepath p .189 .678 .924 r
F P
0 g
s
.45149 .40998 m .43755 .38818 L .46552 .4054 L closepath p .196 .676 .931 r
F P
0 g
s
.42311 .39679 m .43755 .38818 L .45149 .40998 L closepath p .262 .718 .954 r
F P
0 g
s
.50595 .44169 m .47907 .42488 L .49332 .42571 L closepath p .083 .615 .866 r
F P
0 g
s
.46552 .4054 m .49332 .42571 L .47907 .42488 L closepath p .098 .616 .885 r
F P
0 g
s
.50595 .44169 m .49332 .42571 L .52025 .4488 L closepath p 0 .533 .807 r
F P
0 g
s
.42311 .39679 m .40422 .36961 L .43755 .38818 L closepath p .266 .715 .958 r
F P
0 g
s
.65567 .54658 m .6463 .56541 L .66762 .57132 L closepath p .575 .082 .047 r
F P
0 g
s
.5415 .46716 m .53171 .45958 L .54508 .47232 L p .211 0 0 r
F P
0 g
s
.5357 .46144 m .54508 .47232 L .52988 .45793 L p 0 .461 .739 r
F P
0 g
s
.52988 .45793 m .52025 .4488 L .5357 .46144 L p 0 .461 .739 r
F P
0 g
s
.5357 .46144 m .52853 .44595 L .54133 .46485 L p 0 .43 .777 r
F P
0 g
s
.5486 .47741 m .55901 .49031 L .54508 .47232 L p .244 0 0 r
F P
0 g
s
.54508 .47232 m .55583 .47824 L .5415 .46716 L p .211 0 0 r
F P
0 g
s
.54508 .47232 m .56824 .49622 L .55583 .47824 L closepath p .331 0 0 r
F P
0 g
s
.55901 .49031 m .56824 .49622 L .555 .48257 L p .293 0 0 r
F P
0 g
s
.555 .48257 m .54508 .47232 L .55901 .49031 L p .293 0 0 r
F P
0 g
s
.54508 .47232 m .5357 .46144 L .5486 .47741 L p .244 0 0 r
F P
0 g
s
.55146 .47979 m .55901 .49031 L .54539 .47344 L p 0 .393 .704 r
F P
0 g
s
.54539 .47344 m .5357 .46144 L .55146 .47979 L p 0 .393 .704 r
F P
0 g
s
.55146 .47979 m .54602 .46514 L .55725 .48585 L p 0 .368 .754 r
F P
0 g
s
.55901 .49031 m .55146 .47979 L .56293 .49787 L p .316 0 0 r
F P
0 g
s
.56293 .49787 m .57308 .51387 L .55901 .49031 L p .316 0 0 r
F P
0 g
s
.55901 .49031 m .58068 .51915 L .56824 .49622 L closepath p .412 0 0 r
F P
0 g
s
.55901 .49031 m .57308 .51387 L .58068 .51915 L closepath p .379 0 0 r
F P
0 g
s
.55725 .48585 m .56644 .5028 L .55146 .47979 L p 0 .368 .754 r
F P
0 g
s
.54133 .46485 m .55146 .47979 L .5357 .46144 L p 0 .43 .777 r
F P
0 g
s
.16167 .32054 m .19451 .33625 L .1663 .34938 L closepath p .418 .803 .99 r
F P
0 g
s
.1663 .34938 m .1342 .33333 L .16167 .32054 L closepath p .422 .806 .99 r
F P
0 g
s
.57879 .49756 m .55583 .47824 L .56824 .49622 L closepath p .362 0 0 r
F P
0 g
s
.90477 .36433 m .92804 .32265 L .93974 .34596 L closepath p .581 .886 .991 r
F P
0 g
s
.57879 .49756 m .59014 .52057 L .60126 .51805 L closepath p .48 0 0 r
F P
0 g
s
.56824 .49622 m .59014 .52057 L .57879 .49756 L closepath p .464 0 0 r
F P
0 g
s
.85495 .40181 m .876 .38331 L .85212 .41518 L closepath p .434 .873 .887 r
F P
0 g
s
.85495 .40181 m .85212 .41518 L .83132 .43297 L closepath p .449 .882 .869 r
F P
0 g
s
.80733 .46347 m .78404 .48865 L .7916 .4866 L closepath p 0 0 0 r
F P
0 g
s
.7916 .4866 m .78404 .48865 L .76943 .51459 L closepath p 0 0 0 r
F P
0 g
s
.22136 .32264 m .25165 .33725 L .22634 .34917 L closepath p .414 .797 .991 r
F P
0 g
s
.19451 .33625 m .22136 .32264 L .22634 .34917 L closepath p .42 .802 .991 r
F P
0 g
s
.61094 .54547 m .62336 .54031 L .60126 .51805 L closepath p .559 .016 0 r
F P
0 g
s
.60126 .51805 m .59014 .52057 L .61094 .54547 L closepath p .561 0 0 r
F P
0 g
s
.33159 .36247 m .3334 .34009 L .3531 .35911 L closepath p .382 .774 .988 r
F P
0 g
s
.33159 .36247 m .30622 .34494 L .3334 .34009 L closepath p .39 .782 .988 r
F P
0 g
s
.28002 .35242 m .27832 .32725 L .30622 .34494 L closepath p .403 .788 .99 r
F P
0 g
s
.28002 .35242 m .25165 .33725 L .27832 .32725 L closepath p .412 .795 .991 r
F P
0 g
s
.92804 .32265 m .90477 .36433 L .89691 .34516 L closepath p .561 .905 .975 r
F P
0 g
s
.876 .38331 m .89691 .34516 L .90477 .36433 L closepath p .572 .909 .974 r
F P
0 g
s
.81078 .45976 m .8009 .47675 L .80997 .46142 L p 0 0 0 r
F P
0 g
s
.81078 .45976 m .83132 .43297 L .81167 .45794 L p 0 0 0 r
F P
0 g
s
.8224 .44041 m .83132 .43297 L .81078 .45976 L closepath p .358 .826 .768 r
F P
0 g
s
.80997 .46142 m .8224 .44041 L .81078 .45976 L p 0 0 0 r
F P
0 g
s
.81167 .45794 m .80733 .46347 L .81078 .45976 L p 0 0 0 r
F P
0 g
s
.40422 .36961 m .37684 .37184 L .36677 .3425 L closepath p .343 .75 .981 r
F P
0 g
s
.36677 .3425 m .37684 .37184 L .3531 .35911 L closepath p .372 .769 .986 r
F P
0 g
s
.63025 .57089 m .6463 .56541 L .62336 .54031 L closepath p .601 .1 .034 r
F P
0 g
s
.63025 .57089 m .62336 .54031 L .61094 .54547 L closepath p .621 .102 0 r
F P
0 g
s
.70669 .56357 m .71742 .58483 L .72709 .55237 L closepath p .476 .03 .111 r
F P
0 g
s
.69858 .59504 m .71742 .58483 L .70669 .56357 L closepath p .476 .03 .11 r
F P
0 g
s
.69858 .59504 m .70669 .56357 L .68681 .56964 L closepath p .542 .104 .158 r
F P
0 g
s
.68681 .56964 m .68049 .59963 L .69858 .59504 L closepath p .543 .099 .145 r
F P
0 g
s
.68681 .56964 m .66762 .57132 L .68049 .59963 L closepath p .589 .15 .175 r
F P
0 g
s
.73641 .5674 m .74787 .53616 L .72709 .55237 L closepath p .383 0 .024 r
F P
0 g
s
.72709 .55237 m .71742 .58483 L .73641 .5674 L closepath p .382 0 .037 r
F P
0 g
s
.68049 .59963 m .66762 .57132 L .66286 .59987 L closepath p .592 .139 .147 r
F P
0 g
s
.66762 .57132 m .6463 .56541 L .66286 .59987 L closepath p .624 .175 .165 r
F P
0 g
s
.75526 .54685 m .76943 .51459 L .74787 .53616 L closepath p .262 0 0 r
F P
0 g
s
.73641 .5674 m .75526 .54685 L .74787 .53616 L closepath p .26 0 0 r
F P
0 g
s
.36677 .3425 m .3531 .35911 L .3334 .34009 L closepath p .372 .77 .986 r
F P
0 g
s
.19451 .33625 m .18944 .3076 L .22136 .32264 L closepath p .421 .802 .991 r
F P
0 g
s
.16167 .32054 m .18944 .3076 L .19451 .33625 L closepath p .426 .805 .992 r
F P
0 g
s
.89691 .34516 m .876 .38331 L .86977 .37233 L closepath p .54 .916 .941 r
F P
0 g
s
.85495 .40181 m .86977 .37233 L .876 .38331 L closepath p .543 .917 .941 r
F P
0 g
s
.6463 .56541 m .64664 .59888 L .66286 .59987 L closepath p .624 .157 .125 r
F P
0 g
s
.63025 .57089 m .64664 .59888 L .6463 .56541 L closepath p .655 .184 .122 r
F P
0 g
s
.25165 .33725 m .24862 .31044 L .27832 .32725 L closepath p .416 .795 .992 r
F P
0 g
s
.25165 .33725 m .22136 .32264 L .24862 .31044 L closepath p .424 .8 .993 r
F P
0 g
s
.30622 .34494 m .30605 .3189 L .3334 .34009 L closepath p .402 .782 .991 r
F P
0 g
s
.30622 .34494 m .27832 .32725 L .30605 .3189 L closepath p .412 .791 .992 r
F P
0 g
s
.77685 .51886 m .7916 .4866 L .76943 .51459 L closepath p .077 0 0 r
F P
0 g
s
.75526 .54685 m .77685 .51886 L .76943 .51459 L closepath p .073 0 0 r
F P
0 g
s
.58068 .51915 m .59014 .52057 L .56824 .49622 L closepath p .449 0 0 r
F P
0 g
s
.84519 .40548 m .85495 .40181 L .83132 .43297 L closepath p .45 .883 .87 r
F P
0 g
s
.84519 .40548 m .83132 .43297 L .8224 .44041 L closepath p .409 .856 .783 r
F P
0 g
s
.52025 .4488 m .51007 .43203 L .5357 .46144 L closepath p 0 .485 .795 r
F P
0 g
s
.49332 .42571 m .51007 .43203 L .52025 .4488 L closepath p .047 .579 .859 r
F P
0 g
s
.46552 .4054 m .47927 .39987 L .49332 .42571 L closepath p .18 .659 .93 r
F P
0 g
s
.45257 .38413 m .47927 .39987 L .46552 .4054 L closepath p .182 .657 .932 r
F P
0 g
s
.43755 .38818 m .45257 .38413 L .46552 .4054 L closepath p .254 .701 .958 r
F P
0 g
s
.59014 .52057 m .60082 .54761 L .61094 .54547 L closepath p .559 0 0 r
F P
0 g
s
.58068 .51915 m .60082 .54761 L .59014 .52057 L closepath p .546 0 0 r
F P
0 g
s
.49332 .42571 m .47927 .39987 L .51007 .43203 L closepath p .077 .587 .887 r
F P
0 g
s
.43755 .38818 m .42817 .36278 L .45257 .38413 L closepath p .262 .7 .962 r
F P
0 g
s
.42817 .36278 m .43755 .38818 L .40422 .36961 L closepath p .314 .733 .975 r
F P
0 g
s
.84519 .40548 m .86977 .37233 L .85495 .40181 L closepath p .503 .909 .884 r
F P
0 g
s
.61944 .57552 m .63025 .57089 L .61094 .54547 L closepath p .623 .105 0 r
F P
0 g
s
.61094 .54547 m .60082 .54761 L .61944 .57552 L closepath p .632 .093 0 r
F P
0 g
s
.42817 .36278 m .40422 .36961 L .39748 .33622 L closepath p .326 .731 .98 r
F P
0 g
s
.40422 .36961 m .36677 .3425 L .39748 .33622 L closepath p .365 .758 .987 r
F P
0 g
s
.8009 .47675 m .79134 .49912 L .7994 .4822 L p 0 0 0 r
F P
0 g
s
.8009 .47675 m .81078 .45976 L .80232 .47161 L p 0 0 0 r
F P
0 g
s
.81209 .45554 m .8224 .44041 L .81474 .45336 L p .357 .814 .705 r
F P
0 g
s
.81474 .45336 m .8009 .47675 L .81209 .45554 L p .357 .814 .705 r
F P
0 g
s
.7994 .4822 m .81209 .45554 L .8009 .47675 L p 0 0 0 r
F P
0 g
s
.81209 .45554 m .80266 .47346 L .80955 .45764 L p .473 .871 .696 r
F P
0 g
s
.80955 .45764 m .82116 .43099 L .81209 .45554 L p .473 .871 .696 r
F P
0 g
s
.80232 .47161 m .7916 .4866 L .8009 .47675 L p 0 0 0 r
F P
0 g
s
.77685 .51886 m .8009 .47675 L .7916 .4866 L closepath p 0 0 0 r
F P
0 g
s
.91611 .29887 m .92804 .32265 L .89691 .34516 L closepath p .631 .905 .984 r
F P
0 g
s
.63428 .59782 m .64664 .59888 L .63025 .57089 L closepath p .652 .173 .102 r
F P
0 g
s
.63428 .59782 m .63025 .57089 L .61944 .57552 L closepath p .677 .184 .07 r
F P
0 g
s
.69858 .59504 m .7085 .62334 L .71742 .58483 L closepath p .537 .138 .233 r
F P
0 g
s
.69858 .59504 m .69136 .62977 L .7085 .62334 L closepath p .541 .136 .222 r
F P
0 g
s
.69858 .59504 m .68049 .59963 L .69136 .62977 L closepath p .593 .194 .257 r
F P
0 g
s
.68049 .59963 m .67534 .63223 L .69136 .62977 L closepath p .598 .189 .237 r
F P
0 g
s
.67534 .63223 m .68049 .59963 L .66286 .59987 L closepath p .638 .232 .258 r
F P
0 g
s
.73641 .5674 m .71742 .58483 L .72727 .60809 L closepath p .46 .048 .171 r
F P
0 g
s
.7085 .62334 m .72727 .60809 L .71742 .58483 L closepath p .458 .05 .176 r
F P
0 g
s
.3334 .34009 m .33547 .3128 L .36677 .3425 L closepath p .397 .773 .992 r
F P
0 g
s
.3334 .34009 m .30605 .3189 L .33547 .3128 L closepath p .409 .785 .993 r
F P
0 g
s
.74365 .57792 m .75526 .54685 L .73641 .5674 L closepath p .351 0 .064 r
F P
0 g
s
.72727 .60809 m .74365 .57792 L .73641 .5674 L closepath p .348 0 .093 r
F P
0 g
s
.21751 .29452 m .24862 .31044 L .22136 .32264 L closepath p .425 .8 .993 r
F P
0 g
s
.22136 .32264 m .18944 .3076 L .21751 .29452 L closepath p .431 .805 .993 r
F P
0 g
s
.6603 .63263 m .67534 .63223 L .66286 .59987 L closepath p .643 .222 .228 r
F P
0 g
s
.6603 .63263 m .66286 .59987 L .64664 .59888 L closepath p .674 .252 .237 r
F P
0 g
s
.27832 .32725 m .27656 .29918 L .30605 .3189 L closepath p .418 .791 .994 r
F P
0 g
s
.27832 .32725 m .24862 .31044 L .27656 .29918 L closepath p .428 .798 .994 r
F P
0 g
s
.91611 .29887 m .89691 .34516 L .88658 .32974 L closepath p .622 .929 .969 r
F P
0 g
s
.86977 .37233 m .88658 .32974 L .89691 .34516 L closepath p .621 .929 .969 r
F P
0 g
s
.75526 .54685 m .75745 .56307 L .77685 .51886 L closepath p .207 0 0 r
F P
0 g
s
.74365 .57792 m .75745 .56307 L .75526 .54685 L closepath p .213 0 0 r
F P
0 g
s
.64538 .62844 m .6603 .63263 L .64664 .59888 L closepath p .676 .236 .198 r
F P
0 g
s
.63428 .59782 m .64538 .62844 L .64664 .59888 L closepath p .699 .259 .201 r
F P
0 g
s
.59283 .54707 m .60082 .54761 L .58068 .51915 L closepath p .54 0 0 r
F P
0 g
s
.57308 .51387 m .59283 .54707 L .58068 .51915 L closepath p .496 0 0 r
F P
0 g
s
.36567 .30703 m .39748 .33622 L .36677 .3425 L closepath p .382 .757 .991 r
F P
0 g
s
.36567 .30703 m .36677 .3425 L .33547 .3128 L closepath p .405 .777 .993 r
F P
0 g
s
.61944 .57552 m .60082 .54761 L .61074 .57833 L closepath p .633 .096 0 r
F P
0 g
s
.59283 .54707 m .61074 .57833 L .60082 .54761 L closepath p .625 .065 0 r
F P
0 g
s
.85856 .36952 m .88658 .32974 L .86977 .37233 L closepath p .602 .944 .931 r
F P
0 g
s
.84519 .40548 m .85856 .36952 L .86977 .37233 L closepath p .581 .937 .931 r
F P
0 g
s
.52853 .44595 m .5357 .46144 L .51007 .43203 L closepath p 0 .529 .842 r
F P
0 g
s
.45257 .38413 m .45229 .36497 L .47927 .39987 L closepath p .261 .684 .965 r
F P
0 g
s
.45257 .38413 m .42817 .36278 L .45229 .36497 L closepath p .292 .712 .972 r
F P
0 g
s
.62937 .61454 m .63428 .59782 L .61944 .57552 L closepath p .688 .21 .112 r
F P
0 g
s
.62937 .61454 m .61944 .57552 L .61074 .57833 L closepath p .702 .193 .021 r
F P
0 g
s
.83363 .41393 m .85856 .36952 L .84519 .40548 L closepath p .555 .933 .863 r
F P
0 g
s
.83363 .41393 m .84519 .40548 L .8224 .44041 L closepath p .486 .901 .845 r
F P
0 g
s
.52853 .44595 m .51007 .43203 L .50356 .41089 L closepath p .053 .549 .886 r
F P
0 g
s
.50356 .41089 m .51007 .43203 L .47927 .39987 L closepath p .151 .627 .925 r
F P
0 g
s
.83363 .41393 m .8224 .44041 L .81209 .45554 L closepath p .444 .867 .743 r
F P
0 g
s
.77685 .51886 m .77183 .54397 L .79134 .49912 L closepath p 0 0 0 r
F P
0 g
s
.8009 .47675 m .77685 .51886 L .79134 .49912 L closepath p 0 0 0 r
F P
0 g
s
.77685 .51886 m .75745 .56307 L .77183 .54397 L closepath p .031 0 0 r
F P
0 g
s
.62937 .61454 m .64538 .62844 L .63428 .59782 L closepath p .702 .228 .123 r
F P
0 g
s
.7085 .62334 m .69136 .62977 L .70095 .6651 L closepath p .583 .222 .322 r
F P
0 g
s
.69136 .62977 m .68418 .66246 L .70095 .6651 L closepath p .593 .22 .305 r
F P
0 g
s
.67534 .63223 m .68418 .66246 L .69136 .62977 L closepath p .637 .274 .338 r
F P
0 g
s
.67261 .66871 m .68418 .66246 L .67534 .63223 L closepath p .649 .268 .31 r
F P
0 g
s
.67261 .66871 m .67534 .63223 L .6603 .63263 L closepath p .678 .298 .319 r
F P
0 g
s
.74365 .57792 m .74198 .60344 L .75745 .56307 L closepath p .307 0 .104 r
F P
0 g
s
.72727 .60809 m .74198 .60344 L .74365 .57792 L closepath p .368 0 .119 r
F P
0 g
s
.7085 .62334 m .71543 .65734 L .72727 .60809 L closepath p .52 .161 .298 r
F P
0 g
s
.7085 .62334 m .70095 .6651 L .71543 .65734 L closepath p .53 .163 .288 r
F P
0 g
s
.30605 .3189 m .30527 .28879 L .33547 .3128 L closepath p .42 .785 .995 r
F P
0 g
s
.30605 .3189 m .27656 .29918 L .30527 .28879 L closepath p .433 .795 .996 r
F P
0 g
s
.24862 .31044 m .2459 .28129 L .27656 .29918 L closepath p .43 .798 .995 r
F P
0 g
s
.21751 .29452 m .2459 .28129 L .24862 .31044 L closepath p .438 .804 .995 r
F P
0 g
s
.42817 .36278 m .42466 .33312 L .45229 .36497 L closepath p .321 .713 .981 r
F P
0 g
s
.42817 .36278 m .39748 .33622 L .42466 .33312 L closepath p .36 .744 .988 r
F P
0 g
s
.47927 .39987 m .47763 .37423 L .50356 .41089 L closepath p .204 .636 .95 r
F P
0 g
s
.45229 .36497 m .47763 .37423 L .47927 .39987 L closepath p .256 .682 .964 r
F P
0 g
s
.65877 .66788 m .67261 .66871 L .6603 .63263 L closepath p .685 .289 .291 r
F P
0 g
s
.64538 .62844 m .65877 .66788 L .6603 .63263 L closepath p .712 .315 .295 r
F P
0 g
s
.74198 .60344 m .72727 .60809 L .72759 .64337 L closepath p .411 .054 .245 r
F P
0 g
s
.72727 .60809 m .71543 .65734 L .72759 .64337 L closepath p .455 .085 .247 r
F P
0 g
s
.75519 .58898 m .77183 .54397 L .75745 .56307 L closepath p .159 0 0 r
F P
0 g
s
.75745 .56307 m .74198 .60344 L .75519 .58898 L closepath p .209 0 .001 r
F P
0 g
s
.64538 .62844 m .64266 .65548 L .65877 .66788 L closepath p .72 .295 .239 r
F P
0 g
s
.64538 .62844 m .62937 .61454 L .64266 .65548 L closepath p .742 .32 .247 r
F P
0 g
s
.60348 .57566 m .61074 .57833 L .59283 .54707 L closepath p .621 .06 0 r
F P
0 g
s
.58808 .54617 m .60348 .57566 L .59283 .54707 L closepath p .562 0 0 r
F P
0 g
s
.57308 .51387 m .58808 .54617 L .59283 .54707 L closepath p .478 0 0 r
F P
0 g
s
.90394 .2746 m .91611 .29887 L .88658 .32974 L closepath p .68 .92 .973 r
F P
0 g
s
.61839 .61112 m .62937 .61454 L .61074 .57833 L closepath p .697 .184 .006 r
F P
0 g
s
.60348 .57566 m .61839 .61112 L .61074 .57833 L closepath p .703 .172 0 r
F P
0 g
s
.56644 .5028 m .57308 .51387 L .56017 .49353 L p 0 .322 .67 r
F P
0 g
s
.56017 .49353 m .55146 .47979 L .56644 .5028 L p 0 .322 .67 r
F P
0 g
s
.56644 .5028 m .56189 .48818 L .57196 .51098 L p 0 .307 .731 r
F P
0 g
s
.57308 .51387 m .56644 .5028 L .58808 .54617 L closepath p .399 0 0 r
F P
0 g
s
.57196 .51098 m .57788 .52441 L .56644 .5028 L p 0 .307 .731 r
F P
0 g
s
.54602 .46514 m .55146 .47979 L .52853 .44595 L closepath p 0 .474 .824 r
F P
0 g
s
.39491 .30078 m .42466 .33312 L .39748 .33622 L closepath p .378 .742 .991 r
F P
0 g
s
.39491 .30078 m .39748 .33622 L .36567 .30703 L closepath p .41 .766 .995 r
F P
0 g
s
.33547 .3128 m .33462 .27898 L .36567 .30703 L closepath p .42 .776 .996 r
F P
0 g
s
.33547 .3128 m .30527 .28879 L .33462 .27898 L closepath p .438 .791 .997 r
F P
0 g
s
.64266 .65548 m .62937 .61454 L .63455 .65521 L closepath p .747 .285 .153 r
F P
0 g
s
.61839 .61112 m .63455 .65521 L .62937 .61454 L closepath p .764 .294 .127 r
F P
0 g
s
.74198 .60344 m .74086 .63497 L .75519 .58898 L closepath p .286 0 .156 r
F P
0 g
s
.74198 .60344 m .72759 .64337 L .74086 .63497 L closepath p .337 0 .165 r
F P
0 g
s
.87389 .31804 m .90394 .2746 L .88658 .32974 L closepath p .68 .95 .956 r
F P
0 g
s
.85856 .36952 m .87389 .31804 L .88658 .32974 L closepath p .672 .948 .957 r
F P
0 g
s
.54602 .46514 m .52853 .44595 L .52441 .42692 L closepath p .019 .503 .874 r
F P
0 g
s
.52853 .44595 m .50356 .41089 L .52441 .42692 L closepath p .111 .582 .914 r
F P
0 g
s
.27461 .26792 m .30527 .28879 L .27656 .29918 L closepath p .436 .795 .996 r
F P
0 g
s
.27656 .29918 m .2459 .28129 L .27461 .26792 L closepath p .447 .803 .997 r
F P
0 g
s
.83363 .41393 m .81209 .45554 L .82116 .43099 L closepath p .52 .915 .812 r
F P
0 g
s
.81209 .45554 m .79134 .49912 L .80266 .47346 L closepath p .399 .827 .666 r
F P
0 g
s
.78341 .52294 m .80266 .47346 L .79134 .49912 L closepath p 0 0 0 r
F P
0 g
s
.78341 .52294 m .79134 .49912 L .77183 .54397 L closepath p 0 0 0 r
F P
0 g
s
.45142 .33591 m .47763 .37423 L .45229 .36497 L closepath p .307 .687 .977 r
F P
0 g
s
.45142 .33591 m .45229 .36497 L .42466 .33312 L closepath p .349 .724 .986 r
F P
0 g
s
.84181 .37828 m .85856 .36952 L .83363 .41393 L closepath p .621 .956 .909 r
F P
0 g
s
.84181 .37828 m .83363 .41393 L .82116 .43099 L closepath p .626 .961 .851 r
F P
0 g
s
.84181 .37828 m .87389 .31804 L .85856 .36952 L closepath p .664 .969 .912 r
F P
0 g
s
.76671 .57 m .78341 .52294 L .77183 .54397 L closepath p 0 0 0 r
F P
0 g
s
.75519 .58898 m .76671 .57 L .77183 .54397 L closepath p .004 0 0 r
F P
0 g
s
.50356 .41089 m .50157 .38711 L .52441 .42692 L closepath p .174 .596 .94 r
F P
0 g
s
.47763 .37423 m .50157 .38711 L .50356 .41089 L closepath p .239 .653 .961 r
F P
0 g
s
.60348 .57566 m .60728 .595 L .61839 .61112 L closepath p .649 .091 0 r
F P
0 g
s
.58808 .54617 m .60728 .595 L .60348 .57566 L closepath p .65 .092 0 r
F P
0 g
s
.75235 .61739 m .76671 .57 L .75519 .58898 L closepath p .119 0 .03 r
F P
0 g
s
.74086 .63497 m .75235 .61739 L .75519 .58898 L closepath p .18 0 .052 r
F P
0 g
s
.61839 .61112 m .62482 .64656 L .63455 .65521 L closepath p .749 .25 .039 r
F P
0 g
s
.61839 .61112 m .60728 .595 L .62482 .64656 L closepath p .757 .248 0 r
F P
0 g
s
.36567 .30703 m .36414 .26931 L .39491 .30078 L closepath p .422 .764 .996 r
F P
0 g
s
.36567 .30703 m .33462 .27898 L .36414 .26931 L closepath p .446 .784 .998 r
F P
0 g
s
.57788 .52441 m .57731 .51725 L .58326 .53391 L p 0 .271 .73 r
F P
0 g
s
.57788 .52441 m .58808 .54617 L .57225 .51445 L p 0 .275 .672 r
F P
0 g
s
.59543 .56795 m .58808 .54617 L .57788 .52441 L closepath p .474 0 0 r
F P
0 g
s
.58326 .53391 m .59543 .56795 L .57788 .52441 L p 0 .271 .73 r
F P
0 g
s
.57225 .51445 m .56644 .5028 L .57788 .52441 L p 0 .275 .672 r
F P
0 g
s
.54602 .46514 m .56189 .48818 L .56644 .5028 L closepath p 0 .42 .808 r
F P
0 g
s
.59543 .56795 m .60728 .595 L .58808 .54617 L closepath p .47 0 0 r
F P
0 g
s
.30364 .25439 m .33462 .27898 L .30527 .28879 L closepath p .443 .79 .998 r
F P
0 g
s
.27461 .26792 m .30364 .25439 L .30527 .28879 L closepath p .457 .801 .998 r
F P
0 g
s
.45142 .33591 m .42466 .33312 L .42355 .29751 L closepath p .376 .722 .989 r
F P
0 g
s
.42355 .29751 m .42466 .33312 L .39491 .30078 L closepath p .412 .752 .994 r
F P
0 g
s
.54323 .44758 m .56189 .48818 L .54602 .46514 L closepath p 0 .455 .86 r
F P
0 g
s
.52441 .42692 m .54323 .44758 L .54602 .46514 L closepath p .08 .537 .903 r
F P
0 g
s
.50157 .38711 m .47763 .37423 L .4778 .34508 L closepath p .292 .657 .971 r
F P
0 g
s
.4778 .34508 m .47763 .37423 L .45142 .33591 L closepath p .337 .7 .982 r
F P
0 g
s
.81517 .43371 m .84181 .37828 L .82116 .43099 L closepath p .597 .948 .818 r
F P
0 g
s
.80266 .47346 m .81517 .43371 L .82116 .43099 L closepath p .614 .954 .817 r
F P
0 g
s
.80266 .47346 m .79232 .49656 L .81517 .43371 L closepath p .545 .9 .686 r
F P
0 g
s
.80239 .47415 m .78341 .52294 L .79232 .49656 L p .484 .866 .661 r
F P
0 g
s
.78341 .52294 m .77537 .54956 L .79232 .49656 L closepath p 0 0 0 r
F P
0 g
s
.77537 .54956 m .78341 .52294 L .77931 .5345 L p 0 0 0 r
F P
0 g
s
.77931 .5345 m .76671 .57 L .77537 .54956 L p 0 0 0 r
F P
0 g
s
.77537 .54956 m .76833 .57585 L .77182 .5631 L p 0 0 0 r
F P
0 g
s
.77926 .53595 m .79232 .49656 L .77537 .54956 L closepath p .503 .853 .594 r
F P
0 g
s
.77182 .5631 m .77926 .53595 L .77537 .54956 L p 0 0 0 r
F P
0 g
s
.79232 .49656 m .77926 .53595 L .78428 .51445 L p .69 .963 .73 r
F P
0 g
s
.78428 .51445 m .78634 .50564 L .79232 .49656 L p .69 .963 .73 r
F P
0 g
s
.79232 .49656 m .80266 .47346 L .80239 .47415 L p .484 .866 .661 r
F P
0 g
s
.76671 .57 m .76202 .59378 L .77537 .54956 L closepath p 0 0 0 r
F P
0 g
s
.75235 .61739 m .76202 .59378 L .76671 .57 L closepath p 0 0 0 r
F P
0 g
s
.61342 .62311 m .62482 .64656 L .60728 .595 L closepath p .678 .132 0 r
F P
0 g
s
.61342 .62311 m .60728 .595 L .59543 .56795 L closepath p .67 .127 0 r
F P
0 g
s
.52346 .40528 m .54323 .44758 L .52441 .42692 L closepath p .153 .556 .928 r
F P
0 g
s
.50157 .38711 m .52346 .40528 L .52441 .42692 L closepath p .218 .616 .952 r
F P
0 g
s
.87389 .31804 m .89152 .24983 L .90394 .2746 L closepath p .727 .932 .957 r
F P
0 g
s
.56189 .48818 m .57731 .51725 L .57788 .52441 L closepath p 0 .354 .778 r
F P
0 g
s
.86062 .30627 m .89152 .24983 L .87389 .31804 L closepath p .733 .964 .939 r
F P
0 g
s
.84181 .37828 m .86062 .30627 L .87389 .31804 L closepath p .73 .963 .94 r
F P
0 g
s
.82894 .3742 m .86062 .30627 L .84181 .37828 L closepath p .723 .984 .9 r
F P
0 g
s
.82894 .3742 m .84181 .37828 L .81517 .43371 L closepath p .735 .986 .897 r
F P
0 g
s
.39491 .30078 m .39368 .2604 L .42355 .29751 L closepath p .426 .749 .994 r
F P
0 g
s
.39491 .30078 m .36414 .26931 L .39368 .2604 L closepath p .457 .774 .997 r
F P
0 g
s
.36414 .26931 m .33462 .27898 L .333 .24071 L closepath p .451 .782 .998 r
F P
0 g
s
.33462 .27898 m .30364 .25439 L .333 .24071 L closepath p .47 .797 .999 r
F P
0 g
s
.56011 .47189 m .57731 .51725 L .56189 .48818 L closepath p 0 .407 .844 r
F P
0 g
s
.56011 .47189 m .56189 .48818 L .54323 .44758 L closepath p .056 .493 .891 r
F P
0 g
s
.4778 .34508 m .45142 .33591 L .45267 .30009 L closepath p .373 .698 .983 r
F P
0 g
s
.45142 .33591 m .42355 .29751 L .45267 .30009 L closepath p .411 .734 .991 r
F P
0 g
s
.52346 .40528 m .50157 .38711 L .50265 .36046 L closepath p .277 .623 .96 r
F P
0 g
s
.50157 .38711 m .4778 .34508 L .50265 .36046 L closepath p .324 .67 .976 r
F P
0 g
s
.77537 .54956 m .76202 .59378 L .76833 .57585 L closepath p 0 0 0 r
F P
0 g
s
.59543 .56795 m .61093 .61544 L .61342 .62311 L closepath p 0 .079 .546 r
F P
0 g
s
.59291 .55418 m .61093 .61544 L .59543 .56795 L closepath p 0 .145 .643 r
F P
0 g
s
.59291 .55418 m .59543 .56795 L .57731 .51725 L closepath p 0 .279 .738 r
F P
0 g
s
.54323 .44758 m .54322 .42841 L .56011 .47189 L closepath p .136 .515 .914 r
F P
0 g
s
.54323 .44758 m .52346 .40528 L .54322 .42841 L closepath p .197 .577 .94 r
F P
0 g
s
.82894 .3742 m .81517 .43371 L .80355 .4495 L closepath p .719 .988 .831 r
F P
0 g
s
.79232 .49656 m .80355 .4495 L .81517 .43371 L closepath p .712 .986 .83 r
F P
0 g
s
.80355 .4495 m .79232 .49656 L .78634 .50564 L closepath p .671 .958 .731 r
F P
0 g
s
.5752 .4977 m .59291 .55418 L .57731 .51725 L closepath p 0 .35 .818 r
F P
0 g
s
.5752 .4977 m .57731 .51725 L .56011 .47189 L closepath p .045 .453 .878 r
F P
0 g
s
.4778 .34508 m .48108 .31262 L .50265 .36046 L closepath p .369 .669 .974 r
F P
0 g
s
.4778 .34508 m .45267 .30009 L .48108 .31262 L closepath p .405 .709 .985 r
F P
0 g
s
.52559 .38172 m .54322 .42841 L .52346 .40528 L closepath p .264 .588 .947 r
F P
0 g
s
.52346 .40528 m .50265 .36046 L .52559 .38172 L closepath p .308 .636 .965 r
F P
0 g
s
.36414 .26931 m .36269 .22688 L .39368 .2604 L closepath p .462 .772 .997 r
F P
0 g
s
.36414 .26931 m .333 .24071 L .36269 .22688 L closepath p .486 .791 .999 r
F P
0 g
s
.42401 .25397 m .45267 .30009 L .42355 .29751 L closepath p .432 .729 .988 r
F P
0 g
s
.42355 .29751 m .39368 .2604 L .42401 .25397 L closepath p .467 .76 .994 r
F P
0 g
s
.55936 .45338 m .5752 .4977 L .56011 .47189 L closepath p .111 .47 .893 r
F P
0 g
s
.55936 .45338 m .56011 .47189 L .54322 .42841 L closepath p .184 .54 .927 r
F P
0 g
s
.77926 .53595 m .76833 .57585 L .77378 .55174 L closepath p .621 .913 .633 r
F P
0 g
s
.77926 .53595 m .77378 .55174 L .78634 .50564 L closepath p .641 .921 .639 r
F P
0 g
s
.59291 .55418 m .60792 .59794 L .61093 .61544 L closepath p 0 .234 .731 r
F P
0 g
s
.86062 .30627 m .87883 .22454 L .89152 .24983 L closepath p .768 .939 .938 r
F P
0 g
s
.59291 .55418 m .59041 .52837 L .60792 .59794 L closepath p 0 .308 .799 r
F P
0 g
s
.5752 .4977 m .59041 .52837 L .59291 .55418 L closepath p .084 .435 .875 r
F P
0 g
s
.54725 .41048 m .55936 .45338 L .54322 .42841 L closepath p .26 .555 .931 r
F P
0 g
s
.54725 .41048 m .54322 .42841 L .52559 .38172 L closepath p .29 .6 .95 r
F P
0 g
s
.50265 .36046 m .50655 .33188 L .52559 .38172 L closepath p .357 .638 .961 r
F P
0 g
s
.50265 .36046 m .48108 .31262 L .50655 .33188 L closepath p .391 .678 .974 r
F P
0 g
s
.84839 .2901 m .87883 .22454 L .86062 .30627 L closepath p .777 .965 .924 r
F P
0 g
s
.82894 .3742 m .84839 .2901 L .86062 .30627 L closepath p .785 .966 .92 r
F P
0 g
s
.5752 .4977 m .57063 .45833 L .59041 .52837 L closepath p .139 .449 .881 r
F P
0 g
s
.55936 .45338 m .57063 .45833 L .5752 .4977 L closepath p .246 .533 .92 r
F P
0 g
s
.79042 .47415 m .80764 .41122 L .80355 .4495 L closepath p .801 .999 .844 r
F P
0 g
s
.80764 .41122 m .82894 .3742 L .80355 .4495 L closepath p .785 .989 .889 r
F P
0 g
s
.80355 .4495 m .78634 .50564 L .79042 .47415 L closepath p .793 .999 .843 r
F P
0 g
s
.78634 .50564 m .77837 .52348 L .79042 .47415 L closepath p .795 .996 .782 r
F P
0 g
s
.78634 .50564 m .77378 .55174 L .77837 .52348 L closepath p .795 .996 .782 r
F P
0 g
s
.45843 .25371 m .48108 .31262 L .45267 .30009 L closepath p .441 .704 .977 r
F P
0 g
s
.42401 .25397 m .45843 .25371 L .45267 .30009 L closepath p .472 .74 .987 r
F P
0 g
s
.39273 .21289 m .42401 .25397 L .39368 .2604 L closepath p .474 .757 .992 r
F P
0 g
s
.39368 .2604 m .36269 .22688 L .39273 .21289 L closepath p .504 .781 .995 r
F P
0 g
s
.54725 .41048 m .57063 .45833 L .55936 .45338 L closepath p .235 .542 .926 r
F P
0 g
s
.82472 .34767 m .82894 .3742 L .80764 .41122 L closepath p .809 .984 .889 r
F P
0 g
s
.82472 .34767 m .84839 .2901 L .82894 .3742 L closepath p .786 .979 .908 r
F P
0 g
s
.60437 .56346 m .60792 .59794 L .59041 .52837 L closepath p .108 .403 .855 r
F P
0 g
s
.54725 .41048 m .52559 .38172 L .53035 .35503 L closepath p .351 .606 .944 r
F P
0 g
s
.53035 .35503 m .52559 .38172 L .50655 .33188 L closepath p .383 .648 .962 r
F P
0 g
s
.60437 .56346 m .59041 .52837 L .59113 .50083 L closepath p .179 .423 .858 r
F P
0 g
s
.59113 .50083 m .59041 .52837 L .57063 .45833 L closepath p .266 .512 .903 r
F P
0 g
s
.50655 .33188 m .48108 .31262 L .48859 .2862 L closepath p .429 .675 .966 r
F P
0 g
s
.45843 .25371 m .48859 .2862 L .48108 .31262 L closepath p .447 .707 .977 r
F P
0 g
s
.79042 .47415 m .77837 .52348 L .77582 .50635 L closepath p .849 .988 .854 r
F P
0 g
s
.54725 .41048 m .55409 .38906 L .57063 .45833 L closepath p .343 .57 .923 r
F P
0 g
s
.53035 .35503 m .55409 .38906 L .54725 .41048 L closepath p .367 .614 .945 r
F P
0 g
s
.79042 .44536 m .79042 .47415 L .77582 .50635 L closepath p .851 .987 .854 r
F P
0 g
s
.80764 .41122 m .79042 .47415 L .79042 .44536 L closepath p .831 .98 .88 r
F P
0 g
s
.59113 .50083 m .57063 .45833 L .57652 .43093 L closepath p .32 .521 .895 r
F P
0 g
s
.55409 .38906 m .57652 .43093 L .57063 .45833 L closepath p .358 .577 .924 r
F P
0 g
s
.42401 .25397 m .42311 .19873 L .45843 .25371 L closepath p .484 .735 .982 r
F P
0 g
s
.39273 .21289 m .42311 .19873 L .42401 .25397 L closepath p .522 .767 .986 r
F P
0 g
s
.59113 .50083 m .60837 .55034 L .60437 .56346 L closepath p .3 .488 .876 r
F P
0 g
s
.53035 .35503 m .50655 .33188 L .51121 .29637 L closepath p .42 .646 .951 r
F P
0 g
s
.50655 .33188 m .48859 .2862 L .51121 .29637 L closepath p .45 .683 .964 r
F P
0 g
s
.80764 .41122 m .79042 .44536 L .80724 .37539 L closepath p .838 .977 .879 r
F P
0 g
s
.82472 .34767 m .80764 .41122 L .80724 .37539 L closepath p .816 .967 .902 r
F P
0 g
s
.84839 .2901 m .86588 .19871 L .87883 .22454 L closepath p .797 .942 .922 r
F P
0 g
s
.83096 .28387 m .84839 .2901 L .82472 .34767 L closepath p .81 .953 .911 r
F P
0 g
s
.83096 .28387 m .86588 .19871 L .84839 .2901 L closepath p .802 .955 .916 r
F P
0 g
s
.59626 .47657 m .60837 .55034 L .59113 .50083 L closepath p .34 .499 .871 r
F P
0 g
s
.59626 .47657 m .59113 .50083 L .57652 .43093 L closepath p .375 .55 .899 r
F P
0 g
s
.53035 .35503 m .53736 .32239 L .55409 .38906 L closepath p .419 .616 .932 r
F P
0 g
s
.53035 .35503 m .51121 .29637 L .53736 .32239 L closepath p .447 .657 .95 r
F P
0 g
s
.79042 .44536 m .77582 .50635 L .77357 .4858 L closepath p .858 .962 .871 r
F P
0 g
s
.83096 .28387 m .82472 .34767 L .80724 .37539 L closepath p .821 .964 .9 r
F P
0 g
s
.59626 .47657 m .61316 .52804 L .60837 .55034 L closepath p .378 .521 .874 r
F P
0 g
s
.48859 .2862 m .48591 .23869 L .51121 .29637 L closepath p .462 .678 .958 r
F P
0 g
s
.45843 .25371 m .48591 .23869 L .48859 .2862 L closepath p .501 .708 .963 r
F P
0 g
s
.56215 .35923 m .57652 .43093 L .55409 .38906 L closepath p .413 .584 .91 r
F P
0 g
s
.53736 .32239 m .56215 .35923 L .55409 .38906 L closepath p .438 .624 .931 r
F P
0 g
s
.79042 .44536 m .77357 .4858 L .78748 .41945 L closepath p .861 .956 .869 r
F P
0 g
s
.78748 .41945 m .80724 .37539 L .79042 .44536 L closepath p .838 .951 .891 r
F P
0 g
s
.45385 .18441 m .48591 .23869 L .45843 .25371 L closepath p .498 .714 .968 r
F P
0 g
s
.45843 .25371 m .42311 .19873 L .45385 .18441 L closepath p .54 .745 .971 r
F P
0 g
s
.58452 .40163 m .59626 .47657 L .57652 .43093 L closepath p .419 .557 .887 r
F P
0 g
s
.58452 .40163 m .57652 .43093 L .56215 .35923 L closepath p .444 .599 .91 r
F P
0 g
s
.80724 .37539 m .78748 .41945 L .80178 .34494 L closepath p .842 .943 .888 r
F P
0 g
s
.80724 .37539 m .80178 .34494 L .83096 .28387 L closepath p .819 .938 .906 r
F P
0 g
s
.51121 .29637 m .51864 .24386 L .53736 .32239 L closepath p .477 .652 .936 r
F P
0 g
s
.51121 .29637 m .48591 .23869 L .51864 .24386 L closepath p .508 .69 .95 r
F P
0 g
s
.59626 .47657 m .60364 .45009 L .61316 .52804 L closepath p .427 .534 .863 r
F P
0 g
s
.58452 .40163 m .60364 .45009 L .59626 .47657 L closepath p .451 .575 .888 r
F P
0 g
s
.78748 .41945 m .77357 .4858 L .77163 .46341 L closepath p .854 .928 .876 r
F P
0 g
s
.60364 .45009 m .61874 .49037 L .61316 .52804 L closepath p .469 .559 .865 r
F P
0 g
s
.54759 .29407 m .56215 .35923 L .53736 .32239 L closepath p .472 .623 .917 r
F P
0 g
s
.51864 .24386 m .54759 .29407 L .53736 .32239 L closepath p .488 .656 .935 r
F P
0 g
s
.78748 .41945 m .77163 .46341 L .77991 .40269 L closepath p .855 .915 .872 r
F P
0 g
s
.80178 .34494 m .78748 .41945 L .77991 .40269 L closepath p .832 .917 .893 r
F P
0 g
s
.83096 .28387 m .85265 .17234 L .86588 .19871 L closepath p .808 .942 .914 r
F P
0 g
s
.57258 .31685 m .58452 .40163 L .56215 .35923 L closepath p .476 .6 .897 r
F P
0 g
s
.54759 .29407 m .57258 .31685 L .56215 .35923 L closepath p .498 .635 .914 r
F P
0 g
s
.77991 .40269 m .77163 .46341 L .76058 .45931 L closepath p .837 .889 .878 r
F P
0 g
s
.80178 .34494 m .8156 .2544 L .83096 .28387 L closepath p .822 .932 .904 r
F P
0 g
s
.8156 .2544 m .85265 .17234 L .83096 .28387 L closepath p .804 .932 .917 r
F P
0 g
s
.48591 .23869 m .48494 .16993 L .51864 .24386 L closepath p .512 .687 .946 r
F P
0 g
s
.45385 .18441 m .48494 .16993 L .48591 .23869 L closepath p .554 .722 .952 r
F P
0 g
s
.58452 .40163 m .59569 .37811 L .60364 .45009 L closepath p .483 .58 .876 r
F P
0 g
s
.57258 .31685 m .59569 .37811 L .58452 .40163 L closepath p .494 .611 .897 r
F P
0 g
s
.6133 .42737 m .61874 .49037 L .60364 .45009 L closepath p .491 .562 .856 r
F P
0 g
s
.59569 .37811 m .6133 .42737 L .60364 .45009 L closepath p .504 .594 .877 r
F P
0 g
s
.6133 .42737 m .62842 .48036 L .61874 .49037 L closepath p .53 .588 .857 r
F P
0 g
s
.77991 .40269 m .78897 .33514 L .80178 .34494 L closepath p .832 .906 .889 r
F P
0 g
s
.78897 .33514 m .8156 .2544 L .80178 .34494 L closepath p .81 .913 .91 r
F P
0 g
s
.76866 .39402 m .77991 .40269 L .76058 .45931 L closepath p .835 .875 .872 r
F P
0 g
s
.78897 .33514 m .77991 .40269 L .76866 .39402 L closepath p .817 .884 .893 r
F P
0 g
s
.76866 .39402 m .76058 .45931 L .75094 .44899 L closepath p .818 .853 .874 r
F P
0 g
s
.54759 .29407 m .5471 .23103 L .57258 .31685 L closepath p .502 .633 .911 r
F P
0 g
s
.51864 .24386 m .5471 .23103 L .54759 .29407 L closepath p .531 .658 .917 r
F P
0 g
s
.62502 .40747 m .63999 .46407 L .62842 .48036 L closepath p .554 .602 .856 r
F P
0 g
s
.62502 .40747 m .62842 .48036 L .6133 .42737 L closepath p .538 .59 .854 r
F P
0 g
s
.5164 .15527 m .5471 .23103 L .51864 .24386 L closepath p .528 .667 .925 r
F P
0 g
s
.51864 .24386 m .48494 .16993 L .5164 .15527 L closepath p .565 .696 .929 r
F P
0 g
s
.62502 .40747 m .6133 .42737 L .60819 .35893 L closepath p .544 .611 .87 r
F P
0 g
s
.60819 .35893 m .6133 .42737 L .59569 .37811 L closepath p .524 .597 .869 r
F P
0 g
s
.60819 .35893 m .59569 .37811 L .59163 .31674 L closepath p .529 .617 .884 r
F P
0 g
s
.57258 .31685 m .59163 .31674 L .59569 .37811 L closepath p .528 .617 .884 r
F P
0 g
s
.57258 .31685 m .57914 .23728 L .59163 .31674 L closepath p .525 .624 .893 r
F P
0 g
s
.57258 .31685 m .5471 .23103 L .57914 .23728 L closepath p .546 .647 .9 r
F P
0 g
s
.76866 .39402 m .75094 .44899 L .75653 .38572 L closepath p .814 .838 .868 r
F P
0 g
s
.75653 .38572 m .75094 .44899 L .74274 .43458 L closepath p .798 .819 .869 r
F P
0 g
s
.63913 .38573 m .65347 .43728 L .63999 .46407 L closepath p .589 .622 .852 r
F P
0 g
s
.62502 .40747 m .63913 .38573 L .63999 .46407 L closepath p .569 .606 .85 r
F P
0 g
s
.78897 .33514 m .76866 .39402 L .77464 .33314 L closepath p .815 .87 .887 r
F P
0 g
s
.76866 .39402 m .75653 .38572 L .77464 .33314 L closepath p .797 .849 .89 r
F P
0 g
s
.79261 .28278 m .8156 .2544 L .78897 .33514 L closepath p .807 .891 .904 r
F P
0 g
s
.79261 .28278 m .78897 .33514 L .77464 .33314 L closepath p .798 .882 .907 r
F P
0 g
s
.8156 .2544 m .83913 .14539 L .85265 .17234 L closepath p .8 .942 .92 r
F P
0 g
s
.62098 .3298 m .63913 .38573 L .62502 .40747 L closepath p .575 .627 .866 r
F P
0 g
s
.62098 .3298 m .62502 .40747 L .60819 .35893 L closepath p .555 .612 .865 r
F P
0 g
s
.60234 .29436 m .62098 .3298 L .60819 .35893 L closepath p .562 .631 .877 r
F P
0 g
s
.60819 .35893 m .59163 .31674 L .60234 .29436 L closepath p .544 .619 .878 r
F P
0 g
s
.75653 .38572 m .74274 .43458 L .74213 .38386 L closepath p .787 .794 .858 r
F P
0 g
s
.74213 .38386 m .74274 .43458 L .72601 .44046 L closepath p .766 .778 .863 r
F P
0 g
s
.79261 .28278 m .80254 .21385 L .8156 .2544 L closepath p .796 .909 .918 r
F P
0 g
s
.80254 .21385 m .83913 .14539 L .8156 .2544 L closepath p .786 .908 .925 r
F P
0 g
s
.59163 .31674 m .57914 .23728 L .60234 .29436 L closepath p .551 .638 .89 r
F P
0 g
s
.65589 .37264 m .66628 .43942 L .65347 .43728 L closepath p .62 .643 .851 r
F P
0 g
s
.65589 .37264 m .65347 .43728 L .63913 .38573 L closepath p .601 .624 .847 r
F P
0 g
s
.54823 .14044 m .57914 .23728 L .5471 .23103 L closepath p .544 .649 .903 r
F P
0 g
s
.5164 .15527 m .54823 .14044 L .5471 .23103 L closepath p .572 .674 .908 r
F P
0 g
s
.77464 .33314 m .75653 .38572 L .75964 .32064 L closepath p .793 .833 .883 r
F P
0 g
s
.75964 .32064 m .75653 .38572 L .74213 .38386 L closepath p .772 .814 .886 r
F P
0 g
s
.7266 .38278 m .74213 .38386 L .72601 .44046 L closepath p .757 .759 .854 r
F P
0 g
s
.7266 .38278 m .72601 .44046 L .71148 .43565 L closepath p .733 .74 .857 r
F P
0 g
s
.6743 .37004 m .68152 .43668 L .66628 .43942 L closepath p .642 .659 .851 r
F P
0 g
s
.6743 .37004 m .66628 .43942 L .65589 .37264 L closepath p .632 .648 .848 r
F P
0 g
s
.63943 .30414 m .65589 .37264 L .63913 .38573 L closepath p .601 .645 .865 r
F P
0 g
s
.63913 .38573 m .62098 .3298 L .63943 .30414 L closepath p .582 .628 .863 r
F P
0 g
s
.77464 .33314 m .78068 .27872 L .79261 .28278 L closepath p .796 .869 .902 r
F P
0 g
s
.77464 .33314 m .75964 .32064 L .78068 .27872 L closepath p .774 .842 .904 r
F P
0 g
s
.69344 .36957 m .69921 .42459 L .68152 .43668 L closepath p .671 .683 .852 r
F P
0 g
s
.6743 .37004 m .69344 .36957 L .68152 .43668 L closepath p .662 .67 .847 r
F P
0 g
s
.7266 .38278 m .71148 .43565 L .71115 .37404 L closepath p .728 .729 .851 r
F P
0 g
s
.71115 .37404 m .71148 .43565 L .69921 .42459 L closepath p .705 .711 .853 r
F P
0 g
s
.78068 .27872 m .80254 .21385 L .79261 .28278 L closepath p .775 .879 .923 r
F P
0 g
s
.74096 .3386 m .75964 .32064 L .74213 .38386 L closepath p .761 .788 .875 r
F P
0 g
s
.74096 .3386 m .74213 .38386 L .7266 .38278 L closepath p .746 .778 .879 r
F P
0 g
s
.62098 .3298 m .61164 .23042 L .63943 .30414 L closepath p .586 .647 .877 r
F P
0 g
s
.60234 .29436 m .61164 .23042 L .62098 .3298 L closepath p .561 .631 .878 r
F P
0 g
s
.60234 .29436 m .57914 .23728 L .61164 .23042 L closepath p .567 .639 .882 r
F P
0 g
s
.71115 .37404 m .69921 .42459 L .69344 .36957 L closepath p .696 .696 .846 r
F P
0 g
s
.65589 .37264 m .66275 .29991 L .6743 .37004 L closepath p .628 .665 .867 r
F P
0 g
s
.66275 .29991 m .65589 .37264 L .63943 .30414 L closepath p .612 .648 .862 r
F P
0 g
s
.58043 .12543 m .61164 .23042 L .57914 .23728 L closepath p .565 .647 .89 r
F P
0 g
s
.54823 .14044 m .58043 .12543 L .57914 .23728 L closepath p .576 .657 .892 r
F P
0 g
s
.74096 .3386 m .7266 .38278 L .72765 .32326 L closepath p .741 .765 .872 r
F P
0 g
s
.7266 .38278 m .71115 .37404 L .72765 .32326 L closepath p .715 .743 .874 r
F P
0 g
s
.7734 .24524 m .78068 .27872 L .75964 .32064 L closepath p .774 .857 .912 r
F P
0 g
s
.74839 .28204 m .7734 .24524 L .75964 .32064 L closepath p .744 .81 .904 r
F P
0 g
s
.74096 .3386 m .74839 .28204 L .75964 .32064 L closepath p .756 .812 .897 r
F P
0 g
s
.68602 .30259 m .69344 .36957 L .6743 .37004 L closepath p .655 .689 .869 r
F P
0 g
s
.68602 .30259 m .6743 .37004 L .66275 .29991 L closepath p .643 .672 .863 r
F P
0 g
s
.7734 .24524 m .80254 .21385 L .78068 .27872 L closepath p .77 .856 .915 r
F P
0 g
s
.70849 .3083 m .72765 .32326 L .71115 .37404 L closepath p .711 .734 .869 r
F P
0 g
s
.70849 .3083 m .71115 .37404 L .69344 .36957 L closepath p .685 .715 .872 r
F P
0 g
s
.69344 .36957 m .68602 .30259 L .70849 .3083 L closepath p .676 .7 .865 r
F P
0 g
s
.74096 .3386 m .72765 .32326 L .74839 .28204 L closepath p .725 .773 .89 r
F P
0 g
s
.66275 .29991 m .63943 .30414 L .6462 .2073 L closepath p .606 .665 .881 r
F P
0 g
s
.63943 .30414 m .61164 .23042 L .6462 .2073 L closepath p .588 .647 .876 r
F P
0 g
s
.80254 .21385 m .82532 .11785 L .83913 .14539 L closepath p .77 .939 .937 r
F P
0 g
s
.79038 .16717 m .82532 .11785 L .80254 .21385 L closepath p .749 .887 .942 r
F P
0 g
s
.7734 .24524 m .79038 .16717 L .80254 .21385 L closepath p .753 .888 .94 r
F P
0 g
s
.74839 .28204 m .72765 .32326 L .73008 .25469 L closepath p .727 .78 .895 r
F P
0 g
s
.73008 .25469 m .72765 .32326 L .70849 .3083 L closepath p .691 .748 .895 r
F P
0 g
s
.61303 .11025 m .6462 .2073 L .61164 .23042 L closepath p .587 .659 .887 r
F P
0 g
s
.58043 .12543 m .61303 .11025 L .61164 .23042 L closepath p .577 .65 .886 r
F P
0 g
s
.6747 .24139 m .68602 .30259 L .66275 .29991 L closepath p .633 .69 .885 r
F P
0 g
s
.6462 .2073 m .6747 .24139 L .66275 .29991 L closepath p .621 .672 .877 r
F P
0 g
s
.75815 .20823 m .7734 .24524 L .74839 .28204 L closepath p .742 .834 .921 r
F P
0 g
s
.74839 .28204 m .73008 .25469 L .75815 .20823 L closepath p .703 .785 .915 r
F P
0 g
s
.70257 .2267 m .73008 .25469 L .70849 .3083 L closepath p .688 .739 .89 r
F P
0 g
s
.68602 .30259 m .70257 .2267 L .70849 .3083 L closepath p .66 .721 .894 r
F P
0 g
s
.6747 .24139 m .70257 .2267 L .68602 .30259 L closepath p .643 .694 .882 r
F P
0 g
s
.75815 .20823 m .79038 .16717 L .7734 .24524 L closepath p .722 .832 .933 r
F P
0 g
s
.73503 .17628 m .75815 .20823 L .73008 .25469 L closepath p .704 .798 .923 r
F P
0 g
s
.70257 .2267 m .73503 .17628 L .73008 .25469 L closepath p .657 .751 .919 r
F P
0 g
s
.6747 .24139 m .67607 .16542 L .70257 .2267 L closepath p .64 .711 .899 r
F P
0 g
s
.6462 .2073 m .67607 .16542 L .6747 .24139 L closepath p .603 .679 .895 r
F P
0 g
s
.6462 .2073 m .64601 .09488 L .67607 .16542 L closepath p .603 .68 .895 r
F P
0 g
s
.61303 .11025 m .64601 .09488 L .6462 .2073 L closepath p .576 .657 .892 r
F P
0 g
s
.79038 .16717 m .8112 .08971 L .82532 .11785 L closepath p .714 .929 .962 r
F P
0 g
s
.75815 .20823 m .77134 .12562 L .79038 .16717 L closepath p .711 .866 .956 r
F P
0 g
s
.73503 .17628 m .77134 .12562 L .75815 .20823 L closepath p .662 .801 .949 r
F P
0 g
s
.70257 .2267 m .70774 .13599 L .73503 .17628 L closepath p .657 .754 .921 r
F P
0 g
s
.70257 .2267 m .67607 .16542 L .70774 .13599 L closepath p .611 .709 .914 r
F P
0 g
s
.77134 .12562 m .8112 .08971 L .79038 .16717 L closepath p .689 .863 .964 r
F P
0 g
s
.67607 .16542 m .67939 .07933 L .70774 .13599 L closepath p .611 .714 .918 r
F P
0 g
s
.67607 .16542 m .64601 .09488 L .67939 .07933 L closepath p .571 .677 .911 r
F P
0 g
s
.73503 .17628 m .74035 .10394 L .77134 .12562 L closepath p .661 .81 .955 r
F P
0 g
s
.73503 .17628 m .70774 .13599 L .74035 .10394 L closepath p .608 .757 .949 r
F P
0 g
s
.70774 .13599 m .71318 .06359 L .74035 .10394 L closepath p .608 .759 .951 r
F P
0 g
s
.70774 .13599 m .67939 .07933 L .71318 .06359 L closepath p .56 .71 .941 r
F P
0 g
s
.79677 .06094 m .8112 .08971 L .77134 .12562 L closepath p .623 .902 .985 r
F P
0 g
s
.77134 .12562 m .76754 .07258 L .79677 .06094 L closepath p .616 .838 .985 r
F P
0 g
s
.74035 .10394 m .76754 .07258 L .77134 .12562 L closepath p .585 .817 .988 r
F P
0 g
s
.74035 .10394 m .74739 .04765 L .76754 .07258 L closepath p .587 .813 .986 r
F P
0 g
s
.74035 .10394 m .71318 .06359 L .74739 .04765 L closepath p .534 .753 .977 r
F P
0 g
s
.76754 .07258 m .78201 .03152 L .79677 .06094 L closepath p .513 .857 .994 r
F P
0 g
s
.76754 .07258 m .74739 .04765 L .78201 .03152 L closepath p .486 .791 .999 r
F P
0 g
s
.25 Mabswid
.79826 0 m
.99671 .40534 L
s
.99671 .40534 m
1 .76812 L
s
1 .76812 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.00338 .37369 m
0 .74137 L
s
0 .74137 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.79826 0 m
.00338 .37369 L
s
.00338 .37369 m
.79826 0 L
s
.01971 .36602 m
.0195 .38939 L
s
[(0)] .02012 .31927 -0.03303 1 Mshowa
.29667 .23581 m
.29658 .25919 L
s
[(2)] .29685 .18904 -0.01416 1 Mshowa
.60686 .08998 m
.6069 .11337 L
s
[(4)] .60677 .04321 .00709 1 Mshowa
.125 Mabswid
.08613 .33479 m
.08602 .34881 L
s
.15438 .3027 m
.15429 .31673 L
s
.22453 .26972 m
.22446 .28375 L
s
.37088 .20092 m
.37084 .21495 L
s
.44725 .16502 m
.44723 .17905 L
s
.52587 .12806 m
.52588 .14209 L
s
.69031 .05075 m
.69036 .06478 L
s
.77634 .01031 m
.77641 .02434 L
s
gsave
.37156 -0.0797 -66.5429 -15.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(x) show
70.187500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
0 .74137 m
.2826 1 L
s
.00774 .74845 m
.00796 .72508 L
s
[(0)] .00732 .7952 .03362 -1 Mshowa
.1058 .8382 m
.10597 .81482 L
s
[(1)] .10546 .88495 .02692 -1 Mshowa
.1951 .91992 m
.19523 .89654 L
s
[(2)] .19484 .96668 .02085 -1 Mshowa
.27675 .99465 m
.27685 .97127 L
s
[(3)] .27656 1.04142 .01531 -1 Mshowa
.125 Mabswid
.02812 .7671 m
.02824 .75308 L
s
.0481 .78539 m
.04822 .77136 L
s
.0677 .80333 m
.06781 .7893 L
s
.08693 .82093 m
.08704 .8069 L
s
.12432 .85515 m
.12442 .84112 L
s
.1425 .87179 m
.1426 .85776 L
s
.16035 .88812 m
.16044 .8741 L
s
.17788 .90417 m
.17797 .89014 L
s
.21201 .9354 m
.21208 .92137 L
s
.22862 .9506 m
.22869 .93657 L
s
.24494 .96554 m
.24501 .95151 L
s
.26099 .98022 m
.26105 .96619 L
s
gsave
.14967 1.16054 -66.7909 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.125000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(y) show
70.312500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/0P10`4000010?ooo`000<c<c;no_`00
E@00ZW=cLiVIVNfXL<V>GoOI`7eI>ooooiVIVH:2P]7AdTi>C_ooohVHZWB1T<[AfDI=E_ooofQ_PUQN
K[ZmaS4e?_oooiBDU:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJF
UWemOFATI4];Bc8b<P0T000TE@0TZP0To`190019E@19ZP19o`1]001]E@1]ZP1]o`2B002BE@2BZP2B
o`2f002fE@2fZP2fo`3K003KE@3KZP3Ko`3oE@3oZUD005D0EED0ZUD0oeDT05DTEEDTZUDToeE905E9
EEE9ZUE9oeE]05E]EEE]ZUE]oeFB05FBEEFBZUFBoeFf05FfEEFfZUFfoeGK05GKEEGKZUGKoeGo05Go
EEGoZUGoojX00:X0EJX0ZZX0ojXT0:XTEJXTZZXTojY90:Y9EJY9ZZY9ojY]0:Y]EJY]ZZY]ojZB0:ZB
EJZBZZZBojZf0:ZfEJZfZZZfoj[K0:[KEJ[KZZ[Koj[o0:[oEJ[oZZ[oool0EOl0Z_lT0?lTEOlTZ_lT
oom90?m9EOm9Z_m9oom]0?m]EOm]Z_m]oonB0?nBEOnBZ_nBoonf0?nfEOnfZ_nfoooK0?oKEOoKZ_oK
ooooEOooZ/d0003=0<g=0000cLd0c@3=cNGUiDe=C@00:P00O`00e00Z000Z:P0ZE@0ZO`0ZZP0Ze01E
001E:P1EE@1EO`1EZP1Ee01o001o:P1oE@1oO`1oZP1oe02Z002Z:P2ZE@2ZO`2ZZP2Ze03D003D:P3D
E@3DO`3DZP3De2X002X0:RX0EBX0ObX0ZRX0e2XZ02XZ:RXZEBXZObXZZRXZe2YE02YE:RYEEBYEObYE
ZRYEe2Yo02Yo:RYoEBYoObYoZRYoe2ZZ02ZZ:RZZEBZZObZZZRZZe2[D02[D:R[DEB[DOb[DZR[De5D0
:UD0OeD0e5DZ05DZ:UDZEEDZOeDZZUDZe5EE05EE:UEEEOl0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000;h00`4200<1jP000;l000<100000P000`40003Y0000`0001040007/0000
`@020Nd000310081k@000<0000@10001k0000;l000<100000P000`40003Y0000_P030@800`7Z0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`1;0081I0000?l0
B@020@03004006<0003o04L00P430081H`000?l0A@020@D00P5S0000o`0>00@1<@030@H000<10040
HP000?l040000`40000^00810P000`40000400030@010680003o010000<10000;0020@@000<10000
10001040005Q0000o`0;00L1:`020@H000<1000010001040005Q0000o`0;00030@00008000<10000
9`030@P000<1000010001@4000010600003o00`000D100000@0W00814P001@4000010600003o00`0
00D100000@0U00813@020@D000<100000P000`40001M0000o`0=00040@000B<00P4=008100<T0@00
10000`40000200030@0005d0003o00d000@100018@020@d00P401@0V0BH100@000<100000`000`40
001L0000o`0>00030@0101l00P4=008100LT02@00@0100@000<100000`000`40001L0000o`0>0003
0@0101d00P4=008100X09P0V02H19P010`000`40000400030@0005/0003o00l00P4K008130030@0<
900T02@0904V02H10`000`40000400030@0005/0003o02T00`4<0081010V0@409P0V02H00@0V02H1
0P000`40000500030@0005X0003o02L00`4<0081018090010040900T02@19P0V004200030@0000D0
00<10000FP000?l09@020@8000<1000020020@0G9P0V02H19P0102H09P0102H09P01004020000`40
001I0000o`0S008110000`400006008101T0900T02@00@0V0B@0900102H09P0V0@0100P000<10000
F@000?l08@020@H000<100000`030@0H9P0V02H09P0102H09P4V02H19P0V02H00`4900030@0005P0
003o01l00P4800040@00008101PT0@40900T02@0904V02H09P4T00409P030@03004100T000<10000
F0000?l07@020@`00P400`0V00020@0?9P0V02H09P0102H09P0V008100@00@410`000`400@0:0003
0@0005L0003o01/00P4<008101LT02@0904T0B@0900T02@19P0V02H09P020@0802H09P0V0BH20@T0
00<10000E`000?l060030@`00P40200V02H09P010P003P4V02H09P0102H09P0V0P4600030@000081
2P000`40001F0000o`0F00813@020@0X900T02@0900T0B@00@0T02@0904V02H09P4V0BH09P0V0040
9P4V0@T000<10000EP000?l050020@d00P40300102H09P0V02H00@<000l102H09P0102H09P4V0040
1P001P4000400@X000<10000E@000?l04P020@d00P400b@090020@0X02@0900T02@1900T0040900T
0BH09P4V02H19P0V02H19P0V0BH00@T000<10000E@000?l040020@d00P401@0V02H0008100TV02H0
9P0V00401@003@409P0102H00@0V00401@001@400001008000<1000020000`40001D0000o`0>0081
3@020@0b900T02@0904T0B@0900T02@1900T02@1900102H00@0V02H19P0V00409P0V0BH09P490003
0@0005@0003o00`00`4;00<101@09P0V02H09P0100409P0V02H00@H000`19P4V00409P0V00440003
0@00008000<100000P4:00030@0005<0003o00X00P4200030@0000L00P407P0102@0900T02@0904T
0040900T02@1900T02@0908101@00@0V02H09P4V02H00@0V02H19P8100<00@0020000`40001C0000
o`0700<110000`400005008100<V02H00P402P0V02H09P0V004200080BH09P0V004700<100T09P0V
02H09P400P000`400003008100@00@012P000`40001B0000o`0500811`000`400003008100D0900T
00020@0^900T02@0900102@0904T02@00@0T02@0900T00409P0V02H09P0102H19P0V008100HV0040
9P4900030@000580003o00<00P4900040@000081018V02H09P0V0@019P0V02H09P4400060@0V02H1
20040@0:9P0V02H09P400@@000<100400P001040004900030@000580003o00400P4;00@10480900T
02@090010040900T02@00@0T02@00@0T0040900T02@00@0102H19P0V02H00@0102H00@0V0BH19P0V
004900030@000540003n008130020@0G00409P0V02H09P0V0@019P0V02H09P401@001@4V02H100H0
00@100010`002@4V02H09P400@0200070@0000400@0400030@0000L000<10000D@000?`00P4<0081
00<T02@00P40;00T02@0900T0040904T02@0900102@0900T0B@00@0T02@00@0T00409P0V0P400`0V
00020@0@9P4102H09P4V0BH09P0V0@T000<10000D0000?X00P4<008100D09P0V00020@0:9P0V02H0
9P0V0@8000P102H09P0V0@H000@102H110000`40000200030@0000@000<102H00`4600811P000`40
000700030@000500003h008130020@07900T02@090020@0_02@0900T02@00@0T0040900T0040900T
02@00@0102@0904T02@19P0V02H09P0010401bH09P0V02H00P401@0V02H000<12@000`40001?0000
m@030@`00P405@0V02H09P0V00400@0V02H09P0V0@0300070BH09P0V0@0700030BH100<000<10000
0P000`40000800030@0000H00P4500<12@000`40001?0000l`020@d00P40:B@1900T02@0900T0B@1
900T02@0900102@00@0T02@00@0T02@0900T00810180904T02@0904V02H09P0V02H20@0802H09P0V
02H20@0402H09P8100<09P402@000`40001>0000l@030@`00P400`0V00020@0D9P0V02H09P010040
9P0V02H09P4400060@0V02H120020@03004000D000<100001`030@L000<100000P000`400@020003
0@0000L000<10000CP000>l00P4200030@0000P00P401B@0900T008102L0900T02@0904T0040900T
02@00@0T02@1900T0040900T02@090000P404R@0900T02@19P0V02H09P01008101<V02H09P0V0BH0
9P0100409P0100T000<10000C@000>d00P4400030@0000H00P401`0V02H09P000P4052H09P0V02H0
0@0V0BH09P0V02H110001P409P0V0@T000<100001P000`40000500030@01008000<1000010001@40
0001008000D100000@0900030@0004d0003[00811P000`400004008100TT02@0900T02@00P409P0T
02@0900T0B@0904T02@0900102@0900102@00@0T02@0900T0`40:2@0900T02@19P0V02H00@0V0BH0
0@0V02H09P4V00409P4V02H09P4900030@0004`0003Y008120000`400002008102400@0V02H09P0V
00400@0V02H09P0102H00@0V02H09P401@001@4V02H100H00P400`010@0700030@0000@000@10001
0`000`40000200810P001040004500030@0000L000<10000C0000>L00P4=008100<T02@00P40C@0T
02@0900T0B@1900T02@0904T02@1900T02@00@0T02@0904T0B@0900T004090010040900T02@19P0V
02H19P0V0BH09P4V02H0008100dV0BH09P4V02H09P0100T000<10000B`000>@00`4=008100D09P0V
00020@0N9P0V02H09P0100409P0V02H00@0V02H19P0V02H11P000`400@0500030@00008000@10001
1@000`40000200030@00008000D100000@0400<110000`40000300812@000`40001;0000hP020@h0
0P401b@0900T02@00P40FP0T02@0900T0B@00@0T02@0904T02@00@0T02@00@0T02@0904T0B@09001
02@0900102@1900T02@19P0V0BH09P0V0BH09P0102H19P4V02H00@0V02H09P<12@000`40001:0000
YP080C800P4=00<100T09P0V02H09P000P407RH09P0V02H00@0V0BH09P0V00409P0V0BH09P0V0@L0
0P4300030@0000@000D100000@0400050@0000401@000`40000200060@000@0110000`4000030081
00<00@0020000`40001:0000Y`000`40000300030@0002h00P4=008103<00@0T02@0900T02@1904T
02@0900T0B@00@0T02@0904T02@0904T02@00@0T02@090000P409B@10@0T02@0900102@0904T0040
9P0102H09P0102H09P0V0BH00`402RH09P0102H09P020@8000<100001`000`4000190000Z0000`40
000a008130030@039P0V008102T09P0V02H09P0100409P0V02H00@0V00409P0V00409P0V00409P0V
0@08008120000`40000300040@000@8000<1000010000`40000400050@01004010000`4000020003
0@01008000<100001`000`4000190000Z@000`40000^008130020@06900T02@00P40ER@0900T02@0
904T0B@0900T02@1900T0B@0900T0B@0900T0B@0900102@0900T02@1900T02@0900T0B@0900T0040
0@0102H09P0V00409P0V02H00P403RH19P0V00409P0V0BH10`000`40000700030@0004P0002Z0003
0@0002/00`4:00<103<09P0V02H09P4V0BH09P0V02H00@0V0BH09P0V00409P0102H09P0102H09P0V
0BH09P401`020@T000<100001@020@03004000H000<100001@001040004300030@00008000<10040
10000`40000700030@0004P0002[00030@0002P00P4200030@0000H00P40?@0102@0900T02@00@01
02@0900T02@1900102@0900T0B@0904T02@0904T02@0900102@00@0T02@090000`40400T02@0900T
0B@0900T02@30@0?9P0V02H09P0102H09P0V008100h09P4V02H19P0V0BH00@D000<100001`000`40
00170000[0000`40000T00<110000`400004008100<V02H00P40<`0V02H09P0V0BH19P0V02H09P01
02H19P0V02H00@0V02H19P0V00409P0V02H19P0V0@0600040@010@P000<100001P000`4000070003
0@0000<000H10041004300050@0000400P000`40000300030@0000L000<10000A`000:d000<10000
8@020@L000<100000P020@0502@090000P40CB@0900T02@00@0T0B@0900T02@1900102@0900T0B@0
900102@0904T02@0900T0B@00@0T02@00@0T0B@1900T02@0904T02@0900T00<100hV02H09P0V0BH0
9P0V008100lV0BH19P0V0BH00@0V004010030@T000<10000AP000:L000<100000`000`40000O0081
30020@0l9P0V02H09P4V0BH09P0V02H19P0102H09P0V00409P0102H09P0102H09P4V02H00@0V02H0
9P0102H11@001P4000400@L000<100001@030@L000<100000`001`400@0100400P001@4000010080
00<100000`000`400@0900030@0004H0002W00030@0000<000<100007@020@d007X102@0900T02@0
0@0102@0900T0040900102@0900T0B@0904T02@00@0T02@0904T02@1900T02@090010040900T0B@0
904T0040900T02@1900T02@00@0100409P0V02H19P0V00409P4V0BH19P0102H19P0V0@@00P420003
0@0000L000<10000A@000:P000D100000@0N00813@040@0m02H09P0V02H19P0102H09P0V0BH09P4V
02H09P4V02H00@0V02H19P0V02H00@0V00409P0V02H09P4V0@0300030@00008000@100011P000`40
000400030@01008000<100000`001@40000100<000<100000P4200040@000@<000<100000P020@80
00<100001`000`4000150000Z@030Ad00P4=0081070T0041900T02@0900102@1900T02@00@0T02@1
900T0040900T0040900102@0900T0B@0904T02@0900T00400@0T0B@0900T0B@0904T02@0904T02@0
904T0040900102H09P4V02H19P0V0BH00P402RH00@0102H09P4300030@0100<000<100001`000`40
00140000a`020@d00P40@@0V02H19P4V02H09P0V0BH09P4V02H09P4V02H00@0V02H19P0V02H19P0V
0BH09P0V00409P0102H09P0V02H000810P000`40000300050@0000401@000`40000200050@000040
0`000`40000200040@000@<000@100000P401001004400030@00008000<100400`000`4000070003
0@0004@0003500813@020@13900T02@00@0T0B@0900T0040900102@0900102@0904T02@00@0T02@0
0@0T0040900T02@00@0T0B@0900T02@090020@0j0040900T02@1900T0B@0900T0B@0900102@00@0T
02@19P0V0BH19P0V00409P0V0BH19P4V02H00@<000<100401@000`40000700030@0004<0003200<1
3@020@1502H09P0V02H19P0102H09P0V0BH09P0102H09P4V02H09P4V02H19P0V02H00@0V0BH09P0V
02H19P0102H09P0V02H000<11`000`40000200030@00008000D100000@0400030@0000<000D10040
0@0500030@00008000@100411@001`40000100401@000`40000700030@0004<0003000<13@020@18
904102@0900T02@1900102@0900102@0904T02@0904T02@00@0T0040900T02@1900102@0900T0040
904T02@0900T02@00P408B@0900T02@1900T0040900T0B@0904T02@00@0T02@090020@0>00409P0V
00409P0V0BH20@0602H09P010P001040004500030@0000L000<10000@`000;h00P4200030@0000P0
0`400`0V00020@169P0V02H09P0102H00@0V02H00@0V00409P0V00409P0V00409P4V02H09P0102H1
9P0V02H09P0100409P0V02H09P0V0@P000<100000`001040004300030@0000<000<1000010030@H0
00<100000P030@H000@100010P000`40000200<12@000`4000120000_0020@@000<100001P020@1;
02@0900T0B@1900T02@0904T02@1900T02@1900T0040900T0B@0900T0B@00@0T02@0904T0040900T
02@0904T0B@0900T02@000@102lT02@0900T0B@0900T0B@00@0T0040900T0040900T02@00@0V02H0
9P4V02H09P030@069P0V02H10P001@40000100@000<19P402@000`4000120000^P020@H000<10000
0`030@1B9P0V02H09P0102H19P0V02H00@0V02H19P0V00409P0V0BH09P0102H09P0102H19P0V02H0
9P4V0BH09P0V02H00@0102H09P0V00409P400@H000<1000010000`400@0200030@0000@000<10000
10020@H000<100000`020@H000@100010P000`40000200810P000`40000700030@000440002h0081
2000104000020@16904102@0900T02@1900102@0900T0B@0900102@0904T02@0904T02@1900T02@0
0@0102@0900T00400@0T02@0900T008102dT02@0900102@00@0102@0900T0B@0900T00400@0102@0
900T0040900T02@00`402bH09P0V0BH09P0V008100L09P0V02H1008000@1000110020@039P0100T0
00<10000@@000;H00P4<008104P09P0100409P0V02H00@0V00409P0V00409P0V00409P0102H09P01
02H00@0V02H09P4V0BH09P0V02H19P4V02H09P0V02H20@0402H09P8100@09P0V0P400`0100040003
0@0000D00P400`01000500030@0000@000@100411@000`40000200@11@000`400@0300030@000080
0P4300030@0000L000<10000@0000;@00P4<008103DT02@0904T0B@0900T02@1900T0040900T0B@0
900T0040904T02@0900102@1900T02@090020@0802@0900T02@20@0902@0900T02@0008100XT02@1
900T02@00P403b@00@0T02@1900T02@090030@0>900T02@0904T02@090020@0<9P0102H09P4V02H0
10401BH09P01008000D100000@0300070@019P0V0@0900030@000400002a00<130020@0g02H09P0V
00409P4V02H09P0102H09P4V02H00@0V02H09P4V00409P0V02H19P0102H09P0V00020@089P0V02H0
9P020@099P0V02H09P0V008100X09P4V02H09P0V0P4300050@00004020020@L000<100000`000`40
0@0200050@00004010001@400@0100@000<1004010001`400001004010000`40000700030@0003l0
002_00813@020@13900T02@0900T0B@0904T02@0904T02@0904T02@1900T02@0904T0B@0900T02@1
904T02@0900T0040900T02@090020@0:02@0900T02@090<100PT02@0900T008100lT02@00@0T0B@0
900T02@00P407@0T02@0900T0B@0900102@19P0102H00@0V0040008100TV00409P01004010003040
00400@0V02H00@T000<10000?`000:d00P4=008102P00@4V02H09P0V00409P0V0BH09P0102H09P01
02H00@0V02H09P0V0P404@0V02H09P0100409P0V02H000@100HV02H09P020@089P0V02H09P030@0<
02H09P0V02H00@011@000`400@05008100<00@401`000`40000200040@000@<000@100010P000`40
00020@8000@100000P4400070@0000400@0400<12@000`40000n0000Z`020@d00P40:R@0904T0B@0
900T02@1900T0040900T0B@0900T0040904T02@0900T008100PT02@0900T008100DT02@090020@04
004090D100DT004090060@0J02@1900T02@0900T0B@1900T02@1904T02@20@0Q02@00@0102@0900T
0B@0904T02@19P0V0BH00@0102H0008100PV00409P400@@000h100000@0102H09P4V0@T000<10000
?P000:T00P4=008102d09P0V00409P4V02H09P0102H09P0102H00@0V02H09P0102H19P0V02H09P00
0`401P0V02H09P8100<09P000P400bH09P020@0602H09P0V1@4022H09P0V02H00P402BH09P0V02H0
0@0200030@0000<00P400`010@0400810P000`40000300050@0000400`000`40000200040@010@<0
00<100400P000`4000020@D000@100010P00104000020@8000<100001`000`40000m0000Y`020@03
004000T00`40;B@0900T02@1900T0B@0900T0B@0900T0040904T02@0900T0B@00@0T02@090030@03
900100@100@0904T0P401@0T02@0008101<T02@0900T0040900T02@0900T00810180900T02@0904T
0040900T02@20@0602@0900T0P405@0T0040900T0B@00@0T02@19P0V00020@0=9P4V02H19P4V02H1
9P020@D000@100010P0010409P020@039P0100T000<10000?@000:D00P4300030@0000H00P40;RH1
0@0V02H09P0102H09P4V02H00@0V02H09P0100409P0V02H09P4V0BH09P020@0:9P010BH09P0V00@1
00LV02H09P0V008100P09P0V02H09P8100P09P0V02H09P8100X00@0V02H09P010P000`4000040003
0@0000D00P4400070@0000400@0500030@0000<00P4400030@0100<000<100401P000`400@030006
0@000@010`000`40000700030@0003`0002S00811@000`40000300<102X090010040900T02@1900T
02@1900T0B@0900T02@1904T02@0900T02@20@0802@10@0T02@20@0802@0900T02@20@0702@0900T
00020@08900T02@090020@08900T02@090020@0K904T02@0900T0B@00@0T02@0904T02@0900T0081
0280900T0040904T0B@0900102H09P0V0BH09P0102H19P0V0`4500040@000@8000/19P0V0BH19P0V
0@0900030@0003`0002P00<11`00104000020@0R02H09P0V0BH00@0V02H09P4V02H09P4V00409P0V
02H09P8100T09P0V02H09P00104042H09P0V00400@0V02H09P020@0A9P0V02H09P4V0BH09P0V02H0
0`402BH09P0V02H19P020@0802H09P0V004200030@0000<000<100401P000`400@0500040@010@D0
00<100000`020@@000@1004110000`40000400030@0100<000@100000P4500030@0000L000<10000
>`0009h00P4:000W0@010B@0900T02@00@0T0040900T0040900T02@1904T02@0900T00h10180900T
02@0904T0B@0900T02@30@0@900T02@00@0102@0900T00<102d0900T02@00@010040900T02@19001
02@090010040900T0040904T02@090000`402`0T02@00@0V02H0008100HV02H00@020@059P0V0040
1P000`400@03000<0BH00@0102H09P012@000`40000k0000W0020@`010403@0V02H09P0V0BH09P00
0P402bH09P4V02H09P0V008100<09P4010403BH00@4V02H09P0V02H00P404@0V02H09P0102H19P0V
02H000<101009P0V02H19P4V02H09P0V0P40500102H09P0V0BH19P4V02H09P010`000`4000020003
0@0100D000@100011`000`40000400030@0000800P401@01000100800P4300811P000`400@030006
0@000@0110030@T000<10000>P0009X00P4<00<100<00@000P4062@0900T0040900T02@1900102@0
900T00@100LT02@0900T008100T0900T02@090000`40E00T02@0904T0040900T00400@0102@09001
02@1900T02@1904T0B@0900T00400@0T0B@0900T0B@0904T02@00@0102@0900102@1900T02@09081
01<0900T00409P0100400@0V0BH0008100<V02H00`45008110003@4V0BH00@0V02H19P402@000`40
000j0000V0020@`00P407@0V02H19P0V0BH09P0V0BH09P0V02H19P4V02H000<101H00@4V02H09P0V
00400@0V02H09P0V0`40>BH09P0V00409P4V02H09P4V0BH19P0V02H19P0102H09P0100409P4V02H0
9P4V0BH00@0V02H00@0300030@00008000<100401@001040004600<11@001040004200030@010080
00<100000P4200@110000`400@0400030@01008000@100000P4200030@0000L000<10000>@0009H0
0P4<008101TV02H09P0102@0904T02@00@0T02@0900T008107<00@4102@0904T0B@0900T02@1904T
02@0900T0B@1904T02@0904T02@1900T0040904T0B@0900102@1900T02@1904T0040900T0B@00@0T
0B@0900T0B@0904T02@1900T0B@0900102@1900T02@0008101@T0B@0900102H19P0100400@0V0081
00@V0@010P00104000020@@000X100409P0102H00P400bH00@0900030@0003T0002C00<130020@0K
02H09P0V02H19P0V02H19P0V0BH09P0V02H000<105P09P0V02H00@0102H09P0V00409P4V02H09P01
02H19P4V02H00@0V00409P0V0BH00@0102H09P4V02H19P0V00409P4V0BH09P0102H19P0V0BH09P01
0`001@40000100<000D100000@0300030@0000<000<100400P001`40000100400`001@400@0100<0
0P4010010@4200040@00008110000`400@0200070@0000400@0300030@0000L000<10000>0000980
0P4<0081014V0@419P0V02H09P4T02@090020@060040900T1040QP010B@0900T02@1900102@0900T
0B@0904T02@0904T0040904T02@1900T0040900102@1900102@00@0T0040900T0B@00@0T0B@0904T
0040900102@0904T02@1900T0B@0904T02@1900T0B@0900T0B@1900T0B@19P4V02H19P0V0P400`0V
00030@03004100<000<100000P4400090@0102H19P0V008100D09P0V0@0900030@0003P0002?00<1
00<00@002@020@0602H00@0V0P403@0V02H00@0V02H09P000P40HbH10@4V02H09P4V0BH09P0V0040
9P0102H09P0102H00@0V02H00@0V00409P4V00409P0V0BH09P4V00409P0102H19P0V0BH09P0102H1
9P0102H00@0V0BH09P4V02H00@0400040@000@<000D100000@0300030@0000<000<100400`001040
0@4400050@0000400`000`400@0200810`001040004500810`001P4000400@D000<100001`000`40
000g0000S@020@@000<100001P020@0G9P0V02H1900T00409P0V0B@0900T02@00`40PR@0900T02@0
0@0T0B@0900T0B@0900102@0904T02@00@0T02@1900T0B@0904T0B@0900102@00@0T0040904T0040
900102@0904T0040900102@1900102@0904T02@1900T00400@0T02@1900T0B@0904T02@00@0T0B@0
90030@0<02H09P4V02H19P0V0`42008110030@D00P403BH09P4V00400@0V02H00P4900030@0003L0
002;00811P000`40000300<100d09P0V02H00@0V02H0008100LV00409P0V00<106@V0@409P0V02H0
9P4V02H19P0V00409P0V00409P0102H09P4V02H00@0V00409P0100409P0V00409P4V02H19P0V0BH1
9P0V00409P0102H19P0V0BH00@0V0BH09P0102H11@000`400@0500030@0100@000<100000P001040
004500030@0000<000@100000`4020010@400@0110020@H000<100000P000`400@0200030@000080
00<100402@000`40000f0000R@020@P000@100000P404P010@409P0V02H1900T02@090<103PT0@41
900T00400@0T02@0900102@00@0T02@1900T02@00@0T0B@0900T0B@0904T02@00@0T008102<T02@0
904T0B@0900102@00@0102@0904T02@1900102@090020@0g02@0904T02@00@0102@090010040900T
00400@0T02@1900T0B@0904T02@0904V02H09P4V00020@0:9P010@0100400@D00P44008100d09P01
02H19P0102H0008100<V0@0020000`40000f0000D0030C@00P4<008100HV02H19P020@0=9P0V02H1
9P0V02H09P030@0h9P0V02H09P4V00409P0V02H19P0V00409P0102H09P0V0BH00@0V02H09P4V0040
9P0V0BH09P020@059P0V02H00P402@0V02H19P0V00020@0@9P0V00409P0102H00@0V008100`V02H0
0@0V02H19P4500030@0100D000<1004010001@40000100<000<100000P020@@000@100000P430004
0@0100811P000`400@020@03004000<000<100400P001`40000100400P000`40000700030@0003D0
001?00050@000040<@020@`00P405@0V02H00@0T02@19P0V0040900T00020@0m904102@0900T0040
900102@0900102@0900102@1900T02@0904T0B@0900T00400@0T02@00@0T02@190020@0302@00081
00TT02@0904T02@00P40?`0T02@00@0T0B@0904T02@00@0T02@1900T00400@0T02@00@0102@09001
0040900T00400@0T02@1900T00030@0;02H00@0V0BH19P000P401@010@0100D00P44008101<V02H0
0@0102H19P0V0BH19P0100T000<10000=@0004h000<100000`000`40000/00812`030@0=9P0V02H0
9P4V02H09P020@0602H19P0V0P404@0V00400@0V02H09P4V02H0008101DV02H19P0V02H00@0102H0
9P0V02H00P401`0V02H09P000P402RH09P0V00409P020@039P01008102T00@0V02H00@0V02H19P0V
02H19P0102H00@0V02H19P0V00409P0V00020@H000<1000010000`400@0500030@0100<000<10000
0P020@030040008000<100000P43008100D00@400@0500811@020@<00P4300060@000@0110000`40
000700030@0003@0001>00030@0000<000<10000:P020@/00P404RH10@4V02H09P0V0B@0900T00<1
02<00@4T02@0904T0040900T0040900T02@1900102@0900T00020@04900T00X100<T02@00P404@0T
02@0904T00400@0T02@0008100XT02@0904T02@00P403b@090010040900T0B@090020@0:02@0904T
02@090810200900T02@1900T02@0904T02@090010040900102@0908100d00@0V00400@0102H00081
00D00@400@0400<110001P409P4V008100hV02H19P0100409P0V0@T000<10000=00004h000<10000
0`000`40000W00<12`020@0602H00@0V0P403@0V02H00@0V02H09P000`402`0V02H09P0102H00081
01@V02H19P0V02H09P4V0BH09P0V00L100D09P0V00020@039P0V00H101409P0V02H19P4V00409P0V
00020@089P0V00409P020@069P4V02H00P4022H09P0102H00`402`0V00409P0V02H100<01`440003
0@0000@00P4400090@0000400@01008000H10001004200050@0100400P000`40000200<110001@40
004000810`001040004200030@0000<000<100001`000`40000c0000CP000`40000300030@0002D0
0P4<008100XV02H09P4V02H00P4022H09P4T02@00`4070010@40900T02@1900T02@1900102@0900T
02@20@0302@100<100`0904102@0900T02@20@0702@0900T00040@0?900T00400@0T0B@0900T00<1
00lT02@00@0T0B@1904T02@00P402@0T02@1900100020@0:900T0B@0900T00@100LT02@1900T00H1
00D0900T00020@0L900T0040900100409P4V0BH19P0102H19P400@8000@100000`44008100<09P00
0`403P0V00400@0V00409P0V0`4800030@0003<0001>00030@0000<000<100008`020@03004000T0
0P403P0V02H09P0102H09P0V0P400`0100030@0N02H09P4V00409P0V00409P0V02H19P4V02H09P0V
1@406BH09P0V00400@0V02H09P0100409P0V02H00P40800V0@409P4V0BH09P4V02H19P4V0BH09P4V
00400@0V0P405`010BH09P0V0BH19P4V0BH00@0V02H000811P000`40000500030@0100<100<00@00
10001040004200080@000@01004200030@01008000<100400P00104000030@@00P4400030@000080
00<100400P000`400002008100<00@001`000`40000c0000CP000`40000300030@0002400P430003
0@0000H01@403`0V02H09P0102H09P0V00030@0;02@0900T004090000P402b@0904T02@0900T0081
00<09040104092@00@4T02@0900T0B@V0@0T02@0904T9P40900T004V0@0T008101`T0B@0900102@0
0@0102@190010040904T02@00`401@0T02@0008100dT00400@0T0B@0900T008100H0900T02@20@0<
02@0900T0B@0900T0P40400T02@00@0T0B@00@0V0BH20@0802H00@0102H30@8000D1000100020@<0
0P401BH09P0V008100<09P000P403BH09P4V02H19P4V004020000`40000b0000CP000`4000030003
0@0001l00P4400030@0000D00P401@0V0BH0008100/V02H09P4V02H09P020@0J00410@0V02H09P4V
02H09P4V00409P0V02H50@0[9P0V02H09P4V0BH09P0V00409P0102H09P0102H19P0V02H19P4V02H0
9P020@0K02H09P4V02H19P0102H19P4V0BH00@0V02H0008100DV02H09P020@0>02H00@0100409P0V
0BH20@D00P4500811P020@@000<100400P001P40000100810`001P400@01008100<00@000P001040
0@43008110001@400040008110001P4000400@<000<100001P000`40000b0000CP000`4000030003
0@0001d00P4600030@0000<00P402BH09P0102H09P020@0702H00@0V00020@0C9P0V0B@00@0T02@0
0@0T02@090030@03900100<1034T0041900T02@090019R@1900T02@192HT9P40900T0B@V904T02@0
0BHT0B@0900T00810100900T0B@00@0T0B@0904T0P402@0T0040900T00070@0?02@0904T0B@00@0T
0B@0008100DT02@090020@0502@090000P401B@0900T008100l00@0T00400@0T0BH09P000P4012H0
9P020@049P01008100T00@000@0100400P002`4V0BH09P0102H000<100l09P0102H00@0102H09P40
20000`40000a0000C`001@40000101h00P4<008100d09P0V02H19P0V02H0008100<V0BH00P402@0V
02H00@0V00020@0;9P0V0BH09P0V02H01040=@0V02H09P4V0BH09P0V02H19P0V0BH09P0102H09P4V
02H00@0V02H19P0V0BH00@0V02H000@100dV00409P4V00409P0V008102009P0V0BH09P0V0BH09P0V
0BH09P0102H19P4V00409P8110000`400@0500811@020@800P400`010@0300030@0000800P440081
0P001P400@400@8000D100400@0200030@0100<00P4400030@00008000@100010P000`4000020003
0@0000H000<10000<@0005000`4M008130020@0A9P0V02H09P0102H09P0V02H00`405BH09P0V02H1
900T02@1900102@090030@0i904102@0900T004V904T02@090019R@V904T02@00BHT9R@1900T0B@V
92HT0B@00BHT9P40900T008101/V904T00400@0T0040900T0B@0900102@090000P4012@090020@0>
900T0B@00@01004090020@0C900T00400@0T02@00@0102@090020@0402@090<1018T02@19P0V0040
9P0V0BH19P020@0600400@010P000`400@02008100@09P0V0P400`0V00030@0@02H00@0102H19P0V
02H00@P000<10000<00006d00`4<008101800@4102H09P0V00409P0V02H50@0=9P0V02H09P4V02H0
9P030@0:9P410BH09P01008103/V02H09P4V02H19P0V02H19P0V02H19P0V0BH09P0V0BH00@0V02H0
0@0V0BH09P0102H00@0102H19P020@0902H09P4V02H0008100PV02H19P0V00<100@09P0V0P401`0V
00409P000`401`0V02H100400`000`400@0500030@0100@00P4500030@00008000@100000`430081
0`020@0800400@01004200050@0100400`020@@000@100010P020@<000<100000`020@P000<10000
<00006/00P4=008100HV02H19P020@0:9P0V02H19P0V008100DV004090020@0<02H09P0102@0900T
0`40AB@0900T02@1900T0B@090019R@V904T02@00BHT9R@V904T004V92HT9R@1904T9R@V92H1004V
92HT0B@0904T0B@V90030@09900T0040900T008100X0900T0B@0904T0P40300T00400@0T0B@090<1
020T02@00@0102@00@0T0B@0904T004090010040900T008100DT02@19P020@030040008100HV0@40
9P020@03004000810P002`4000400@0102H0008100<V02H00P400`0V00030@0802H00@0V02H30@03
9P4000L000<10000;`0006T00P4=008100X09P0V00409P0V0P401P0V00409P8100T09P0V00409P00
0P401bH09P4V02H00`40;bH10@4V02H09P0102H09P4V02H19P0V02H19P0V0BH09P0V02H19P4V02H0
9P0V008100L09P0V02H0008101/V02H09P4V0BH09P4V02H00@0V02H19P0V0BH00P402@0V00409P01
00020@0F9P0V0BH19P0102H09P4V02H09P400@<000@100010`001040004300030@0100@00P401P01
00400@800P43008110001@40004000810P000`4000020@8000H100010043008110000`400002000:
0@00004V00400@P000<10000;`0006L00P4<00<100hV02H09P0V0BH09P0V0081014V0@419P0V02H0
9P4T02@090020@030040008100LT02@00@0T008100X0900T0B@0900T0P4050019R@V92HT0B@00BHT
9R@V92HT0P4012HT9R@90@0392HT0081010V92HT9P400BHT9P4V92HT0P40500T02@1900100400@0T
0B@0904T0P40;P0T02@1904T0B@0900102@090010040900102@1900T0B@00@0T00400@0T02@20@03
02@1008100<T02@00P40100V02H60@800`43000?0@01004V0BH09P4V0BH0008100@V02H00`403P0V
00409P4V00409P0120000`40000^0000I@030@/00P404bH10@4V02H09P0V0BH09P0V02H00`4042H0
9P0V02H00@0V02H09P030@0;02H09P0V0BH09P000P402RH00@0V02H09P020@0A9P0V02H09P4V0BH0
9P0V02H01`4012H09P020@039P0V00H100D09P0V00020@089P0V00409P030@0Q02H00@0V0BH19P4V
00409P0100400@0V00400@0102H000<100L09P0V0@0100<000@100010`001040004300030@0100@0
0P4300030@0000800P4400030@0000800P4010010@4200030@0000810P001P4000400@<00P440003
0@01008000`100019P0V0BH09P4800030@0002h0001S00810P000`400007008100H09P0102H20@0;
02H09P0102H09P000P400bH190030@0;9P0V02H1900T02@00P4040010@40900T0040900T02@30@04
900T00T100@V92HT0P400bHT0@040@0D92H10B@V92HT9R@1904T9R@V92H60@0;90419R@V904T9R@0
0P408RH102@190010040904T00400@0T0B@1900T0B@1904T02@30@0Y900T0040904T0040904T02@1
900102@1900102@00@0102@00@0T02@00P400`0V00020@<00P4010010@4300030@00008101009P4V
00400@0V0BH19P0V0P40100V0BH30@0:9P0V0BH09P0V0@P000<10000;@0006400P4400030@0000D0
0P402RH09P0V0BH09P020@079P0V0BH09P020@0802H09P4V02H20@0702H09P4V00020@069P0V0BH0
0P402bH09P0102H09P0V00H100@09P0V0P400`0V00080@099P0V02H09P4V008100h09P0V00409P4V
02H09P8100@09P0V0`4022H09P0V0BH00P408bH00@0V0BH19P0102H19P4V0BH00@0V0BH00@0V0BH1
9P4V008100D09P0V0@0200040@000@<000@100010P001040004300060@000@010P020@<000<10040
0`020@<00P4200810`001040004300030@01008000H10001004200810`030@0<02H09P0102H09P01
20000`40000]0000G`020@H000<100000P030@0>02H09P0V00409P0V02H20@030040008100`V02H0
9P0102@090030@0>00410@0T02@00@0T02@20@0602@1900T0`404b@10BHT9R@V92H19P4V92HT9R@0
10408BHT9R@V92HT0B@V0BHT9R@192HT0B@V92H19P4V92HT9P030@1@9R@V0BH19R@192H100400@0T
0B@00@010040904T0040904T00400@0100400@0T0040904T004090010040900102@1900T0B@00@01
02@20@0902@00@0102H000810`00104000020@8002X100000BH00@0V0BH00@0100409P4V0BH09P01
02H09P0102H09P0V004800030@0002`0001L00<120000`4000020@0C00410@0V02H09P0102H09P0V
00030@0A02H09P0V02H19P0V02H09P000`403@0V02H09P4V02H09P000`402@010BH09P0100020@0c
9P0V02H19P0102H09P4V00400@0V02H09P0102H00@0V02H19P0102H09P4V00409P0V008100D00@0V
00020@069P0102H00`401`0V00409P000P403RH19P0V0BH19P0102H00`402@0V0BH19P0V0@020004
0@000@<000<100400`001040004300060@000@010P020@<000<100400P020@<00P4300810P000`40
00020@8000<100400P001P4000400@800P44008100/V02H09P4V02H09P020@03004000H000<10000
;00005X00P4<008100HV02H19P020@0;9P0V02H19P0V02H00P404`010@409P0V02H00@0T02@09000
0`403`010B@0900T0B@0900T00030@0M9R@V92HT0B@V904T9R@192HT9P4V92H19R@192H00P408R@V
92HT0B@V92H19R@V0BHT9P4V92H19R@192HT9P4V0BH20@0:90419R@V0BHT9P8100PT02@1900T0081
00dT02@00@0T0B@1900T00<101TT00400@0T00409001004090010040900T008100`0900T0B@19001
02@20@0802@00@0102H20@<00P4300810P000`4000020@0D9P0100409P4V0BH00@0100409P020@0:
9P4V02H19P0V008100@V0BH120000`40000[0000F0020@`00P402P0V02H00@0V02H20@0702H00@0V
00020@069P0V0BH00P404bH09P0V00409P0V00409P0102H00P401`0V00409P000`40;@010@409P0V
00409P0V0BH00@0V02H00@0102H09P4V02H19P0V00409P0V00020@0G9P4V02H00@0V0BH09P4V02H1
9P4V02H00`404BH09P4V02H00@0V00409P0V008101T09P0V0BH00@0102H09P4V02H19P0102H100<0
0P4400030@0100@00P4400030@01008000D100400@0200030@01008000<100400P000`400@030081
0P000`400@020@800P4300090@0100400@0100<00P402BH00@0V00409P020@0702H19P0V0@070003
0@0002/0001F008130020@0>9P0V02H09P4V02H09P020@039P4V008100X09P0V00409P0V0P404`0V
02H1900T0040900T0B@090000P400b@190020@0B9R@V904T9P4V92HT0B@V92HT0`400b@V90040@0:
9R@V92H19R@V908100XV904T9R@V92HT0P4082HT9R@V0BH19R@V904T9P4V904T9R@V0BHT9P4V92HT
0P404@0T0040900T0B@0900102@00081018T02@00@0T0040904T004090020@0;900T00400@0T02@0
0P40600T02@1904T00400@0102@19P4V00400@8000<100400P030@0500400@000P401bH19P4V02H0
0P40200V0@409P4V0P402@0102H00@0100020@099P0V0BH09P0100P000<10000:P0005@00P4<0081
01800@4102H09P0V0BH09P0V02H30@0>9P0V02H09P4V02H09P020@059P0102H00P40300V02H00@0V
02H09P<100XV02H09P0102H00P4022H00@0V02H01P400bH09P030@0C9P0V00409P0V00409P4V02H0
9P060@0302H0008100lV02H00@0V0BH00@0V02H00P401`0V00409P000P404BH09P4V02H00@0V02H1
9P0V008100<09P000P402RH09P4V02H19P4300811@000`400003008110020@<000<100400P001@40
0@01008000<100400P000`400@02008100P00@400@0100<10`000`400@02008100d00@019P0102H0
9P4V008100`09P0V00409P0V004700030@0002X0001A00<130020@069P0V0BH00P402RH09P0102H0
9P020@0C9P410BH09P0V02H19P0V02H09P020@0=0041900T02@0900100040@0?90410B@V92HT0B@V
92HT00<100<T0@400P400bHT0@020@0792HT9P4V90020@099R@192HT9R@V00<100<V0@400P401R@V
0@4T9PH100hT9R@V0BH19R@V0BHT9P<100LV904T9P4V008100HT004090030@0<02@00@0T00400@0T
0`407b@00@0T00400@0T02@1900T02@1900T02@1900T02@00P40500T00400@0T0B@19P4V0BH00@01
0P001P400@0100<10P000`400@020@0302H0008100HV00419P020@0H9P0V0BH09P0V0BH09P0V0040
9P0V02H120000`40000Y0000D0020@`00P402P0V02H00@0V02H20@0602H19P0V0P401P0V00409P81
00l09P0V00409P0V02H09P003@40200V02H00@0V0P402@0V00409P0V00040@0B9P0V02H00@0V0BH0
9P4V02H00P401BH19P0V00H100L09P0V02H0008100DV02H09P040@0302H0008100LV02H00@0V0081
01/00@0102H19P4V0BH19P0V0B@19P0V0BH09P000P400bH190020@0902H19P0V0BH100@000<10000
2`4400810`020@<00P4200030@01008000<100400P000`400@020@030041008000<100000P440081
0P020@0;00419P0V0BH09P000`403@0V02H00@0V02H09P000P4700030@0002T0001>008100<00@00
20030@0>9P0V02H09P4V02H09P030@0=00419P0V02H19P0V00020@099P0V0BH09P0V00@100/00@4T
02@0900T00030@0902@0904T9R@V008100<T0B@010400bHT0@020@0992HT9R@192HT008100PV904T
9R@V90D100@T9R@V0`403BHT9R@V904T0B@V92H00P400b@V0@040@0K9R@V904T9P4V0BH19P4V0BH1
9P40904T0040008100LT02@1900T00<100/T0041900T0B@090020@0402@090<100/T02@1900T02@1
90030@06904T02@00P400b@090020@0<02H19P4V00400@010P030@030041008000<100000P401RH0
9P0100<100X00@4V02H19P0V0`403BH00@0V00409P0V02H0104800030@0002P0001<00810`000`40
0005008100<V0BH00P403@0V02H00@0V02H09P000`403P0V02H09P0102H09P0V0P400`0100040@07
9P0V02H19P020@0D02H09P0V0BH00@0V00409P0V02H30@099P0V02H00@0V008100T09P0V0BH09P00
0`400`010@020@0;9P410BH09P0V0BH00`401bH09P0102H00P402@0V0BH19P0V00020@0<9P0V02H1
9P4V0BH00P403BH19P0100400@0V0B@00P40200V00409P0V0P402`0T0BH19P0102H000810`020@D0
00<100000`000`400002008110020@<00P4200810`020@8000@100400P400`01000200030@0000<1
0P030@0400409P8101D09P0102H19P0102H19P0V0BH09P000P401BH00@0100L000<10000:00004`0
0`4200030@0000<00P401`0V02H19P000P4022H09P4V02H00`400`0100020@0>9P0V02H00@0V02H0
9P030@0N02H09P0V02H1900T0040900T0040900102@192HT0`403R@10@4T9R@V904T9R@V0P4022@V
0BHT9R@V0`404RHT9R@V0BHT0B@V92H19P4V9081014V92H19R@V0BH19R@192HT9P030@0;9R@V0BH1
9R@192H00P408b@V904T0B@1904T00400@0T0B@0900102@00@0102@1900T008100l00@4T0B@0900T
0B@090000P401B@0900T00<100LT0041900T008100L09P4102H000810P0010400@020@0300400080
00H1000102H30@0<9P010BH00@0100400P403BH00@0V00409P4V02H00P401`0V00409P4020000`40
000W0000@@030@P000<100400P00104000020@0;9P0V02H00@0V02H00P400`0100030@0802H09P4V
02H30@089P0V0BH09P030@0K00419P0V02H09P0102H09P0102H00@0V02H000<100X00@4V02H00@0V
0P402P0V02H19P0V02H30@039P41008101L00@4102H09P4V02H19P0102H00@0V00020@069P4V02H0
0P401bH09P4V02H00P401@010BH0008100dV00409P0102H00@0V008100D00@0T00020@0E9P0102H0
9P4T02@19P4V00409P0V00810`000`400002008110020@D000<100000`4400050@0000400`020@<0
00@10040104200030@00008110000`4000020@0600419P4V0P403@0V00409P0V0BH19P000P402RH0
9P0102H09P4700030@0002L0001000050@0000401`001P4000400@800`403@0V02H09P4V02H09P00
0`403P0V02H09P0102H09P0V0P400`0100030@0602H09P4V0P403@0V02H09P4T02@090000P401b@1
900T02@010402bHT9R@V904T9R@V008100PT9P4V92HT9P@1014T9R@V904T9P4V92HT0B@V90030@09
92HT9P4V92HT00<100<T9P400P4032HT9P4V92H19P4V90@100HV92H19R@20@069R@V0BHT0`402b@0
904T0B@1900T00<1014T0040904T0B@0900102@090020@0302@0008100XT02@00@0102@00P4012@0
900:0@039P0V00810`020@03004000<10P002040004V02H00P4012H10@020@0?9P0V0BH09P0102H0
9P0V00@100`09P0V02H19P0V004800030@0002H0000o00030@0000<000<100001000104000020@<0
00<100401040100V0BH40@0C00410@0V02H09P4V02H09P0V00030@0J02H09P0V00409P0102H09P01
02H09P0V02H20@030041008100@V00400P402bH09P0V0BH09P0V00<100<V0@400P400`0V0@020@08
9P0V00409P020@079P0102H09P030@0:9P0V00409P0V00<100<09P000P4032H00@0V0BH00@0V0081
00HV02H19P030@0502H19P000P4032@0900100400@0V00<100/0904V00400@0V00020@@00P430003
0@0100<000<100400`020@@000<100000P00104000020@800P43008100<00@000`4200040@010@@0
00<100000P401BH10@0V008100X09P0102H09P0V0`403bH09P0V02H19P0V02H00@0700030@0002H0
000o00030@0000<000<1000010000`40000200030@00008000H102H09P040@089P0V02H19P020@0<
9P0V02H19P0V02H01@406@0V02H00@0V02H00@0V02H1900T02@090000`402RHT9R@V904T9R@20@09
9R@V0BHT9R@V00@101<T9R@V92H19R@192HT9P4V92HT00<100LT0@41904T008100HV904T9R@20@0E
9P4192HT0B@192H19P4V904T9R@V008100DT9R@190030@0>92H19R@V0@0T02@190020@0490010081
00/T0040904T0B@090020@0402@0908100l0900100409001004090000P400b@090020@0302@00081
00DT02@100030@@000D1000100030@8000L100419P0V008100<00@000P400bH19P030@079P0102H0
9P070@0<9P0V02H19P0V02H00P4800030@0002D0000o00030@0000<000<1000010000`4000030003
0@00008001d102H09P0100409P0V00409P0V0BH09P0102H09P020@0502H19P000P402RH09P4V02H0
9P020@079P4V02H09P040@0?00410@0V02H00@0V02H0008100<V0BH010400`0V0@020@099P0V02H1
9P0V008100P09P4V02H09P<100PV02H09P4V00@100dV0BH09P4V0BH09P4V008100X00@4V02H19P0V
0`401RH00@01008100DV00409P020@0D02@00@0T0BH19P4T0B@1904V02H20@0402H19P810`020@<0
00<100400`000`400@0300810`020@<00P4200810`00104000020@801@400`0100030@<00P401001
0@020@0:9P4102H19P4V008100lV00419P0V0BH09P4V02H01@400bH1000600030@0002D0000o0003
0@0000<000<1000010000`40000300030@00008001@19P0V00409P0102H00@0V02H09P8100D09P4V
00020@099P0V02H19P0V008100X00@0V02H09P0V0P400`010@020@0E900T02@192H19R@V904T9R@V
92HT00<100TT9R@V92H19R@00P402BHT9P4V92HT9P030@039P41008100<T0B@00P402BHT9P4V92HT
9P030@069R@V904T0P401BH19R@V00<100TV92HT0B@V0BH00`4052H10B@1904T0B@1904T02@00@0T
0`400b@090020@0502@190000P4012@190020@03900T008101@090010040904T0040904T0B@09081
00<090000P4012@00@4300810P020@03004000810P020@0600410BH0104012H10@030@0302H000D1
0RH20@0600409P0V0P400`010@020@079P0V0BH00@0800030@0002@0000o00030@0000<000<10000
10000`40000400030@00008000T19P0V0BH09P000P402RH19P0V02H09P050@0=02H09P0V00409P0V
000=0@0=02H09P0V02H00@0V00020@079P0102H09P030@0>9P410BH09P0V0BH09P020@059P4V02H0
1P402@0V02H09P4V00020@069P0102H00P4022H10@4V02H10RH00`400@039P8100`V02H00@0V0B@1
9P030@0400409P8100<00@000P401R@0904T00<100<090000P401B@00@0V00<100<V02H00P430081
0`000`400@0200040@000@8000<100400P000`400@0200030@0100800P42008100@00@410P001040
0@4200D10`020@0500400@000P401BH19P0V008100l09P0V0BH00@0V02H09P000P402bH09P0V0040
9P0100L000<10000900003l000<100000`000`40000400030@0000D000<100000P002P4V0BH09P0V
02H20@0502H09P001`403@0V02H09P0V0BH09P000P402BH1900T02@090020@0@9P410@0T02@0904T
9R@V90<101LT0@4192HT9P4V904T9R@V0BHT9R@V90050@0692HT9P4V0P402B@V92H19R@V90030@09
90419R@V0BHT008100DV904T9P020@0590419R@00P402bHT9P4V0BHT0B@V008100@T9P4V0P400b@1
90020@0<02@00@0T0040900T0P401P0T0B@0908100<090000P400b@090020@0N02@00@0102@19001
02@1904T00400@0T0B@190010`020@<000<100000P40100100050@0302H100<100<00@400P4012H1
9P030@@V00X19P0V0BH09P0V1@402RH09P0102H09P4800030@0002<0000o00030@0000<000<10000
10000`40000600030@00008040401@0V02H000H101XV02H00@0V00409P0102H09P0V0BH00@0V0081
00XV02H00@0V02H010402bH09P0V02H19P0V008100H09P4V02H30@0=9P410BH09P0V0BH09P020@09
02H19P0V02H000<100T09P0V00409P000`429P<112H40@0@9P010B@V0@0V0040904V008102@T0040
0@0T0B@0904T02@00@0100409P0102H09P4V0@01004200040@000@8000@100010P000`400@020003
0@01008000@100400P4300810`000`4000020@040040008100<00@400P001@400@4V00D100<V0@40
0P403@0V02H19P0V00409P000P401BH19P0V008100T00@0V02H09P401P020BD0000o00030@0000<0
00<1000010020@P000<100000P002@4V02H09P0V00020@0?9P0V02H09P0100409P0V00@101P00@41
02H09P0102@0904T02@192HT9R@30@03904T00@100@V904T0P402bHT9R@V0BHT9R@V00<100XV0@4T
9R@V0BHT0P402BHT9P4V92HT9P020@03904T00@100<V90400`4022HT9P4V92HT0`400b@V90020@03
9R@V008100<T0@400P401RHT9P4V9081014V904T9P4V92H19P400@0T00020@0:900T0B@1904T0081
02LT02@00@010040904T0B@00@0T0B@00@0102@1900100400@0T0B@00P4200030@0100800`401@01
000100800`400`0100020@0>9P410BH19P4V0BH19P439P0B0BH09P0102H00@0V00409P0V1P40100V
0044008100<00@009000040000D100000@0700030@0000P000<100000P007@4V02H09P0102H19P0V
02H09P4V00409P0V02H000@101LV02H09P0102H19P0V02H19P0V02H09P030@0:9P0V02H00@0V0081
00TV02H19P0V02H010402`0V02H09P4V02H0008100PV00409P0V00@100XV02H09P0102H010401RH0
0@4V00810RH00`4V0@029P030B@1008V0P401R@V904V00<100l00@0V0B@0904T0B@00@000`402P0T
00400@0102H30@089P0V0BH1004200030@01008000<1004010020@8000<100400P000`400@020004
0@010@800P4300D100<00@400P0010400@4200070@010@0100040@069P4V00400P401bH09P0V0BH0
0P401`0V00409P000P401RH19P0V00P10`000`40000S0000@@030@P000<100002@000`400002000P
0BH09P0V0BH09P4V02H09P4V02H00@0V02H09P4V0BH50@0A9P4T0B@0900T004V92HT9R@00`403B@1
0BHT9R@192HT9R@00P400bH19P040@0492H19P8100/T9R@V904T9R@V90030@039041008100@V904T
0P402bHT9R@V0BHT9R@V00<100/V92HT0B@192H19P020@06904T0B@V0`4012HT0B@20@059P4192H0
0P402b@V0BHT0B@1904T008100D0904T00020@03904T008101H0900100400@0T0B@190010040900T
0P403P0T0B@1904T0040904T0P4200810`0010400042008100@00@400P402bH10@0100400@01008V
00@19P0V0P401P0V02H09P<100LV02H09P4V00810bH00`4V9P039P@11P000`40000S0000C0000`40
000:00030@00008002@19P0V0BH09P0V0BH09P0102H09P0102H09P4V00409P0V02H50@0:9P0V02H0
0@0V008100HV00409P020@099P0V0BH09P0V00@100X09P0V02H19P0V0P402@0V00409P0V00040@0:
9P0V02H00@0V008100/V02H19P0V02H19P030@059P0V0BH00P402BH19RH19P4T0@029P080BH19P40
0BH20@03904V008100H0904T02@20@030040008100DT004090020@030040008100HV02H19P020@80
00<100400P000`400@0400810P001`400@0100400`020@0600400@010P020@8000<100000P401`01
0@400@001`4020010BH19P0V0P400`0V00080@039P0V00<100PV02H00@0V00@12@000`40000S0000
C0000`40000;00030@00008002P19P0102H09P0V0BH00@0V02H09P4V02H19P0V00409P0V02H19P4T
1@400b@190020@0:9R@V904T9R@V908100<V0BH0104012@V0BH20@0:92HT9P4V92HT9P8100<T0B@0
10400bHT0@020@0;92HT9R@192HT9R@00P401BHT0B@V008100DT9R@190020@059R@192H00P400b@V
90020@049R@1908100<V0BH00P400b@190030@0890419R@192H20@03904T008100D0904T00030@0:
02@10@0T0040908100P0904T0B@0908100P0900T0B@0908100P0904T0B@0908100H00@0100420081
0P000`4000060@039P4100H100P00@4V0BH09PH10RH20@@V1`439P030BH100<12`000`40000S0000
C0000`40000<00030@0000800P40:P0V02H09P0V0BH19P0V02H09P4V0BH09P0V00409P0V0BH00@0V
02H09P@100d09P0V02H19P0V02H000<100X09P0V02H19P0V0P402@0V00409P0V00040@099P0V02H0
0@0V008100/09P0V0BH09P0V00030@0900409P0V02H1008V0`429PT100HT0@4V0BH30@049R@19081
00T00@4T004090000P402R@00@01004090020@0:900100409P4V00810P000`400@03008110001040
00020@8000<100400`020@0600400@010P020@030041008010400`0100090@0400419PD100@V02H0
0P401RH19P0V00<100@09P0V1P4>00030@0002<0001<00030@0000d000<100000P0G0@0D9P0V02H0
0@0V0BH09P019R@V92H30@0;9P4192HT9P4V92H00`403bH10@4V92HT9P4V92HT9P020@03904T00@1
00<V90400P402R@V92HT0B@V92H20@0:92H19R@V92HT9PD100XV92HT9R@192HT0`400b@V90020@03
9R@V00D100dV0BHT0B@V0BHT0B@V008100DT0B@V90020@0F02@1904T0B@0904T02@1904T00409081
00P0904T0B@0908100P090410040908100P0904102@0908100<00@400P001P400@0100@100<00@40
0P400`010@070@039P41008100@T0BH10bH20@8V00<19RH00`401bHT0B@V92H0104@00030@0002<0
001<00030@0000h000<100000P005P4V02H09P0V0BH19P0V02H09P4V0BH50@0E9P4V0BH09P0V0040
9P0102H00@0V00<100<V0BH00P401`0V02H19P000P402RH09P4V02H09P040@099P0V02H00@0V0081
00X09P0102H09P0V0`400bH00@040@0302H100P100<V0BH00P4012H10@429P<10RH0204V0@4T0B@V
0P4012@V0BH20@0?904102@1900T0B@00@0T00810140900102@00@0102H19P0V0@0300030@000080
00@1000110030@<00P4400030@0000810P0010400@040@03004000@100<00@400`401BH10@0V0081
00L09P4V0040008100@V02H00`401`0V02H19P00104C00030@0002<0001<00030@0000h000<10000
0`006P4V02H09P0102H19P0V02H00@0102H09P0V1`401BH09P4T008100XV92HT0B@V92HT0`402b@V
92HT9P4V92HT008100@V0BHT10400bHT0@020@0:92HT9R@192HT9P8100HT0B@V92H70@0:9R@V92HT
9P4V908100@V92HT1P400bHT0@020@03904T00@100HV904T0B@30@03904T008101<V0BH19P4V92H1
02@00@01004000<101@0904T0B@00@0T0040900102@090<11`00104T02@20@0302@0008100@V0@01
0P000`400@040@0500410@000`400`0100020@048`419P8100@T9P410RH30@039P4100H100@T9R@V
104F00030@0002<0001<00030@0000l000<100000`00804V02H00@0V02H19P0V00409P0102H09P0V
0BH19P0V1P402@0V02H00@0V00020@039P4V008100/09P0V00409P0V00040@099P0V02H00@0V0081
00T09P0102H09P001`40100V02H30@0<9P0V02H09P4V02H00P439P030BH000X100<T9P401@4012@1
0BH20@039041008100P00@4T0B@0908100P0904T0B@190<100/T00409P4V0BH09P020@8000<10000
0`4:008100<00@000`400`0100020@<0104010010@030@04004100/100D00@0V00050@0800419P4V
02H60AP000<100008`0004`00P4A00030@00008002<102H09P4V02H00@0V02H19P0V00409P0V0BH0
0@0V02H09P020@039P41008100<T9P400P401bHT9P4V92H00P401b@V904T9R@00P400bH19P020@0:
92HT9R@192HT9P8100<T0B@0104012HT0B@20@089R@V92H19R@20@0<9R@V92H19R@V92HT0P401RH1
9R@V908100@V904T1`400b@10@050@03904100<100<T0B@00P402RHT0B@1900102@30@0@904102@1
90010040900100D13`030@D00`400`010@020@030040008100H00@400@030@049P419PD100<V0BH0
1@402bHT0@4V904T9R@V00<16`000`40000S0000C0000`40000A00030@00008002T102H19P0V02H0
0@0V0BH09P0V00409P4V02H00@0V02H19P0102H09P040@0:02H09P4V02H09P8100<00@000P401bH0
9P4V02H00P402@0V00409P0V00040@0:9P0V02H00@0V008100PV02H19P0V008100XV02H19P0V02H0
4`400bHT0@060@03904T008100/V0@4100400@0100020@04904T00<100h00@4T00409P4V0BH00Ah0
10401@010040008100<00@000P400`010@070@0400419PH100<00@000P401bH09P0102H00`4M0003
0@0002<0001<00030@00018000<100000P00:@400@0V02H09P0100409P0V02H19P4V02H09P0102H1
9P0V0B@V92HT008100`V0@4T9R@192HT9R@40@0;9R@V92H19R@V92H00P400b@190040@039R@10081
00/T9R@V904T9R@V90020@089P4V92HT9R@20@049P4V90@100@V904T0P401bHT9R@V0BH03@403BH1
9P4V904T0B@192H00P400b@190020@090040900102@0008100XT00400@0102@170040@<00P4200<1
0P000`4000030@039P4100@100DV0@419P020@0392HT008100LV92HT9P4V008180000`40000S0000
C0000`40000C00030@00008060403RH09P4V0BH09P0V0BH00P401BH00@0V008100<00@0010400bH0
0@020@0<02H09P0102H09P0V0`402BH09P0V00409P020@0:02H00@0V02H09Pl100XV02H09P0102H0
0`400bH19P039P<10RH70@059P4V0BH00`4012H10B@20@030040008100TT0B@0904T02@00P401@0T
0BH0008100<V0@006P020@<000@100000P4200030@01008000H100400@0<0@03004100/18P000`40
000S0000C0000`40000D00030@000080018102H09P0V00400@0V02H09P020@039P0V00P100l09P01
9P4V92HT0B@V92H00`402BHT9R@V904T9P020@0792HT0B@V90030@0>90410B@V92HT0B@V92H20@04
904T9P@100<T0B@00P4022HT9R@V904T1`401R@V0BHT9QL100XV0@419P4V0BHT0P403BHT0B@1904T
004090000P401B@00@0T008100<00@005`030@H000<100000P401P010@01008100D00@0100070@03
9P4100H100<T9R@0104U00030@0002<0001<00030@0001D000<100000P006P409P0V02H19P0102H0
9P0102H19P0V02H00P400bH00@070@059P0V0BH00P403P010@409P0V00409P0V0P400`0100020@07
9P0V00409P020@0902H00@0V02H000D101P09P0V00409P0102H09P0102H00@0V02H60@0502H00@00
10439P030BH100h100<T0@400`402B@00@01004090020@0902@1904T0BH000<15`020@X00`420081
00<00@000P401@010@4000T100HV00419P050BP000<100008`0004`000<100005P000`400002000U
0@0V02H19P0V00409P0V0BH09P4V02H00@0V0B@V92H19P4V90050@0692HT0B@V0P402B@V904T9R@V
90040@0;9R@V92HT0B@V92H00P400b@190040@039R@1008101@T9R@V0BHT9R@V0BHT9P4V92HT9P81
00DT0@4V90020@0:9R@V0BHT9R@V91<100DT0@4190020@059R@V0BH00P403B@10@4T00400@0102@0
104C00<13P020@04004000@100@00@4V20400bH10@080BX000<100008`0004`000<100005`000`40
0002000a0@0V0BH09P0V00409P4V02H00@0V02H19P0102H00@0V0BH09P0100409P0102H09P020@03
004000@100@V00400P402RH09P0102H09P040@099P0V02H00@0V008100T09P4V02H09P000P401BH1
9P0V00T100LV00419P4V00D100LV00409P0V00<10`0:0@0802@1900T0B@20@0>0041904T0B@1904V
02H20@038P4000l00P4C008100<00@001P400`010@040@039P4100L1;@000`40000S0000C0000`40
000H00030@00008001d100409P0V02H00@0102H09P0102H19P0V004V0@049P070B@V904T9P020@07
904T9R@V90040@0:9R@V92HT0B@V90<100<T0B@0104012HT0B@20@099R@V904T9R@V010100TT9R@V
0BHT9R@02@401b@V92HT0B@00`47008100<V0BH0104022@V0@4T9P4V0P400b@190020@0400409081
00D090018P030@`00`4F00060@01004010400bH10@060@039R@100<1;`000`40000S0000C0020AX0
00<100000P0?0@0802H00@0102H50@0=9P0V02H19P4V02H09P020@030041008100<V00400P40300V
02H00@0V02H09P<100XV02H09P0102H00`400`0100030@0800409P0V02H20@03004000<100T09P0V
0BH09P000`401P0102H09PL12`070@03904T008100T0904T00400@000P4012@190020@039P0100H1
2@020AT000<100400P400`010@080@039P4100<1<P000`40000S0000C0000`40000I00030@0000<0
00h102H09P0V00400@0V00P10bH00`4V0@030@069P4V0BHT1P402BHT9R@V904T9P020@0792HT0B@V
90030@03904T008100XV92HT0B@V92HT10402BHT9R@V0BHT9P020@0892HT0B@V92H40@03904T0081
00/V92HT0B@V92HT9P020@l00`401R@10BH19P<100@V0BHT0P40100T0B@20@0702@10@018P080@@0
0`4K00030@01008100<00@402`4e00030@0002<0001<00030@0001X000<100000`004@409P0V02H1
9P0102H09P0V00<100XV02H00@0V02H01P4012H09P030@0:02H09P0102H09P8100<00@000P4022H0
9P0102H00`400`0100040@049P0100@100<V0BH00P403P0V02H00@0V02H09P0V10402P0V02H09P4V
02H20A801`403R@10@4T0040904T00400P401B@00@4V00@100DR0B818P050Ah000<100400P400`01
00090@03004003@000<100008`0004`000<100006`000`400003000D0@0V02H19P0V00409P0V0BH0
9P439P030BHV00<V0P402bH19R@V92H19R@V008100TT9P4V92HT9R@010403BHT9R@V904T9R@V92H0
0`4032HT9R@V904T9R@V90D100/T9R@V92H19R@V90030@04904T9P@100@T9P4V0P4E00D100@T0@4V
0P403R@192H102@1904T0B@00`400b810@0;0A`01@4200030@0100D11@000`40000b00030@0002<0
001<00030@0001`000<100000P00504V02H00@0V02H19P0V00409P0V0P401@0V0BH0008100lV02H0
0@0V02H19P0V02H00`400bH00@030@049P01008100PV02H00@0V00<100<00@000P4022H09P0102H0
0`40100102H40@0402H19P<100/V02H00@0V02H09P040AP000<190400`403P010@400@0T0B@00@0T
0P401`0T0@4R0B801`400b818P040AH02@400`010@050@L000<10000<P000`40000S0000C0000`40
000M00030@00008001D19P0102H09P0V0BH00@0V02H09P001@469P0<0B@V904T9R@V92HT10402RHT
9R@V904T9R@30@03904T008100PV92HT0B@V90<100<T0B@00P402bHT9R@192HT9R@V00D100/V92HT
9R@V0BHT9P030A/01@401BH10@4V008100PT0@41904T008100<T0@400P400b810@030@038P4100d1
2@0>0@03004000D12P000`40000a00030@0002<0001<00030@0001h000<100000P004@4V0BH09P0V
02H19P4V02H000P100hV02H09P0V0BH00@0V00@100<V0BH00P40300V02H00@0V02H09P<100dV02H0
9P0102H09P0V00<100`V02H09P0102H09P020@049P4V00D100@09P4V0P4O00030BH100D100`T0@40
0@01004090030@038P4100<100DR0@418P050@078P4R0B818P0I0@03004100@130000`40000`0003
0@0002<0001<00030@0001l000<100000P0@0@0802H09P4V0BH60@8V0P403BH10@4V92HT9P4V92H0
0P401b@V904T9R@00`400b@190020@089R@V904T9R@30@03904T00<100LT9R@V0BHT008100/V92H1
9R@V92HT9P040B40104042H10B@1904T0B@1900102@80@038P4R00<100<R0@408`400`010@030@`0
00<10000<0000`40000S0000C0000`40000P00030@00008001d19P0V02H09P4V0BH09P0V02H00@01
02H09P0V00050@0>02H09P0V02H19P0V02H20@030040008100PV02H00@0V00<100<00@000P402BH0
9P0102H09P030@0@9P410BH09P0V0BH09P0V00819@060@039P4000<100L00@4T0B@000D100TR0@41
8P4R0B800`401B818P4R01l100<00@400P4=00030@0002l000<100008`0004`000<100008@000`40
0002000F0BH09P0V00409P4V02H09P0102H00@PV0P404BH10B@V92HT9P4V92HT9R@V00<100dV92HT
9R@192HT9R@V00<100hV92HT9R@192HT9R@V90@100XV92HT9R@192HT0P4X00<100TT0B@192H19R@0
0P400`0100050@058P410B801@400b810@0X0@h000<10000;`000`40000S0000;0030@P00P4600P1
1@030B8000<100000P008P4V02H00@0V02H19P0V00409P0V00409P0V02H19P0102H30@089P0V0BH0
9P030@030040008100PV02H00@0V00<100<00@000`40200V02H19P0V0`4012H19P040@049P010081
:P050@079R@1904T00050@03004100<100LR0B818P4R00D100TR0B818P4R0B800`400b818P0D0@03
004100813`000`40000^00030@0002<0000[00050@0000401`020@L000<100000`000`4000030003
0@0002<000<100000P005P4V00409P0V02H19P0102H09P0V00449P050BHV9P401BH20@039P4V00<1
00DV92HT0@039P<100<T0B@00P401BHT9R@100@V0`400bH19P020@0592HT9P401RH40Bd01@401RH1
9P4V908100/00@018P4R0B818P030@038P4R02L100@00@414@000`40000]00030@0002<0000Z0003
0@0000<000<100003P000`40000800030@0002@000<100000P009@4V0BH09P0V02H19P4V02H09P0V
0BH09P4V02H09P4V02H09P000`403P0V02H09P0102H09P0V0`403RH09P0V00409P0V02H00`402`0V
02H09P4V02H000<1<0050@03904T00<100DT0@4190060@038P4R00<100<R0@400`400b818P0=0@03
8P4R00l10P030A4000<10000;@000`40000S0000:P000`40000300030@0000l000<100001`000`40
000T00030@0000<040401`0V02H19P401RH00`4V9P029P8100<V0BH00`401BHT9R@100<V0`400bH1
9P020@059R@V904012H30@039P4V00@100@T9P4V0`4d00T100<T9P400P401@010@4R00D100LR0B81
8P4R00<100<R0B807P400`018`020A8000<10000;0000`40000S0000:P000`40000300030@000100
00<100001P000`40000U00030@0000<000l19P0V02H09P4V0BH09P002P401RH09P01008100PV02H1
9P0V00<100<00@000P4022H09P0102H00`400`0100030@0:02H09P4V02H09P@1=`050@069P410@0T
0`4022@10B818P4R0`401b818P4R0B800`402B810@4R0B818P030@058P410B800`400b810@090@04
00410A<000<10000;0000`40000S0000:P000`40000300030@00014000<100001@000`40000V0003
0@00008001T102H09P0V00409P4V02H09P0V0BH19P0V00H11BH00`4V9P059P<11RH00`4V9P049P<1
00PV92HT9R@V0@@V0P4j00X100<V0BH01@401B810@4R00L100<R0B808P4010010@4D00030@0002/0
00<100008`0002X000<100000`000`40000B00030@0000@000<100009`000`400002000N0@0V02H0
0@0V02H19P0V02H19P0V0BH09P0V00400P4022H09P4V02H00`400`0100020@089P0V00409P030@03
004000<100L09P0V0BH00081>@030@03004100L100<T0B@01P401B810@4R00D100lR0B818P4R0B81
8P4R0B801@401B818P4R00D100<R0@402@4D00030@0002/000<100008`0002X000<100000`000`40
000C00030@0000<000<10000:0000`400002000O0@0V00409P0V00409P0V00409P0V0BH09P0102H0
9P020@039P4V00<112H0104V9RH30@039P4V008112H00`4V9P029P<100@V0@41>@020@D00P401B@1
0@4V00<100<V0@401@401b818P4R0B80:`4F00030@0002X000<100008`0002X000<100000`000`40
000=00030@0000<000<100000`000`40000Y00030@000080020100409P0V02H00@0V00409P0V0040
9P0V0BH09P0V00<100h09P0V02H00@0V02H09P<100hV02H09P0102H09P0V0081>0030@L01P4032@1
0@01004R0B818P<100<R0@400`401B818P4R00<100DR0B818P030@038P4R00<100LR0B818P4R00`1
5`000`40000Y00030@0002<0000Z00030@0000<000<100003@000`40000300030@0000<000<10000
:P00804000019P4V02H09P0V00400@0V02H09P0102H19P0V10400`0100020@DV00<19RH00RH20@03
9P4V00<112H0104V9RH30CP00P4;00<100<V0BH00P400b@190020@058P4R0B801@402B810@4R0B81
8P030@038P4R00<100<R0@405`4G00030@0002T000<100008`0002/000D100000@0A00050@000040
1P020B`000@100004@401RH09P0100@100TV02H09P4V02H00P401P0V00409P8100P09P0102H09P<1
00<V0BH00P4h00<13@050@03904T008100<00@400`401b818P4R0B801@401B818P4R00<100DR0B81
8P030@038P4R00L100DR0@418P070AT000<10000:0000`40000S0000;0030A<00`4700030@0002`0
01@100000BH09P0V02H00@0102H09PL100hV02H09P0V00409P0V008100@V0@411BH00`4V9P059P<1
>0020A401@401B@10@4T00@100<R0@400`401B818P4R00<100DR0@418P030@058P4R0B80604I0003
0@0002P000<100008`0004`000<10000;@005`4000019P0V02H00@0V00409P0V02H00081014V0@40
9P0V02H19P0V02H09P030@0<9P0V02H00@0V02H00P4h00<14`050@079P410@018P050@098P410B81
8P4R00<100DR0B818P050@038P4R00<100<R0B800`401B810@4R00<100<R0@400`4J00030@0002L0
00<100008`0004`000<10000;P00504000019P0V02H19P0V0BH09P0V0P401@0V0BH000<100P09P01
02H09P<100<V0BH00P449P040BHV9P81=`030AL01P401R@10B818P<100DR0B818P050@038P4R00<1
00<R0B801`400b810@0C0A`000<100009P000`40000S0000C0000`40000_000M0@00004V02H19P0V
02H19P0V0BH09P0102H09P000P400bH19P030@089P0V00409P030@0300400081=`020AX01P400`01
0@030@058P410B800`401b810@4R0B800`400b810@050@058P4R0B800`400b818P030@038P4100<1
00<R0@400P4L00030@0002H000<100008`0004`000<10000<0007P4000019P4V02H09P0V0BH19P0V
02H00@0V02H09P<100hV02H09P0V0BH09P0V00<1=P030Ad01P401BH10@4R00D100DR0B818P0;0@07
8P410B818P0G0Ad000<100009@000`40000S0000C0000`40000`00030@0000803@402RH09P0V0BH0
9P020@039P4V00<100PV02H00@0V00<1=`020B001P4012@10@438P0=0B810B818P4R0B818P030@05
8P4R0B800`401b818P4R0B800`401B818P4R00<100<R0@400`4N00030@0002D000<100008`0004`0
00<10000<@000`40000200080BH09P0V02H20@0302H100<100@09P4V0P40200V00409P0V0`400bH1
9P020CL00`4S00<100<V0BH00`401b818P4R0B800`401b818P4R0B800`401B818P4R00<100<R0@40
504O00030@0002@000<100008`0004`000<10000<P000`400002000?0BH09P0V00400@0V02H000@1
00dV02H09P4V02H09P0V00<1=`020BH00`401B@1904R00810R801`4R8P4R0B801@400b810@030@05
8P410B801@400b810@030@038P4100/17`000`40000T00030@0002<0001;00030@0103@000<10000
0P003P4V02H00@0V0BH09P0V0`402bH10@0V02H19P0V00<1=`030BT00`4032@10B818P4R0B818PD1
00LR0@418P4R00<100DR0B818P050@038P4100l18@000`40000S00030@0002<0001;0081=P000`40
0002000C0BH00@0V02H19P0V0BH00@0V00030@0300400081>0020B/010401BH10B81008R0P448P<1
00TR0B818P4R0B801@400b810@030@058P4R0B801@400b818P060B8000<100008P000`40000S0000
B`000`40000f00030@00008000T19P4V02H00@000P402RH09P4V02H09P030CL00`4Z00<100D00@41
90040@038P4100<100<R0@400`400b818P030@038P4R00T100<R0B804@4R00030@00028000<10000
8`0004/000<10000=`000`40000200P100X09P0V00409P0V0P4h0081:P030@D01@400b810@028P81
00HR0@4R0B860@038P4100D100<R0B800`402B818P4R0B818P0:0B<000<100008@000`40000S0000
B`000`40000h00030@00008000H19P0V02H30@069P0V0BH00P4g00<1:P020@P01@400b810@050@03
8P4R00<100<R0B801@402B810@4R0B818P030@038P4100h190000`40000Q00030@0002<0001;0003
0@0003T000<100000P00204V02H19P0V0P4010010@4g0081:P030@/020428P810b820@DR00H18P4R
0B830@078P410B818P050@038P4100T19@000`40000P00030@0002<0001;00030@0003X000h10000
0@0V00409P0V0081=P030BX00P4>00P100dR0@418P4R0B818P4R00<100lR0B818P4R0B818P4R0B80
0`400b810@0:0BH000<100007`000`40000S0000B`000`40000k000:0@0000409P4V00<1=@030BX0
0`4@00D10R801P4R8P4R0@8R00<18P400b840@078P410B818P030@038P4100<100LR0@418P4R00H1
9`000`40000O00030@0002<0000Y00@100<00@406`000`40000l00040@00008100<V0@40=P020B/0
0P4C00H100<R0@401@400b818P030@038P4R00<100<R0B800`401B818P4R00<100<R0B802`4X0003
0@0001h000<100008`0002P000<100000P020A`000<10000?0000`4000020081=@030BX00`4F00<1
00`R0B818P4R0B818P428P030B8R008100<R0B800`401B810@4R00<100DW0B818P050@038P4R00@1
:0000`40000N00030@0002<0000X00030@0000<000<100006P030Cd000<10000=P020BX00`4I00<1
00<R0@401@404b818P4R0B818P4R0B818P4R0B801@401B810@4R00`1:P000`40000N00030@000280
000X00030@0000<000<100006P000`40000n00030@0003800`4Z008170030@068P418P4U0P428P81
0b820@038P4R00<100<R0B800`400bL19`050@058P4R0B801`4Z00030@0001h000<100008P0002P0
00<100000`000`40000J00030@0003l000<10000;`020BX00`4O00@100DR0@418P050@038P4R00<1
00<R0B802@401b818P4R0B80304[00030@0001d000<100008P0002P000<100000`000`40000J0003
0@00040000<10000:`030BX00P4R00@100LR0@4U8P4R010100<W0BL00`401bL10@4R0B800`400b81
0@040Bd000<1000070000`40000R0000:0000`40000300030@0001X000<10000@@000`40000X0081
:P030B@010401B810@4R00<100DR0B818P030@038P4R00<100/R0B818P4R0B818P0=0Bd000<10000
70000`40000R00009`020@@00P4L00030@00048000<1000090030BT00`4X00050B810B800P401B81
0@0R008100<08P000P4012808P030@039`4100<100DW0B818P050@039`4R00@1;P000`40000K0003
0@000280001;00030@0004<000<100008@020BX00P4[00D100LR0@418P4R00<1014R0@418P4R0B81
8P4R0B818P030@038P4R00<100<R0@40104_00030@0001/000<100008P0004/000<10000A0000`40
000M00<1:@030Bd00P4012D10B820@049B818P8100P08P0100408P@100/W0@419`4W0B818P030@05
9`410B800`4`00030@0001X000<100008P0004/000<10000A@000`40000J0081:P020C002@401b81
8P4R0B801@401B818P4R00<100<R0@401`400b810@050C4000<100006P000`40000R0000:`030@P0
0P4:00@11@030DH000<100005P030BT00`4c00@100HR9@418RD20@098P4R0B800@0R008100LW0BL1
9`4W00D100LW0@419`4R00@1<P000`40000I00030@000280000Z00050@0000401`020@`000<10000
10000`40001700030@0001<00P4Y00<1=P020@038P4R00L100/R0@418P4R0B818P050@038P4R00<1
00DR0@418P060C@000<1000060000`40000R0000:@000`40000300030@00018000<1000010000`40
001700030@0001000`4Y0081>@001P4R0@4R9@8100HR00418RD20@038P0R008100TW0BL19`4W0BL0
0`400bL19`080CD000<1000060000`40000R0000:@000`40000300030@0000d01`4500030@0004P0
00<1000030030BT00`4l00L100DR0B818P030@038P4R00<100DR0B818P030@098P4R0B818P4R00D1
=P000`40000G00030@000280000Y00030@0000<000<100003@000`40000200030@0000@000<10000
B@000`4000090081:P020Cl010401RD19@408P<100@R0B802@400`0100030@039`4100L1=`000`40
000G00030@000280000Y00030@0000<000<100003P001@40000100H000<10000BP000`40000500<1
:@030D8020400b810@050@0;8P410B818P4R0B801@400b818P050CT000<100005P000`40000R0000
:@000`40000300030@0000h000D100000@0600030@0004/000<100000P020BT00`55008100<R0B80
0P402`0R0@408P4R0280008100<R02800P401`01004W0BL01`4k00030@0001D000<100008P0002T0
00<100000`000`40000?00040@000@H000<10000C0040BT00P5900<100DR0@418P030@058P410B80
0`400b810@030@038P4100D100<R0@40?@000`40000E00030@000280000Y00030@0000<000<10000
3`001040004600030@0004/00`4X00<1B`020@049@418P8100@08P4R0P400`0R00020@0>8P010040
0@019`4W0BL20Cl000<1000050000`40000R0000:@000`40000300030@00010000<100401P000`40
001800<10`000`40000S0081C`040@078P410B818P030@058P410B801@400b818P060D0000<10000
50000`40000R0000:P001@40000101<000<100401P030DH00P4700030@0001l00`5B00<1018R0@40
8P4R0B818P0100400@050@039`410081@P000`40000C00030@000280000[00<15@020@H000<10000
@`030@X000<1000070020EH00P403B810@4R0B818P4R0B801@401b818P4R0B800P5300030@0001<0
00<100008P0004/000<10000@@020@h000<1000060030ET010400`0R00020@068P010B800`400`01
0@030@039`4104D000<100004P000`40000R0000B`000`40000n00<14@000`40000D00<1H0000`4R
0@020@0;8P410B818P4R0B800`400b810@1700030@00014000<100008P0004/000<10000?0020A@0
00<100004P020F@010401R800@4R00D100@00@41B@000`40000A00030@000280001;00030@0003T0
0`4G00030@0000h00`5X008100<R0B800`4400<1C0000`40000@00030@000280001;00030@0003L0
0P4K00030@0000/00P5/00D1E@000`40000@00030@000280001;00030@0003@00`4N00030@0000L0
0`7900030@0000l000<100008P0004/000<10000<@030B8000<100000`030Ld000<100003P000`40
000R0000B`000`40000_00819P000`4000020M0000<100003P000`40000R0000B`030B`00`4Y0081
d`000`40000=00030@000280001;00030@0002X00P4/00030@000=8000<100003@000`40000R0000
B`000`40000W00<1;P000`40003C00030@0000`000<100008P0004/000<100009@020C4000<10000
d`000`40000<00030@000280001;00030@0002800`4c00030@000=@000<100002`000`40000R0000
B`000`40000P0081=P000`40003E00030@0000X000<100008P0004/000<100007@030CP000<10000
e@000`40000:00030@000280001;00030@0001/00P4k00030@000=H000<100002@000`40000R0000
B`000`40000H00<1?@000`40003F00030@0000T000<100008P0004/00P4G0081@0000`40003G0003
0@0000P000<100008P0004/000<100004`030D8000<10000f0000`40000700030@000280001;0003
0@0001400P5500030@000=P000<100001`000`40000R0000B`000`40000>00<1A`000`40003I0003
0@0000H000<100008P0004/000<100002`030DX000<10000f@000`40000600030@000280001;0004
0@000@P00P5=00030@000=X000<100001@000`40000R0000B`001040004500<1C`000`40003J0003
0@0000D000<100008P0004/000@100010`020E8000<10000f`000`40000400030@000280001;0003
0@0000@1E0000`40003L00030@0000<000<100008P0002/00`4800812@030@L0105G00030@000=`0
00<100000`000`40000R0000:P001@40000100L00P4800050@0000401P020@03004005H000<10000
g@000`40000200030@000280000Y00030@0000<000<100003@000`40000300030@0000@000<10040
E`000`40003M00030@00008000<100008P0002T000<100000`000`40000=00030@0000<000<10000
1@020EL000<10000gP001@40000102@0000Y00030@0000<000<100003@000`40000300030@0000H0
00<10000E@000`40003O00040@000B@0000Y00030@0000<000<100003@020@<000<1000020000`40
000200030@0004l000<10000g`001040004T0000:@000`40000300030@0000d000<100400P4<0081
0P000`40001?00030@000>0000<10040900002T000<100000`000`40000=00030@00010000<10040
D@000`40003P00030@0102@0000Y00030@0000<000<100003P000`40000@0081D@000`40003Q0081
900002T000<100000`000`40000>00030@00014000<10000C`000`40003Q0081900002X000D10000
0@0B00030@0001000P5@00030@000>400P4T0000:`030A@0104@00030@00008000<10000B0000`40
003N00<19P0005L000D100000@1:00030@000=/00`4Y0000F0001040005:00030@000=P00`4/0000
F@000`400@1:00030@000=D00`4_0000FP020DX000<10000d@040C80001K00030@0004P000<10000
cP030CH0001L00810`000`40001200030@000</00`4i0000GP001040005400030@000<P00`4l0000
G`000`400@1400030@000<D00`4o0000H0020D@000<10000`P030D80001Q00030@00048000<10000
_`030DD0001<00<14`000`40000200030@0003`000<10000_0030DP0001;00050@0000404`001@40
000103h000<10000^@030D/0001:00030@0000<000<100004@001040004n00030@000;H00`5>0000
BP000`40000300030@00018000<100401@000`40000f00030@000;<00`5A0000BP000`4000030003
0@0001<00P4500030@0003H000<10000[`040E@0001:00030@0000<000<1000050020@@000<10000
=P000`40002/00<1F00004X000<100000`000`40000F00050@000040>0000`40002Y00<1F`0004X0
00<100000`000`40000G00040@000CP000<10000YP030Eh0001:00030@0000<000<1000060000`40
0@0h00030@000:<00`5Q0000BP000`40000300030@0001T00P4h00030@000:000`5T0000B`001@40
000101d000<10000=P000`40002M00<1I`0004`00`4O00030@00008000<10000<0000`40002J00<1
JP0006l000D100000@0b00030@0009L00`5]0000L0001040004b00030@0009@00`5`0000L@000`40
0@0b00030@000900105c0000LP020C8000<10000S@030GL0001c0081<@000`40002:00<1NP0007D0
00D100000@0/00030@0008L00`5m0000MP001040004/00030@0008@00`600000M`000`400@0/0003
0@0008400`630000N0020B`000<10000OP030HH0001i00030@0002X000<10000N`030HT0001j0005
0@0000409`000`40001h00<1S00007/000@100019`000`40001e00<1S`0007`000<100409`000`40
001b00<1TP0007d00P4W00030@0006h0106E0000OP020BH000<10000J`030IT0002000050@000040
8@000`40001X00<1W000084000@100018@000`40001U00<1W`0006T01`4B00030@0100D000<10000
6@000`40001R00<1XP0006`000<1000050020@D000<100006@000`40001O00<1Y@0006`000<10000
5@000`40000300030@0001T000<10000G0030JP0001/00030@0001H000<100000P000`40000I0003
0@0005T00`6[0000K0000`40000G00050@0000406`000`40001F00<1[P0006`000<1000060001040
004K00030@0005<00`6a0000K0000`40000I00030@0101/000<10000D0030K@0001/00030@0001X0
0P4K00030@0004`0106g0000K0000`40000L00050@0000405P000`40001900<1^`0006`000<10000
7@001040004F00030@0004H00`6n0000J@020@03004001l000<100405P000`40001300<1`@0006/0
0P4Q00815P000`40001000<1a00008l000<1000050000`40000m00<1a`00090000D100000@0A0003
0@0003X00`7:0000T@001040004A00030@0003L00`7=0000TP000`400@0A00030@0003@00`7@0000
T`020A4000<10000<@030M<0002D008140000`40000]00@1eP0009H000D100000@0;00030@0002X0
0`7J0000U`001040004;00030@0002L00`7M0000V0000`400@0;00030@0002@00`7P0000V@020@/0
00<100008@030N<0002J00030@00008000<1000010000`40000N00<1iP0009/000D100000@060003
0@0001/00`7Y0000W0001040004400030@0101X00`7/0000W@000`400@0400030@0101L00`7_0000
Q@080A400P4400030@0101@00`7b0000QP000`40000300030@00010000<100000P000`400@0A00<1
m@0008L000<100005P020@8000<100403@040OP0002800030@0001L000D100400@0:00<1o00008T0
00<100005`020@03004000H00`7o0000RP000`40000G00030@0100@00`7o00<0002;00030@0001L0
0P4010010@7o00H0002<00030@0001L00P7o00T0002600030@0000<000<10000o`0R0000QP000`40
000300030@000?l08P0008L000D100000@3o02D0002800<1o`0V0000o`2a0000o`2a0000o`2a0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000X@040Ol0
30000:0000<100000P000`40003o00T0002V00030@000?l020000:H000<10000o`080000YP000`40
003o00P0002U00030@000?l02@000:800`7o00`0002U00030@000?l02@000:H000<10000o`080000
YP000`40003o00P0002P00030@00008000<10000o`090000X@040Ol030000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@0007L01@7o03D0001j00030@000?l0=00007X000<10000o`0d0000N`000`40
003o03<0001k00030@000?l0<`0007X000<10040o`0d0000NP000`400@3o03@0001i00050@000040
o`0c0000N@001@4000010?l0<`0007P000<100000`000`40003o0300001h00030@0000<000<10000
o`0`0000M`030@<00`7o0340003o0;40003o0;40003o0;400000\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.234686, 0.31168, 0.00465498, 
  0.00125278}}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Level curves", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(levels = Table[j, {j, 0, 1, 0.05}]; \n
    glevel = LevelGraphics2D[points, elements, levels]; \nShow[glevel]; \)], 
  "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0.0147151 0.196201 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .01472 m
.02381 .05396 L
s
.02381 .01472 m
.0619 .01472 L
s
.02381 .05396 m
.02381 .05396 L
s
.0619 .01472 m
.0619 .01472 L
s
.02381 .05396 m
.02381 .0932 L
s
.02381 .05396 m
.02381 .0932 L
s
.0619 .01472 m
.1 .01472 L
s
.1 .01472 m
.0619 .01472 L
s
.02381 .0932 m
.02381 .0932 L
s
.1 .01472 m
.1 .01472 L
s
.02381 .13244 m
.02381 .0932 L
s
.02381 .13244 m
.02381 .0932 L
s
.1 .01472 m
.1381 .01472 L
s
.1381 .01472 m
.1 .01472 L
s
.02381 .13244 m
.02381 .13244 L
s
.1381 .01472 m
.1381 .01472 L
s
.02381 .17168 m
.02381 .13244 L
s
.02381 .17168 m
.02381 .13244 L
s
.1381 .01472 m
.17619 .01472 L
s
.1381 .01472 m
.17619 .01472 L
s
.02381 .17168 m
.02381 .17168 L
s
.17619 .01472 m
.17619 .01472 L
s
.02381 .21092 m
.02381 .17168 L
s
.02381 .21092 m
.02381 .17168 L
s
.17619 .01472 m
.21429 .01472 L
s
.17619 .01472 m
.21429 .01472 L
s
.02381 .21092 m
.02381 .21092 L
s
.21429 .01472 m
.21429 .01472 L
s
.02381 .25016 m
.02381 .21092 L
s
.02381 .25016 m
.02381 .21092 L
s
.21429 .01472 m
.25238 .01472 L
s
.25238 .01472 m
.21429 .01472 L
s
.02381 .25016 m
.02381 .25016 L
s
.25238 .01472 m
.25238 .01472 L
s
.02381 .2894 m
.02381 .25016 L
s
.02381 .2894 m
.02381 .25016 L
s
.25238 .01472 m
.29048 .01472 L
s
.25238 .01472 m
.29048 .01472 L
s
.02381 .2894 m
.02381 .2894 L
s
.29048 .01472 m
.29048 .01472 L
s
.02381 .32864 m
.02381 .2894 L
s
.02381 .2894 m
.02381 .32864 L
s
.29048 .01472 m
.32857 .01472 L
s
.32857 .01472 m
.29048 .01472 L
s
.02381 .32864 m
.02381 .32864 L
s
.32857 .01472 m
.32857 .01472 L
s
.02381 .36788 m
.02381 .32864 L
s
.02381 .32864 m
.02381 .36788 L
s
.32857 .01472 m
.36667 .01472 L
s
.32857 .01472 m
.36667 .01472 L
s
.02381 .36788 m
.02381 .36788 L
s
.36667 .01472 m
.36667 .01472 L
s
.1 .40712 m
.1 .40712 L
s
.02381 .40712 m
.02381 .36788 L
s
.02381 .40712 m
.02381 .36788 L
s
.0619 .40712 m
.0619 .40712 L
s
.1381 .40712 m
.1381 .40712 L
s
.1 .40712 m
.0619 .40712 L
s
.1 .40712 m
.0619 .40712 L
s
.1381 .40712 m
.1 .40712 L
s
.1 .40712 m
.1381 .40712 L
s
.36667 .01472 m
.40476 .01472 L
s
.40476 .01472 m
.36667 .01472 L
s
.0619 .40712 m
.02381 .40712 L
s
.0619 .40712 m
.02381 .40712 L
s
.17619 .40712 m
.1381 .40712 L
s
.17619 .40712 m
.1381 .40712 L
s
.17619 .40712 m
.17619 .40712 L
s
.40476 .01472 m
.40476 .01472 L
s
.21429 .40712 m
.17619 .40712 L
s
.21429 .40712 m
.17619 .40712 L
s
.21429 .40712 m
.21429 .40712 L
s
.40476 .01472 m
.44286 .01472 L
s
.40476 .01472 m
.44286 .01472 L
s
.25238 .40712 m
.21429 .40712 L
s
.21429 .40712 m
.25238 .40712 L
s
.44286 .01472 m
.44286 .01472 L
s
.25238 .40712 m
.25238 .40712 L
s
.44286 .01472 m
.48095 .01472 L
s
.44286 .01472 m
.48095 .01472 L
s
.29048 .40712 m
.25238 .40712 L
s
.29048 .40712 m
.25238 .40712 L
s
.29048 .40712 m
.29048 .40712 L
s
.48095 .01472 m
.48095 .01472 L
s
.32857 .40712 m
.29048 .40712 L
s
.29048 .40712 m
.32857 .40712 L
s
.32857 .40712 m
.32857 .40712 L
s
.48095 .01472 m
.51905 .01472 L
s
.48095 .01472 m
.51905 .01472 L
s
.36667 .40712 m
.32857 .40712 L
s
.36667 .40712 m
.32857 .40712 L
s
.51905 .01472 m
.51905 .01472 L
s
.36667 .40712 m
.36667 .40712 L
s
.51905 .01472 m
.55714 .01472 L
s
.55714 .01472 m
.51905 .01472 L
s
.40476 .40712 m
.36667 .40712 L
s
.36667 .40712 m
.40476 .40712 L
s
.40476 .40712 m
.40476 .40712 L
s
.55714 .01472 m
.55714 .01472 L
s
.44286 .40712 m
.40476 .40712 L
s
.44286 .40712 m
.40476 .40712 L
s
.44286 .40712 m
.44286 .40712 L
s
.55714 .01472 m
.59524 .01472 L
s
.55714 .01472 m
.59524 .01472 L
s
.48095 .40712 m
.44286 .40712 L
s
.48095 .40712 m
.44286 .40712 L
s
.59524 .01472 m
.59524 .01472 L
s
.48095 .40712 m
.48095 .40712 L
s
.59524 .01472 m
.63333 .01472 L
s
.63333 .01472 m
.59524 .01472 L
s
.51905 .40712 m
.48095 .40712 L
s
.51905 .40712 m
.48095 .40712 L
s
.51905 .40712 m
.51905 .40712 L
s
.63333 .01472 m
.63333 .01472 L
s
.55714 .40712 m
.51905 .40712 L
s
.51905 .40712 m
.55714 .40712 L
s
.55714 .40712 m
.55714 .40712 L
s
.63333 .01472 m
.67143 .01472 L
s
.63333 .01472 m
.67143 .01472 L
s
.59524 .40712 m
.55714 .40712 L
s
.55714 .40712 m
.59524 .40712 L
s
.67143 .01472 m
.67143 .01472 L
s
.59524 .40712 m
.59524 .40712 L
s
.67143 .01472 m
.70952 .01472 L
s
.70952 .01472 m
.67143 .01472 L
s
.59524 .40712 m
.59524 .40712 L
s
.70952 .01472 m
.70952 .01472 L
s
.59524 .40712 m
.59524 .40712 L
s
.59524 .44636 m
.59524 .40712 L
s
.59524 .40712 m
.59524 .44636 L
s
.70952 .01472 m
.74762 .01472 L
s
.74762 .01472 m
.70952 .01472 L
s
.59524 .44636 m
.59524 .44636 L
s
.74762 .01472 m
.74762 .01472 L
s
.59524 .4856 m
.59524 .44636 L
s
.59524 .4856 m
.59524 .44636 L
s
.74762 .01472 m
.78571 .01472 L
s
.78571 .01472 m
.74762 .01472 L
s
.59524 .4856 m
.59524 .4856 L
s
.78571 .01472 m
.78571 .01472 L
s
.59524 .52484 m
.59524 .4856 L
s
.59524 .52484 m
.59524 .4856 L
s
.78571 .01472 m
.82381 .01472 L
s
.82381 .01472 m
.78571 .01472 L
s
.59524 .52484 m
.59524 .52484 L
s
.82381 .01472 m
.82381 .01472 L
s
.59524 .56408 m
.59524 .52484 L
s
.59524 .52484 m
.59524 .56408 L
s
.82381 .01472 m
.8619 .01472 L
s
.82381 .01472 m
.8619 .01472 L
s
.59524 .56408 m
.59524 .56408 L
s
.8619 .01472 m
.8619 .01472 L
s
.59524 .60332 m
.59524 .56408 L
s
.59524 .60332 m
.59524 .56408 L
s
.63333 .60332 m
.63333 .60332 L
s
.8619 .01472 m
.9 .01472 L
s
.8619 .01472 m
.9 .01472 L
s
.67143 .60332 m
.67143 .60332 L
s
.67143 .60332 m
.63333 .60332 L
s
.67143 .60332 m
.63333 .60332 L
s
.63333 .60332 m
.59524 .60332 L
s
.63333 .60332 m
.59524 .60332 L
s
.9 .01472 m
.9 .01472 L
s
.70952 .60332 m
.67143 .60332 L
s
.67143 .60332 m
.70952 .60332 L
s
.70952 .60332 m
.70952 .60332 L
s
.9 .01472 m
.9381 .01472 L
s
.9 .01472 m
.9381 .01472 L
s
.74762 .60332 m
.70952 .60332 L
s
.70952 .60332 m
.74762 .60332 L
s
.74762 .60332 m
.74762 .60332 L
s
.9381 .01472 m
.9381 .01472 L
s
.9381 .01472 m
.97619 .01472 L
s
.9381 .01472 m
.97619 .01472 L
s
.78571 .60332 m
.74762 .60332 L
s
.74762 .60332 m
.78571 .60332 L
s
.97619 .05396 m
.97619 .05396 L
s
.78571 .60332 m
.78571 .60332 L
s
.97619 .0932 m
.97619 .0932 L
s
.97619 .01472 m
.97619 .05396 L
s
.97619 .01472 m
.97619 .05396 L
s
.97619 .05396 m
.97619 .0932 L
s
.97619 .0932 m
.97619 .05396 L
s
.82381 .60332 m
.78571 .60332 L
s
.82381 .60332 m
.78571 .60332 L
s
.97619 .0932 m
.97619 .13244 L
s
.97619 .0932 m
.97619 .13244 L
s
.97619 .13244 m
.97619 .13244 L
s
.82381 .60332 m
.82381 .60332 L
s
.97619 .13244 m
.97619 .17168 L
s
.97619 .13244 m
.97619 .17168 L
s
.8619 .60332 m
.82381 .60332 L
s
.82381 .60332 m
.8619 .60332 L
s
.8619 .60332 m
.8619 .60332 L
s
.97619 .17168 m
.97619 .17168 L
s
.9 .60332 m
.8619 .60332 L
s
.8619 .60332 m
.9 .60332 L
s
.97619 .17168 m
.97619 .21092 L
s
.97619 .17168 m
.97619 .21092 L
s
.9 .60332 m
.9 .60332 L
s
.97619 .21092 m
.97619 .21092 L
s
.97619 .4856 m
.97619 .4856 L
s
.97619 .52484 m
.97619 .52484 L
s
.9381 .60332 m
.9 .60332 L
s
.9 .60332 m
.9381 .60332 L
s
.97619 .4856 m
.97619 .52484 L
s
.97619 .52484 m
.97619 .4856 L
s
.97619 .21092 m
.97619 .25016 L
s
.97619 .21092 m
.97619 .25016 L
s
.9381 .60332 m
.9381 .60332 L
s
.97619 .44636 m
.97619 .4856 L
s
.97619 .44636 m
.97619 .4856 L
s
.97619 .44636 m
.97619 .44636 L
s
.97619 .52484 m
.97619 .56408 L
s
.97619 .52484 m
.97619 .56408 L
s
.97619 .56408 m
.97619 .56408 L
s
.97619 .25016 m
.97619 .25016 L
s
.97619 .40712 m
.97619 .40712 L
s
.97619 .60332 m
.9381 .60332 L
s
.97619 .60332 m
.9381 .60332 L
s
.97619 .40712 m
.97619 .44636 L
s
.97619 .44636 m
.97619 .40712 L
s
.97619 .56408 m
.97619 .60332 L
s
.97619 .56408 m
.97619 .60332 L
s
.97619 .36788 m
.97619 .36788 L
s
.97619 .36788 m
.97619 .40712 L
s
.97619 .40712 m
.97619 .36788 L
s
.97619 .25016 m
.97619 .2894 L
s
.97619 .25016 m
.97619 .2894 L
s
.97619 .32864 m
.97619 .32864 L
s
.97619 .2894 m
.97619 .2894 L
s
.97619 .32864 m
.97619 .36788 L
s
.97619 .32864 m
.97619 .36788 L
s
.97619 .2894 m
.97619 .32864 L
s
.97619 .2894 m
.97619 .32864 L
s
.41193 .1595 m
.41095 .16144 L
s
.41095 .16144 m
.40373 .1857 L
s
.41465 .15577 m
.41193 .1595 L
s
.40182 .1921 m
.39933 .22268 L
s
.40648 .25923 m
.40345 .24119 L
s
.41465 .15577 m
.42875 .13118 L
s
.4552 .10316 m
.43602 .12166 L
s
.40188 .19154 m
.40373 .1857 L
s
.40345 .24119 m
.39933 .22268 L
s
.42875 .13118 m
.43602 .12166 L
s
.40182 .1921 m
.40188 .19154 L
s
.40648 .25923 m
.41205 .269 L
s
.48098 .086 m
.4552 .10316 L
s
.41205 .269 m
.42975 .29992 L
s
.48285 .08429 m
.48098 .086 L
s
.42975 .29992 m
.43252 .30272 L
s
.51138 .06863 m
.48437 .08331 L
s
.48285 .08429 m
.48437 .08331 L
s
.43252 .30272 m
.43704 .30645 L
s
.51138 .06863 m
.53085 .06299 L
s
.45464 .32555 m
.43704 .30645 L
s
.54204 .05709 m
.53085 .06299 L
s
.46726 .33569 m
.45464 .32555 L
s
.57611 .04713 m
.55961 .0516 L
s
.46726 .33569 m
.48042 .34444 L
s
.55961 .0516 m
.54204 .05709 L
s
.57611 .04713 m
.58221 .04713 L
s
.48042 .34444 m
.49009 .34788 L
s
.49009 .34788 m
.50885 .36025 L
s
.61243 .03977 m
.58221 .04713 L
s
.50885 .36025 m
.53864 .37148 L
s
.62081 .03977 m
.61243 .03977 L
s
.53864 .37148 m
.53943 .372 L
s
.62081 .03977 m
.64557 .03441 L
s
.54017 .3724 m
.53943 .372 L
s
.54017 .3724 m
.57104 .38414 L
s
.64557 .03441 m
.66261 .03441 L
s
.57104 .38414 m
.58647 .38414 L
s
.68219 .03104 m
.66261 .03441 L
s
.60463 .39447 m
.58647 .38414 L
s
.70045 .03104 m
.68219 .03104 L
s
.61291 .40428 m
.60463 .39447 L
s
.71727 .02892 m
.70045 .03104 L
s
.62093 .42603 m
.61291 .40428 L
s
.62093 .42603 m
.62093 .43614 L
s
.73988 .02892 m
.71727 .02892 L
s
.62093 .43614 m
.62842 .45691 L
s
.75471 .02819 m
.73988 .02892 L
s
.62842 .45691 m
.63303 .46582 L
s
.77851 .02819 m
.75471 .02819 L
s
.63303 .46582 m
.64061 .47624 L
s
.79453 .02894 m
.77851 .02819 L
s
.64061 .47624 m
.64841 .49322 L
s
.81712 .02894 m
.79453 .02894 L
s
.64841 .49322 m
.65833 .50515 L
s
.67015 .51602 m
.65833 .50515 L
s
.81712 .02894 m
.83568 .03144 L
s
.67015 .51602 m
.68964 .52803 L
s
.83568 .03144 m
.85197 .03144 L
s
.69883 .53217 m
.68964 .52803 L
s
.85197 .03144 m
.87645 .03707 L
s
.71852 .53889 m
.69883 .53217 L
s
.87645 .03707 m
.8852 .03707 L
s
.73536 .54177 m
.71852 .53889 L
s
.8852 .03707 m
.90284 .04346 L
s
.90284 .04346 m
.91494 .04772 L
s
.73536 .54177 m
.74747 .5442 L
s
.75572 .54466 m
.74747 .5442 L
s
.91494 .04772 m
.93561 .06868 L
s
.78757 .54538 m
.75572 .54466 L
s
.93998 .07205 m
.93561 .06868 L
s
.79327 .54528 m
.78757 .54538 L
s
.93998 .07205 m
.94089 .07324 L
s
.94089 .07485 m
.95477 .10708 L
s
.82586 .54111 m
.79327 .54528 L
s
.94089 .07485 m
.94089 .07324 L
s
.95477 .10708 m
.95477 .12017 L
s
.82586 .54111 m
.82702 .541 L
s
.95477 .12017 m
.96007 .14162 L
s
.8605 .53178 m
.83052 .53991 L
s
.82702 .541 m
.83052 .53991 L
s
.87461 .52286 m
.8605 .53178 L
s
.96007 .14162 m
.96007 .16226 L
s
.89185 .51426 m
.87461 .52286 L
s
.96007 .16226 m
.96242 .17923 L
s
.91867 .48548 m
.89995 .5043 L
s
.89995 .5043 m
.89185 .51426 L
s
.96242 .17923 m
.96242 .20304 L
s
.92658 .4644 m
.91867 .48548 L
s
.93425 .45243 m
.92658 .4644 L
s
.96242 .20304 m
.96304 .2195 L
s
.94128 .42567 m
.94361 .42073 L
s
.94128 .42567 m
.93425 .45243 L
s
.96304 .2195 m
.96304 .2437 L
s
.94771 .40173 m
.94361 .42073 L
s
.96304 .2437 m
.96219 .26064 L
s
.94771 .40173 m
.95054 .38861 L
s
.95614 .35504 m
.95054 .38533 L
s
.95054 .38861 m
.95054 .38533 L
s
.96219 .26064 m
.96219 .28107 L
s
.95998 .31861 m
.95614 .343 L
s
.96219 .28107 m
.95998 .30079 L
s
.95614 .343 m
.95614 .35504 L
s
.95998 .30079 m
.95998 .31861 L
s
.51369 .13406 m
.50398 .14747 L
s
.49763 .15999 m
.49021 .17875 L
s
.49763 .15999 m
.50398 .14747 L
s
.53046 .11514 m
.53491 .11129 L
s
.48776 .18655 m
.48427 .21455 L
s
.51369 .13406 m
.53046 .11514 L
s
.49021 .17875 m
.48776 .18655 L
s
.48714 .24651 m
.48414 .21847 L
s
.48427 .21455 m
.48414 .21847 L
s
.53491 .11129 m
.54192 .10677 L
s
.49528 .27272 m
.48917 .25426 L
s
.48917 .25426 m
.48714 .24651 L
s
.50705 .29123 m
.49528 .27272 L
s
.54192 .10677 m
.55903 .09264 L
s
.58518 .07822 m
.56782 .08678 L
s
.55903 .09264 m
.56782 .08678 L
s
.51076 .2993 m
.50705 .29123 L
s
.51494 .30485 m
.53237 .32427 L
s
.51494 .30485 m
.51076 .2993 L
s
.60306 .06995 m
.58518 .07822 L
s
.60837 .06868 m
.60306 .06995 L
s
.5508 .34005 m
.53237 .32427 L
s
.5508 .34005 m
.55442 .3424 L
s
.60837 .06868 m
.62538 .06164 L
s
.62538 .06164 m
.65057 .05589 L
s
.55442 .3424 m
.56037 .34561 L
s
.56037 .34561 m
.58161 .35958 L
s
.65057 .05589 m
.656 .05391 L
s
.58161 .35958 m
.58635 .36418 L
s
.65875 .05302 m
.656 .05391 L
s
.69139 .04752 m
.65875 .05302 L
s
.61403 .38182 m
.59499 .371 L
s
.59499 .371 m
.58635 .36418 L
s
.69139 .04752 m
.69264 .04717 L
s
.63059 .40145 m
.61403 .38182 L
s
.72502 .04312 m
.69745 .04659 L
s
.69264 .04717 m
.69745 .04659 L
s
.6345 .41205 m
.63059 .40145 L
s
.73214 .04312 m
.72502 .04312 L
s
.64631 .42649 m
.6345 .41205 L
s
.76179 .04166 m
.73214 .04312 L
s
.66067 .44635 m
.64631 .42649 L
s
.77131 .04166 m
.76179 .04166 L
s
.66501 .45021 m
.66067 .44635 L
s
.80334 .04316 m
.77131 .04166 L
s
.68423 .46659 m
.66501 .45021 L
s
.68824 .47049 m
.68423 .46659 L
s
.81042 .04316 m
.80334 .04316 L
s
.68824 .47049 m
.68961 .47157 L
s
.81042 .04316 m
.83128 .04598 L
s
.71704 .48596 m
.68961 .47157 L
s
.83128 .04598 m
.84265 .04749 L
s
.71704 .48596 m
.71928 .48698 L
s
.84265 .04749 m
.84752 .04921 L
s
.7493 .49484 m
.73027 .48958 L
s
.84752 .04921 m
.87349 .05494 L
s
.71928 .48698 m
.73027 .48958 L
s
.7493 .49484 m
.75369 .4951 L
s
.87349 .05494 m
.89636 .06818 L
s
.89636 .06818 m
.90072 .06972 L
s
.77965 .49726 m
.75369 .4951 L
s
.79866 .49478 m
.77965 .49726 L
s
.90072 .06972 m
.9035 .07254 L
s
.9035 .07254 m
.92319 .09073 L
s
.79866 .49478 m
.80821 .494 L
s
.92319 .09073 m
.93012 .10858 L
s
.93012 .10858 m
.93625 .11901 L
s
.80821 .494 m
.81698 .49143 L
s
.81698 .49143 m
.83836 .48565 L
s
.83836 .48565 m
.86545 .46904 L
s
.86757 .46796 m
.86545 .46904 L
s
.93625 .11901 m
.94252 .14502 L
s
.94395 .15081 m
.94252 .14502 L
s
.87121 .46441 m
.86757 .46796 L
s
.94395 .15081 m
.94395 .15285 L
s
.89118 .44664 m
.87121 .46441 L
s
.94395 .15285 m
.94864 .18679 L
s
.89649 .43785 m
.89118 .44664 L
s
.9073 .42431 m
.89649 .43785 L
s
.94864 .18679 m
.94864 .19516 L
s
.91866 .40162 m
.9156 .40649 L
s
.9156 .40649 m
.9073 .42431 L
s
.92142 .3952 m
.91866 .40162 L
s
.94864 .19516 m
.94989 .22808 L
s
.92934 .37528 m
.92142 .3952 L
s
.94989 .22808 m
.94989 .23724 L
s
.93799 .34438 m
.93203 .36127 L
s
.93203 .36127 m
.92934 .37528 L
s
.94989 .23724 m
.94819 .27113 L
s
.94307 .31269 m
.93799 .34438 L
s
.94424 .30873 m
.94307 .31269 L
s
.94819 .27113 m
.94819 .27275 L
s
.94681 .285 m
.94424 .30873 L
s
.94819 .27275 m
.94681 .285 L
s
.53362 .21464 m
.53354 .21758 L
s
.53683 .199 m
.53362 .21464 L
s
.53354 .21758 m
.53547 .23557 L
s
.55561 .14779 m
.5405 .1784 L
s
.53683 .199 m
.53983 .18073 L
s
.55712 .14517 m
.57414 .12434 L
s
.53983 .18073 m
.5405 .1784 L
s
.55712 .14517 m
.55561 .14779 L
s
.53547 .23557 m
.53638 .24808 L
s
.59435 .1064 m
.58075 .11733 L
s
.57414 .12434 m
.58075 .11733 L
s
.54452 .27487 m
.5368 .2498 L
s
.5368 .2498 m
.53638 .24808 L
s
.55687 .29756 m
.55097 .28745 L
s
.60947 .09567 m
.61437 .09301 L
s
.59435 .1064 m
.60947 .09567 L
s
.55097 .28745 m
.54452 .27487 L
s
.55687 .29756 m
.56479 .30758 L
s
.61437 .09301 m
.62215 .08944 L
s
.62215 .08944 m
.63515 .08201 L
s
.56479 .30758 m
.56908 .31434 L
s
.65882 .07261 m
.64233 .07862 L
s
.63515 .08201 m
.64233 .07862 L
s
.56908 .31434 m
.58511 .33162 L
s
.58542 .33196 m
.6047 .35066 L
s
.67601 .06705 m
.65882 .07261 L
s
.68237 .06598 m
.67601 .06705 L
s
.6047 .35066 m
.61521 .35895 L
s
.5852 .33174 m
.58511 .33162 L
s
.58542 .33196 m
.5852 .33174 L
s
.68237 .06598 m
.69944 .06117 L
s
.61521 .35895 m
.6254 .36875 L
s
.69944 .06117 m
.72159 .0585 L
s
.64258 .38738 m
.63211 .37558 L
s
.63211 .37558 m
.6254 .36875 L
s
.72159 .0585 m
.72908 .05697 L
s
.65146 .39585 m
.64258 .38738 L
s
.65146 .39585 m
.66453 .41081 L
s
.72908 .05697 m
.73478 .05636 L
s
.73478 .05636 m
.76158 .05517 L
s
.66453 .41081 m
.67164 .41642 L
s
.67164 .41642 m
.69075 .4323 L
s
.76158 .05517 m
.76599 .05485 L
s
.76599 .05485 m
.76996 .05497 L
s
.76996 .05497 m
.80222 .05627 L
s
.69526 .43474 m
.69075 .4323 L
s
.69526 .43474 m
.71841 .44734 L
s
.80222 .05627 m
.8085 .05709 L
s
.82142 .0597 m
.8085 .05709 L
s
.83515 .06153 m
.82142 .0597 L
s
.71841 .44734 m
.7236 .44888 L
s
.7236 .44888 m
.7465 .45595 L
s
.83515 .06153 m
.85925 .07006 L
s
.85925 .07006 m
.86309 .0709 L
s
.7465 .45595 m
.75818 .45687 L
s
.75818 .45687 m
.77448 .45863 L
s
.86309 .0709 m
.86749 .07345 L
s
.86749 .07345 m
.88987 .0861 L
s
.77448 .45863 m
.80163 .45516 L
s
.80187 .45514 m
.80163 .45516 L
s
.90254 .1005 m
.88987 .0861 L
s
.80204 .45508 m
.80187 .45514 L
s
.90254 .1005 m
.90805 .10612 L
s
.82858 .44794 m
.80204 .45508 L
s
.91565 .12051 m
.90805 .10612 L
s
.92063 .12897 m
.91565 .12051 L
s
.82858 .44794 m
.84867 .4362 L
s
.84867 .4362 m
.85333 .43397 L
s
.92063 .12897 m
.92207 .13496 L
s
.86043 .42783 m
.85333 .43397 L
s
.92207 .13496 m
.9285 .15234 L
s
.87607 .41539 m
.86043 .42783 L
s
.9285 .15234 m
.93112 .16226 L
s
.87607 .41539 m
.88021 .40991 L
s
.89434 .39113 m
.8966 .38846 L
s
.93112 .16226 m
.93366 .18207 L
s
.88021 .40991 m
.89434 .39113 L
s
.8972 .38758 m
.8966 .38846 L
s
.93366 .18207 m
.93518 .19004 L
s
.90965 .36531 m
.8972 .38758 L
s
.93518 .19004 m
.93577 .19545 L
s
.93577 .19545 m
.93698 .2292 L
s
.91744 .34657 m
.90965 .36531 L
s
.93698 .2292 m
.93705 .23409 L
s
.92136 .33544 m
.91744 .34657 L
s
.93648 .24113 m
.93705 .23409 L
s
.92271 .32704 m
.92136 .33544 L
s
.93543 .26551 m
.93648 .24113 L
s
.93084 .29962 m
.92271 .32704 L
s
.93543 .26551 m
.93293 .28035 L
s
.93293 .28035 m
.93084 .29962 L
s
.57041 .20105 m
.57712 .17715 L
s
.58995 .15156 m
.57927 .17121 L
s
.57712 .17715 m
.57927 .17121 L
s
.56971 .20693 m
.56867 .23053 L
s
.59465 .14414 m
.60569 .13116 L
s
.56971 .20693 m
.57041 .20105 L
s
.59465 .14414 m
.58995 .15156 L
s
.56967 .23922 m
.57279 .25692 L
s
.56967 .23922 m
.56867 .23053 L
s
.62296 .11498 m
.6151 .12135 L
s
.57868 .27362 m
.5812 .27989 L
s
.60569 .13116 m
.6151 .12135 L
s
.57868 .27362 m
.57279 .25692 L
s
.6399 .10284 m
.64167 .10183 L
s
.62296 .11498 m
.6399 .10284 L
s
.5812 .27989 m
.58561 .28718 L
s
.64167 .10183 m
.64536 .10008 L
s
.59197 .30029 m
.58561 .28718 L
s
.6055 .31912 m
.59953 .31086 L
s
.64536 .10008 m
.66204 .0907 L
s
.61056 .32413 m
.6055 .31912 L
s
.59953 .31086 m
.59197 .30029 L
s
.68261 .0821 m
.66816 .08764 L
s
.66816 .08764 m
.66204 .0907 L
s
.70625 .07517 m
.70338 .07598 L
s
.63499 .35136 m
.63899 .35521 L
s
.70338 .07598 m
.68261 .0821 L
s
.61056 .32413 m
.62156 .33764 L
s
.70625 .07517 m
.70864 .07488 L
s
.62156 .33764 m
.63499 .35136 L
s
.63899 .35521 m
.64281 .35909 L
s
.70864 .07488 m
.73 .07053 L
s
.64281 .35909 m
.65495 .37107 L
s
.73 .07053 m
.74623 .06879 L
s
.65495 .37107 m
.66837 .38364 L
s
.74623 .06879 m
.74681 .06876 L
s
.66837 .38364 m
.674 .38789 L
s
.74681 .06876 m
.76443 .06746 L
s
.674 .38789 m
.69256 .40221 L
s
.76443 .06746 m
.78026 .06795 L
s
.78026 .06795 m
.79274 .06845 L
s
.69705 .4046 m
.69256 .40221 L
s
.71803 .41603 m
.69705 .4046 L
s
.79274 .06845 m
.81058 .07077 L
s
.72409 .41794 m
.71803 .41603 L
s
.82292 .07326 m
.81058 .07077 L
s
.72409 .41794 m
.74412 .4245 L
s
.83325 .07618 m
.82292 .07326 L
s
.74412 .4245 m
.75633 .42575 L
s
.84842 .08126 m
.83325 .07618 L
s
.77016 .42752 m
.75633 .42575 L
s
.84842 .08126 m
.85864 .08617 L
s
.79463 .42489 m
.77016 .42752 L
s
.87692 .09867 m
.85864 .08617 L
s
.79463 .42489 m
.79555 .42485 L
s
.88015 .10106 m
.87692 .09867 L
s
.79615 .42467 m
.79555 .42485 L
s
.79615 .42467 m
.82011 .41877 L
s
.88015 .10106 m
.88728 .10924 L
s
.88728 .10924 m
.89591 .11807 L
s
.82011 .41877 m
.83934 .40839 L
s
.89875 .12345 m
.89591 .11807 L
s
.83934 .40839 m
.84215 .40715 L
s
.89875 .12345 m
.9076 .13723 L
s
.84548 .40462 m
.84215 .40715 L
s
.84548 .40462 m
.86026 .39414 L
s
.91266 .15134 m
.9076 .13723 L
s
.87404 .38002 m
.86026 .39414 L
s
.91266 .15134 m
.91427 .15517 L
s
.87404 .38002 m
.87594 .37778 L
s
.91427 .15517 m
.91815 .16886 L
s
.87993 .37192 m
.89089 .3576 L
s
.87594 .37778 m
.87993 .37192 L
s
.91959 .1751 m
.91815 .16886 L
s
.92208 .18819 m
.91959 .1751 L
s
.89089 .3576 m
.89349 .3535 L
s
.92208 .18819 m
.92396 .2054 L
s
.89349 .3535 m
.90285 .33388 L
s
.92396 .2054 m
.92443 .21868 L
s
.90749 .32518 m
.90285 .33388 L
s
.92443 .21868 m
.9247 .23608 L
s
.91656 .29344 m
.90749 .32518 L
s
.92298 .25723 m
.9247 .23608 L
s
.91744 .2905 m
.91656 .29344 L
s
.91755 .28946 m
.91933 .27888 L
s
.91755 .28946 m
.91744 .2905 L
s
.92278 .25928 m
.92298 .25723 L
s
.91933 .27888 m
.92278 .25928 L
s
.59982 .19255 m
.60627 .17424 L
s
.59845 .19949 m
.59587 .21871 L
s
.61783 .15289 m
.60922 .16734 L
s
.59845 .19949 m
.59982 .19255 L
s
.60922 .16734 m
.60627 .17424 L
s
.59606 .22805 m
.59718 .24382 L
s
.59606 .22805 m
.59587 .21871 L
s
.63165 .1354 m
.62323 .1449 L
s
.61783 .15289 m
.62323 .1449 L
s
.60083 .25919 m
.60289 .26668 L
s
.60083 .25919 m
.59718 .24382 L
s
.6404 .12634 m
.64753 .12066 L
s
.63165 .1354 m
.6404 .12634 L
s
.60289 .26668 m
.60837 .27853 L
s
.66593 .10775 m
.66166 .11015 L
s
.66166 .11015 m
.64753 .12066 L
s
.67811 .10166 m
.66593 .10775 L
s
.61175 .28754 m
.60837 .27853 L
s
.61448 .29229 m
.62305 .30694 L
s
.61175 .28754 m
.61448 .29229 L
s
.68569 .09737 m
.67811 .10166 L
s
.62305 .30694 m
.63503 .32191 L
s
.70698 .08928 m
.68786 .09634 L
s
.68569 .09737 m
.68786 .09634 L
s
.63944 .32749 m
.65265 .34234 L
s
.63503 .32191 m
.63721 .32507 L
s
.71641 .08636 m
.70698 .08928 L
s
.72832 .08405 m
.71641 .08636 L
s
.63944 .32749 m
.63721 .32507 L
s
.65265 .34234 m
.66992 .35835 L
s
.72832 .08405 m
.73714 .08238 L
s
.66992 .35835 m
.67131 .3597 L
s
.73714 .08238 m
.74478 .08132 L
s
.76287 .08008 m
.76187 .08015 L
s
.76187 .08015 m
.74478 .08132 L
s
.67217 .36034 m
.67131 .3597 L
s
.69472 .37736 m
.67217 .36034 L
s
.76287 .08008 m
.76604 .08017 L
s
.69603 .37808 m
.69472 .37736 L
s
.71819 .39017 m
.69603 .37808 L
s
.76604 .08017 m
.78471 .08026 L
s
.72148 .39129 m
.71819 .39017 L
s
.78864 .08062 m
.78471 .08026 L
s
.80918 .0834 m
.78864 .08062 L
s
.74232 .39846 m
.72148 .39129 L
s
.80918 .0834 m
.81257 .08415 L
s
.75093 .39959 m
.74232 .39846 L
s
.83327 .09034 m
.81257 .08415 L
s
.75093 .39959 m
.76657 .40193 L
s
.8349 .09096 m
.83327 .09034 L
s
.85544 .1013 m
.85299 .10012 L
s
.76657 .40193 m
.78633 .40042 L
s
.8349 .09096 m
.85299 .10012 L
s
.85577 .10152 m
.85544 .1013 L
s
.78633 .40042 m
.79031 .40025 L
s
.87257 .11395 m
.85577 .10152 L
s
.79031 .40025 m
.79515 .39882 L
s
.81331 .39494 m
.79515 .39882 L
s
.87688 .1189 m
.87257 .11395 L
s
.83174 .38569 m
.81331 .39494 L
s
.87688 .1189 m
.88505 .12777 L
s
.83174 .38569 m
.83413 .38464 L
s
.88505 .12777 m
.895 .14458 L
s
.83524 .38379 m
.83413 .38464 L
s
.895 .14458 m
.89524 .14493 L
s
.85301 .37176 m
.83524 .38379 L
s
.89533 .14511 m
.89524 .14493 L
s
.85301 .37176 m
.85687 .36817 L
s
.89807 .15169 m
.90406 .16593 L
s
.89533 .14511 m
.89807 .15169 L
s
.87215 .35179 m
.85687 .36817 L
s
.90406 .16593 m
.9042 .16641 L
s
.87215 .35179 m
.87283 .35089 L
s
.90487 .16929 m
.9042 .16641 L
s
.9091 .18597 m
.90487 .16929 L
s
.87422 .3487 m
.88505 .33355 L
s
.87283 .35089 m
.87422 .3487 L
s
.90962 .18853 m
.9091 .18597 L
s
.89037 .32316 m
.88505 .33355 L
s
.90962 .18853 m
.91214 .20963 L
s
.89516 .31416 m
.89037 .32316 L
s
.91246 .21358 m
.91214 .20963 L
s
.91231 .23547 m
.91235 .23807 L
s
.89516 .31416 m
.89745 .30616 L
s
.91246 .21358 m
.91231 .23547 L
s
.89745 .30616 m
.90488 .28712 L
s
.91228 .23888 m
.91235 .23807 L
s
.90805 .273 m
.90488 .28712 L
s
.91017 .26101 m
.91228 .23888 L
s
.90805 .273 m
.91017 .26101 L
s
.61888 .21038 m
.62248 .19268 L
s
.63038 .17157 m
.62398 .18711 L
s
.62398 .18711 m
.62248 .19268 L
s
.61848 .21725 m
.6183 .23389 L
s
.64092 .15367 m
.63334 .16547 L
s
.61848 .21725 m
.61888 .21038 L
s
.63038 .17157 m
.63334 .16547 L
s
.61997 .24505 m
.62193 .25598 L
s
.6183 .23389 m
.61997 .24505 L
s
.65271 .13956 m
.64571 .14705 L
s
.62872 .27497 m
.62918 .2762 L
s
.64092 .15367 m
.64571 .14705 L
s
.62872 .27497 m
.62193 .25598 L
s
.66958 .12418 m
.65861 .13281 L
s
.65861 .13281 m
.65271 .13956 L
s
.62918 .2762 m
.63019 .27794 L
s
.68808 .11199 m
.67927 .11698 L
s
.70687 .10341 m
.70392 .1045 L
s
.67927 .11698 m
.66958 .12418 L
s
.68808 .11199 m
.70392 .1045 L
s
.63019 .27794 m
.63867 .29541 L
s
.64555 .30527 m
.65109 .31329 L
s
.71713 .10023 m
.70687 .10341 L
s
.64555 .30527 m
.63867 .29541 L
s
.65897 .32186 m
.65109 .31329 L
s
.71713 .10023 m
.72533 .09738 L
s
.66583 .33015 m
.65897 .32186 L
s
.67291 .33642 m
.68038 .34369 L
s
.67291 .33642 m
.66583 .33015 L
s
.72696 .09693 m
.72533 .09738 L
s
.74284 .09358 m
.72696 .09693 L
s
.69457 .35422 m
.68038 .34369 L
s
.69792 .35675 m
.69457 .35422 L
s
.74284 .09358 m
.74769 .0929 L
s
.75987 .09165 m
.74769 .0929 L
s
.71773 .36764 m
.69792 .35675 L
s
.71892 .36829 m
.71773 .36764 L
s
.77088 .0914 m
.77916 .09144 L
s
.77088 .0914 m
.75987 .09165 L
s
.72525 .37042 m
.71892 .36829 L
s
.79373 .09277 m
.77916 .09144 L
s
.79963 .09357 m
.79373 .09277 L
s
.74111 .37628 m
.72525 .37042 L
s
.79963 .09357 m
.81434 .09682 L
s
.82069 .09872 m
.81434 .09682 L
s
.74422 .3768 m
.74111 .37628 L
s
.76369 .38025 m
.74422 .3768 L
s
.83424 .10388 m
.82069 .09872 L
s
.83424 .10388 m
.841 .1073 L
s
.76369 .38025 m
.77529 .37981 L
s
.77529 .37981 m
.7859 .37967 L
s
.85179 .11369 m
.841 .1073 L
s
.7859 .37967 m
.80085 .37587 L
s
.85909 .11956 m
.85179 .11369 L
s
.80758 .37469 m
.80085 .37587 L
s
.85909 .11956 m
.86614 .12567 L
s
.81312 .37216 m
.80758 .37469 L
s
.87274 .13331 m
.86614 .12567 L
s
.82949 .36557 m
.81312 .37216 L
s
.87697 .13869 m
.87274 .13331 L
s
.83769 .35966 m
.82949 .36557 L
s
.88221 .14644 m
.87697 .13869 L
s
.83769 .35966 m
.85131 .34953 L
s
.8878 .15687 m
.88221 .14644 L
s
.85131 .34953 m
.85726 .34356 L
s
.89077 .16402 m
.8878 .15687 L
s
.85726 .34356 m
.87069 .32728 L
s
.89077 .16402 m
.89492 .17586 L
s
.89659 .18243 m
.89492 .17586 L
s
.87069 .32728 m
.87149 .32617 L
s
.89953 .19698 m
.89659 .18243 L
s
.87236 .32445 m
.87149 .32617 L
s
.88336 .30686 m
.87236 .32445 L
s
.89953 .19698 m
.90029 .20331 L
s
.88968 .29097 m
.88336 .30686 L
s
.90154 .2185 m
.90029 .20331 L
s
.90148 .22749 m
.90154 .2185 L
s
.88968 .29097 m
.89237 .28408 L
s
.90148 .22749 m
.90108 .23974 L
s
.89237 .28408 m
.89452 .2745 L
s
.89812 .26175 m
.89452 .2745 L
s
.89897 .25596 m
.90108 .23974 L
s
.89812 .26175 m
.89897 .25596 L
s
.63873 .20366 m
.64298 .18743 L
s
.63815 .20714 m
.63673 .22612 L
s
.65081 .16939 m
.64415 .18367 L
s
.63873 .20366 m
.63815 .20714 L
s
.64415 .18367 m
.64298 .18743 L
s
.63722 .23189 m
.63877 .24721 L
s
.66008 .15459 m
.65364 .16424 L
s
.63673 .22612 m
.63722 .23189 L
s
.65081 .16939 m
.65364 .16424 L
s
.64207 .25881 m
.64433 .26655 L
s
.67275 .14092 m
.66008 .15459 L
s
.64207 .25881 m
.63877 .24721 L
s
.64433 .26655 m
.65115 .28069 L
s
.67275 .14092 m
.67327 .14024 L
s
.67357 .13995 m
.69085 .12591 L
s
.70779 .11632 m
.69456 .12344 L
s
.67327 .14024 m
.67357 .13995 L
s
.65115 .28069 m
.65278 .28461 L
s
.69085 .12591 m
.69456 .12344 L
s
.72457 .10966 m
.71644 .11248 L
s
.66333 .30181 m
.65381 .28627 L
s
.6732 .31315 m
.67807 .31902 L
s
.71644 .11248 m
.70779 .11632 L
s
.65381 .28627 m
.65278 .28461 L
s
.6732 .31315 m
.66333 .30181 L
s
.67807 .31902 m
.68251 .32296 L
s
.73715 .10618 m
.72457 .10966 L
s
.74063 .10544 m
.73715 .10618 L
s
.68251 .32296 m
.69453 .33359 L
s
.75478 .10346 m
.74063 .10544 L
s
.69453 .33359 m
.70085 .33844 L
s
.75679 .10313 m
.75478 .10346 L
s
.70085 .33844 m
.71373 .34613 L
s
.72006 .34973 m
.71373 .34613 L
s
.75718 .10309 m
.75679 .10313 L
s
.77407 .10223 m
.75718 .10309 L
s
.73723 .3562 m
.72006 .34973 L
s
.73723 .3562 m
.7405 .35741 L
s
.792 .10336 m
.7788 .10236 L
s
.7788 .10236 m
.77407 .10223 L
s
.7405 .35741 m
.751 .35915 L
s
.79692 .10419 m
.792 .10336 L
s
.80981 .10708 m
.79692 .10419 L
s
.76145 .36138 m
.751 .35915 L
s
.81507 .10853 m
.80981 .10708 L
s
.82796 .1135 m
.81507 .10853 L
s
.76145 .36138 m
.76486 .36138 L
s
.78221 .36176 m
.76486 .36138 L
s
.8337 .11608 m
.82796 .1135 L
s
.84431 .12254 m
.8337 .11608 L
s
.79647 .35891 m
.78221 .36176 L
s
.80195 .35795 m
.79647 .35891 L
s
.84787 .12502 m
.84431 .12254 L
s
.80925 .35461 m
.80195 .35795 L
s
.85817 .13427 m
.84787 .12502 L
s
.82271 .34974 m
.80925 .35461 L
s
.86048 .13651 m
.85817 .13427 L
s
.86932 .14753 m
.86048 .13651 L
s
.82271 .34974 m
.8286 .34631 L
s
.8286 .34631 m
.84318 .33588 L
s
.87222 .15171 m
.86932 .14753 L
s
.87821 .16247 m
.87222 .15171 L
s
.85076 .32899 m
.84318 .33588 L
s
.88016 .16683 m
.87821 .16247 L
s
.85857 .3204 m
.85076 .32899 L
s
.88447 .17907 m
.88016 .16683 L
s
.86912 .30526 m
.85857 .3204 L
s
.88601 .18418 m
.88447 .17907 L
s
.86912 .30526 m
.87176 .30103 L
s
.88881 .19772 m
.88601 .18418 L
s
.8741 .29515 m
.87176 .30103 L
s
.88992 .20442 m
.88881 .19772 L
s
.8741 .29515 m
.88135 .28037 L
s
.89091 .21803 m
.88992 .20442 L
s
.88135 .28037 m
.88475 .26875 L
s
.89105 .22208 m
.89091 .21803 L
s
.88695 .26097 m
.88475 .26875 L
s
.88998 .24063 m
.89105 .22208 L
s
.88695 .26097 m
.8895 .2437 L
s
.8895 .2437 m
.88993 .24129 L
s
.88993 .24129 m
.88998 .24063 L
s
.65308 .21755 m
.6529 .21991 L
s
.65421 .21074 m
.65308 .21755 L
s
.6529 .21991 m
.65347 .22664 L
s
.65421 .21074 m
.6555 .19885 L
s
.6555 .19885 m
.66076 .18321 L
s
.65389 .24018 m
.65347 .22664 L
s
.66928 .16615 m
.66115 .18211 L
s
.66076 .18321 m
.66115 .18211 L
s
.68009 .15156 m
.67156 .16271 L
s
.68887 .14302 m
.69424 .13866 L
s
.65492 .24471 m
.65801 .25809 L
s
.70894 .12897 m
.70786 .12955 L
s
.67156 .16271 m
.66928 .16615 L
s
.65492 .24471 m
.65389 .24018 L
s
.68009 .15156 m
.68887 .14302 L
s
.66281 .26954 m
.66509 .27505 L
s
.69424 .13866 m
.70786 .12955 L
s
.66281 .26954 m
.65801 .25809 L
s
.7121 .12757 m
.70894 .12897 L
s
.67106 .28471 m
.66509 .27505 L
s
.7121 .12757 m
.72417 .12148 L
s
.67427 .29055 m
.67106 .28471 L
s
.67591 .29271 m
.68657 .30617 L
s
.73915 .11644 m
.72802 .11999 L
s
.67591 .29271 m
.67427 .29055 L
s
.72802 .11999 m
.72417 .12148 L
s
.68657 .30617 m
.69288 .31239 L
s
.69288 .31239 m
.70366 .32166 L
s
.74547 .11522 m
.73915 .11644 L
s
.7538 .11385 m
.74547 .11522 L
s
.71057 .32646 m
.70366 .32166 L
s
.71057 .32646 m
.72128 .33304 L
s
.76381 .11277 m
.7538 .11385 L
s
.76965 .11248 m
.76381 .11277 L
s
.72128 .33304 m
.73069 .3372 L
s
.7854 .1129 m
.76965 .11248 L
s
.74022 .34109 m
.73069 .3372 L
s
.79882 .11511 m
.78661 .11306 L
s
.80028 .11543 m
.79882 .11511 L
s
.75472 .34423 m
.74022 .34109 L
s
.75472 .34423 m
.76004 .34537 L
s
.81495 .11948 m
.80028 .11543 L
s
.81642 .12004 m
.81495 .11948 L
s
.76004 .34537 m
.77076 .34535 L
s
.77076 .34535 m
.77946 .34592 L
s
.82363 .12328 m
.81642 .12004 L
s
.78432 .34506 m
.77946 .34592 L
s
.79622 .34351 m
.78432 .34506 L
s
.84357 .13494 m
.83886 .13166 L
s
.84459 .13585 m
.84357 .13494 L
s
.80933 .33901 m
.79622 .34351 L
s
.80933 .33901 m
.81302 .33767 L
s
.8555 .14645 m
.84459 .13585 L
s
.85669 .14793 m
.8555 .14645 L
s
.81302 .33767 m
.82354 .33155 L
s
.83048 .32794 m
.82354 .33155 L
s
.86174 .15523 m
.85669 .14793 L
s
.83185 .32697 m
.83048 .32794 L
s
.84616 .31492 m
.83185 .32697 L
s
.8724 .17495 m
.87046 .1706 L
s
.84616 .31492 m
.85004 .31065 L
s
.87273 .17588 m
.8724 .17495 L
s
.85004 .31065 m
.85896 .29882 L
s
.87726 .191 m
.87273 .17588 L
s
.87761 .19268 m
.87726 .191 L
s
.86351 .29203 m
.85896 .29882 L
s
.88009 .20758 m
.87761 .19268 L
s
.87085 .27649 m
.86351 .29203 L
s
.87091 .27639 m
.87085 .27649 L
s
.88112 .22396 m
.88069 .21194 L
s
.87091 .27639 m
.87094 .2763 L
s
.88083 .22904 m
.88112 .22396 L
s
.87094 .2763 m
.87646 .26037 L
s
.87808 .25148 m
.87992 .24113 L
s
.87992 .24113 m
.88083 .22904 L
s
.87646 .26037 m
.87808 .25148 L
s
.66803 .20641 m
.66862 .201 L
s
.66862 .201 m
.6699 .1972 L
s
.6773 .17752 m
.67535 .18134 L
s
.6699 .1972 m
.67535 .18134 L
s
.66616 .22105 m
.66803 .20641 L
s
.66616 .22105 m
.66735 .23434 L
s
.6848 .16622 m
.6773 .17752 L
s
.66735 .23434 m
.66751 .23594 L
s
.6848 .16622 m
.68684 .16271 L
s
.66751 .23594 m
.67029 .25018 L
s
.68752 .16182 m
.69814 .14995 L
s
.71089 .14022 m
.70171 .14695 L
s
.67635 .26741 m
.6718 .25458 L
s
.68684 .16271 m
.68752 .16182 L
s
.72444 .13243 m
.71799 .13569 L
s
.70171 .14695 m
.69814 .14995 L
s
.6718 .25458 m
.67029 .25018 L
s
.68354 .27963 m
.68309 .27882 L
s
.71799 .13569 m
.71089 .14022 L
s
.68309 .27882 m
.67635 .26741 L
s
.68654 .2836 m
.68354 .27963 L
s
.73625 .12786 m
.72444 .13243 L
s
.69302 .2925 m
.68654 .2836 L
s
.70659 .30635 m
.69386 .29347 L
s
.69386 .29347 m
.69302 .2925 L
s
.70852 .30792 m
.70659 .30635 L
s
.7226 .31789 m
.70852 .30792 L
s
.7226 .31789 m
.72569 .31953 L
s
.72569 .31953 m
.74003 .32623 L
s
.74003 .32623 m
.74573 .32791 L
s
.74573 .32791 m
.75881 .3313 L
s
.7707 .33194 m
.75881 .3313 L
s
.7707 .33194 m
.77856 .33245 L
s
.78538 .33125 m
.77856 .33245 L
s
.79578 .32968 m
.78538 .33125 L
s
.79578 .32968 m
.80375 .32733 L
s
.80375 .32733 m
.81477 .32266 L
s
.82047 .31969 m
.81477 .32266 L
s
.83337 .31063 m
.82047 .31969 L
s
.83447 .30971 m
.83337 .31063 L
s
.83447 .30971 m
.8376 .30626 L
s
.84604 .29773 m
.8376 .30626 L
s
.84604 .29773 m
.8483 .29484 L
s
.8483 .29484 m
.85552 .28433 L
s
.85552 .28433 m
.85861 .27874 L
s
.85861 .27874 m
.86224 .27197 L
s
.86224 .27197 m
.86542 .26282 L
s
.68524 .18721 m
.68086 .19998 L
s
.67866 .22163 m
.6796 .20523 L
s
.67997 .23376 m
.68066 .24086 L
s
.68086 .19998 m
.6796 .20523 L
s
.69292 .17231 m
.68789 .18098 L
s
.68066 .24086 m
.68092 .2422 L
s
.70085 .16186 m
.70246 .16006 L
s
.68789 .18098 m
.68524 .18721 L
s
.71292 .15133 m
.71255 .15161 L
s
.69292 .17231 m
.70085 .16186 L
s
.67864 .22364 m
.67866 .22163 L
s
.67997 .23376 m
.67864 .22364 L
s
.71255 .15161 m
.70246 .16006 L
s
.68438 .25229 m
.68092 .2422 L
s
.71436 .15041 m
.71292 .15133 L
s
.68695 .26038 m
.68438 .25229 L
s
.68972 .26521 m
.69614 .27687 L
s
.69614 .27687 m
.69792 .27932 L
s
.68723 .26105 m
.68695 .26038 L
s
.68972 .26521 m
.68723 .26105 L
s
.71029 .29333 m
.70779 .2908 L
s
.70779 .2908 m
.69792 .27932 L
s
.72226 .30307 m
.71029 .29333 L
s
.72409 .30437 m
.72226 .30307 L
s
.72409 .30437 m
.73913 .31237 L
s
.73913 .31237 m
.73987 .31271 L
s
.73987 .31271 m
.74575 .31444 L
s
.74575 .31444 m
.75769 .31823 L
s
.75769 .31823 m
.75887 .31842 L
s
.75887 .31842 m
.77654 .32029 L
s
.77654 .32029 m
.78076 .32011 L
s
.78076 .32011 m
.79462 .31778 L
s
.79462 .31778 m
.79992 .31629 L
s
.79992 .31629 m
.81175 .31164 L
s
.81817 .30823 m
.81175 .31164 L
s
.82552 .30354 m
.81817 .30823 L
s
.83335 .29689 m
.82552 .30354 L
s
.83335 .29689 m
.83422 .296 L
s
.83422 .296 m
.83848 .29057 L
s
.69314 .19613 m
.69268 .19746 L
s
.6895 .2118 m
.68958 .21045 L
s
.69268 .19746 m
.68958 .21045 L
s
.6962 .18893 m
.69314 .19613 L
s
.68937 .22762 m
.6895 .2118 L
s
.68937 .22762 m
.69024 .23431 L
s
.69255 .24426 m
.69487 .25155 L
s
.69024 .23431 m
.69255 .24426 L
s
.69853 .26029 m
.69487 .25155 L
s
.69853 .26029 m
.70177 .2657 L
s
.70177 .2657 m
.70825 .27503 L
s
.70825 .27503 m
.71404 .28174 L
s
.72659 .29294 m
.72043 .28781 L
s
.72043 .28781 m
.71404 .28174 L
s
.73394 .29783 m
.72659 .29294 L
s
.73394 .29783 m
.73859 .30038 L
s
.73859 .30038 m
.75044 .30496 L
s
.75044 .30496 m
.75691 .30702 L
s
.75691 .30702 m
.76874 .30884 L
s
.76874 .30884 m
.77349 .30935 L
s
.77349 .30935 m
.78507 .30884 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012PP10`4000010?ooo`000<c<c;no_`00
E@00ZW=cLiVIVNfXL<V>GoOI`7eI>ooooiVIVH:2P]7AdTi>C_ooohVHZWB1T<[AfDI=E_ooofQ_PUQN
K[ZmaS4e?_oooiBDU:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJF
UWemOFATI4];Bc8b<P0T000TE@0TZP0To`190019E@19ZP19o`1]001]E@1]ZP1]o`2B002BE@2BZP2B
o`2f002fE@2fZP2fo`3K003KE@3KZP3Ko`3oE@3oZUD005D0EED0ZUD0oeDT05DTEEDTZUDToeE905E9
EEE9ZUE9oeE]05E]EEE]ZUE]oeFB05FBEEFBZUFBoeFf05FfEEFfZUFfoeGK05GKEEGKZUGKoeGo05Go
EEGoZUGoojX00:X0EJX0ZZX0ojXT0:XTEJXTZZXTojY90:Y9EJY9ZZY9ojY]0:Y]EJY]ZZY]ojZB0:ZB
EJZBZZZBojZf0:ZfEJZfZZZfoj[K0:[KEJ[KZZ[Koj[o0:[oEJ[oZZ[oool0EOl0Z_lT0?lTEOlTZ_lT
oom90?m9EOm9Z_m9oom]0?m]EOm]Z_m]oonB0?nBEOnBZ_nBoonf0?nfEOnfZ_nfoooK0?oKEOoKZ_oK
ooooEOooZ/d0003=0<g=0000cLd0c@3=cNGUiDe=C@00:P00O`00e00Z000Z:P0ZE@0ZO`0ZZP0Ze01E
001E:P1EE@1EO`1EZP1Ee01o001o:P1oE@1oO`1oZP1oe02Z002Z:P2ZE@2ZO`2ZZP2Ze03D003D:P3D
E@3DO`3DZP3De2X002X0:RX0EBX0ObX0ZRX0e2XZ02XZ:RXZEBXZObXZZRXZe2YE02YE:RYEEBYEObYE
ZRYEe2Yo02Yo:RYoEBYoObYoZRYoe2ZZ02ZZ:RZZEBZZObZZZRZZe2[D02[D:R[DEB[DOb[DZR[De5D0
:UD0OeD0e5DZ05DZ:UDZEEDZOeDZZUDZe5EE05EE:UEEEOl0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@0000X0o`6L0@/0000:00030@000?l0V0000`40000900002P000`40003o09P000<10000
2@0000X000<10000o`2H00030@0000T0000:00030@000?l0V0000`40000900002P000`40003o09P0
00<100002@0000X000<10000o`0U0381@@000`40000900002P000`40003o01D0404b00d1=0000`40
000900002P000`40003o00T0305?00@1<0000`40000900002P000`40003o00D0105O00@1;0000`40
000900002P000`40003o0040105W00L19@000`40000900002P000`40003h00P1LP040B4000<10000
2@0000X000<10000m0040C00;@4Q00<17P000`40000900002P000`40003`00@1;P060Bd01P4N00<1
6`000`40000900002P000`40003Z00H1:@090CT0104M00816@000`40000900002P000`40003V00@1
:0070DH0104K00030@0001H000<100002@0000X000<10000hP040BP0105A00H15P000`40000E0003
0@0000T0000:00030@000=h0104Z00819`0@0B@00`4D00030@0001@000<100002@0000X000<10000
g0020B/00`4O00X1400=0AX00P4C00030@0001<000<100002@0000X000<10000fP020BL01P4L00H1
9`050AL00P4A00030@0001<000<100002@0000X000<10000f0020BD0104K00L1<P050A@000<10000
3`000`40000B00030@0000T0000:00030@000=D00`4U00816`040Ch00`4B008140000`40000A0003
0@0000T0000:00030@000=40104V008160050DD00P4B00813`000`40000@00030@0000T0000:0003
0@000<h00`4W00<15`030A`06`4E00<14@000`40000=00030@0000l000<100002@0000X000<10000
c0020BL00`4H00816@060A/01`4A00<13`000`40000=00030@0000h000<100002@0000X000<10000
bP020BL00P4H00<15`040BP0104@00813P020@h00P4>00030@0000T0000:00030@000<P00P4W0081
5P040A<01`4`00@13P020@h000<1000030000`40000<00030@0000T0000:00030@000<H00P4V00<1
5P020A@00`4k00813P020@d000<100002`000`40000<00030@0000T0000:00030@000<@00P4V0081
5`020A@00P4J018150020@h000<100002`000`40000;00030@0000/000<100002@0000X000<10000
``000`40000T008160000`40000A00<14`090A80104B00813@020@`000<100002P000`40000;0003
0@0000T0000:00030@000<<000<100008P020AP00P4@00@14P040Al01P4>00813@020@/000<10000
2P000`40000:00030@0000T0000:00030@000<8000<100008@020AP00P4@00814@050BT0104<0081
3@000`40000900030@0000T000<100002P000`40000900002P000`40003000818P020AP00P4A0003
0@0000h00`4b00<12`020@`000<100002@000`40000900030@0000T000<100002@0000X000<10000
_`000`40000Q00030@0001H00P4B00030@0000`00`4F00d15@000`40000:00030@0000X000<10000
2@000`40000800030@0000T000<100002@0000X000<10000_@020B<000<100005@020A800P4=0081
40090@d02@4=00812`020@X000<100002@000`40000800030@0000T000<100002@0000X000<10000
^`020B@000<100005@000`40000@00813@020@l00`4O00D12P020@/00P4900030@0000T000<10000
20000`40000800030@0000T0000:00030@000;X000<100008`000`40000D00814@020@d00P4>00<1
9`030@T00P4;00030@0000L000<1000020000`40000800030@0000P000<100002@0000X000<10000
^0020BD000<1000050000`40000>00<130030@`0104]00<120030@T00P4800030@0000P000<10000
20000`40000700030@0000T0000:00030@000;L000<1000090000`40000C008140000`40000;0081
3@020A005`4=00812@000`40000800030@0000H00P4800030@0000P000<100001`000`4000090000
2P000`40002f00030@0002@000<100004`000`40000>008130020@d00P4<00H15`020@d00P480003
0@0000L000<1000020000`40000600030@0000P000<100001P000`40000900002P000`40002e0003
0@0002@000<100004`000`40000>00030@0000/000<100002`020@`00P4O00H12@000`4000060081
20000`40000700030@0000H000<1000020000`40000600030@0000T0000:00030@000;@000<10000
8`020A@00P4@00030@0000/000<100002P020@X0104W008120000`40000700030@0000H000<10000
1`000`40000600030@0000L000<100001P000`40000900002P000`40002b008190020AD000<10000
3P020@`00P4:00<12@030A003`4>00811`020@P000<100001P000`40000600030@0000H000<10000
1`000`40000600030@0000T0000:00030@000;4000<100008`000`40000D00030@0000h000<10000
2`000`40000800812P020@d01P4?00D12`020@L000<100001P000`40000600030@0000H000<10000
1P000`40000600030@0000H000<100002@0000X000<10000/0000`40000S00030@0001@000<10000
3P000`40000;00030@0000P000<100002@000`40000900@16P030@X000<100001@020@L000<10000
1P000`40000500030@0000H000<100001P000`40000600030@0000T0000:00030@000:l000<10000
8`000`40000D00030@0000h000<100002P020@T00P4:008120040B400`4800811`000`4000050003
0@0000H000<100001@000`40000600030@0000D000<100001P000`40000900002P000`40002^0003
0@0002<000<1000050000`40000>00030@0000X000<1000020000`40000800812@000`40000V0081
20020@H000<100001@000`40000500030@0000D000<100001P000`40000500030@0000H000<10000
2@0000X000<10000[@000`40000T00030@0001<000<100003P000`40000900812P000`4000070081
2@020B/000<100001`020@D00P4600030@0000D000<100001@000`40000500030@0000D000<10000
1P000`40000900002P000`40002]00030@0002<000<100004`000`40000=00812`000`4000090003
0@0000L000<100001`020CT000<1000010000`40000400030@0000@000<100001@000`4000060003
0@0000D000<100001@000`40000900002P000`40002/00030@0002<000<100004`000`40000=0003
0@0000X000<1000020020@T000<100001@030@T00P4a00030@0000@000<1000010000`4000040003
0@0000D000<100001@000`40000500030@0000D000<100002@0000X000<10000Z`000`40000S0003
0@0001@000<1000030000`40000;00030@0000L000<1000020000`40000500030@0000P00P4]0003
0@0000@000<100000`000`40000500030@0000<000<100001@000`40000500030@0000D000<10000
1@000`40000900002P000`40002Z00030@0002<000<1000050000`40000<00030@0000/000<10000
1`000`40000700811P020@P00`4`00030@0000@000<100000`000`40000400030@0000@000<10000
1@000`40000400030@0000D000<100001@000`40000900002P000`40002Z00030@00028000<10000
50000`40000=00030@0000X000<100001`000`40000700030@0000D000<100001P020C@000<10000
10000`40000300030@0000@000<100000`000`40000500030@0000@000<100001@000`4000050003
0@0000T0000:00030@000:T000<100008`000`40000C00030@0000d000<100002P000`4000080003
0@0000H000<100001@000`4000050081=`000`40000400030@0000<000<100000`000`4000040003
0@0000@000<1000010000`40000600030@0000@000<100002@0000X000<10000Z@000`40000R0003
0@0001@000<1000030000`40000:00030@0000P000<100001@020@H00P4700030@0003P000<10000
10000`40000200030@0000@000<100000`000`40000500030@0000@000<100001@000`4000040003
0@0000T0000:00030@000:P000<100008P000`40000D00030@0000d000<100002@000`4000080003
0@0000D000<100001@000`4000050081?0000`40000300030@0000<000<100000`000`4000040003
0@0000@000<1000010000`40000500030@0000@000<100002@0000X000<10000Z0000`40000R0003
0@0001<000<100003@000`40000900030@0000P000<100001@000`40000500030@0000D000<10000
?@000`40000300030@00008000<1000010000`40000300030@0000@000<1000010000`4000050003
0@0000@000<100002@0000X000<10000Y`000`40000R00030@0001<000<100003@000`4000090003
0@0000P000<100001P000`40000400030@0000@00P5100030@0000<000<100000P000`4000030003
0@0000@000<100000`000`40000400030@0000D000<1000010000`40000900002P000`40002W0003
0@00024000<100004`000`40000>00030@0000T000<100001`000`40000600030@0000@000<10000
10000`40000400030@0003/000<100000P000`40000300030@0000<000<100000`000`4000040003
0@0000@000<1000010000`40000400030@0000T0000:00030@000:H000<100008P000`40000C0003
0@0000d000<100002@000`40000800030@0000D000<1000010000`40000400030@0000@00P4n0003
0@00008000<100000P000`40000300030@0000@000<100000`000`40000400030@0000@000<10000
10000`40000900002P000`40002V00030@00024000<100004`000`40000=00030@0000T000<10000
20000`40000500030@0000@000<1000010000`4000030081@@000`40000200030@00008000<10000
0P000`40000400030@0000<000<1000010000`40000400030@0000@000<100002@0000X000<10000
Y@000`40000R00030@0001<000<100003@000`40000900030@0000L000<100001@000`4000050003
0@0000@000<100000P000`40001600030@0000<000<100000P000`40000300030@0000<000<10000
10000`40000400030@0000@000<100002@0000X000<10000Y0000`40000R00030@0001<000<10000
3@000`40000900030@0000L000<100001P000`40000400030@0000@000<100000P000`4000180003
0@00008000<100000P000`40000400030@0000<000<100000`000`40000400030@0000@000<10000
2@0000X000<10000Y0000`40000R00030@0001<000<100003@000`40000900030@0000L000<10000
1@000`40000500030@0000<000<100000P000`40001900030@0000<000<100000P000`4000030003
0@0000<000<1000010000`40000300030@0000@000<100002@0000X000<10000Y0000`40000Q0003
0@0001<000<100003@000`40000900030@0000L000<100001@000`40000500030@0000<000<10000
0P000`40001;00030@00008000<100000P000`40000300030@0000<000<1000010000`4000030003
0@0000@000<100002@0000X000<10000X`000`40000R00030@0001<000<1000030000`40000:0003
0@0000H000<100001P000`40000400030@0000@000<100000P000`40001<00050@0000401@000`40
000300030@00008000<1000010000`40000300030@0000@000<100002@0000X000<10000X`000`40
000R00030@00018000<100003@000`40000900030@0000L000<100001@000`40000400030@0000@0
00<100000P000`40001=00030@00008000<100000P000`40000300030@0000<000<100000`000`40
000300030@0000@000<100002@0000X000<10000X`000`40000Q00030@0001<000<1000030000`40
000:00030@0000H000<100001P000`40000400030@0000<000<100000`000`40001800030@000080
00<100000P000`40000200030@0000<000<100000`000`40000300030@0000<000<1000010000`40
000900002P000`40002S00030@00024000<100004P000`40000=00030@0000T000<100001`000`40
000500030@0000@000<1000010000`40000200030@0004T000<100000P000`40000300050@000040
1P000`40000200030@0000<000<1000010000`40000300030@0000T0000:00030@000:8000<10000
8P000`40000B00030@0000`000<100002P000`40000600030@0000H000<1000010000`4000030003
0@00008000<10000B`000`40000200030@00008000<100000P000`40000300030@00008000<10000
0`000`40000400030@0000<000<100002@0000X000<10000XP000`40000Q00030@00018000<10000
3@000`40000:00030@0000H000<100001@000`40000400030@0000<000<100000`000`40001;0003
0@00008000<100000P000`40000200030@0000<000<100000`000`40000200030@0000@000<10000
0`000`40000900002P000`40002R00030@00024000<100004P000`40000=00030@0000T000<10000
1P000`40000600030@0000@000<100000`000`40000200030@0004d000D100000@0500050@000040
1@000`40000300030@0000<000<100000`000`40000300030@0000T0000:00030@000:8000<10000
8@000`40000B00030@0000`000<100002P000`40000600030@0000D000<1000010000`4000030003
0@0000<000<10000C@001@40000100D000D100000@0500030@0000<000<100000`000`4000030003
0@0000<000<100002@0000X000<10000X@000`40000Q00030@0001<000<1000030000`4000090003
0@0000H000<100001P000`40000400030@0000<000<100000P000`40001>00030@00008000<10000
0P000`40000200030@0000<000<100000P000`40000300030@0000<000<100000`000`4000090000
2P000`40002Q00030@00024000<100004`000`40000<00030@0000T000<100001P000`4000060003
0@0000<000<100000`000`40000300030@0004h000<100000P000`40000200030@00008000<10000
0`000`40000200030@0000<000<100000`000`40000300030@0000T0000:00030@000:4000<10000
8@000`40000C00030@0000`000<1000020000`40000700030@0000H000<100000`000`4000030003
0@0000<000D100000@1=00050@00004010000`40000200030@0000<000<100000P000`4000030003
0@0000<000<100000`000`40000900002P000`40002P00030@00028000<100004P000`40000<0003
0@0000T000<100001`000`40000500030@0000@000<100000`000`40000200030@00008000<10000
B`001@40000100D000D100000@0500030@00008000<100000`000`40000300030@0000<000<10000
2@0000X000<10000X0000`40000R00030@00018000<1000030000`40000900030@0000L000<10000
1@000`40000400030@00008000<100000`001@40000104h000<100000P000`40000200050@000040
1@000`40000200030@0000<000<100000`000`40000300030@0000T0000:00030@000:0000<10000
8@000`40000C00030@0000`000<1000020000`40000800030@0000D000<100000`000`4000030003
0@0000<000D100000@1>00030@00008000<100000P001@40000100D000<100000P000`4000030003
0@0000<000<100000`000`40000900002P000`40002P00030@00024000<100004`000`40000;0003
0@0000T000<100001`000`40000600030@0000<000<100000`000`40000200030@00008000<10000
C@001@40000100@000D100000@0500030@00008000<100000`000`40000300030@0000<000<10000
2@0000X000<10000X0000`40000Q00030@0001<000<100002`000`40000900030@0000L000<10000
1@000`40000400030@0000<000<100000P000`40000200030@0004d000D100000@0400030@000080
00<100000P000`40000200030@0000<000<100000`000`40000300030@0000T0000:00030@000:00
00<100008@000`40000B00030@0000`000<100002@000`40000700030@0000D000<1000010000`40
000300030@00008000<100000P000`40001=00030@00008000D100000@0400030@00008000<10000
0P000`40000300030@0000<000<100000`000`40000900002P000`40002P00030@00024000<10000
4P000`40000<00030@0000T000<100001`000`40000500030@0000<000<100000`000`4000020003
0@00008000<10000C`001@40000100<000<100000P000`40000200030@00008000<100000`000`40
000300030@0000<000<100002@0000X000<10000W`000`40000R00030@00018000<1000030000`40
000900030@0000H000<100001@000`40000400030@0000<000<100000P000`40000200030@0004l0
00D100000@0300030@00008000<100000P000`40000200030@0000<000<100000`000`4000030003
0@0000T0000:00030@0009l000<100008@000`40000C00030@0000`000<1000020000`4000070003
0@0000D000<1000010000`40000300030@00008000<100000P000`40001C00050@00004010000`40
000200030@00008000<100000`000`40000300030@0000<000<100002@0000X000<10000W`000`40
000Q00030@00018000<100003@000`40000800030@0000L000<100001@000`40000400030@0000<0
00<100000P000`40000200030@0004l000D100000@0400050@00004010000`40000200030@0000<0
00<100000`000`40000300030@0000T0000:00030@0009l000<100008@000`40000B00030@0000d0
00<1000020000`40000700030@0000D000<1000010000`40000200030@0000<000<100000P000`40
001?00050@00004010001@40000100D000<100000P000`40000300030@00008000<100000`000`40
000900002P000`40002O00030@00024000<100004P000`40000=00030@0000P000<100001`000`40
000500030@0000@000<100000P000`40000300030@00008000<10000C`001@40000100@000D10000
0@0500030@00008000<100000`000`40000200030@0000<000<100002@0000X000<10000W`000`40
000Q00030@00018000<100003@000`40000800030@0000L000<100001@000`40000400030@000080
00<100000`000`40000200030@0004l000D100000@0400050@0000401@000`40000200030@0000<0
00<100000P000`40000300030@0000T0000:00030@0009l000<100008@000`40000B00030@0000d0
00<1000020000`40000700030@0000D000<1000010000`40000200030@0000<000D100000@1B0005
0@00004010001@40000100D000<100000P000`40000300030@00008000<100000`000`4000090000
2P000`40002O00030@00024000<100004P000`40000=00030@0000P000<100001`000`4000050003
0@0000@000<100000P000`40000300050@000040DP001@40000100@000D100000@0500030@000080
00<100000`000`40000200030@0000<000<100002@0000X000<10000W`000`40000Q00030@000180
00<100003@000`40000800030@0000L000<100001@000`40000400030@00008000<100000`001@40
0001058000D100000@0400050@0000401@000`40000200030@0000<000<100000P000`4000030003
0@0000T0000:00030@0009l000<100008@000`40000C00030@0000`000<1000020000`4000070003
0@0000D000<1000010000`40000300030@00008000D100000@1B00050@00004010001@40000100D0
00<100000P000`40000300030@00008000<100000`000`40000900002P000`40002O00030@000240
00<100004`000`40000<00030@0000P000<100001`000`40000500030@0000@000<100000`000`40
000200050@000040DP001@40000100@000D100000@0500030@00008000<100000`000`4000020003
0@0000<000<100002@0000X000<10000X0000`40000P00030@0001<000<1000030000`4000080003
0@0000L000<100001@000`40000400030@0000<000<100000P000`40000200030@0004l000D10000
0@0400050@0000401@000`40000200030@0000<000<100000P000`40000300030@0000T0000:0003
0@000:0000<100008@000`40000B00030@0000`000<1000020000`40000700030@0000D000<10000
10000`40000300030@00008000<100000P000`40001>00030@00008000<100000P001@40000100D0
00<100000P000`40000300030@00008000<100000`000`40000900002P000`40002P00030@000240
00<100004P000`40000<00030@0000T000<100001P000`40000500030@0000@000<100000`000`40
000300050@000040D0000`40000200030@00008000D100000@0500030@00008000<100000`000`40
000200030@0000<000<100002@0000X000<10000X0000`40000Q00030@00018000<1000030000`40
000900030@0000L000<100001@000`40000300030@0000<000<100000`000`40000200030@0004d0
00<100000P000`40000200050@0000401@000`40000200030@0000<000<100000P000`4000030003
0@0000T0000:00030@000:0000<100008@000`40000C00030@0000/000<100002@000`4000070003
0@0000D000<100000`000`40000300030@0000<000<100000P000`40001=00030@00008000<10000
0P001@40000100D000<100000P000`40000300030@00008000<100000`000`40000900002P000`40
002P00030@00024000<100004`000`40000;00030@0000T000<100001`000`40000500030@0000@0
00<100000`000`40000200030@00008000<10000C@000`40000200030@00008000D100000@050003
0@00008000<100000`000`40000200030@0000<000<100002@0000X000<10000X@000`40000P0003
0@0001<000<100002`000`40000900030@0000L000<100001@000`40000400030@0000<000<10000
0P000`40000200030@0004d000<100000P000`40000200050@00004010000`40000300030@0000<0
00<100000P000`40000300030@0000T0000:00030@000:4000<1000080000`40000C00030@0000`0
00<1000020000`40000700030@0000D000<1000010000`40000300030@0000<000<100000P000`40
001<00050@00004010000`40000200030@00008000<100000`000`40000300030@00008000<10000
0`000`40000900002P000`40002Q00030@00024000<100004P000`40000<00030@0000P000<10000
1`000`40000500030@0000@000<100000`000`40000300030@00008000<10000C0001@40000100@0
00D100000@0500030@0000<000<100000`000`40000200030@0000<000<100002@0000X000<10000
X@000`40000Q00030@0001<000<100002`000`40000900030@0000H000<100001P000`4000040003
0@0000<000<100000P000`40000200030@0004`000D100000@0400050@0000401@000`4000020003
0@0000@000<100000P000`40000300030@0000T0000:00030@000:8000<1000080000`40000C0003
0@0000/000<100002@000`40000600030@0000H000<1000010000`40000300030@00008000<10000
0P000`40001<00050@00004010001@40000100D000<100000P000`40000300030@0000<000<10000
0`000`40000900002P000`40002R00030@00024000<100004P000`40000;00030@0000T000<10000
1P000`40000600030@0000@000<1000010000`40000200030@00008000<10000BP001@40000100D0
00D100000@0500030@00008000<100000`000`40000300030@0000<000<100002@0000X000<10000
XP000`40000Q00030@00018000<1000030000`40000800030@0000L000<100001@000`4000040003
0@0000@000<100000P020@<000<10000BP001@40000100@000<100000P000`40000300030@000080
00<100000`000`40000300030@0000<000<100002@0000X000<10000X`000`40000P00030@0001<0
00<100002`000`40000900030@0000H000<100001P000`40000400030@0000<000<100000`000`40
000200030@0004@000<100000P001@40000100@000<100000P000`40000300030@00008000<10000
0`000`40000300030@0000<000<100002@0000X000<10000X`000`40000Q00030@00018000<10000
30000`40000800030@0000H000<100001P000`40000400030@0000@000<100000`000`4000020003
0@0004<000D100000@0400030@00008000<100000P000`40000200030@0000<000<100000`000`40
000300030@0000<000<100002@0000X000<10000Y0000`40000P00030@00018000<1000030000`40
000900030@0000H000<100001@000`40000500030@0000<000<100000`000`40000300030@000480
00D100000@0400030@00008000<100000P000`40000200030@0000<000<100000`000`4000030003
0@0000<000<100002@0000X000<10000Y@000`40000P00030@00014000<100003@000`4000080003
0@0000H000<100001P000`40000400030@0000@000<100000`000`40000200030@00044000<10000
0P000`40000200050@0000401@000`40000200030@0000<000<100000`000`40000300030@0000<0
00<100002@0000X000<10000Y@000`40000P00030@00018000<1000030000`40000900030@0000H0
00<100001@000`40000400030@0000@000<100000`000`40000300030@00040000D100000@040003
0@00008000<100000`000`40000200030@0000<000<100000`000`40000300030@0000<000<10000
2@0000X000<10000YP000`40000P00030@00014000<1000030000`40000900030@0000H00P470003
0@0000@000<1000010000`40000300030@0000<000<10000?P000`40000200030@00008000<10000
0P000`40000300050@0000401@000`40000400030@0000<000<100000`000`40000900002P000`40
002V00030@00020000<100004P000`40000<00030@0000T000<100001P000`40000500030@0000D0
00<100000`000`40000300030@0000@000<10000?@000`40000200050@0000401@000`4000020003
0@00008000<100000`000`40000400030@0000<000<100000`000`40000900002P000`40002W0003
0@00020000<100004@000`40000<00030@0000T000<100001P000`40000600030@0000@000<10000
10000`40000300030@0000<000<10000?0000`40000200030@00008000<100000`000`4000020003
0@00008000<100000`000`40000300030@0000@000<100000`000`40000900002P000`40002W0003
0@00020000<100004P000`40000<00030@0000P000<100001`000`40000500030@0000D000<10000
0`000`40000400030@0000<000<10000>`000`40000200030@00008000<100000P000`4000030003
0@00008000<100000`000`40000300030@0000@000<100000`000`40000900002P000`40002X0003
0@00020000<100004@000`40000<00030@0000T000<100001P000`40000600030@0000D000<10000
0`000`40000400030@0000<00P4j00030@00008000<100000P000`40000300030@00008000<10000
0`000`40000300030@0000<000<1000010000`40000300030@0000T0000:00030@000:T000<10000
80000`40000A00030@0000`000<1000020000`40000700030@0000D000<100001@000`4000040003
0@0000<000<100001@000`40000f00030@00008000<100000`000`40000300030@00008000<10000
0`000`40000300030@0000<000<1000010000`40000300030@0000T0000:00030@000:T000<10000
80000`40000A00030@0000`000<1000020000`40000700030@0000H000<100001@000`4000040003
0@0000<000<100001@000`40000e00030@00008000<100000P000`40000300030@0000<000<10000
0`000`40000300030@0000<000<1000010000`40000300030@0000T0000:00030@000:X000<10000
80000`40000A00030@0000`000<1000020000`40000700030@0000D000<100001@000`4000050003
0@0000<000<1000010000`40000_00030@00008000<100000P000`40000300030@0000<000<10000
0P000`40000400030@00008000<1000010000`40000400030@0000<000<100002@0000X000<10000
ZP000`40000Q00030@00010000<1000030000`40000900030@0000H000<100001P000`4000050003
0@0000@000<1000010000`40000400030@0002d000<100000P000`40000200030@0000<000<10000
10000`40000200030@0000<000<100000`000`40000400030@0000<000<1000010000`4000090000
2P000`40002[00030@00020000<100004@000`40000<00030@0000P000<100001`000`4000060003
0@0000D000<1000010000`40000400811@020B/00P4500030@00008000<100000`000`4000030003
0@0000<000<100000`000`40000300030@0000@000<100000`000`40000400030@0000T0000:0003
0@000:/000<100008@000`40000@00030@0000`000<100002@000`40000700030@0000D000<10000
1P000`40000400030@0000D000<1000010020BP000<1000010000`40000200030@0000@000<10000
0`000`40000200030@0000<000<1000010000`40000300030@0000@000<1000010000`4000090000
2P000`40002/00030@00020000<100004@000`40000<00030@0000P000<100001`000`4000060003
0@0000D000<100001@020@H00P4600<190000`400003008110000`40000400030@0000<000<10000
0`000`40000300030@0000@000<100000`000`40000400030@0000@000<100002@0000X000<10000
[@000`40000P00030@00014000<1000030000`40000800030@0000L000<100001P000`4000050003
0@0000H000<100001@000`40000600818@000`40000300030@0000@000<1000010000`4000030003
0@00008000<1000010000`40000400030@0000<000<1000010000`40000400030@0000T0000:0003
0@000:h00P4P00030@00018000<100002`000`40000800030@0000P000<100001P000`4000050003
0@0000H000<1000010020@T00`4L00811@000`40000400030@0000@000<100000`000`4000030003
0@0000@000<100000`000`40000400030@0000@000<1000010000`40000900002P000`40002`0003
0@0001h000<100004P000`40000;00030@0000P000<1000020000`40000600030@0000D000<10000
1P000`40000500812P070@@00`4=00030@0000D000<1000010000`40000300030@0000@000<10000
0`000`40000300030@0000@000<1000010000`40000400030@0000@000<100002@0000X000<10000
/@000`40000N00030@00018000<100002`000`40000800030@0000L000<100001`000`4000050003
0@0000H00P4700813`040@h00P4600811P000`40000300030@0000@000<1000010000`4000030003
0@0000@000<1000010000`40000400030@0000@000<100002@0000X000<10000/@000`40000O0003
0@00014000<100002`000`40000900030@0000L000<100001`000`40000500030@0000L000<10000
1P030A`00P4700030@0000D000<1000010000`40000400030@0000<000<1000010000`4000040003
0@0000@000<1000010000`40000400030@0000T0000:00030@000;8000<100007`000`40000A0003
0@0000/000<1000020000`40000800030@0000L000<100001@000`40000700812@030AL00P470081
1`000`40000400030@0000@000<1000010000`40000300030@0000@000<100001@000`4000040003
0@0000@000<100002@0000X000<10000/`000`40000O00030@00014000<100002P000`4000090003
0@0000L000<100001`000`40000600812@000`40000900813P070@P000<100001P000`4000040003
0@0000@000<100001@000`40000300030@0000@000<100001@000`40000400030@0000@000<10000
2@0000X000<10000]0000`40000O00030@00014000<100002P000`40000900030@0000L000<10000
1`000`40000700030@0000L000<100002P0>0@d00P4700811P000`40000400030@0000D000<10000
10000`40000400030@0000@000<100001@000`40000400030@0000T0000:00030@000;@000<10000
80000`40000A00030@0000X00P4:00030@0000L000<100001`000`40000700030@0000L00P4R00<1
20000`40000500030@0000D000<100001@000`40000400030@0000@000<1000010000`4000050003
0@0000@000<100002@0000X000<10000]@000`40000P00030@00014000<100002`000`4000080003
0@0000L000<100001`000`40000700812@020Ah00P4:00030@0000H000<1000010000`4000050003
0@0000@000<1000010000`40000500030@0000D000<1000010000`40000900002P000`40002f0003
0@00020000<100004@000`40000;00030@0000L000<100001`000`40000800030@0000P000<10000
20030AP00`4;00030@0000H000<1000010020@L000<1000010000`40000400030@0000D000<10000
10000`40000500030@0000T0000:00030@000;L00P4Q00030@00014000<100002P000`4000080003
0@0000L000<1000020020@T000<100002P040A00104<008120000`40000400030@0000H000<10000
10000`40000500030@0000D000<1000010000`40000500030@0000T0000:00030@000;T000<10000
7`000`40000A00030@0000X000<1000020000`40000700030@0000T000<100001`020@h02@4400<1
3P020@T000<100001@000`40000600030@0000@000<100001@000`40000400030@0000D000<10000
1@000`40000900002P000`40002j008180020A8000<100002P000`40000800030@0000L00P4:0003
0@0000P000<1000050040@l00P4:00030@0000D000<100001P000`40000400030@0000D000<10000
1@000`40000500030@0000D000<100002@0000X000<10000_0000`40000O00030@00010000<10000
2@000`40000900030@0000P000<1000020000`40000800819P000`40000900030@0000D000<10000
1P000`40000500030@0000D000<100001@000`40000500030@0000D000<100002@0000X000<10000
_@020B0000<1000040000`40000900030@0000T000<1000020000`40000800030@0000T00`4O00@1
2P020@L000<100001`000`40000400030@0000H000<100001@000`40000400030@0000H000<10000
2@0000X000<10000_`000`40000N00030@00010000<100002@000`40000900030@0000P00P490081
30040AP00`4=00030@0000L000<100001P000`40000500030@0000H000<1000010000`4000050003
0@0000H000<100002@0000X000<10000`0020Al00P4A00030@0000T000<100002@000`4000090003
0@0000P000<100003@040A00104?00030@0000L000<100001P000`40000500030@0000H000<10000
1@000`40000500030@0000H000<100002@0000X000<10000`P020Al00P4@00030@0000T000<10000
2@000`40000900030@0000P00P4A01014@020@T000<100001P000`40000600030@0000H000<10000
1@000`40000500030@0000H000<100002@0000X000<10000a0030Ah00P4?00030@0000T000<10000
2@000`40000900030@0000T00P4^00812P000`40000700030@0000D000<100001`000`4000050003
0@0000D000<100001P000`40000900002P000`40003700030@0001d000<100003@000`4000090003
0@0000T000<100002@020@/00`4Y00812`000`40000700030@0000H000<100001`000`4000040003
0@0000H000<100001P000`40000900002P000`40003800817P020@h000<100002@000`4000090003
0@0000X000<100002`020B@00`4;00812@000`40000600030@0000L000<100001@000`4000060003
0@0000H000<100002@0000X000<10000bP000`40000M00030@0000`00P4:00812P000`40000:0003
0@0000`00P4P00813@000`40000800030@0000L000<100001`000`40000500030@0000H000<10000
1P000`40000900002P000`40003;00030@0001d00P4>00030@0000P000<100002@000`40000:0003
0@0000d01@4I00813@020@X000<100001`000`40000700030@0000H000<100001P000`4000060003
0@0000T0000:00030@000<`00P4O00030@0000`000<1000020000`40000900812`020A80104?00H1
3P000`40000900030@0000L000<1000020000`40000500030@0000L000<100001P000`4000090000
2P000`40003>00817P000`40000<00030@0000P000<100002@000`40000;0081500?0A<000<10000
2P000`40000700030@0000L000<100001P000`40000700030@0000H000<100002@0000X000<10000
d0030A`000<1000030000`40000800030@0000T00P4=0081<`000`40000:00030@0000L000<10000
20000`40000600030@0000H000<100001`000`40000900002P000`40003C00@16@000`40000<0003
0@0000P00P4;00030@0000`00P4_008130000`40000700030@0000P000<100001`000`4000060003
0@0000L000<100002@0000X000<10000e`030AL00P4=00030@0000T000<100002@000`40000=0081
:`020@d000<100001`000`40000800030@0000P000<100001P000`40000700030@0000T0000:0003
0@000=X00P4G00030@0000/000<100002@000`40000900813`030BH00P4>00030@0000P000<10000
20000`40000700030@0000L000<100001`000`40000900002P000`40003K00815`020@`000<10000
2@000`40000:00030@0000l0104O00<13`000`40000800030@0000P000<1000020000`4000070003
0@0000L000<100002@0000X000<10000g@030AH000<100002P000`40000900030@0000X00P4C0081
5`060A000P4;00030@0000P000<1000020000`40000700030@0000L000<100002@0000X000<10000
h0020AD00P4;00030@0000P000<1000030000`40000B00@140030A@00P4<00030@0000P000<10000
2@000`40000600030@0000P000<100002@0000X000<10000hP030A@00P4:00030@0000P000<10000
30020AH0404F00030@0000/000<100002@000`40000800030@0000L000<1000020000`4000090000
2P000`40003U00<14`000`40000700030@0000T000<100003@020CP00P4=00030@0000T000<10000
2@000`40000700030@0000P000<100002@0000X000<10000j0080@`000<100001`000`4000090081
3`020CD000<1000030000`40000:00030@0000P000<1000020000`40000800030@0000T0000:0003
0@000?000P4;00030@0000L000<100002P020@l00P4a00813P000`40000:00030@0000T000<10000
20000`40000800030@0000T0000:00030@000?800P4:00030@0000L000<100002`000`40000>0081
;@020@l000<100002`000`40000800030@0000T000<1000020000`40000900002P000`40003d0081
20000`40000800030@0000/000<100003`020BT00P4@00030@0000/000<100002@000`4000090003
0@0000P000<100002@0000X000<10000mP020@L000<1000020000`40000;00030@0001000`4T0081
4@000`40000;00030@0000T000<100002P000`40000800030@0000T0000:00030@000?P000<10000
1@000`40000800030@0000/00P4C00<170050A8000<1000030000`40000900030@0000T000<10000
2@000`40000900002P000`40003i00030@0000D000<1000020000`40000<00030@0001<00`4D00D1
5@020@h000<100002P000`40000900030@0000T000<100002@0000X000<10000n@000`4000060003
0@0000P000<1000030000`40000E00H11`070AT000<100003@000`40000:00030@0000X000<10000
2@000`40000900002P000`40003j00030@0000D000<100002@000`40000<00030@0001X01`4O0003
0@0000d000<100002`000`40000:00030@0000T000<100002@0000X000<10000n`000`4000050003
0@0000T000<1000030020Ch00P4@00030@0000X000<100002`000`40000900030@0000T0000:0?L1
1`000`40000500030@0000X000<100003@020CX00P4A00030@0000/000<100002P000`40000:0003
0@0000T0003o004000<100001P000`40000500030@0000X000<100003P020CH00P4B00030@0000/0
00<100002`000`40000:00030@0000T0003o004000<100001P000`40000500030@0000/000<10000
3`020C800P4C00030@0000`000<100002`000`40000:00030@0000T0003o004000<100001P000`40
000600030@0000/000<1000040020Bl000<100004P000`40000<00030@0000`000<100002P000`40
00090000o`0100030@0000L000<100001P000`40000;00030@000140104Y008150000`40000=0003
0@0000/000<100002`000`4000090000o`0100030@0000L000<100001P000`40000<00030@0001@0
0`4S00<15@000`40000=00030@0000`000<100002`000`4000090000o`0100030@0000L000<10000
1`000`40000<00030@0001H00`4J00H15P020A0000<1000030000`40000;00030@0000T0003o0040
00<1000020000`40000700030@0000`000<1000060040A401@4K00030@0000l000<100003@000`40
000;00030@0000T0003o004000<1000020000`40000800030@0000`00P4L00D11P060Al000<10000
3`000`40000=00030@0000`000<100002@000?l00@000`40000800030@0000T000<100003@000`40
000N00H190000`40000@00030@0000d000<1000030000`4000090000o`0100030@0000P000<10000
2@000`40000>00030@0004D00P4B00030@0000d000<100003@000`4000090000o`0100030@0000P0
00<100002P000`40000>00030@0004800P4C00030@0000h000<100003@000`4000090000o`010003
0@0000P000<100002`000`40000>0081@@000`40000C00030@0000h000<100003@000`4000090000
o`0100030@0000P000<100002`000`40000@0081?@020AD000<100003`000`40000=00030@0000T0
003o004000<100002@000`40000;00030@0001400P4i00815P000`40000?00030@0000h000<10000
2@000?l00@000`40000900030@0000`000<100004P020CD00P4H00030@0000l000<100003P000`40
00090000o`0100030@0000T000<1000030000`40000D0081<@020AT000<1000040000`40000>0003
0@0000T0003o004000<100002P000`40000<00030@0001D00P4]00816P000`40000A00030@0000h0
00<100002@000?l00@000`40000:00030@0000d00P4G00<19P040A/000<100004@000`40000?0003
0@0000T0003o004000<100002P000`40000?00030@0001L00`4O00@17P000`40000B00030@0000l0
00<100002@000?l00@000`40000;00030@0000l000<100006@040AH01@4P008150000`40000@0003
0@0000T0003o004000<100002`000`40000@00030@0001`02P4600H190000`40000D00030@000100
00<100002@000?l00@000`40000;00030@0001400P4V00H1:@000`40000D00030@00014000<10000
2@000?l00@000`40000<00030@00018000<10000D@000`40000E00030@00014000<100002@000?l0
0@000`40000<00030@0001<000<10000C`000`40000E00030@00018000<100002@000?l00@000`40
000=00030@0001<000<10000C@000`40000F00030@00018000<100002@000?l00@000`40000>0003
0@0001<000<10000BP020AP000<100004`000`4000090000o`0100030@0000h000<1000050020DP0
0P4J00030@0001<000<100002@000?l00@000`40000?00030@0001D00P5500030@0001T000<10000
50000`4000090000o`0100030@00010000<100005P000`40001100030@0001X000<1000050000`40
00090000o`0100030@00010000<100005`020Cl00P4L00030@0001D000<100002@000?l00@000`40
000A00030@0001P00P4l00030@0001`000<100005@000`4000090000o`0100030@00014000<10000
6P000`40000h00030@0001`000<100005P000`4000090000o`0100030@00018000<100006P020CH0
0P4O00030@0001H000<100002@000?l00@000`40000B00030@0001`0104_00<180000`40000G0003
0@0000T0003o004000<100004`000`40000O00@19P050B8000<1000060000`4000090000o`010003
0@0001<000<100008`020B00104V00030@0001T000<100002@000?l00@000`40000D00030@0002@0
0`4D00T1:@000`40000J00030@0000T0003o004000<100005@000`40000V00T11`040C8000<10000
6P000`4000090000o`0100030@0001H000<10000;P070CD000<100006`000`4000090000o`010003
0@0001H000<10000J@000`40000L00030@0000T0003o004000<100005`000`40001W00030@0001d0
00<100002@000?l00@000`40000H00030@0006D000<100007P000`4000090000o`0100030@0001T0
00<10000H`000`40000O00030@0000T0003o004000<100006P000`40001R00030@0001l000<10000
2@000?l00@000`40000K00030@00060000<1000080000`4000090000o`0100030@0001`000<10000
G@020B<000<100002@000?l00@000`40000M00030@0005X00P4U00030@0000T0003o004000<10000
7P020EP00P4W00030@0000T0003o004000<1000080020E@00P4Y00030@0000T0003o004000<10000
8P020E000P4[00030@0000T0003o004000<1000090020D`00P4]00030@0000T0003o004000<10000
9P000`4000170081;`000`4000090000o`0100030@0002L00P550081<@000`4000090000o`010003
0@0002T00`4o00<1<`000`4000090000o`0100030@0002`0104g00@1=P000`4000090000o`010003
0@000300104`00<1>P000`4000090000o`0100030@0003@0104U00L1?@000`4000090000o`010003
0@0003P0104J00L1A0000`4000090000o`0100030@0003`06P5;00030@0000T0003o004000<10000
X@000`4000090000o`0100030@000:4000<100002@000?l00@000`40002Q00030@0000T0003o0040
00<10000X@000`4000090000o`0100030@000:4000<100002@000?l00@000`40002Q00030@0000T0
003o004000<10000X@000`4000090000o`0100030@000:4000<100002@000?l00@000`40002Q0003
0@0000T0003o004000<10000X@000`4000090000o`0100030@000:4000<100002@000?l00@000`40
002Q00030@0000T0003o004000<10000X@000`4000090000o`0100030@000:4000<100002@000?l0
0@000`40002Q00030@0000T0003o004000<10000X@000`4000090000o`0100030@000:4000<10000
2@000?l00@000`40002Q00030@0000T0003o004000<10000X@000`4000090000o`0100030@000:40
00<100002@000?l00@000`40002Q00030@0000T0003o004000<10000X@000`4000090000o`010003
0@000:4000<100002@000?l00@000`40002Q00030@0000T0003o0040Y@4;0000o`2a0000o`2a0000
o`2a0000o`2a0000o`2a0000o`2a0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.129923, -0.0750159, 
  0.0183057, 0.0177716}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(levels = Table[j, {j, 0, 1, 0.05}]; \n
    glevel3D = LevelGraphics3D[points, elements, levels]; \nShow[glevel3D]; 
    \)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .58689 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12  scalefont  setfont
% Scaling calculations
5.55112e-17 1.00617 -0.177348 1.00617 [
[ 0 0 0 0 ]
[ 1 .58689 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.00963 .27532 m
0 .35096 L
s
0 .35096 m
.27583 .58689 L
s
.27583 .58689 m
.2797 .51566 L
s
.2797 .51566 m
.00963 .27532 L
s
.78654 0 m
.9906 .3013 L
s
.9906 .3013 m
1 .37651 L
s
1 .37651 m
.79279 .07943 L
s
.79279 .07943 m
.78654 0 L
s
.00963 .27532 m
0 .35096 L
s
0 .35096 m
.79279 .07943 L
s
.79279 .07943 m
.78654 0 L
s
.78654 0 m
.00963 .27532 L
s
.2797 .51566 m
.9906 .3013 L
s
.9906 .3013 m
1 .37651 L
s
1 .37651 m
.27583 .58689 L
s
.27583 .58689 m
.2797 .51566 L
s
0 0 m
1 0 L
1 .58689 L
0 .58689 L
closepath
clip
newpath
.5 Mabswid
.23531 .42862 m
.21009 .43662 L
s
.21009 .43662 m
.19357 .42179 L
s
.26073 .42055 m
.23531 .42862 L
s
.23531 .42862 m
.23531 .42862 L
s
.19357 .42179 m
.19357 .42179 L
s
.1768 .40674 m
.19357 .42179 L
s
.26073 .42055 m
.28636 .41241 L
s
.26073 .42055 m
.26073 .42055 L
s
.31219 .40421 m
.28636 .41241 L
s
.28636 .41241 m
.28636 .41241 L
s
.1768 .40674 m
.1768 .40674 L
s
.15978 .39146 m
.1768 .40674 L
s
.33824 .39594 m
.31219 .40421 L
s
.31219 .40421 m
.31219 .40421 L
s
.33824 .39594 m
.3645 .38761 L
s
.33824 .39594 m
.33824 .39594 L
s
.15978 .39146 m
.15978 .39146 L
s
.15978 .39146 m
.1425 .37595 L
s
.70741 .38115 m
.67966 .38954 L
s
.67966 .38954 m
.66632 .374 L
s
.39097 .37921 m
.3645 .38761 L
s
.3645 .38761 m
.3645 .38761 L
s
.73538 .37269 m
.70741 .38115 L
s
.70741 .38115 m
.70741 .38115 L
s
.39097 .37921 m
.41766 .37073 L
s
.39097 .37921 m
.39097 .37921 L
s
.1425 .37595 m
.1425 .37595 L
s
.1425 .37595 m
.12496 .36021 L
s
.66632 .374 m
.66632 .374 L
s
.65277 .35822 m
.66632 .374 L
s
.73538 .37269 m
.76358 .36417 L
s
.73538 .37269 m
.73538 .37269 L
s
.44458 .36219 m
.41766 .37073 L
s
.41766 .37073 m
.41766 .37073 L
s
.76358 .36417 m
.79202 .35557 L
s
.76358 .36417 m
.76358 .36417 L
s
.44458 .36219 m
.47171 .35357 L
s
.44458 .36219 m
.44458 .36219 L
s
.12496 .36021 m
.12496 .36021 L
s
.12496 .36021 m
.10715 .34422 L
s
.65277 .35822 m
.65277 .35822 L
s
.65277 .35822 m
.63902 .3422 L
s
.79202 .35557 m
.82069 .3469 L
s
.79202 .35557 m
.79202 .35557 L
s
.49908 .34489 m
.47171 .35357 L
s
.47171 .35357 m
.47171 .35357 L
s
.84961 .33816 m
.82069 .3469 L
s
.82069 .3469 m
.82069 .3469 L
s
.52667 .33613 m
.49908 .34489 L
s
.49908 .34489 m
.49908 .34489 L
s
.10715 .34422 m
.10715 .34422 L
s
.10715 .34422 m
.08905 .32798 L
s
.63902 .3422 m
.63902 .3422 L
s
.63902 .3422 m
.62505 .32593 L
s
.84961 .33816 m
.87876 .32934 L
s
.84961 .33816 m
.84961 .33816 L
s
.5545 .3273 m
.52667 .33613 L
s
.52667 .33613 m
.52667 .33613 L
s
.73246 .34393 m
.72416 .34434 L
s
.70549 .34387 m
.72416 .34434 L
s
.74951 .34217 m
.73246 .34393 L
s
.70549 .34387 m
.69293 .3421 L
s
.76317 .33947 m
.74951 .34217 L
s
.6814 .33946 m
.69293 .3421 L
s
.76317 .33947 m
.77314 .33767 L
s
.66966 .33421 m
.6814 .33946 L
s
.87876 .32934 m
.90817 .32045 L
s
.87876 .32934 m
.87876 .32934 L
s
.08905 .32798 m
.08905 .32798 L
s
.08905 .32798 m
.07068 .31148 L
s
.77954 .33595 m
.77314 .33767 L
s
.5545 .3273 m
.58256 .31839 L
s
.5545 .3273 m
.5545 .3273 L
s
.80403 .32884 m
.77954 .33595 L
s
.62505 .32593 m
.62505 .32593 L
s
.61086 .3094 m
.62505 .32593 L
s
.66033 .33162 m
.66966 .33421 L
s
.65374 .32898 m
.66033 .33162 L
s
.90817 .32045 m
.93782 .31148 L
s
.90817 .32045 m
.90817 .32045 L
s
.64074 .32201 m
.65374 .32898 L
s
.80835 .32747 m
.80403 .32884 L
s
.58256 .31839 m
.61086 .3094 L
s
.58256 .31839 m
.58256 .31839 L
s
.83204 .31805 m
.80835 .32747 L
s
.63711 .31774 m
.64074 .32201 L
s
.96773 .30244 m
.93782 .31148 L
s
.93782 .31148 m
.93782 .31148 L
s
.07068 .31148 m
.07068 .31148 L
s
.05201 .29473 m
.07068 .31148 L
s
.61086 .3094 m
.61086 .3094 L
s
.83204 .31805 m
.8329 .31773 L
s
.63711 .31774 m
.62324 .31041 L
s
.8329 .31773 m
.83525 .31644 L
s
.73094 .32681 m
.70519 .3272 L
s
.70378 .32707 m
.70519 .3272 L
s
.70378 .32707 m
.69939 .32637 L
s
.73094 .32681 m
.73299 .32672 L
s
.73299 .32672 m
.74224 .32523 L
s
.69939 .32637 m
.67913 .32398 L
s
.8559 .30586 m
.83525 .31644 L
s
.50239 .31639 m
.51083 .31556 L
s
.48933 .31579 m
.50239 .31639 L
s
.52939 .31636 m
.55569 .31408 L
s
.51083 .31556 m
.52939 .31636 L
s
.48933 .31579 m
.47612 .31448 L
s
.75854 .32297 m
.74224 .32523 L
s
.47612 .31448 m
.45569 .31053 L
s
.55717 .31411 m
.55647 .31411 L
s
.55569 .31408 m
.55647 .31411 L
s
.55717 .31411 m
.58451 .31178 L
s
.67913 .32398 m
.67449 .32337 L
s
.67449 .32337 m
.6566 .31835 L
s
.75854 .32297 m
.76199 .32204 L
s
.9561 .28554 m
.96773 .30244 L
s
.78263 .31681 m
.76199 .32204 L
s
.58451 .31178 m
.59604 .30811 L
s
.4509 .31002 m
.45569 .31053 L
s
.62324 .31041 m
.61345 .30819 L
s
.44776 .30948 m
.4509 .31002 L
s
.42238 .30047 m
.44776 .30948 L
s
.6566 .31835 m
.64251 .31501 L
s
.61345 .30819 m
.59604 .30811 L
s
.79648 .31123 m
.78263 .31681 L
s
.86408 .29865 m
.8559 .30586 L
s
.64251 .31501 m
.63572 .31142 L
s
.05201 .29473 m
.05201 .29473 L
s
.03305 .27771 m
.05201 .29473 L
s
.63572 .31142 m
.61606 .30697 L
s
.79648 .31123 m
.80361 .30861 L
s
.41432 .2976 m
.42238 .30047 L
s
.80361 .30861 m
.80958 .30541 L
s
.87487 .29078 m
.86408 .29865 L
s
.41432 .2976 m
.40472 .29053 L
s
.61606 .30697 m
.59773 .30688 L
s
.59773 .30688 m
.58869 .30602 L
s
.69698 .31668 m
.69268 .31671 L
s
.67249 .31448 m
.69268 .31671 L
s
.69698 .31668 m
.71908 .31654 L
s
.71908 .31654 m
.72359 .31596 L
s
.58341 .30518 m
.58869 .30602 L
s
.72359 .31596 m
.74362 .31352 L
s
.9561 .28554 m
.9561 .28554 L
s
.94428 .26838 m
.9561 .28554 L
s
.80958 .30541 m
.82434 .29776 L
s
.56221 .30427 m
.58341 .30518 L
s
.66507 .31377 m
.67249 .31448 L
s
.55654 .30432 m
.56221 .30427 L
s
.55294 .30417 m
.55654 .30432 L
s
.55294 .30417 m
.53312 .3018 L
s
.64974 .31048 m
.66507 .31377 L
s
.74362 .31352 m
.75294 .31112 L
s
.51258 .29761 m
.53312 .3018 L
s
.75294 .31112 m
.76614 .30796 L
s
.8781 .28448 m
.87487 .29078 L
s
.40472 .29053 m
.39405 .28347 L
s
.64974 .31048 m
.63992 .30897 L
s
.63992 .30897 m
.62766 .3064 L
s
.76614 .30796 m
.78608 .2999 L
s
.82434 .29776 m
.84018 .28394 L
s
.03305 .27771 m
.05952 .26838 L
s
.51258 .29761 m
.5073 .29622 L
s
.62766 .3064 m
.62005 .30507 L
s
.5073 .29622 m
.50137 .29361 L
s
.6087 .30329 m
.62005 .30507 L
s
.88696 .27163 m
.8781 .28448 L
s
.50137 .29361 m
.48527 .2884 L
s
.38313 .26941 m
.39405 .28347 L
s
.59764 .30224 m
.6087 .30329 L
s
.57602 .29879 m
.59764 .30224 L
s
.78625 .29984 m
.78608 .2999 L
s
.78637 .29977 m
.78625 .29984 L
s
.68802 .30978 m
.68371 .30983 L
s
.66434 .30811 m
.68371 .30983 L
s
.80475 .29019 m
.78637 .29977 L
s
.70821 .30966 m
.68802 .30978 L
s
.71356 .30902 m
.70821 .30966 L
s
.71356 .30902 m
.73133 .30702 L
s
.57602 .29879 m
.57577 .29875 L
s
.57577 .29875 m
.57565 .29872 L
s
.55695 .29508 m
.57565 .29872 L
s
.94428 .26838 m
.93227 .25093 L
s
.94428 .26838 m
.94428 .26838 L
s
.08622 .25896 m
.05952 .26838 L
s
.05952 .26838 m
.05952 .26838 L
s
.48527 .2884 m
.47335 .28181 L
s
.65848 .30763 m
.66434 .30811 L
s
.64364 .30543 m
.65848 .30763 L
s
.73133 .30702 m
.74124 .30461 L
s
.62996 .30317 m
.64364 .30543 L
s
.55112 .29318 m
.55695 .29508 L
s
.75261 .30202 m
.74124 .30461 L
s
.8415 .28294 m
.84018 .28394 L
s
.54126 .29073 m
.55112 .29318 L
s
.62561 .3024 m
.62996 .30317 L
s
.84322 .28049 m
.8415 .28294 L
s
.62112 .3017 m
.62561 .3024 L
s
.77082 .29491 m
.75261 .30202 L
s
.47335 .28181 m
.46872 .27889 L
s
.88658 .2603 m
.88696 .27163 L
s
.60576 .29898 m
.62112 .3017 L
s
.54126 .29073 m
.53288 .28774 L
s
.85327 .26768 m
.84322 .28049 L
s
.80475 .29019 m
.81668 .28009 L
s
.38313 .26941 m
.38294 .26915 L
s
.38294 .26915 m
.38185 .2661 L
s
.11315 .24947 m
.08622 .25896 L
s
.08622 .25896 m
.08622 .25896 L
s
.46872 .27889 m
.45513 .26721 L
s
.59229 .29576 m
.60576 .29898 L
s
.53288 .28774 m
.5231 .28378 L
s
.37695 .25347 m
.38185 .2661 L
s
.59229 .29576 m
.58653 .2948 L
s
.67804 .30503 m
.67677 .30504 L
s
.67677 .30504 m
.65319 .3031 L
s
.69952 .30491 m
.67804 .30503 L
s
.77082 .29491 m
.77153 .29467 L
s
.70247 .3046 m
.69952 .30491 L
s
.58653 .2948 m
.57884 .29264 L
s
.77194 .29444 m
.77153 .29467 L
s
.72121 .30265 m
.70247 .3046 L
s
.77194 .29444 m
.78877 .28596 L
s
.5231 .28378 m
.50737 .27459 L
s
.65319 .3031 m
.65229 .30303 L
s
.65072 .30277 m
.65229 .30303 L
s
.63152 .29999 m
.65072 .30277 L
s
.57884 .29264 m
.56904 .28984 L
s
.72832 .30104 m
.72121 .30265 L
s
.92007 .23321 m
.93227 .25093 L
s
.93227 .25093 m
.93227 .25093 L
s
.72832 .30104 m
.74136 .29825 L
s
.8889 .25303 m
.88658 .2603 L
s
.81668 .28009 m
.81962 .27795 L
s
.11315 .24947 m
.14031 .23989 L
s
.11315 .24947 m
.11315 .24947 L
s
.61583 .2967 m
.63152 .29999 L
s
.56904 .28984 m
.55914 .28562 L
s
.74136 .29825 m
.75634 .29273 L
s
.82321 .27346 m
.81962 .27795 L
s
.85464 .26244 m
.85327 .26768 L
s
.45363 .26543 m
.45513 .26721 L
s
.61583 .2967 m
.61321 .29618 L
s
.61036 .29533 m
.61321 .29618 L
s
.78877 .28596 m
.80048 .27658 L
s
.44477 .25198 m
.45363 .26543 L
s
.55296 .28349 m
.55914 .28562 L
s
.5955 .29168 m
.61036 .29533 L
s
.50737 .27459 m
.50637 .27393 L
s
.50069 .26858 m
.50637 .27393 L
s
.37686 .25235 m
.37695 .25347 L
s
.83153 .264 m
.82321 .27346 L
s
.88599 .23915 m
.8889 .25303 L
s
.5486 .28133 m
.55296 .28349 L
s
.3773 .24999 m
.37686 .25235 L
s
.85886 .25364 m
.85464 .26244 L
s
.75634 .29273 m
.75941 .29167 L
s
.58333 .2873 m
.5955 .29168 L
s
.75941 .29167 m
.76272 .28984 L
s
.69143 .30172 m
.67208 .30179 L
s
.67208 .30179 m
.66855 .3015 L
s
.69259 .30171 m
.69143 .30172 L
s
.6983 .3011 m
.69259 .30171 L
s
.5486 .28133 m
.53758 .27536 L
s
.66855 .3015 m
.65369 .30035 L
s
.71278 .29978 m
.6983 .3011 L
s
.14031 .23989 m
.16772 .23022 L
s
.14031 .23989 m
.14031 .23989 L
s
.3773 .24999 m
.37734 .23526 L
s
.64521 .29898 m
.65369 .30035 L
s
.77574 .28362 m
.76272 .28984 L
s
.7154 .29924 m
.71278 .29978 L
s
.7319 .29597 m
.7154 .29924 L
s
.49202 .26099 m
.50069 .26858 L
s
.64521 .29898 m
.6374 .29796 L
s
.6374 .29796 m
.629 .29599 L
s
.57941 .28587 m
.58333 .2873 L
s
.80048 .27658 m
.80229 .27533 L
s
.57941 .28587 m
.57334 .2827 L
s
.629 .29599 m
.61969 .29414 L
s
.7319 .29597 m
.74088 .29291 L
s
.52822 .26821 m
.53758 .27536 L
s
.80409 .27337 m
.80229 .27533 L
s
.83153 .264 m
.83302 .2605 L
s
.85971 .24347 m
.85886 .25364 L
s
.92007 .23321 m
.90766 .21519 L
s
.92007 .23321 m
.92007 .23321 L
s
.61238 .29195 m
.61969 .29414 L
s
.78721 .2749 m
.77574 .28362 L
s
.80409 .27337 m
.81234 .26496 L
s
.56454 .27878 m
.57334 .2827 L
s
.74088 .29291 m
.7492 .29022 L
s
.4434 .24783 m
.44477 .25198 L
s
.61238 .29195 m
.60333 .28928 L
s
.49089 .25964 m
.49202 .26099 L
s
.19537 .22047 m
.16772 .23022 L
s
.16772 .23022 m
.16772 .23022 L
s
.83302 .2605 m
.83812 .24842 L
s
.48702 .25176 m
.49089 .25964 L
s
.88599 .23915 m
.88629 .23631 L
s
.7492 .29022 m
.75972 .28481 L
s
.70291 .29846 m
.6871 .29984 L
s
.6871 .29984 m
.68084 .29981 L
s
.66796 .29958 m
.68084 .29981 L
s
.66124 .299 m
.66796 .29958 L
s
.59009 .2836 m
.60333 .28928 L
s
.56005 .27594 m
.56454 .27878 L
s
.44124 .23739 m
.4434 .24783 L
s
.64793 .29726 m
.66124 .299 L
s
.52822 .26821 m
.52401 .26455 L
s
.70291 .29846 m
.70591 .29819 L
s
.70591 .29819 m
.7148 .29636 L
s
.88348 .22655 m
.88629 .23631 L
s
.64301 .29661 m
.64793 .29726 L
s
.37734 .23526 m
.37872 .22905 L
s
.63702 .29521 m
.64301 .29661 L
s
.56005 .27594 m
.55123 .26994 L
s
.72384 .29476 m
.7148 .29636 L
s
.81862 .25509 m
.81234 .26496 L
s
.78721 .2749 m
.78875 .27382 L
s
.51539 .25359 m
.52401 .26455 L
s
.63702 .29521 m
.62507 .29261 L
s
.76465 .2826 m
.75972 .28481 L
s
.72384 .29476 m
.72653 .29391 L
s
.86058 .24046 m
.85971 .24347 L
s
.78935 .27316 m
.78875 .27382 L
s
.7404 .28977 m
.72653 .29391 L
s
.58864 .28307 m
.59009 .2836 L
s
.79945 .26325 m
.78935 .27316 L
s
.58781 .28264 m
.58864 .28307 L
s
.58781 .28264 m
.57521 .27581 L
s
.76818 .28008 m
.76465 .2826 L
s
.48702 .25176 m
.48187 .24266 L
s
.62507 .29261 m
.61178 .28805 L
s
.19537 .22047 m
.22326 .21064 L
s
.19537 .22047 m
.19537 .22047 L
s
.8607 .23682 m
.86058 .24046 L
s
.38523 .21555 m
.37872 .22905 L
s
.54416 .2631 m
.55123 .26994 L
s
.77896 .273 m
.76818 .28008 L
s
.75072 .28493 m
.7404 .28977 L
s
.83812 .24842 m
.83904 .24663 L
s
.44124 .23739 m
.44085 .23001 L
s
.68229 .29876 m
.69123 .29827 L
s
.67146 .29849 m
.68229 .29876 L
s
.88348 .22655 m
.88152 .21976 L
s
.86056 .22562 m
.8607 .23682 L
s
.67146 .29849 m
.66426 .29807 L
s
.61178 .28805 m
.61035 .28756 L
s
.70015 .29764 m
.69123 .29827 L
s
.83923 .24607 m
.83904 .24663 L
s
.65236 .29661 m
.66426 .29807 L
s
.60759 .28621 m
.61035 .28756 L
s
.71275 .29543 m
.70015 .29764 L
s
.84186 .23265 m
.83923 .24607 L
s
.90766 .21519 m
.89505 .19687 L
s
.90766 .21519 m
.90766 .21519 L
s
.6451 .29541 m
.65236 .29661 L
s
.59686 .28156 m
.60759 .28621 L
s
.56944 .27137 m
.57521 .27581 L
s
.81862 .25509 m
.81939 .25358 L
s
.71275 .29543 m
.71737 .29461 L
s
.63169 .29205 m
.6451 .29541 L
s
.75475 .28312 m
.75072 .28493 L
s
.51539 .25359 m
.51357 .25073 L
s
.81939 .25358 m
.82063 .24989 L
s
.71737 .29461 m
.72587 .29192 L
s
.48187 .24266 m
.48143 .24142 L
s
.79945 .26325 m
.80132 .26065 L
s
.54416 .2631 m
.54031 .2589 L
s
.78349 .26826 m
.77896 .273 L
s
.75942 .27979 m
.75475 .28312 L
s
.25141 .20071 m
.22326 .21064 L
s
.22326 .21064 m
.22326 .21064 L
s
.48049 .22338 m
.48143 .24142 L
s
.63169 .29205 m
.62959 .29149 L
s
.56378 .26673 m
.56944 .27137 L
s
.51357 .25073 m
.5091 .23803 L
s
.88152 .21976 m
.88045 .21825 L
s
.44266 .2213 m
.44085 .23001 L
s
.72587 .29192 m
.73299 .29 L
s
.59211 .27864 m
.59686 .28156 L
s
.80803 .24926 m
.80132 .26065 L
s
.62959 .29149 m
.62486 .28968 L
s
.82063 .24989 m
.82463 .24051 L
s
.87511 .20264 m
.88045 .21825 L
s
.76848 .27419 m
.75942 .27979 L
s
.73656 .2884 m
.73299 .29 L
s
.53465 .24945 m
.54031 .2589 L
s
.62486 .28968 m
.61652 .28683 L
s
.59211 .27864 m
.58504 .27423 L
s
.78349 .26826 m
.79094 .26018 L
s
.74551 .28471 m
.73656 .2884 L
s
.67808 .29833 m
.68113 .29829 L
s
.67808 .29833 m
.66327 .29738 L
s
.68113 .29829 m
.69494 .29771 L
s
.39744 .20079 m
.38523 .21555 L
s
.85812 .21841 m
.86056 .22562 L
s
.55736 .2591 m
.56378 .26673 L
s
.69494 .29771 m
.70006 .29704 L
s
.66327 .29738 m
.66117 .29716 L
s
.61262 .28492 m
.61652 .28683 L
s
.66117 .29716 m
.64631 .29457 L
s
.70006 .29704 m
.71168 .29528 L
s
.72139 .29259 m
.71168 .29528 L
s
.61262 .28492 m
.60446 .28095 L
s
.75442 .27937 m
.74551 .28471 L
s
.57786 .26766 m
.58504 .27423 L
s
.76848 .27419 m
.77201 .27114 L
s
.84192 .22193 m
.84186 .23265 L
s
.64631 .29457 m
.64529 .29435 L
s
.64529 .29435 m
.63682 .29197 L
s
.25141 .20071 m
.2798 .1907 L
s
.25141 .20071 m
.25141 .20071 L
s
.44266 .2213 m
.44759 .20808 L
s
.72139 .29259 m
.72785 .29085 L
s
.82463 .24051 m
.82536 .23795 L
s
.63682 .29197 m
.63296 .29093 L
s
.77201 .27114 m
.78013 .26266 L
s
.59687 .27567 m
.60446 .28095 L
s
.79094 .26018 m
.79368 .25593 L
s
.85736 .20895 m
.85812 .21841 L
s
.53465 .24945 m
.53294 .24591 L
s
.63296 .29093 m
.6323 .29068 L
s
.73287 .28859 m
.72785 .29085 L
s
.80803 .24926 m
.80828 .24868 L
s
.6323 .29068 m
.62265 .28723 L
s
.88224 .17825 m
.89505 .19687 L
s
.89505 .19687 m
.89505 .19687 L
s
.55736 .2591 m
.55496 .25585 L
s
.80828 .24868 m
.80865 .24731 L
s
.5091 .23803 m
.50835 .23472 L
s
.82536 .23795 m
.82631 .22641 L
s
.75442 .27937 m
.75691 .27782 L
s
.80865 .24731 m
.81224 .23751 L
s
.74064 .28525 m
.73287 .28859 L
s
.57521 .26515 m
.57786 .26766 L
s
.75691 .27782 m
.76321 .27237 L
s
.67456 .29859 m
.68947 .29841 L
s
.67456 .29859 m
.67263 .29846 L
s
.67263 .29846 m
.65951 .29708 L
s
.68947 .29841 m
.69018 .29838 L
s
.69018 .29838 m
.6955 .29768 L
s
.53294 .24591 m
.53061 .23576 L
s
.6955 .29768 m
.7064 .29639 L
s
.62265 .28723 m
.61414 .28257 L
s
.79368 .25593 m
.79893 .24496 L
s
.55496 .25585 m
.55074 .24676 L
s
.50865 .22279 m
.50835 .23472 L
s
.65951 .29708 m
.65658 .29657 L
s
.65658 .29657 m
.64444 .29388 L
s
.59687 .27567 m
.59431 .27387 L
s
.7064 .29639 m
.70741 .29618 L
s
.56887 .25607 m
.57521 .26515 L
s
.87114 .19696 m
.87511 .20264 L
s
.48058 .22176 m
.48049 .22338 L
s
.70741 .29618 m
.72223 .29257 L
s
.74064 .28525 m
.74621 .28217 L
s
.84137 .2157 m
.84192 .22193 L
s
.30846 .1806 m
.2798 .1907 L
s
.2798 .1907 m
.2798 .1907 L
s
.59431 .27387 m
.5887 .26778 L
s
.48058 .22176 m
.48516 .20752 L
s
.39813 .19949 m
.39744 .20079 L
s
.78385 .25756 m
.78013 .26266 L
s
.6421 .29322 m
.64444 .29388 L
s
.39813 .19949 m
.39887 .19863 L
s
.63258 .28956 m
.6421 .29322 L
s
.76754 .26895 m
.76321 .27237 L
s
.61414 .28257 m
.61126 .28095 L
s
.74621 .28217 m
.75343 .27724 L
s
.72223 .29257 m
.72555 .29142 L
s
.41327 .18449 m
.39887 .19863 L
s
.85091 .19261 m
.85736 .20895 L
s
.44759 .20808 m
.4494 .20509 L
s
.72555 .29142 m
.73585 .28686 L
s
.60356 .27516 m
.61126 .28095 L
s
.8662 .18435 m
.87114 .19696 L
s
.76831 .26816 m
.76754 .26895 L
s
.81301 .23129 m
.81224 .23751 L
s
.63258 .28956 m
.62903 .28831 L
s
.68027 .29989 m
.6726 .29954 L
s
.68027 .29989 m
.68493 .29987 L
s
.68493 .29987 m
.6961 .29896 L
s
.6726 .29954 m
.66577 .29877 L
s
.77565 .25894 m
.76831 .26816 L
s
.82713 .22113 m
.82631 .22641 L
s
.78718 .25158 m
.78385 .25756 L
s
.83964 .2114 m
.84137 .2157 L
s
.58562 .26485 m
.5887 .26778 L
s
.4494 .20509 m
.45291 .20112 L
s
.6961 .29896 m
.70203 .29825 L
s
.54969 .2438 m
.55074 .24676 L
s
.66577 .29877 m
.6584 .29764 L
s
.62903 .28831 m
.62854 .28807 L
s
.62854 .28807 m
.62341 .28505 L
s
.53012 .23178 m
.53061 .23576 L
s
.70203 .29825 m
.71219 .29609 L
s
.75698 .27443 m
.75343 .27724 L
s
.56794 .25431 m
.56887 .25607 L
s
.65126 .29607 m
.6584 .29764 L
s
.73585 .28686 m
.73959 .28483 L
s
.79893 .24496 m
.79919 .24424 L
s
.79931 .24322 m
.79919 .24424 L
s
.58483 .26371 m
.58562 .26485 L
s
.56794 .25431 m
.56485 .24569 L
s
.71219 .29609 m
.71618 .29512 L
s
.64257 .29347 m
.65126 .29607 L
s
.60281 .27447 m
.60356 .27516 L
s
.54849 .23489 m
.54969 .2438 L
s
.50929 .21808 m
.50865 .22279 L
s
.30846 .1806 m
.33737 .1704 L
s
.30846 .1806 m
.30846 .1806 L
s
.62341 .28505 m
.61678 .28133 L
s
.80217 .23215 m
.79931 .24322 L
s
.59669 .26867 m
.60281 .27447 L
s
.76421 .26696 m
.75698 .27443 L
s
.73959 .28483 m
.74748 .2797 L
s
.45291 .20112 m
.46023 .18985 L
s
.71618 .29512 m
.72532 .29196 L
s
.58305 .26029 m
.58483 .26371 L
s
.837 .19624 m
.83964 .2114 L
s
.53103 .22275 m
.53012 .23178 L
s
.8237 .20373 m
.82713 .22113 L
s
.81382 .22582 m
.81301 .23129 L
s
.63793 .29184 m
.64257 .29347 L
s
.79048 .2418 m
.78718 .25158 L
s
.8692 .15932 m
.88224 .17825 L
s
.88224 .17825 m
.88224 .17825 L
s
.63793 .29184 m
.63168 .28879 L
s
.61606 .28078 m
.61678 .28133 L
s
.58305 .26029 m
.57939 .25445 L
s
.50929 .21808 m
.51262 .20904 L
s
.48516 .20752 m
.48741 .20245 L
s
.61276 .27771 m
.61606 .28078 L
s
.77565 .25894 m
.77731 .25596 L
s
.75146 .27649 m
.74748 .2797 L
s
.59669 .26867 m
.59244 .26144 L
s
.41327 .18449 m
.42589 .17658 L
s
.86024 .1758 m
.8662 .18435 L
s
.62703 .28605 m
.63168 .28879 L
s
.81382 .22582 m
.81294 .22146 L
s
.61276 .27771 m
.60778 .2734 L
s
.76477 .26626 m
.76421 .26696 L
s
.56428 .24364 m
.56485 .24569 L
s
.77731 .25596 m
.78041 .24815 L
s
.54849 .23489 m
.54838 .23126 L
s
.76477 .26626 m
.7661 .26386 L
s
.85054 .19041 m
.85091 .19261 L
s
.75575 .27245 m
.75146 .27649 L
s
.49351 .19362 m
.48741 .20245 L
s
.62135 .28208 m
.62703 .28605 L
s
.57939 .25445 m
.57737 .2458 L
s
.33737 .1704 m
.36654 .16011 L
s
.33737 .1704 m
.33737 .1704 L
s
.53103 .22275 m
.53204 .21756 L
s
.56385 .23522 m
.56428 .24364 L
s
.84462 .17832 m
.85054 .19041 L
s
.76994 .25772 m
.7661 .26386 L
s
.80182 .22301 m
.80217 .23215 L
s
.81294 .22146 m
.81244 .21048 L
s
.46023 .18985 m
.46462 .1847 L
s
.60778 .2734 m
.60701 .27247 L
s
.79048 .2418 m
.79116 .23913 L
s
.54961 .22371 m
.54838 .23126 L
s
.60701 .27247 m
.60135 .26464 L
s
.75974 .26736 m
.75575 .27245 L
s
.51262 .20904 m
.51598 .20187 L
s
.59244 .26144 m
.59078 .25877 L
s
.61806 .27924 m
.62135 .28208 L
s
.43204 .17072 m
.42589 .17658 L
s
.79102 .23576 m
.79116 .23913 L
s
.85542 .16566 m
.86024 .1758 L
s
.53482 .21083 m
.53204 .21756 L
s
.83213 .18644 m
.837 .19624 L
s
.59078 .25877 m
.5903 .25668 L
s
.78173 .24378 m
.78041 .24815 L
s
.47459 .1759 m
.46462 .1847 L
s
.61806 .27924 m
.61201 .2719 L
s
.76994 .25772 m
.77076 .25579 L
s
.49351 .19362 m
.50094 .18441 L
s
.79102 .23576 m
.79176 .22689 L
s
.75974 .26736 m
.76014 .26673 L
s
.5903 .25668 m
.58838 .24785 L
s
.82341 .20209 m
.8237 .20373 L
s
.57737 .2458 m
.57723 .24518 L
s
.56385 .23522 m
.56403 .23204 L
s
.76014 .26673 m
.76167 .2631 L
s
.77076 .25579 m
.77293 .249 L
s
.57735 .23555 m
.57723 .24518 L
s
.54961 .22371 m
.55073 .21931 L
s
.44338 .16329 m
.43204 .17072 L
s
.36654 .16011 m
.39598 .14973 L
s
.36654 .16011 m
.36654 .16011 L
s
.82314 .20157 m
.82341 .20209 L
s
.80182 .22301 m
.80162 .21903 L
s
.51956 .19673 m
.51598 .20187 L
s
.78225 .23455 m
.78173 .24378 L
s
.82314 .20157 m
.82095 .19603 L
s
.8692 .15932 m
.8692 .15932 L
s
.85595 .14006 m
.8692 .15932 L
s
.60029 .26187 m
.60135 .26464 L
s
.56528 .22579 m
.56403 .23204 L
s
.53482 .21083 m
.53806 .20417 L
s
.81014 .20289 m
.81244 .21048 L
s
.84158 .17202 m
.84462 .17832 L
s
.59876 .25477 m
.60029 .26187 L
s
.55326 .21389 m
.55073 .21931 L
s
.61201 .2719 m
.61154 .27125 L
s
.80162 .21903 m
.80004 .2139 L
s
.84847 .15568 m
.85542 .16566 L
s
.60894 .26441 m
.61154 .27125 L
s
.48518 .16704 m
.47459 .1759 L
s
.82909 .17859 m
.83213 .18644 L
s
.51956 .19673 m
.528 .18632 L
s
.77293 .249 m
.77348 .24557 L
s
.82095 .19603 m
.817 .18565 L
s
.79176 .22689 m
.79047 .22047 L
s
.45439 .15661 m
.44338 .16329 L
s
.50094 .18441 m
.5037 .1818 L
s
.58842 .24554 m
.58838 .24785 L
s
.5699 .21592 m
.56528 .22579 L
s
.53806 .20417 m
.5414 .19954 L
s
.84102 .17123 m
.84158 .17202 L
s
.5778 .23333 m
.57735 .23555 L
s
.79852 .20685 m
.80004 .2139 L
s
.55523 .2091 m
.55326 .21389 L
s
.77348 .24557 m
.77404 .24143 L
s
.58994 .23824 m
.58842 .24554 L
s
.83075 .15417 m
.84102 .17123 L
s
.5037 .1818 m
.50837 .17797 L
s
.81014 .20289 m
.8077 .19656 L
s
.78226 .23448 m
.78225 .23455 L
s
.42569 .13925 m
.39598 .14973 L
s
.39598 .14973 m
.39598 .14973 L
s
.78226 .23448 m
.78226 .23443 L
s
.78226 .23443 m
.78114 .22541 L
s
.59842 .25116 m
.59876 .25477 L
s
.58017 .22577 m
.5778 .23333 L
s
.60879 .26367 m
.60894 .26441 L
s
.60841 .2595 m
.60879 .26367 L
s
.82909 .17859 m
.8215 .16642 L
s
.54838 .19067 m
.5414 .19954 L
s
.78952 .21617 m
.79047 .22047 L
s
.50837 .17797 m
.5156 .17082 L
s
.45439 .15661 m
.45918 .15494 L
s
.77404 .24143 m
.77339 .23626 L
s
.48884 .16497 m
.48518 .16704 L
s
.5605 .20201 m
.55523 .2091 L
s
.84337 .14711 m
.84847 .15568 L
s
.59904 .24575 m
.59842 .25116 L
s
.4799 .143 m
.45918 .15494 L
s
.528 .18632 m
.52898 .18536 L
s
.48884 .16497 m
.49935 .15686 L
s
.8077 .19656 m
.80167 .18527 L
s
.817 .18565 m
.81645 .1846 L
s
.79852 .20685 m
.79718 .20383 L
s
.52898 .18536 m
.5312 .18351 L
s
.58994 .23824 m
.59018 .23604 L
s
.81645 .1846 m
.81047 .17383 L
s
.78952 .21617 m
.78549 .2072 L
s
.5699 .21592 m
.57004 .21545 L
s
.57004 .21545 m
.57017 .21523 L
s
.5312 .18351 m
.54066 .17441 L
s
.57017 .21523 m
.57839 .20385 L
s
.58017 .22577 m
.5838 .21936 L
s
.78114 .22541 m
.7793 .22074 L
s
.59018 .23604 m
.59037 .23544 L
s
.85595 .14006 m
.85595 .14006 L
s
.84247 .12048 m
.85595 .14006 L
s
.79718 .20383 m
.79318 .19544 L
s
.59037 .23544 m
.59419 .22696 L
s
.59904 .24575 m
.60128 .23871 L
s
.42569 .13925 m
.45568 .12868 L
s
.42569 .13925 m
.42569 .13925 L
s
.5156 .17082 m
.5199 .1672 L
s
.56536 .19592 m
.5605 .20201 L
s
.55081 .18835 m
.54838 .19067 L
s
.83501 .13509 m
.84337 .14711 L
s
.49935 .15686 m
.51701 .14707 L
s
.7793 .22074 m
.77711 .21528 L
s
.53058 .15973 m
.5199 .1672 L
s
.5581 .18207 m
.55081 .18835 L
s
.8215 .16642 m
.81952 .1628 L
s
.5838 .21936 m
.58638 .21583 L
s
.82757 .14965 m
.83075 .15417 L
s
.78549 .2072 m
.78498 .20592 L
s
.48655 .14067 m
.4799 .143 L
s
.60128 .23871 m
.60187 .23742 L
s
.80167 .18527 m
.80143 .18486 L
s
.57042 .19111 m
.56536 .19592 L
s
.54431 .17124 m
.54066 .17441 L
s
.80048 .1836 m
.80143 .18486 L
s
.78914 .18937 m
.79318 .19544 L
s
.78498 .20592 m
.78479 .20559 L
s
.8041 .16536 m
.81047 .17383 L
s
.59588 .22458 m
.59419 .22696 L
s
.78479 .20559 m
.77903 .19631 L
s
.81774 .16041 m
.81952 .1628 L
s
.60667 .2307 m
.60187 .23742 L
s
.58638 .21583 m
.59374 .20777 L
s
.79282 .17284 m
.80048 .1836 L
s
.48655 .14067 m
.50401 .13111 L
s
.77711 .21528 m
.77352 .20921 L
s
.57839 .20385 m
.58038 .20165 L
s
.48594 .11801 m
.45568 .12868 L
s
.45568 .12868 m
.45568 .12868 L
s
.56245 .17791 m
.5581 .18207 L
s
.55365 .16454 m
.54431 .17124 L
s
.8048 .14399 m
.81774 .16041 L
s
.60063 .21888 m
.59588 .22458 L
s
.81497 .13357 m
.82757 .14965 L
s
.54207 .15223 m
.53058 .15973 L
s
.57042 .19111 m
.58018 .18313 L
s
.58822 .1946 m
.58038 .20165 L
s
.51701 .14707 m
.52053 .14458 L
s
.78914 .18937 m
.78598 .18497 L
s
.82874 .12701 m
.83501 .13509 L
s
.56245 .17791 m
.56378 .17681 L
s
.60687 .23047 m
.60667 .2307 L
s
.77352 .20921 m
.77192 .20667 L
s
.60767 .22965 m
.60687 .23047 L
s
.57636 .16808 m
.56378 .17681 L
s
.52237 .14337 m
.52053 .14458 L
s
.79899 .15892 m
.8041 .16536 L
s
.60458 .21479 m
.60063 .21888 L
s
.59438 .2072 m
.59374 .20777 L
s
.54635 .13151 m
.52237 .14337 L
s
.56792 .15577 m
.55365 .16454 L
s
.59638 .20567 m
.59438 .2072 L
s
.77903 .19631 m
.77744 .19438 L
s
.77192 .20667 m
.7672 .20095 L
s
.78598 .18497 m
.77946 .17785 L
s
.84247 .12048 m
.84247 .12048 L
s
.82876 .10056 m
.84247 .12048 L
s
.50401 .13111 m
.51766 .12633 L
s
.54676 .14993 m
.54207 .15223 L
s
.59638 .20567 m
.60375 .19943 L
s
.59369 .19037 m
.58822 .1946 L
s
.77744 .19438 m
.7717 .18799 L
s
.79282 .17284 m
.79113 .17098 L
s
.58213 .18179 m
.58018 .18313 L
s
.60853 .21147 m
.60458 .21479 L
s
.48594 .11801 m
.51648 .10724 L
s
.48594 .11801 m
.48594 .11801 L
s
.54676 .14993 m
.55857 .14278 L
s
.58916 .17742 m
.58213 .18179 L
s
.78143 .16123 m
.79113 .17098 L
s
.82037 .11498 m
.82874 .12701 L
s
.79125 .15088 m
.79899 .15892 L
s
.53202 .11912 m
.51766 .12633 L
s
.59917 .18677 m
.59369 .19037 L
s
.61634 .20605 m
.60853 .21147 L
s
.58283 .16403 m
.57636 .16808 L
s
.60629 .19766 m
.60375 .19943 L
s
.77654 .17494 m
.77946 .17785 L
s
.76511 .199 m
.7576 .19238 L
s
.58916 .17742 m
.59469 .17375 L
s
.56991 .15457 m
.56792 .15577 L
s
.7717 .18799 m
.76833 .18501 L
s
.812 .12936 m
.81497 .13357 L
s
.80238 .14145 m
.8048 .14399 L
s
.56991 .15457 m
.57173 .15376 L
s
.61396 .19288 m
.60629 .19766 L
s
.60804 .1816 m
.59917 .18677 L
s
.57173 .15376 m
.58734 .14562 L
s
.55857 .14278 m
.57545 .13523 L
s
.77654 .17494 m
.76879 .16855 L
s
.5916 .15958 m
.58283 .16403 L
s
.54635 .13151 m
.54721 .13098 L
s
.76833 .18501 m
.76104 .17911 L
s
.54721 .13098 m
.55086 .12934 L
s
.79826 .13787 m
.80238 .14145 L
s
.59583 .17308 m
.59469 .17375 L
s
.60746 .16703 m
.59583 .17308 L
s
.54679 .11395 m
.53202 .11912 L
s
.79589 .11346 m
.812 .12936 L
s
.57184 .11979 m
.55086 .12934 L
s
.79125 .15088 m
.78444 .14502 L
s
.78143 .16123 m
.78008 .1601 L
s
.54731 .09637 m
.51648 .10724 L
s
.51648 .10724 m
.51648 .10724 L
s
.61865 .19055 m
.61396 .19288 L
s
.78008 .1601 m
.77045 .15295 L
s
.61058 .18028 m
.60804 .1816 L
s
.7576 .19238 m
.75669 .19165 L
s
.5916 .15958 m
.59811 .15629 L
s
.78899 .12859 m
.79826 .13787 L
s
.75669 .19165 m
.74732 .18545 L
s
.62139 .17539 m
.61058 .18028 L
s
.81236 .107 m
.82037 .11498 L
s
.62493 .18758 m
.61865 .19055 L
s
.76879 .16855 m
.76499 .16575 L
s
.76104 .17911 m
.75788 .177 L
s
.55958 .10808 m
.54679 .11395 L
s
.57545 .13523 m
.58094 .13235 L
s
.58734 .14562 m
.5999 .1402 L
s
.59811 .15629 m
.60394 .15363 L
s
.60746 .16703 m
.61115 .16535 L
s
.78444 .14502 m
.78096 .14229 L
s
.75788 .177 m
.74974 .17212 L
s
.63273 .1843 m
.62493 .18758 L
s
.75715 .161 m
.76499 .16575 L
s
.82876 .10056 m
.82876 .10056 L
s
.81481 .08029 m
.82876 .10056 L
s
.62066 .16133 m
.61115 .16535 L
s
.58094 .13235 m
.58534 .13044 L
s
.62291 .17467 m
.62139 .17539 L
s
.6175 .14826 m
.60394 .15363 L
s
.7727 .13647 m
.78096 .14229 L
s
.62321 .17454 m
.62291 .17467 L
s
.77045 .15295 m
.76883 .15169 L
s
.74732 .18545 m
.7467 .18508 L
s
.63678 .16943 m
.62321 .17454 L
s
.7467 .18508 m
.74346 .18347 L
s
.57765 .11777 m
.57184 .11979 L
s
.58534 .13044 m
.60684 .12224 L
s
.63743 .18254 m
.63273 .1843 L
s
.76854 .15149 m
.76883 .15169 L
s
.75827 .14601 m
.76854 .15149 L
s
.57804 .10162 m
.55958 .10808 L
s
.5999 .1402 m
.60037 .14002 L
s
.60037 .14002 m
.61432 .13439 L
s
.57843 .08539 m
.54731 .09637 L
s
.54731 .09637 m
.54731 .09637 L
s
.62962 .15813 m
.62066 .16133 L
s
.74974 .17212 m
.74649 .17049 L
s
.80496 .09639 m
.81236 .107 L
s
.60137 .10856 m
.57765 .11777 L
s
.78317 .12429 m
.78899 .12859 L
s
.75715 .161 m
.75202 .15825 L
s
.6506 .17839 m
.63743 .18254 L
s
.74649 .17049 m
.73746 .1668 L
s
.61829 .14794 m
.6175 .14826 L
s
.61829 .14794 m
.62094 .1471 L
s
.64074 .16818 m
.63678 .16943 L
s
.75202 .15825 m
.74344 .15458 L
s
.76995 .13499 m
.7727 .13647 L
s
.62962 .15813 m
.63646 .15584 L
s
.62094 .1471 m
.63638 .14189 L
s
.65205 .165 m
.64074 .16818 L
s
.79516 .11242 m
.79589 .11346 L
s
.77139 .11725 m
.78317 .12429 L
s
.75354 .14346 m
.75827 .14601 L
s
.76995 .13499 m
.76049 .12926 L
s
.73042 .17826 m
.72344 .17603 L
s
.66259 .17576 m
.65166 .17814 L
s
.61432 .13439 m
.62757 .13016 L
s
.58991 .09698 m
.57804 .10162 L
s
.73746 .1668 m
.73347 .16552 L
s
.79516 .11242 m
.79185 .10986 L
s
.64905 .15242 m
.63646 .15584 L
s
.60684 .12224 m
.61034 .12082 L
s
.73347 .16552 m
.7219 .16247 L
s
.74344 .15458 m
.73615 .15216 L
s
.75354 .14346 m
.7534 .1434 L
s
.75306 .14329 m
.7534 .1434 L
s
.72344 .17603 m
.72188 .17562 L
s
.73842 .13759 m
.75306 .14329 L
s
.65645 .16404 m
.65205 .165 L
s
.66393 .17552 m
.66259 .17576 L
s
.72188 .17562 m
.70873 .17337 L
s
.6777 .17345 m
.66393 .17552 L
s
.61034 .12082 m
.61365 .11975 L
s
.77704 .09826 m
.79185 .10986 L
s
.66829 .16189 m
.65645 .16404 L
s
.63978 .14096 m
.63638 .14189 L
s
.65426 .15116 m
.64905 .15242 L
s
.61365 .11975 m
.64088 .11119 L
s
.73615 .15216 m
.73058 .15063 L
s
.60922 .10584 m
.60137 .10856 L
s
.62757 .13016 m
.6381 .12687 L
s
.65791 .13656 m
.63978 .14096 L
s
.67914 .17333 m
.6777 .17345 L
s
.70873 .17337 m
.70753 .1732 L
s
.7219 .16247 m
.71912 .16199 L
s
.68641 .17294 m
.67914 .17333 L
s
.60953 .09011 m
.58991 .09698 L
s
.70753 .1732 m
.70234 .17287 L
s
.65426 .15116 m
.66776 .14866 L
s
.71912 .16199 m
.70697 .16023 L
s
.74791 .12475 m
.76049 .12926 L
s
.67323 .16114 m
.66829 .16189 L
s
.79269 .0868 m
.80496 .09639 L
s
.60985 .07432 m
.57843 .08539 L
s
.57843 .08539 m
.57843 .08539 L
s
.73058 .15063 m
.72215 .14864 L
s
.6859 .16002 m
.67323 .16114 L
s
.76797 .1146 m
.77139 .11725 L
s
.63635 .09738 m
.60922 .10584 L
s
.70697 .16023 m
.70305 .15998 L
s
.69169 .15971 m
.6859 .16002 L
s
.70305 .15998 m
.69169 .15971 L
s
.6381 .12687 m
.65383 .12297 L
s
.67379 .14783 m
.66776 .14866 L
s
.71391 .14753 m
.72215 .14864 L
s
.72821 .1354 m
.73842 .13759 L
s
.6871 .14661 m
.67379 .14783 L
s
.71391 .14753 m
.70556 .14663 L
s
.65791 .13656 m
.66103 .13598 L
s
.76797 .1146 m
.76063 .11168 L
s
.68071 .13324 m
.66103 .13598 L
s
.70556 .14663 m
.69407 .14644 L
s
.6871 .14661 m
.69407 .14644 L
s
.74791 .12475 m
.7435 .12281 L
s
.72821 .1354 m
.72271 .13411 L
s
.81481 .08029 m
.81481 .08029 L
s
.80061 .05967 m
.81481 .08029 L
s
.72271 .13411 m
.70386 .13255 L
s
.6651 .12072 m
.65383 .12297 L
s
.6231 .08585 m
.60953 .09011 L
s
.73291 .12075 m
.7435 .12281 L
s
.76063 .11168 m
.74884 .10645 L
s
.68231 .1331 m
.68071 .13324 L
s
.68231 .1331 m
.701 .13262 L
s
.64088 .11119 m
.64642 .10981 L
s
.70351 .13253 m
.701 .13262 L
s
.70386 .13255 m
.70351 .13253 L
s
.72381 .1186 m
.73291 .12075 L
s
.65822 .10744 m
.64642 .10981 L
s
.67486 .11925 m
.6651 .12072 L
s
.76799 .09463 m
.77704 .09826 L
s
.78791 .07997 m
.79269 .0868 L
s
.64224 .09533 m
.63635 .09738 L
s
.68954 .11752 m
.67486 .11925 L
s
.6419 .07927 m
.6231 .08585 L
s
.72381 .1186 m
.72017 .1183 L
s
.67041 .10442 m
.65822 .10744 L
s
.72017 .1183 m
.70002 .11711 L
s
.64156 .06314 m
.60985 .07432 L
s
.60985 .07432 m
.60985 .07432 L
s
.68954 .11752 m
.70002 .11711 L
s
.74207 .10513 m
.74884 .10645 L
s
.64224 .09533 m
.66076 .09074 L
s
.75547 .08734 m
.76799 .09463 L
s
.74207 .10513 m
.72592 .10133 L
s
.67041 .10442 m
.69395 .10187 L
s
.76414 .06708 m
.78791 .07997 L
s
.69395 .10187 m
.69751 .1012 L
s
.66076 .09074 m
.67088 .08822 L
s
.7022 .10125 m
.72592 .10133 L
s
.69751 .1012 m
.7022 .10125 L
s
.6419 .07927 m
.65842 .07515 L
s
.67088 .08822 m
.67565 .08771 L
s
.67565 .08771 m
.69974 .08312 L
s
.7319 .08357 m
.75547 .08734 L
s
.65842 .07515 m
.6721 .07036 L
s
.64156 .06314 m
.67357 .05185 L
s
.64156 .06314 m
.64156 .06314 L
s
.72846 .0829 m
.7319 .08357 L
s
.72418 .08337 m
.72846 .0829 L
s
.69974 .08312 m
.72418 .08337 L
s
.80061 .05967 m
.78616 .03869 L
s
.80061 .05967 m
.80061 .05967 L
s
.6721 .07036 m
.69495 .06612 L
s
.76414 .06708 m
.76354 .06623 L
s
.76232 .06587 m
.76354 .06623 L
s
.69495 .06612 m
.70236 .06353 L
s
.76232 .06587 m
.75732 .06537 L
s
.73175 .06037 m
.75732 .06537 L
s
.70236 .06353 m
.71981 .06169 L
s
.67357 .05185 m
.70589 .04045 L
s
.67357 .05185 m
.67357 .05185 L
s
.71981 .06169 m
.73175 .06037 L
s
.70589 .04045 m
.73852 .02894 L
s
.70589 .04045 m
.70589 .04045 L
s
.77146 .01733 m
.78616 .03869 L
s
.78616 .03869 m
.78616 .03869 L
s
.73852 .02894 m
.77146 .01733 L
s
.73852 .02894 m
.73852 .02894 L
s
.25 Mabswid
.78654 0 m
.9906 .3013 L
s
.9906 .3013 m
1 .37651 L
s
1 .37651 m
.79279 .07943 L
s
.79279 .07943 m
.78654 0 L
s
.00963 .27532 m
0 .35096 L
s
0 .35096 m
.79279 .07943 L
s
.79279 .07943 m
.78654 0 L
s
.78654 0 m
.00963 .27532 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{285.812, 168.438},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006/0000o0P10`4000010?ooo`000<c<c;no_`00
E@00ZW=cLiVIVNfXL<V>GoOI`7eI>ooooiVIVH:2P]7AdTi>C_ooohVHZWB1T<[AfDI=E_ooofQ_PUQN
K[ZmaS4e?_oooiBDU:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJF
UWemOFATI4];Bc8b<P0T000TE@0TZP0To`190019E@19ZP19o`1]001]E@1]ZP1]o`2B002BE@2BZP2B
o`2f002fE@2fZP2fo`3K003KE@3KZP3Ko`3oE@3oZUD005D0EED0ZUD0oeDT05DTEEDTZUDToeE905E9
EEE9ZUE9oeE]05E]EEE]ZUE]oeFB05FBEEFBZUFBoeFf05FfEEFfZUFfoeGK05GKEEGKZUGKoeGo05Go
EEGoZUGoojX00:X0EJX0ZZX0ojXT0:XTEJXTZZXTojY90:Y9EJY9ZZY9ojY]0:Y]EJY]ZZY]ojZB0:ZB
EJZBZZZBojZf0:ZfEJZfZZZfoj[K0:[KEJ[KZZ[Koj[o0:[oEJ[oZZ[oool0EOl0Z_lT0?lTEOlTZ_lT
oom90?m9EOm9Z_m9oom]0?m]EOm]Z_m]oonB0?nBEOnBZ_nBoonf0?nfEOnfZ_nfoooK0?oKEOoKZ_oK
ooooEOooZ/d0003=0<g=0000cLd0c@3=cNGUiDe=C@00:P00O`00e00Z000Z:P0ZE@0ZO`0ZZP0Ze01E
001E:P1EE@1EO`1EZP1Ee01o001o:P1oE@1oO`1oZP1oe02Z002Z:P2ZE@2ZO`2ZZP2Ze03D003D:P3D
E@3DO`3DZP3De2X002X0:RX0EBX0ObX0ZRX0e2XZ02XZ:RXZEBXZObXZZRXZe2YE02YE:RYEEBYEObYE
ZRYEe2Yo02Yo:RYoEBYoObYoZRYoe2ZZ02ZZ:RZZEBZZObZZZRZZe2[D02[D:R[DEB[DOb[DZR[De5D0
:UD0OeD0e5DZ05DZ:UDZEEDZOeDZZUDZe5EE05EE:UEEEOl0D0020E/0003o04d00`400`010@1J0000
o`1:00<110020EX0003o04P00P4700030@0105T0003o04D00`4900040@000EP0003o04800`4<0004
0@000EP0003o03l00`4@00040@000EL0003o03`00`4C00050@000040EP000?l0>P020@d00P470005
0@000040EP000?l0=`030@`00`4200030@0000@000<100000P000`40001C0000o`0d00<130030@H0
00<100000`000`40000300030@000580003o03400`4=00812@000`40000300030@0000<000<10000
DP000?l0;P030@/000<100400P4<00030@00008000<1000010000`40001A0000o`0/00813@030A00
00D100000@0700030@000500003o02T00`4;00@150001040004700030@000500003o02H00`4;00<1
60001040004800030@0004l0003o02<00`4:00@170000`400@0900030@0004h0003o02000`4:00<1
8@020@T000<10000CP000?l07@030@/00P4U00812@000`40001=0000o`0K00812`030BL00P4:0003
0@0004`0003o01P00`4:00<1:`000`40000800030@0004`0003o01D00`4;0081;P020@X000<10000
B`000?l04P030@/00`4`00812`000`40001:0000o`0?00<12`030C<000<100402`000`4000190000
o`0=00812`030CH000@100012P000`4000190000o`0:00<12`020CT000@100012`000`4000180000
o`0700<12`030Al01@4G00050@0000402`000`4000170000o`0400<12`030AL02`4500H14@000`40
000200030@0000P000<10000A`000?l00@030@/00`4F00@15P060@/000<100000P000`4000090003
0@0004H0003m00<12`030A@01@4O00030@0000X000<100000P000`40000900030@0004D0003k0081
30020A<0104U00812P000`40000300030@0000P000<10000A@000?P00`4;00<14P030B/00P480003
0@0000<000<100002@000`4000140000m@030@/00`4B00<1<0020@H000<1000010000`4000090003
0@0004<0003b00<130020A40104e008110000`40000500030@0000P000<10000@`000>l00`4<00<1
40030C/00P400`010@0700812P000`4000120000k@020@`00`4@00<1?P030@8000<100001@000`40
000900030@000440003Z00<130020A400P4K01<140030@<000<1004020000`40000800030@000440
003W00<130030A000`4G00H14`050@P00`4700030@0100L000<100002@000`4000100000i0030@`0
0`4@00<14`070Ah00`4200<12`000`400@0700030@0000T000<10000?`000>400`4;00@140030A80
104W00<13`020@P000<1000020000`40000o0000g`020@/00`4A00<14@040BT00P4300813`000`40
000600030@0000P000<10000?P000=`00`4:00<14P020A400`4Z00<11`020@h000<100001@000`40
000900030@0003d0003I00<12P030A800`4@00<1:P030@`00`4<00030@0000D000<1000020000`40
000m0000eP030@X00`4C00813`040A/0204700<14P000`40000900030@0000H000<1000020000`40
000l0000d`030@/00P4B00@13@040AT01P4800T150000`40000900030@0000H000<1000020000`40
000k0000d0030@/00`4@00@13`020AH01`4?00<11@050A000P4900811`000`40000800030@0003/0
003>00812`030A000`4A00814`050A<00`4=00<13`000`40000700030@0000H000<1000020000`40
000j0000b`030@/00P4A00814P020A40104E00<14`020@h000<100001`000`40000600030@0000P0
00<10000>@000<P00`4;00<14@020A800P4@00<15P030AP00`4<008120000`40000500030@0000P0
00<10000>@000<D00`4;00<14P020A400`4>00@15P030Ah00P4;00030@0000H00P4700030@0000P0
00<10000>0000<800`4<00814P030A400P4=00@15`0=0AT000<100002@000`40000600811`000`40
000800030@0003L00030008130030A400`4?00D13@020AH01@4=00L14`020@X000<100001P000`40
000500030@0000P000<10000=`000;d00`4;00<14P020A000P4>00@150060AL0104A00812@000`40
000500811`000`40000800030@0003H0002j00<12P040A<00P4@00813P020A801P400`010@0N00<1
40000`40000700030@0000D00P4700030@0000P000<10000=@000;L00`4:00<15@020A000P4>0081
3`050@@00`4S00@13@020@P000<100001@000`40000500030@0000P000<10000=@000;@00`4:00<1
5@030A000P4>00813`020@H00`4Z00@12`000`40000600030@0000D000<100001@000`4000080003
0@0003@0002b00812`020AH00P4@00<13@030@l00P4500<1<@000`40000900030@0000H000<10000
10020@L000<1000020000`40000c0000[`030@X00`4F008140020@h00P4?00<110030@d04P4F0081
2P000`40000600030@0000@00P4700030@0000L000<10000<`000:`00`4;00815`020A000P4>0081
3P030@@00`4;00D14P080A0000<1000020000`40000600030@0000@00P4600030@0000P000<10000
<P000:T00`4<00815`020A4000<100002`030@d00`4400<12@050Al0104=00030@0000P000<10000
1@000`40000400030@0100H000<1000020000`40000a0000YP030@`00`4G00814@020@`00P4=00<1
10030@P0104X00@12P020@T000<100001@000`40000400811`000`40000700030@000340002S00<1
30030AP00P4A008130020@d00P4400<12@020C000P4:008120000`40000500030@0000@00P470003
0@0000L000<10000<0000:400P4;00@16@020A400P4<00813@020@<00`4800@1<`000`40000:0003
0@0000H000<100001@000`40000300811`000`40000800030@0002l0002N00<12P030A/00P4A0081
30020@d00P4200<120030A@0504@00812`000`40000600030@0000@000<1000010020@L000<10000
1`000`40000_0000V`030@X00`4L00814@020@`00P4=008100@00@412P020A401@4D00<13`020@X0
00<100001@000`40000500030@0000@000<100001@000`40000800030@0002h0002H00<12`020Ad0
0P4A00813@000`40000;00D12P030@d01P4L00H12`020@T000<100001@000`40000500030@0000<0
0P4700030@0000P000<10000;@0009D00`4;00<17@020A400P4=008130040@X00`4>0081:0020@/0
00<100001`000`40000500030@0000D000<100000`000`40000600030@0000P000<10000;00009<0
0P4;00<17`000`40000@00030@0000/00P4=00<12@030@h00`4/00812P000`40000700030@0000D0
00<100001@000`40000300030@0000D000<1000020000`40000/0000T0030@/00P4P00814P000`40
000;00030@0000X00`4:00813@040AL0104F00812@020@P000<1000010000`40000500030@0000@0
00<100001@000`40000800030@0002/0002=00<12`030B000P4C00030@0000/000<1000020040@X0
0P4<00<1400;0@@02P4>00812@020@L000<1000010000`40000500030@0000<000<100001P000`40
000800030@0002X0002:00<12`030B8000<100004P000`40000;00030@0000H00`400`010@0:0003
0@0000X00P4?00@16@030@d00P4900030@0000@000<100001@000`40000400030@0000@000<10000
1@000`40000800030@0002X0002700<12`030B<00P4C00813@000`40000400<10P020@X00P4;0081
3P030B000`4<008120000`40000400030@0000D000<1000010000`40000400030@0000D000<10000
20000`40000Y0000Q@020@/00`4T008150000`40000;008110030@<00P4;00030@0000T00P4=00<1
9P030@/000<100001P000`40000400030@0000@000<1000010000`40000400030@0000H000<10000
20000`40000X0000PP030@X00`4V00030@0001800P4=00040@0000<11@000`40000:00030@0000P0
0P4=0081;0030@T000<100001P000`40000300030@0000D000<1000010000`40000400030@0000H0
00<100001`000`40000X0000O`030@X00`4X00030@0001400P4>00030@01008120000`4000080081
2@020@`00`4?00P11P060@h000<100001`020@L000<100000`000`40000400030@0000@000<10000
10020@L000<1000020000`40000W0000O0030@T0104[00030@00010000<100003@030@X000<10000
20000`40000700812P040@d01@4D00D12P020@T000<100001@000`40000300030@0000@000<10000
10000`40000400030@0000H000<1000020000`40000V0000N@030@T00`4]00814P000`40000;00<1
30000`40000700812@000`40000800813@040Ah00P4:008120000`40000400030@0000@000<10000
0`000`40000500030@0000@000<100001P000`40000700030@0002H0001f00<12@030Bh00P4C0003
0@0000T00`400`01000;00030@0000L000<100001`020@X000<100002P030C0000<100001P000`40
000400030@0000<000<1000010000`40000400030@0000@00P4800030@0000L000<100009@0007@0
0P4:0081;`020A@000<100001`030@<000<100002P000`40000700030@0000L000<1000020020@X0
0`4d00811`000`40000300030@0000@000<100000`000`40000500030@0000@000<100001P000`40
000800030@0002@0001a00<12@030C0000<100004`000`40000600811@000`40000900812@000`40
000800030@0000H00P4:0081;P030@P000<100001@000`40000300030@0000@000<100000`000`40
000400030@0000D000<100001P000`40000700030@0002@0001^00<12P020C8000<100004`000`40
000400<11P000`40000900030@0000P000<100001`020@P000<1000020020C<000<100001P000`40
000400030@0000@000<100000`000`40000300030@0000@000<100001@020@L000<1000020000`40
000S0000J`030@/00P4c00030@0001<000<100000P030@T000<1000020000`40000800030@0000H0
0P4800812@020CH00P4700030@0000@000<100000`000`40000300030@0000<000<100001@000`40
000500811`000`40000800030@000280001X00<12`030C@000<100004`000`4000030@/000<10000
2@000`40000700030@0000H000<100001P020@T00P4j00811P000`40000400030@0000<000<10000
0`000`40000300030@0000@000<100001@020@P000<100001`000`40000R0000IP020@/00`4f0003
0@0001@00`4=00030@0000T000<100001`000`40000600030@0000H000<1000020000`40000l0003
0@0000@000<100000`000`40000300030@0000<000<100000`000`40000400030@0000H00P470003
0@0000P000<100008@0006<00`4:00<1>0000`40000B00<13`000`40000900030@0000P000<10000
1@000`40000600030@0000L00P5000030@0000@000<100000`000`40000300030@0000<000<10000
0`000`40000300030@0000H000<100401`000`40000800030@000200001P00<12P030CX000<10000
40030@8000<100003@000`40000800030@0000P000<100001@000`40000600030@0000H00P530003
0@0000<000<100000`000`40000300030@0000<000<100000`000`40000400030@0000H00P480003
0@0000L000<10000800005d00`4:00<1?0020@l00`4400030@0000d000<100002@000`4000070003
0@0000H000<100001@000`40000600030@0004@000<100000`000`40000300030@0000<000<10000
0`000`40000300030@0000<000<100001P000`400@0700030@0000P000<100007`0005X00`4:00<1
?P000`40000<00<11`000`40000<00030@0000T000<1000020000`40000500030@0000D000<10000
1P000`40001600030@00008000<100000`000`40000300030@0000<000<100000`000`4000040003
0@0000H000<100401`000`40000800030@0001h0001G00<12P030D0000<100002P030@T000<10000
3@000`40000900030@0000L000<100001@000`40000500030@0000H000<100001@000`40000o0003
0@0000<000<100000P000`40000300030@0000<000<100000`000`40000400030@0000H000<10040
20000`40000700030@0001h0001E00812P030D<000<100001`030@/000<100003@000`4000090003
0@0000L000<100001P000`40000400030@0000D00P4700030@00044000<100000P000`4000030003
0@0000<000<100000P000`40000300030@0000D000<100001P000`400@0800030@0000L000<10000
7@0005800`4900<1A@000`40000500<13P000`40000=00030@0000T000<100001P000`4000060003
0@0000D000<1000010000`40000600030@00048000<100000P000`40000300030@0000<000<10000
0P000`40000400030@0000@000<100001P001040004700030@0000P000<10000700004l00`4900<1
A`000`40000300<140000`40000=00030@0000T000<100001`000`40000600030@0000@000<10000
10000`4000050081AP000`40000200030@0000<000<100000P000`40000200030@0000@000<10000
10000`40000600040@000@P000<100001`000`40000L0000C0030@T00`5900030@0000800P4C0003
0@0000d000<100002@000`40000600030@0000H000<1000010000`40000400030@0000D000<10000
A`001@40000100D000<100000P000`40000300030@0000<000<1000010000`40000600050@000040
20000`40000700030@0001/0001900<12@030D`000<100400P4E00030@0000d000<1000020000`40
000700030@0000H000<1000010000`40000400030@0000@000<10000B0000`40000200030@000080
00<100000P000`40000300030@0000<000<100001@000`40000600050@00004020000`4000070003
0@0001X0001700812@030Dh00`4G00030@0000d000<100002@000`40000700030@0000D000<10000
1@000`40000300030@0000@000<10000BP000`40000200050@00004010000`40000300030@0000@0
00<1000010000`40000600810`000`40000500030@0000L000<100006P0004@00`490081CP040AT0
00<100003@000`40000900030@0000L000<100001@000`40000400030@0000@000<100000`000`40
001;00030@00008000D100000@0400030@0000<000<1000010000`40000400030@0000L000D10000
0@0800030@0000L000<100006@0004400`4900<1C@030@8000<1000060000`40000=00030@0000T0
00<100001`000`40000500030@0000@000<100000`000`40000300030@0004`000<100000P001@40
000100@000<100000`000`40000400030@0000D000<100001P000`40000200030@0000D000<10000
20000`40000H0000?P030@X00P5=00<11@000`40000H00030@0000d000<100002@000`4000070003
0@0000D000<100000`000`40000400030@0000<000<10000C@001@40000100<000<100000`000`40
000200030@0000@000<100001@000`40000600030@0000<000<100001@000`40000700030@0001P0
000k00<12`020D`00`4700030@0001P000<100003P000`40000900030@0000H000<100001@000`40
000400030@0000@000<100000`000`40000200030@0004P000D100000@0400030@00008000<10000
0P000`40000400030@0000D000<100001P000`40000300030@0000H000<100001`000`40000G0000
>@020@/00`5;00<12P000`40000H00030@0000h000<100002@000`40000600030@0000D000<10000
10000`40000300030@0000<000<100000`000`40001800050@00004010001@40000100D000<10000
10000`40000500030@0000L000<100000`000`40000500030@0000P000<100005P0003H00`4:00<1
B`030@d000<1000060000`40000>00030@0000T000<100001P000`40000500030@0000@000<10000
0`000`40000300030@00008000<10000B@001@40000100@000D100000@0400030@0000D000<10000
1@000`40000700030@0000@000<100001@000`40000700030@0001H0000c00<12`020D/00`4@0003
0@0001P000<100003P000`40000900030@0000H000<100001@000`40000400030@00008000<10000
0`000`40000300030@0004D000D100000@0300050@00004010000`40000200030@0000D000<10000
1@000`40000800030@0000<000<100001P000`40000700030@0001D0000`00<12`030DX00`4C0003
0@0001P000<100003P000`40000900030@0000H000<100001@000`40000400030@00008000<10000
0`000`40000200030@0004H000D100000@0300050@00004010000`40000200030@0000@000<10000
1P000`40000700030@0000D000<100001@000`40000800030@0001@0000]00<12`030DX00`4F0003
0@0001P000<100003P000`40000900030@0000H000<100001P000`40000300030@00008000<10000
0`000`40000200030@0004L000@100010`001040004500030@00008000<1000010000`4000060003
0@0000L000<100001P000`40000500030@0000L000<10000500002X00`4;00<1BP030AT000<10000
60000`40000?00030@0000T000<100001P000`40000500030@0000<000<100000P000`4000020003
0@00008000<10000B0001040004300040@000@@000<100000`000`40000300030@0000H000<10000
2@000`40000600030@0000@000<1000020000`40000C0000:0020@/00`5;008170000`40000I0003
0@0000h000<100002@000`40000600030@0000D000<100000`000`40000200030@00008000<10000
0P000`40001700050@0000400`001040004400030@00008000<1000010000`40000600030@0000T0
00<100001P000`40000500030@0000P000<100004P0002D00`4:00<1B`030Ah000<100006@000`40
000?00030@0000T000<100001@000`40000500030@0000<000<100000P000`40000200030@000080
00<10000A`001040004300050@0000400`000`40000300030@0000@000<100001P000`4000090003
0@0000L000<100001@000`40000700030@000180000R00<12P030D/00`4Q00030@0001T000<10000
3`000`40000900030@0000H000<1000010000`40000300030@00008000<100000P001@40000104X0
00@100010`001040004400030@00008000<100001@000`40000600030@0000X000<100001`000`40
000400812@000`40000A00007`030@X00`5;00<190000`40000J00030@0000l000<1000020000`40
000700030@0000@000<100000`000`40000200050@0000400`000`40001800040@000@8000D10000
0@0300030@0000<000<1000010000`40000700030@0000X000<100001`000`40000500030@0000P0
00<10000400001`00`4:00<1B`030BL000<100006`000`40000>00030@0000T000<100001P000`40
000400030@0000<000<100000P000`40000200050@000040B0001040004300040@000@@000<10000
0`000`40000400030@0000L000<100002P000`40000800030@0000D000<1000020000`40000?0000
6P020@X00`5;00<1:P000`40000K00030@0000l000<100002@000`40000600030@0000@000<10000
0`000`40000200050@0000400`000`40001600040@000@8000D100000@0300030@0000<000<10000
1@000`40000600030@0000/000<100002@000`40000400030@0000P000<100003`0001L00`4900<1
B`030Bh000<100006`000`40000>00030@0000X000<100001P000`40000400030@0000<000D10000
0@0300050@0000400`000`40001100040@000@8000D100000@0400030@0000<000<1000010000`40
000700030@0000X000<100002P000`40000500030@0000P000<100003P0001@00`4900<1B`030C40
00<1000070000`40000>00030@0000X000<100001@000`40000500030@00008000<100000P001@40
0001008000D100000@1300030@0100<000@1000110000`40000300030@0000D000<100001P000`40
000;00030@0000/000<100001@000`40000800030@0000d0000A00<12P020D/00`4e00030@0001/0
00<100003`000`40000:00030@0000D000<100001@000`40000200030@00008000D100000@020004
0@000D0000<100400P001040004300050@0000401P000`40000400030@0000L000<100002`000`40
000<00030@0000D000<100001`000`40000=00003P030@X00`5:00<1>0000`40000L00030@0000l0
0P4:00030@0000H000<100001@000`40000200050@0000400`001040004200030@0003h00`401@01
000100<000D100000@0600030@0000@000<100001`000`40000<00030@0000`000<100001@000`40
000800030@0000`0000<00812P030DX00`4k00030@0001d000<1000040000`40000800030@0000H0
00<100001@000`40000200050@0000400`001040004200030@0003`00P4200040@000@<000D10000
0@0500811`000`40000700030@0000`000<100003@000`40000500030@0000P000<100002`0000T0
0`4900<1BP030Cl000<100007@000`40000@00030@0000P000<100001P000`40000500030@000080
0P4200810P000`400@0200030@0003`00P400`0100020@<000D100000@0500030@0000H000<10000
1`000`40000=00030@0000h000<100001@000`40000700030@0000/0000600<12P020DX00`530003
0@0001d00P4A00030@0000P00P4700811P020@@000L100000@01008000@10001>`030@0400400@@0
00D100000@0500030@0000H000<100001`000`40000=00030@00010000<1000010000`4000080003
0@0000X0000400812P030DX00P5600030@0001l000<100003P000`40000:00030@0000H00P460003
0@0000800P4200050@010@000`4j00030@0100@10`020@<000<100000`020@L000<1000020000`40
000=00030@00010000<100001@000`40000800030@0000T00004008120020DX00`5900030@0001l0
00<100003P020@/00P4800811@020@@000<100000P4010010@020CP000H10041004300810`020@D0
00<100001P000`40000800030@0000d000<100004P000`40000500030@0000L000<100002@0000@0
00<1004020000`40001500<1C@000`40000O00030@0000l00P4;008120020@D00P4300050@000040
0P0010400@4f00H10`00104000020@D00P4800030@0000P000<100003@000`40000D00030@0000D0
00<100001`000`400008000010001040004800030@0004400`5A00030@0001l000<1000040020@/0
104600811@020@8000@100000P400`010@020C<010400`01000200810`000`40000400030@0000L0
00<1000020000`40000>00030@0001@000<100001@000`40000800030@0000L0000400040@000081
1`000`40000m00<1E@020B000P4B00813@020@H00P45008100<00@400`070Bl0104010010@030@<0
0P4600030@0000H00P4;00030@0000d000<100005P000`400005008120000`40000700000`000`40
000400030@0000D00P4j00<1FP020B000P4B00<130020@H00`44008100<00@402P4[00L100<00@40
10020@H00P4800030@0000X000<100003P000`40000G00030@0000D000<100001`000`4000060000
0`000`40000500030@0000H000<10000=0030El000<100007`020A<00P4<00<11P020@@00`4300P1
:00:0@@00P46008120020@`000<100003P000`40000H00030@0000D000<1000020000`4000050000
0`000`40000600030@0000H000<10000<0030F<000<1000080040A400`4<00@110050@800`4300P1
80090@03004100@00`46008120020@`00P4A00030@0001T000<100001@000`40000700030@0000D0
000300030@0000L000<100001P000`40000/00<1I`020B@00P4B00L12@030@H010400`010@020@<0
2@4G00X10P020@@00P4600<120020@d000<100003`020Ad000<1000010000`40000800030@0000@0
000300030@0000P000<100001P000`40000X00<1K0000`40000S00L14P040@P0104600T10`0=0@T0
2`4200D110020@D00`4900813P000`40000>00817`000`40000500030@0000P000<100000`0000<0
00<100002@000`40000600030@0002@00`5`0081:P040A801`4500H11@0:0@<07@4500@11@020@X0
0P4>008140000`40000P00030@0000D000<100001`000`40000300000`000`40000:00030@0000H0
00<1000080030GD00P4/00H14`070@@0104600@100<00@405P4900@11@040@X00P4>00814@000`40
000R00030@0000@000<100001`020@@0000300030@0000/000<100001P020Ad00`5j00<1;`0=0@d0
1@4300H110080A002P4500@130020@h00P4B00030@0002@000<100000P030@D00`400`0100030000
0P000`40000=008120000`40000H0081P0020CX06P400`010@040@H0404900H13P020@h00P4C0003
0@0002D01@4400@110000`40000200000P000`40000?00030@0000H000<1000050030H@00`4o00/1
00<00@400P4200@10`0:0A001@4@00@13@030A<00P4U00D110030@P000<100000P00008000<10000
40000`40000600030@0001000`6:00L1=P020@H00`4200<10P030@<01P47010140050@h00`4D0081
90030@D0104;00030@000080000200030@00014000<100001P000`40000<00<1U0040B/01`4400@1
0`000`40000200810`020@L01@4L00H14@020AD00P4S00<11@040@l000<100000@00008000<10000
4P000`40000600030@0000P00`6K02/11`040@P000<100000`020@<00`4900H140060AD00P4E0081
8@040@D00`4400030@0000`000<100000@00008000<100004`000`40000600030@0000@00`7=00<1
3@000`400004008110050@X0404H00<14`040B000`4600<120000`40000;00030@00004000020003
0@0001@000<100001P020@0400410Ld00`4@00030@0000H000<100001P040BX0104D00818@030@D0
104;00030@0000/000<100000@00008000<100005@000`40000500<1c@030A@000<100001P000`40
000900<180070AD00`4P00<11@030A0000<100002P000`400001000000<00@0060020@800`430003
0@000<L00`4H00811`020@`01P4D00H16P020Al0104500<150000`40000900030@00004000000`01
000I00<11`000`40003300<170000`4000070081400D0Ad00`4N00<11@040AL000<100002@000`40
0001000000<00@005P030@8000<100001P000`40002o00<180000`4000080081?@020Ah00`4500<1
70000`40000800030@00004000000`01000C00<11P000`40000600030@000;/00`4T00030@0000T0
0`4f00@17@030@D00`4P00030@0000P00@41004000000`01000@00<12P000`40000600030@000;L0
0`4X00030@0000/0104_00<17P030@D00`4T00030@0000L00@41004000000`01000>00813P000`40
000600030@000;<00`4[00030@0000l00P4Z00<17P030@@0104W00030@0000L00@41004000000`01
000;00<14@000`40000600030@000:l00`4_00030@000100104O00L17P030@@00`4/00030@0000H0
0@41004000000`01000800<15@000`40000600030@000:/00`4c008150050AL00`4R00<110030C00
00<100001@010@400@0000030@0000D00`4I00030@0000H000<10000Y`030CL000<100005`0G0B40
104300@1<`000`40000500410@01000000<100000P030Ad00P4700030@000:80104k00030@0004X0
104400<1>0000`40000400410@01000000<100400P4R00030@0000D000<10000W@040D0000<10000
AP040@@00`4l00030@0000<00@41004000020BH000<100001@000`40002H00@1A@000`40001100@1
10040Cl000<1000010010@4000000`01000V00030@0000D00P6E00<1BP000`40000l00@11@030D@0
00<100000`010@40000200030@0002D000<100001P000`40002?00<1CP000`40000h00<11P030DP0
00<100000P010@40000300819P000`40000600030@0008`00P5B00030@0003@00`4600<1B`000`40
000200410@0000D000<1000090000`40000600030@0008P00`5E00030@0003000`4500@1C`001@40
00010040000600030@0002@000<100001P000`40002400<1F0000`40000]00<11@030E@000@10001
0@0000L000<1000090000`40000600030@0008000`5L00030@0002T00`4500<1F0000`400@010000
20000`40000T00811`000`40001l00<1H0000`40000U00<110040E/000<100400@0000T000<10000
9@000`40000500030@0007P00`5T00030@0002400`4400<1H0010@400@4000X00P4V00030@0000D0
00<10000M0030FL000<1000070000`400@020@@00`5T0081000<00030@0002@000<100001@000`40
001_00@1J`020AX0104300@1I`020@003@000`40000T00030@0000D00P5[00@1L0000`40000D00@1
10030F/00P4000h000<1000090000`40000500030@0006H00`5e00030@0000l0104500<1JP040@80
000?00030@0002@000<100001@000`40001R00<1N@000`40000;00<11@040FX00`46000040000`40
000T00030@0000D000<10000GP030G`000<1000020030@D00`5Z00@12@00014000<1000090020@H0
0P5K00<1P0000`40000400<11@030FX00`4=00004P020BH000<100001@000`40001E00<1Q0000`40
00030@D00`5Z00<1400001@000<1000090000`4000050081DP030HP00P4400@1J@040A<0000E0003
0@0002@000<100001@000`40001>0081SP030FX00`4G00005P000`40000T00030@0000D000<10000
B0000`400@020Hd00`5Y00@16P0001L000<1000090000`40000500030@0004H00`6<00@1J@030Ah0
000H00030@0002@000<100001@020D80106<00<1J@040B40000I00819@000`40000600030@0003/0
106=00<1J@030BD0000K00030@0002<000<100001P000`40000g00<1S@040FP0104X000070000`40
000S00811`000`40000c00<1S@030FT00`4/00007@000`40000T00030@0000D000<10000<0020Hd0
0`5Y00<1;`0001h000<1000090000`40000500030@0002X000<100400P6<00<1J0040C80000O0003
0@0002@000<100001@000`40000X00<1R`040FP00`4f000080020BD000<100001@000`40000S00@1
R`030FP0104i00008P000`40000S00030@0000D000<100007P040H`00`5X00<1?@0002<000<10000
8`000`40000500816`030H`0105W00@1@00002@000<100008`000`40000500030@0001D0106<00<1
J0030D@0000U00030@0002<00P4600030@000100106=00<1I`040DL0000V00030@0002@000<10000
10000`40000<00<1S@040FL00`5;00009`000`40000T00030@0000@00P4900<1S@030FP00`5>0000
:0020BD000<100001@000`4000040081S@030FL0105A0000:P000`40000S00030@0000D000<10000
0`6;00@1I`030ED0000[00030@0002<000<100001@020H/00`5W00@1F00002`000<100008`000`40
002>00<1I`030E`0000]00030@0002<000<10000RP030FH0105O0000;P000`40000S00030@0008D0
105V00<1H`0002l00P4T0081PP030FH0105V0000<@000`40000S00030@0007`00`5V00<1JP000380
00<100008`000`40001g00@1IP030Fd0000c00030@0002<000<10000L`030FH0105`0000=0000`40
000S00030@0006l00`5V00<1M00003D000<100008`000`40001Z00@1I@040GL0000f008190000`40
001V00<1IP030G/0000h00030@00028000<10000HP030FD0105n0000>@000`40000R00030@0005h0
0`5U00<1PP0003X000<100008P020EX0105T00@1Q@0003/000<100008`000`40001D00<1I@030HT0
000l00030@0002<000<10000D0030FD00`6<0000?@000`40000S00030@0004/0105T00@1S`0003h0
0P4T00030@0004L00`5U00<1T`00040000<100008P000`40001300<1I0040IH0001100030@000280
00<10000?P040F@00`6J0000@P000`40000R00030@0003X00`5T00@1W@0004<000<100008P020CL0
0`5T00<1X@0004@000<100008`000`40000`00@1H`040J@00015008190000`40000/00<1I0030JP0
001700030@00028000<10000:0030F@00`6[0000B0000`40000R00030@0002@00`5S00@1[P0004T0
00<100008P000`40000O00@1H`030K80001:00030@00028000<100006`030F<0106e0000B`000`40
000R00030@0001L00`5S00<1^@0004`00P4S00030@000180105R00@1_00004h000<100008@020@l0
0`5S00<1`00004l000<100008P000`40000900<1HP040L<0001@00030@00028000<1000010040F80
0`770000D@000`40000R00030@0000<1H`030LX0001B00030@0002800P5R00@1c@0005<000<10000
8@000`40001N00<1d@0005@00P4Q00030@0005X0107D0000EP000`40000N00030@0005L00`7H0000
E`000`40000M00030@0005<0107K0000F0000`40000L00030@0005000`7O0000F@000`40000K0003
0@0004`0107R0000FP000`40000J00030@0004T00`7V0000F`020AT000<10000A`030NT0001M0003
0@0001H000<10000@`040N`0001N00030@0001D000<10000@0030O00001O00030@0001@000<10000
?0040O<0001P00030@0001<000<10000>@030OL0001Q00030@00018000<10000=@040OX0001R0081
4P000`40000b00<1oP0006@000<100003`000`40000^00@1o`020000I@000`40000>00030@0002/0
0`7o00H0001V00030@0000d000<10000:0030Ol02@0006L000<1000030000`40000T00@1o`0<0000
J0000`40000;00030@0002400`7o0100001Y00030@0000X000<100007@040Ol04`0006X00P4:0003
0@0001X00`7o01L0001/00030@0000L000<100005P040Ol06P0006d000<100001@000`40000D00<1
o`0N0000KP000`40000400030@000100107o0240001_00030@0000<000<100003@030Ol09@000700
00<100000P000`40000:00<1o`0X0000L@020@8000<100001P040Ol0:`0007<000<100401@030Ol0
;`0007@00P400`010@020Ol0<P0007D00P7o03H00000\
\>"],
  ImageRangeCache->{{{0, 427.688}, {251.625, 0}} -> {-4.96939*^-6, 0.175538, 
  0.00348576, 0.00348576}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(g = FEMSurfaceNoMesh[points, elements]; \n
    solution = 
      Show[g, \n\t\tAxes \[Rule] True, 
        AxesLabel \[Rule] {"\<x\>", "\<y\>", "\<u\>"}, AspectRatio \[Rule] 1, 
        PlotRange \[Rule] All, ViewPoint \[Rule] {1.582, \(-2.9\), 0.647}]; 
    Show[solution, glevel3D]; \)], "Input"],

Cell[CellGroupData[{

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12  scalefont  setfont
% Scaling calculations
0 1.007 -1.29925 3.74173 [
[.02012 .31927 -3.48107 -10.8 ]
[.02012 .31927 3.71893 0 ]
[.29685 .18904 -3.54904 -10.8 ]
[.29685 .18904 3.65096 0 ]
[.60677 .04321 -3.62554 -10.8 ]
[.60677 .04321 3.57446 0 ]
[.37156 -0.0797 -5.54294 -11.375 ]
[.37156 -0.0797 5.64456 0 ]
[.00732 .7952 -3.72102 0 ]
[.00732 .7952 3.47898 10.8 ]
[.10546 .88495 -3.69692 0 ]
[.10546 .88495 3.50308 10.8 ]
[.19484 .96668 -3.67505 0 ]
[.19484 .96668 3.52495 10.8 ]
[.27656 1.04142 -3.6551 0 ]
[.27656 1.04142 3.5449 10.8 ]
[.14967 1.16054 -5.79094 0 ]
[.14967 1.16054 5.52156 11.375 ]
[-0.00919 .38824 -7.2 -4.71937 ]
[-0.00919 .38824 0 6.08063 ]
[-0.01029 .50705 -21.6 -4.75471 ]
[-0.01029 .50705 0 6.04529 ]
[-0.0114 .62638 -21.6 -4.79022 ]
[-0.0114 .62638 0 6.00978 ]
[-0.01252 .74624 -21.6 -4.82591 ]
[-0.01252 .74624 0 5.97409 ]
[-0.07332 .58937 -11.1875 -5.02518 ]
[-0.07332 .58937 0 6.34982 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.00338 .37369 m
.79826 0 L
s
.01971 .36602 m
.0195 .38939 L
s
[(0)] .02012 .31927 -0.03303 1 Mshowa
.29667 .23581 m
.29658 .25919 L
s
[(2)] .29685 .18904 -0.01416 1 Mshowa
.60686 .08998 m
.6069 .11337 L
s
[(4)] .60677 .04321 .00709 1 Mshowa
.125 Mabswid
.08613 .33479 m
.08602 .34881 L
s
.15438 .3027 m
.15429 .31673 L
s
.22453 .26972 m
.22446 .28375 L
s
.37088 .20092 m
.37084 .21495 L
s
.44725 .16502 m
.44723 .17905 L
s
.52587 .12806 m
.52588 .14209 L
s
.69031 .05075 m
.69036 .06478 L
s
.77634 .01031 m
.77641 .02434 L
s
gsave
.37156 -0.0797 -66.5429 -15.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(x) show
70.187500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
0 .74137 m
.2826 1 L
s
.00774 .74845 m
.00796 .72508 L
s
[(0)] .00732 .7952 .03362 -1 Mshowa
.1058 .8382 m
.10597 .81482 L
s
[(1)] .10546 .88495 .02692 -1 Mshowa
.1951 .91992 m
.19523 .89654 L
s
[(2)] .19484 .96668 .02085 -1 Mshowa
.27675 .99465 m
.27685 .97127 L
s
[(3)] .27656 1.04142 .01531 -1 Mshowa
.125 Mabswid
.02812 .7671 m
.02824 .75308 L
s
.0481 .78539 m
.04822 .77136 L
s
.0677 .80333 m
.06781 .7893 L
s
.08693 .82093 m
.08704 .8069 L
s
.12432 .85515 m
.12442 .84112 L
s
.1425 .87179 m
.1426 .85776 L
s
.16035 .88812 m
.16044 .8741 L
s
.17788 .90417 m
.17797 .89014 L
s
.21201 .9354 m
.21208 .92137 L
s
.22862 .9506 m
.22869 .93657 L
s
.24494 .96554 m
.24501 .95151 L
s
.26099 .98022 m
.26105 .96619 L
s
gsave
.14967 1.16054 -66.7909 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.125000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(y) show
70.312500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
.00338 .37369 m
0 .74137 L
s
.0033 .38239 m
.00954 .37946 L
s
[(0)] -0.00919 .38824 1 -0.12604 Mshowa
.0022 .5015 m
.00845 .49872 L
s
[(0.2)] -0.01029 .50705 1 -0.1195 Mshowa
.0011 .62113 m
.00736 .61851 L
s
[(0.4)] -0.0114 .62638 1 -0.11292 Mshowa
0 .74129 m
.00626 .73882 L
s
[(0.6)] -0.01252 .74624 1 -0.10631 Mshowa
.125 Mabswid
.00303 .41212 m
.00677 .41038 L
s
.00275 .44188 m
.0065 .44017 L
s
.00248 .47167 m
.00623 .46999 L
s
.00193 .53136 m
.00568 .52972 L
s
.00166 .56125 m
.00541 .55963 L
s
.00138 .59118 m
.00513 .58958 L
s
.00083 .65112 m
.00458 .64957 L
s
.00055 .68115 m
.00431 .67962 L
s
.00028 .7112 m
.00403 .7097 L
s
gsave
-0.07332 .58937 -72.1875 -9.02518 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(u) show
70.187500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
.00338 .37369 m
0 .74137 L
s
0 .74137 m
.2826 1 L
s
.2826 1 m
.28388 .67992 L
s
.28388 .67992 m
.00338 .37369 L
s
.79826 0 m
.99671 .40534 L
s
.99671 .40534 m
1 .76812 L
s
1 .76812 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.00338 .37369 m
0 .74137 L
s
0 .74137 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.79826 0 m
.00338 .37369 L
s
.28388 .67992 m
.99671 .40534 L
s
.99671 .40534 m
1 .76812 L
s
1 .76812 m
.2826 1 L
s
.2826 1 m
.28388 .67992 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.21767 .56799 m .23719 .57651 L .21234 .58667 L .41 .805 .986 r
F
.21767 .56799 m .21234 .58667 L .19533 .56796 L F
.2356 .55712 m .26229 .56624 L .23719 .57651 L .409 .805 .985 r
F
.2356 .55712 m .23719 .57651 L .21767 .56799 L .407 .804 .985 r
F
.21767 .56799 m .19533 .56796 L .20803 .55132 L F
.19533 .56796 m .17801 .54892 L .20803 .55132 L .407 .804 .986 r
F
.2356 .55712 m .21767 .56799 L .20803 .55132 L .408 .804 .986 r
F
.28763 .55588 m .26229 .56624 L .25995 .5459 L .406 .805 .984 r
F
.25995 .5459 m .26229 .56624 L .2356 .55712 L .404 .804 .984 r
F
.2356 .55712 m .22875 .53405 L .25995 .5459 L .405 .804 .985 r
F
.22875 .53405 m .2356 .55712 L .20803 .55132 L .404 .803 .985 r
F
.28638 .53588 m .31322 .54541 L .28763 .55588 L .403 .805 .983 r
F
.28763 .55588 m .25995 .5459 L .28638 .53588 L .401 .803 .983 r
F
.22875 .53405 m .20803 .55132 L .19313 .53118 L .405 .803 .985 r
F
.20803 .55132 m .17801 .54892 L .19313 .53118 L F
.17801 .54892 m .16038 .52953 L .19313 .53118 L .405 .802 .985 r
F
.31252 .52591 m .33907 .53484 L .31322 .54541 L .399 .805 .981 r
F
.31252 .52591 m .31322 .54541 L .28638 .53588 L .396 .803 .981 r
F
.25944 .52536 m .28638 .53588 L .25995 .5459 L .402 .803 .983 r
F
.25944 .52536 m .25995 .5459 L .22875 .53405 L .4 .802 .983 r
F
.2859 .51607 m .31252 .52591 L .28638 .53588 L .397 .803 .982 r
F
.2859 .51607 m .28638 .53588 L .25944 .52536 L .396 .801 .982 r
F
.36518 .52417 m .33907 .53484 L .3386 .51588 L .393 .804 .979 r
F
.3386 .51588 m .33907 .53484 L .31252 .52591 L .39 .801 .979 r
F
.23694 .51413 m .25944 .52536 L .22875 .53405 L .401 .802 .984 r
F
.23694 .51413 m .22875 .53405 L .20878 .51206 L .4 .801 .984 r
F
.19313 .53118 m .20878 .51206 L .22875 .53405 L .402 .802 .984 r
F
.19313 .53118 m .17507 .51029 L .20878 .51206 L .402 .802 .985 r
F
.17507 .51029 m .19313 .53118 L .16038 .52953 L .403 .802 .985 r
F
.17507 .51029 m .16038 .52953 L .14242 .50977 L F
.69079 .50771 m .7088 .51518 L .68086 .52594 L .37 .8 .968 r
F
.69079 .50771 m .68086 .52594 L .66769 .50613 L .367 .782 .979 r
F
.3386 .51588 m .31252 .52591 L .31218 .50669 L .392 .802 .98 r
F
.31218 .50669 m .31252 .52591 L .2859 .51607 L .389 .8 .98 r
F
.36456 .5058 m .39155 .51338 L .36518 .52417 L .386 .803 .976 r
F
.36518 .52417 m .3386 .51588 L .36456 .5058 L .381 .799 .975 r
F
.25941 .50524 m .2859 .51607 L .25944 .52536 L .397 .801 .982 r
F
.25941 .50524 m .25944 .52536 L .23694 .51413 L .396 .8 .982 r
F
.71364 .4986 m .73701 .5043 L .7088 .51518 L .306 .78 .929 r
F
.71364 .4986 m .7088 .51518 L .69079 .50771 L .287 .759 .937 r
F
.36456 .5058 m .3386 .51588 L .33812 .4973 L .384 .8 .977 r
F
.3386 .51588 m .31218 .50669 L .33812 .4973 L .381 .797 .977 r
F
.28593 .4965 m .31218 .50669 L .2859 .51607 L .392 .8 .981 r
F
.28593 .4965 m .2859 .51607 L .25941 .50524 L .39 .799 .981 r
F
.41818 .50249 m .39155 .51338 L .39025 .49561 L .375 .801 .97 r
F
.39025 .49561 m .39155 .51338 L .36456 .5058 L .369 .796 .97 r
F
.25941 .50524 m .23694 .51413 L .22888 .49313 L .398 .801 .983 r
F
.23694 .51413 m .20878 .51206 L .22888 .49313 L F
.22888 .49313 m .20878 .51206 L .19035 .48837 L .399 .8 .984 r
F
.17507 .51029 m .19035 .48837 L .20878 .51206 L .401 .801 .984 r
F
.17507 .51029 m .14242 .50977 L .15554 .48985 L .402 .801 .985 r
F
.15554 .48985 m .14242 .50977 L .12412 .48965 L F
.17507 .51029 m .15554 .48985 L .19035 .48837 L .401 .801 .984 r
F
.68306 .49472 m .69079 .50771 L .66769 .50613 L .285 .753 .942 r
F
.66769 .50613 m .65427 .48594 L .68306 .49472 L .304 .747 .964 r
F
.71364 .4986 m .69079 .50771 L .68306 .49472 L .298 .762 .946 r
F
.76552 .49332 m .73701 .5043 L .74542 .49262 L .239 .748 .875 r
F
.74542 .49262 m .73701 .5043 L .71364 .4986 L .234 .739 .894 r
F
.31207 .48777 m .33812 .4973 L .31218 .50669 L .384 .798 .978 r
F
.31207 .48777 m .31218 .50669 L .28593 .4965 L .383 .797 .978 r
F
.39025 .49561 m .36456 .5058 L .36342 .48797 L .373 .797 .972 r
F
.36456 .5058 m .33812 .4973 L .36342 .48797 L .369 .794 .972 r
F
.28593 .4965 m .25941 .50524 L .25976 .48516 L .393 .799 .982 r
F
.25976 .48516 m .25941 .50524 L .22888 .49313 L .392 .799 .982 r
F
.41582 .48507 m .44509 .49149 L .41818 .50249 L .36 .798 .962 r
F
.41818 .50249 m .39025 .49561 L .41582 .48507 L .35 .791 .961 r
F
.79431 .48222 m .76552 .49332 L .77274 .4844 L .197 .724 .837 r
F
.76552 .49332 m .74542 .49262 L .77274 .4844 L .194 .72 .849 r
F
.36342 .48797 m .33812 .4973 L .33753 .47919 L .374 .795 .975 r
F
.33753 .47919 m .33812 .4973 L .31207 .48777 L .372 .793 .974 r
F
.31207 .48777 m .28593 .4965 L .28634 .47719 L .386 .797 .98 r
F
.28634 .47719 m .28593 .4965 L .25976 .48516 L .386 .796 .98 r
F
.41582 .48507 m .39025 .49561 L .3876 .47846 L .357 .792 .966 r
F
.39025 .49561 m .36342 .48797 L .3876 .47846 L .352 .789 .965 r
F
.47227 .48037 m .44509 .49149 L .4421 .47533 L .336 .791 .948 r
F
.4421 .47533 m .44509 .49149 L .41582 .48507 L .321 .78 .946 r
F
.22888 .49313 m .23653 .47216 L .25976 .48516 L .395 .799 .983 r
F
.21064 .47333 m .23653 .47216 L .22888 .49313 L .395 .798 .983 r
F
.21064 .47333 m .22888 .49313 L .19035 .48837 L .398 .8 .984 r
F
.13594 .46996 m .15554 .48985 L .12412 .48965 L .402 .801 .985 r
F
.13594 .46996 m .12412 .48965 L .10548 .46916 L F
.15554 .48985 m .16548 .47023 L .19035 .48837 L .401 .801 .985 r
F
.15554 .48985 m .13594 .46996 L .16548 .47023 L F
.72964 .48925 m .74542 .49262 L .71364 .4986 L .173 .705 .844 r
F
.72964 .48925 m .71364 .4986 L .70078 .48873 L .19 .709 .877 r
F
.70078 .48873 m .71364 .4986 L .68306 .49472 L .168 .694 .868 r
F
.21064 .47333 m .19035 .48837 L .18827 .46843 L .399 .8 .984 r
F
.18827 .46843 m .19035 .48837 L .16548 .47023 L F
.70078 .48873 m .68306 .49472 L .67115 .48139 L .199 .703 .905 r
F
.68306 .49472 m .65427 .48594 L .67115 .48139 L .175 .688 .896 r
F
.67115 .48139 m .65427 .48594 L .6406 .46537 L .212 .691 .934 r
F
.33753 .47919 m .31207 .48777 L .31196 .4691 L .377 .794 .977 r
F
.31196 .4691 m .31207 .48777 L .28634 .47719 L .376 .793 .977 r
F
.3876 .47846 m .36342 .48797 L .36179 .47112 L .36 .79 .969 r
F
.36179 .47112 m .36342 .48797 L .33753 .47919 L .358 .788 .969 r
F
.8234 .47101 m .79431 .48222 L .80081 .47447 L .176 .711 .816 r
F
.80081 .47447 m .79431 .48222 L .77274 .4844 L .174 .709 .822 r
F
.25976 .48516 m .26146 .4657 L .28634 .47719 L .389 .797 .981 r
F
.25976 .48516 m .23653 .47216 L .26146 .4657 L F
.46871 .46489 m .49974 .46914 L .47227 .48037 L .301 .779 .923 r
F
.47227 .48037 m .4421 .47533 L .46871 .46489 L .278 .762 .921 r
F
.77274 .4844 m .74542 .49262 L .75404 .48596 L .162 .7 .818 r
F
.72964 .48925 m .75404 .48596 L .74542 .49262 L .164 .699 .839 r
F
.4421 .47533 m .41582 .48507 L .41123 .46652 L .333 .784 .954 r
F
.41582 .48507 m .3876 .47846 L .41123 .46652 L .326 .779 .953 r
F
.36179 .47112 m .33753 .47919 L .33707 .46141 L .365 .79 .973 r
F
.33753 .47919 m .31196 .4691 L .33707 .46141 L .364 .789 .973 r
F
.28634 .47719 m .28641 .45801 L .31196 .4691 L .382 .794 .979 r
F
.28634 .47719 m .26146 .4657 L .28641 .45801 L F
.82998 .46291 m .8528 .45968 L .8234 .47101 L .183 .715 .823 r
F
.8234 .47101 m .80081 .47447 L .82998 .46291 L .184 .716 .821 r
F
.41123 .46652 m .3876 .47846 L .38238 .46443 L .338 .781 .962 r
F
.38238 .46443 m .3876 .47846 L .36179 .47112 L .34 .782 .962 r
F
.22006 .46039 m .23653 .47216 L .21064 .47333 L .396 .799 .983 r
F
.22006 .46039 m .21064 .47333 L .19977 .45655 L .396 .798 .984 r
F
.18827 .46843 m .19977 .45655 L .21064 .47333 L .398 .8 .984 r
F
.77641 .47865 m .80081 .47447 L .77274 .4844 L .145 .689 .792 r
F
.77274 .4844 m .75404 .48596 L .77641 .47865 L .134 .68 .799 r
F
.16548 .47023 m .13594 .46996 L .14768 .45096 L .402 .801 .985 r
F
.14768 .45096 m .13594 .46996 L .11746 .44923 L F
.13594 .46996 m .10548 .46916 L .11746 .44923 L .403 .801 .985 r
F
.11746 .44923 m .10548 .46916 L .08649 .44827 L .403 .802 .985 r
F
.50071 .45832 m .52748 .4578 L .49974 .46914 L .24 .75 .873 r
F
.50071 .45832 m .49974 .46914 L .46871 .46489 L .213 .728 .877 r
F
.17528 .45309 m .18827 .46843 L .16548 .47023 L .401 .8 .985 r
F
.16548 .47023 m .14768 .45096 L .17528 .45309 L F
.23826 .45427 m .26146 .4657 L .23653 .47216 L .392 .797 .983 r
F
.23826 .45427 m .23653 .47216 L .22006 .46039 L F
.46871 .46489 m .4421 .47533 L .43915 .46232 L .298 .769 .936 r
F
.4421 .47533 m .41123 .46652 L .43915 .46232 L .287 .761 .934 r
F
.17528 .45309 m .19977 .45655 L .18827 .46843 L .399 .799 .984 r
F
.72964 .48925 m .70078 .48873 L .71784 .48671 L .054 .62 .764 r
F
.71784 .48671 m .70078 .48873 L .68902 .48284 L .078 .635 .803 r
F
.70078 .48873 m .67115 .48139 L .68902 .48284 L .027 .595 .775 r
F
.67115 .48139 m .65898 .47062 L .68902 .48284 L .066 .616 .83 r
F
.67115 .48139 m .6406 .46537 L .65898 .47062 L .01 .574 .799 r
F
.65898 .47062 m .6406 .46537 L .62666 .44441 L .081 .602 .878 r
F
.74589 .48497 m .75404 .48596 L .72964 .48925 L .081 .641 .763 r
F
.74589 .48497 m .72964 .48925 L .71784 .48671 L .094 .65 .786 r
F
.38238 .46443 m .36179 .47112 L .36178 .45485 L .35 .784 .967 r
F
.36179 .47112 m .33707 .46141 L .36178 .45485 L .348 .783 .967 r
F
.31196 .4691 m .31155 .45034 L .33707 .46141 L .372 .79 .976 r
F
.31196 .4691 m .28641 .45801 L .31155 .45034 L F
.28641 .45801 m .26146 .4657 L .26102 .44641 L .387 .795 .981 r
F
.23826 .45427 m .26102 .44641 L .26146 .4657 L F
.74589 .48497 m .77641 .47865 L .75404 .48596 L .121 .671 .782 r
F
.88249 .44824 m .8528 .45968 L .85927 .44978 L .216 .736 .854 r
F
.85927 .44978 m .8528 .45968 L .82998 .46291 L .22 .74 .847 r
F
.8081 .46744 m .82998 .46291 L .80081 .47447 L .159 .7 .797 r
F
.8081 .46744 m .80081 .47447 L .77641 .47865 L .158 .699 .801 r
F
.52935 .45294 m .50071 .45832 L .50351 .45184 L .139 .682 .818 r
F
.52935 .45294 m .50351 .45184 L .5197 .45284 L .02 0 0 r
F
.52935 .45294 m .53908 .45305 L .52748 .4578 L .146 .69 .781 r
F
.50071 .45832 m .52935 .45294 L .52748 .4578 L .113 .664 .791 r
F
.52935 .45294 m .5197 .45284 L .53151 .45356 L .02 0 0 r
F
.52935 .45294 m .55551 .44633 L .53908 .45305 L .146 .69 .781 r
F
.23826 .45427 m .22006 .46039 L .21503 .445 L .395 .798 .984 r
F
.21503 .445 m .22006 .46039 L .19977 .45655 L .396 .799 .984 r
F
.41123 .46652 m .42207 .44857 L .43915 .46232 L .298 .765 .941 r
F
.38869 .45096 m .42207 .44857 L .41123 .46652 L .297 .76 .946 r
F
.41123 .46652 m .38238 .46443 L .38869 .45096 L .329 .777 .957 r
F
.18946 .43845 m .21503 .445 L .19977 .45655 L .397 .798 .984 r
F
.19977 .45655 m .17528 .45309 L .18946 .43845 L .4 .8 .985 r
F
.38238 .46443 m .36178 .45485 L .38869 .45096 L .327 .773 .959 r
F
.50071 .45832 m .46871 .46489 L .47556 .45323 L .224 .734 .886 r
F
.4513 .4549 m .47556 .45323 L .46871 .46489 L .217 .724 .896 r
F
.4513 .4549 m .46871 .46489 L .43915 .46232 L .252 .746 .911 r
F
.33707 .46141 m .33684 .44337 L .36178 .45485 L .358 .785 .972 r
F
.33707 .46141 m .31155 .45034 L .33684 .44337 L F
.28551 .43846 m .31155 .45034 L .28641 .45801 L .379 .791 .979 r
F
.28641 .45801 m .26102 .44641 L .28551 .43846 L .38 .792 .979 r
F
.17528 .45309 m .16146 .4335 L .18946 .43845 L .4 .8 .985 r
F
.17528 .45309 m .14768 .45096 L .16146 .4335 L .403 .801 .985 r
F
.23826 .45427 m .23525 .43504 L .26102 .44641 L .391 .796 .983 r
F
.23826 .45427 m .21503 .445 L .23525 .43504 L .393 .797 .983 r
F
.16146 .4335 m .14768 .45096 L .13083 .42926 L .403 .801 .985 r
F
.14768 .45096 m .11746 .44923 L .13083 .42926 L .404 .802 .986 r
F
.13083 .42926 m .11746 .44923 L .09856 .4272 L .404 .802 .985 r
F
.08649 .44827 m .09856 .4272 L .11746 .44923 L .405 .802 .986 r
F
.09856 .4272 m .08649 .44827 L .06713 .42698 L .405 .802 .985 r
F
.9125 .43667 m .88249 .44824 L .88959 .43516 L .269 .764 .9 r
F
.88249 .44824 m .85927 .44978 L .88959 .43516 L .275 .771 .889 r
F
.85927 .44978 m .82998 .46291 L .83573 .45437 L .2 .728 .828 r
F
.8081 .46744 m .83573 .45437 L .82998 .46291 L .202 .729 .825 r
F
.4513 .4549 m .43915 .46232 L .42207 .44857 L .269 .749 .931 r
F
.58384 .43475 m .55551 .44633 L .55746 .45008 L 0 0 0 r
F
.55746 .45008 m .55551 .44633 L .52935 .45294 L .046 0 0 r
F
.71784 .48671 m .68902 .48284 L .70662 .48767 L .088 0 0 r
F
.70662 .48767 m .68902 .48284 L .67812 .48168 L .068 0 0 r
F
.65898 .47062 m .67812 .48168 L .68902 .48284 L .159 0 0 r
F
.65898 .47062 m .64796 .46566 L .67812 .48168 L .122 0 0 r
F
.62666 .44441 m .64796 .46566 L .65898 .47062 L .246 0 0 r
F
.74589 .48497 m .71784 .48671 L .73444 .4871 L .025 0 0 r
F
.73444 .4871 m .71784 .48671 L .70662 .48767 L .015 0 0 r
F
.62666 .44441 m .61246 .42304 L .64796 .46566 L 0 .451 .766 r
F
.79061 .47197 m .8081 .46744 L .77641 .47865 L .103 .658 .744 r
F
.79061 .47197 m .77641 .47865 L .76245 .48143 L .104 .659 .745 r
F
.76245 .48143 m .77641 .47865 L .74908 .48431 L .042 .611 .709 r
F
.76245 .48143 m .74981 .4881 L .7762 .47847 L 0 0 0 r
F
.76245 .48143 m .7762 .47847 L .77681 .47825 L F
.76245 .48143 m .74908 .48431 L .74589 .48497 L .042 .611 .709 r
F
.76245 .48143 m .74589 .48497 L .73444 .4871 L .043 .612 .711 r
F
.50071 .45832 m .47556 .45323 L .50351 .45184 L .121 .666 .824 r
F
.36178 .45485 m .36288 .43762 L .38869 .45096 L .34 .777 .966 r
F
.36178 .45485 m .33684 .44337 L .36288 .43762 L .339 .776 .966 r
F
.31069 .43105 m .33684 .44337 L .31155 .45034 L .368 .786 .977 r
F
.31069 .43105 m .31155 .45034 L .28551 .43846 L .37 .788 .977 r
F
.20758 .42467 m .23525 .43504 L .21503 .445 L .395 .797 .984 r
F
.18946 .43845 m .20758 .42467 L .21503 .445 L .398 .799 .985 r
F
.26102 .44641 m .25893 .42606 L .28551 .43846 L .386 .793 .982 r
F
.26102 .44641 m .23525 .43504 L .25893 .42606 L .388 .795 .982 r
F
.94282 .42499 m .9125 .43667 L .92005 .42062 L .329 .788 .944 r
F
.92005 .42062 m .9125 .43667 L .88959 .43516 L .335 .799 .932 r
F
.8653 .43729 m .88959 .43516 L .85927 .44978 L .263 .766 .879 r
F
.8653 .43729 m .85927 .44978 L .83573 .45437 L .269 .771 .871 r
F
.17817 .41749 m .20758 .42467 L .18946 .43845 L .399 .798 .985 r
F
.16146 .4335 m .17817 .41749 L .18946 .43845 L .403 .8 .986 r
F
.61246 .42304 m .58384 .43475 L .5871 .45572 L .216 0 0 r
F
.58384 .43475 m .55746 .45008 L .5871 .45572 L .287 0 0 r
F
.81514 .46116 m .83573 .45437 L .8081 .46744 L .162 .702 .784 r
F
.81514 .46116 m .8081 .46744 L .79061 .47197 L .162 .702 .783 r
F
.38869 .45096 m .39092 .43294 L .42207 .44857 L .314 .766 .956 r
F
.38869 .45096 m .36288 .43762 L .39092 .43294 L .312 .764 .956 r
F
.47556 .45323 m .4513 .4549 L .45083 .44403 L .236 .733 .91 r
F
.4513 .4549 m .42207 .44857 L .45083 .44403 L .229 .728 .909 r
F
.55746 .45008 m .52935 .45294 L .53151 .45356 L 0 0 0 r
F
.16146 .4335 m .14804 .41148 L .17817 .41749 L .403 .8 .986 r
F
.16146 .4335 m .13083 .42926 L .14804 .41148 L .405 .802 .986 r
F
.3362 .42444 m .36288 .43762 L .33684 .44337 L .353 .779 .973 r
F
.3362 .42444 m .33684 .44337 L .31069 .43105 L .355 .781 .973 r
F
.28406 .41811 m .31069 .43105 L .28551 .43846 L .378 .789 .98 r
F
.28406 .41811 m .28551 .43846 L .25893 .42606 L .381 .791 .981 r
F
.23082 .4131 m .25893 .42606 L .23525 .43504 L .392 .795 .984 r
F
.20758 .42467 m .23082 .4131 L .23525 .43504 L .396 .797 .984 r
F
.14804 .41148 m .13083 .42926 L .11228 .40274 L .405 .802 .986 r
F
.13083 .42926 m .09856 .4272 L .11228 .40274 L .407 .802 .986 r
F
.47843 .44485 m .50351 .45184 L .47556 .45323 L .164 .692 .862 r
F
.45083 .44403 m .47843 .44485 L .47556 .45323 L .154 .683 .865 r
F
.09856 .4272 m .06713 .42698 L .07742 .40603 L .407 .803 .986 r
F
.0474 .40528 m .07742 .40603 L .06713 .42698 L .407 .804 .986 r
F
.42207 .44857 m .42192 .42805 L .45083 .44403 L .266 .742 .936 r
F
.42207 .44857 m .39092 .43294 L .42192 .42805 L .265 .742 .936 r
F
.07742 .40603 m .11228 .40274 L .09856 .4272 L .407 .803 .986 r
F
.70662 .48767 m .67812 .48168 L .69623 .49363 L .245 0 0 r
F
.69623 .49363 m .67812 .48168 L .66908 .48765 L .24 0 0 r
F
.66908 .48765 m .67812 .48168 L .64796 .46566 L .354 0 0 r
F
.70662 .48767 m .72306 .49351 L .73444 .4871 L .154 0 0 r
F
.70662 .48767 m .69623 .49363 L .72306 .49351 L .15 0 0 r
F
.94601 .40968 m .97346 .41318 L .94282 .42499 L .377 .801 .972 r
F
.92005 .42062 m .94601 .40968 L .94282 .42499 L .392 .816 .968 r
F
.73444 .4871 m .74981 .4881 L .76245 .48143 L .066 0 0 r
F
.73444 .4871 m .72306 .49351 L .74981 .4881 L F
.66908 .48765 m .64796 .46566 L .64205 .4761 L .371 0 0 r
F
.64205 .4761 m .64796 .46566 L .61246 .42304 L .543 0 0 r
F
.8982 .41726 m .92005 .42062 L .88959 .43516 L .336 .8 .933 r
F
.88959 .43516 m .8653 .43729 L .8982 .41726 L .341 .809 .919 r
F
.83677 .44623 m .8653 .43729 L .83573 .45437 L .227 .746 .83 r
F
.83677 .44623 m .83573 .45437 L .81514 .46116 L F
.77681 .47825 m .79061 .47197 L .76245 .48143 L 0 0 0 r
F
.64205 .4761 m .61246 .42304 L .61895 .47259 L .618 .126 .061 r
F
.61895 .47259 m .61246 .42304 L .5871 .45572 L .554 .092 .111 r
F
.36199 .41863 m .39092 .43294 L .36288 .43762 L .332 .769 .967 r
F
.36199 .41863 m .36288 .43762 L .3362 .42444 L .335 .772 .967 r
F
.23082 .4131 m .20758 .42467 L .19734 .39831 L .398 .797 .985 r
F
.19734 .39831 m .20758 .42467 L .17817 .41749 L .402 .799 .986 r
F
.5871 .45572 m .55746 .45008 L .55748 .45956 L .186 0 0 r
F
.55746 .45008 m .53151 .45356 L .55748 .45956 L .201 0 0 r
F
.80293 .46654 m .79061 .47197 L .80443 .46462 L .115 .667 .736 r
F
.77681 .47825 m .80443 .46462 L .79061 .47197 L .106 .659 .715 r
F
.78953 .47412 m .80443 .46462 L .77681 .47825 L 0 0 0 r
F
.78953 .47412 m .80605 .46255 L .80443 .46462 L .153 .692 .718 r
F
.80605 .46255 m .81732 .45464 L .80443 .46462 L F
.81514 .46116 m .80293 .46654 L .80443 .46462 L .115 .667 .736 r
F
.30947 .41094 m .3362 .42444 L .31069 .43105 L .367 .783 .978 r
F
.28406 .41811 m .30947 .41094 L .31069 .43105 L .37 .786 .978 r
F
.257 .40453 m .28406 .41811 L .25893 .42606 L .387 .792 .983 r
F
.23082 .4131 m .257 .40453 L .25893 .42606 L .391 .794 .983 r
F
.49252 .43899 m .50459 .44475 L .50572 .44816 L 0 .549 .762 r
F
.50572 .44816 m .53151 .45356 L .50702 .45206 L .051 .616 .776 r
F
.50572 .44816 m .53571 .45753 L .53151 .45356 L .115 0 0 r
F
.50459 .44475 m .51924 .45173 L .50572 .44816 L 0 .549 .762 r
F
.50572 .44816 m .50702 .45206 L .50351 .45184 L .051 .616 .776 r
F
.47843 .44485 m .50572 .44816 L .50351 .45184 L .04 .606 .779 r
F
.83677 .44623 m .81514 .46116 L .80443 .46462 L .184 .717 .774 r
F
.16618 .39833 m .19734 .39831 L .17817 .41749 L .403 .799 .986 r
F
.14804 .41148 m .16618 .39833 L .17817 .41749 L .406 .801 .987 r
F
.9314 .3994 m .94601 .40968 L .92005 .42062 L .4 .818 .972 r
F
.8982 .41726 m .9314 .3994 L .92005 .42062 L .4 .831 .957 r
F
.94601 .40968 m .96244 .39121 L .97346 .41318 L .44 .821 .99 r
F
.45083 .44403 m .45571 .43335 L .47843 .44485 L .2 .708 .901 r
F
.45083 .44403 m .42192 .42805 L .45571 .43335 L .195 .702 .903 r
F
.87249 .42034 m .8982 .41726 L .8653 .43729 L .322 .801 .905 r
F
.87249 .42034 m .8653 .43729 L .83677 .44623 L .329 .809 .892 r
F
.42192 .42805 m .39092 .43294 L .38738 .41303 L .299 .752 .957 r
F
.38738 .41303 m .39092 .43294 L .36199 .41863 L .309 .759 .958 r
F
.14804 .41148 m .14017 .38741 L .16618 .39833 L .406 .801 .987 r
F
.14804 .41148 m .11228 .40274 L .14017 .38741 L .409 .803 .987 r
F
.33472 .40448 m .36199 .41863 L .3362 .42444 L .352 .775 .975 r
F
.30947 .41094 m .33472 .40448 L .3362 .42444 L .357 .779 .975 r
F
.28406 .41811 m .28276 .39705 L .30947 .41094 L .38 .787 .982 r
F
.257 .40453 m .28276 .39705 L .28406 .41811 L .384 .79 .982 r
F
.22946 .38942 m .257 .40453 L .23082 .4131 L .396 .795 .985 r
F
.23082 .4131 m .19734 .39831 L .22946 .38942 L .4 .797 .986 r
F
.9314 .3994 m .96244 .39121 L .94601 .40968 L .423 .829 .977 r
F
.69623 .49363 m .71234 .50487 L .72306 .49351 L .283 0 0 r
F
.68706 .50544 m .71234 .50487 L .69623 .49363 L .284 0 0 r
F
.69623 .49363 m .66908 .48765 L .68706 .50544 L .374 0 0 r
F
.53571 .45753 m .55748 .45956 L .53151 .45356 L .104 0 0 r
F
.74981 .4881 m .72306 .49351 L .7377 .49928 L .186 0 0 r
F
.72306 .49351 m .71234 .50487 L .7377 .49928 L .188 0 0 r
F
.66908 .48765 m .66216 .50131 L .68706 .50544 L .38 0 0 r
F
.66908 .48765 m .64205 .4761 L .66216 .50131 L .459 0 0 r
F
.11228 .40274 m .07742 .40603 L .08177 .38366 L .41 .803 .987 r
F
.07742 .40603 m .05749 .38985 L .08177 .38366 L .411 .804 .987 r
F
.07742 .40603 m .0474 .40528 L .05749 .38985 L .41 .804 .987 r
F
.02728 .38315 m .05749 .38985 L .0474 .40528 L .41 .805 .986 r
F
.77681 .47825 m .74981 .4881 L .76331 .48894 L .079 0 0 r
F
.74981 .4881 m .7377 .49928 L .76331 .48894 L .084 0 0 r
F
.61895 .47259 m .5871 .45572 L .60254 .48431 L .444 0 0 r
F
.5871 .45572 m .5768 .47319 L .60254 .48431 L .442 0 0 r
F
.5871 .45572 m .55748 .45956 L .5768 .47319 L .358 0 0 r
F
.4796 .43859 m .50572 .44816 L .47843 .44485 L .113 .652 .851 r
F
.45571 .43335 m .4796 .43859 L .47843 .44485 L .112 .651 .851 r
F
.64113 .49466 m .66216 .50131 L .64205 .4761 L .478 0 0 r
F
.64113 .49466 m .64205 .4761 L .61895 .47259 L .501 0 0 r
F
.77681 .47825 m .76331 .48894 L .78953 .47412 L 0 0 0 r
F
.10977 .37338 m .14017 .38741 L .11228 .40274 L .41 .803 .987 r
F
.11228 .40274 m .08177 .38366 L .10977 .37338 L .412 .804 .987 r
F
.91068 .38831 m .9314 .3994 L .8982 .41726 L .42 .836 .967 r
F
.87249 .42034 m .91068 .38831 L .8982 .41726 L .425 .851 .949 r
F
.81616 .45795 m .80443 .46462 L .81732 .45464 L .168 .706 .759 r
F
.7999 .46898 m .81732 .45464 L .78953 .47412 L 0 0 0 r
F
.81848 .45134 m .81732 .45464 L .7999 .46898 L .222 .74 .736 r
F
.8259 .44429 m .81732 .45464 L .81848 .45134 L F
.83677 .44623 m .81616 .45795 L .81732 .45464 L .168 .706 .759 r
F
.38738 .41303 m .36199 .41863 L .35915 .39858 L .331 .763 .97 r
F
.33472 .40448 m .35915 .39858 L .36199 .41863 L .34 .77 .971 r
F
.30947 .41094 m .30809 .39002 L .33472 .40448 L .369 .781 .98 r
F
.30947 .41094 m .28276 .39705 L .30809 .39002 L .375 .785 .981 r
F
.87249 .42034 m .83677 .44623 L .84699 .43033 L .281 .782 .849 r
F
.83677 .44623 m .81732 .45464 L .84699 .43033 L .282 .784 .824 r
F
.28276 .39705 m .257 .40453 L .2565 .38298 L .391 .791 .985 r
F
.2565 .38298 m .257 .40453 L .22946 .38942 L .395 .794 .985 r
F
.16618 .39833 m .16813 .37703 L .19734 .39831 L .407 .8 .987 r
F
.14017 .38741 m .16813 .37703 L .16618 .39833 L .409 .802 .987 r
F
.5768 .47319 m .55748 .45956 L .55626 .46687 L .274 0 0 r
F
.55626 .46687 m .55748 .45956 L .53571 .45753 L .265 0 0 r
F
.64113 .49466 m .61895 .47259 L .62799 .50241 L .522 0 0 r
F
.61895 .47259 m .60254 .48431 L .62799 .50241 L .52 0 0 r
F
.42192 .42805 m .43777 .41696 L .45571 .43335 L .238 .722 .93 r
F
.40845 .40609 m .43777 .41696 L .42192 .42805 L .243 .719 .938 r
F
.42192 .42805 m .38738 .41303 L .40845 .40609 L .283 .744 .95 r
F
.19734 .39831 m .20214 .36725 L .22946 .38942 L .405 .798 .987 r
F
.19734 .39831 m .16813 .37703 L .20214 .36725 L .407 .8 .987 r
F
.9314 .3994 m .9512 .36881 L .96244 .39121 L .476 .839 .993 r
F
.91068 .38831 m .9512 .36881 L .9314 .3994 L .462 .855 .976 r
F
.51975 .45127 m .51924 .45173 L .50595 .44169 L 0 .478 .716 r
F
.51924 .45173 m .53571 .45753 L .51871 .45222 L 0 .539 .721 r
F
.54426 .46528 m .53571 .45753 L .51924 .45173 L .183 0 0 r
F
.54426 .46528 m .5292 .45712 L .53171 .45958 L .131 0 0 r
F
.5292 .45712 m .51924 .45173 L .53171 .45958 L F
.53171 .45958 m .51924 .45173 L .51975 .45127 L 0 .478 .716 r
F
.53171 .45958 m .50595 .44169 L .52025 .4488 L 0 .52 .767 r
F
.53171 .45958 m .52025 .4488 L .53419 .46201 L .177 0 0 r
F
.53171 .45958 m .53419 .46201 L .54508 .47232 L F
.51924 .45173 m .51871 .45222 L .50572 .44816 L 0 .539 .721 r
F
.49252 .43899 m .50572 .44816 L .4796 .43859 L .055 .613 .809 r
F
.05749 .38985 m .05358 .3709 L .08177 .38366 L .411 .805 .987 r
F
.02728 .38315 m .05358 .3709 L .05749 .38985 L .413 .806 .987 r
F
.45571 .43335 m .46597 .42709 L .4796 .43859 L .163 .68 .89 r
F
.46597 .42709 m .45571 .43335 L .43777 .41696 L .166 .677 .898 r
F
.7377 .49928 m .71234 .50487 L .72629 .51551 L .301 0 0 r
F
.70258 .52154 m .72629 .51551 L .71234 .50487 L .303 0 0 r
F
.70258 .52154 m .71234 .50487 L .68706 .50544 L .39 0 0 r
F
.70258 .52154 m .68706 .50544 L .67892 .52312 L .392 0 0 r
F
.68706 .50544 m .66216 .50131 L .67892 .52312 L .462 0 0 r
F
.7503 .50473 m .76331 .48894 L .7377 .49928 L .193 0 0 r
F
.7377 .49928 m .72629 .51551 L .7503 .50473 L .197 0 0 r
F
.33472 .40448 m .33286 .38373 L .35915 .39858 L .356 .772 .978 r
F
.33472 .40448 m .30809 .39002 L .33286 .38373 L .364 .779 .979 r
F
.66216 .50131 m .65383 .52076 L .67892 .52312 L .463 0 0 r
F
.64113 .49466 m .65383 .52076 L .66216 .50131 L .509 0 0 r
F
.28155 .37537 m .30809 .39002 L .28276 .39705 L .384 .786 .984 r
F
.28155 .37537 m .28276 .39705 L .2565 .38298 L .39 .791 .985 r
F
.13753 .36119 m .16813 .37703 L .14017 .38741 L .411 .802 .988 r
F
.10977 .37338 m .13753 .36119 L .14017 .38741 L .413 .804 .988 r
F
.64113 .49466 m .62799 .50241 L .65383 .52076 L .508 0 0 r
F
.76331 .48894 m .77476 .48916 L .78953 .47412 L .065 0 0 r
F
.7503 .50473 m .77476 .48916 L .76331 .48894 L .072 0 0 r
F
.38123 .39275 m .40845 .40609 L .38738 .41303 L .302 .746 .963 r
F
.38738 .41303 m .35915 .39858 L .38123 .39275 L .321 .759 .967 r
F
.87249 .42034 m .87719 .40019 L .91068 .38831 L .398 .843 .93 r
F
.87249 .42034 m .84699 .43033 L .87719 .40019 L .413 .855 .918 r
F
.55626 .46687 m .53571 .45753 L .54426 .46528 L .188 0 0 r
F
.49252 .43899 m .4796 .43859 L .46597 .42709 L .084 .626 .849 r
F
.23377 .37002 m .2565 .38298 L .22946 .38942 L .401 .795 .987 r
F
.23377 .37002 m .22946 .38942 L .20214 .36725 L .403 .797 .987 r
F
.08177 .38366 m .08016 .35851 L .10977 .37338 L .413 .804 .988 r
F
.05358 .3709 m .08016 .35851 L .08177 .38366 L .414 .806 .988 r
F
.77476 .48916 m .7999 .46898 L .78953 .47412 L 0 0 0 r
F
.5901 .49376 m .60254 .48431 L .5768 .47319 L .424 0 0 r
F
.5901 .49376 m .5768 .47319 L .56703 .47918 L .408 0 0 r
F
.5768 .47319 m .55626 .46687 L .56703 .47918 L .33 0 0 r
F
.60254 .48431 m .61381 .5114 L .62799 .50241 L .498 0 0 r
F
.60254 .48431 m .5901 .49376 L .61381 .5114 L .499 0 0 r
F
.55626 .46687 m .54426 .46528 L .56703 .47918 L .304 0 0 r
F
.8259 .44429 m .84699 .43033 L .81732 .45464 L .223 .743 .766 r
F
.30685 .36821 m .33286 .38373 L .30809 .39002 L .376 .78 .984 r
F
.30685 .36821 m .30809 .39002 L .28155 .37537 L .384 .786 .984 r
F
.38123 .39275 m .35915 .39858 L .35596 .37814 L .339 .761 .975 r
F
.35915 .39858 m .33286 .38373 L .35596 .37814 L .352 .771 .977 r
F
.28155 .37537 m .2565 .38298 L .25495 .36125 L .396 .791 .987 r
F
.23377 .37002 m .25495 .36125 L .2565 .38298 L .402 .795 .987 r
F
.1663 .34938 m .20214 .36725 L .16813 .37703 L .411 .801 .989 r
F
.13753 .36119 m .1663 .34938 L .16813 .37703 L .415 .803 .989 r
F
.65383 .52076 m .62799 .50241 L .63696 .53085 L .53 .006 0 r
F
.62799 .50241 m .61381 .5114 L .63696 .53085 L .538 .002 0 r
F
.13753 .36119 m .10977 .37338 L .10704 .34599 L .414 .804 .988 r
F
.08016 .35851 m .10704 .34599 L .10977 .37338 L .416 .806 .988 r
F
.71589 .53695 m .72629 .51551 L .70258 .52154 L .399 0 0 r
F
.69408 .54332 m .71589 .53695 L .70258 .52154 L .4 0 0 r
F
.69408 .54332 m .70258 .52154 L .67892 .52312 L .471 0 0 r
F
.93974 .34596 m .9512 .36881 L .91068 .38831 L .526 .863 .994 r
F
.93974 .34596 m .91068 .38831 L .90477 .36433 L .5 .877 .974 r
F
.87719 .40019 m .90477 .36433 L .91068 .38831 L .514 .885 .973 r
F
.72629 .51551 m .73812 .52587 L .7503 .50473 L .304 0 0 r
F
.71589 .53695 m .73812 .52587 L .72629 .51551 L .305 0 0 r
F
.67294 .54559 m .69408 .54332 L .67892 .52312 L .471 0 0 r
F
.67294 .54559 m .67892 .52312 L .65383 .52076 L .523 .012 0 r
F
.43777 .41696 m .45149 .40998 L .46597 .42709 L .208 .698 .924 r
F
.43777 .41696 m .42311 .39679 L .45149 .40998 L .214 .696 .931 r
F
.40845 .40609 m .42311 .39679 L .43777 .41696 L .269 .73 .95 r
F
.85212 .41518 m .87719 .40019 L .84699 .43033 L .361 .831 .873 r
F
.85212 .41518 m .84699 .43033 L .8259 .44429 L .372 .839 .856 r
F
.39887 .38762 m .42311 .39679 L .40845 .40609 L .275 .728 .956 r
F
.39887 .38762 m .40845 .40609 L .38123 .39275 L .308 .748 .965 r
F
.49252 .43899 m .46597 .42709 L .47907 .42488 L .12 .647 .874 r
F
.47907 .42488 m .46597 .42709 L .45149 .40998 L .131 .648 .89 r
F
.49252 .43899 m .50595 .44169 L .51924 .45173 L .007 .573 .791 r
F
.49252 .43899 m .47907 .42488 L .50595 .44169 L .03 .581 .827 r
F
.77476 .48916 m .7503 .50473 L .76082 .50982 L .181 0 0 r
F
.7503 .50473 m .73812 .52587 L .76082 .50982 L .184 0 0 r
F
.67294 .54559 m .65383 .52076 L .65567 .54658 L .524 .015 0 r
F
.65383 .52076 m .63696 .53085 L .65567 .54658 L .552 .036 0 r
F
.7999 .46898 m .77476 .48916 L .78404 .48865 L .028 0 0 r
F
.78404 .48865 m .77476 .48916 L .76082 .50982 L .033 0 0 r
F
.56703 .47918 m .54426 .46528 L .55583 .47824 L .278 0 0 r
F
.54426 .46528 m .53171 .45958 L .55583 .47824 L .256 0 0 r
F
.33159 .36247 m .35596 .37814 L .33286 .38373 L .366 .772 .983 r
F
.33159 .36247 m .33286 .38373 L .30685 .36821 L .375 .78 .984 r
F
.30685 .36821 m .28155 .37537 L .28002 .35242 L .392 .787 .987 r
F
.28002 .35242 m .28155 .37537 L .25495 .36125 L .399 .792 .988 r
F
.23377 .37002 m .22634 .34917 L .25495 .36125 L .404 .795 .988 r
F
.23377 .37002 m .20214 .36725 L .22634 .34917 L .411 .799 .989 r
F
.19451 .33625 m .22634 .34917 L .20214 .36725 L F
.1663 .34938 m .19451 .33625 L .20214 .36725 L .417 .803 .99 r
F
.57879 .49756 m .5901 .49376 L .56703 .47918 L .402 0 0 r
F
.57879 .49756 m .56703 .47918 L .55583 .47824 L .382 0 0 r
F
.5901 .49376 m .60126 .51805 L .61381 .5114 L .492 0 0 r
F
.5901 .49376 m .57879 .49756 L .60126 .51805 L .487 0 0 r
F
.39887 .38762 m .38123 .39275 L .37684 .37184 L .322 .749 .973 r
F
.38123 .39275 m .35596 .37814 L .37684 .37184 L .34 .762 .976 r
F
.7916 .4866 m .80684 .46528 L .80733 .46347 L 0 0 0 r
F
.8259 .44429 m .80789 .46139 L .80733 .46347 L F
.83132 .43297 m .8259 .44429 L .80733 .46347 L .312 .799 .765 r
F
.80684 .46528 m .81078 .45976 L .80733 .46347 L 0 0 0 r
F
.80789 .46139 m .7999 .46898 L .80733 .46347 L F
.80733 .46347 m .7999 .46898 L .78404 .48865 L F
.13753 .36119 m .1342 .33333 L .1663 .34938 L .416 .803 .989 r
F
.13753 .36119 m .10704 .34599 L .1342 .33333 L .419 .806 .989 r
F
.62336 .54031 m .63696 .53085 L .61381 .5114 L .548 .02 0 r
F
.61381 .5114 m .60126 .51805 L .62336 .54031 L .556 .011 0 r
F
.876 .38331 m .90477 .36433 L .87719 .40019 L .469 .881 .938 r
F
.85212 .41518 m .876 .38331 L .87719 .40019 L .486 .892 .934 r
F
.83132 .43297 m .85212 .41518 L .8259 .44429 L .303 .796 .788 r
F
.25165 .33725 m .28002 .35242 L .25495 .36125 L .405 .792 .989 r
F
.25165 .33725 m .25495 .36125 L .22634 .34917 L .412 .797 .99 r
F
.69408 .54332 m .70669 .56357 L .71589 .53695 L .478 0 .035 r
F
.68681 .56964 m .70669 .56357 L .69408 .54332 L .478 0 .032 r
F
.68681 .56964 m .69408 .54332 L .67294 .54559 L .536 .061 .073 r
F
.65567 .54658 m .63696 .53085 L .6463 .56541 L .576 .083 .048 r
F
.62336 .54031 m .6463 .56541 L .63696 .53085 L .591 .079 .001 r
F
.33159 .36247 m .30685 .36821 L .30622 .34494 L .387 .781 .988 r
F
.30685 .36821 m .28002 .35242 L .30622 .34494 L .396 .788 .988 r
F
.72709 .55237 m .73812 .52587 L .71589 .53695 L .398 0 0 r
F
.72709 .55237 m .71589 .53695 L .70669 .56357 L .399 0 0 r
F
.3531 .35911 m .37684 .37184 L .35596 .37814 L .352 .762 .981 r
F
.33159 .36247 m .3531 .35911 L .35596 .37814 L .367 .773 .983 r
F
.76082 .50982 m .73812 .52587 L .74787 .53616 L .291 0 0 r
F
.73812 .52587 m .72709 .55237 L .74787 .53616 L .292 0 0 r
F
.67294 .54559 m .66762 .57132 L .68681 .56964 L .536 .055 .061 r
F
.67294 .54559 m .65567 .54658 L .66762 .57132 L .575 .098 .084 r
F
.76082 .50982 m .76943 .51459 L .78404 .48865 L .147 0 0 r
F
.76082 .50982 m .74787 .53616 L .76943 .51459 L .148 0 0 r
F
.39887 .38762 m .40422 .36961 L .42311 .39679 L .315 .742 .972 r
F
.39887 .38762 m .37684 .37184 L .40422 .36961 L .319 .748 .972 r
F
.45149 .40998 m .46552 .4054 L .47907 .42488 L .189 .678 .924 r
F
.45149 .40998 m .43755 .38818 L .46552 .4054 L .196 .676 .931 r
F
.42311 .39679 m .43755 .38818 L .45149 .40998 L .262 .718 .954 r
F
.50595 .44169 m .47907 .42488 L .49332 .42571 L .083 .615 .866 r
F
.46552 .4054 m .49332 .42571 L .47907 .42488 L .098 .616 .885 r
F
.50595 .44169 m .49332 .42571 L .52025 .4488 L 0 .533 .807 r
F
.42311 .39679 m .40422 .36961 L .43755 .38818 L .266 .715 .958 r
F
.65567 .54658 m .6463 .56541 L .66762 .57132 L .575 .082 .047 r
F
.54508 .47232 m .5415 .46716 L .53171 .45958 L .211 0 0 r
F
.54508 .47232 m .52988 .45793 L .5357 .46144 L 0 .461 .739 r
F
.52988 .45793 m .52025 .4488 L .5357 .46144 L F
.52853 .44595 m .54133 .46485 L .5357 .46144 L 0 .43 .777 r
F
.54508 .47232 m .5486 .47741 L .55901 .49031 L .244 0 0 r
F
.54508 .47232 m .55583 .47824 L .5415 .46716 L .211 0 0 r
F
.54508 .47232 m .56824 .49622 L .55583 .47824 L .331 0 0 r
F
.55901 .49031 m .56824 .49622 L .555 .48257 L .293 0 0 r
F
.55901 .49031 m .555 .48257 L .54508 .47232 L F
.54508 .47232 m .5357 .46144 L .5486 .47741 L .244 0 0 r
F
.55901 .49031 m .54539 .47344 L .55146 .47979 L 0 .393 .704 r
F
.54539 .47344 m .5357 .46144 L .55146 .47979 L F
.55725 .48585 m .55146 .47979 L .54602 .46514 L 0 .368 .754 r
F
.55901 .49031 m .55146 .47979 L .56293 .49787 L .316 0 0 r
F
.55901 .49031 m .56293 .49787 L .57308 .51387 L F
.55901 .49031 m .58068 .51915 L .56824 .49622 L .412 0 0 r
F
.55901 .49031 m .57308 .51387 L .58068 .51915 L .379 0 0 r
F
.56644 .5028 m .55146 .47979 L .55725 .48585 L 0 .368 .754 r
F
.54133 .46485 m .55146 .47979 L .5357 .46144 L 0 .43 .777 r
F
.16167 .32054 m .19451 .33625 L .1663 .34938 L .418 .803 .99 r
F
.1663 .34938 m .1342 .33333 L .16167 .32054 L .422 .806 .99 r
F
.57879 .49756 m .55583 .47824 L .56824 .49622 L .362 0 0 r
F
.90477 .36433 m .92804 .32265 L .93974 .34596 L .581 .886 .991 r
F
.57879 .49756 m .59014 .52057 L .60126 .51805 L .48 0 0 r
F
.56824 .49622 m .59014 .52057 L .57879 .49756 L .464 0 0 r
F
.85495 .40181 m .876 .38331 L .85212 .41518 L .434 .873 .887 r
F
.85495 .40181 m .85212 .41518 L .83132 .43297 L .449 .882 .869 r
F
.80733 .46347 m .78404 .48865 L .7916 .4866 L 0 0 0 r
F
.7916 .4866 m .78404 .48865 L .76943 .51459 L F
.22136 .32264 m .25165 .33725 L .22634 .34917 L .414 .797 .991 r
F
.19451 .33625 m .22136 .32264 L .22634 .34917 L .42 .802 .991 r
F
.61094 .54547 m .62336 .54031 L .60126 .51805 L .559 .016 0 r
F
.60126 .51805 m .59014 .52057 L .61094 .54547 L .561 0 0 r
F
.33159 .36247 m .3334 .34009 L .3531 .35911 L .382 .774 .988 r
F
.33159 .36247 m .30622 .34494 L .3334 .34009 L .39 .782 .988 r
F
.28002 .35242 m .27832 .32725 L .30622 .34494 L .403 .788 .99 r
F
.28002 .35242 m .25165 .33725 L .27832 .32725 L .412 .795 .991 r
F
.92804 .32265 m .90477 .36433 L .89691 .34516 L .561 .905 .975 r
F
.876 .38331 m .89691 .34516 L .90477 .36433 L .572 .909 .974 r
F
.81078 .45976 m .8009 .47675 L .80997 .46142 L 0 0 0 r
F
.83132 .43297 m .81167 .45794 L .81078 .45976 L F
.8224 .44041 m .83132 .43297 L .81078 .45976 L .358 .826 .768 r
F
.81078 .45976 m .80997 .46142 L .8224 .44041 L 0 0 0 r
F
.81167 .45794 m .80733 .46347 L .81078 .45976 L F
.40422 .36961 m .37684 .37184 L .36677 .3425 L .343 .75 .981 r
F
.36677 .3425 m .37684 .37184 L .3531 .35911 L .372 .769 .986 r
F
.63025 .57089 m .6463 .56541 L .62336 .54031 L .601 .1 .034 r
F
.63025 .57089 m .62336 .54031 L .61094 .54547 L .621 .102 0 r
F
.70669 .56357 m .71742 .58483 L .72709 .55237 L .476 .03 .111 r
F
.69858 .59504 m .71742 .58483 L .70669 .56357 L .476 .03 .11 r
F
.69858 .59504 m .70669 .56357 L .68681 .56964 L .542 .104 .158 r
F
.68681 .56964 m .68049 .59963 L .69858 .59504 L .543 .099 .145 r
F
.68681 .56964 m .66762 .57132 L .68049 .59963 L .589 .15 .175 r
F
.73641 .5674 m .74787 .53616 L .72709 .55237 L .383 0 .024 r
F
.72709 .55237 m .71742 .58483 L .73641 .5674 L .382 0 .037 r
F
.68049 .59963 m .66762 .57132 L .66286 .59987 L .592 .139 .147 r
F
.66762 .57132 m .6463 .56541 L .66286 .59987 L .624 .175 .165 r
F
.75526 .54685 m .76943 .51459 L .74787 .53616 L .262 0 0 r
F
.73641 .5674 m .75526 .54685 L .74787 .53616 L .26 0 0 r
F
.36677 .3425 m .3531 .35911 L .3334 .34009 L .372 .77 .986 r
F
.19451 .33625 m .18944 .3076 L .22136 .32264 L .421 .802 .991 r
F
.16167 .32054 m .18944 .3076 L .19451 .33625 L .426 .805 .992 r
F
.89691 .34516 m .876 .38331 L .86977 .37233 L .54 .916 .941 r
F
.85495 .40181 m .86977 .37233 L .876 .38331 L .543 .917 .941 r
F
.6463 .56541 m .64664 .59888 L .66286 .59987 L .624 .157 .125 r
F
.63025 .57089 m .64664 .59888 L .6463 .56541 L .655 .184 .122 r
F
.25165 .33725 m .24862 .31044 L .27832 .32725 L .416 .795 .992 r
F
.25165 .33725 m .22136 .32264 L .24862 .31044 L .424 .8 .993 r
F
.30622 .34494 m .30605 .3189 L .3334 .34009 L .402 .782 .991 r
F
.30622 .34494 m .27832 .32725 L .30605 .3189 L .412 .791 .992 r
F
.77685 .51886 m .7916 .4866 L .76943 .51459 L .077 0 0 r
F
.75526 .54685 m .77685 .51886 L .76943 .51459 L .073 0 0 r
F
.58068 .51915 m .59014 .52057 L .56824 .49622 L .449 0 0 r
F
.84519 .40548 m .85495 .40181 L .83132 .43297 L .45 .883 .87 r
F
.84519 .40548 m .83132 .43297 L .8224 .44041 L .409 .856 .783 r
F
.52025 .4488 m .51007 .43203 L .5357 .46144 L 0 .485 .795 r
F
.49332 .42571 m .51007 .43203 L .52025 .4488 L .047 .579 .859 r
F
.46552 .4054 m .47927 .39987 L .49332 .42571 L .18 .659 .93 r
F
.45257 .38413 m .47927 .39987 L .46552 .4054 L .182 .657 .932 r
F
.43755 .38818 m .45257 .38413 L .46552 .4054 L .254 .701 .958 r
F
.59014 .52057 m .60082 .54761 L .61094 .54547 L .559 0 0 r
F
.58068 .51915 m .60082 .54761 L .59014 .52057 L .546 0 0 r
F
.49332 .42571 m .47927 .39987 L .51007 .43203 L .077 .587 .887 r
F
.43755 .38818 m .42817 .36278 L .45257 .38413 L .262 .7 .962 r
F
.42817 .36278 m .43755 .38818 L .40422 .36961 L .314 .733 .975 r
F
.84519 .40548 m .86977 .37233 L .85495 .40181 L .503 .909 .884 r
F
.61944 .57552 m .63025 .57089 L .61094 .54547 L .623 .105 0 r
F
.61094 .54547 m .60082 .54761 L .61944 .57552 L .632 .093 0 r
F
.42817 .36278 m .40422 .36961 L .39748 .33622 L .326 .731 .98 r
F
.40422 .36961 m .36677 .3425 L .39748 .33622 L .365 .758 .987 r
F
.7994 .4822 m .8009 .47675 L .79134 .49912 L 0 0 0 r
F
.80232 .47161 m .8009 .47675 L .81078 .45976 L F
.81209 .45554 m .8224 .44041 L .81474 .45336 L .357 .814 .705 r
F
.81209 .45554 m .81474 .45336 L .8009 .47675 L F
.81209 .45554 m .8009 .47675 L .7994 .4822 L 0 0 0 r
F
.80955 .45764 m .81209 .45554 L .80266 .47346 L .473 .871 .696 r
F
.82116 .43099 m .81209 .45554 L .80955 .45764 L F
.7916 .4866 m .8009 .47675 L .80232 .47161 L 0 0 0 r
F
.77685 .51886 m .8009 .47675 L .7916 .4866 L F
.91611 .29887 m .92804 .32265 L .89691 .34516 L .631 .905 .984 r
F
.63428 .59782 m .64664 .59888 L .63025 .57089 L .652 .173 .102 r
F
.63428 .59782 m .63025 .57089 L .61944 .57552 L .677 .184 .07 r
F
.69858 .59504 m .7085 .62334 L .71742 .58483 L .537 .138 .233 r
F
.69858 .59504 m .69136 .62977 L .7085 .62334 L .541 .136 .222 r
F
.69858 .59504 m .68049 .59963 L .69136 .62977 L .593 .194 .257 r
F
.68049 .59963 m .67534 .63223 L .69136 .62977 L .598 .189 .237 r
F
.67534 .63223 m .68049 .59963 L .66286 .59987 L .638 .232 .258 r
F
.73641 .5674 m .71742 .58483 L .72727 .60809 L .46 .048 .171 r
F
.7085 .62334 m .72727 .60809 L .71742 .58483 L .458 .05 .176 r
F
.3334 .34009 m .33547 .3128 L .36677 .3425 L .397 .773 .992 r
F
.3334 .34009 m .30605 .3189 L .33547 .3128 L .409 .785 .993 r
F
.74365 .57792 m .75526 .54685 L .73641 .5674 L .351 0 .064 r
F
.72727 .60809 m .74365 .57792 L .73641 .5674 L .348 0 .093 r
F
.21751 .29452 m .24862 .31044 L .22136 .32264 L .425 .8 .993 r
F
.22136 .32264 m .18944 .3076 L .21751 .29452 L .431 .805 .993 r
F
.6603 .63263 m .67534 .63223 L .66286 .59987 L .643 .222 .228 r
F
.6603 .63263 m .66286 .59987 L .64664 .59888 L .674 .252 .237 r
F
.27832 .32725 m .27656 .29918 L .30605 .3189 L .418 .791 .994 r
F
.27832 .32725 m .24862 .31044 L .27656 .29918 L .428 .798 .994 r
F
.91611 .29887 m .89691 .34516 L .88658 .32974 L .622 .929 .969 r
F
.86977 .37233 m .88658 .32974 L .89691 .34516 L .621 .929 .969 r
F
.75526 .54685 m .75745 .56307 L .77685 .51886 L .207 0 0 r
F
.74365 .57792 m .75745 .56307 L .75526 .54685 L .213 0 0 r
F
.64538 .62844 m .6603 .63263 L .64664 .59888 L .676 .236 .198 r
F
.63428 .59782 m .64538 .62844 L .64664 .59888 L .699 .259 .201 r
F
.59283 .54707 m .60082 .54761 L .58068 .51915 L .54 0 0 r
F
.57308 .51387 m .59283 .54707 L .58068 .51915 L .496 0 0 r
F
.36567 .30703 m .39748 .33622 L .36677 .3425 L .382 .757 .991 r
F
.36567 .30703 m .36677 .3425 L .33547 .3128 L .405 .777 .993 r
F
.61944 .57552 m .60082 .54761 L .61074 .57833 L .633 .096 0 r
F
.59283 .54707 m .61074 .57833 L .60082 .54761 L .625 .065 0 r
F
.85856 .36952 m .88658 .32974 L .86977 .37233 L .602 .944 .931 r
F
.84519 .40548 m .85856 .36952 L .86977 .37233 L .581 .937 .931 r
F
.52853 .44595 m .5357 .46144 L .51007 .43203 L 0 .529 .842 r
F
.45257 .38413 m .45229 .36497 L .47927 .39987 L .261 .684 .965 r
F
.45257 .38413 m .42817 .36278 L .45229 .36497 L .292 .712 .972 r
F
.62937 .61454 m .63428 .59782 L .61944 .57552 L .688 .21 .112 r
F
.62937 .61454 m .61944 .57552 L .61074 .57833 L .702 .193 .021 r
F
.83363 .41393 m .85856 .36952 L .84519 .40548 L .555 .933 .863 r
F
.83363 .41393 m .84519 .40548 L .8224 .44041 L .486 .901 .845 r
F
.52853 .44595 m .51007 .43203 L .50356 .41089 L .053 .549 .886 r
F
.50356 .41089 m .51007 .43203 L .47927 .39987 L .151 .627 .925 r
F
.83363 .41393 m .8224 .44041 L .81209 .45554 L .444 .867 .743 r
F
.77685 .51886 m .77183 .54397 L .79134 .49912 L 0 0 0 r
F
.8009 .47675 m .77685 .51886 L .79134 .49912 L F
.77685 .51886 m .75745 .56307 L .77183 .54397 L .031 0 0 r
F
.62937 .61454 m .64538 .62844 L .63428 .59782 L .702 .228 .123 r
F
.7085 .62334 m .69136 .62977 L .70095 .6651 L .583 .222 .322 r
F
.69136 .62977 m .68418 .66246 L .70095 .6651 L .593 .22 .305 r
F
.67534 .63223 m .68418 .66246 L .69136 .62977 L .637 .274 .338 r
F
.67261 .66871 m .68418 .66246 L .67534 .63223 L .649 .268 .31 r
F
.67261 .66871 m .67534 .63223 L .6603 .63263 L .678 .298 .319 r
F
.74365 .57792 m .74198 .60344 L .75745 .56307 L .307 0 .104 r
F
.72727 .60809 m .74198 .60344 L .74365 .57792 L .368 0 .119 r
F
.7085 .62334 m .71543 .65734 L .72727 .60809 L .52 .161 .298 r
F
.7085 .62334 m .70095 .6651 L .71543 .65734 L .53 .163 .288 r
F
.30605 .3189 m .30527 .28879 L .33547 .3128 L .42 .785 .995 r
F
.30605 .3189 m .27656 .29918 L .30527 .28879 L .433 .795 .996 r
F
.24862 .31044 m .2459 .28129 L .27656 .29918 L .43 .798 .995 r
F
.21751 .29452 m .2459 .28129 L .24862 .31044 L .438 .804 .995 r
F
.42817 .36278 m .42466 .33312 L .45229 .36497 L .321 .713 .981 r
F
.42817 .36278 m .39748 .33622 L .42466 .33312 L .36 .744 .988 r
F
.47927 .39987 m .47763 .37423 L .50356 .41089 L .204 .636 .95 r
F
.45229 .36497 m .47763 .37423 L .47927 .39987 L .256 .682 .964 r
F
.65877 .66788 m .67261 .66871 L .6603 .63263 L .685 .289 .291 r
F
.64538 .62844 m .65877 .66788 L .6603 .63263 L .712 .315 .295 r
F
.74198 .60344 m .72727 .60809 L .72759 .64337 L .411 .054 .245 r
F
.72727 .60809 m .71543 .65734 L .72759 .64337 L .455 .085 .247 r
F
.75519 .58898 m .77183 .54397 L .75745 .56307 L .159 0 0 r
F
.75745 .56307 m .74198 .60344 L .75519 .58898 L .209 0 .001 r
F
.64538 .62844 m .64266 .65548 L .65877 .66788 L .72 .295 .239 r
F
.64538 .62844 m .62937 .61454 L .64266 .65548 L .742 .32 .247 r
F
.60348 .57566 m .61074 .57833 L .59283 .54707 L .621 .06 0 r
F
.58808 .54617 m .60348 .57566 L .59283 .54707 L .562 0 0 r
F
.57308 .51387 m .58808 .54617 L .59283 .54707 L .478 0 0 r
F
.90394 .2746 m .91611 .29887 L .88658 .32974 L .68 .92 .973 r
F
.61839 .61112 m .62937 .61454 L .61074 .57833 L .697 .184 .006 r
F
.60348 .57566 m .61839 .61112 L .61074 .57833 L .703 .172 0 r
F
.57308 .51387 m .56017 .49353 L .56644 .5028 L 0 .322 .67 r
F
.56017 .49353 m .55146 .47979 L .56644 .5028 L F
.57196 .51098 m .56644 .5028 L .56189 .48818 L 0 .307 .731 r
F
.57308 .51387 m .56644 .5028 L .58808 .54617 L .399 0 0 r
F
.57788 .52441 m .56644 .5028 L .57196 .51098 L 0 .307 .731 r
F
.54602 .46514 m .55146 .47979 L .52853 .44595 L 0 .474 .824 r
F
.39491 .30078 m .42466 .33312 L .39748 .33622 L .378 .742 .991 r
F
.39491 .30078 m .39748 .33622 L .36567 .30703 L .41 .766 .995 r
F
.33547 .3128 m .33462 .27898 L .36567 .30703 L .42 .776 .996 r
F
.33547 .3128 m .30527 .28879 L .33462 .27898 L .438 .791 .997 r
F
.64266 .65548 m .62937 .61454 L .63455 .65521 L .747 .285 .153 r
F
.61839 .61112 m .63455 .65521 L .62937 .61454 L .764 .294 .127 r
F
.74198 .60344 m .74086 .63497 L .75519 .58898 L .286 0 .156 r
F
.74198 .60344 m .72759 .64337 L .74086 .63497 L .337 0 .165 r
F
.87389 .31804 m .90394 .2746 L .88658 .32974 L .68 .95 .956 r
F
.85856 .36952 m .87389 .31804 L .88658 .32974 L .672 .948 .957 r
F
.54602 .46514 m .52853 .44595 L .52441 .42692 L .019 .503 .874 r
F
.52853 .44595 m .50356 .41089 L .52441 .42692 L .111 .582 .914 r
F
.27461 .26792 m .30527 .28879 L .27656 .29918 L .436 .795 .996 r
F
.27656 .29918 m .2459 .28129 L .27461 .26792 L .447 .803 .997 r
F
.83363 .41393 m .81209 .45554 L .82116 .43099 L .52 .915 .812 r
F
.81209 .45554 m .79134 .49912 L .80266 .47346 L .399 .827 .666 r
F
.78341 .52294 m .80266 .47346 L .79134 .49912 L 0 0 0 r
F
.78341 .52294 m .79134 .49912 L .77183 .54397 L F
.45142 .33591 m .47763 .37423 L .45229 .36497 L .307 .687 .977 r
F
.45142 .33591 m .45229 .36497 L .42466 .33312 L .349 .724 .986 r
F
.84181 .37828 m .85856 .36952 L .83363 .41393 L .621 .956 .909 r
F
.84181 .37828 m .83363 .41393 L .82116 .43099 L .626 .961 .851 r
F
.84181 .37828 m .87389 .31804 L .85856 .36952 L .664 .969 .912 r
F
.76671 .57 m .78341 .52294 L .77183 .54397 L 0 0 0 r
F
.75519 .58898 m .76671 .57 L .77183 .54397 L .004 0 0 r
F
.50356 .41089 m .50157 .38711 L .52441 .42692 L .174 .596 .94 r
F
.47763 .37423 m .50157 .38711 L .50356 .41089 L .239 .653 .961 r
F
.60348 .57566 m .60728 .595 L .61839 .61112 L .649 .091 0 r
F
.58808 .54617 m .60728 .595 L .60348 .57566 L .65 .092 0 r
F
.75235 .61739 m .76671 .57 L .75519 .58898 L .119 0 .03 r
F
.74086 .63497 m .75235 .61739 L .75519 .58898 L .18 0 .052 r
F
.61839 .61112 m .62482 .64656 L .63455 .65521 L .749 .25 .039 r
F
.61839 .61112 m .60728 .595 L .62482 .64656 L .757 .248 0 r
F
.36567 .30703 m .36414 .26931 L .39491 .30078 L .422 .764 .996 r
F
.36567 .30703 m .33462 .27898 L .36414 .26931 L .446 .784 .998 r
F
.58326 .53391 m .57788 .52441 L .57731 .51725 L 0 .271 .73 r
F
.57225 .51445 m .57788 .52441 L .58808 .54617 L 0 .275 .672 r
F
.59543 .56795 m .58808 .54617 L .57788 .52441 L .474 0 0 r
F
.59543 .56795 m .57788 .52441 L .58326 .53391 L 0 .271 .73 r
F
.56644 .5028 m .57788 .52441 L .57225 .51445 L 0 .275 .672 r
F
.54602 .46514 m .56189 .48818 L .56644 .5028 L 0 .42 .808 r
F
.59543 .56795 m .60728 .595 L .58808 .54617 L .47 0 0 r
F
.30364 .25439 m .33462 .27898 L .30527 .28879 L .443 .79 .998 r
F
.27461 .26792 m .30364 .25439 L .30527 .28879 L .457 .801 .998 r
F
.45142 .33591 m .42466 .33312 L .42355 .29751 L .376 .722 .989 r
F
.42355 .29751 m .42466 .33312 L .39491 .30078 L .412 .752 .994 r
F
.54323 .44758 m .56189 .48818 L .54602 .46514 L 0 .455 .86 r
F
.52441 .42692 m .54323 .44758 L .54602 .46514 L .08 .537 .903 r
F
.50157 .38711 m .47763 .37423 L .4778 .34508 L .292 .657 .971 r
F
.4778 .34508 m .47763 .37423 L .45142 .33591 L .337 .7 .982 r
F
.81517 .43371 m .84181 .37828 L .82116 .43099 L .597 .948 .818 r
F
.80266 .47346 m .81517 .43371 L .82116 .43099 L .614 .954 .817 r
F
.80266 .47346 m .79232 .49656 L .81517 .43371 L .545 .9 .686 r
F
.78341 .52294 m .79232 .49656 L .80239 .47415 L .484 .866 .661 r
F
.78341 .52294 m .77537 .54956 L .79232 .49656 L 0 0 0 r
F
.78341 .52294 m .77931 .5345 L .77537 .54956 L F
.77931 .5345 m .76671 .57 L .77537 .54956 L F
.77182 .5631 m .77537 .54956 L .76833 .57585 L F
.77926 .53595 m .79232 .49656 L .77537 .54956 L .503 .853 .594 r
F
.77926 .53595 m .77537 .54956 L .77182 .5631 L 0 0 0 r
F
.77926 .53595 m .78428 .51445 L .79232 .49656 L .69 .963 .73 r
F
.78428 .51445 m .78634 .50564 L .79232 .49656 L F
.80239 .47415 m .79232 .49656 L .80266 .47346 L .484 .866 .661 r
F
.76671 .57 m .76202 .59378 L .77537 .54956 L 0 0 0 r
F
.75235 .61739 m .76202 .59378 L .76671 .57 L F
.61342 .62311 m .62482 .64656 L .60728 .595 L .678 .132 0 r
F
.61342 .62311 m .60728 .595 L .59543 .56795 L .67 .127 0 r
F
.52346 .40528 m .54323 .44758 L .52441 .42692 L .153 .556 .928 r
F
.50157 .38711 m .52346 .40528 L .52441 .42692 L .218 .616 .952 r
F
.87389 .31804 m .89152 .24983 L .90394 .2746 L .727 .932 .957 r
F
.56189 .48818 m .57731 .51725 L .57788 .52441 L 0 .354 .778 r
F
.86062 .30627 m .89152 .24983 L .87389 .31804 L .733 .964 .939 r
F
.84181 .37828 m .86062 .30627 L .87389 .31804 L .73 .963 .94 r
F
.82894 .3742 m .86062 .30627 L .84181 .37828 L .723 .984 .9 r
F
.82894 .3742 m .84181 .37828 L .81517 .43371 L .735 .986 .897 r
F
.39491 .30078 m .39368 .2604 L .42355 .29751 L .426 .749 .994 r
F
.39491 .30078 m .36414 .26931 L .39368 .2604 L .457 .774 .997 r
F
.36414 .26931 m .33462 .27898 L .333 .24071 L .451 .782 .998 r
F
.33462 .27898 m .30364 .25439 L .333 .24071 L .47 .797 .999 r
F
.56011 .47189 m .57731 .51725 L .56189 .48818 L 0 .407 .844 r
F
.56011 .47189 m .56189 .48818 L .54323 .44758 L .056 .493 .891 r
F
.4778 .34508 m .45142 .33591 L .45267 .30009 L .373 .698 .983 r
F
.45142 .33591 m .42355 .29751 L .45267 .30009 L .411 .734 .991 r
F
.52346 .40528 m .50157 .38711 L .50265 .36046 L .277 .623 .96 r
F
.50157 .38711 m .4778 .34508 L .50265 .36046 L .324 .67 .976 r
F
.77537 .54956 m .76202 .59378 L .76833 .57585 L 0 0 0 r
F
.59543 .56795 m .61093 .61544 L .61342 .62311 L 0 .079 .546 r
F
.59291 .55418 m .61093 .61544 L .59543 .56795 L 0 .145 .643 r
F
.59291 .55418 m .59543 .56795 L .57731 .51725 L 0 .279 .738 r
F
.54323 .44758 m .54322 .42841 L .56011 .47189 L .136 .515 .914 r
F
.54323 .44758 m .52346 .40528 L .54322 .42841 L .197 .577 .94 r
F
.82894 .3742 m .81517 .43371 L .80355 .4495 L .719 .988 .831 r
F
.79232 .49656 m .80355 .4495 L .81517 .43371 L .712 .986 .83 r
F
.80355 .4495 m .79232 .49656 L .78634 .50564 L .671 .958 .731 r
F
.5752 .4977 m .59291 .55418 L .57731 .51725 L 0 .35 .818 r
F
.5752 .4977 m .57731 .51725 L .56011 .47189 L .045 .453 .878 r
F
.4778 .34508 m .48108 .31262 L .50265 .36046 L .369 .669 .974 r
F
.4778 .34508 m .45267 .30009 L .48108 .31262 L .405 .709 .985 r
F
.52559 .38172 m .54322 .42841 L .52346 .40528 L .264 .588 .947 r
F
.52346 .40528 m .50265 .36046 L .52559 .38172 L .308 .636 .965 r
F
.36414 .26931 m .36269 .22688 L .39368 .2604 L .462 .772 .997 r
F
.36414 .26931 m .333 .24071 L .36269 .22688 L .486 .791 .999 r
F
.42401 .25397 m .45267 .30009 L .42355 .29751 L .432 .729 .988 r
F
.42355 .29751 m .39368 .2604 L .42401 .25397 L .467 .76 .994 r
F
.55936 .45338 m .5752 .4977 L .56011 .47189 L .111 .47 .893 r
F
.55936 .45338 m .56011 .47189 L .54322 .42841 L .184 .54 .927 r
F
.77926 .53595 m .76833 .57585 L .77378 .55174 L .621 .913 .633 r
F
.77926 .53595 m .77378 .55174 L .78634 .50564 L .641 .921 .639 r
F
.59291 .55418 m .60792 .59794 L .61093 .61544 L 0 .234 .731 r
F
.86062 .30627 m .87883 .22454 L .89152 .24983 L .768 .939 .938 r
F
.59291 .55418 m .59041 .52837 L .60792 .59794 L 0 .308 .799 r
F
.5752 .4977 m .59041 .52837 L .59291 .55418 L .084 .435 .875 r
F
.54725 .41048 m .55936 .45338 L .54322 .42841 L .26 .555 .931 r
F
.54725 .41048 m .54322 .42841 L .52559 .38172 L .29 .6 .95 r
F
.50265 .36046 m .50655 .33188 L .52559 .38172 L .357 .638 .961 r
F
.50265 .36046 m .48108 .31262 L .50655 .33188 L .391 .678 .974 r
F
.84839 .2901 m .87883 .22454 L .86062 .30627 L .777 .965 .924 r
F
.82894 .3742 m .84839 .2901 L .86062 .30627 L .785 .966 .92 r
F
.5752 .4977 m .57063 .45833 L .59041 .52837 L .139 .449 .881 r
F
.55936 .45338 m .57063 .45833 L .5752 .4977 L .246 .533 .92 r
F
.79042 .47415 m .80764 .41122 L .80355 .4495 L .801 .999 .844 r
F
.80764 .41122 m .82894 .3742 L .80355 .4495 L .785 .989 .889 r
F
.80355 .4495 m .78634 .50564 L .79042 .47415 L .793 .999 .843 r
F
.78634 .50564 m .77837 .52348 L .79042 .47415 L .795 .996 .782 r
F
.78634 .50564 m .77378 .55174 L .77837 .52348 L F
.45843 .25371 m .48108 .31262 L .45267 .30009 L .441 .704 .977 r
F
.42401 .25397 m .45843 .25371 L .45267 .30009 L .472 .74 .987 r
F
.39273 .21289 m .42401 .25397 L .39368 .2604 L .474 .757 .992 r
F
.39368 .2604 m .36269 .22688 L .39273 .21289 L .504 .781 .995 r
F
.54725 .41048 m .57063 .45833 L .55936 .45338 L .235 .542 .926 r
F
.82472 .34767 m .82894 .3742 L .80764 .41122 L .809 .984 .889 r
F
.82472 .34767 m .84839 .2901 L .82894 .3742 L .786 .979 .908 r
F
.60437 .56346 m .60792 .59794 L .59041 .52837 L .108 .403 .855 r
F
.54725 .41048 m .52559 .38172 L .53035 .35503 L .351 .606 .944 r
F
.53035 .35503 m .52559 .38172 L .50655 .33188 L .383 .648 .962 r
F
.60437 .56346 m .59041 .52837 L .59113 .50083 L .179 .423 .858 r
F
.59113 .50083 m .59041 .52837 L .57063 .45833 L .266 .512 .903 r
F
.50655 .33188 m .48108 .31262 L .48859 .2862 L .429 .675 .966 r
F
.45843 .25371 m .48859 .2862 L .48108 .31262 L .447 .707 .977 r
F
.79042 .47415 m .77837 .52348 L .77582 .50635 L .849 .988 .854 r
F
.54725 .41048 m .55409 .38906 L .57063 .45833 L .343 .57 .923 r
F
.53035 .35503 m .55409 .38906 L .54725 .41048 L .367 .614 .945 r
F
.79042 .44536 m .79042 .47415 L .77582 .50635 L .851 .987 .854 r
F
.80764 .41122 m .79042 .47415 L .79042 .44536 L .831 .98 .88 r
F
.59113 .50083 m .57063 .45833 L .57652 .43093 L .32 .521 .895 r
F
.55409 .38906 m .57652 .43093 L .57063 .45833 L .358 .577 .924 r
F
.42401 .25397 m .42311 .19873 L .45843 .25371 L .484 .735 .982 r
F
.39273 .21289 m .42311 .19873 L .42401 .25397 L .522 .767 .986 r
F
.59113 .50083 m .60837 .55034 L .60437 .56346 L .3 .488 .876 r
F
.53035 .35503 m .50655 .33188 L .51121 .29637 L .42 .646 .951 r
F
.50655 .33188 m .48859 .2862 L .51121 .29637 L .45 .683 .964 r
F
.80764 .41122 m .79042 .44536 L .80724 .37539 L .838 .977 .879 r
F
.82472 .34767 m .80764 .41122 L .80724 .37539 L .816 .967 .902 r
F
.84839 .2901 m .86588 .19871 L .87883 .22454 L .797 .942 .922 r
F
.83096 .28387 m .84839 .2901 L .82472 .34767 L .81 .953 .911 r
F
.83096 .28387 m .86588 .19871 L .84839 .2901 L .802 .955 .916 r
F
.59626 .47657 m .60837 .55034 L .59113 .50083 L .34 .499 .871 r
F
.59626 .47657 m .59113 .50083 L .57652 .43093 L .375 .55 .899 r
F
.53035 .35503 m .53736 .32239 L .55409 .38906 L .419 .616 .932 r
F
.53035 .35503 m .51121 .29637 L .53736 .32239 L .447 .657 .95 r
F
.79042 .44536 m .77582 .50635 L .77357 .4858 L .858 .962 .871 r
F
.83096 .28387 m .82472 .34767 L .80724 .37539 L .821 .964 .9 r
F
.59626 .47657 m .61316 .52804 L .60837 .55034 L .378 .521 .874 r
F
.48859 .2862 m .48591 .23869 L .51121 .29637 L .462 .678 .958 r
F
.45843 .25371 m .48591 .23869 L .48859 .2862 L .501 .708 .963 r
F
.56215 .35923 m .57652 .43093 L .55409 .38906 L .413 .584 .91 r
F
.53736 .32239 m .56215 .35923 L .55409 .38906 L .438 .624 .931 r
F
.79042 .44536 m .77357 .4858 L .78748 .41945 L .861 .956 .869 r
F
.78748 .41945 m .80724 .37539 L .79042 .44536 L .838 .951 .891 r
F
.45385 .18441 m .48591 .23869 L .45843 .25371 L .498 .714 .968 r
F
.45843 .25371 m .42311 .19873 L .45385 .18441 L .54 .745 .971 r
F
.58452 .40163 m .59626 .47657 L .57652 .43093 L .419 .557 .887 r
F
.58452 .40163 m .57652 .43093 L .56215 .35923 L .444 .599 .91 r
F
.80724 .37539 m .78748 .41945 L .80178 .34494 L .842 .943 .888 r
F
.80724 .37539 m .80178 .34494 L .83096 .28387 L .819 .938 .906 r
F
.51121 .29637 m .51864 .24386 L .53736 .32239 L .477 .652 .936 r
F
.51121 .29637 m .48591 .23869 L .51864 .24386 L .508 .69 .95 r
F
.59626 .47657 m .60364 .45009 L .61316 .52804 L .427 .534 .863 r
F
.58452 .40163 m .60364 .45009 L .59626 .47657 L .451 .575 .888 r
F
.78748 .41945 m .77357 .4858 L .77163 .46341 L .854 .928 .876 r
F
.60364 .45009 m .61874 .49037 L .61316 .52804 L .469 .559 .865 r
F
.54759 .29407 m .56215 .35923 L .53736 .32239 L .472 .623 .917 r
F
.51864 .24386 m .54759 .29407 L .53736 .32239 L .488 .656 .935 r
F
.78748 .41945 m .77163 .46341 L .77991 .40269 L .855 .915 .872 r
F
.80178 .34494 m .78748 .41945 L .77991 .40269 L .832 .917 .893 r
F
.83096 .28387 m .85265 .17234 L .86588 .19871 L .808 .942 .914 r
F
.57258 .31685 m .58452 .40163 L .56215 .35923 L .476 .6 .897 r
F
.54759 .29407 m .57258 .31685 L .56215 .35923 L .498 .635 .914 r
F
.77991 .40269 m .77163 .46341 L .76058 .45931 L .837 .889 .878 r
F
.80178 .34494 m .8156 .2544 L .83096 .28387 L .822 .932 .904 r
F
.8156 .2544 m .85265 .17234 L .83096 .28387 L .804 .932 .917 r
F
.48591 .23869 m .48494 .16993 L .51864 .24386 L .512 .687 .946 r
F
.45385 .18441 m .48494 .16993 L .48591 .23869 L .554 .722 .952 r
F
.58452 .40163 m .59569 .37811 L .60364 .45009 L .483 .58 .876 r
F
.57258 .31685 m .59569 .37811 L .58452 .40163 L .494 .611 .897 r
F
.6133 .42737 m .61874 .49037 L .60364 .45009 L .491 .562 .856 r
F
.59569 .37811 m .6133 .42737 L .60364 .45009 L .504 .594 .877 r
F
.6133 .42737 m .62842 .48036 L .61874 .49037 L .53 .588 .857 r
F
.77991 .40269 m .78897 .33514 L .80178 .34494 L .832 .906 .889 r
F
.78897 .33514 m .8156 .2544 L .80178 .34494 L .81 .913 .91 r
F
.76866 .39402 m .77991 .40269 L .76058 .45931 L .835 .875 .872 r
F
.78897 .33514 m .77991 .40269 L .76866 .39402 L .817 .884 .893 r
F
.76866 .39402 m .76058 .45931 L .75094 .44899 L .818 .853 .874 r
F
.54759 .29407 m .5471 .23103 L .57258 .31685 L .502 .633 .911 r
F
.51864 .24386 m .5471 .23103 L .54759 .29407 L .531 .658 .917 r
F
.62502 .40747 m .63999 .46407 L .62842 .48036 L .554 .602 .856 r
F
.62502 .40747 m .62842 .48036 L .6133 .42737 L .538 .59 .854 r
F
.5164 .15527 m .5471 .23103 L .51864 .24386 L .528 .667 .925 r
F
.51864 .24386 m .48494 .16993 L .5164 .15527 L .565 .696 .929 r
F
.62502 .40747 m .6133 .42737 L .60819 .35893 L .544 .611 .87 r
F
.60819 .35893 m .6133 .42737 L .59569 .37811 L .524 .597 .869 r
F
.60819 .35893 m .59569 .37811 L .59163 .31674 L .529 .617 .884 r
F
.57258 .31685 m .59163 .31674 L .59569 .37811 L .528 .617 .884 r
F
.57258 .31685 m .57914 .23728 L .59163 .31674 L .525 .624 .893 r
F
.57258 .31685 m .5471 .23103 L .57914 .23728 L .546 .647 .9 r
F
.76866 .39402 m .75094 .44899 L .75653 .38572 L .814 .838 .868 r
F
.75653 .38572 m .75094 .44899 L .74274 .43458 L .798 .819 .869 r
F
.63913 .38573 m .65347 .43728 L .63999 .46407 L .589 .622 .852 r
F
.62502 .40747 m .63913 .38573 L .63999 .46407 L .569 .606 .85 r
F
.78897 .33514 m .76866 .39402 L .77464 .33314 L .815 .87 .887 r
F
.76866 .39402 m .75653 .38572 L .77464 .33314 L .797 .849 .89 r
F
.79261 .28278 m .8156 .2544 L .78897 .33514 L .807 .891 .904 r
F
.79261 .28278 m .78897 .33514 L .77464 .33314 L .798 .882 .907 r
F
.8156 .2544 m .83913 .14539 L .85265 .17234 L .8 .942 .92 r
F
.62098 .3298 m .63913 .38573 L .62502 .40747 L .575 .627 .866 r
F
.62098 .3298 m .62502 .40747 L .60819 .35893 L .555 .612 .865 r
F
.60234 .29436 m .62098 .3298 L .60819 .35893 L .562 .631 .877 r
F
.60819 .35893 m .59163 .31674 L .60234 .29436 L .544 .619 .878 r
F
.75653 .38572 m .74274 .43458 L .74213 .38386 L .787 .794 .858 r
F
.74213 .38386 m .74274 .43458 L .72601 .44046 L .766 .778 .863 r
F
.79261 .28278 m .80254 .21385 L .8156 .2544 L .796 .909 .918 r
F
.80254 .21385 m .83913 .14539 L .8156 .2544 L .786 .908 .925 r
F
.59163 .31674 m .57914 .23728 L .60234 .29436 L .551 .638 .89 r
F
.65589 .37264 m .66628 .43942 L .65347 .43728 L .62 .643 .851 r
F
.65589 .37264 m .65347 .43728 L .63913 .38573 L .601 .624 .847 r
F
.54823 .14044 m .57914 .23728 L .5471 .23103 L .544 .649 .903 r
F
.5164 .15527 m .54823 .14044 L .5471 .23103 L .572 .674 .908 r
F
.77464 .33314 m .75653 .38572 L .75964 .32064 L .793 .833 .883 r
F
.75964 .32064 m .75653 .38572 L .74213 .38386 L .772 .814 .886 r
F
.7266 .38278 m .74213 .38386 L .72601 .44046 L .757 .759 .854 r
F
.7266 .38278 m .72601 .44046 L .71148 .43565 L .733 .74 .857 r
F
.6743 .37004 m .68152 .43668 L .66628 .43942 L .642 .659 .851 r
F
.6743 .37004 m .66628 .43942 L .65589 .37264 L .632 .648 .848 r
F
.63943 .30414 m .65589 .37264 L .63913 .38573 L .601 .645 .865 r
F
.63913 .38573 m .62098 .3298 L .63943 .30414 L .582 .628 .863 r
F
.77464 .33314 m .78068 .27872 L .79261 .28278 L .796 .869 .902 r
F
.77464 .33314 m .75964 .32064 L .78068 .27872 L .774 .842 .904 r
F
.69344 .36957 m .69921 .42459 L .68152 .43668 L .671 .683 .852 r
F
.6743 .37004 m .69344 .36957 L .68152 .43668 L .662 .67 .847 r
F
.7266 .38278 m .71148 .43565 L .71115 .37404 L .728 .729 .851 r
F
.71115 .37404 m .71148 .43565 L .69921 .42459 L .705 .711 .853 r
F
.78068 .27872 m .80254 .21385 L .79261 .28278 L .775 .879 .923 r
F
.74096 .3386 m .75964 .32064 L .74213 .38386 L .761 .788 .875 r
F
.74096 .3386 m .74213 .38386 L .7266 .38278 L .746 .778 .879 r
F
.62098 .3298 m .61164 .23042 L .63943 .30414 L .586 .647 .877 r
F
.60234 .29436 m .61164 .23042 L .62098 .3298 L .561 .631 .878 r
F
.60234 .29436 m .57914 .23728 L .61164 .23042 L .567 .639 .882 r
F
.71115 .37404 m .69921 .42459 L .69344 .36957 L .696 .696 .846 r
F
.65589 .37264 m .66275 .29991 L .6743 .37004 L .628 .665 .867 r
F
.66275 .29991 m .65589 .37264 L .63943 .30414 L .612 .648 .862 r
F
.58043 .12543 m .61164 .23042 L .57914 .23728 L .565 .647 .89 r
F
.54823 .14044 m .58043 .12543 L .57914 .23728 L .576 .657 .892 r
F
.74096 .3386 m .7266 .38278 L .72765 .32326 L .741 .765 .872 r
F
.7266 .38278 m .71115 .37404 L .72765 .32326 L .715 .743 .874 r
F
.7734 .24524 m .78068 .27872 L .75964 .32064 L .774 .857 .912 r
F
.74839 .28204 m .7734 .24524 L .75964 .32064 L .744 .81 .904 r
F
.74096 .3386 m .74839 .28204 L .75964 .32064 L .756 .812 .897 r
F
.68602 .30259 m .69344 .36957 L .6743 .37004 L .655 .689 .869 r
F
.68602 .30259 m .6743 .37004 L .66275 .29991 L .643 .672 .863 r
F
.7734 .24524 m .80254 .21385 L .78068 .27872 L .77 .856 .915 r
F
.70849 .3083 m .72765 .32326 L .71115 .37404 L .711 .734 .869 r
F
.70849 .3083 m .71115 .37404 L .69344 .36957 L .685 .715 .872 r
F
.69344 .36957 m .68602 .30259 L .70849 .3083 L .676 .7 .865 r
F
.74096 .3386 m .72765 .32326 L .74839 .28204 L .725 .773 .89 r
F
.66275 .29991 m .63943 .30414 L .6462 .2073 L .606 .665 .881 r
F
.63943 .30414 m .61164 .23042 L .6462 .2073 L .588 .647 .876 r
F
.80254 .21385 m .82532 .11785 L .83913 .14539 L .77 .939 .937 r
F
.79038 .16717 m .82532 .11785 L .80254 .21385 L .749 .887 .942 r
F
.7734 .24524 m .79038 .16717 L .80254 .21385 L .753 .888 .94 r
F
.74839 .28204 m .72765 .32326 L .73008 .25469 L .727 .78 .895 r
F
.73008 .25469 m .72765 .32326 L .70849 .3083 L .691 .748 .895 r
F
.61303 .11025 m .6462 .2073 L .61164 .23042 L .587 .659 .887 r
F
.58043 .12543 m .61303 .11025 L .61164 .23042 L .577 .65 .886 r
F
.6747 .24139 m .68602 .30259 L .66275 .29991 L .633 .69 .885 r
F
.6462 .2073 m .6747 .24139 L .66275 .29991 L .621 .672 .877 r
F
.75815 .20823 m .7734 .24524 L .74839 .28204 L .742 .834 .921 r
F
.74839 .28204 m .73008 .25469 L .75815 .20823 L .703 .785 .915 r
F
.70257 .2267 m .73008 .25469 L .70849 .3083 L .688 .739 .89 r
F
.68602 .30259 m .70257 .2267 L .70849 .3083 L .66 .721 .894 r
F
.6747 .24139 m .70257 .2267 L .68602 .30259 L .643 .694 .882 r
F
.75815 .20823 m .79038 .16717 L .7734 .24524 L .722 .832 .933 r
F
.73503 .17628 m .75815 .20823 L .73008 .25469 L .704 .798 .923 r
F
.70257 .2267 m .73503 .17628 L .73008 .25469 L .657 .751 .919 r
F
.6747 .24139 m .67607 .16542 L .70257 .2267 L .64 .711 .899 r
F
.6462 .2073 m .67607 .16542 L .6747 .24139 L .603 .679 .895 r
F
.6462 .2073 m .64601 .09488 L .67607 .16542 L .603 .68 .895 r
F
.61303 .11025 m .64601 .09488 L .6462 .2073 L .576 .657 .892 r
F
.79038 .16717 m .8112 .08971 L .82532 .11785 L .714 .929 .962 r
F
.75815 .20823 m .77134 .12562 L .79038 .16717 L .711 .866 .956 r
F
.73503 .17628 m .77134 .12562 L .75815 .20823 L .662 .801 .949 r
F
.70257 .2267 m .70774 .13599 L .73503 .17628 L .657 .754 .921 r
F
.70257 .2267 m .67607 .16542 L .70774 .13599 L .611 .709 .914 r
F
.77134 .12562 m .8112 .08971 L .79038 .16717 L .689 .863 .964 r
F
.67607 .16542 m .67939 .07933 L .70774 .13599 L .611 .714 .918 r
F
.67607 .16542 m .64601 .09488 L .67939 .07933 L .571 .677 .911 r
F
.73503 .17628 m .74035 .10394 L .77134 .12562 L .661 .81 .955 r
F
.73503 .17628 m .70774 .13599 L .74035 .10394 L .608 .757 .949 r
F
.70774 .13599 m .71318 .06359 L .74035 .10394 L .608 .759 .951 r
F
.70774 .13599 m .67939 .07933 L .71318 .06359 L .56 .71 .941 r
F
.79677 .06094 m .8112 .08971 L .77134 .12562 L .623 .902 .985 r
F
.77134 .12562 m .76754 .07258 L .79677 .06094 L .616 .838 .985 r
F
.74035 .10394 m .76754 .07258 L .77134 .12562 L .585 .817 .988 r
F
.74035 .10394 m .74739 .04765 L .76754 .07258 L .587 .813 .986 r
F
.74035 .10394 m .71318 .06359 L .74739 .04765 L .534 .753 .977 r
F
.76754 .07258 m .78201 .03152 L .79677 .06094 L .513 .857 .994 r
F
.76754 .07258 m .74739 .04765 L .78201 .03152 L .486 .791 .999 r
F
0 g
.79826 0 m
.99671 .40534 L
s
.99671 .40534 m
1 .76812 L
s
1 .76812 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.00338 .37369 m
0 .74137 L
s
0 .74137 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.79826 0 m
.00338 .37369 L
s
.00338 .37369 m
.79826 0 L
s
.01971 .36602 m
.0195 .38939 L
s
[(0)] .02012 .31927 -0.03303 1 Mshowa
.29667 .23581 m
.29658 .25919 L
s
[(2)] .29685 .18904 -0.01416 1 Mshowa
.60686 .08998 m
.6069 .11337 L
s
[(4)] .60677 .04321 .00709 1 Mshowa
.125 Mabswid
.08613 .33479 m
.08602 .34881 L
s
.15438 .3027 m
.15429 .31673 L
s
.22453 .26972 m
.22446 .28375 L
s
.37088 .20092 m
.37084 .21495 L
s
.44725 .16502 m
.44723 .17905 L
s
.52587 .12806 m
.52588 .14209 L
s
.69031 .05075 m
.69036 .06478 L
s
.77634 .01031 m
.77641 .02434 L
s
gsave
.37156 -0.0797 -66.5429 -15.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(x) show
70.187500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
0 .74137 m
.2826 1 L
s
.00774 .74845 m
.00796 .72508 L
s
[(0)] .00732 .7952 .03362 -1 Mshowa
.1058 .8382 m
.10597 .81482 L
s
[(1)] .10546 .88495 .02692 -1 Mshowa
.1951 .91992 m
.19523 .89654 L
s
[(2)] .19484 .96668 .02085 -1 Mshowa
.27675 .99465 m
.27685 .97127 L
s
[(3)] .27656 1.04142 .01531 -1 Mshowa
.125 Mabswid
.02812 .7671 m
.02824 .75308 L
s
.0481 .78539 m
.04822 .77136 L
s
.0677 .80333 m
.06781 .7893 L
s
.08693 .82093 m
.08704 .8069 L
s
.12432 .85515 m
.12442 .84112 L
s
.1425 .87179 m
.1426 .85776 L
s
.16035 .88812 m
.16044 .8741 L
s
.17788 .90417 m
.17797 .89014 L
s
.21201 .9354 m
.21208 .92137 L
s
.22862 .9506 m
.22869 .93657 L
s
.24494 .96554 m
.24501 .95151 L
s
.26099 .98022 m
.26105 .96619 L
s
gsave
.14967 1.16054 -66.7909 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.125000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(y) show
70.312500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/0P10`4000010?ooo`000<c<c;no_`00
E@00ZW=cLiVIVNfXL<V>GoOI`7eI>ooooiVIVH:2P]7AdTi>C_ooohVHZWB1T<[AfDI=E_ooofQ_PUQN
K[ZmaS4e?_oooiBDU:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJF
UWemOFATI4];Bc8b<P0T000TE@0TZP0To`190019E@19ZP19o`1]001]E@1]ZP1]o`2B002BE@2BZP2B
o`2f002fE@2fZP2fo`3K003KE@3KZP3Ko`3oE@3oZUD005D0EED0ZUD0oeDT05DTEEDTZUDToeE905E9
EEE9ZUE9oeE]05E]EEE]ZUE]oeFB05FBEEFBZUFBoeFf05FfEEFfZUFfoeGK05GKEEGKZUGKoeGo05Go
EEGoZUGoojX00:X0EJX0ZZX0ojXT0:XTEJXTZZXTojY90:Y9EJY9ZZY9ojY]0:Y]EJY]ZZY]ojZB0:ZB
EJZBZZZBojZf0:ZfEJZfZZZfoj[K0:[KEJ[KZZ[Koj[o0:[oEJ[oZZ[oool0EOl0Z_lT0?lTEOlTZ_lT
oom90?m9EOm9Z_m9oom]0?m]EOm]Z_m]oonB0?nBEOnBZ_nBoonf0?nfEOnfZ_nfoooK0?oKEOoKZ_oK
ooooEOooZ/d0003=0<g=0000cLd0c@3=cNGUiDe=C@00:P00O`00e00Z000Z:P0ZE@0ZO`0ZZP0Ze01E
001E:P1EE@1EO`1EZP1Ee01o001o:P1oE@1oO`1oZP1oe02Z002Z:P2ZE@2ZO`2ZZP2Ze03D003D:P3D
E@3DO`3DZP3De2X002X0:RX0EBX0ObX0ZRX0e2XZ02XZ:RXZEBXZObXZZRXZe2YE02YE:RYEEBYEObYE
ZRYEe2Yo02Yo:RYoEBYoObYoZRYoe2ZZ02ZZ:RZZEBZZObZZZRZZe2[D02[D:R[DEB[DOb[DZR[De5D0
:UD0OeD0e5DZ05DZ:UDZEEDZOeDZZUDZe5EE05EE:UEEEOl0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000;h00`4200<1jP000;l000<100000P000`40003Y0000`0001040007/0000
`@020Nd000310081k@000<0000@10001k0000;l000<100000P000`40003Y0000_P030@800`7Z0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`1;0081I0000?l0
B@020@03004006<0003o04L00P430081H`000?l0A@020@D00P5S0000o`0>00@1<@030@H000<10040
HP000?l040000`40000^00810P000`40000400030@010680003o010000<10000;0020@@000<10000
10001040005Q0000o`0;00L1:`020@H000<1000010001040005Q0000o`0;00030@00008000<10000
9`030@P000<1000010001@4000010600003o00`000D100000@0W00814P001@4000010600003o00`0
00D100000@0U008150000`40000200030@0005d0003o00d000@100018`020@l000<T000010000`40
000200030@0005d0003o00d000@100018@020@h000DV02H09P0500030@0000<000<10000G0000?l0
3P000`400@0O00813`001b@0900T02H010000`40000300030@0005`0003o00h000<100407@020@h0
00TV02H09P0V02H01@000`40000400030@0005/0003o00l00P4K00813@003B@0900T02@0900V02H0
10000`40000400030@0005/0003o02T00`4>000?9P0V02H09P0V02H09P0V00<000<100001@000`40
001J0000o`0W00<13@004b@0900V02@0900T02@09P0V02H00P000`40000500030@0005X0003o02D0
0P4200030@0000X001<V02H09P0V02H09P0V02H09P0V00<000<100001P000`40001I0000o`0S0081
10000`400007000G900T02@0900V02H0900T02@09P0V02H00P000`40000600030@0005T0003o0240
0P4600030@0000@001dV02H09P0V02H09P0V02H09P0V02H09P0V02H00@0900030@0005P0003o01l0
0P4800030@0000<001hT02@0900T02@0900V02H09P0T02H09P0V02H09P4900030@0005P0003o01d0
0P4=00039P0V00<001<V02H09P0V02H09P0V02H09P0V00L000<1000020000`40001G0000o`0K0081
3P008R@0900T02@0900T02@0900V02H09P0V02H09P0V02H09P4:00030@0005L0003o01P00`4=0007
9P0V02H09P05000?9P0V02H09P0V02H09P0V00T000<100002@000`40001F0000o`0F00813`009b@0
900T02@0900T02@0900T02@09P0V02H09P0V02H09P0V02H19P0:00030@0005H0003o01@00P4>000;
9P0002H09P0V02H01`003BH09P0V02H09P0V02H02@000`40000:00030@0005D0003o01800P4?000[
900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P0V02H19P0;00030@0005D0003o0100
0P4>00059P0V02H00`001bH09P0V02H01`003BH09P0V02H09P0V02H02@000`40000;00030@0005@0
003o00h00P4?000a900T02@0900T02@0900T02@0900T02@0900T02H09P0V02H09P0V02H09P0V0BH0
9P0:00030@0005@0003o00`00`4=00099P0V02H09P0V00D000DV02H09P09000;9P0V02H09P0V02H0
2@000`40000<00030@0005<0003o00X00P4200030@0000P003LT02@0900T02@0900T02@0900T02@0
900T02@0900T02@09P0V02H09P0V02H09P0V02H19P0V00/000<10000D`000?l01`030@@000<10000
1`003BH09P0V02H09P0V02H01@001BH09P0V00/000TV02H09P0V02H02@000`40000=00030@000580
003o00D00P4700030@0000@003dT02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V02H0
9P0V02H09P0V02H09P0V0BH09P0V00X000<10000DP000?l00`020@T000<100000`001bH09P0V02H0
0`001bH09P0V02H01`000bH09P0=00079P0V02H09P0900030@0000d000<10000DP000?l00@020@/0
04@1000T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P0V
02H09P4V02H09P/000<10000D@000?h00P4=000=0@0V02H09P0V02H09P0300079P0V02H09P070003
9P0V00l000DV02H09P0900030@0000h000<10000D@000?`00P4>0017900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900V02H09P0V02H09P0V02H09P0V0BH09P0V02H02P000`40
001@0000nP020@d0014V02H09P0V02H09P0V02H09P0500059P0V02H02@000bH0000A00039P0000L0
00<100003`000`40001@0000n0020@h004/T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900V02H09P0V02H09P0V02H09P0V0BH09P0V02H02`000`40001?0000m@030@d0
01DV02H09P0V02H09P0V02H09P0V02H01@001BH09P0V00T000<V00006`000`40000@00030@0004l0
003c00813`00DB@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900V02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0:00030@0004h0003a00<13@006BH09P0V02H0
9P0V02H09P0V02H09P0V02H01@001BH09P0V02L000<100004@000`40001>0000k`020@8000<10000
2P00EB@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
9P0V02H09P0V02H09P0V02H09P4V02H09P0V02H02`000`40001=0000k@020@@000<100001`007BH0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V00L000<V02H09`000`40000B00030@0004d0003[0081
1P000`400006001K900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@09P0V02H09P0V02H09P0V02H09P4V02H09P0V02H09P0:00030@0004`0003Y0081
20000`400003000Q9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V00L000<V02H09`000`40
000C00030@0004`0003W00813`00Gb@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900V02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V00/0
00<10000B`000>@00`4>000U9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P090003
9P0002D000<1000050000`40001;0000hP020A0006<T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02H09P0V02H09P0V02H09P0V02H1
9P0V02H09P0V02H030000`40001:0000YP080C800P4?000Y9P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H02@000bH0000U00030@0001D000<10000BP000:L000<100000`000`40
000^00813P00JB@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P0V02H09P4V02H09P0V02H09P0=0003
0@0004T0002X00030@0003400P4=000_9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H02@000bH0000U00030@0001H000<10000B@000:T000<10000;P020@h006dT02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900V02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V00h000<10000B0000:X0
00<10000:`030@`003<V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H0<@000`40000G00030@0004P0002[00030@0002P00P4200030@0000L007<T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900V02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V00l000<10000A`000:`0
00<1000090030@@000<100001P00=bH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H0<@000`40000H00030@0004L0002]00030@0002400P4700030@0000<0
07DT02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P0V02H09P4V02H09P0V02H0
4P000`4000160000Y`000`40000300030@0001l00P4>000k9P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0<@000`40000I00030@0004H0002W0003
0@0000<000<100007@020@d007TT02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H0
9P0V02H09P0V02H19P0V02H09P0V01<000<10000A@000:P000D100000@0N00813P00?bH09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0a0003
0@0001X000<10000A@000:T00`4M00813`00Nb@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0D00030@0004@0003700813P00@bH09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
<@000`40000K00030@0004@0003500813`00Ob@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02H09P0V02H09P0V02H09P4V02H09P0V02H05@000`4000130000`P030@h004LV02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0a00030@0001`000<10000@`000<000`4?0021900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P4V02H09P0V00d000<V00001`000`40
00130000_P020@8000<100002P00BbH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0a00030@0001d000<10000@P000;`0
0P4400030@0000L008LT02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900V02H09P0V02H09P0V0BH09P0V02H03@000bH0000800030@000480002j00811P000`40
0004001A9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V034000<100007P000`4000110000^0020@P000<10000
0`00Rb@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900V02H09P0V02H09P0V0BH09P0V02H03@000bH09P0900030@000440002f00813@00EBH09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H0<@000`40000O00030@000400002d00813P00Sb@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H0
9P4V02H09P0V00d000<V02H02P000`4000100000/@030@d005TV02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0a00030@00020000<10000?`000:l00P4?002A900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P4V02H0
9P0?00059P0V02H02@000`40000o0000[@020@h005dV02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H0<@000`40000Q00030@0003h0002[00813`00UB@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V02H09P0V02H09P0V0BH0
9P0V00l000DV02H09P0:00030@0003h0002Y00813P00HBH09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H0<@000`40000R00030@0003d0002W008100<00@002P00Vb@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900V02H09P0V02H09P0V0BH09P0V00d000TV02H09P0V02H02@000`40000m0000Y@020@<000<10000
2000IBH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V034000<100008`000`40
000l0000X`020@D000<100001@00Wb@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P4V02H0
9P0=00099P0V02H09P0V00X000<10000?0000:000`4700030@00008006/V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0a00030@0002@000<10000>`0009h00P4:002U
0@0T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02H09P0V02H09P0V02H19P0V00l000/V02H09P0V02H0
9P0900030@0003/0002L00813000KP4V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V<`000`40000U00030@0003X0002J00813P00Y@409P0T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@09P0V02H09P0V02H09P0V0BH09P0?000;9P0V02H09P0V02H02P000`40000j0000V0020@d0
074V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0c0003
0@0002H000<10000>@0009H00P4>002W9P0V02H09P0T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H0
9P0V02H09P0V0BH04@003BH09P0V02H09P0V02H02@000`40000i0000T`030@d007DV02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0<`000`40000W0003
0@0003P0002B00813P00ZRH0900V02H09P0V02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V02H09P0V02H0
9P0V02H14P003BH09P0V02H09P0V02H02P000`40000h0000S`030@03004000X007TV02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V03<000<10000
:0000`40000g0000S@020@@000<100002000[RH09P0V02@0900V02H09P0T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@09P0V02H09P0V02H09P0V0A00014V02H09P0V02H09P0V02H09P0900030@0003L0002;0081
1P000`400005001m9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H0<`000`40000Y00030@0003H00029008120000`400002002a9P0V02H0
9P0V02H0900T02@0900V02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V02H09P0V02H09P0V008000<10000
3P004BH09P0V02H09P0V02H09P0V00X000<10000=P0005000`4d00813P00PBH09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0c0003
0@0002X000<10000=@0004l000D100000@0a00813@00]BH09P0V02H0900T02H09P0T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02H09P0V02H09P0V02H00P000`40000>000C9P0V02H09P0V02H0
9P0V02H09P0900030@0003D0001>00030@0000<000<10000;0020@`008LV02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
<`000`40000[00030@0003@0001>00030@0000<000<10000:P020@d00;TV02H09P0V02H09P0T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P0200030@0000h0
01<V02H09P0V02H09P0V02H09P0V00X000<10000=00004h000<100000`000`40000W00<13000RbH0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H0<`000`40000/00030@0003<0001>00030@0000<000<100009@020@h0
0;dV02H09P0V02H09P0V02H0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@092HT02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V02H0
9P0V02H09P0V02H00P000`40000<000G9P0V02H09P0V02H09P0V02H09P0V02H02@000`40000c0000
CP000`40000300030@0002<00P400`01000:002?9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0<`000`40
000/00030@0003<0001>00030@0000<000<100008@020@<000<100002000`BH09P0V02H09P0V02H0
9P0V02H0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@V900T02@0
92HT9R@0900T02@V900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V02H09P0V02H09P0V02H0
0P000`40000<000G9P0V02H09P0V02H09P0V02H09P0V02H02P000`40000b0000CP000`4000030003
0@0001l00P4400030@0000H009<V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0<`000`40000]0003
0@000380001>00030@0000<000<100007@020@H000<100001@00aBH09P0V02H09P0V02H09P0V02H0
9P0V02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@V900T02@0900T9R@V900T02@V
92HT9R@0900T9R@0900T02@V900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P0V0080
00<100002P006bH09P0V02H09P0V02H09P0V02H09P0V02H09P0900030@000340001?00050@000040
7P020@d009LV02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V03<000<10000;P000`40000a0000
D0030Ad00P4>00399P0V02H09P0V02H09P0V02H09P0V02H09P0V02@0900T02@0900T02@0900T02@0
900T02@092HT02@0900T02@V92HT02@0900T9R@V92HT02@V92HT9R@0900T9R@V900T02@V92HT02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@09P0V02H09P0V02H09P0V008000<100002P006bH09P0V02H09P0V02H0
9P0V02H09P0V02H09P0:00030@000300001]00<13@00VbH09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V03<000<100009@000bH0000700030@000300001[00813`00cBH09P0V02H09P0V02H0
9P0V02H09P0V02@09P0V02H0900T02@0900T02@0900T02@0900T02@0900T9R@V900T02@092HT9R@V
900T02@V92HT9R@V92HT9R@V92HT02@V92HT9R@092HT9R@V92HT02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V02@0
9P0V02H09P0V02H00P000`400008000O9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P090003
0@0002l0001Y00813P00WbH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0c0003
0@0002<000DV02H09P0800030@0002l0001W00813@00dBH09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02@0900T02H0900T02@0900T02@0900T02@0900T02@0900T9R@V92HT02@092HT9R@V92HT02@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@092HT9R@V92HT9R@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H0
9P0V00@000<1000020007bH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H02P000`40000^0000
I@030@d00:<V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V03<000<10000
8@001bH09P0V02H02@000`40000^0000H`020@8000<100002000dBH09P0V02H09P0V02H09P0V02H0
9P0V02@09P0V02H09P0T02@0900T02@0900T02@0900T02@0900T02@0900T02@092HT9R@V92HT02@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02H09P0V00P000<100001P008bH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V00T0
00<10000;@0006400P4400030@0000L00:LV02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0c00030@0001l000/V02H09P0V02H09P0800030@0002d0001O00811P000`40
0004003E9P0V02H09P0V02H09P0V02H09P0V02H09P0V02@0900T02@09P0T02@0900T02@0900T02@0
900T02@092HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900V02H09P0V00P000<100001P008bH09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V00X000<10000;00005`00`48002a0@0002H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V03<000<100006`003bH0
9P0V02H09P0V02H09P0900030@0002`0001J00813P00fBH09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0T02@0900T02@0900T02@0900T02@0900T9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02H09P0V02H020000`400006000U9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0900030@0002/0001H00813@00/BH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0c00030@0001T001<V02H09P0V02H09P0V02H09P0V00P000<10000
:`0005H00P4>003M9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0T02@0900T02@0
900T02@0900T02@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02H09P0V02H020000`40
0006000W9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V00P000<10000:P0005@0
0P4=002e9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0900V02H09P0V02H09P0V02H0
9P0V02H09P0c00030@0001L001DV02H09P0V02H09P0V02H09P0V02H02@000`40000Z0000D@030@h0
0=lV02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0T02@0900T02@0900T02@0
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02H09P0V00X000<100001P009bH0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0900030@0002T0001@00813@00^BH0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0T02H09P0V02H09P0V02@09P0V02H0
9P0V03<000<100005@006BH09P0V02H09P0V02H09P0V02H09P0V02H020000`40000Y0000CP020@03
004000T00>DV02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0T02@0
900T02@0900T02@092HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V02H09P0V00X0
00<100001P00:BH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V00P000<10000
:00004`00P4300030@0000L00;dV02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02@0900V02H09P0V02@0900V02H09P0V02H0<`000`40000?000O9P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0900030@0002P0001<00<10P000`400004003Y9P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02@0900T02@0900T02@0900T9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02H09P0V02H02P000`400006000Y9P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H02@000`40000W0000@@030@P000<100400P000`40
000300319P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02@0900T02H0
9P0V02H0900T02H09P0V02H09P0c00030@0000l0024V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H020000`40000W0000@0001@40000100L000H100010042003]9P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0T02@0900T02@0900T02@092HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900V02H09P0V00X000<100001P00:bH09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H020000`40000V0000?`000`4000030003
0@0000@000@100000P4300339P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0T02H0
9P0V02H0900T02@0900V02H0900T02@09P0V02H09P0V03<000<100003@008bH09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V00T000<100009P0003l000<100000`000`40000400030@000080
00<100000P00ibH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02@0900T02@092HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900300039P0000P000<1000010006RH09P0V02H09P0V02H09P0V02H09P0V02H00bH0400V02H09P0V
02H09P0V02H900030@0002D0000o00030@0000<000<1000010000`40000300030@00008008HV02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V00DV00T09P0V02H09P000RH0:b@09P0V02@09P0V02H09P0T02@0900T02H09P0T02@0
900V02H09P0V02H0<`000`40000=000U9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0800030@0002D0000o00030@0000<000<1000010000`40000300030@0000<00>DV02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0900T02@0900T9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V00h000<10000100052H09P0V02H0
9P0V02H09P0V02H01bH0500V02H09P0V02H09P0V02H09P0V20000`40000T0000?`000`4000030003
0@0000@000<1000010000`40000300209P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P059P0302HV00<V00L09P0V02H0008V02dT9RH0
9P0T02@09P0V02@0900T02@0900T02@0900T02@09P0V02H09P0V02H0<@000`40000;000W9P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V00T000<10000900003l000<100000`000`40
000400030@0000D000<100000`00hbH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0T02@0900T02@V900T02@0900T9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900V00h000<100000P0052H09P0V02H09P0V02H09P0V02H01bH05P0V02H09P0V02H09P0V
02H09P0V02H900030@0002<0000o00030@0000<000<1000010000`40000600030@0000<0080V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V00dV00<09P000RH0;B@V900V02@0900V02H0900T02@0900T02@0900T02@0900V02H09P0V02H0
9P0a00030@0000L002dV00009P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
1`020BD0000o00030@0000<000<1000010020@P000<100000`00hBH09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H0900T02@0900T02@V92HT9R@092HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@09P0>00030@00008001@V02H09P0V02H09P0V02H09P0V00<V
01`09P0V02H09P0V02H09P0V02H09P0V02H09P0V10020@03004002@0001000050@0000401`000`40
000800030@0000<0080V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V00XV00<T9RH00bH00b@V90029P0X02@V900V02H0900T02@0900T
02@0900T02@0900V02H0900V02H09S4000<100002@00:bH09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H00P030@<000<100008`0004400`4800030@0000T000<100000`00gbH0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0900T02@0900T9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02H03P004@4002H09P0V02H09P0S
02H000<V00<09P000bH04@0V02H09P0V02H09P0V02H000/V0`4600030@0002<0001<00030@0000X0
00<100000`00N2H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V00<V00<09P0032H0<B@V9RHT9R@V900T9R@V900V02@0900T02@0900T02@0900T02@0
9P0T02@09P0V02H0<@000`400005000/9P0002H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P020@T000<100008`0004`000<100002`000`400003000T9P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H00RH0]B@V900T02@092HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02H040003`4002H09P0V02H09P0V00039P0502H09P006RH30@/0
00<100008`0004`000<1000030000`400003001h9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H01BH00`0V9P089P0392HT00<V00TT9R@V92HT9R@0
0RH08P0T02@0900T02@0900T02@0900T02H0900T02H09P0V02H^00812@008bH09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V00800P4>00030@0002<0001<00030@0000d000<100000`00abH0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@01@003B@0900T02@0900T02H03@020@8000`102H09P0V02H08`039P04
02HV90lV00TT9R@V92HT9R@00`4@00030@0002<0001<00030@0000h000<100000`00L2H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P059P0302H000`V03DT9RHV
92HT9R@V92HT9R@V92HT02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0Y00<110000`40
0002000S9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H00`4C00030@0002<0001<0003
0@0000h000<100001000`bH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT02@0
900T02@0900T02@0900T02@0900T02@0900T02@09009000;900T02@0900T02H020020@P000P19P0V
02H08`DV00<T9R@02bH02R@V92HT9R@V92H20AH000<100008`0004`000<100003`000`400004001b
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H01bH00`0V
9P089P0c92HT9R@V92HT9R@V92HT9R@V900T02@0900T02@0900T02@0900T02@0900V02H09P0V02@0
0`4:00030@0000<001`V02H09P0V02H09P0V02H09P0V02H09P0V02H00`4H00030@0002<0001<0081
4@000`400004000R9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V008V094T9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92H19R@V92HT9R@V92HT9R@V92HT9R@V900T02@0
900T02@0900T02@0900T01l000@V00000P4>00050@0V02<00bH00`4V90059P0?92HT9R@V92HT9R@V
92HT00816`000`40000S0000C0000`40000A00030@0000@007`V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H022H0<B@V92HT9R@V92HT9R@V
92HT9R@0900T02@0900T02@0900T02@0900T02H09P0V02H07`030A0001d19P0V02H09P0V02H09P0V
02H09P0V02H09P0V00020Ad000<100008`0004`000<100004P000`400004002c9P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92H19P4V
0BH19P4V0BHT9R@V92HT9R@V92HT9R@V92HT02@0900T02@0900T02@0900T02@07@030A<001`V0B<V
0BH19R@V92HT9R@V92HT9R@V92HT9R@180000`40000S0000C0000`40000C00030@00008007XV02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V00hV
02/T9R@V92HT9R@V92HT9R@0900T02@0900T02@0900T02@0900T02H09P0V01`00P4H000F0BH09P0V
02H09P0V02H09P0V02H09P818P000`40000S0000C0000`40000D00030@000080028V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H00RH0SB@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9P4V0BH19P4V0BHT9R@V92HT9R@V92HT9R@V900T02@0900T02@0900T02@0900J00<15P005bH0
8`019P4V0BHT9R@V92HT9R@V92HT00819@000`40000S0000C0000`40000E00030@00008007hV02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0<9P0W92HT0B@V92HT9R@V900T02@0900T02@0900T02@0900T02H09P0V01L00P4L00039P01008V
00`09P0V02H09P0V02H20BP000<100008`0004`000<100005P000`400002000J9P0V02H09P0V02H0
9P0V02H09P0V02H09P029P0392HT00<V08hT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92H19P4V0BH19P0V0BHT9R@V92HT9R@V92HT9R@V900T02@0900T02@0900T02@150030A/0014V02HV
004T9R@V92HT9R@V00030BX000<100008`0004`000<100005`000`40000200259P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P06000U8R@V92HT9R@V900T02@0900T02@0900T02@0900T02H09P018P0A00817`000bH09P029P03
0BH000DV00<00@40;@000`40000S0000C0000`40000H00030@00008001XV02H09P0V02H09P0V02H0
9P0V02H09P0V00PV060T9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92H90081024V92HT
9R@V92HT9R@V92HT02@0900T02@0900T02@00B800P4=00<17P0022H09P019R@V0P4012@V92H30Bl0
00<100008`0004`00P4J00030@00008007dV02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0>000U8P4V92HT9R@0900T02@0900T02@0900T
02@0900V02HR0B818P0:008190069P030BHV0081<P000`40000S0000C0000`40000I00030@0000<0
018V02H09P0V02H09P0V02H09P0@9P1I92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@04@0082@V
92HT9R@V92HT9R@V92HT02@0900T02@0900T8P4R104600<18`002BH08bHT9R@V90020CD000<10000
8`0004`000<100006P000`400003001g9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H050009BHT9R@V92HT02@0900T02@0900T02@0900T02HR0B81
8P4R0B800P400`010@0X00@V0`400`01000d00030@0002<0001<00030@0001/000<100000`0052H0
9P0V02H09P0V02H09P0V02H032H0EB@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@05P02900M9R@V92HT
9R@V92HT9R@0900T02@0900T8P4R0B802@4O00030@0000<000HV02<090020@D000<10000<P000`40
000S0000C0000`40000L00030@0000<0074V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0J000Q9R@V92HT02@0900T02@0900T02@0900T8P4R0B818P4R00<1
00<R0B800`4I00@120000bH10@0700030@00038000<100008`0004`000<100007@000`400003000D
9P0V02H09P0V02H09P0V02H09P0<9P1>92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V7@007BHT9R@V92HT9R@V92HT
02@0900T02@R0B818P4R00<100<R0@402@4A00T11@000bH09P0;00030@00034000<100008`0004`0
00<100007P000`400003001Y9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V024001XT9R@V900T9R@0900T02@0900T02@R0B818P<101<R0B818P4R0B818P4R0B818P4R01H1
50000`40000`00030@0002<0001<00030@0001l000<100000`0052H09P0V02H09P0V02H09P0V02H0
32H0AR@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT000F92HT9R@V92HT9R@V92HT02@092818PD100LR0B818P4R02<11@000bH0
000<00030@00030000<100008`0004`000<1000080000`400003001S9P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H001<V92HT9R@V900T02@0900T02@000D101<R0B818P4R0B81
8P4R0B818P4R01X15@000`40000_00030@0002<0001<00030@00024000<100000`0052H09P0V02H0
9P0V02H09P0V02H03RH0?R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V:@004BH19R@V92HT9R@V92HT02@000<100dR0B818P4R0B818P4R02<1
1@000b<0000=00030@0002l000<100008`0002`00`4800811P080@D00`4R00030@0000<005dV02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0;0003BHT9R@V900T02@090000`408@018P4R0B81
8P4R0B818P4R0B818P4R0B818P4R0B818P0B0AH000<10000;P000`40000S0000:`001@40000100L0
0P4700030@0000<000<100000`000`40000S00030@0000<001@V02H09P0V02H09P0V02H09P0V010V
00/T9R@V92HT9R@V90079P0992HT9R@V92HT00LV00TT9R@V92HT9R@02RH_008100TT9R@V92HT9R@0
0P405@0T02@R0B818P4R0B818P4R0B818P0Q0@<000<S02<04@000`40000]00030@0002<0000Z0003
0@0000<000<100003P000`40000800030@0002@000<100000`00EBH09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H0<`001b@V92HT9R@00`40:R@0900T004R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R
0B818Ph15`000`40000]00030@0002<0000Z00030@0000<000<100003`000`40000700030@0002@0
00<10000100052H09P0V02H09P0V02H09P0V02H052H01B@V92HT00/V00DT9R@V90079P0392HV00<V
00<T9R@012Hf0005904V92H00P407R@V92HT9R@00B818P4R0B818P4R0B818P4R0B818Qh10P000b<0
000B00030@0002`000<100008`0002X000<100000`000`40000@00030@0000H000<100009@000`40
0004001=9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H0>P00<b810@4V92HT02@0900T8P4R0B818P4R0B818P4R0B81
8P4R0B818P4R0B818P4R0B818P030@038P4100<100<R0@400`4G00030@0002`000<100008`0002X0
00<100000`000`40000A00030@0000D000<100009P000`400004000J9P0V02H09P0V02H09P0V02H0
9P0V02H09P0R9P0792HT9R@V90079S/00P42900Q9R@V92HT9R@V90018P4R0B818P4R0B818P4R0B81
8P4R01h10P000b<0000C00030@0002/000<100008`0002X000<100000`000`40000B00030@0000@0
00<100009`000`40000400179P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0>P030@8003H19R@V92HT02@0900T8P4R0B818P4R
0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B830@058P410B800`4H00030@0002/000<10000
8`0002X000<100000`000`40000C00030@0000<000<10000:0000`400002000P9P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0R9P0492HV9SX00P45000Q904T92HT9R@V92HT004R0B818P4R0B81
8P4R0B818P4R02010P000b<0000D00030@0002X000<100008`0002X000<100000`000`40000=0003
0@0000<000<100000`000`40000Y00030@00008004<V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V03T00`48000k9R@V92HT02@0900T
8P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B801@4J00030@0002T0
00<100008`0002X000<100000`000`40000=00030@0000<000<100000`000`40000Z00030@000080
028V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H07RHi00812`00:P4T92HT9R@V92HT8P4R
0B818P4R0B818P4R0B818P4R0B818P4R0B818QH100<S00005`000`40000Y00030@0002<0000[0005
0@0000404@001@40000100H00P4/00030@00008003/V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0j00<13P00>R@V92HT9R@090018P4R0B818P4R
0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B830@048P408aT000<10000:0000`40
000S0000;0030A<00`4700030@0002`000<100000P0092H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V01@V>P020A4002/T0BHT9R@V92HT8P4R0B818P4R0B818P4R0B818P4R0B818P4R0B81
8P4R01@100<S000060000`40000X00030@0002<0001<00030@0002d000<100000P00=BH09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V03T00`4D000k9R@V92HT
02@00B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B800`4L0003
0@0002L000<100008`0004`000<10000;P000`400002000X9P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V00XV>0030AL000<192@00R@09BHT9R@R0B818P4R0B818P4R0B818P4R0B81
8P4R0B818P4R0B8050400b<0000J00030@0002H000<100008`0004`000<10000;`000`400002000_
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0>0020A/000TT9R@V
900T00401B80;04R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0`4M0003
0@0002H000<100008`0004`000<10000<0000`400002000[9P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0h00<17@009b@192@V92HT8P4R0B818P4R0B818P4R0B818P4R0B81
8P4R0B818P0E0Al000<100009@000`40000S0000C0000`40000`00030@0000<002LV02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0>@020B4000HV92HT02@98P0X0B818P4R0B818P4R
0B818P4R0B818P4R0B818P4R0B818P4R0B818P@17`000`40000U00030@0002<0001<00030@000340
00<100000`008bH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V03T00`4S000V0B@19R@V
0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B8D0B4000<1000090000`40000S0000C0000`40
000b00030@0000<001lV02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V03X00P4W000592HT9R@0
2b809P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R104P00030@0002@000<10000
8`0004/000<10040=0000`400003000M9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0>0030BT0
02DT0BHT9P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R00<100<R0B803`4Q00030@0002<0
00<100008`0004/00P4f00030@0000<001TV02H09P0V02H09P0V02H09P0V02H09P0V03T00P4[0081
00<V92H03R808P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B850B<000<100008P000`40
000S0000B`000`40000f00030@0000<001DV02H09P0V02H09P0V02H09P0V02H0>@030BX00`42000X
0B@T9P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818Q018`000`40000R00030@0002<0
001;00030@0003L000<100000`004BH09P0V02H09P0V02H09P0V03X00P4Z00<11@0290039R@R010R
01h18P4R0B818P4R0B818P4R0B818P4R0B818P4R0B850B@000<100008@000`40000S0000B`000`40
000h00030@0000<000lV02H09P0V02H09P0V02H0>0030BX00P49008T02DV0B818P4R0B818P4R0B81
8P4R0B818P4R0B818P4R0B818P4R010190000`40000Q00030@0002<0001;00030@0003T000<10000
0`002bH09P0V02H09P0V03T00P4Z00<12`000b@V900C8P0L0B818P4R0B818P4R0B818P4R0B818P4R
0B818P@19P000`40000P00030@0002<0001;00030@0003X000<100000`002BH09P0V02H09P0g00<1
:P020@h002LT0BH18P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B800`400b810@090BL0
00<100007`000`40000S0000B`000`40000k00030@0000<000DV02H09P0g00<1:P030A4000<V9280
0b800bDR8P0<8P0L0B818P4R0B818P4R0B818P4R0B818P4R0B818P@19`000`40000O00030@0002<0
000Y00@100<00@406`000`40000l00030@0000<000<V0000=P020B/00P4D000Z0BH18P4R0B818P4R
0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R2P4Y00030@0001h000<100008`0002P000<10000
0P020A`000<10000?0000`40000i00<1:P030AH00R@C8P0L0B818P4R0B818P4R0B819`4R0B818P4R
0B818P<1:@000`40000N00030@0002<0000X00030@0000<000<100006P030Cd000<10000=P020BX0
0`4J000Y904R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B802P4Z00030@0001h0
00<100008P0002P000<100000`000`40000J00030@0003h000<10000<P030BX00P4M0003928R00@R
00<U8R802R80704R0B818P4R0B819`4W0BL18P4R0B818P4R0B820B/000<100007P000`40000R0000
:0000`40000300030@0001X000<10000?`000`40000_0081:P030Al002LT0B818P4R0B818P4R0B81
8P4R0B818P4R0B818P4R0B818P4R0B802`4/00030@0001d000<100008P0002P000<100000`000`40
000J00030@00040000<10000:`030BX00P4R00030BHR00@R00DU8RDR00068P0L0B818P4R0BL19`4W
0BL19`4R0B818P4R0B818P@1;@000`40000L00030@000280000X00030@0000<000<100006P000`40
001100030@0002P00P4Z00<19@009P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R
2P4^00030@0001`000<100008P0002L00P44008170000`40001200030@0002@00`4Y00<1:0000b@R
8P058P0U9B808RDR02808P0R02800BL19`4W0BL19`4R0B818P4R0BL18P030Bl000<100006`000`40
000R0000B`000`40001300030@0002400P4Z0081:`00:P4R0B818P4R0B818P4R0B818P4R0B818P4R
0B818P4R0B818P4R0B818PH1;`000`40000K00030@000280001;00030@0004@000<100007@030BT0
0`4^00038RDR00@R02DU8P0R9B808P0R02808P019`4W0BL19`4W0B818P4W0BL18P4R0081<@000`40
000J00030@000280001;00030@0004D000<100006P020BX00P4a000Y8P4R0B818P4R0B818P4R0B81
8P4R0B818P4R0B818P4R0B818P4R0B801@4a00030@0001X000<100008P0002/00`4800812P040@D0
0`5600030@0001H00`4Y00<1<`00:R8U8RDR9B808RDR9B8U8P0R02808RL19`4W0BL19`4W0B819`4W
0BL18P<1<`000`40000I00030@000280000Z00050@0000401`020@`000<1000010000`4000170003
0@0001<00P4Y00<1=P009b818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P060C@0
00<1000060000`40000R0000:@000`40000300030@00018000<1000010000`40001700030@000100
0`4Y0081>P028P0W9B8U8P0R028U8RDR02808P0R9`4W0BL19`4W0BL19`4W0BL19`4R0081=@000`40
000H00030@000280000Y00030@0000<000<100003@070@D000<10000B0000`40000<00<1:@030C`0
02H18P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P@1=`000`40000G00030@000280
000Y00030@0000<000<100003@000`40000200030@0000@000<10000B@000`4000090081:P020D00
02PR9B8U8RDR02808RDR02808P019`4W0@0100400BL19`4W0BL18P4R>0000`40000G00030@000280
000Y00030@0000<000<100003P001@40000100H000<10000BP000`40000500<1:@030D8002DR0B81
8P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0081>P000`40000F00030@000280000Y0003
0@0000<000<100003P001@40000100H000<10000B`000`4000020081:@030DH002DU8RDR9B808P0R
02808P0R02808P0R00400@019`4W0BL19`4R03`000<100005@000`40000R0000:@000`4000030003
0@0000l000@100011P000`40001<00@1:@020DT002818P4R0B818P4R0B818P4R0B818P4R0B818P4R
0B818P4R0P4m00030@0001D000<100008P0002T000<100000`000`40000?00040@000@H000<10000
B`030BP00`5<000Q8RDR9B808P0R02808P0R02808P0R00400@019`4W0BL1040000<1000050000`40
000R0000:@000`40000300030@00010000<100401P000`40001800<10`000`40000S0081C`008281
8P4R0B818P4R0B818P4R0B818P4R0B818P4R0B81@@000`40000D00030@000280000Z00050@000040
4`000`400@0600<1AP020@L000<100007`030E8001hU8RDR9B808P0R02808P0R02800@0100400BL1
9`5300030@0001<000<100008P0002/00`4E00811P000`40001300<12P000`40000L0081EP007281
8P4R0B818P4R0B818P4R0B818P4R0B818P5400030@0001<000<100008P0004/000<10000@@020@h0
00<1000060030ET001XU8RDR02808P0R02808P0R00400@019`4W0DH000<100004P000`40000R0000
B`000`40000n00<14@000`40000D00<1H@0052818P4R0B818P4R0B818P4R0B81B0000`40000A0003
0@000280001;00030@0003`00P4D00030@0001800P5V000@8P0R02808P0R00400@019dX000<10000
4@000`40000R0000B`000`40000i00<15`000`40000>00<1J0001b818P4R0B80E0000`40000@0003
0@000280001;00030@0003L00P4K00030@0000/00P5]00038P0R05H000<1000040000`40000R0000
B`000`40000d00<17P000`40000700<1b@000`40000?00030@000280001;00030@0003400`4R0003
0@0000<00`7=00030@0000h000<100008P0004/000<10000;`020BH000<100000P7@00030@0000h0
00<100008P0004/00`4/00<1:@020M<000<100003@000`40000R0000B`000`40000Z0081;0000`40
003B00030@0000d000<100008P0004/000<100009`030Bh000<10000d`000`40000<00030@000280
001;00030@0002D00P4a00030@000=<000<1000030000`40000R0000B`000`40000R00<1<`000`40
003D00030@0000/000<100008P0004/000<1000080020CH000<10000e@000`40000:00030@000280
001;00030@0001d00`4h00030@000=D000<100002P000`40000R0000B`000`40000K0081>`000`40
003F00030@0000T000<100008P0004/000<1000060030Cd000<10000eP000`40000900030@000280
001;00815`020D0000<10000e`000`40000800030@000280001;00030@0001<00`5200030@000=P0
00<100001`000`40000R0000B`000`40000A0081A@000`40003H00030@0000L000<100008P0004/0
00<100003P030DL000<10000f@000`40000600030@000280001;00030@0000/00`5:00030@000=T0
00<100001P000`40000R0000B`00104000480081C@000`40003J00030@0000D000<100008P0004/0
00@100011@030Dl000<10000fP000`40000500030@000280001;00040@000@<00P5B00030@000=/0
00<1000010000`40000R0000B`000`4000040E@000<10000g0000`40000300030@000280000[00<1
20020@T00`4700@1E`000`40003L00030@0000<000<100008P0002X000D100000@07008120001@40
000100H00P400`01001F00030@000=d000<100000P000`40000R0000:@000`40000300030@0000d0
00<100000`000`40000400030@0105L000<10000g@000`40000200030@000280000Y00030@0000<0
00<100003@000`40000300030@0000D00P5G00030@000=h000D100000@0T0000:@000`4000030003
0@0000d000<100000`000`40000600030@0005D000<10000g`001040004T0000:@000`4000030003
0@0000d00P4300030@0000P000<100000P000`40001?00030@000=l000@10001900002T000<10000
0`000`40000=00030@01008130020@8000<10000C`000`40003P00030@0102@0000Y00030@0000<0
00<100003@000`40000@00030@01054000<10000h0000`400@0T0000:@000`40000300030@0000h0
00<1000040020E4000<10000h@020B@0000Y00030@0000<000<100003P000`40000A00030@0004l0
00<10000h@020B@0000Z00050@0000404P000`40000@0081D0000`40003Q0081900002/00`4D00@1
40000`40000200030@0004P000<10000gP030BH0001G00050@000040BP000`40003K00<1:@0005P0
00@10001BP000`40003H00<1;00005T000<10040BP000`40003E00<1;`0005X00P5:00030@000=40
104b0000F`000`40001800030@000<h00`4f0000G0020@<000<10000@P000`40003;00<1>@0005h0
00@10001A0000`40003800<1?00005l000<10040A0000`40003500<1?`0006000P5400030@000<80
0`520000H@000`40001200030@000;l00`550000C0030A<000<100000P000`40000l00030@000;`0
0`580000B`001@40000101<000D100000@0n00030@000;T00`5;0000BP000`40000300030@000140
00@10001?P000`40002f00<1CP0004X000<100000`000`40000B00030@0100D000<10000=P000`40
002c00<1D@0004X000<100000`000`40000C00811@000`40000f00030@000:l0105D0000BP000`40
000300030@0001@00P4400030@0003H000<10000[0030EP0001:00030@0000<000<100005P001@40
000103P000<10000Z@030E/0001:00030@0000<000<100005`001040004h00030@000:H00`5N0000
BP000`40000300030@0001P000<10040>0000`40002S00<1H@0004X000<100000`000`40000I0081
>0000`40002P00<1I00004/000D100000@0M00030@0003H000<10000W@030FL0001<00<17`000`40
000200030@00030000<10000VP030FX0001_00050@000040<P000`40002G00<1K@00070000@10001
<P000`40002D00<1L000074000<10040<P000`40002@00@1L`0007800P4b00030@0008d00`5g0000
L`020C4000<10000RP030GX0001e00050@000040;0000`40002700<1O@0007H000@10001;0000`40
002400<1P00007L000<10040;0000`40002100<1P`0007P00P4/00030@0007h00`660000N@000`40
000Z00030@0007/00`690000NP001@40000102L000<10000N0030H`0001k00040@000BL000<10000
M@030Hl0001l00030@0102L000<10000LP030I80001m00819`000`40001^00@1U@0007h00P4V0003
0@0006/00`6I0000P0001@400001024000<10000J0030I`0002100040@000B4000<10000I@030Il0
001Y00L14P000`400@0500030@0001T000<10000HP030J80001/00030@0001@00P4500030@0001T0
00<10000G`030JD0001/00030@0001D000<100000`000`40000I00030@0005`00`6X0000K0000`40
000F00030@00008000<100006@000`40001I00<1Z`0006`000<100005`001@40000101/000<10000
EP030Jh0001/00030@0001P000@100016`000`40001C00<1/@0006`000<100006@000`400@0K0003
0@0005000`6d0000K0000`40000J00816`000`40001<00@1]`0006`000<1000070001@40000101H0
00<10000B@030K/0001/00030@0001d000@100015P000`40001600<1_P0006T00P400`01000O0003
0@0101H000<10000@`030L40001[00818@020AH000<10000@0030L@0002?00030@0001@000<10000
?@030LL0002@00050@0000404@000`40000j00<1bP00094000@100014@000`40000g00<1c@000980
00<100404@000`40000d00<1d00009<00P4A00030@0003400`7C0000U0020A0000<10000;@040MH0
002F00050@0000402`000`40000Z00<1fP0009L000@100012`000`40000W00<1g@0009P000<10040
2`000`40000T00<1h00009T00P4;00030@0002400`7S0000VP000`40000200030@0000@000<10000
7P030NH0002K00050@0000401P000`40000K00<1j@0009`000@1000110000`400@0J00<1k00009d0
00<1004010000`400@0G00<1k`0008D0204A008110000`400@0D00<1lP0008H000<100000`000`40
000@00030@00008000<100404@030OD0002700030@0001H00P4200030@0100d0107h0000R0000`40
000G00050@0100402P030O`0002900030@0001L00P400`01000600<1o`0008X000<100005`000`40
0@0400<1o`030000R`000`40000G008100@00@41o`060000S0000`40000G0081o`090000QP000`40
000300030@000?l08P0008H000<100000`000`40003o0280002700050@000040o`0U0000R0030Ol0
9P000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000:40107o00`0002P00030@00008000<10000o`090000YP000`40003o00P0
002V00030@000?l020000:H000<10000o`080000Y@000`40003o00T0002R00<1o`0<0000Y@000`40
003o00T0002V00030@000?l020000:H000<10000o`080000X0000`40000200030@000?l02@000:40
107o00`0003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40
003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40
003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40001g00D1o`0e0000NP000`40003o03@0
001j00030@000?l0=00007/000<10000o`0c0000N`000`40003o03<0001j00030@010?l0=00007X0
00<10040o`0d0000N@001@4000010?l0<`0007T000D100000@3o03<0001h00030@0000<000<10000
o`0`0000N0000`40000300030@000?l0<00007L00`4300<1o`0a0000o`2a0000o`2a0000o`2a0000
\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.234686, 0.31168, 0.00465498, 
  0.00125278}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12  scalefont  setfont
% Scaling calculations
0 1.007 -1.29925 3.74173 [
[.02012 .31927 -3.48107 -10.8 ]
[.02012 .31927 3.71893 0 ]
[.29685 .18904 -3.54904 -10.8 ]
[.29685 .18904 3.65096 0 ]
[.60677 .04321 -3.62554 -10.8 ]
[.60677 .04321 3.57446 0 ]
[.37156 -0.0797 -5.54294 -11.375 ]
[.37156 -0.0797 5.64456 0 ]
[.00732 .7952 -3.72102 0 ]
[.00732 .7952 3.47898 10.8 ]
[.10546 .88495 -3.69692 0 ]
[.10546 .88495 3.50308 10.8 ]
[.19484 .96668 -3.67505 0 ]
[.19484 .96668 3.52495 10.8 ]
[.27656 1.04142 -3.6551 0 ]
[.27656 1.04142 3.5449 10.8 ]
[.14967 1.16054 -5.79094 0 ]
[.14967 1.16054 5.52156 11.375 ]
[-0.00919 .38824 -7.2 -4.71937 ]
[-0.00919 .38824 0 6.08063 ]
[-0.01029 .50705 -21.6 -4.75471 ]
[-0.01029 .50705 0 6.04529 ]
[-0.0114 .62638 -21.6 -4.79022 ]
[-0.0114 .62638 0 6.00978 ]
[-0.01252 .74624 -21.6 -4.82591 ]
[-0.01252 .74624 0 5.97409 ]
[-0.07332 .58937 -11.1875 -5.02518 ]
[-0.07332 .58937 0 6.34982 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.00338 .37369 m
.79826 0 L
s
.01971 .36602 m
.0195 .38939 L
s
[(0)] .02012 .31927 -0.03303 1 Mshowa
.29667 .23581 m
.29658 .25919 L
s
[(2)] .29685 .18904 -0.01416 1 Mshowa
.60686 .08998 m
.6069 .11337 L
s
[(4)] .60677 .04321 .00709 1 Mshowa
.125 Mabswid
.08613 .33479 m
.08602 .34881 L
s
.15438 .3027 m
.15429 .31673 L
s
.22453 .26972 m
.22446 .28375 L
s
.37088 .20092 m
.37084 .21495 L
s
.44725 .16502 m
.44723 .17905 L
s
.52587 .12806 m
.52588 .14209 L
s
.69031 .05075 m
.69036 .06478 L
s
.77634 .01031 m
.77641 .02434 L
s
gsave
.37156 -0.0797 -66.5429 -15.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(x) show
70.187500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
0 .74137 m
.2826 1 L
s
.00774 .74845 m
.00796 .72508 L
s
[(0)] .00732 .7952 .03362 -1 Mshowa
.1058 .8382 m
.10597 .81482 L
s
[(1)] .10546 .88495 .02692 -1 Mshowa
.1951 .91992 m
.19523 .89654 L
s
[(2)] .19484 .96668 .02085 -1 Mshowa
.27675 .99465 m
.27685 .97127 L
s
[(3)] .27656 1.04142 .01531 -1 Mshowa
.125 Mabswid
.02812 .7671 m
.02824 .75308 L
s
.0481 .78539 m
.04822 .77136 L
s
.0677 .80333 m
.06781 .7893 L
s
.08693 .82093 m
.08704 .8069 L
s
.12432 .85515 m
.12442 .84112 L
s
.1425 .87179 m
.1426 .85776 L
s
.16035 .88812 m
.16044 .8741 L
s
.17788 .90417 m
.17797 .89014 L
s
.21201 .9354 m
.21208 .92137 L
s
.22862 .9506 m
.22869 .93657 L
s
.24494 .96554 m
.24501 .95151 L
s
.26099 .98022 m
.26105 .96619 L
s
gsave
.14967 1.16054 -66.7909 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.125000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(y) show
70.312500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
.00338 .37369 m
0 .74137 L
s
.0033 .38239 m
.00954 .37946 L
s
[(0)] -0.00919 .38824 1 -0.12604 Mshowa
.0022 .5015 m
.00845 .49872 L
s
[(0.2)] -0.01029 .50705 1 -0.1195 Mshowa
.0011 .62113 m
.00736 .61851 L
s
[(0.4)] -0.0114 .62638 1 -0.11292 Mshowa
0 .74129 m
.00626 .73882 L
s
[(0.6)] -0.01252 .74624 1 -0.10631 Mshowa
.125 Mabswid
.00303 .41212 m
.00677 .41038 L
s
.00275 .44188 m
.0065 .44017 L
s
.00248 .47167 m
.00623 .46999 L
s
.00193 .53136 m
.00568 .52972 L
s
.00166 .56125 m
.00541 .55963 L
s
.00138 .59118 m
.00513 .58958 L
s
.00083 .65112 m
.00458 .64957 L
s
.00055 .68115 m
.00431 .67962 L
s
.00028 .7112 m
.00403 .7097 L
s
gsave
-0.07332 .58937 -72.1875 -9.02518 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(u) show
70.187500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
.00338 .37369 m
0 .74137 L
s
0 .74137 m
.2826 1 L
s
.2826 1 m
.28388 .67992 L
s
.28388 .67992 m
.00338 .37369 L
s
.79826 0 m
.99671 .40534 L
s
.99671 .40534 m
1 .76812 L
s
1 .76812 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.00338 .37369 m
0 .74137 L
s
0 .74137 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.79826 0 m
.00338 .37369 L
s
.28388 .67992 m
.99671 .40534 L
s
.99671 .40534 m
1 .76812 L
s
1 .76812 m
.2826 1 L
s
.2826 1 m
.28388 .67992 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.21767 .56799 m .23719 .57651 L .21234 .58667 L .41 .805 .986 r
F
.21767 .56799 m .21234 .58667 L .19533 .56796 L F
.2356 .55712 m .26229 .56624 L .23719 .57651 L .409 .805 .985 r
F
.2356 .55712 m .23719 .57651 L .21767 .56799 L .407 .804 .985 r
F
0 g
.5 Mabswid
.21234 .58667 m
.23719 .57651 L
s
.23719 .57651 m
.23719 .57651 L
s
.21767 .56799 m .19533 .56796 L .20803 .55132 L .407 .804 .985 r
F
.19533 .56796 m .17801 .54892 L .20803 .55132 L .407 .804 .986 r
F
0 g
.19533 .56796 m
.21234 .58667 L
s
.19533 .56796 m
.19533 .56796 L
s
.2356 .55712 m .21767 .56799 L .20803 .55132 L .408 .804 .986 r
F
.28763 .55588 m .26229 .56624 L .25995 .5459 L .406 .805 .984 r
F
.25995 .5459 m .26229 .56624 L .2356 .55712 L .404 .804 .984 r
F
0 g
.23719 .57651 m
.26229 .56624 L
s
.26229 .56624 m
.26229 .56624 L
s
.2356 .55712 m .22875 .53405 L .25995 .5459 L .405 .804 .985 r
F
.22875 .53405 m .2356 .55712 L .20803 .55132 L .404 .803 .985 r
F
.28638 .53588 m .31322 .54541 L .28763 .55588 L .403 .805 .983 r
F
.28763 .55588 m .25995 .5459 L .28638 .53588 L .401 .803 .983 r
F
0 g
.26229 .56624 m
.28763 .55588 L
s
.28763 .55588 m
.28763 .55588 L
s
.22875 .53405 m .20803 .55132 L .19313 .53118 L .405 .803 .985 r
F
.20803 .55132 m .17801 .54892 L .19313 .53118 L F
0 g
.17833 .54926 m
.19533 .56796 L
s
.17801 .54892 m .16038 .52953 L .19313 .53118 L .405 .802 .985 r
F
0 g
.17801 .54892 m
.17801 .54892 L
s
.17801 .54892 m
.17833 .54926 L
s
.31252 .52591 m .33907 .53484 L .31322 .54541 L .399 .805 .981 r
F
.31252 .52591 m .31322 .54541 L .28638 .53588 L .396 .803 .981 r
F
0 g
.28763 .55588 m
.31322 .54541 L
s
.31322 .54541 m
.31322 .54541 L
s
.25944 .52536 m .28638 .53588 L .25995 .5459 L .402 .803 .983 r
F
.25944 .52536 m .25995 .5459 L .22875 .53405 L .4 .802 .983 r
F
.2859 .51607 m .31252 .52591 L .28638 .53588 L .397 .803 .982 r
F
.2859 .51607 m .28638 .53588 L .25944 .52536 L .396 .801 .982 r
F
.36518 .52417 m .33907 .53484 L .3386 .51588 L .393 .804 .979 r
F
.3386 .51588 m .33907 .53484 L .31252 .52591 L .39 .801 .979 r
F
0 g
.31322 .54541 m
.33907 .53484 L
s
.33907 .53484 m
.33907 .53484 L
s
.23694 .51413 m .25944 .52536 L .22875 .53405 L .401 .802 .984 r
F
.23694 .51413 m .22875 .53405 L .20878 .51206 L .4 .801 .984 r
F
.19313 .53118 m .20878 .51206 L .22875 .53405 L .402 .802 .984 r
F
.19313 .53118 m .17507 .51029 L .20878 .51206 L .402 .802 .985 r
F
.17507 .51029 m .19313 .53118 L .16038 .52953 L .403 .802 .985 r
F
.17507 .51029 m .16038 .52953 L .14242 .50977 L F
0 g
.16038 .52953 m
.16038 .52953 L
s
.16038 .52953 m
.17801 .54892 L
s
.69079 .50771 m .7088 .51518 L .68086 .52594 L .37 .8 .968 r
F
.69079 .50771 m .68086 .52594 L .66769 .50613 L .367 .782 .979 r
F
.3386 .51588 m .31252 .52591 L .31218 .50669 L .392 .802 .98 r
F
.31218 .50669 m .31252 .52591 L .2859 .51607 L .389 .8 .98 r
F
.36456 .5058 m .39155 .51338 L .36518 .52417 L .386 .803 .976 r
F
.36518 .52417 m .3386 .51588 L .36456 .5058 L .381 .799 .975 r
F
0 g
.33907 .53484 m
.36518 .52417 L
s
.36518 .52417 m
.36518 .52417 L
s
.25941 .50524 m .2859 .51607 L .25944 .52536 L .397 .801 .982 r
F
.25941 .50524 m .25944 .52536 L .23694 .51413 L .396 .8 .982 r
F
.71364 .4986 m .73701 .5043 L .7088 .51518 L .306 .78 .929 r
F
.71364 .4986 m .7088 .51518 L .69079 .50771 L .287 .759 .937 r
F
0 g
.68086 .52594 m
.7088 .51518 L
s
.7088 .51518 m
.7088 .51518 L
s
.36456 .5058 m .3386 .51588 L .33812 .4973 L .384 .8 .977 r
F
.3386 .51588 m .31218 .50669 L .33812 .4973 L .381 .797 .977 r
F
.28593 .4965 m .31218 .50669 L .2859 .51607 L .392 .8 .981 r
F
.28593 .4965 m .2859 .51607 L .25941 .50524 L .39 .799 .981 r
F
.41818 .50249 m .39155 .51338 L .39025 .49561 L .375 .801 .97 r
F
.39025 .49561 m .39155 .51338 L .36456 .5058 L .369 .796 .97 r
F
0 g
.36518 .52417 m
.39155 .51338 L
s
.39155 .51338 m
.39155 .51338 L
s
.25941 .50524 m .23694 .51413 L .22888 .49313 L .398 .801 .983 r
F
.23694 .51413 m .20878 .51206 L .22888 .49313 L F
.22888 .49313 m .20878 .51206 L .19035 .48837 L .399 .8 .984 r
F
.17507 .51029 m .19035 .48837 L .20878 .51206 L .401 .801 .984 r
F
.17507 .51029 m .14242 .50977 L .15554 .48985 L .402 .801 .985 r
F
.15554 .48985 m .14242 .50977 L .12412 .48965 L F
0 g
.14242 .50977 m
.16038 .52953 L
s
.14242 .50977 m
.14242 .50977 L
s
.17507 .51029 m .15554 .48985 L .19035 .48837 L .401 .801 .984 r
F
.68306 .49472 m .69079 .50771 L .66769 .50613 L .285 .753 .942 r
F
.66769 .50613 m .65427 .48594 L .68306 .49472 L .304 .747 .964 r
F
0 g
.66769 .50613 m
.68086 .52594 L
s
.66769 .50613 m
.66769 .50613 L
s
.71364 .4986 m .69079 .50771 L .68306 .49472 L .298 .762 .946 r
F
.76552 .49332 m .73701 .5043 L .74542 .49262 L .239 .748 .875 r
F
.74542 .49262 m .73701 .5043 L .71364 .4986 L .234 .739 .894 r
F
0 g
.7088 .51518 m
.73701 .5043 L
s
.73701 .5043 m
.73701 .5043 L
s
.31207 .48777 m .33812 .4973 L .31218 .50669 L .384 .798 .978 r
F
.31207 .48777 m .31218 .50669 L .28593 .4965 L .383 .797 .978 r
F
.39025 .49561 m .36456 .5058 L .36342 .48797 L .373 .797 .972 r
F
.36456 .5058 m .33812 .4973 L .36342 .48797 L .369 .794 .972 r
F
.28593 .4965 m .25941 .50524 L .25976 .48516 L .393 .799 .982 r
F
.25976 .48516 m .25941 .50524 L .22888 .49313 L .392 .799 .982 r
F
.41582 .48507 m .44509 .49149 L .41818 .50249 L .36 .798 .962 r
F
.41818 .50249 m .39025 .49561 L .41582 .48507 L .35 .791 .961 r
F
0 g
.39155 .51338 m
.41818 .50249 L
s
.41818 .50249 m
.41818 .50249 L
s
.79431 .48222 m .76552 .49332 L .77274 .4844 L .197 .724 .837 r
F
.76552 .49332 m .74542 .49262 L .77274 .4844 L .194 .72 .849 r
F
0 g
.73701 .5043 m
.76552 .49332 L
s
.76552 .49332 m
.76552 .49332 L
s
.36342 .48797 m .33812 .4973 L .33753 .47919 L .374 .795 .975 r
F
.33753 .47919 m .33812 .4973 L .31207 .48777 L .372 .793 .974 r
F
.31207 .48777 m .28593 .4965 L .28634 .47719 L .386 .797 .98 r
F
.28634 .47719 m .28593 .4965 L .25976 .48516 L .386 .796 .98 r
F
.41582 .48507 m .39025 .49561 L .3876 .47846 L .357 .792 .966 r
F
.39025 .49561 m .36342 .48797 L .3876 .47846 L .352 .789 .965 r
F
.47227 .48037 m .44509 .49149 L .4421 .47533 L .336 .791 .948 r
F
.4421 .47533 m .44509 .49149 L .41582 .48507 L .321 .78 .946 r
F
0 g
.41818 .50249 m
.44509 .49149 L
s
.44509 .49149 m
.44509 .49149 L
s
.22888 .49313 m .23653 .47216 L .25976 .48516 L .395 .799 .983 r
F
.21064 .47333 m .23653 .47216 L .22888 .49313 L .395 .798 .983 r
F
.21064 .47333 m .22888 .49313 L .19035 .48837 L .398 .8 .984 r
F
.13594 .46996 m .15554 .48985 L .12412 .48965 L .402 .801 .985 r
F
.13594 .46996 m .12412 .48965 L .10548 .46916 L F
0 g
.12412 .48965 m
.14242 .50977 L
s
.12412 .48965 m
.12412 .48965 L
s
.15554 .48985 m .16548 .47023 L .19035 .48837 L .401 .801 .985 r
F
.15554 .48985 m .13594 .46996 L .16548 .47023 L F
.72964 .48925 m .74542 .49262 L .71364 .4986 L .173 .705 .844 r
F
.72964 .48925 m .71364 .4986 L .70078 .48873 L .19 .709 .877 r
F
.70078 .48873 m .71364 .4986 L .68306 .49472 L .168 .694 .868 r
F
.21064 .47333 m .19035 .48837 L .18827 .46843 L .399 .8 .984 r
F
.18827 .46843 m .19035 .48837 L .16548 .47023 L F
.70078 .48873 m .68306 .49472 L .67115 .48139 L .199 .703 .905 r
F
.68306 .49472 m .65427 .48594 L .67115 .48139 L .175 .688 .896 r
F
.67115 .48139 m .65427 .48594 L .6406 .46537 L .212 .691 .934 r
F
0 g
.65427 .48594 m
.65427 .48594 L
s
.65427 .48594 m
.66769 .50613 L
s
.33753 .47919 m .31207 .48777 L .31196 .4691 L .377 .794 .977 r
F
.31196 .4691 m .31207 .48777 L .28634 .47719 L .376 .793 .977 r
F
.3876 .47846 m .36342 .48797 L .36179 .47112 L .36 .79 .969 r
F
.36179 .47112 m .36342 .48797 L .33753 .47919 L .358 .788 .969 r
F
.8234 .47101 m .79431 .48222 L .80081 .47447 L .176 .711 .816 r
F
.80081 .47447 m .79431 .48222 L .77274 .4844 L .174 .709 .822 r
F
0 g
.76552 .49332 m
.79431 .48222 L
s
.79431 .48222 m
.79431 .48222 L
s
.25976 .48516 m .26146 .4657 L .28634 .47719 L .389 .797 .981 r
F
.25976 .48516 m .23653 .47216 L .26146 .4657 L F
.46871 .46489 m .49974 .46914 L .47227 .48037 L .301 .779 .923 r
F
.47227 .48037 m .4421 .47533 L .46871 .46489 L .278 .762 .921 r
F
0 g
.44509 .49149 m
.47227 .48037 L
s
.47227 .48037 m
.47227 .48037 L
s
.77274 .4844 m .74542 .49262 L .75404 .48596 L .162 .7 .818 r
F
.72964 .48925 m .75404 .48596 L .74542 .49262 L .164 .699 .839 r
F
.4421 .47533 m .41582 .48507 L .41123 .46652 L .333 .784 .954 r
F
.41582 .48507 m .3876 .47846 L .41123 .46652 L .326 .779 .953 r
F
.36179 .47112 m .33753 .47919 L .33707 .46141 L .365 .79 .973 r
F
.33753 .47919 m .31196 .4691 L .33707 .46141 L .364 .789 .973 r
F
.28634 .47719 m .28641 .45801 L .31196 .4691 L .382 .794 .979 r
F
.28634 .47719 m .26146 .4657 L .28641 .45801 L F
.82998 .46291 m .8528 .45968 L .8234 .47101 L .183 .715 .823 r
F
.8234 .47101 m .80081 .47447 L .82998 .46291 L .184 .716 .821 r
F
0 g
.79431 .48222 m
.8234 .47101 L
s
.8234 .47101 m
.8234 .47101 L
s
.41123 .46652 m .3876 .47846 L .38238 .46443 L .338 .781 .962 r
F
.38238 .46443 m .3876 .47846 L .36179 .47112 L .34 .782 .962 r
F
.22006 .46039 m .23653 .47216 L .21064 .47333 L .396 .799 .983 r
F
.22006 .46039 m .21064 .47333 L .19977 .45655 L .396 .798 .984 r
F
.18827 .46843 m .19977 .45655 L .21064 .47333 L .398 .8 .984 r
F
.77641 .47865 m .80081 .47447 L .77274 .4844 L .145 .689 .792 r
F
.77274 .4844 m .75404 .48596 L .77641 .47865 L .134 .68 .799 r
F
.16548 .47023 m .13594 .46996 L .14768 .45096 L .402 .801 .985 r
F
.14768 .45096 m .13594 .46996 L .11746 .44923 L F
0 g
.12083 .48603 m
.12412 .48965 L
s
.13594 .46996 m .10548 .46916 L .11746 .44923 L .403 .801 .985 r
F
.11746 .44923 m .10548 .46916 L .08649 .44827 L .403 .802 .985 r
F
0 g
.10548 .46916 m
.10548 .46916 L
s
.10548 .46916 m
.12083 .48603 L
s
.50071 .45832 m .52748 .4578 L .49974 .46914 L .24 .75 .873 r
F
.50071 .45832 m .49974 .46914 L .46871 .46489 L .213 .728 .877 r
F
0 g
.47227 .48037 m
.49974 .46914 L
s
.49974 .46914 m
.49974 .46914 L
s
.17528 .45309 m .18827 .46843 L .16548 .47023 L .401 .8 .985 r
F
.16548 .47023 m .14768 .45096 L .17528 .45309 L F
.23826 .45427 m .26146 .4657 L .23653 .47216 L .392 .797 .983 r
F
.23826 .45427 m .23653 .47216 L .22006 .46039 L F
.46871 .46489 m .4421 .47533 L .43915 .46232 L .298 .769 .936 r
F
.4421 .47533 m .41123 .46652 L .43915 .46232 L .287 .761 .934 r
F
.17528 .45309 m .19977 .45655 L .18827 .46843 L .399 .799 .984 r
F
.72964 .48925 m .70078 .48873 L .71784 .48671 L .054 .62 .764 r
F
.71784 .48671 m .70078 .48873 L .68902 .48284 L .078 .635 .803 r
F
.70078 .48873 m .67115 .48139 L .68902 .48284 L .027 .595 .775 r
F
.67115 .48139 m .65898 .47062 L .68902 .48284 L .066 .616 .83 r
F
.67115 .48139 m .6406 .46537 L .65898 .47062 L .01 .574 .799 r
F
.65898 .47062 m .6406 .46537 L .62666 .44441 L .081 .602 .878 r
F
0 g
.6406 .46537 m
.6406 .46537 L
s
.6406 .46537 m
.65427 .48594 L
s
.74589 .48497 m .75404 .48596 L .72964 .48925 L .081 .641 .763 r
F
.74589 .48497 m .72964 .48925 L .71784 .48671 L .094 .65 .786 r
F
.38238 .46443 m .36179 .47112 L .36178 .45485 L .35 .784 .967 r
F
.36179 .47112 m .33707 .46141 L .36178 .45485 L .348 .783 .967 r
F
.31196 .4691 m .31155 .45034 L .33707 .46141 L .372 .79 .976 r
F
.31196 .4691 m .28641 .45801 L .31155 .45034 L F
.28641 .45801 m .26146 .4657 L .26102 .44641 L .387 .795 .981 r
F
.23826 .45427 m .26102 .44641 L .26146 .4657 L F
.74589 .48497 m .77641 .47865 L .75404 .48596 L .121 .671 .782 r
F
.88249 .44824 m .8528 .45968 L .85927 .44978 L .216 .736 .854 r
F
.85927 .44978 m .8528 .45968 L .82998 .46291 L .22 .74 .847 r
F
0 g
.8234 .47101 m
.8528 .45968 L
s
.8528 .45968 m
.8528 .45968 L
s
.72617 .4885 m
.73452 .48796 L
s
.70739 .48794 m
.72617 .4885 L
s
.8081 .46744 m .82998 .46291 L .80081 .47447 L .159 .7 .797 r
F
.8081 .46744 m .80081 .47447 L .77641 .47865 L .158 .699 .801 r
F
0 g
.73452 .48796 m
.75172 .48568 L
s
.69477 .48572 m
.70739 .48794 L
s
.52935 .45294 m .50071 .45832 L .50351 .45184 L .139 .682 .818 r
F
.52935 .45294 m .50351 .45184 L .5197 .45284 L .02 0 0 r
F
.52935 .45294 m .53908 .45305 L .52748 .4578 L .146 .69 .781 r
F
.50071 .45832 m .52935 .45294 L .52748 .4578 L .113 .664 .791 r
F
.52935 .45294 m .5197 .45284 L .53151 .45356 L .02 0 0 r
F
.52935 .45294 m .55551 .44633 L .53908 .45305 L .146 .69 .781 r
F
0 g
.52748 .4578 m
.52748 .4578 L
s
.52748 .4578 m
.53908 .45305 L
s
.49974 .46914 m
.52003 .46084 L
s
.52003 .46084 m
.52748 .4578 L
s
.23826 .45427 m .22006 .46039 L .21503 .445 L .395 .798 .984 r
F
.21503 .445 m .22006 .46039 L .19977 .45655 L .396 .799 .984 r
F
0 g
.68321 .48237 m
.69477 .48572 L
s
.75172 .48568 m
.76554 .4822 L
s
.41123 .46652 m .42207 .44857 L .43915 .46232 L .298 .765 .941 r
F
.38869 .45096 m .42207 .44857 L .41123 .46652 L .297 .76 .946 r
F
.41123 .46652 m .38238 .46443 L .38869 .45096 L .329 .777 .957 r
F
.18946 .43845 m .21503 .445 L .19977 .45655 L .397 .798 .984 r
F
.19977 .45655 m .17528 .45309 L .18946 .43845 L .4 .8 .985 r
F
.38238 .46443 m .36178 .45485 L .38869 .45096 L .327 .773 .959 r
F
.50071 .45832 m .46871 .46489 L .47556 .45323 L .224 .734 .886 r
F
.4513 .4549 m .47556 .45323 L .46871 .46489 L .217 .724 .896 r
F
.4513 .4549 m .46871 .46489 L .43915 .46232 L .252 .746 .911 r
F
0 g
.76554 .4822 m
.77562 .47989 L
s
.33707 .46141 m .33684 .44337 L .36178 .45485 L .358 .785 .972 r
F
.33707 .46141 m .31155 .45034 L .33684 .44337 L F
.28551 .43846 m .31155 .45034 L .28641 .45801 L .379 .791 .979 r
F
.28641 .45801 m .26102 .44641 L .28551 .43846 L .38 .792 .979 r
F
.17528 .45309 m .16146 .4335 L .18946 .43845 L .4 .8 .985 r
F
.17528 .45309 m .14768 .45096 L .16146 .4335 L .403 .801 .985 r
F
0 g
.77562 .47989 m
.77806 .47906 L
s
.77806 .47906 m
.78211 .47768 L
s
.23826 .45427 m .23525 .43504 L .26102 .44641 L .391 .796 .983 r
F
.23826 .45427 m .21503 .445 L .23525 .43504 L .393 .797 .983 r
F
.16146 .4335 m .14768 .45096 L .13083 .42926 L .403 .801 .985 r
F
.14768 .45096 m .11746 .44923 L .13083 .42926 L .404 .802 .986 r
F
.13083 .42926 m .11746 .44923 L .09856 .4272 L .404 .802 .985 r
F
0 g
.09416 .4567 m
.10548 .46916 L
s
.08649 .44827 m .09856 .4272 L .11746 .44923 L .405 .802 .986 r
F
.09856 .4272 m .08649 .44827 L .06713 .42698 L .405 .802 .985 r
F
0 g
.08649 .44827 m
.08649 .44827 L
s
.08649 .44827 m
.09416 .4567 L
s
.9125 .43667 m .88249 .44824 L .88959 .43516 L .269 .764 .9 r
F
.88249 .44824 m .85927 .44978 L .88959 .43516 L .275 .771 .889 r
F
0 g
.8528 .45968 m
.88249 .44824 L
s
.88249 .44824 m
.88249 .44824 L
s
.85927 .44978 m .82998 .46291 L .83573 .45437 L .2 .728 .828 r
F
.8081 .46744 m .83573 .45437 L .82998 .46291 L .202 .729 .825 r
F
.4513 .4549 m .43915 .46232 L .42207 .44857 L .269 .749 .931 r
F
.58384 .43475 m .55551 .44633 L .55746 .45008 L 0 0 0 r
F
.55746 .45008 m .55551 .44633 L .52935 .45294 L .046 0 0 r
F
0 g
.53908 .45305 m
.54267 .45159 L
s
.55551 .44633 m
.55551 .44633 L
s
.71784 .48671 m .68902 .48284 L .70662 .48767 L .088 0 0 r
F
.70662 .48767 m .68902 .48284 L .67812 .48168 L .068 0 0 r
F
0 g
.68034 .48074 m
.68288 .48219 L
s
.65898 .47062 m .67812 .48168 L .68902 .48284 L .159 0 0 r
F
.65898 .47062 m .64796 .46566 L .67812 .48168 L .122 0 0 r
F
0 g
.66559 .47364 m
.67145 .4757 L
s
.66208 .47241 m
.66559 .47364 L
s
.62666 .44441 m .64796 .46566 L .65898 .47062 L .246 0 0 r
F
.74589 .48497 m .71784 .48671 L .73444 .4871 L .025 0 0 r
F
.73444 .4871 m .71784 .48671 L .70662 .48767 L .015 0 0 r
F
.62666 .44441 m .61246 .42304 L .64796 .46566 L 0 .451 .766 r
F
0 g
.62666 .44441 m
.62666 .44441 L
s
.62666 .44441 m
.6406 .46537 L
s
.79061 .47197 m .8081 .46744 L .77641 .47865 L .103 .658 .744 r
F
.79061 .47197 m .77641 .47865 L .76245 .48143 L .104 .659 .745 r
F
0 g
.78211 .47768 m
.80697 .46853 L
s
.76245 .48143 m .77641 .47865 L .74908 .48431 L .042 .611 .709 r
F
.76245 .48143 m .74981 .4881 L .7762 .47847 L 0 0 0 r
F
.76245 .48143 m .7762 .47847 L .77681 .47825 L F
.76245 .48143 m .74908 .48431 L .74589 .48497 L .042 .611 .709 r
F
.76245 .48143 m .74589 .48497 L .73444 .4871 L .043 .612 .711 r
F
.50071 .45832 m .47556 .45323 L .50351 .45184 L .121 .666 .824 r
F
.36178 .45485 m .36288 .43762 L .38869 .45096 L .34 .777 .966 r
F
.36178 .45485 m .33684 .44337 L .36288 .43762 L .339 .776 .966 r
F
0 g
.65547 .46904 m
.66208 .47241 L
s
.31069 .43105 m .33684 .44337 L .31155 .45034 L .368 .786 .977 r
F
.31069 .43105 m .31155 .45034 L .28551 .43846 L .37 .788 .977 r
F
.20758 .42467 m .23525 .43504 L .21503 .445 L .395 .797 .984 r
F
.18946 .43845 m .20758 .42467 L .21503 .445 L .398 .799 .985 r
F
.26102 .44641 m .25893 .42606 L .28551 .43846 L .386 .793 .982 r
F
.26102 .44641 m .23525 .43504 L .25893 .42606 L .388 .795 .982 r
F
0 g
.6446 .46162 m
.65547 .46904 L
s
.80697 .46853 m
.81137 .46676 L
s
.94282 .42499 m .9125 .43667 L .92005 .42062 L .329 .788 .944 r
F
.92005 .42062 m .9125 .43667 L .88959 .43516 L .335 .799 .932 r
F
0 g
.88249 .44824 m
.9125 .43667 L
s
.9125 .43667 m
.9125 .43667 L
s
.8653 .43729 m .88959 .43516 L .85927 .44978 L .263 .766 .879 r
F
.8653 .43729 m .85927 .44978 L .83573 .45437 L .269 .771 .871 r
F
.17817 .41749 m .20758 .42467 L .18946 .43845 L .399 .798 .985 r
F
.16146 .4335 m .17817 .41749 L .18946 .43845 L .403 .8 .986 r
F
.61246 .42304 m .58384 .43475 L .5871 .45572 L .216 0 0 r
F
.58384 .43475 m .55746 .45008 L .5871 .45572 L .287 0 0 r
F
0 g
.55551 .44633 m
.58384 .43475 L
s
.58384 .43475 m
.58384 .43475 L
s
.81514 .46116 m .83573 .45437 L .8081 .46744 L .162 .702 .784 r
F
.81514 .46116 m .8081 .46744 L .79061 .47197 L .162 .702 .783 r
F
0 g
.81137 .46676 m
.83554 .45465 L
s
.38869 .45096 m .39092 .43294 L .42207 .44857 L .314 .766 .956 r
F
.38869 .45096 m .36288 .43762 L .39092 .43294 L .312 .764 .956 r
F
.47556 .45323 m .4513 .4549 L .45083 .44403 L .236 .733 .91 r
F
.4513 .4549 m .42207 .44857 L .45083 .44403 L .229 .728 .909 r
F
.55746 .45008 m .52935 .45294 L .53151 .45356 L 0 0 0 r
F
.16146 .4335 m .14804 .41148 L .17817 .41749 L .403 .8 .986 r
F
.16146 .4335 m .13083 .42926 L .14804 .41148 L .405 .802 .986 r
F
.3362 .42444 m .36288 .43762 L .33684 .44337 L .353 .779 .973 r
F
.3362 .42444 m .33684 .44337 L .31069 .43105 L .355 .781 .973 r
F
.28406 .41811 m .31069 .43105 L .28551 .43846 L .378 .789 .98 r
F
.28406 .41811 m .28551 .43846 L .25893 .42606 L .381 .791 .981 r
F
.23082 .4131 m .25893 .42606 L .23525 .43504 L .392 .795 .984 r
F
.20758 .42467 m .23082 .4131 L .23525 .43504 L .396 .797 .984 r
F
.14804 .41148 m .13083 .42926 L .11228 .40274 L .405 .802 .986 r
F
.13083 .42926 m .09856 .4272 L .11228 .40274 L .407 .802 .986 r
F
.47843 .44485 m .50351 .45184 L .47556 .45323 L .164 .692 .862 r
F
.45083 .44403 m .47843 .44485 L .47556 .45323 L .154 .683 .865 r
F
.09856 .4272 m .06713 .42698 L .07742 .40603 L .407 .803 .986 r
F
.0474 .40528 m .07742 .40603 L .06713 .42698 L .407 .804 .986 r
F
0 g
.06713 .42698 m
.08649 .44827 L
s
.06713 .42698 m
.06713 .42698 L
s
.70724 .48762 m
.7332 .48708 L
s
.70582 .48745 m
.70724 .48762 L
s
.42207 .44857 m .42192 .42805 L .45083 .44403 L .266 .742 .936 r
F
.42207 .44857 m .39092 .43294 L .42192 .42805 L .265 .742 .936 r
F
.07742 .40603 m .11228 .40274 L .09856 .4272 L .407 .803 .986 r
F
0 g
.7332 .48708 m
.73518 .48695 L
s
.70662 .48767 m .67812 .48168 L .69623 .49363 L .245 0 0 r
F
0 g
.70141 .48658 m
.70582 .48745 L
s
.69623 .49363 m .67812 .48168 L .66908 .48765 L .24 0 0 r
F
.66908 .48765 m .67812 .48168 L .64796 .46566 L .354 0 0 r
F
0 g
.68103 .4836 m
.70141 .48658 L
s
.70662 .48767 m .72306 .49351 L .73444 .4871 L .154 0 0 r
F
.70662 .48767 m .69623 .49363 L .72306 .49351 L .15 0 0 r
F
.94601 .40968 m .97346 .41318 L .94282 .42499 L .377 .801 .972 r
F
.92005 .42062 m .94601 .40968 L .94282 .42499 L .392 .816 .968 r
F
0 g
.9125 .43667 m
.94282 .42499 L
s
.94282 .42499 m
.94282 .42499 L
s
.73444 .4871 m .74981 .4881 L .76245 .48143 L .066 0 0 r
F
.73444 .4871 m .72306 .49351 L .74981 .4881 L F
.66908 .48765 m .64796 .46566 L .64205 .4761 L .371 0 0 r
F
.64205 .4761 m .64796 .46566 L .61246 .42304 L .543 0 0 r
F
.8982 .41726 m .92005 .42062 L .88959 .43516 L .336 .8 .933 r
F
.88959 .43516 m .8653 .43729 L .8982 .41726 L .341 .809 .919 r
F
0 g
.83554 .45465 m
.83642 .45424 L
s
.74732 .48457 m
.76111 .48214 L
s
.74463 .48504 m
.74732 .48457 L
s
.83677 .44623 m .8653 .43729 L .83573 .45437 L .227 .746 .83 r
F
.83677 .44623 m .83573 .45437 L .81514 .46116 L F
0 g
.83642 .45424 m
.83873 .45265 L
s
.83873 .45265 m
.83883 .45258 L
s
.77681 .47825 m .79061 .47197 L .76245 .48143 L 0 0 0 r
F
.64205 .4761 m .61246 .42304 L .61895 .47259 L .618 .126 .061 r
F
0 g
.62488 .44531 m
.63882 .45469 L
s
.61895 .47259 m .61246 .42304 L .5871 .45572 L .554 .092 .111 r
F
0 g
.58384 .43475 m
.61246 .42304 L
s
.61246 .42304 m
.61246 .42304 L
s
.50289 .45326 m
.51062 .45228 L
s
.48971 .45253 m
.50289 .45326 L
s
.67636 .48284 m
.68103 .4836 L
s
.5114 .45218 m
.53014 .45317 L
s
.47637 .45087 m
.48971 .45253 L
s
.65838 .47651 m
.67636 .48284 L
s
.76111 .48214 m
.7646 .48095 L
s
.36199 .41863 m .39092 .43294 L .36288 .43762 L .332 .769 .967 r
F
.36199 .41863 m .36288 .43762 L .3362 .42444 L .335 .772 .967 r
F
.23082 .4131 m .20758 .42467 L .19734 .39831 L .398 .797 .985 r
F
.19734 .39831 m .20758 .42467 L .17817 .41749 L .402 .799 .986 r
F
0 g
.45572 .44585 m
.47637 .45087 L
s
.5871 .45572 m .55746 .45008 L .55748 .45956 L .186 0 0 r
F
.55746 .45008 m .53151 .45356 L .55748 .45956 L .201 0 0 r
F
0 g
.53014 .45317 m
.55667 .45018 L
s
.55667 .45018 m
.55746 .45022 L
s
.55746 .45022 m
.55817 .45021 L
s
.55817 .45021 m
.58577 .44716 L
s
.80293 .46654 m .79061 .47197 L .80443 .46462 L .115 .667 .736 r
F
.77681 .47825 m .80443 .46462 L .79061 .47197 L .106 .659 .715 r
F
.78953 .47412 m .80443 .46462 L .77681 .47825 L 0 0 0 r
F
.78953 .47412 m .80605 .46255 L .80443 .46462 L .153 .692 .718 r
F
.80605 .46255 m .81732 .45464 L .80443 .46462 L F
.81514 .46116 m .80293 .46654 L .80443 .46462 L .115 .667 .736 r
F
0 g
.7646 .48095 m
.7743 .47786 L
s
.7743 .47786 m
.78363 .47488 L
s
.30947 .41094 m .3362 .42444 L .31069 .43105 L .367 .783 .978 r
F
.28406 .41811 m .30947 .41094 L .31069 .43105 L .37 .786 .978 r
F
.257 .40453 m .28406 .41811 L .25893 .42606 L .387 .792 .983 r
F
.23082 .4131 m .257 .40453 L .25893 .42606 L .391 .794 .983 r
F
.49252 .43899 m .50459 .44475 L .50572 .44816 L 0 .549 .762 r
F
.50572 .44816 m .53151 .45356 L .50702 .45206 L .051 .616 .776 r
F
.50572 .44816 m .53571 .45753 L .53151 .45356 L .115 0 0 r
F
.50459 .44475 m .51924 .45173 L .50572 .44816 L 0 .549 .762 r
F
.50572 .44816 m .50702 .45206 L .50351 .45184 L .051 .616 .776 r
F
.47843 .44485 m .50572 .44816 L .50351 .45184 L .04 .606 .779 r
F
0 g
.58577 .44716 m
.59742 .44242 L
s
.6442 .4723 m
.65838 .47651 L
s
.45088 .4452 m
.45572 .44585 L
s
.61501 .44249 m
.62488 .44531 L
s
.83677 .44623 m .81514 .46116 L .80443 .46462 L .184 .717 .774 r
F
0 g
.4477 .44453 m
.45088 .4452 L
s
.16618 .39833 m .19734 .39831 L .17817 .41749 L .403 .799 .986 r
F
.14804 .41148 m .16618 .39833 L .17817 .41749 L .406 .801 .987 r
F
.9314 .3994 m .94601 .40968 L .92005 .42062 L .4 .818 .972 r
F
0 g
.94282 .42499 m
.94681 .42345 L
s
.8982 .41726 m .9314 .3994 L .92005 .42062 L .4 .831 .957 r
F
.94601 .40968 m .96244 .39121 L .97346 .41318 L .44 .821 .99 r
F
0 g
.94681 .42345 m
.97346 .41318 L
s
.45083 .44403 m .45571 .43335 L .47843 .44485 L .2 .708 .901 r
F
.45083 .44403 m .42192 .42805 L .45571 .43335 L .195 .702 .903 r
F
0 g
.42196 .433 m
.4477 .44453 L
s
.59742 .44242 m
.61501 .44249 L
s
.63739 .46775 m
.6442 .4723 L
s
.87249 .42034 m .8982 .41726 L .8653 .43729 L .322 .801 .905 r
F
.87249 .42034 m .8653 .43729 L .83677 .44623 L .329 .809 .892 r
F
0 g
.83883 .45258 m
.86003 .43894 L
s
.42192 .42805 m .39092 .43294 L .38738 .41303 L .299 .752 .957 r
F
.38738 .41303 m .39092 .43294 L .36199 .41863 L .309 .759 .958 r
F
.14804 .41148 m .14017 .38741 L .16618 .39833 L .406 .801 .987 r
F
.14804 .41148 m .11228 .40274 L .14017 .38741 L .409 .803 .987 r
F
.33472 .40448 m .36199 .41863 L .3362 .42444 L .352 .775 .975 r
F
.30947 .41094 m .33472 .40448 L .3362 .42444 L .357 .779 .975 r
F
0 g
.86003 .43894 m
.86854 .42965 L
s
.28406 .41811 m .28276 .39705 L .30947 .41094 L .38 .787 .982 r
F
.257 .40453 m .28276 .39705 L .28406 .41811 L .384 .79 .982 r
F
.22946 .38942 m .257 .40453 L .23082 .4131 L .396 .795 .985 r
F
.23082 .4131 m .19734 .39831 L .22946 .38942 L .4 .797 .986 r
F
.9314 .3994 m .96244 .39121 L .94601 .40968 L .423 .829 .977 r
F
.69623 .49363 m .71234 .50487 L .72306 .49351 L .283 0 0 r
F
.68706 .50544 m .71234 .50487 L .69623 .49363 L .284 0 0 r
F
.69623 .49363 m .66908 .48765 L .68706 .50544 L .374 0 0 r
F
.53571 .45753 m .55748 .45956 L .53151 .45356 L .104 0 0 r
F
.74981 .4881 m .72306 .49351 L .7377 .49928 L .186 0 0 r
F
.72306 .49351 m .71234 .50487 L .7377 .49928 L .188 0 0 r
F
.66908 .48765 m .66216 .50131 L .68706 .50544 L .38 0 0 r
F
.66908 .48765 m .64205 .4761 L .66216 .50131 L .459 0 0 r
F
0 g
.80401 .46516 m
.80695 .4638 L
s
.8025 .46585 m
.80401 .46516 L
s
.11228 .40274 m .07742 .40603 L .08177 .38366 L .41 .803 .987 r
F
.07742 .40603 m .05749 .38985 L .08177 .38366 L .411 .804 .987 r
F
.07742 .40603 m .0474 .40528 L .05749 .38985 L .41 .804 .987 r
F
.02728 .38315 m .05749 .38985 L .0474 .40528 L .41 .805 .986 r
F
0 g
.0474 .40528 m
.0474 .40528 L
s
.0474 .40528 m
.06713 .42698 L
s
.77681 .47825 m .74981 .4881 L .76331 .48894 L .079 0 0 r
F
.74981 .4881 m .7377 .49928 L .76331 .48894 L .084 0 0 r
F
.61895 .47259 m .5871 .45572 L .60254 .48431 L .444 0 0 r
F
0 g
.61758 .46213 m
.63025 .46572 L
s
.63025 .46572 m
.63739 .46775 L
s
.5871 .45572 m .5768 .47319 L .60254 .48431 L .442 0 0 r
F
.5871 .45572 m .55748 .45956 L .5768 .47319 L .358 0 0 r
F
0 g
.69468 .49563 m
.69902 .49557 L
s
.67435 .49287 m
.69468 .49563 L
s
.69902 .49557 m
.72132 .49535 L
s
.4796 .43859 m .50572 .44816 L .47843 .44485 L .113 .652 .851 r
F
.45571 .43335 m .4796 .43859 L .47843 .44485 L .112 .651 .851 r
F
.64113 .49466 m .66216 .50131 L .64205 .4761 L .478 0 0 r
F
.64113 .49466 m .64205 .4761 L .61895 .47259 L .501 0 0 r
F
0 g
.72132 .49535 m
.72587 .49462 L
s
.72587 .49462 m
.74612 .4915 L
s
.41378 .42933 m
.42196 .433 L
s
.77681 .47825 m .76331 .48894 L .78953 .47412 L 0 0 0 r
F
.10977 .37338 m .14017 .38741 L .11228 .40274 L .41 .803 .987 r
F
.11228 .40274 m .08177 .38366 L .10977 .37338 L .412 .804 .987 r
F
0 g
.66688 .49199 m
.67435 .49287 L
s
.59907 .46206 m
.61758 .46213 L
s
.58995 .46099 m
.59907 .46206 L
s
.65145 .48789 m
.66688 .49199 L
s
.58462 .45993 m
.58995 .46099 L
s
.91068 .38831 m .9314 .3994 L .8982 .41726 L .42 .836 .967 r
F
.87249 .42034 m .91068 .38831 L .8982 .41726 L .425 .851 .949 r
F
0 g
.74612 .4915 m
.75558 .48846 L
s
.81616 .45795 m .80443 .46462 L .81732 .45464 L .168 .706 .759 r
F
.7999 .46898 m .81732 .45464 L .78953 .47412 L 0 0 0 r
F
.81848 .45134 m .81732 .45464 L .7999 .46898 L .222 .74 .736 r
F
.8259 .44429 m .81732 .45464 L .81848 .45134 L F
.83677 .44623 m .81616 .45795 L .81732 .45464 L .168 .706 .759 r
F
0 g
.80695 .4638 m
.81306 .45971 L
s
.56321 .45882 m
.58462 .45993 L
s
.81306 .45971 m
.81627 .45764 L
s
.38738 .41303 m .36199 .41863 L .35915 .39858 L .331 .763 .97 r
F
.33472 .40448 m .35915 .39858 L .36199 .41863 L .34 .77 .971 r
F
.30947 .41094 m .30809 .39002 L .33472 .40448 L .369 .781 .98 r
F
.30947 .41094 m .28276 .39705 L .30809 .39002 L .375 .785 .981 r
F
0 g
.55385 .45872 m
.55748 .45889 L
s
.55748 .45889 m
.56321 .45882 L
s
.87249 .42034 m .83677 .44623 L .84699 .43033 L .281 .782 .849 r
F
0 g
.86854 .42965 m
.87974 .41947 L
s
.83677 .44623 m .81732 .45464 L .84699 .43033 L .282 .784 .824 r
F
0 g
.81627 .45764 m
.8282 .44994 L
s
.28276 .39705 m .257 .40453 L .2565 .38298 L .391 .791 .985 r
F
.2565 .38298 m .257 .40453 L .22946 .38942 L .395 .794 .985 r
F
0 g
.75558 .48846 m
.76897 .48445 L
s
.16618 .39833 m .16813 .37703 L .19734 .39831 L .407 .8 .987 r
F
.14017 .38741 m .16813 .37703 L .16618 .39833 L .409 .802 .987 r
F
.5768 .47319 m .55748 .45956 L .55626 .46687 L .274 0 0 r
F
.55626 .46687 m .55748 .45956 L .53571 .45753 L .265 0 0 r
F
0 g
.53383 .45575 m
.55385 .45872 L
s
.64156 .48601 m
.65145 .48789 L
s
.64113 .49466 m .61895 .47259 L .62799 .50241 L .522 0 0 r
F
.61895 .47259 m .60254 .48431 L .62799 .50241 L .52 0 0 r
F
.42192 .42805 m .43777 .41696 L .45571 .43335 L .238 .722 .93 r
F
.40845 .40609 m .43777 .41696 L .42192 .42805 L .243 .719 .938 r
F
.42192 .42805 m .38738 .41303 L .40845 .40609 L .283 .744 .95 r
F
0 g
.40396 .42024 m
.41378 .42933 L
s
.19734 .39831 m .20214 .36725 L .22946 .38942 L .405 .798 .987 r
F
.19734 .39831 m .16813 .37703 L .20214 .36725 L .407 .8 .987 r
F
0 g
.62921 .4828 m
.64156 .48601 L
s
.9314 .3994 m .9512 .36881 L .96244 .39121 L .476 .839 .993 r
F
0 g
.96244 .39121 m
.97346 .41318 L
s
.91068 .38831 m .9512 .36881 L .9314 .3994 L .462 .855 .976 r
F
0 g
.96244 .39121 m
.96244 .39121 L
s
.51975 .45127 m .51924 .45173 L .50595 .44169 L 0 .478 .716 r
F
.51924 .45173 m .53571 .45753 L .51871 .45222 L 0 .539 .721 r
F
.54426 .46528 m .53571 .45753 L .51924 .45173 L .183 0 0 r
F
.54426 .46528 m .5292 .45712 L .53171 .45958 L .131 0 0 r
F
.5292 .45712 m .51924 .45173 L .53171 .45958 L F
.53171 .45958 m .51924 .45173 L .51975 .45127 L 0 .478 .716 r
F
.53171 .45958 m .50595 .44169 L .52025 .4488 L 0 .52 .767 r
F
.53171 .45958 m .52025 .4488 L .53419 .46201 L .177 0 0 r
F
.53171 .45958 m .53419 .46201 L .54508 .47232 L F
.51924 .45173 m .51871 .45222 L .50572 .44816 L 0 .539 .721 r
F
0 g
.52569 .45367 m
.53383 .45575 L
s
.52567 .45367 m
.52569 .45367 L
s
.52534 .45358 m
.52567 .45367 L
s
.49252 .43899 m .50572 .44816 L .4796 .43859 L .055 .613 .809 r
F
0 g
.62154 .48114 m
.62921 .4828 L
s
.05749 .38985 m .05358 .3709 L .08177 .38366 L .411 .805 .987 r
F
.02728 .38315 m .05358 .3709 L .05749 .38985 L .413 .806 .987 r
F
0 g
.02728 .38315 m
.0474 .40528 L
s
.45571 .43335 m .46597 .42709 L .4796 .43859 L .163 .68 .89 r
F
.46597 .42709 m .45571 .43335 L .43777 .41696 L .166 .677 .898 r
F
0 g
.50772 .44869 m
.51307 .45046 L
s
.61009 .47892 m
.62154 .48114 L
s
.7377 .49928 m .71234 .50487 L .72629 .51551 L .301 0 0 r
F
.70258 .52154 m .72629 .51551 L .71234 .50487 L .303 0 0 r
F
.70258 .52154 m .71234 .50487 L .68706 .50544 L .39 0 0 r
F
.70258 .52154 m .68706 .50544 L .67892 .52312 L .392 0 0 r
F
.68706 .50544 m .66216 .50131 L .67892 .52312 L .462 0 0 r
F
.7503 .50473 m .76331 .48894 L .7377 .49928 L .193 0 0 r
F
.7377 .49928 m .72629 .51551 L .7503 .50473 L .197 0 0 r
F
0 g
.50171 .44538 m
.50736 .44849 L
s
.50736 .44849 m
.50772 .44869 L
s
.33472 .40448 m .33286 .38373 L .35915 .39858 L .356 .772 .978 r
F
.33472 .40448 m .30809 .39002 L .33286 .38373 L .364 .779 .979 r
F
.66216 .50131 m .65383 .52076 L .67892 .52312 L .463 0 0 r
F
.64113 .49466 m .65383 .52076 L .66216 .50131 L .509 0 0 r
F
.28155 .37537 m .30809 .39002 L .28276 .39705 L .384 .786 .984 r
F
.28155 .37537 m .28276 .39705 L .2565 .38298 L .39 .791 .985 r
F
0 g
.59893 .47762 m
.61009 .47892 L
s
.13753 .36119 m .16813 .37703 L .14017 .38741 L .411 .802 .988 r
F
.10977 .37338 m .13753 .36119 L .14017 .38741 L .413 .804 .988 r
F
.64113 .49466 m .62799 .50241 L .65383 .52076 L .508 0 0 r
F
0 g
.68563 .50854 m
.68998 .50847 L
s
.66611 .50645 m
.68563 .50854 L
s
.68998 .50847 m
.71035 .50828 L
s
.76331 .48894 m .77476 .48916 L .78953 .47412 L .065 0 0 r
F
0 g
.76897 .48445 m
.7893 .47425 L
s
.7503 .50473 m .77476 .48916 L .76331 .48894 L .072 0 0 r
F
.38123 .39275 m .40845 .40609 L .38738 .41303 L .302 .746 .963 r
F
0 g
.39305 .41116 m
.40396 .42024 L
s
.38738 .41303 m .35915 .39858 L .38123 .39275 L .321 .759 .967 r
F
0 g
.71035 .50828 m
.71575 .50747 L
s
.87249 .42034 m .87719 .40019 L .91068 .38831 L .398 .843 .93 r
F
0 g
.87974 .41947 m
.88323 .41133 L
s
.87249 .42034 m .84699 .43033 L .87719 .40019 L .413 .855 .918 r
F
.55626 .46687 m .53571 .45753 L .54426 .46528 L .188 0 0 r
F
.49252 .43899 m .4796 .43859 L .46597 .42709 L .084 .626 .849 r
F
0 g
.71575 .50747 m
.73372 .50494 L
s
.23377 .37002 m .2565 .38298 L .22946 .38942 L .401 .795 .987 r
F
.23377 .37002 m .22946 .38942 L .20214 .36725 L .403 .797 .987 r
F
0 g
.48541 .43877 m
.50171 .44538 L
s
.08177 .38366 m .08016 .35851 L .10977 .37338 L .413 .804 .988 r
F
.05358 .3709 m .08016 .35851 L .08177 .38366 L .414 .806 .988 r
F
0 g
.02728 .38315 m
.05358 .3709 L
s
.66021 .50586 m
.66611 .50645 L
s
.64526 .50315 m
.66021 .50586 L
s
.77476 .48916 m .7999 .46898 L .78953 .47412 L 0 0 0 r
F
0 g
.7893 .47425 m
.78948 .47417 L
s
.78948 .47417 m
.7896 .47409 L
s
.7896 .47409 m
.8049 .46423 L
s
.73372 .50494 m
.74377 .50191 L
s
.5901 .49376 m .60254 .48431 L .5768 .47319 L .424 0 0 r
F
.5901 .49376 m .5768 .47319 L .56703 .47918 L .408 0 0 r
F
0 g
.5771 .47332 m
.59893 .47762 L
s
.5768 .47319 m .55626 .46687 L .56703 .47918 L .33 0 0 r
F
0 g
.57685 .47327 m
.5771 .47332 L
s
.57674 .47323 m
.57685 .47327 L
s
.63147 .50036 m
.64526 .50315 L
s
.60254 .48431 m .61381 .5114 L .62799 .50241 L .498 0 0 r
F
.60254 .48431 m .5901 .49376 L .61381 .5114 L .499 0 0 r
F
0 g
.74377 .50191 m
.7553 .49866 L
s
.55626 .46687 m .54426 .46528 L .56703 .47918 L .304 0 0 r
F
0 g
.55785 .46869 m
.57674 .47323 L
s
.8259 .44429 m .84699 .43033 L .81732 .45464 L .223 .743 .766 r
F
0 g
.8282 .44994 m
.84466 .43225 L
s
.62708 .49941 m
.63147 .50036 L
s
.55195 .4663 m
.55785 .46869 L
s
.30685 .36821 m .33286 .38373 L .30809 .39002 L .376 .78 .984 r
F
.30685 .36821 m .30809 .39002 L .28155 .37537 L .384 .786 .984 r
F
0 g
.62256 .49855 m
.62708 .49941 L
s
.38123 .39275 m .35915 .39858 L .35596 .37814 L .339 .761 .975 r
F
.35915 .39858 m .33286 .38373 L .35596 .37814 L .352 .771 .977 r
F
.28155 .37537 m .2565 .38298 L .25495 .36125 L .396 .791 .987 r
F
.23377 .37002 m .25495 .36125 L .2565 .38298 L .402 .795 .987 r
F
.1663 .34938 m .20214 .36725 L .16813 .37703 L .411 .801 .989 r
F
.13753 .36119 m .1663 .34938 L .16813 .37703 L .415 .803 .989 r
F
0 g
.60707 .49519 m
.62256 .49855 L
s
.542 .46322 m
.55195 .4663 L
s
.65383 .52076 m .62799 .50241 L .63696 .53085 L .53 .006 0 r
F
.62799 .50241 m .61381 .5114 L .63696 .53085 L .538 .002 0 r
F
.13753 .36119 m .10977 .37338 L .10704 .34599 L .414 .804 .988 r
F
.08016 .35851 m .10704 .34599 L .10977 .37338 L .416 .806 .988 r
F
0 g
.05358 .3709 m
.08016 .35851 L
s
.05358 .3709 m
.05358 .3709 L
s
.59348 .49119 m
.60707 .49519 L
s
.71589 .53695 m .72629 .51551 L .70258 .52154 L .399 0 0 r
F
.69408 .54332 m .71589 .53695 L .70258 .52154 L .4 0 0 r
F
.69408 .54332 m .70258 .52154 L .67892 .52312 L .471 0 0 r
F
.93974 .34596 m .9512 .36881 L .91068 .38831 L .526 .863 .994 r
F
0 g
.9512 .36881 m
.96244 .39121 L
s
.93974 .34596 m .91068 .38831 L .90477 .36433 L .5 .877 .974 r
F
.87719 .40019 m .90477 .36433 L .91068 .38831 L .514 .885 .973 r
F
0 g
.88323 .41133 m
.89268 .39469 L
s
.72629 .51551 m .73812 .52587 L .7503 .50473 L .304 0 0 r
F
.71589 .53695 m .73812 .52587 L .72629 .51551 L .305 0 0 r
F
0 g
.53352 .45946 m
.542 .46322 L
s
.67294 .54559 m .69408 .54332 L .67892 .52312 L .471 0 0 r
F
.67294 .54559 m .67892 .52312 L .65383 .52076 L .523 .012 0 r
F
.43777 .41696 m .45149 .40998 L .46597 .42709 L .208 .698 .924 r
F
.43777 .41696 m .42311 .39679 L .45149 .40998 L .214 .696 .931 r
F
.40845 .40609 m .42311 .39679 L .43777 .41696 L .269 .73 .95 r
F
.85212 .41518 m .87719 .40019 L .84699 .43033 L .361 .831 .873 r
F
.85212 .41518 m .84699 .43033 L .8259 .44429 L .372 .839 .856 r
F
0 g
.84466 .43225 m
.84602 .43097 L
s
.9512 .36881 m
.9512 .36881 L
s
.84602 .43097 m
.84784 .42783 L
s
.67859 .52439 m
.67987 .52438 L
s
.67987 .52438 m
.70154 .52419 L
s
.70154 .52419 m
.70452 .52379 L
s
.39887 .38762 m .42311 .39679 L .40845 .40609 L .275 .728 .956 r
F
.39887 .38762 m .40845 .40609 L .38123 .39275 L .308 .748 .965 r
F
0 g
.38174 .393 m
.39305 .41116 L
s
.49252 .43899 m .46597 .42709 L .47907 .42488 L .12 .647 .874 r
F
0 g
.47329 .43037 m
.48541 .43877 L
s
.47907 .42488 m .46597 .42709 L .45149 .40998 L .131 .648 .89 r
F
0 g
.46858 .42665 m
.47329 .43037 L
s
.70452 .52379 m
.72346 .52135 L
s
.49252 .43899 m .50595 .44169 L .51924 .45173 L .007 .573 .791 r
F
.49252 .43899 m .47907 .42488 L .50595 .44169 L .03 .581 .827 r
F
.77476 .48916 m .7503 .50473 L .76082 .50982 L .181 0 0 r
F
0 g
.7553 .49866 m
.77386 .48973 L
s
.7503 .50473 m .73812 .52587 L .76082 .50982 L .184 0 0 r
F
.67294 .54559 m .65383 .52076 L .65567 .54658 L .524 .015 0 r
F
.65383 .52076 m .63696 .53085 L .65567 .54658 L .552 .036 0 r
F
0 g
.65482 .52204 m
.67859 .52439 L
s
.58767 .49001 m
.59348 .49119 L
s
.65391 .52196 m
.65482 .52204 L
s
.65234 .52165 m
.65386 .52195 L
s
.65386 .52195 m
.65391 .52196 L
s
.52955 .45747 m
.53352 .45946 L
s
.52937 .45737 m
.52955 .45747 L
s
.63299 .51825 m
.65234 .52165 L
s
.72346 .52135 m
.73067 .51934 L
s
.7999 .46898 m .77476 .48916 L .78404 .48865 L .028 0 0 r
F
.78404 .48865 m .77476 .48916 L .76082 .50982 L .033 0 0 r
F
0 g
.77386 .48973 m
.77458 .48943 L
s
.56703 .47918 m .54426 .46528 L .55583 .47824 L .278 0 0 r
F
.54426 .46528 m .53171 .45958 L .55583 .47824 L .256 0 0 r
F
0 g
.77458 .48943 m
.775 .48915 L
s
.33159 .36247 m .35596 .37814 L .33286 .38373 L .366 .772 .983 r
F
.33159 .36247 m .33286 .38373 L .30685 .36821 L .375 .78 .984 r
F
.30685 .36821 m .28155 .37537 L .28002 .35242 L .392 .787 .987 r
F
.28002 .35242 m .28155 .37537 L .25495 .36125 L .399 .792 .988 r
F
.23377 .37002 m .22634 .34917 L .25495 .36125 L .404 .795 .988 r
F
.23377 .37002 m .20214 .36725 L .22634 .34917 L .411 .799 .989 r
F
.19451 .33625 m .22634 .34917 L .20214 .36725 L F
.1663 .34938 m .19451 .33625 L .20214 .36725 L .417 .803 .99 r
F
0 g
.73067 .51934 m
.74388 .51586 L
s
.57879 .49756 m .5901 .49376 L .56703 .47918 L .402 0 0 r
F
0 g
.57991 .48732 m
.58767 .49001 L
s
.57879 .49756 m .56703 .47918 L .55583 .47824 L .382 0 0 r
F
0 g
.57002 .48385 m
.57991 .48732 L
s
.5901 .49376 m .60126 .51805 L .61381 .5114 L .492 0 0 r
F
.5901 .49376 m .57879 .49756 L .60126 .51805 L .487 0 0 r
F
.39887 .38762 m .38123 .39275 L .37684 .37184 L .322 .749 .973 r
F
0 g
.38154 .39266 m
.38174 .393 L
s
.38123 .39275 m .35596 .37814 L .37684 .37184 L .34 .762 .976 r
F
0 g
.38077 .39003 m
.38154 .39266 L
s
.38039 .38872 m
.38077 .39003 L
s
.7916 .4866 m .80684 .46528 L .80733 .46347 L 0 0 0 r
F
.8259 .44429 m .80789 .46139 L .80733 .46347 L F
.83132 .43297 m .8259 .44429 L .80733 .46347 L .312 .799 .765 r
F
.80684 .46528 m .81078 .45976 L .80733 .46347 L 0 0 0 r
F
.80789 .46139 m .7999 .46898 L .80733 .46347 L F
.80733 .46347 m .7999 .46898 L .78404 .48865 L F
0 g
.775 .48915 m
.79223 .47849 L
s
.13753 .36119 m .1342 .33333 L .1663 .34938 L .416 .803 .989 r
F
.13753 .36119 m .10704 .34599 L .1342 .33333 L .419 .806 .989 r
F
0 g
.74388 .51586 m
.75913 .509 L
s
.56001 .47859 m
.57002 .48385 L
s
.08016 .35851 m
.10704 .34599 L
s
.08016 .35851 m
.08016 .35851 L
s
.62336 .54031 m .63696 .53085 L .61381 .5114 L .548 .02 0 r
F
.61381 .5114 m .60126 .51805 L .62336 .54031 L .556 .011 0 r
F
0 g
.61718 .51423 m
.63299 .51825 L
s
.61453 .51359 m
.61718 .51423 L
s
.876 .38331 m .90477 .36433 L .87719 .40019 L .469 .881 .938 r
F
0 g
.89268 .39469 m
.89271 .38001 L
s
.85212 .41518 m .876 .38331 L .87719 .40019 L .486 .892 .934 r
F
0 g
.61165 .51254 m
.6142 .51347 L
s
.6142 .51347 m
.61453 .51359 L
s
.59667 .50806 m
.61165 .51254 L
s
.67379 .54242 m
.69331 .54229 L
s
.67024 .54208 m
.67379 .54242 L
s
.69331 .54229 m
.69448 .54228 L
s
.83132 .43297 m .85212 .41518 L .8259 .44429 L .303 .796 .788 r
F
0 g
.84784 .42783 m
.85845 .4114 L
s
.25165 .33725 m .28002 .35242 L .25495 .36125 L .405 .792 .989 r
F
.25165 .33725 m .25495 .36125 L .22634 .34917 L .412 .797 .99 r
F
.69408 .54332 m .70669 .56357 L .71589 .53695 L .478 0 .035 r
F
0 g
.69448 .54228 m
.70024 .54152 L
s
.68681 .56964 m .70669 .56357 L .69408 .54332 L .478 0 .032 r
F
.68681 .56964 m .69408 .54332 L .67294 .54559 L .536 .061 .073 r
F
.65567 .54658 m .63696 .53085 L .6463 .56541 L .576 .083 .048 r
F
0 g
.65996 .54113 m
.67024 .54208 L
s
.65525 .5407 m
.65996 .54113 L
s
.62336 .54031 m .6463 .56541 L .63696 .53085 L .591 .079 .001 r
F
.33159 .36247 m .30685 .36821 L .30622 .34494 L .387 .781 .988 r
F
.30685 .36821 m .28002 .35242 L .30622 .34494 L .396 .788 .988 r
F
.72709 .55237 m .73812 .52587 L .71589 .53695 L .398 0 0 r
F
.72709 .55237 m .71589 .53695 L .70669 .56357 L .399 0 0 r
F
0 g
.70024 .54152 m
.71488 .53988 L
s
.55377 .47593 m
.56001 .47859 L
s
.3531 .35911 m .37684 .37184 L .35596 .37814 L .352 .762 .981 r
F
.33159 .36247 m .3531 .35911 L .35596 .37814 L .367 .773 .983 r
F
0 g
.64672 .53905 m
.65429 .54052 L
s
.65429 .54052 m
.65525 .5407 L
s
.82384 .44642 m
.82761 .44072 L
s
.75913 .509 m
.76226 .50769 L
s
.76082 .50982 m .73812 .52587 L .74787 .53616 L .291 0 0 r
F
.73812 .52587 m .72709 .55237 L .74787 .53616 L .292 0 0 r
F
.67294 .54559 m .66762 .57132 L .68681 .56964 L .536 .055 .061 r
F
.67294 .54559 m .65567 .54658 L .66762 .57132 L .575 .098 .084 r
F
0 g
.71488 .53988 m
.71753 .53921 L
s
.71753 .53921 m
.73424 .53518 L
s
.63884 .53781 m
.64672 .53905 L
s
.63039 .53543 m
.63564 .53691 L
s
.63564 .53691 m
.63884 .53781 L
s
.76082 .50982 m .76943 .51459 L .78404 .48865 L .147 0 0 r
F
0 g
.76226 .50769 m
.76565 .50542 L
s
.76082 .50982 m .74787 .53616 L .76943 .51459 L .148 0 0 r
F
.39887 .38762 m .40422 .36961 L .42311 .39679 L .315 .742 .972 r
F
.39887 .38762 m .37684 .37184 L .40422 .36961 L .319 .748 .972 r
F
0 g
.37518 .37234 m
.38039 .38872 L
s
.45149 .40998 m .46552 .4054 L .47907 .42488 L .189 .678 .924 r
F
0 g
.4547 .41171 m
.46858 .42665 L
s
.45149 .40998 m .43755 .38818 L .46552 .4054 L .196 .676 .931 r
F
.42311 .39679 m .43755 .38818 L .45149 .40998 L .262 .718 .954 r
F
0 g
.45316 .40943 m
.4547 .41171 L
s
.50595 .44169 m .47907 .42488 L .49332 .42571 L .083 .615 .866 r
F
.46552 .4054 m .49332 .42571 L .47907 .42488 L .098 .616 .885 r
F
.50595 .44169 m .49332 .42571 L .52025 .4488 L 0 .533 .807 r
F
0 g
.50766 .44288 m
.52361 .45448 L
s
.50664 .44204 m
.50766 .44288 L
s
.42311 .39679 m .40422 .36961 L .43755 .38818 L .266 .715 .958 r
F
.65567 .54658 m .6463 .56541 L .66762 .57132 L .575 .082 .047 r
F
.54508 .47232 m .5415 .46716 L .53171 .45958 L .211 0 0 r
F
.54508 .47232 m .52988 .45793 L .5357 .46144 L 0 .461 .739 r
F
.52988 .45793 m .52025 .4488 L .5357 .46144 L F
.52853 .44595 m .54133 .46485 L .5357 .46144 L 0 .43 .777 r
F
.54508 .47232 m .5486 .47741 L .55901 .49031 L .244 0 0 r
F
.54508 .47232 m .55583 .47824 L .5415 .46716 L .211 0 0 r
F
.54508 .47232 m .56824 .49622 L .55583 .47824 L .331 0 0 r
F
.55901 .49031 m .56824 .49622 L .555 .48257 L .293 0 0 r
F
.55901 .49031 m .555 .48257 L .54508 .47232 L F
.54508 .47232 m .5357 .46144 L .5486 .47741 L .244 0 0 r
F
.55901 .49031 m .54539 .47344 L .55146 .47979 L 0 .393 .704 r
F
.54539 .47344 m .5357 .46144 L .55146 .47979 L F
.55725 .48585 m .55146 .47979 L .54602 .46514 L 0 .368 .754 r
F
.55901 .49031 m .55146 .47979 L .56293 .49787 L .316 0 0 r
F
.55901 .49031 m .56293 .49787 L .57308 .51387 L F
.55901 .49031 m .58068 .51915 L .56824 .49622 L .412 0 0 r
F
.55901 .49031 m .57308 .51387 L .58068 .51915 L .379 0 0 r
F
.56644 .5028 m .55146 .47979 L .55725 .48585 L 0 .368 .754 r
F
.54133 .46485 m .55146 .47979 L .5357 .46144 L 0 .43 .777 r
F
0 g
.55092 .47419 m
.55377 .47593 L
s
.16167 .32054 m .19451 .33625 L .1663 .34938 L .418 .803 .99 r
F
.1663 .34938 m .1342 .33333 L .16167 .32054 L .422 .806 .99 r
F
0 g
.50087 .43526 m
.50664 .44204 L
s
.54507 .47036 m
.54891 .47293 L
s
.57879 .49756 m .55583 .47824 L .56824 .49622 L .362 0 0 r
F
0 g
.621 .53319 m
.63039 .53543 L
s
.73424 .53518 m
.74337 .53141 L
s
.10704 .34599 m
.1342 .33333 L
s
.10704 .34599 m
.10704 .34599 L
s
.90477 .36433 m .92804 .32265 L .93974 .34596 L .581 .886 .991 r
F
0 g
.93974 .34596 m
.93974 .34596 L
s
.93974 .34596 m
.9512 .36881 L
s
.57879 .49756 m .59014 .52057 L .60126 .51805 L .48 0 0 r
F
.56824 .49622 m .59014 .52057 L .57879 .49756 L .464 0 0 r
F
0 g
.58439 .50267 m
.59667 .50806 L
s
.85495 .40181 m .876 .38331 L .85212 .41518 L .434 .873 .887 r
F
0 g
.85845 .4114 m
.86001 .40466 L
s
.85495 .40181 m .85212 .41518 L .83132 .43297 L .449 .882 .869 r
F
0 g
.58044 .50091 m
.58439 .50267 L
s
.80733 .46347 m .78404 .48865 L .7916 .4866 L 0 0 0 r
F
0 g
.79223 .47849 m
.80433 .46671 L
s
.7916 .4866 m .78404 .48865 L .76943 .51459 L 0 0 0 r
F
0 g
.76565 .50542 m
.77895 .49769 L
s
.22136 .32264 m .25165 .33725 L .22634 .34917 L .414 .797 .991 r
F
.19451 .33625 m .22136 .32264 L .22634 .34917 L .42 .802 .991 r
F
.61094 .54547 m .62336 .54031 L .60126 .51805 L .559 .016 0 r
F
0 g
.61364 .53053 m
.621 .53319 L
s
.60126 .51805 m .59014 .52057 L .61094 .54547 L .561 0 0 r
F
0 g
.57432 .497 m
.58044 .50091 L
s
.33159 .36247 m .3334 .34009 L .3531 .35911 L .382 .774 .988 r
F
.33159 .36247 m .30622 .34494 L .3334 .34009 L .39 .782 .988 r
F
0 g
.74337 .53141 m
.75183 .52811 L
s
.28002 .35242 m .27832 .32725 L .30622 .34494 L .403 .788 .99 r
F
.28002 .35242 m .25165 .33725 L .27832 .32725 L .412 .795 .991 r
F
0 g
.68885 .56226 m
.70482 .56056 L
s
.68254 .56223 m
.68885 .56226 L
s
.66955 .56199 m
.68254 .56223 L
s
.92804 .32265 m .90477 .36433 L .89691 .34516 L .561 .905 .975 r
F
.876 .38331 m .89691 .34516 L .90477 .36433 L .572 .909 .974 r
F
0 g
.89271 .38001 m
.89534 .37055 L
s
.60451 .52727 m
.61364 .53053 L
s
.66278 .5613 m
.66955 .56199 L
s
.49203 .42564 m
.50087 .43526 L
s
.64938 .55922 m
.66278 .5613 L
s
.80433 .46671 m
.80619 .46514 L
s
.81078 .45976 m .8009 .47675 L .80997 .46142 L 0 0 0 r
F
.83132 .43297 m .81167 .45794 L .81078 .45976 L F
.8224 .44041 m .83132 .43297 L .81078 .45976 L .358 .826 .768 r
F
.81078 .45976 m .80997 .46142 L .8224 .44041 L 0 0 0 r
F
.81167 .45794 m .80733 .46347 L .81078 .45976 L F
0 g
.8314 .43548 m
.83633 .42868 L
s
.82761 .44072 m
.8314 .43548 L
s
.70482 .56056 m
.70785 .56022 L
s
.40422 .36961 m .37684 .37184 L .36677 .3425 L .343 .75 .981 r
F
.36677 .3425 m .37684 .37184 L .3531 .35911 L .372 .769 .986 r
F
0 g
.37506 .37088 m
.37518 .37234 L
s
.63025 .57089 m .6463 .56541 L .62336 .54031 L .601 .1 .034 r
F
.63025 .57089 m .62336 .54031 L .61094 .54547 L .621 .102 0 r
F
0 g
.37506 .37088 m
.37546 .36781 L
s
.56756 .49331 m
.57432 .497 L
s
.56542 .49215 m
.56756 .49331 L
s
.80619 .46514 m
.80807 .46268 L
s
.64565 .55864 m
.64938 .55922 L
s
.64442 .55845 m
.64565 .55864 L
s
.59117 .52034 m
.60451 .52727 L
s
.6384 .55676 m
.64442 .55845 L
s
.70669 .56357 m .71742 .58483 L .72709 .55237 L .476 .03 .111 r
F
0 g
.70785 .56022 m
.71685 .558 L
s
.69858 .59504 m .71742 .58483 L .70669 .56357 L .476 .03 .11 r
F
.69858 .59504 m .70669 .56357 L .68681 .56964 L .542 .104 .158 r
F
.68681 .56964 m .68049 .59963 L .69858 .59504 L .543 .099 .145 r
F
.68681 .56964 m .66762 .57132 L .68049 .59963 L .589 .15 .175 r
F
.73641 .5674 m .74787 .53616 L .72709 .55237 L .383 0 .024 r
F
.72709 .55237 m .71742 .58483 L .73641 .5674 L .382 0 .037 r
F
0 g
.71685 .558 m
.726 .55604 L
s
.80807 .46268 m
.81131 .45868 L
s
.81131 .45868 m
.81177 .45811 L
s
.62636 .55365 m
.6384 .55676 L
s
.68049 .59963 m .66762 .57132 L .66286 .59987 L .592 .139 .147 r
F
.66762 .57132 m .6463 .56541 L .66286 .59987 L .624 .175 .165 r
F
0 g
.726 .55604 m
.72872 .55501 L
s
.75526 .54685 m .76943 .51459 L .74787 .53616 L .262 0 0 r
F
0 g
.75183 .52811 m
.76258 .52145 L
s
.73641 .5674 m .75526 .54685 L .74787 .53616 L .26 0 0 r
F
.36677 .3425 m .3531 .35911 L .3334 .34009 L .372 .77 .986 r
F
0 g
.56399 .49104 m
.56542 .49215 L
s
.56089 .48864 m
.56399 .49104 L
s
.19451 .33625 m .18944 .3076 L .22136 .32264 L .421 .802 .991 r
F
.16167 .32054 m .18944 .3076 L .19451 .33625 L .426 .805 .992 r
F
.89691 .34516 m .876 .38331 L .86977 .37233 L .54 .916 .941 r
F
0 g
.89289 .35249 m
.89534 .37055 L
s
.85495 .40181 m .86977 .37233 L .876 .38331 L .543 .917 .941 r
F
0 g
.86001 .40466 m
.86459 .39334 L
s
.72872 .55501 m
.74281 .54996 L
s
.6463 .56541 m .64664 .59888 L .66286 .59987 L .624 .157 .125 r
F
.63025 .57089 m .64664 .59888 L .6463 .56541 L .655 .184 .122 r
F
0 g
.1342 .33333 m
.16167 .32054 L
s
.1342 .33333 m
.1342 .33333 L
s
.25165 .33725 m .24862 .31044 L .27832 .32725 L .416 .795 .992 r
F
.25165 .33725 m .22136 .32264 L .24862 .31044 L .424 .8 .993 r
F
.30622 .34494 m .30605 .3189 L .3334 .34009 L .402 .782 .991 r
F
.30622 .34494 m .27832 .32725 L .30605 .3189 L .412 .791 .992 r
F
0 g
.557 .48541 m
.56089 .48864 L
s
.77685 .51886 m .7916 .4866 L .76943 .51459 L .077 0 0 r
F
0 g
.77895 .49769 m
.79077 .48683 L
s
.75526 .54685 m .77685 .51886 L .76943 .51459 L .073 0 0 r
F
0 g
.76258 .52145 m
.76761 .51874 L
s
.58068 .51915 m .59014 .52057 L .56824 .49622 L .449 0 0 r
F
0 g
.58971 .51969 m
.59117 .52034 L
s
.84519 .40548 m .85495 .40181 L .83132 .43297 L .45 .883 .87 r
F
.84519 .40548 m .83132 .43297 L .8224 .44041 L .409 .856 .783 r
F
0 g
.83633 .42868 m
.83795 .42422 L
s
.68391 .58342 m
.69292 .58282 L
s
.67299 .58312 m
.68391 .58342 L
s
.52025 .4488 m .51007 .43203 L .5357 .46144 L 0 .485 .795 r
F
0 g
.52444 .45223 m
.52871 .45682 L
s
.49332 .42571 m .51007 .43203 L .52025 .4488 L .047 .579 .859 r
F
0 g
.66573 .58263 m
.67299 .58312 L
s
.58888 .51916 m
.58971 .51969 L
s
.69292 .58282 m
.70193 .58204 L
s
.65376 .58092 m
.66573 .58263 L
s
.83795 .42422 m
.8435 .40883 L
s
.46552 .4054 m .47927 .39987 L .49332 .42571 L .18 .659 .93 r
F
0 g
.49088 .42393 m
.49203 .42564 L
s
.45257 .38413 m .47927 .39987 L .46552 .4054 L .182 .657 .932 r
F
.43755 .38818 m .45257 .38413 L .46552 .4054 L .254 .701 .958 r
F
0 g
.44402 .39217 m
.45316 .40943 L
s
.59014 .52057 m .60082 .54761 L .61094 .54547 L .559 0 0 r
F
.58068 .51915 m .60082 .54761 L .59014 .52057 L .546 0 0 r
F
0 g
.57615 .51081 m
.58888 .51916 L
s
.49332 .42571 m .47927 .39987 L .51007 .43203 L .077 .587 .887 r
F
0 g
.48689 .4139 m
.49088 .42393 L
s
.43755 .38818 m .42817 .36278 L .45257 .38413 L .262 .7 .962 r
F
0 g
.44258 .38683 m
.44402 .39217 L
s
.42817 .36278 m .43755 .38818 L .40422 .36961 L .314 .733 .975 r
F
0 g
.76761 .51874 m
.77124 .51563 L
s
.70193 .58204 m
.71467 .57937 L
s
.79077 .48683 m
.79236 .48549 L
s
.74281 .54996 m
.75334 .54407 L
s
.64644 .5795 m
.65376 .58092 L
s
.84519 .40548 m .86977 .37233 L .85495 .40181 L .503 .909 .884 r
F
0 g
.86459 .39334 m
.8658 .38022 L
s
.61944 .57552 m .63025 .57089 L .61094 .54547 L .623 .105 0 r
F
.61094 .54547 m .60082 .54761 L .61944 .57552 L .632 .093 0 r
F
0 g
.61298 .54815 m
.62636 .55365 L
s
.42817 .36278 m .40422 .36961 L .39748 .33622 L .326 .731 .98 r
F
.40422 .36961 m .36677 .3425 L .39748 .33622 L .365 .758 .987 r
F
0 g
.37523 .34863 m
.37546 .36781 L
s
.7994 .4822 m .8009 .47675 L .79134 .49912 L 0 0 0 r
F
.80232 .47161 m .8009 .47675 L .81078 .45976 L F
.81209 .45554 m .8224 .44041 L .81474 .45336 L .357 .814 .705 r
F
.81209 .45554 m .81474 .45336 L .8009 .47675 L F
.81209 .45554 m .8009 .47675 L .7994 .4822 L 0 0 0 r
F
.80955 .45764 m .81209 .45554 L .80266 .47346 L .473 .871 .696 r
F
.82116 .43099 m .81209 .45554 L .80955 .45764 L F
.7916 .4866 m .8009 .47675 L .80232 .47161 L 0 0 0 r
F
.77685 .51886 m .8009 .47675 L .7916 .4866 L F
0 g
.79236 .48549 m
.79298 .48467 L
s
.61153 .54756 m
.61298 .54815 L
s
.79298 .48467 m
.80147 .47468 L
s
.80147 .47468 m
.80282 .47309 L
s
.71467 .57937 m
.71934 .57838 L
s
.60875 .54593 m
.6115 .54754 L
s
.6115 .54754 m
.61153 .54756 L
s
.77124 .51563 m
.78231 .50691 L
s
.81852 .44861 m
.82332 .43964 L
s
.81726 .45096 m
.81852 .44861 L
s
.81666 .45209 m
.81726 .45096 L
s
.59794 .54032 m
.60875 .54593 L
s
.91611 .29887 m .92804 .32265 L .89691 .34516 L .631 .905 .984 r
F
0 g
.92804 .32265 m
.92804 .32265 L
s
.92804 .32265 m
.93974 .34596 L
s
.63428 .59782 m .64664 .59888 L .63025 .57089 L .652 .173 .102 r
F
.63428 .59782 m .63025 .57089 L .61944 .57552 L .677 .184 .07 r
F
0 g
.63295 .57551 m
.64644 .5795 L
s
.69858 .59504 m .7085 .62334 L .71742 .58483 L .537 .138 .233 r
F
.69858 .59504 m .69136 .62977 L .7085 .62334 L .541 .136 .222 r
F
.69858 .59504 m .68049 .59963 L .69136 .62977 L .593 .194 .257 r
F
.68049 .59963 m .67534 .63223 L .69136 .62977 L .598 .189 .237 r
F
.67534 .63223 m .68049 .59963 L .66286 .59987 L .638 .232 .258 r
F
0 g
.75334 .54407 m
.75745 .54187 L
s
.73641 .5674 m .71742 .58483 L .72727 .60809 L .46 .048 .171 r
F
0 g
.71934 .57838 m
.72796 .57515 L
s
.7085 .62334 m .72727 .60809 L .71742 .58483 L .458 .05 .176 r
F
0 g
.63084 .57484 m
.63295 .57551 L
s
.3334 .34009 m .33547 .3128 L .36677 .3425 L .397 .773 .992 r
F
.3334 .34009 m .30605 .3189 L .33547 .3128 L .409 .785 .993 r
F
0 g
.16167 .32054 m
.16167 .32054 L
s
.62608 .57268 m
.63084 .57484 L
s
.74365 .57792 m .75526 .54685 L .73641 .5674 L .351 0 .064 r
F
.72727 .60809 m .74365 .57792 L .73641 .5674 L .348 0 .093 r
F
0 g
.72796 .57515 m
.73519 .57284 L
s
.57033 .50537 m
.57615 .51081 L
s
.21751 .29452 m .24862 .31044 L .22136 .32264 L .425 .8 .993 r
F
.22136 .32264 m .18944 .3076 L .21751 .29452 L .431 .805 .993 r
F
0 g
.16167 .32054 m
.18944 .3076 L
s
.6603 .63263 m .67534 .63223 L .66286 .59987 L .643 .222 .228 r
F
.6603 .63263 m .66286 .59987 L .64664 .59888 L .674 .252 .237 r
F
0 g
.67955 .60558 m
.68262 .60553 L
s
.66464 .60449 m
.67955 .60558 L
s
.68262 .60553 m
.69655 .60481 L
s
.27832 .32725 m .27656 .29918 L .30605 .3189 L .418 .791 .994 r
F
.27832 .32725 m .24862 .31044 L .27656 .29918 L .428 .798 .994 r
F
0 g
.69655 .60481 m
.70172 .60401 L
s
.73519 .57284 m
.73882 .57091 L
s
.61768 .56929 m
.62608 .57268 L
s
.66252 .60423 m
.66464 .60449 L
s
.91611 .29887 m .89691 .34516 L .88658 .32974 L .622 .929 .969 r
F
.86977 .37233 m .88658 .32974 L .89691 .34516 L .621 .929 .969 r
F
0 g
.89289 .35249 m
.8933 .34878 L
s
.70172 .60401 m
.71347 .60191 L
s
.75526 .54685 m .75745 .56307 L .77685 .51886 L .207 0 0 r
F
0 g
.75745 .54187 m
.76223 .53781 L
s
.74365 .57792 m .75745 .56307 L .75526 .54685 L .213 0 0 r
F
0 g
.89079 .33603 m
.8933 .34878 L
s
.73882 .57091 m
.74793 .56647 L
s
.71347 .60191 m
.72329 .5987 L
s
.64538 .62844 m .6603 .63263 L .64664 .59888 L .676 .236 .198 r
F
0 g
.64758 .60121 m
.66252 .60423 L
s
.63428 .59782 m .64538 .62844 L .64664 .59888 L .699 .259 .201 r
F
0 g
.64655 .60094 m
.64758 .60121 L
s
.59283 .54707 m .60082 .54761 L .58068 .51915 L .54 0 0 r
F
0 g
.59316 .53679 m
.59794 .54032 L
s
.57308 .51387 m .59283 .54707 L .58068 .51915 L .496 0 0 r
F
0 g
.5649 .49997 m
.57033 .50537 L
s
.36567 .30703 m .39748 .33622 L .36677 .3425 L .382 .757 .991 r
F
0 g
.37523 .34863 m
.37651 .34051 L
s
.36567 .30703 m .36677 .3425 L .33547 .3128 L .405 .777 .993 r
F
0 g
.63803 .59814 m
.64655 .60094 L
s
.58603 .53145 m
.59316 .53679 L
s
.54089 .46749 m
.5448 .47271 L
s
.61944 .57552 m .60082 .54761 L .61074 .57833 L .633 .096 0 r
F
0 g
.61376 .56701 m
.61768 .56929 L
s
.59283 .54707 m .61074 .57833 L .60082 .54761 L .625 .065 0 r
F
.85856 .36952 m .88658 .32974 L .86977 .37233 L .602 .944 .931 r
F
.84519 .40548 m .85856 .36952 L .86977 .37233 L .581 .937 .931 r
F
0 g
.8658 .38022 m
.86679 .37634 L
s
.52853 .44595 m .5357 .46144 L .51007 .43203 L 0 .529 .842 r
F
0 g
.51566 .43845 m
.52444 .45223 L
s
.45257 .38413 m .45229 .36497 L .47927 .39987 L .261 .684 .965 r
F
.45257 .38413 m .42817 .36278 L .45229 .36497 L .292 .712 .972 r
F
0 g
.44027 .37337 m
.44258 .38683 L
s
.86679 .37634 m
.86703 .37164 L
s
.60555 .56226 m
.61376 .56701 L
s
.72329 .5987 m
.72984 .59662 L
s
.62937 .61454 m .63428 .59782 L .61944 .57552 L .688 .21 .112 r
F
.62937 .61454 m .61944 .57552 L .61074 .57833 L .702 .193 .021 r
F
.83363 .41393 m .85856 .36952 L .84519 .40548 L .555 .933 .863 r
F
.83363 .41393 m .84519 .40548 L .8224 .44041 L .486 .901 .845 r
F
0 g
.8435 .40883 m
.84449 .40654 L
s
.82332 .43964 m
.82414 .43774 L
s
.63434 .59699 m
.63803 .59814 L
s
.63414 .59693 m
.63434 .59699 L
s
.84449 .40654 m
.84471 .40583 L
s
.52853 .44595 m .51007 .43203 L .50356 .41089 L .053 .549 .886 r
F
0 g
.5138 .43484 m
.51566 .43845 L
s
.50356 .41089 m .51007 .43203 L .47927 .39987 L .151 .627 .925 r
F
0 g
.4816 .40231 m
.48689 .4139 L
s
.83363 .41393 m .8224 .44041 L .81209 .45554 L .444 .867 .743 r
F
0 g
.82414 .43774 m
.82551 .43309 L
s
.77685 .51886 m .77183 .54397 L .79134 .49912 L 0 0 0 r
F
.8009 .47675 m .77685 .51886 L .79134 .49912 L F
0 g
.78231 .50691 m
.78701 .50107 L
s
.77685 .51886 m .75745 .56307 L .77183 .54397 L .031 0 0 r
F
0 g
.76223 .53781 m
.77152 .53099 L
s
.63349 .59663 m
.63414 .59693 L
s
.78701 .50107 m
.79476 .49111 L
s
.8123 .45537 m
.81256 .45486 L
s
.80969 .46058 m
.8123 .45537 L
s
.67589 .62875 m
.69091 .62852 L
s
.67394 .62861 m
.67589 .62875 L
s
.66074 .62703 m
.67394 .62861 L
s
.62378 .59256 m
.63349 .59663 L
s
.69091 .62852 m
.69163 .62848 L
s
.62937 .61454 m .64538 .62844 L .63428 .59782 L .702 .228 .123 r
F
.7085 .62334 m .69136 .62977 L .70095 .6651 L .583 .222 .322 r
F
0 g
.69163 .62848 m
.69699 .62766 L
s
.69136 .62977 m .68418 .66246 L .70095 .6651 L .593 .22 .305 r
F
.67534 .63223 m .68418 .66246 L .69136 .62977 L .637 .274 .338 r
F
.67261 .66871 m .68418 .66246 L .67534 .63223 L .649 .268 .31 r
F
.67261 .66871 m .67534 .63223 L .6603 .63263 L .678 .298 .319 r
F
0 g
.18944 .3076 m
.18944 .3076 L
s
.74365 .57792 m .74198 .60344 L .75745 .56307 L .307 0 .104 r
F
0 g
.74793 .56647 m
.75704 .56004 L
s
.72727 .60809 m .74198 .60344 L .74365 .57792 L .368 0 .119 r
F
0 g
.72984 .59662 m
.73495 .59394 L
s
.6578 .62644 m
.66074 .62703 L
s
.77152 .53099 m
.77517 .52727 L
s
.7085 .62334 m .71543 .65734 L .72727 .60809 L .52 .161 .298 r
F
.7085 .62334 m .70095 .6651 L .71543 .65734 L .53 .163 .288 r
F
0 g
.69699 .62766 m
.708 .62612 L
s
.30605 .3189 m .30527 .28879 L .33547 .3128 L .42 .785 .995 r
F
.30605 .3189 m .27656 .29918 L .30527 .28879 L .433 .795 .996 r
F
0 g
.51213 .42902 m
.5138 .43484 L
s
.24862 .31044 m .2459 .28129 L .27656 .29918 L .43 .798 .995 r
F
.21751 .29452 m .2459 .28129 L .24862 .31044 L .438 .804 .995 r
F
0 g
.18944 .3076 m
.21751 .29452 L
s
.6456 .62332 m
.6578 .62644 L
s
.42817 .36278 m .42466 .33312 L .45229 .36497 L .321 .713 .981 r
F
0 g
.43979 .36383 m
.44027 .37337 L
s
.42817 .36278 m .39748 .33622 L .42466 .33312 L .36 .744 .988 r
F
0 g
.708 .62612 m
.70902 .62587 L
s
.47927 .39987 m .47763 .37423 L .50356 .41089 L .204 .636 .95 r
F
0 g
.48115 .40073 m
.4816 .40231 L
s
.45229 .36497 m .47763 .37423 L .47927 .39987 L .256 .682 .964 r
F
0 g
.73495 .59394 m
.74286 .58997 L
s
.65877 .66788 m .67261 .66871 L .6603 .63263 L .685 .289 .291 r
F
.64538 .62844 m .65877 .66788 L .6603 .63263 L .712 .315 .295 r
F
.74198 .60344 m .72727 .60809 L .72759 .64337 L .411 .054 .245 r
F
.72727 .60809 m .71543 .65734 L .72759 .64337 L .455 .085 .247 r
F
0 g
.70902 .62587 m
.72401 .62162 L
s
.75704 .56004 m
.75959 .55818 L
s
.64324 .62255 m
.6456 .62332 L
s
.72401 .62162 m
.72738 .62027 L
s
.75519 .58898 m .77183 .54397 L .75745 .56307 L .159 0 0 r
F
0 g
.75959 .55818 m
.76608 .55161 L
s
.75745 .56307 m .74198 .60344 L .75519 .58898 L .209 0 .001 r
F
0 g
.74286 .58997 m
.74855 .5863 L
s
.67379 .65292 m
.6815 .65331 L
s
.6815 .65331 m
.6862 .65328 L
s
.6862 .65328 m
.69745 .6522 L
s
.64538 .62844 m .64266 .65548 L .65877 .66788 L .72 .295 .239 r
F
.64538 .62844 m .62937 .61454 L .64266 .65548 L .742 .32 .247 r
F
0 g
.63368 .61829 m
.64324 .62255 L
s
.66692 .65205 m
.67379 .65292 L
s
.60348 .57566 m .61074 .57833 L .59283 .54707 L .621 .06 0 r
F
.58808 .54617 m .60348 .57566 L .59283 .54707 L .562 0 0 r
F
0 g
.59791 .55593 m
.60555 .56226 L
s
.57308 .51387 m .58808 .54617 L .59283 .54707 L .478 0 0 r
F
0 g
.57879 .52347 m
.58603 .53145 L
s
.69745 .6522 m
.70344 .65138 L
s
.59533 .55378 m
.59791 .55593 L
s
.65952 .65076 m
.66692 .65205 L
s
.90394 .2746 m .91611 .29887 L .88658 .32974 L .68 .92 .973 r
F
0 g
.91611 .29887 m
.91611 .29887 L
s
.88905 .32716 m
.89079 .33603 L
s
.91611 .29887 m
.92804 .32265 L
s
.61839 .61112 m .62937 .61454 L .61074 .57833 L .697 .184 .006 r
F
.60348 .57566 m .61839 .61112 L .61074 .57833 L .703 .172 0 r
F
0 g
.61523 .58705 m
.62378 .59256 L
s
.57308 .51387 m .56017 .49353 L .56644 .5028 L 0 .322 .67 r
F
.56017 .49353 m .55146 .47979 L .56644 .5028 L F
.57196 .51098 m .56644 .5028 L .56189 .48818 L 0 .307 .731 r
F
.57308 .51387 m .56644 .5028 L .58808 .54617 L .399 0 0 r
F
.57788 .52441 m .56644 .5028 L .57196 .51098 L 0 .307 .731 r
F
0 g
.57613 .52043 m
.57879 .52347 L
s
.70344 .65138 m
.7137 .64886 L
s
.61233 .58514 m
.61523 .58705 L
s
.55569 .48629 m
.55812 .49029 L
s
.65234 .64895 m
.65952 .65076 L
s
.74855 .5863 m
.75593 .58044 L
s
.54602 .46514 m .55146 .47979 L .52853 .44595 L 0 .474 .824 r
F
0 g
.53515 .45572 m
.54089 .46749 L
s
.58969 .54647 m
.59533 .55378 L
s
.39491 .30078 m .42466 .33312 L .39748 .33622 L .378 .742 .991 r
F
.39491 .30078 m .39748 .33622 L .36567 .30703 L .41 .766 .995 r
F
0 g
.37651 .34051 m
.38291 .32284 L
s
.33547 .3128 m .33462 .27898 L .36567 .30703 L .42 .776 .996 r
F
.33547 .3128 m .30527 .28879 L .33462 .27898 L .438 .791 .997 r
F
.64266 .65548 m .62937 .61454 L .63455 .65521 L .747 .285 .153 r
F
.61839 .61112 m .63455 .65521 L .62937 .61454 L .764 .294 .127 r
F
0 g
.63011 .61683 m
.63368 .61829 L
s
.62963 .61655 m
.63011 .61683 L
s
.64361 .64597 m
.65234 .64895 L
s
.7137 .64886 m
.71774 .64773 L
s
.62447 .61302 m
.62963 .61655 L
s
.74198 .60344 m .74086 .63497 L .75519 .58898 L .286 0 .156 r
F
.74198 .60344 m .72759 .64337 L .74086 .63497 L .337 0 .165 r
F
0 g
.72738 .62027 m
.73785 .61491 L
s
.87389 .31804 m .90394 .2746 L .88658 .32974 L .68 .95 .956 r
F
0 g
.88802 .32517 m
.88905 .32716 L
s
.85856 .36952 m .87389 .31804 L .88658 .32974 L .672 .948 .957 r
F
0 g
.86703 .37164 m
.86727 .35715 L
s
.21751 .29452 m
.21751 .29452 L
s
.54602 .46514 m .52853 .44595 L .52441 .42692 L .019 .503 .874 r
F
0 g
.5334 .4513 m
.53515 .45572 L
s
.52853 .44595 m .50356 .41089 L .52441 .42692 L .111 .582 .914 r
F
0 g
.50921 .41882 m
.51213 .42902 L
s
.71774 .64773 m
.727 .64405 L
s
.27461 .26792 m .30527 .28879 L .27656 .29918 L .436 .795 .996 r
F
.27656 .29918 m .2459 .28129 L .27461 .26792 L .447 .803 .997 r
F
0 g
.21751 .29452 m
.2459 .28129 L
s
.83363 .41393 m .81209 .45554 L .82116 .43099 L .52 .915 .812 r
F
0 g
.81256 .45486 m
.81283 .45413 L
s
.82551 .43309 m
.82985 .42123 L
s
.81209 .45554 m .79134 .49912 L .80266 .47346 L .399 .827 .666 r
F
0 g
.79476 .49111 m
.79532 .4901 L
s
.81283 .45413 m
.81325 .45242 L
s
.73785 .61491 m
.74166 .61253 L
s
.78341 .52294 m .80266 .47346 L .79134 .49912 L 0 0 0 r
F
.78341 .52294 m .79134 .49912 L .77183 .54397 L F
0 g
.77517 .52727 m
.78359 .51694 L
s
.45142 .33591 m .47763 .37423 L .45229 .36497 L .307 .687 .977 r
F
.45142 .33591 m .45229 .36497 L .42466 .33312 L .349 .724 .986 r
F
0 g
.43979 .36383 m
.44152 .35255 L
s
.84181 .37828 m .85856 .36952 L .83363 .41393 L .621 .956 .909 r
F
0 g
.84471 .40583 m
.84781 .38868 L
s
.84181 .37828 m .83363 .41393 L .82116 .43099 L .626 .961 .851 r
F
0 g
.82985 .42123 m
.83066 .41799 L
s
.84181 .37828 m .87389 .31804 L .85856 .36952 L .664 .969 .912 r
F
0 g
.84781 .38868 m
.8482 .37494 L
s
.86502 .34782 m
.86727 .35715 L
s
.63896 .64409 m
.64361 .64597 L
s
.78359 .51694 m
.78748 .51071 L
s
.76671 .57 m .78341 .52294 L .77183 .54397 L 0 0 0 r
F
.75519 .58898 m .76671 .57 L .77183 .54397 L .004 0 0 r
F
0 g
.76608 .55161 m
.77053 .54749 L
s
.75593 .58044 m
.75959 .57709 L
s
.77053 .54749 m
.77133 .54653 L
s
.63269 .64058 m
.63896 .64409 L
s
.50356 .41089 m .50157 .38711 L .52441 .42692 L .174 .596 .94 r
F
0 g
.50844 .41464 m
.50921 .41882 L
s
.47763 .37423 m .50157 .38711 L .50356 .41089 L .239 .653 .961 r
F
0 g
.48005 .37765 m
.48115 .40073 L
s
.60348 .57566 m .60728 .595 L .61839 .61112 L .649 .091 0 r
F
0 g
.61781 .60867 m
.62447 .61302 L
s
.58808 .54617 m .60728 .595 L .60348 .57566 L .65 .092 0 r
F
0 g
.60459 .57829 m
.61233 .58514 L
s
.58658 .54294 m
.58969 .54647 L
s
.60384 .57746 m
.60459 .57829 L
s
.61709 .60803 m
.61781 .60867 L
s
.75235 .61739 m .76671 .57 L .75519 .58898 L .119 0 .03 r
F
0 g
.75959 .57709 m
.76706 .56819 L
s
.74086 .63497 m .75235 .61739 L .75519 .58898 L .18 0 .052 r
F
0 g
.74166 .61253 m
.74973 .60649 L
s
.61839 .61112 m .62482 .64656 L .63455 .65521 L .749 .25 .039 r
F
0 g
.62803 .63742 m
.63269 .64058 L
s
.61839 .61112 m .60728 .595 L .62482 .64656 L .757 .248 0 r
F
0 g
.61378 .60443 m
.61709 .60803 L
s
.74973 .60649 m
.75381 .60272 L
s
.36567 .30703 m .36414 .26931 L .39491 .30078 L .422 .764 .996 r
F
.36567 .30703 m .33462 .27898 L .36414 .26931 L .446 .784 .998 r
F
.58326 .53391 m .57788 .52441 L .57731 .51725 L 0 .271 .73 r
F
.57225 .51445 m .57788 .52441 L .58808 .54617 L 0 .275 .672 r
F
.59543 .56795 m .58808 .54617 L .57788 .52441 L .474 0 0 r
F
.59543 .56795 m .57788 .52441 L .58326 .53391 L 0 .271 .73 r
F
.56644 .5028 m .57788 .52441 L .57225 .51445 L 0 .275 .672 r
F
0 g
.58651 .54283 m
.58658 .54294 L
s
.54602 .46514 m .56189 .48818 L .56644 .5028 L 0 .42 .808 r
F
0 g
.55141 .47509 m
.55569 .48629 L
s
.56878 .50723 m
.56971 .50937 L
s
.59543 .56795 m .60728 .595 L .58808 .54617 L .47 0 0 r
F
0 g
.59768 .57059 m
.60384 .57746 L
s
.62233 .63284 m
.62803 .63742 L
s
.30364 .25439 m .33462 .27898 L .30527 .28879 L .443 .79 .998 r
F
.27461 .26792 m .30364 .25439 L .30527 .28879 L .457 .801 .998 r
F
.45142 .33591 m .42466 .33312 L .42355 .29751 L .376 .722 .989 r
F
0 g
.44152 .35255 m
.4464 .33539 L
s
.42355 .29751 m .42466 .33312 L .39491 .30078 L .412 .752 .994 r
F
0 g
.38291 .32284 m
.3951 .30344 L
s
.2459 .28129 m
.27461 .26792 L
s
.2459 .28129 m
.2459 .28129 L
s
.54323 .44758 m .56189 .48818 L .54602 .46514 L 0 .455 .86 r
F
0 g
.55035 .47142 m
.55141 .47509 L
s
.52441 .42692 m .54323 .44758 L .54602 .46514 L .08 .537 .903 r
F
0 g
.53102 .43861 m
.5334 .4513 L
s
.50157 .38711 m .47763 .37423 L .4778 .34508 L .292 .657 .971 r
F
0 g
.48005 .37765 m
.48013 .37557 L
s
.4778 .34508 m .47763 .37423 L .45142 .33591 L .337 .7 .982 r
F
.81517 .43371 m .84181 .37828 L .82116 .43099 L .597 .948 .818 r
F
0 g
.83066 .41799 m
.83198 .40337 L
s
.80266 .47346 m .81517 .43371 L .82116 .43099 L .614 .954 .817 r
F
0 g
.81325 .45242 m
.81717 .44016 L
s
.80068 .47855 m
.80327 .4723 L
s
.81717 .44016 m
.81813 .43236 L
s
.80327 .4723 m
.80355 .47141 L
s
.80266 .47346 m .79232 .49656 L .81517 .43371 L .545 .9 .686 r
F
0 g
.80355 .47141 m
.8037 .47015 L
s
.78341 .52294 m .79232 .49656 L .80239 .47415 L .484 .866 .661 r
F
.78341 .52294 m .77537 .54956 L .79232 .49656 L 0 0 0 r
F
.78341 .52294 m .77931 .5345 L .77537 .54956 L F
.77931 .5345 m .76671 .57 L .77537 .54956 L F
.77182 .5631 m .77537 .54956 L .76833 .57585 L F
.77926 .53595 m .79232 .49656 L .77537 .54956 L .503 .853 .594 r
F
.77926 .53595 m .77537 .54956 L .77182 .5631 L 0 0 0 r
F
.77926 .53595 m .78428 .51445 L .79232 .49656 L .69 .963 .73 r
F
.78428 .51445 m .78634 .50564 L .79232 .49656 L F
.80239 .47415 m .79232 .49656 L .80266 .47346 L .484 .866 .661 r
F
0 g
.76706 .56819 m
.76765 .56735 L
s
.79459 .49151 m
.79461 .49144 L
s
.79101 .50341 m
.79459 .49151 L
s
.77133 .54653 m
.77899 .53541 L
s
.77899 .53541 m
.77976 .53381 L
s
.76671 .57 m .76202 .59378 L .77537 .54956 L 0 0 0 r
F
0 g
.76765 .56735 m
.76904 .56449 L
s
.75235 .61739 m .76202 .59378 L .76671 .57 L 0 0 0 r
F
0 g
.75381 .60272 m
.75823 .59796 L
s
.61342 .62311 m .62482 .64656 L .60728 .595 L .678 .132 0 r
F
0 g
.61903 .62955 m
.62233 .63284 L
s
.61342 .62311 m .60728 .595 L .59543 .56795 L .67 .127 0 r
F
0 g
.60878 .59939 m
.61378 .60443 L
s
.59342 .56198 m
.59768 .57059 L
s
.608 .59829 m
.60878 .59939 L
s
.52346 .40528 m .54323 .44758 L .52441 .42692 L .153 .556 .928 r
F
0 g
.53051 .43362 m
.53102 .43861 L
s
.50157 .38711 m .52346 .40528 L .52441 .42692 L .218 .616 .952 r
F
0 g
.50844 .41464 m
.50869 .39952 L
s
.87389 .31804 m .89152 .24983 L .90394 .2746 L .727 .932 .957 r
F
0 g
.90394 .2746 m
.90394 .2746 L
s
.88312 .3047 m
.88802 .32517 L
s
.90394 .2746 m
.91611 .29887 L
s
.75823 .59796 m
.76238 .59197 L
s
.56189 .48818 m .57731 .51725 L .57788 .52441 L 0 .354 .778 r
F
0 g
.56566 .49671 m
.56878 .50723 L
s
.86062 .30627 m .89152 .24983 L .87389 .31804 L .733 .964 .939 r
F
0 g
.87926 .29724 m
.88312 .3047 L
s
.84181 .37828 m .86062 .30627 L .87389 .31804 L .73 .963 .94 r
F
0 g
.84785 .36694 m
.8482 .37494 L
s
.86457 .33553 m
.86502 .34782 L
s
.3951 .30344 m
.39578 .30173 L
s
.59305 .56128 m
.59342 .56198 L
s
.82894 .3742 m .86062 .30627 L .84181 .37828 L .723 .984 .9 r
F
0 g
.84621 .36141 m
.84785 .36694 L
s
.85853 .31428 m
.86457 .33553 L
s
.82894 .3742 m .84181 .37828 L .81517 .43371 L .735 .986 .897 r
F
0 g
.81813 .43236 m
.81911 .4255 L
s
.83198 .40337 m
.83297 .39666 L
s
.39491 .30078 m .39368 .2604 L .42355 .29751 L .426 .749 .994 r
F
0 g
.39578 .30173 m
.39653 .30059 L
s
.39491 .30078 m .36414 .26931 L .39368 .2604 L .457 .774 .997 r
F
.36414 .26931 m .33462 .27898 L .333 .24071 L .451 .782 .998 r
F
.33462 .27898 m .30364 .25439 L .333 .24071 L .47 .797 .999 r
F
.56011 .47189 m .57731 .51725 L .56189 .48818 L 0 .407 .844 r
F
0 g
.56508 .49419 m
.56566 .49671 L
s
.56011 .47189 m .56189 .48818 L .54323 .44758 L .056 .493 .891 r
F
0 g
.54913 .46041 m
.55035 .47142 L
s
.4778 .34508 m .45142 .33591 L .45267 .30009 L .373 .698 .983 r
F
.45142 .33591 m .42355 .29751 L .45267 .30009 L .411 .734 .991 r
F
0 g
.4464 .33539 m
.44822 .3315 L
s
.27461 .26792 m
.30364 .25439 L
s
.27461 .26792 m
.27461 .26792 L
s
.52346 .40528 m .50157 .38711 L .50265 .36046 L .277 .623 .96 r
F
0 g
.50869 .39952 m
.50931 .39354 L
s
.50157 .38711 m .4778 .34508 L .50265 .36046 L .324 .67 .976 r
F
0 g
.48013 .37557 m
.4847 .35728 L
s
.83 .37454 m
.83297 .39666 L
s
.44822 .3315 m
.45175 .32632 L
s
.59169 .5585 m
.59175 .55881 L
s
.77537 .54956 m .76202 .59378 L .76833 .57585 L 0 0 0 r
F
0 g
.76238 .59197 m
.7628 .59122 L
s
.76904 .56449 m
.77307 .55717 L
s
.7628 .59122 m
.76442 .58695 L
s
.59543 .56795 m .61093 .61544 L .61342 .62311 L 0 .079 .546 r
F
0 g
.61297 .62104 m
.61903 .62955 L
s
.59291 .55418 m .61093 .61544 L .59543 .56795 L 0 .145 .643 r
F
0 g
.60396 .59174 m
.608 .59829 L
s
.59291 .55418 m .59543 .56795 L .57731 .51725 L 0 .279 .738 r
F
0 g
.59127 .55632 m
.59169 .5585 L
s
.57829 .51998 m
.58031 .53043 L
s
.77307 .55717 m
.77395 .55487 L
s
.54323 .44758 m .54322 .42841 L .56011 .47189 L .136 .515 .914 r
F
0 g
.54901 .4559 m
.54913 .46041 L
s
.54323 .44758 m .52346 .40528 L .54322 .42841 L .197 .577 .94 r
F
0 g
.53051 .43362 m
.53142 .4223 L
s
.60127 .58585 m
.60233 .5891 L
s
.61251 .6203 m
.61297 .62104 L
s
.60992 .61235 m
.6125 .62029 L
s
.78406 .52237 m
.78551 .51709 L
s
.77395 .55487 m
.77483 .55182 L
s
.82894 .3742 m .81517 .43371 L .80355 .4495 L .719 .988 .831 r
F
0 g
.81833 .42003 m
.81911 .4255 L
s
.79232 .49656 m .80355 .4495 L .81517 .43371 L .712 .986 .83 r
F
0 g
.79461 .49144 m
.79537 .48818 L
s
.8037 .47015 m
.80691 .45643 L
s
.8068 .44508 m
.80691 .45643 L
s
.80355 .4495 m .79232 .49656 L .78634 .50564 L .671 .958 .731 r
F
0 g
.79531 .48405 m
.79537 .48818 L
s
.5752 .4977 m .59291 .55418 L .57731 .51725 L 0 .35 .818 r
F
0 g
.57815 .51924 m
.57829 .51998 L
s
.58936 .54578 m
.59127 .55632 L
s
.5752 .4977 m .57731 .51725 L .56011 .47189 L .045 .453 .878 r
F
0 g
.56466 .48389 m
.56508 .49419 L
s
.4778 .34508 m .48108 .31262 L .50265 .36046 L .369 .669 .974 r
F
0 g
.4847 .35728 m
.48696 .35075 L
s
.4778 .34508 m .45267 .30009 L .48108 .31262 L .405 .709 .985 r
F
0 g
.45175 .32632 m
.45912 .31163 L
s
.52559 .38172 m .54322 .42841 L .52346 .40528 L .264 .588 .947 r
F
0 g
.53142 .4223 m
.53243 .41578 L
s
.52346 .40528 m .50265 .36046 L .52559 .38172 L .308 .636 .965 r
F
0 g
.50931 .39354 m
.51266 .38203 L
s
.36414 .26931 m .36269 .22688 L .39368 .2604 L .462 .772 .997 r
F
.36414 .26931 m .333 .24071 L .36269 .22688 L .486 .791 .999 r
F
.42401 .25397 m .45267 .30009 L .42355 .29751 L .432 .729 .988 r
F
.42355 .29751 m .39368 .2604 L .42401 .25397 L .467 .76 .994 r
F
0 g
.39653 .30059 m
.41102 .28194 L
s
.30364 .25439 m
.333 .24071 L
s
.30364 .25439 m
.30364 .25439 L
s
.55936 .45338 m .5752 .4977 L .56011 .47189 L .111 .47 .893 r
F
0 g
.56466 .48389 m
.56485 .48 L
s
.55936 .45338 m .56011 .47189 L .54322 .42841 L .184 .54 .927 r
F
0 g
.54901 .4559 m
.55026 .44654 L
s
.77926 .53595 m .76833 .57585 L .77378 .55174 L .621 .913 .633 r
F
.77926 .53595 m .77378 .55174 L .78634 .50564 L .641 .921 .639 r
F
0 g
.7763 .54676 m
.77694 .54266 L
s
.78551 .51709 m
.78627 .50591 L
s
.59291 .55418 m .60792 .59794 L .61093 .61544 L 0 .234 .731 r
F
0 g
.60977 .61149 m
.60992 .61235 L
s
.59977 .57749 m
.60127 .58585 L
s
.60942 .60664 m
.60977 .61149 L
s
.86062 .30627 m .87883 .22454 L .89152 .24983 L .768 .939 .938 r
F
0 g
.89152 .24983 m
.89152 .24983 L
s
.87466 .28063 m
.87926 .29724 L
s
.89152 .24983 m
.90394 .2746 L
s
.41102 .28194 m
.42382 .27146 L
s
.59291 .55418 m .59041 .52837 L .60792 .59794 L 0 .308 .799 r
F
0 g
.59944 .57322 m
.59977 .57749 L
s
.5752 .4977 m .59041 .52837 L .59291 .55418 L .084 .435 .875 r
F
0 g
.58936 .54578 m
.58941 .54303 L
s
.57815 .51924 m
.5783 .50758 L
s
.54725 .41048 m .55936 .45338 L .54322 .42841 L .26 .555 .931 r
F
0 g
.55026 .44654 m
.5514 .44107 L
s
.54725 .41048 m .54322 .42841 L .52559 .38172 L .29 .6 .95 r
F
0 g
.53243 .41578 m
.53525 .4073 L
s
.50265 .36046 m .50655 .33188 L .52559 .38172 L .357 .638 .961 r
F
0 g
.51266 .38203 m
.51607 .3729 L
s
.50265 .36046 m .48108 .31262 L .50655 .33188 L .391 .678 .974 r
F
0 g
.48696 .35075 m
.49313 .33935 L
s
.84839 .2901 m .87883 .22454 L .86062 .30627 L .777 .965 .924 r
F
0 g
.86885 .26935 m
.87466 .28063 L
s
.82894 .3742 m .84839 .2901 L .86062 .30627 L .785 .966 .92 r
F
0 g
.85822 .31141 m
.85853 .31428 L
s
.844 .3419 m
.84621 .36141 L
s
.85256 .29562 m
.85822 .31141 L
s
.5752 .4977 m .57063 .45833 L .59041 .52837 L .139 .449 .881 r
F
0 g
.5783 .50758 m
.57876 .50488 L
s
.55936 .45338 m .57063 .45833 L .5752 .4977 L .246 .533 .92 r
F
0 g
.56485 .48 m
.56613 .47232 L
s
.79042 .47415 m .80764 .41122 L .80355 .4495 L .801 .999 .844 r
F
0 g
.80677 .44374 m
.8068 .44508 L
s
.80764 .41122 m .82894 .3742 L .80355 .4495 L .785 .989 .889 r
F
0 g
.80671 .44014 m
.80677 .44374 L
s
.81814 .40622 m
.81833 .42003 L
s
.80355 .4495 m .78634 .50564 L .79042 .47415 L .793 .999 .843 r
F
0 g
.79531 .48405 m
.7963 .47316 L
s
.79514 .46528 m
.7963 .47316 L
s
.82975 .37245 m
.83 .37454 L
s
.80523 .43377 m
.80671 .44014 L
s
.78634 .50564 m .77837 .52348 L .79042 .47415 L .795 .996 .782 r
F
.78634 .50564 m .77378 .55174 L .77837 .52348 L F
0 g
.78627 .50591 m
.78628 .50583 L
s
.77694 .54266 m
.77761 .53771 L
s
.78628 .50577 m
.78628 .50583 L
s
.45843 .25371 m .48108 .31262 L .45267 .30009 L .441 .704 .977 r
F
0 g
.45912 .31163 m
.46355 .30489 L
s
.42401 .25397 m .45843 .25371 L .45267 .30009 L .472 .74 .987 r
F
0 g
.77707 .53152 m
.77761 .53771 L
s
.39273 .21289 m .42401 .25397 L .39368 .2604 L .474 .757 .992 r
F
.39368 .2604 m .36269 .22688 L .39273 .21289 L .504 .781 .995 r
F
.54725 .41048 m .57063 .45833 L .55936 .45338 L .235 .542 .926 r
F
0 g
.5514 .44107 m
.55398 .43432 L
s
.82472 .34767 m .82894 .3742 L .80764 .41122 L .809 .984 .889 r
F
0 g
.81602 .39666 m
.81814 .40622 L
s
.82472 .34767 m .84839 .2901 L .82894 .3742 L .786 .979 .908 r
F
0 g
.8295 .37178 m
.82975 .37245 L
s
.83933 .32926 m
.844 .3419 L
s
.333 .24071 m
.36269 .22688 L
s
.333 .24071 m
.333 .24071 L
s
.82743 .36472 m
.8295 .37178 L
s
.60437 .56346 m .60792 .59794 L .59041 .52837 L .108 .403 .855 r
F
0 g
.59944 .57322 m
.60009 .56684 L
s
.58941 .54303 m
.59098 .53429 L
s
.42382 .27146 m
.43005 .2637 L
s
.59098 .53429 m
.59123 .53166 L
s
.54725 .41048 m .52559 .38172 L .53035 .35503 L .351 .606 .944 r
F
0 g
.53525 .4073 m
.53854 .39891 L
s
.53035 .35503 m .52559 .38172 L .50655 .33188 L .383 .648 .962 r
F
0 g
.51607 .3729 m
.51971 .36633 L
s
.60437 .56346 m .59041 .52837 L .59113 .50083 L .179 .423 .858 r
F
0 g
.59123 .53166 m
.59143 .53093 L
s
.60009 .56684 m
.6024 .5585 L
s
.59113 .50083 m .59041 .52837 L .57063 .45833 L .266 .512 .903 r
F
0 g
.57876 .50488 m
.58119 .49572 L
s
.56613 .47232 m
.57084 .46018 L
s
.50655 .33188 m .48108 .31262 L .48859 .2862 L .429 .675 .966 r
F
0 g
.49313 .33935 m
.50066 .32743 L
s
.45843 .25371 m .48859 .2862 L .48108 .31262 L .447 .707 .977 r
F
0 g
.46355 .30489 m
.47368 .29337 L
s
.79042 .47415 m .77837 .52348 L .77582 .50635 L .849 .988 .854 r
F
0 g
.78537 .49483 m
.78628 .50577 L
s
.54725 .41048 m .55409 .38906 L .57063 .45833 L .343 .57 .923 r
F
0 g
.55398 .43432 m
.55598 .42835 L
s
.53035 .35503 m .55409 .38906 L .54725 .41048 L .367 .614 .945 r
F
0 g
.53854 .39891 m
.54195 .39306 L
s
.79042 .44536 m .79042 .47415 L .77582 .50635 L .851 .987 .854 r
F
0 g
.78361 .48916 m
.78537 .49483 L
s
.80764 .41122 m .79042 .47415 L .79042 .44536 L .831 .98 .88 r
F
0 g
.79429 .46 m
.79514 .46528 L
s
.80387 .42498 m
.80523 .43377 L
s
.57084 .46018 m
.57099 .45961 L
s
.79042 .44895 m
.79429 .46 L
s
.59113 .50083 m .57063 .45833 L .57652 .43093 L .32 .521 .895 r
F
0 g
.57099 .45961 m
.57112 .45934 L
s
.58119 .49572 m
.5849 .48791 L
s
.55409 .38906 m .57652 .43093 L .57063 .45833 L .358 .577 .924 r
F
0 g
.55598 .42835 m
.56136 .4195 L
s
.42401 .25397 m .42311 .19873 L .45843 .25371 L .484 .735 .982 r
F
0 g
.43005 .2637 m
.44159 .25384 L
s
.39273 .21289 m .42311 .19873 L .42401 .25397 L .522 .767 .986 r
F
0 g
.36269 .22688 m
.39273 .21289 L
s
.36269 .22688 m
.36269 .22688 L
s
.59113 .50083 m .60837 .55034 L .60437 .56346 L .3 .488 .876 r
F
0 g
.6024 .5585 m
.603 .55698 L
s
.59143 .53093 m
.59534 .52077 L
s
.53035 .35503 m .50655 .33188 L .51121 .29637 L .42 .646 .951 r
F
0 g
.51971 .36633 m
.52829 .35302 L
s
.50655 .33188 m .48859 .2862 L .51121 .29637 L .45 .683 .964 r
F
0 g
.50066 .32743 m
.50347 .32406 L
s
.80764 .41122 m .79042 .44536 L .80724 .37539 L .838 .977 .879 r
F
0 g
.80259 .42123 m
.80387 .42498 L
s
.82472 .34767 m .80764 .41122 L .80724 .37539 L .816 .967 .902 r
F
0 g
.8137 .38868 m
.81602 .39666 L
s
.8237 .35146 m
.82743 .36472 L
s
.50347 .32406 m
.50823 .31909 L
s
.84839 .2901 m .86588 .19871 L .87883 .22454 L .797 .942 .922 r
F
0 g
.87883 .22454 m
.87883 .22454 L
s
.84965 .28738 m
.85256 .29562 L
s
.86425 .25594 m
.86885 .26935 L
s
.87883 .22454 m
.89152 .24983 L
s
.83096 .28387 m .84839 .2901 L .82472 .34767 L .81 .953 .911 r
F
0 g
.83646 .31911 m
.83933 .32926 L
s
.83096 .28387 m .86588 .19871 L .84839 .2901 L .802 .955 .916 r
F
0 g
.84911 .28635 m
.84965 .28738 L
s
.85745 .24272 m
.86425 .25594 L
s
.59626 .47657 m .60837 .55034 L .59113 .50083 L .34 .499 .871 r
F
0 g
.59534 .52077 m
.59707 .5179 L
s
.603 .55698 m
.60791 .54901 L
s
.59626 .47657 m .59113 .50083 L .57652 .43093 L .375 .55 .899 r
F
0 g
.5849 .48791 m
.58753 .48361 L
s
.57112 .45934 m
.57952 .44529 L
s
.53035 .35503 m .53736 .32239 L .55409 .38906 L .419 .616 .932 r
F
0 g
.54195 .39306 m
.54906 .38184 L
s
.53035 .35503 m .51121 .29637 L .53736 .32239 L .447 .657 .95 r
F
0 g
.52829 .35302 m
.52929 .35178 L
s
.50823 .31909 m
.5156 .30981 L
s
.52929 .35178 m
.53156 .34941 L
s
.79042 .44536 m .77582 .50635 L .77357 .4858 L .858 .962 .871 r
F
0 g
.78152 .48253 m
.78361 .48916 L
s
.83096 .28387 m .82472 .34767 L .80724 .37539 L .821 .964 .9 r
F
0 g
.82317 .35013 m
.8237 .35146 L
s
.80785 .37442 m
.8137 .38868 L
s
.59626 .47657 m .61316 .52804 L .60837 .55034 L .378 .521 .874 r
F
0 g
.60791 .54901 m
.60811 .54874 L
s
.59707 .5179 m
.60192 .51105 L
s
.60811 .54874 m
.60893 .54776 L
s
.78993 .44738 m
.79042 .44895 L
s
.48859 .2862 m .48591 .23869 L .51121 .29637 L .462 .678 .958 r
F
.45843 .25371 m .48591 .23869 L .48859 .2862 L .501 .708 .963 r
F
0 g
.47368 .29337 m
.48445 .28173 L
s
.48445 .28173 m
.48819 .27901 L
s
.56215 .35923 m .57652 .43093 L .55409 .38906 L .413 .584 .91 r
F
0 g
.56136 .4195 m
.56632 .41189 L
s
.53736 .32239 m .56215 .35923 L .55409 .38906 L .438 .624 .931 r
F
0 g
.54906 .38184 m
.55154 .3789 L
s
.53156 .34941 m
.54121 .33773 L
s
.79042 .44536 m .77357 .4858 L .78748 .41945 L .861 .956 .869 r
F
0 g
.78975 .44697 m
.78993 .44738 L
s
.77801 .47514 m
.78152 .48253 L
s
.78748 .41945 m .80724 .37539 L .79042 .44536 L .838 .951 .891 r
F
0 g
.79874 .41075 m
.80259 .42123 L
s
.45385 .18441 m .48591 .23869 L .45843 .25371 L .498 .714 .968 r
F
.45843 .25371 m .42311 .19873 L .45385 .18441 L .54 .745 .971 r
F
0 g
.44159 .25384 m
.4528 .24495 L
s
.39273 .21289 m
.42311 .19873 L
s
.39273 .21289 m
.39273 .21289 L
s
.55154 .3789 m
.55899 .37094 L
s
.4528 .24495 m
.4577 .24272 L
s
.58452 .40163 m .59626 .47657 L .57652 .43093 L .419 .557 .887 r
F
0 g
.57952 .44529 m
.58155 .44257 L
s
.58753 .48361 m
.59506 .47379 L
s
.58452 .40163 m .57652 .43093 L .56215 .35923 L .444 .599 .91 r
F
0 g
.56632 .41189 m
.57149 .40586 L
s
.80724 .37539 m .78748 .41945 L .80178 .34494 L .842 .943 .888 r
F
0 g
.79477 .40318 m
.79874 .41075 L
s
.80724 .37539 m .80178 .34494 L .83096 .28387 L .819 .938 .906 r
F
0 g
.80763 .37389 m
.80785 .37442 L
s
.81736 .33634 m
.82317 .35013 L
s
.82906 .30334 m
.83646 .31911 L
s
.80669 .3723 m
.80763 .37389 L
s
.51121 .29637 m .51864 .24386 L .53736 .32239 L .477 .652 .936 r
F
0 g
.5156 .30981 m
.51999 .3051 L
s
.51121 .29637 m .48591 .23869 L .51864 .24386 L .508 .69 .95 r
F
0 g
.48819 .27901 m
.49892 .26835 L
s
.59626 .47657 m .60364 .45009 L .61316 .52804 L .427 .534 .863 r
F
0 g
.60192 .51105 m
.60596 .50611 L
s
.58452 .40163 m .60364 .45009 L .59626 .47657 L .451 .575 .888 r
F
0 g
.59506 .47379 m
.59571 .47309 L
s
.58155 .44257 m
.58957 .43386 L
s
.78748 .41945 m .77357 .4858 L .77163 .46341 L .854 .928 .876 r
F
0 g
.77645 .47205 m
.77801 .47514 L
s
.7841 .43553 m
.78975 .44697 L
s
.59571 .47309 m
.59775 .47122 L
s
.60364 .45009 m .61874 .49037 L .61316 .52804 L .469 .559 .865 r
F
0 g
.60596 .50611 m
.60999 .50211 L
s
.77177 .46508 m
.77645 .47205 L
s
.59775 .47122 m
.60529 .46359 L
s
.55899 .37094 m
.56344 .36565 L
s
.82713 .29865 m
.82906 .30334 L
s
.49892 .26835 m
.517 .25542 L
s
.60999 .50211 m
.61797 .49556 L
s
.54759 .29407 m .56215 .35923 L .53736 .32239 L .472 .623 .917 r
F
0 g
.54121 .33773 m
.54494 .33365 L
s
.51864 .24386 m .54759 .29407 L .53736 .32239 L .488 .656 .935 r
F
0 g
.51999 .3051 m
.53092 .29537 L
s
.78748 .41945 m .77163 .46341 L .77991 .40269 L .855 .915 .872 r
F
0 g
.78254 .43314 m
.7841 .43553 L
s
.80178 .34494 m .78748 .41945 L .77991 .40269 L .832 .917 .893 r
F
0 g
.79166 .39767 m
.79477 .40318 L
s
.79914 .35868 m
.80669 .3723 L
s
.42311 .19873 m
.42311 .19873 L
s
.83096 .28387 m .85265 .17234 L .86588 .19871 L .808 .942 .914 r
F
0 g
.86588 .19871 m
.86588 .19871 L
s
.8525 .23134 m
.85745 .24272 L
s
.83912 .26398 m
.84911 .28635 L
s
.86588 .19871 m
.87883 .22454 L
s
.57258 .31685 m .58452 .40163 L .56215 .35923 L .476 .6 .897 r
F
0 g
.56344 .36565 m
.5648 .36426 L
s
.57149 .40586 m
.58147 .39586 L
s
.54759 .29407 m .57258 .31685 L .56215 .35923 L .498 .635 .914 r
F
0 g
.54494 .33365 m
.55451 .32504 L
s
.77991 .40269 m .77163 .46341 L .76058 .45931 L .837 .889 .878 r
F
0 g
.77683 .42526 m
.78254 .43314 L
s
.80178 .34494 m .8156 .2544 L .83096 .28387 L .822 .932 .904 r
F
0 g
.82538 .29555 m
.82549 .29574 L
s
.82549 .29574 m
.82713 .29865 L
s
.81108 .32547 m
.81736 .33634 L
s
.8156 .2544 m .85265 .17234 L .83096 .28387 L .804 .932 .917 r
F
0 g
.83599 .25804 m
.83912 .26398 L
s
.84428 .21535 m
.8525 .23134 L
s
.48591 .23869 m .48494 .16993 L .51864 .24386 L .512 .687 .946 r
F
.45385 .18441 m .48494 .16993 L .48591 .23869 L .554 .722 .952 r
F
0 g
.42311 .19873 m
.45385 .18441 L
s
.4577 .24272 m
.47888 .22679 L
s
.58452 .40163 m .59569 .37811 L .60364 .45009 L .483 .58 .876 r
F
0 g
.58957 .43386 m
.59516 .42861 L
s
.57258 .31685 m .59569 .37811 L .58452 .40163 L .494 .611 .897 r
F
0 g
.58147 .39586 m
.58347 .39417 L
s
.5648 .36426 m
.57769 .35315 L
s
.6133 .42737 m .61874 .49037 L .60364 .45009 L .491 .562 .856 r
F
0 g
.60529 .46359 m
.60789 .46142 L
s
.59569 .37811 m .6133 .42737 L .60364 .45009 L .504 .594 .877 r
F
0 g
.59516 .42861 m
.60077 .42414 L
s
.47888 .22679 m
.4857 .22368 L
s
.58347 .39417 m
.59067 .38867 L
s
.6133 .42737 m .62842 .48036 L .61874 .49037 L .53 .588 .857 r
F
0 g
.60789 .46142 m
.61574 .45556 L
s
.77991 .40269 m .78897 .33514 L .80178 .34494 L .832 .906 .889 r
F
0 g
.79747 .35632 m
.79914 .35868 L
s
.78518 .38877 m
.79166 .39767 L
s
.78897 .33514 m .8156 .2544 L .80178 .34494 L .81 .913 .91 r
F
0 g
.80601 .31721 m
.81108 .32547 L
s
.76866 .39402 m .77991 .40269 L .76058 .45931 L .835 .875 .872 r
F
0 g
.77347 .42157 m
.77683 .42526 L
s
.76218 .45463 m
.76969 .46269 L
s
.78897 .33514 m .77991 .40269 L .76866 .39402 L .817 .884 .893 r
F
0 g
.78227 .38512 m
.78518 .38877 L
s
.78779 .34395 m
.79747 .35632 L
s
.76866 .39402 m .76058 .45931 L .75094 .44899 L .818 .853 .874 r
F
0 g
.76127 .45374 m
.76218 .45463 L
s
.76616 .41427 m
.77347 .42157 L
s
.54759 .29407 m .5471 .23103 L .57258 .31685 L .502 .633 .911 r
F
0 g
.55451 .32504 m
.56914 .31372 L
s
.51864 .24386 m .5471 .23103 L .54759 .29407 L .531 .658 .917 r
F
0 g
.517 .25542 m
.52061 .25214 L
s
.53092 .29537 m
.54269 .28559 L
s
.52061 .25214 m
.52249 .25054 L
s
.62502 .40747 m .63999 .46407 L .62842 .48036 L .554 .602 .856 r
F
.62502 .40747 m .62842 .48036 L .6133 .42737 L .538 .59 .854 r
F
0 g
.61574 .45556 m
.62053 .4527 L
s
.54269 .28559 m
.5475 .28258 L
s
.45385 .18441 m
.45385 .18441 L
s
.62053 .4527 m
.62696 .44906 L
s
.5164 .15527 m .5471 .23103 L .51864 .24386 L .528 .667 .925 r
F
.51864 .24386 m .48494 .16993 L .5164 .15527 L .565 .696 .929 r
F
0 g
.45385 .18441 m
.48494 .16993 L
s
.4857 .22368 m
.50361 .21088 L
s
.62502 .40747 m .6133 .42737 L .60819 .35893 L .544 .611 .87 r
F
.60819 .35893 m .6133 .42737 L .59569 .37811 L .524 .597 .869 r
F
0 g
.60077 .42414 m
.60985 .41772 L
s
.59067 .38867 m
.59634 .38406 L
s
.59634 .38406 m
.59751 .38322 L
s
.60985 .41772 m
.61246 .41607 L
s
.60819 .35893 m .59569 .37811 L .59163 .31674 L .529 .617 .884 r
F
0 g
.59751 .38322 m
.60944 .37558 L
s
.57258 .31685 m .59163 .31674 L .59569 .37811 L .528 .617 .884 r
F
0 g
.57769 .35315 m
.58432 .34798 L
s
.61246 .41607 m
.62353 .40999 L
s
.58432 .34798 m
.59332 .3423 L
s
.57258 .31685 m .57914 .23728 L .59163 .31674 L .525 .624 .893 r
F
.57258 .31685 m .5471 .23103 L .57914 .23728 L .546 .647 .9 r
F
0 g
.56914 .31372 m
.57119 .31217 L
s
.5475 .28258 m
.55963 .27323 L
s
.52249 .25054 m
.54713 .23483 L
s
.57119 .31217 m
.57305 .31113 L
s
.50361 .21088 m
.51764 .20447 L
s
.76866 .39402 m .75094 .44899 L .75653 .38572 L .814 .838 .868 r
F
0 g
.76297 .41166 m
.76616 .41427 L
s
.75653 .38572 m .75094 .44899 L .74274 .43458 L .798 .819 .869 r
F
0 g
.75185 .44617 m
.76127 .45374 L
s
.63913 .38573 m .65347 .43728 L .63999 .46407 L .589 .622 .852 r
F
.62502 .40747 m .63913 .38573 L .63999 .46407 L .569 .606 .85 r
F
0 g
.62696 .44906 m
.63495 .44502 L
s
.75122 .44572 m
.75185 .44617 L
s
.74796 .44375 m
.75122 .44572 L
s
.75477 .40563 m
.76297 .41166 L
s
.78897 .33514 m .76866 .39402 L .77464 .33314 L .815 .87 .887 r
F
0 g
.78642 .34252 m
.78779 .34395 L
s
.76866 .39402 m .75653 .38572 L .77464 .33314 L .797 .849 .89 r
F
0 g
.77449 .37711 m
.78227 .38512 L
s
.63495 .44502 m
.63975 .44285 L
s
.77067 .3736 m
.77449 .37711 L
s
.62353 .40999 m
.62509 .40909 L
s
.79261 .28278 m .8156 .2544 L .78897 .33514 L .807 .891 .904 r
F
0 g
.79829 .30687 m
.80601 .31721 L
s
.81258 .2742 m
.82538 .29555 L
s
.79261 .28278 m .78897 .33514 L .77464 .33314 L .798 .882 .907 r
F
0 g
.77673 .33343 m
.78642 .34252 L
s
.81016 .27089 m
.81258 .2742 L
s
.62509 .40909 m
.6254 .40893 L
s
.63975 .44285 m
.65324 .43774 L
s
.8156 .2544 m .83913 .14539 L .85265 .17234 L .8 .942 .92 r
F
0 g
.85265 .17234 m
.85265 .17234 L
s
.83809 .20458 m
.84428 .21535 L
s
.82351 .23687 m
.83599 .25804 L
s
.85265 .17234 m
.86588 .19871 L
s
.62098 .3298 m .63913 .38573 L .62502 .40747 L .575 .627 .866 r
F
.62098 .3298 m .62502 .40747 L .60819 .35893 L .555 .612 .865 r
F
0 g
.60944 .37558 m
.61323 .37346 L
s
.55963 .27323 m
.57699 .26334 L
s
.79146 .29933 m
.79829 .30687 L
s
.61323 .37346 m
.62299 .36839 L
s
.48494 .16993 m
.48494 .16993 L
s
.60234 .29436 m .62098 .3298 L .60819 .35893 L .562 .631 .877 r
F
.60819 .35893 m .59163 .31674 L .60234 .29436 L .544 .619 .878 r
F
0 g
.59332 .3423 m
.60001 .3381 L
s
.60001 .3381 m
.606 .33469 L
s
.75653 .38572 m .74274 .43458 L .74213 .38386 L .787 .794 .858 r
F
0 g
.75148 .40361 m
.75477 .40563 L
s
.74213 .38386 m .74274 .43458 L .72601 .44046 L .766 .778 .863 r
F
.79261 .28278 m .80254 .21385 L .8156 .2544 L .796 .909 .918 r
F
0 g
.80602 .26622 m
.81016 .27089 L
s
.80254 .21385 m .83913 .14539 L .8156 .2544 L .786 .908 .925 r
F
0 g
.82058 .23131 m
.82351 .23687 L
s
.82982 .18851 m
.83809 .20458 L
s
.59163 .31674 m .57914 .23728 L .60234 .29436 L .551 .638 .89 r
F
0 g
.57305 .31113 m
.58909 .3006 L
s
.54713 .23483 m
.54802 .23413 L
s
.65589 .37264 m .66628 .43942 L .65347 .43728 L .62 .643 .851 r
F
.65589 .37264 m .65347 .43728 L .63913 .38573 L .601 .624 .847 r
F
0 g
.6254 .40893 m
.63931 .40255 L
s
.74232 .39904 m
.75148 .40361 L
s
.54823 .14044 m .57914 .23728 L .5471 .23103 L .544 .649 .903 r
F
0 g
.54802 .23413 m
.55178 .23194 L
s
.5164 .15527 m .54823 .14044 L .5471 .23103 L .572 .674 .908 r
F
0 g
.48494 .16993 m
.5164 .15527 L
s
.51764 .20447 m
.53241 .19479 L
s
.63931 .40255 m
.64337 .40098 L
s
.77464 .33314 m .75653 .38572 L .75964 .32064 L .793 .833 .883 r
F
0 g
.76275 .36764 m
.77067 .3736 L
s
.75964 .32064 m .75653 .38572 L .74213 .38386 L .772 .814 .886 r
F
0 g
.75756 .36419 m
.76275 .36764 L
s
.7266 .38278 m .74213 .38386 L .72601 .44046 L .757 .759 .854 r
F
0 g
.72766 .43468 m
.73474 .43739 L
s
.7266 .38278 m .72601 .44046 L .71148 .43565 L .733 .74 .857 r
F
.6743 .37004 m .68152 .43668 L .66628 .43942 L .642 .659 .851 r
F
.6743 .37004 m .66628 .43942 L .65589 .37264 L .632 .648 .848 r
F
0 g
.65433 .43742 m
.66551 .43449 L
s
.64337 .40098 m
.65498 .39701 L
s
.72607 .43418 m
.72766 .43468 L
s
.66551 .43449 m
.66688 .43419 L
s
.63943 .30414 m .65589 .37264 L .63913 .38573 L .601 .645 .865 r
F
.63913 .38573 m .62098 .3298 L .63943 .30414 L .582 .628 .863 r
F
0 g
.62299 .36839 m
.63219 .36434 L
s
.77464 .33314 m .78068 .27872 L .79261 .28278 L .796 .869 .902 r
F
0 g
.78795 .29582 m
.79146 .29933 L
s
.77464 .33314 m .75964 .32064 L .78068 .27872 L .774 .842 .904 r
F
0 g
.7751 .33183 m
.77673 .33343 L
s
.77481 .33157 m
.7751 .33183 L
s
.7644 .32461 m
.77481 .33157 L
s
.73826 .39745 m
.74232 .39904 L
s
.53241 .19479 m
.54763 .18783 L
s
.69344 .36957 m .69921 .42459 L .68152 .43668 L .671 .683 .852 r
F
.6743 .37004 m .69344 .36957 L .68152 .43668 L .662 .67 .847 r
F
0 g
.66688 .43419 m
.68098 .43162 L
s
.63219 .36434 m
.63922 .36143 L
s
.7266 .38278 m .71148 .43565 L .71115 .37404 L .728 .729 .851 r
F
0 g
.71268 .43145 m
.72607 .43418 L
s
.71115 .37404 m .71148 .43565 L .69921 .42459 L .705 .711 .853 r
F
0 g
.77962 .28832 m
.78795 .29582 L
s
.68098 .43162 m
.68245 .43146 L
s
.68245 .43146 m
.68988 .43096 L
s
.71146 .43124 m
.71268 .43145 L
s
.72649 .39369 m
.73826 .39745 L
s
.70616 .43085 m
.71146 .43124 L
s
.78068 .27872 m .80254 .21385 L .79261 .28278 L .775 .879 .923 r
F
0 g
.79674 .25411 m
.80602 .26622 L
s
.65498 .39701 m
.65949 .3958 L
s
.74096 .3386 m .75964 .32064 L .74213 .38386 L .761 .788 .875 r
F
0 g
.74885 .35959 m
.75756 .36419 L
s
.74096 .3386 m .74213 .38386 L .7266 .38278 L .746 .778 .879 r
F
.62098 .3298 m .61164 .23042 L .63943 .30414 L .586 .647 .877 r
F
.60234 .29436 m .61164 .23042 L .62098 .3298 L .561 .631 .878 r
F
0 g
.606 .33469 m
.61994 .32781 L
s
.61994 .32781 m
.62076 .3274 L
s
.65949 .3958 m
.67163 .39309 L
s
.60234 .29436 m .57914 .23728 L .61164 .23042 L .567 .639 .882 r
F
0 g
.57699 .26334 m
.58264 .25956 L
s
.58909 .3006 m
.60202 .29356 L
s
.58264 .25956 m
.58718 .25704 L
s
.60202 .29356 m
.60249 .29333 L
s
.71115 .37404 m .69921 .42459 L .69344 .36957 L .696 .696 .846 r
F
0 g
.62076 .3274 m
.62348 .32632 L
s
.60249 .29333 m
.61687 .28602 L
s
.62348 .32632 m
.63937 .31963 L
s
.72365 .3931 m
.72649 .39369 L
s
.74142 .35655 m
.74885 .35959 L
s
.5164 .15527 m
.5164 .15527 L
s
.58718 .25704 m
.60934 .24625 L
s
.71124 .39094 m
.72365 .3931 L
s
.67163 .39309 m
.6767 .39215 L
s
.65589 .37264 m .66275 .29991 L .6743 .37004 L .628 .665 .867 r
F
.66275 .29991 m .65589 .37264 L .63943 .30414 L .612 .648 .862 r
F
0 g
.63922 .36143 m
.65215 .35709 L
s
.7596 .32136 m
.7644 .32461 L
s
.6767 .39215 m
.68968 .39073 L
s
.58043 .12543 m .61164 .23042 L .57914 .23728 L .565 .647 .89 r
F
.54823 .14044 m .58043 .12543 L .57914 .23728 L .576 .657 .892 r
F
0 g
.54763 .18783 m
.56083 .17993 L
s
.5164 .15527 m
.54823 .14044 L
s
.55178 .23194 m
.57339 .21926 L
s
.74096 .3386 m .7266 .38278 L .72765 .32326 L .741 .765 .872 r
F
.7266 .38278 m .71115 .37404 L .72765 .32326 L .715 .743 .874 r
F
0 g
.65215 .35709 m
.6575 .35549 L
s
.70723 .39064 m
.71124 .39094 L
s
.68968 .39073 m
.69561 .39032 L
s
.69561 .39032 m
.70723 .39064 L
s
.73574 .35464 m
.74142 .35655 L
s
.57339 .21926 m
.57938 .21657 L
s
.75946 .32129 m
.7596 .32136 L
s
.7734 .24524 m .78068 .27872 L .75964 .32064 L .774 .857 .912 r
F
0 g
.77682 .28641 m
.77962 .28832 L
s
.74839 .28204 m .7734 .24524 L .75964 .32064 L .744 .81 .904 r
F
.74096 .3386 m .74839 .28204 L .75964 .32064 L .756 .812 .897 r
F
0 g
.75911 .32114 m
.75946 .32129 L
s
.68602 .30259 m .69344 .36957 L .6743 .37004 L .655 .689 .869 r
F
.68602 .30259 m .6743 .37004 L .66275 .29991 L .643 .672 .863 r
F
0 g
.6575 .35549 m
.67138 .35231 L
s
.76723 .27901 m
.77682 .28641 L
s
.7734 .24524 m .80254 .21385 L .78068 .27872 L .77 .856 .915 r
F
0 g
.79086 .2485 m
.79674 .25411 L
s
.70849 .3083 m .72765 .32326 L .71115 .37404 L .711 .734 .869 r
F
0 g
.72714 .35214 m
.73574 .35464 L
s
.70849 .3083 m .71115 .37404 L .69344 .36957 L .685 .715 .872 r
F
0 g
.63937 .31963 m
.64286 .31844 L
s
.69344 .36957 m .68602 .30259 L .70849 .3083 L .676 .7 .865 r
F
.74096 .3386 m .72765 .32326 L .74839 .28204 L .725 .773 .89 r
F
0 g
.74421 .31389 m
.75911 .32114 L
s
.64286 .31844 m
.66153 .31277 L
s
.67138 .35231 m
.67756 .35124 L
s
.71871 .35077 m
.72714 .35214 L
s
.67756 .35124 m
.69123 .34967 L
s
.66275 .29991 m .63943 .30414 L .6462 .2073 L .606 .665 .881 r
F
.63943 .30414 m .61164 .23042 L .6462 .2073 L .588 .647 .876 r
F
0 g
.61687 .28602 m
.63052 .28051 L
s
.71016 .34964 m
.71871 .35077 L
s
.56083 .17993 m
.5799 .17121 L
s
.80254 .21385 m .82532 .11785 L .83913 .14539 L .77 .939 .937 r
F
0 g
.83913 .14539 m
.83913 .14539 L
s
.8218 .17782 m
.82982 .18851 L
s
.80444 .21029 m
.82058 .23131 L
s
.83913 .14539 m
.85265 .17234 L
s
.69123 .34967 m
.69838 .34944 L
s
.69838 .34944 m
.71016 .34964 L
s
.63052 .28051 m
.64138 .27622 L
s
.60934 .24625 m
.61295 .24438 L
s
.61295 .24438 m
.61638 .24297 L
s
.66153 .31277 m
.66474 .31202 L
s
.79038 .16717 m .82532 .11785 L .80254 .21385 L .749 .887 .942 r
F
0 g
.80371 .20891 m
.80444 .21029 L
s
.81446 .16358 m
.8218 .17782 L
s
.7734 .24524 m .79038 .16717 L .80254 .21385 L .753 .888 .94 r
F
0 g
.77891 .2393 m
.79086 .2485 L
s
.74839 .28204 m .72765 .32326 L .73008 .25469 L .727 .78 .895 r
F
0 g
.73376 .31111 m
.74421 .31389 L
s
.73008 .25469 m .72765 .32326 L .70849 .3083 L .691 .748 .895 r
F
0 g
.66474 .31202 m
.685 .30847 L
s
.54823 .14044 m
.54823 .14044 L
s
.80037 .20552 m
.80371 .20891 L
s
.61638 .24297 m
.6445 .23168 L
s
.72814 .30947 m
.73376 .31111 L
s
.61303 .11025 m .6462 .2073 L .61164 .23042 L .587 .659 .887 r
F
.58043 .12543 m .61303 .11025 L .61164 .23042 L .577 .65 .886 r
F
0 g
.54823 .14044 m
.58043 .12543 L
s
.57938 .21657 m
.60388 .2043 L
s
.6747 .24139 m .68602 .30259 L .66275 .29991 L .633 .69 .885 r
F
.6462 .2073 m .6747 .24139 L .66275 .29991 L .621 .672 .877 r
F
0 g
.64138 .27622 m
.65761 .27113 L
s
.72726 .30938 m
.72814 .30947 L
s
.685 .30847 m
.68665 .30828 L
s
.5799 .17121 m
.59218 .16494 L
s
.75815 .20823 m .7734 .24524 L .74839 .28204 L .742 .834 .921 r
F
.74839 .28204 m .73008 .25469 L .75815 .20823 L .703 .785 .915 r
F
0 g
.7544 .2732 m
.76723 .27901 L
s
.70257 .2267 m .73008 .25469 L .70849 .3083 L .688 .739 .89 r
F
0 g
.70892 .30754 m
.72726 .30938 L
s
.7088 .30753 m
.70892 .30754 L
s
.68602 .30259 m .70257 .2267 L .70849 .3083 L .66 .721 .894 r
F
0 g
.68665 .30828 m
.70586 .30763 L
s
.70586 .30763 m
.70843 .3075 L
s
.70843 .3075 m
.7088 .30753 L
s
.74989 .2707 m
.7544 .2732 L
s
.6747 .24139 m .70257 .2267 L .68602 .30259 L .643 .694 .882 r
F
0 g
.60388 .2043 m
.61199 .20065 L
s
.75815 .20823 m .79038 .16717 L .7734 .24524 L .722 .832 .933 r
F
0 g
.77544 .23583 m
.77891 .2393 L
s
.78538 .19014 m
.80037 .20552 L
s
.65761 .27113 m
.66923 .26818 L
s
.73903 .26806 m
.74989 .2707 L
s
.76795 .23201 m
.77544 .23583 L
s
.66923 .26818 m
.6793 .26626 L
s
.59218 .16494 m
.6125 .15564 L
s
.72971 .26528 m
.73903 .26806 L
s
.6793 .26626 m
.69444 .26397 L
s
.6445 .23168 m
.65023 .22984 L
s
.72597 .2649 m
.72971 .26528 L
s
.70524 .26342 m
.72597 .2649 L
s
.69444 .26397 m
.70524 .26342 L
s
.58043 .12543 m
.58043 .12543 L
s
.73503 .17628 m .75815 .20823 L .73008 .25469 L .704 .798 .923 r
F
0 g
.75591 .22517 m
.75722 .22591 L
s
.75722 .22591 m
.76795 .23201 L
s
.70257 .2267 m .73503 .17628 L .73008 .25469 L .657 .751 .919 r
F
.6747 .24139 m .67607 .16542 L .70257 .2267 L .64 .711 .899 r
F
.6462 .2073 m .67607 .16542 L .6747 .24139 L .603 .679 .895 r
F
0 g
.65023 .22984 m
.66242 .2267 L
s
.66242 .2267 m
.67504 .2227 L
s
.6462 .2073 m .64601 .09488 L .67607 .16542 L .603 .68 .895 r
F
.61303 .11025 m .64601 .09488 L .6462 .2073 L .576 .657 .892 r
F
0 g
.58043 .12543 m
.61303 .11025 L
s
.61199 .20065 m
.64006 .18934 L
s
.74896 .22345 m
.75591 .22517 L
s
.6125 .15564 m
.62657 .14987 L
s
.64006 .18934 m
.64617 .1866 L
s
.79038 .16717 m .8112 .08971 L .82532 .11785 L .714 .929 .962 r
F
0 g
.82532 .11785 m
.82532 .11785 L
s
.80204 .15071 m
.81446 .16358 L
s
.82532 .11785 m
.83913 .14539 L
s
.64617 .1866 m
.64964 .18548 L
s
.64964 .18548 m
.66537 .18043 L
s
.75815 .20823 m .77134 .12562 L .79038 .16717 L .711 .866 .956 r
F
0 g
.77613 .18533 m
.78538 .19014 L
s
.73503 .17628 m .77134 .12562 L .75815 .20823 L .662 .801 .949 r
F
.70257 .2267 m .70774 .13599 L .73503 .17628 L .657 .754 .921 r
F
0 g
.73237 .21849 m
.7331 .21871 L
s
.7331 .21871 m
.74896 .22345 L
s
.70257 .2267 m .67607 .16542 L .70774 .13599 L .611 .709 .914 r
F
0 g
.67504 .2227 m
.69937 .21928 L
s
.76335 .17564 m
.77613 .18533 L
s
.70288 .21843 m
.70305 .21839 L
s
.69937 .21928 m
.70288 .21843 L
s
.62657 .14987 m
.64609 .14094 L
s
.77134 .12562 m .8112 .08971 L .79038 .16717 L .689 .863 .964 r
F
0 g
.79728 .1415 m
.80204 .15071 L
s
.70789 .21844 m
.73237 .21849 L
s
.70305 .21839 m
.70789 .21844 L
s
.66537 .18043 m
.67586 .17705 L
s
.61303 .11025 m
.61303 .11025 L
s
.67586 .17705 m
.67595 .17703 L
s
.67595 .17703 m
.6808 .17634 L
s
.70464 .17044 m
.70579 .17016 L
s
.6808 .17634 m
.70464 .17044 L
s
.67607 .16542 m .67939 .07933 L .70774 .13599 L .611 .714 .918 r
F
.67607 .16542 m .64601 .09488 L .67939 .07933 L .571 .677 .911 r
F
0 g
.64609 .14094 m
.66325 .13533 L
s
.61303 .11025 m
.64601 .09488 L
s
.73503 .17628 m .74035 .10394 L .77134 .12562 L .661 .81 .955 r
F
.73503 .17628 m .70774 .13599 L .74035 .10394 L .608 .757 .949 r
F
0 g
.73905 .17067 m
.76335 .17564 L
s
.73551 .16979 m
.73905 .17067 L
s
.73506 .16985 m
.73551 .16979 L
s
.73107 .17043 m
.73506 .16985 L
s
.70579 .17016 m
.73107 .17043 L
s
.66325 .13533 m
.67749 .12882 L
s
.64601 .09488 m
.64601 .09488 L
s
.70774 .13599 m .71318 .06359 L .74035 .10394 L .608 .759 .951 r
F
.70774 .13599 m .67939 .07933 L .71318 .06359 L .56 .71 .941 r
F
0 g
.67749 .12882 m
.70125 .12302 L
s
.64601 .09488 m
.67939 .07933 L
s
.79677 .06094 m .8112 .08971 L .77134 .12562 L .623 .902 .985 r
F
0 g
.8112 .08971 m
.8112 .08971 L
s
.77298 .12414 m
.79728 .1415 L
s
.8112 .08971 m
.82532 .11785 L
s
.77134 .12562 m .76754 .07258 L .79677 .06094 L .616 .838 .985 r
F
0 g
.77238 .12298 m
.77298 .12414 L
s
.74035 .10394 m .76754 .07258 L .77134 .12562 L .585 .817 .988 r
F
0 g
.70856 .11968 m
.70898 .11949 L
s
.70125 .12302 m
.70856 .11968 L
s
.77112 .1225 m
.77238 .12298 L
s
.76594 .12184 m
.77112 .1225 L
s
.73953 .11513 m
.73975 .11518 L
s
.73975 .11518 m
.76594 .12184 L
s
.70898 .11949 m
.72711 .11695 L
s
.72711 .11695 m
.73953 .11513 L
s
.67939 .07933 m
.67939 .07933 L
s
.74035 .10394 m .74739 .04765 L .76754 .07258 L .587 .813 .986 r
F
.74035 .10394 m .71318 .06359 L .74739 .04765 L .534 .753 .977 r
F
0 g
.67939 .07933 m
.71318 .06359 L
s
.71318 .06359 m
.71318 .06359 L
s
.76754 .07258 m .78201 .03152 L .79677 .06094 L .513 .857 .994 r
F
0 g
.79677 .06094 m
.8112 .08971 L
s
.76754 .07258 m .74739 .04765 L .78201 .03152 L .486 .791 .999 r
F
0 g
.71318 .06359 m
.74739 .04765 L
s
.78201 .03152 m
.79677 .06094 L
s
.79677 .06094 m
.79677 .06094 L
s
.74739 .04765 m
.78201 .03152 L
s
.74739 .04765 m
.74739 .04765 L
s
.25 Mabswid
.79826 0 m
.99671 .40534 L
s
.99671 .40534 m
1 .76812 L
s
1 .76812 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.00338 .37369 m
0 .74137 L
s
0 .74137 m
.80058 .42512 L
s
.80058 .42512 m
.79826 0 L
s
.79826 0 m
.00338 .37369 L
s
.00338 .37369 m
.79826 0 L
s
.01971 .36602 m
.0195 .38939 L
s
[(0)] .02012 .31927 -0.03303 1 Mshowa
.29667 .23581 m
.29658 .25919 L
s
[(2)] .29685 .18904 -0.01416 1 Mshowa
.60686 .08998 m
.6069 .11337 L
s
[(4)] .60677 .04321 .00709 1 Mshowa
.125 Mabswid
.08613 .33479 m
.08602 .34881 L
s
.15438 .3027 m
.15429 .31673 L
s
.22453 .26972 m
.22446 .28375 L
s
.37088 .20092 m
.37084 .21495 L
s
.44725 .16502 m
.44723 .17905 L
s
.52587 .12806 m
.52588 .14209 L
s
.69031 .05075 m
.69036 .06478 L
s
.77634 .01031 m
.77641 .02434 L
s
gsave
.37156 -0.0797 -66.5429 -15.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(x) show
70.187500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
0 .74137 m
.2826 1 L
s
.00774 .74845 m
.00796 .72508 L
s
[(0)] .00732 .7952 .03362 -1 Mshowa
.1058 .8382 m
.10597 .81482 L
s
[(1)] .10546 .88495 .02692 -1 Mshowa
.1951 .91992 m
.19523 .89654 L
s
[(2)] .19484 .96668 .02085 -1 Mshowa
.27675 .99465 m
.27685 .97127 L
s
[(3)] .27656 1.04142 .01531 -1 Mshowa
.125 Mabswid
.02812 .7671 m
.02824 .75308 L
s
.0481 .78539 m
.04822 .77136 L
s
.0677 .80333 m
.06781 .7893 L
s
.08693 .82093 m
.08704 .8069 L
s
.12432 .85515 m
.12442 .84112 L
s
.1425 .87179 m
.1426 .85776 L
s
.16035 .88812 m
.16044 .8741 L
s
.17788 .90417 m
.17797 .89014 L
s
.21201 .9354 m
.21208 .92137 L
s
.22862 .9506 m
.22869 .93657 L
s
.24494 .96554 m
.24501 .95151 L
s
.26099 .98022 m
.26105 .96619 L
s
gsave
.14967 1.16054 -66.7909 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 15.375000 moveto
 558.000000 15.375000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 15.375000 lineto
closepath
clip newpath
63.000000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.125000 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(y) show
70.312500 12.500000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/0P10`4000010?ooo`000<c<c;no_`00
E@00ZW=cLiVIVNfXL<V>GoOI`7eI>ooooiVIVH:2P]7AdTi>C_ooohVHZWB1T<[AfDI=E_ooofQ_PUQN
K[ZmaS4e?_oooiBDU:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJF
UWemOFATI4];Bc8b<P0T000TE@0TZP0To`190019E@19ZP19o`1]001]E@1]ZP1]o`2B002BE@2BZP2B
o`2f002fE@2fZP2fo`3K003KE@3KZP3Ko`3oE@3oZUD005D0EED0ZUD0oeDT05DTEEDTZUDToeE905E9
EEE9ZUE9oeE]05E]EEE]ZUE]oeFB05FBEEFBZUFBoeFf05FfEEFfZUFfoeGK05GKEEGKZUGKoeGo05Go
EEGoZUGoojX00:X0EJX0ZZX0ojXT0:XTEJXTZZXTojY90:Y9EJY9ZZY9ojY]0:Y]EJY]ZZY]ojZB0:ZB
EJZBZZZBojZf0:ZfEJZfZZZfoj[K0:[KEJ[KZZ[Koj[o0:[oEJ[oZZ[oool0EOl0Z_lT0?lTEOlTZ_lT
oom90?m9EOm9Z_m9oom]0?m]EOm]Z_m]oonB0?nBEOnBZ_nBoonf0?nfEOnfZ_nfoooK0?oKEOoKZ_oK
ooooEOooZ/d0003=0<g=0000cLd0c@3=cNGUiDe=C@00:P00O`00e00Z000Z:P0ZE@0ZO`0ZZP0Ze01E
001E:P1EE@1EO`1EZP1Ee01o001o:P1oE@1oO`1oZP1oe02Z002Z:P2ZE@2ZO`2ZZP2Ze03D003D:P3D
E@3DO`3DZP3De2X002X0:RX0EBX0ObX0ZRX0e2XZ02XZ:RXZEBXZObXZZRXZe2YE02YE:RYEEBYEObYE
ZRYEe2Yo02Yo:RYoEBYoObYoZRYoe2ZZ02ZZ:RZZEBZZObZZZRZZe2[D02[D:R[DEB[DOb[DZR[De5D0
:UD0OeD0e5DZ05DZ:UDZEEDZOeDZZUDZe5EE05EE:UEEEOl0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000;h00`4200<1jP000;l000<100000P000`40003Y0000`0001040007/0000
`@020Nd000310081k@000<0000@10001k0000;l000<100000P000`40003Y0000_P030@800`7Z0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`1;0081I0000?l0
B@020@03004006<0003o04L00P430081H`000?l0A@020@D00P5S0000o`0>00@1<@030@H000<10040
HP000?l040000`40000^00810P000`40000400030@010680003o010000<10000;0020@@000<10000
10001040005Q0000o`0;00L1:`020@H000<1000010001040005Q0000o`0;00030@00008000<10000
9`030@P000<1000010001@4000010600003o00`000D100000@0W00814P001@4000010600003o00`0
00D100000@0U00813@020@D000<100000P000`40001M0000o`0=00040@000B<00P4=008100<T0@00
10000`40000200030@0005d0003o00d000@100018@020@d00P401@0V02H100@000<100000`000`40
001L0000o`0>00030@0101l00P4=008100LT02@0900100@000<100000`000`40001L0000o`0>0003
0@0101d00P4=008100X09P0V02H09P010`000`40000400030@0005/0003o00l00P4K008130030@0<
900T02@0900V02H10`000`40000400030@0005/0003o02T00`4<0081010V02H09P0V02H09P0V02H1
0P000`40000500030@0005X0003o02L00`4<00810180900V02@0900T02@09P0V004200030@0000D0
00<10000FP000?l09@020@8000<1000020020@0G9P0V02H09P0V02H09P0V02H09P01004020000`40
001I0000o`0S008110000`400006008101H0900T02@09P0V02@0900T02H09P0V0`4800030@0005T0
003o02400P4600030@0000<00`406BH09P0V02H09P0V02H09P0V02H09P0V02H00P4900030@0005P0
003o01l00P4800040@00008101hT02@0900T02@0900V02H09P0T02H09P0V02H09P4900030@0005P0
003o01d00P4<008100<09P000P004bH09P0V02H09P0V02H09P0V02H01`000`40000800030@0005L0
003o01/00P4<0081024T02@0900T02@0900T02@09P0V02H09P0V02H09P0V02H00P4900030@0005L0
003o01P00`4<008100H09P0V02H5000?9P0V02H09P0V02H09P0V00T00P4:00030@0005H0003o01H0
0P4=008102PT02@0900T02@0900T02@0900T02H09P0V02H09P0V02H09P0V0BH12@000`40001F0000
o`0D00813@020@<000LV02H09P0V00L000dV02H09P0V02H09P0V00T000<100402P000`40001E0000
o`0B00813@020@0]900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P0V02H19P0100T0
00<10000E@000?l040020@d00P40100V02H300079P0V02H09P07000=9P0V02H09P0V02H09P090004
0@000@X000<10000E0000?l03P020@d00P40<R@0900T02@0900T02@0900T02@0900T02@0900V02H0
9P0V02H09P0V02H09P4V02H12@000`40001D0000o`0<00<12`030@0802H09P0V02H500059P0V02H0
2@002bH09P0V02H09P0V00T000D100000@0:00030@0005<0003o00X00P4200030@0000L00P40>00T
02@0900T02@0900T02@0900T02@0900T02@0900T02H09P0V02H09P0V02H09P0V0BH09P012@000`40
001C0000o`0700<110000`400005008100dV02H09P0V02H09P0V00D000DV02H09P0;00099P0V02H0
9P0V00T000<100000P000`40000800030@000580003o00D00P4700030@0000<00P40?@0T02@0900T
02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P0V02H09P4V02H09P402@000`40
001B0000o`0300812@00104000020@079P0V02H09P0300079P0V02H09P0700039P0V00d000LV02H0
9P0V00T000<100000`000`40000700030@000580003o00400P4;00@10480900T02@0900T02@0900T
02@0900T02@0900T02@0900T02@0900V02H09P0V02H09P0V02H09P0V0BH09P0V004900030@000540
003n008130020@0<02H09P0V02H09P0V0`001bH09P0V02H01`000bH09P0?00059P0V02H02@000`40
000400030@0000L000<10000D@000?`00P4<008102DT02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T00T101XV02H09P0V02H09P0V02H09P4V02H09P0V0@T000<10000D0000?X00P4<0081
01009P0V02H09P0V02H09P0V1@001BH09P0V00T0104900@10`000bH0000700030@0000D000<10000
1`000`40001@0000n0020@`00P408R@0900T02@0900T02@0900T02@0900T02@0900T02@090030@0A
02@0900T02@0900V02H09P001@404P0V02H09P0V0BH09P0V02H00@T000<10000C`000?D00`4<0081
01@09P0V02H09P0V02H09P0V02H09PD000DV02H09P0200@10`000bH0000B00030@0000H000<10000
1P000`40000700030@0004l0003c00813@020@0O900T02@0900T02@0900T02@0900T02@0900T02@0
90030@0M900T02@0900T02@0900T02@09P0V02H09P0V02H00P404@0V02H09P4V02H09P0V02H100T0
00<10000CP000?400`4<008101P09P0V02H09P0V02H09P0V02H09P0V02H500059P010BH08@000`40
000300030@0000L000<100001`000`40001>0000k`020@8000<1000020020@0N900T02@0900T02@0
900T02@0900T02@0900T02@00`409@0T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P00
0P403bH09P4V02H09P0V02H00@0900030@0004d0003]008110000`400006008101`09P0V02H09P0V
02H09P0V02H09P0V02H09P0V0P020@<000<V02H090020@03004000T000<100001`000`40001=0000
j`020@H000<1000010020@0M900T02@0900T02@0900T02@0900T02@0900T02@00`40;R@0900T02@0
900T02@0900T02@0900T02@0900V02H09P0V02H09P0V02H09P020@0<9P0V02H09P0V02H12@000`40
001<0000j@020@P000<100000P020@0M02H09P0V02H09P0V02H09P0V02H09P0V02H09P000P400bH0
000500039P0V02L000<100002@020@P000<10000C0000>L00P4=008101dT02@0900T02@0900T02@0
900T02@0900T02@090020@1202@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0V
02H09P0V02H09P4V02H09P0V02H09P012@000`40001;0000i0030@d00P407@0V02H09P0V02H09P0V
02H09P0V02H09P0V02H0008100DV02H09P0900039P0002D00P4;00030@0000L000<10000B`000>80
0P4>008101dT02@0900T02@0900T02@0900T02@0900T02@090020@1402@0900T02@0900T02@0900T
02@0900T02@0900T02@0900T02H09P0V02H09P0V02H09P0V02H19P4V02H09P0V02H200030@0000L0
00<10000BP000:H0204b00813@030@0L02H09P0V02H09P0V02H09P0V02H09P0V02H09P<100TV02H0
9P0V02H02@000bH0000U00040@000@X000<100001`000`40001:0000Y`000`40000300030@0002h0
0P4=008101d0900T02@0900T02@0900T02@0900T02@0900T00020@19900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H09P0V02H09P4V02H19P0V02H09P030003
0@0000L000<10000B@000:P000<10000<@020@`00`407BH09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V008100h09P0V02H09P0V02H09PT000<V00009@000`40000200030@0000L000<100001`000`40
00190000Z@000`40000^008130020@0N900T02@0900T02@0900T02@0900T02@0900T02@00P40CB@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V02H09P0V02H09P0V02H0
9P0V0BH09P0102H09P0V00@000<100001`000`4000180000ZP000`40000[00<12P030@0N02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V0P404P0V02H09P0V02H09P0V02H09Q`000<100004P000`40
000300030@0000L000<100001`000`4000180000Z`000`40000X00810P000`400006008101l0900T
02@0900T02@0900T02@0900T02@0900T02@0008102<T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900=0@03904100<101/V02H09P0V02H09P0V02H19P0V02H19P0V02H01@000`4000070003
0@0004L0002/00030@0002@00`4400030@0000@00P407bH09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H00P405P0V02H09P0V02H09P0V02H09P0V02H:00D14`040@/000<1000010000`4000070003
0@0000L000<10000A`000:d000<100008@020@L000<100000P020@0O02@0900T02@0900T02@0900T
02@0900T02@0900T00020@0P900T02@0900T02@0900T02@0900T02@0900T02@090030@0`02@0900T
02@0900T02@0900T02H09P0V02H00@0V02H09P0V02H19P0V02H00@0V20000`40000700030@0004H0
002W00030@0000<000<100007`020@`00P407bH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
0P406P0V02H09P0V02H09P0V02H09P0V02H09P0V10040Al00P4800030@0000D000<100001`000`40
000700030@0004H0002W00030@0000<000<100007@020@d0020102@0900T02@0900T02@0900T02@0
900T02@0900T008101dT02@0900T02@0900T02@0900T02@0900T02@090030@0g900T02@0900T02@0
900T02@0900T02@0900T02H09P0V02H09P4V02H09P0V0BH09P0V02H00@0900030@0000L000<10000
A@000:P000D100000@0N00813@020@0M02H09P0V02H09P0V02H09P0V02H09P0V02H09P000`407@0V
02H09P0V02H09P0V02H09P0V02H09P0V02H00081:@020@D000<100001`000`40000600030@0000L0
00<10000A@000:T00`4M00813@020@0M900T02@0900T02@0900T02@0900T02@0900T02@00P407P0T
02@0900T02@0900T02@0900T02@0900T02@0908103d0900T02@0900T02@0900T02@0900T02@0900T
02@0900V02H09P0V02H09P0102H09P4V02H09P0V02H100T000<100001`000`4000140000a`020@d0
0P407@0V02H09P0V02H09P0V02H09P0V02H09P0V02H0008101dV02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P030@039P0V02d000D100000@0:00030@0000H000<100001`000`4000140000a@020@d0
0P407B@0900T02@0900T02@0900T02@0900T02@0900T008101`0900T02@0900T02@0900T02@0900T
02@0900T0`40@b@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V02H09P0V02H0
9P0102H19P0V02H09P0V00402@000`40000700030@0004<0003200<13@020@0j02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V02H09P0V02H09P0V02H09P<100TV02H09P0V02H0
;`000`400@0:00030@0000L000<100001`000`4000130000`0030@d00P40>R@0900T02@0900T02@0
900T02@0900T02@0900T02@1900T02@0900T02@0900T02@0900T02@090020@0l900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0V02H00P401bH09P0V02H0
10000`40000600030@0000L000<10000@`000;h00P4200030@0000P00`407`0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P000P406BH09P0V02H09P0V02H09P0V02H09P0V02H00P403P0V02H09P0V
02H09P0V<@000`40000900030@0000L000<100001`000`4000120000_0020@@000<100001P020@0P
02@0900T02@0900T02@0900T02@0900T02@0900T02@20@0I02@0900T02@0900T02@0900T02@0900T
00020@11900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02H0
9P0V02H09P0V02H00`401BH09P0V00D000<100001@000bH1000800030@000480002j00811P000`40
000300<103/V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H19P0V02H09P0V02H09P0V02H0
9P0V02H09P020@0B02H09P0V02H09P0V02H09P0V<@001040004900030@0000L000<100001`000`40
00110000^0020@P000@100000P40?2@0900T02@0900T02@0900T02@0900T02@0900T02@0900T0B@0
900T02@0900T02@0900T02@0900T008104dT02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02H09P0V02H09P0V02H19P0102H09P0600030@0000@000<V0040
2@000`4000110000]P020@`00P408`0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H00081
02lV02H09P0V02H09P0V02H09P0V02H09P0102H09P0V02H09P0V02H09P0V02H09P0=01414`001@40
000100X000<100001P000`40000700030@000400002d008130020@0m900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@00@0T02@0900T02@0900T02@0900T02@090020@0O02@0900T02@0900T
02@0900T02@0900T02@0900T00050@0A02@0900T02@0900T02@090000P406B@0900V02H09P0V02H0
9P0V0BH09P0102H01`000`40000300049P0V0@T000<10000@0000;400`4<008102@09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H20@0G02H09P0V02H09P0V02H09P0V02H09P000P406BH0
9P0V02H09P0V02H09P0V02H09P0V02H01P020AP0104=00030@0000<000<100001`000`4000060003
0@0000L000<10000?`000:l00P4=008102@T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
90020@0G900T02@0900T02@0900T02@0900T02@00P407P0T02@0900T02@0900T02@0900T02@0900T
02@090<101hT02@0900T02@0900T02@0900T02@0900T02@090020@0B9P0V02H09P0V02H19P0V02H1
2@000`40000200059P0V00402@000`40000o0000[@020@d00P40?@0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0102H09P0V02H09P0V02H09P0V02H09P000P406bH09P0V02H09P0V02H0
9P0V02H09P0V02H09P050B<00P4900030@0000@000<100001`000`40000600030@0000L000<10000
?P000:/00P4=008102DT02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T008103<0900T
02@0900T02@0900T02@0900T0040900T02@0900T02@0900T02@0900T02@090000P40>b@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@09P0V00409P0V02H09P4V02H09P0100T000X10000
9P0V02H12@000`40000n0000Z@020@d00P40?P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V0BH09P0V02H09P0V02H09P0V02H09P0V0P406P0V02H09P0V02H09P0V02H09P0V02H09P0V
0P40100V02HY00030@0000D000<100001@000`40000700030@0000H000<100001`000`40000m0000
Y`020@03004000T00`40?R@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0904T02@0
900T02@0900T02@0900T02@00P406R@0900T02@0900T02@0900T02@0900T02@00P40?B@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02H09P0V00409P0V02H19P0V02H00`000`40
0006000:0BH09P0V02H00@T000<10000?@000:D00P4300030@0000H00P409bH09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P030@0`9P0V02H09P0V02H09P0V02H09P0V00409P0V02H0
9P0V02H09P0V02H09P0V02H00`40200V02H09P0V:`000`40000300030@0000H000<100001`000`40
000600030@0000L000<10000?0000:<00P4500030@0000<00`409`0T02@0900T02@0900T02@0900T
02@0900T02@0900T02@0900T00020@0G900T02@0900T02@0900T02@0900T02@00P40600T02@0900T
02@0900T02@0900T02@0908103L0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T
02@0900T02H09P0V02H0008100TV02H19P0V02H010000`400006000:0@0V02H09P0V0@T000<10000
?0000:000`4700040@00008105/09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
00409P0V02H09P0V02H09P0V02H09P0V02H19P0V02H09P0V02H09P0V02H09P0V02H0008100dV02H0
9P0V02H09P0V02h000@100012@000`40000700030@0000H00P4800030@0003/0002N00812P00G`40
0@4T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900102@0900T02@0900T02@0
900T02@0900T0040900T02@0900T02@0900T02@0900T008103h0900T02@0900T02@0900T02@0900T
02@0900T02@0900T02@0900T02@0900T02@0900V02H09P0V02H09P<100<V02H01`000`400005000;
9P4V02H09P0V00402@000`40000k0000W0020@`00P40F`0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0102H09P0V02H09P0V02H09P0V02H09P0V00409P0V02H09P0V02H09P0V
02H09P000P403bH09P0V02H09P0V02H09P0c00030@0000P000<100001`000`40000600030@0000L0
00<10000>P0009X00P4<00<104@09P0T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T
02@00@0T02@0900T02@0900T02@0900T02@0908101@0900T02@0900T02@0900T02@0908104P0900T
02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02H09P0V02H09P0V
02H09P4V02H800030@0000@000`V00409P0V02H09P4900030@0003X0002H008130020@1K02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H00@0V02H09P0V02H09P0V02H09P0V
02H09P4V02H09P0V02H09P0V02H09P0V00020@0C9P0V02H09P0V02H09P0V02H09P0C00D16`020@X0
00<100001P000`40000700030@0000L000<10000>@0009H00P4<008104HV02H09P0V02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@00@0T02@0900T02@0900T02@0900T02@00P404b@0
900T02@0900T02@0900T02@00P40800T02@0900T02@0900T02@0900T02@0900T02@0900T20401@0T
02@000L101L0900T02H09P0V02H09P0V02H09P4V0@0:00030@0000<000dV02H19P0V02H09P0100T0
00<10000>@0009<00`4<008105/09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H19P0V02H09P0V02H09P0V02H09P0V00409P0V02H09P0V02H09P0V02H0008101LV02H0
9P0V02H09P0V02H09P0V02H09P0800<150030A4000<100402P000`40000600030@0000L000<10000
1`000`40000h0000TP020@`00P40AbH0900V02H09P0V02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@1900T02@0900T02@0900T02@0900T00810300900T02@0900T02@0900T02@00@0T
02@0900T02@0900T02@0900T02@0900T02@50@0J900T02@0900T02@0900T02@0900T02@090040@0>
9P0V02H09P0V02H09P4200030@0000P000<100000P003RH09P4V02H09P0V02H12@000`40000h0000
S`030@03004000T00P40<@0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P000P40:BH09P0V02H09P0V02H09P0V02H09P4V02H09P0V02H09P0V02H09P0V008101X09P0V
02H09P0V02H09P0V02H09P0V02H09P<18`020@/000D100000@0900030@0000H000<100001`020@P0
00<10000=`0008d00P4400030@0000H00P40N2H09P0V02@0900V02H09P0T02@0900T02@0900T02@0
900T02@0900T02@0900T02@00@0T02@0900T02@0900T02@0900T02@1900T02@0900T02@0900T02@0
904T02@0900T02@0900T02@0900T02@0900T008102PT02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900V02H00P402RH09P0V02H09P4400030@0000L001<102H09P0V00409P0V02H09P0100T0
00<10000=`0008/00P4600030@0000<00`40N00V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V00409P0V02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V02H09P0V
02H19P0V02H09P0V02H09P0V02H09P0V02H09P8100<09P00:@000`40000600030@00008000<10000
1`000`40000600030@0000L000<100001`000`40000f0000R@020@P000@100000P40N@0V02H09P0V
02H0900T02@0900V02@0900T02@0900T02@0900T02@0900T02@0900T02@00@0T02@0900T02@0900T
02@0900T02@1900T02@0900T02@0900T02@0904T02@0900T02@0900T02@0900T02@090000P40=B@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900V02H09P0V0BH09P0V008000<10000
0`000`400007000C0BH09P0V02H19P0V02H09P0V0@0900030@0003H0001@00<1=0020@`00P40NBH0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V00409P0V02H0
9P0V02H09P0V02H09P0V00409P0V02H09P0V02H09P0V02H19P0V02H09P0V02H09P0V02H09P0V02H0
0P401P0V02H09R`00P4500030@0000<000<100001`000`40000500030@0000P000<100001`000`40
000e0000C`001@40000103400P4<008107T09P0V02H0900T02H09P0T02@0900T02@0900T02@0900T
02@0900T02@0900T02@0900T02@0904T02@0900T02@0900T02@0900T02@00@0T02@0900T02@0900T
02@0904T02@0900T02@0900T02@0900T02@0008103TT02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02H09P0V02H09P0102H00P000`40000400030@0000H001@19P0V02H09P01
02H09P0V02H00@T000<10000=@0004h000<100000`000`40000/00812`030@1h9P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V02H0
9P0V02H09P0102H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V02H09P0V02H00`402P0V02H09P0V
02H^00030@00008000<1000010000`40000700030@0000D000<1000020000`40000700030@0003@0
001>00030@0000<000<10000:P020@/00P40NBH09P0V02H09P0V02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T0B@0900T02@0900T02@0900T02@0900102@0900T02@0
900T02@0900T0B@0900T02@0900T02@0900T02@00P40?P0T02@0900T02@0900T02@0900T02@0900T
02@0900T02@0900T02@0900T02@0900V02H09P0V02H09P4V0P000`40000500030@0000H001@102H0
9P0V00409P0V02H09P0V0@T000<10000=00004h000<100000`000`40000W00<12`020@2:02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V0BH09P0V
02H09P0V02H09P0V02H09P0102H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V02H09P0V02H19P0V
02H09P0V02H09P0V<0001040004800030@0000D000<100001P000`40000800030@0000L000<10000
<`0004h000<100000`000`40000U008130020@1j9P0V02H09P0V02H09P0V02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900102@0900T02@V900T02@0900T02@00@0T02@0
900T02@0900T02@0904T02@0900T02@0900T02@090020@11900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900V02H09P0V02H09P0V02H00`4800030@0000<001PV02H1
9P0V02H09P4V02H09P0V02H09P4800030@0003<0001>00030@0000<000<100008`020@03004000T0
0P40I`0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V00409P0V02H09P0V02H09P0V02H09P0102H09P0V02H09P0V02H09P000P409BH09P0V02H0
9P0V02H09P0V00409P0V02H09P0V02H09P0V02H05P030AX000<100001`000`40000500030@0000H0
00<1000020000`40000600030@0003<0001>00030@0000<000<100008@020@<000<100001P020@1k
9P0V02H09P0V02H09P0V02H09P0T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@192HT02@0900T9R@V900T02@092HT0B@0900T02@0900T02@0900102@0900T02@0900T02@0
900T008101h0900T02@0900T02@0900T02@0900T02@0900T02@;0@03900T00@101@0900T02H09P0V
02H09P0V02H09P8000<100001`000`400002000I9P0V00409P0V02H00@0V02H09P0V02H00@080003
0@000380001>00030@0000<000<100007`020@@000<100001@020@1k02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V
02H09P0V02H09P4V02H09P0V02H09P0V02H00@0V02H09P0V02H09P0V02H0008101DV02H09P0V02H0
9P0V02H09P0V02H01P050A800`4C00812@000`40000500030@0000D000<100002@000`4000060003
0@000380001>00030@0000<000<100007@020@H000<100000`020@2G9P0V02H09P0V02H09P0V02H0
9P0V02H0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@092HT02@0904T02@V92HT02@0
92HT9R@V900T02@V0@0T02@092HT02@0900T02@1900T02@0900T02@0900T0040900T02@0900T02@0
900T02@0900T02@090030@0J900T02@0900T02@0900T02@0900T02@090030@0>02H09P0V02H09P0V
02H200812@007P409P0V02H00@0V02H09P0102H09P0V02H09P0V0@P000<10000<@0004l000D10000
0@0N008130020@2E02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H00@0V02H09P0V02H09P0V02H09P0V00409P0V02H09P0V02H09P0V
0BH09P0V02H09P0V02H09P0102H09P0V02H09P0V02H09P0V02H09P00104P00813P000`400@090003
0@0000@000<100001P000`40000900030@0000H000<10000<@0005000`4M008130020@2E9P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02@0900T02@0900T02@0900T02@0900T02@092HT02@0900T02@V
92HT0040900T9R@V92HT02@V92HT9R@0900T0B@V900T02@V92HT02@0904T02@0900T02@0900T02@0
0@0T02@0900T02@0900T02@0900T02@00P409P0T02@0900T02@0900T02@0900T02@0900T02@0900T
02@09P0V0P402P0V02H09P0V02H200040@000@P001h19P0V02H09P4V02H09P0V0BH09P0V02H09P0V
004800030@000300001]00<130020@2E02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V02H09P0V02H09P4V
02H09P0V02H09P0V02H19P0V02H09P0V02H09P0V00409P0V02H09P0V02H09P0V02H09P000P400bH0
9P0W00812P001@40000100P000<1000010000`40000500030@0000T000<V0@001`000`40000`0000
J`020@d00P40UBH09P0V02H09P0V02H09P0V02H09P0V02@09P0V02H0900T02@0900T02@0900T02@0
900T02@0900T9R@V900T02@092HT9R@1900T02@V92HT9R@V92HT9R@V92HT004V92HT9R@092HT9R@V
904T02@0900T02@0900T02@00@0T02@0900T02@0900T02@0900T008102h0900T02@0900T02@0900T
02@0900T02@0900T02@0900T02@0900V02@09P0V0P401P0V02H09P8000D100000@06000P9P0102H0
9P0V0BH09P0V02H00@0V02H09P0V02H09P4800030@0002l0001Y00813@020@2E02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H00@0V02H09P0V02H09P0V02H09P0V00409P0V02H09P0V02H09P0102H09P0V02H09P0V02H00@0V
02H09P0V02H09P0V02H09P000P401bH09P0V02H0:`000`40000500030@00008000<100001@000`40
000500030@0000D000<100001P001BH09P0100P000<10000;`0006L00P4<00<1084V02H09P0V02H0
9P0V02H09P0V02H09P0V02H0900T02@09P0T02@0900T02@0900T02@0900T02@0900T02@V92HT9R@0
900T9R@V904T9R@092HT9R@V92HT9R@V92HT9P4V92HT9R@V900T9R@V92H19R@V900T02@0900T02@0
0P404@0T02@0900T02@0900T02@000<103P0900T02@0900T02@0900T02@0900T02@0900T02@0900T
02@0900T02@0900T02H09P0V00409P@000<100000P000`400003000Q9P0102H09P0V00409P0V02H0
0@0V02H09P0V02H09P0100P000<10000;P0006D00`4;00810:<V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V0BH0
9P0V02H09P0V02H09P0V02H09P4V02H09P0V02H09P0V02H19P0V02H09P0V02H09P4V02H09P0V02H0
9P0V02H09P0102H09P0V02H09P0V02d000<100000`000`40000300030@0000D000<1000010000`40
000500030@0000@000PV02H09P0V0@P000<10000;P0006<00P4200030@0000L00P40d00V02H09P0V
02H09P0V02H09P0V02H0900V02H09P0V02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T
9R@V92HT9R@092HT9P4V92HT9R@V92HT9R@V92HT9R@V904T9R@V92HT9R@V92HT9R@192HT9R@V900T
02@0904T02@0900T02@0900T02@0900102@0900T02@0900T02@0900T02@0900T02@0900T02@0900T
02@0900T02@0900T02@09P0V02H300030@00008000<100000`009`4002H09P0V0BH09P0V00409P0V
02H00@0V02H09P0V02H09P0V0@0800030@0002d0001Q008110000`400005008109LV02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P4V02H09P0V02H09P0V02H09P0V00409P0V02H09P0V02H09P4V02H09P0V02H0
9P0V00409P0V02H09P0V02H09P0V008100h09P0V02H09P0V02H09Rl000D100000@0600030@0000@0
00<1000010000`40000500030@00008000/V02H09P0V02H00@0800030@0002d0001O00811P000`40
000200<109L09P0V02H09P0V02H09P0V02H09P0V02H09P0T02@0900T02H0900T02@0900T02@0900T
02@0900T02@V92HT9R@V92HT9R@V92HT9R@V92HT9R@192HT9R@V92HT9R@V92HT9R@V92HT0B@V92HT
9R@V92HT9R@V0BHT9R@V92HT9R@0900102@0900T02@0900T02@0008103/T02@0900T02@0900T02@0
900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P0V02H09P0500040@000@H002L102H0
9P0V0BH09P0V00409P0V02H09P4V02H09P0V02H09P0V004020000`40000/0000G0030@P000<10000
0P40[00V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V00409P0V02H09P0V02H09P0V02H09P4V02H09P0V
02H09P0V00409P0V02H09P0V02H00@0V02H09P0V02H09P0V00409P0V02H09P0V02H09P0V02H;0101
5@001040004700030@0000<000<1000010000`400006000A0@0V02H09P0V02H09P0V00401`000`40
000/0000FP020@`00P40]RH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0T02@0900T02@0
900T02@0900T02@0900T9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT0B@V92HT9R@V92HT9R@V
92HT9R@V0BHT9R@V92HT9R@V92HT0B@V92HT9R@V92HT9P40900T02@0900T02@0900102@0900T02@0
900T02@0900T02@0900T02@00`40400T02@0900T02@0900T02@30@0=900T02@09P0V02H09P060003
0@0100L002L19P0V02H00@0V02H09P4V02H09P0V0BH09P0V02H09P0V02H09P4020000`40000[0000
F0020@`00P40/00V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0102H09P0V02H09P0V02H09P0V
02H19P0V02H09P0V02H09P4V02H09P0V02H09P0102H09P0V02H09P0V02H00@0V02H09P0V02H09P0V
02H09P0V0`050AH0104@00811`000`40000400030@0000@000<100000`0052H00@0V02H09P0V02H0
9P0V02H11`000`40000[0000EP020@`00P40/bH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H0900T02@0900T02@0900T02@0900T9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
904T9R@V92HT9R@V92HT9R@V92HT9P4V92HT9R@V92HT9R@V0BHT9R@V92HT9R@V904T9R@0900T02@0
900T0040900T02@0900T02@0900T02@0900T008101l0900T02@0900T02@0900T02@0900T02@0900T
02@0008100LV02H09P0V00P000<100001@00:04V02H09P0102H09P0V0BH09P0V02H19P0V02H09P0V
02H09P0V004800030@0002X0001D008130020@2c02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0102H09P0V02H09P0V02H09P0V02H19P0V02H09P0V02H09P4V02H09P0V02H09P4V02H09P0V
02H09P0T00409P0V02H09P0V02H09P0V02H09P000P4S00813@000`40000600030@0000<000<10000
10006`40000V02H00@0V02H09P0V02H09P0V02H00@0700030@0002X0001A00<130020@2M9P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0900T02@0900T02@0900T02@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V904T9R@V92HT9R@V92HT9R@V92HT9P4V92HT9R@V
92HT9R@V0BHT9R@V92HT9R@192HT9R@0900T02@090020@0C02@0900T02@0900T02@0900T00030@0Z
02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02H00@0V2P020@L002P102H09P0102H0
9P0V0BH09P0V02H00@0V02H09P0V02H09P0V02H120000`40000Y0000D0020@`00P40/P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V00409P0V02H09P0V02H09P0V02H09P4V02H09P0V
02H09P0102H09P0V02H09P0102H09P0V02H09P0V00409P0V02H09P0V02H0900V02H20@0402H09RL0
00<100002@020@L000<100000`000`400004000L0@0V02H09P0V0BH09P0V02H09P0V02H09P0V0@L0
00<10000:@0004h00P400`01000800<10;8V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H0900T02@0900T02@0900T02@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V904T9R@V92HT9R@V92HT9R@V92HT9P4V92HT9R@V92HT9R@192HT9R@V92HT9P4V92HT9R@V
900T02@00@0T02@0900T02@0900T02@00P40;R@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900V02H09P020@T000<100401P00:BH19P0V02H19P0V02H19P0V02H09P0102H09P0V02H0
9P0V02H09P0100P000<10000:00004`00P4300030@0000D00P40/bH09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H00@0V02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V0040
9P0V02H09P0V0BH09P0V02H09P0V0040900T02H09P0V02H0900T008100P09P0V02H09RX000<10000
1P001040004600030@0000<002<102H09P4V02H09P0V02H19P0V02H09P0V02H09P0V02H00@070003
0@0002P0001<00<10P000`40000300810;<09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0T02@0900T02@0900T02@092HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT
9R@V92HT9R@V92HT9P4V92HT9R@V92HT9R@V92HT9R@V0BHT9R@V92HT9R@V904T9R@V92HT9R@192HT
9R@V92HT9R@00@0T02@0900T02@0900T00020@0c900T02@0900T02@0900T02@0900T02@0900T02@0
900T02@0900T02@0900V02H09P0V008000<100001@0010400045000Z9P4V02H09P4V02H00@0V02H0
9P0V0BH09P0V02H09P0V02H09P0V02H120000`40000W0000@@030@P000<100400P00104000020@31
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P4V02H09P0V02H0
9P0V02H09P0V0BH09P0V02H09P0V00409P0V02H09P0102H09P0V02H09P0102@0900T02H09P0V02H0
0@0T02H09P0V02H09P0/00030@0000@000D100000@0500030@0000<002@102H09P4V02H09P0V0040
9P0V02H09P0V02H09P0V02H09P4700030@0002L0001000050@0000401`001P4000400@800;D102H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02@0900T02@0
900T02@0900T9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V904T9R@V92HT9R@V
92HT9R@V92HT0B@V92HT9R@V92HT0B@V92HT9R@V904T9R@V92HT9R@V904T02@0900T02@0900T0081
03H0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02H09P0V02H40003
0@0000<000D100000@04000[9P4V02H00@0V02H00@0V02H09P0V0BH09P0V02H09P0V02H09P0V02H0
0@0800030@0002H0000o00030@0000<000<100001000104000020@<00<<102H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H00@0V02H09P0V02H09P0V02H09P0102H0
9P0V02H09P0102H09P0V02H00@0V02@09P0V02H19P0T02@0900T02H09P4T02@0900V02H09P0V02H0
;P000`40000200030@00008000<100000`00:P4000019P0V02H19P0V02H09P0102H09P0V02H09P0V
02H09P0V02H00@L000<100009P0003l000<100000`000`40000400030@00008000<100000P00d040
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02@0900T02@0
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V904T9R@V92HT9R@V
92HT9R@V92HT0B@V92HT9R@V92HT0B@V92HT9R@V0BHT9R@V92HT9R@V0BHT02@0900T02@0904T02@0
900T02@0900T02@0900T02@0900T02@0900;0@0<02@0900T02@0900T0`000bH0000400050@000040
100070409P0V00409P0102H09P0102H09P0V02H19P039P0A02H09P0V02H09P0V02H09P4020000`40
000U0000?`000`40000300030@0000@000<100000`000`40000200260@0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0102H09P0V02H09P0V02H09P0V00409P059P09
02H00@0V02H0008V02/T0BH09P0T02H09P0102H0900T02@0900V0BH0900T02@09P0V02H09P0V00P0
1P4;00D14@001@40000100@000<100000`00:`4000019P0V02H19P0V02H09P0102H09P0V02H09P0V
02H09P0V02H09P401`000`40000U0000?`000`40000300030@0000@000<100000`000`4000020034
0BH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0T02@0900T
02@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V0BHT9R@V92HT
9R@V92HT9R@V92H19R@V92HT9R@V904T9R@V92HT9P4V92HT9R@V92H19R@V92HT02@0900T0B@0900T
02@0900T02@0900T02@090D101HT02@0900T02@0900T02@0900T02@00P401B@0900V00/000@10001
10005P409P0V00409P0102H09P0102H09P039P0H0BHV9P0V02H09P0V02H09P0V02H09P0120000`40
000T0000?`000`40000300030@0000@000<1000010000`40000200230BH09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H19P0V02H09P0V02H09P0V02H09P400bH00`0V
9P039P0700409P0V00029P0/904V02H0900T00409P0T02@0900T02@1900T02@0900T02H09P0V02H0
9P020Ad00`4=00040@000@@000<100000`00;04000409P0V00409P0V02H09P4V02H09P0V02H09P0V
02H09P0V02H09P011`000`40000T0000?`000`40000300030@0000@000<100001@000`400002002n
0BH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0T02@0900T02@V900T02@0900T
9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@192HT9R@V92HT
9R@V92HT9R@V904T9R@V92HT9R@192HT9R@V92H19R@V92HT9R@192HT9R@V900T02@1900T02@0900T
02@0900T02@090810280900T02@0900T02@0900T02@0900T02@0900T02@0900T0P4<00030@0100@0
01@V0BH09P4V02H19P0V02H19P0V00@V01X19RH09P0V02H09P0V02H09P0V02H09P0V0@P000<10000
8`0003l000<100000`000`40000400030@0000H000<100000P00PP4V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V0BH09P0V02H09P0V02H09P0V00499P060BHV02H0
0RH0:2@1900V02@0904V02H0900T02@0904T02@0900T02@0900V02H09P020@039P0V02<000<10000
2P020@D000D100000@03000Z0@0V02H19P0V02H09P0102H09P0V02H09P0V02H09P0V02H09P0V02H1
1P020BD0000o00030@0000<000<1000010020@P000<100000P00]`4V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02@0900T02@0900T9R@V92HT02@V92HT9R@V92HT9R@V92HT9R@V92HT
9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92H19R@V92HT9R@V92HT9R@V92HT9P4V92HT9R@V92H1
9R@V92HT9P4V92HT9R@V0BHT9R@V92HT9R@1900T02@0900T02@0900T00030@0Z02@0900T02@0900T
02@0900T02@0900T02@0900T02@0900T02@09P012`020@@001@V0BH09P4V02H19P0V00409P0V00<V
01`19P0V02H09P0V02H09P0V02H09P0V02H09P0110020@03004002@0001000050@0000401`000`40
000800030@00008008819P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H00@0V02H09P0V02H09P0V02H09P0122H00`4T9P049P03904T008V01h092HT0BH09P0T02@0904T
02@0900T02@0900T02H20@0802@09P0V02HU00030@0000T000<100000`00<@40000102H00@0V02H1
9P0V02H09P4V02H09P0V02H09P0V02H09P0V02H09P0V02H01@4300030@0002<0001100<120000`40
000900030@0000800>019P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0T02@0900T
02@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT
9R@V0BHT9R@V92HT9R@V92HT9R@V92H19R@V92HT9R@192HT9R@V0BHT9R@V92H19R@V92HT9R@V0@0T
02@0900T02@0900T0B@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@09P<0
00<1000020005`4002H09P4V02H19P0102H09P4V02H000<V01400@0V02H09P0V02H09P0V000:9P@1
1P000`40000S0000C0000`40000:00030@00008007T19P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H00@0V02H09P0V02H000<V00<09P000RH00`4V9P059P030BHT00<V
02dT0B@V900T9R@1900V02@0900102@0900T02@0900T02@19P0T02@09P0V02H09`020@P00P44000]
0@0V00409P4V02H19P0V02H09P0102H09P0V02H09P0V02H09P0V02H09P0V00<12@000`40000S0000
C0000`40000;00030@00008002D19P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0008V
08HT9R@0900T02@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V0BHT9R@V92HT9R@V92HT9R@V904T9R@V92HT9P4V92HT9P4V92HT9R@192HT9R@V
92H19R@V900T02@0900T008102dT02@0900T02@0900T02@0900T02@0900T02@0900T02@0900T02@0
900T02H020000`400005008100d09P0V0BH09P4V0BH0008V00H102H09P029P030BHV01DV0`4;0003
0@0002<0001<00030@0000`000<100000P00K@4V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P000P402RH09P0V02H09P059P0302H100LV00@192HT0bH02@4V92HT9P4V
90029P0R02@0904T02@0900T02@0900102@09P0T02@09P0V02H09RP000<100000`030@@002L10000
0@0102H00@0V02H09P0102H09P0V02H09P0V02H09P0V02H0104>00030@0002<0001<00030@0000d0
00<100000P00KP4V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0T9R@V92HT9R@V92HT
9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT
0P40F2HT9R@V92HT9R@V92HT9R@192HT9R@V92H19R@V904T9R@V904T9R@V92HT0B@V92HT9R@0900T
0040900T02@0900T02@0900T02@0900T02@0900T02@5000=900T02@0900T02@09P0:00030@000081
0P0030409P0V0BH00@0S0@8V00D102HV90039P030BHV00TV00PT9R@V92HT9P@140000`40000S0000
C0000`40000>00030@00008007419P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0100059P0302H000LV0P439P0e92HV0B@V92H19R@V92H19R@V900T0040900T02@0
900T0040900T02@0900T02H09P0V02H01`060@@0104D00<110020@<000P102H19P4V008101PV02H0
9P0102H09P0V02H09P0V02H09P040A<000<100008`0004`000<100003P000`400003001_0BH09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT
9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V0081010T9R@V92HT9R@V
92HT9R@V0P40>B@V92HT0B@V92H19R@V904T9R@V92H19R@V92HT9R@0900102@0900T02@0900T02@0
900T02@090040@0402@090T000<T0@400P401P0T02@09PP00P400`01000500090BH09P4V0B<100@V
00<T9R@00bH00`4V9P059P0992HT9R@V92HT00<15P000`40000S0000C0000`40000?00030@0000<0
07019P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V0P400`0V
9P059P0302HV00DV03L19RHT9R@192HT0B@V92H19R@V92HT0B@0900T02@0900102@0900T02@0900T
02H09P0V004000815P020@T00`4300030@0000@0024100000@0100400@0V02H09P0102H09P0V02H0
9P0V02H0104H00030@0002<0001<00814@000`400002000T0@0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H00RH0C2@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92H30@0>9R@V92HT9R@V92HT9R@20@079R@V0BHT
9P020@0W92HT0B@V92HT9P4V92HT9R@V92H102@0900T02@0900T02@0900T00817`0012H000020@L0
00<100000`020@0402H18`8100@V0@4T1BH03P4V92HT9R@V92HT9R@V0`4K00030@0002<0001<0003
0@00014000<100000P00M0409P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H00P4022H09P0V02H01RH20@0[92H10B@V0BHT9P4V92HT9R@1900T02@0900102@0
900T02@0900T02@09P030@039P0V01h0104900030@0000<00P403BH00@4V0BH19P0V02H00P40300V
02H09P0V02H09P<17@000`40000S0000C0000`40000B00030@0000800P40LbH09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0T9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@01P402BHT9R@V92H19P030@0^
9P410BH19R@V0BHT9R@192HT9R@V92HT0B@0900T02@0900T02@0904T02@091d01@4;00030@000080
00D19P419P040@0B92HT9R@192HT9R@V92HT9P4V0P4P00030@0002<0001<00030@0001<000<10000
0P00NP409P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H01@479P8102@T9P4192HT0B@V92H19R@0900T02@1900T02@0900T02@090020@059P0V02H0
70020@D00P4900030@00008000<100000P401`0V0BH09P000P402RH09P0V02H09P030B8000<10000
8`0004`000<1000050000`400002000R0@0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V008V
05XT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92H70@059P4V0BH00`4072H19R@V904T9R@V92HT
9P4V900T02@0900T02@20@0602@0900T6P030@T000<100001P001`40000V0B<00P401RH10@4V9081
00PV92HT9R@V90@19@000`40000S0000C0000`40000E00030@00008007h102H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0<9P0W904T0B@V
92H19R@V900T0040900T02@0900T02@1900T02H09P0V01L00P4<00030@0000L000T100000@010040
0RH02P010BH09P0V02H40BP000<100008`0004`000<100005P000`400002000J0@0V02H09P0V02H0
9P0V02H09P0V02H09P029P0392HT00<V06HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92H19P4V0BH30@0U9P4V92HT0B@V92HT9P4V92HT9R@0900T02@1900T02@0900T0@0D00<13`000`40
000600060@009P4V0`402B@10BHT9R@V90050BX000<100008`0004`000<100005`000`4000020026
0@0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P020@<001LR0BHT9P4V92HT02@1900T02@0900T00020@0<900T02@09P0V004R
4@020A<000<100001@001P400001008V0P400bH09P039P@1;@000`40000S0000C0000`40000H0003
0@00008001X102H09P0V02H09P0V02H09P0V02H09P0V00PV05lT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V90040@H00P408BHT0B@V92HT0B@V92HT9R@0900102@0900T02@090018P020@d0
0`4F00030@0000D000@100400`401RH10B@V90@1;`000`40000S0000C0020AX000<100000P00O@40
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V00<12`020@0Q9P4V92HT0040900T02@0900102@0900T02@09P0V8P4R00<12@020AX000<10000
10000`400@069P030BH10081<P000`40000S0000C0000`40000I00030@0000<0018102H09P0V02H0
9P0V02H09P0@9P1H92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9P813`007`4T0B@V92H19R@V92HT
9R@V0@0T02@0900T02@09280204400<17@000`40000300080@019R@V92H30CD000<100008`0004`0
00<100006P000`400003001g0@0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H00P4B000Q9R@192HT0B@0900T02@0904T02@0900T02@09R818P4R00<1
00<R0@400`4P00030@0000<000@1004V1@400`01000d00030@0002<0001<00030@0001/000<10000
0P005@4V02H09P0V02H09P0V02H09P0V000<9P1D92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V0P4E000O
904V92H19R@V92HT9R@1900T02@0900T02@R0B818P090Al000<100400`0012H00@040@D000<10000
<P000`40000S0000C0000`40000L00030@00008007419P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V00030AP0024192HT0B@0900T02@1900T02@0900T02@R0B81
8P4R0B802`4E00H10P000`40000200040BH10@L000<10000<P000`40000S0000C0000`40000M0003
0@00008001D19P0V02H09P0V02H09P0V02H09P0032H0C2@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92H30A`001H192H1
9R@V92HT0B@V900T02@0900T10400b818P030@038P4100/13P0:0@8000H1000V004;00030@000340
00<100008`0004`000<100007P000`400002001Z0BH09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P817`00604V904T02@V904T02@0900T02@092818PT100/R0B818P4R0B81
8P0J0@<000D100000@0<00030@00030000<100008`0004`000<100007`000`400002000E0BH09P0V
02H09P0V02H09P0V02H000`V04DT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@00`4R000F92H19R@V92H19R@V92HT02@092818PD1
00DR0B818P0U0@<000D102H00@0<00030@00030000<100008`0004`000<1000080000`400002001S
0BH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H000819@004bHT0B@V92H102@0900T
02@090001@402B818P4R0B818P070@038P4R01X10`000`40000?00030@0002l000<100008`0004`0
00<100008@000`400002000E0BH09P0V02H09P0V02H09P0V02H000hV03dT9R@V92HT9R@V92HT9R@V
92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT9R@V92HT0081:0004BH19R@V904T
9R@V92HT02@000<100dR0B818P4R0B818P4R02<10`0010408`4>00030@0002l000<100008`0002`0
0`4800811P080@D00`4R00030@00008005d19P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P00
0P4[000=0B@V92H102@0900T000;0@0I8P4R0B818P4R0B818P4R0B818P4R0B818P0B0@<000<10000
40000`40000^00030@0002<0000[00050@0000401`020@L000<100000`000`40000300030@0002<0
00<100000P005@4V02H09P0V02H09P0V02H09P0V000@9P0;92HT9R@V92HT9R@01bH02B@V92HT9R@V
90079P0992HT9R@V92HT00TV0`4]008100TT9P4V92HT9R@00P402`0102@R0B818P4R00D100DR0B81
8P0Q0@<000<S0@404@000`40000]00030@0002<0000Z00030@0000<000<100003P000`4000080003
0@0002@000<100000P00E@4V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P000`4a00070BHT0B@V90030@0@
900102@00B818P4R0B818PT100/R0B818P4R0B818P0D0@@000<1000040000`40000]00030@0002<0
000Z00030@0000<000<100003`000`40000700030@0002@000<100000`005@4V02H09P0V02H09P0V
02H09P0V000D9P0592HT9R@02bH01B@V92HT00LV00<T9RH00bH00b@V90039P<1=0020@039P4V0081
01`T9R@V904T004R0B818P4R0B818P4R0B818P4R804200038`40018000<10000;0000`40000S0000
:P000`40000300030@00010000<100001P000`40000U00030@0000<004d19P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
00030CP000hR0@419R@V900T0B@0908102<R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B81
8P030@038P4100<100<R0@400`4300030@00014000<10000;0000`40000S0000:P000`4000030003
0@00014000<100001@000`40000V00030@00008001`102H09P0V02H09P0V02H09P0V02H09P0V02H0
8RH01b@V92HT9R@01RH20CX00P403B@19R@V92HT0B@V90001@404B818P4R0B818P4R0B818P4R01h1
0P000b<0000C00030@0002/000<100008`0002X000<100000`000`40000B00030@0000@000<10000
9`000`40000200180@0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H00P4i00<10P020@0>92HT9R@1900T02@R0B870@0I8P4R0B81
8P4R0B818P4R0B818P4R0B818P050@058P410B801`4200030@0001<000<10000:`000`40000S0000
:P000`40000300030@0001<000<100000`000`40000X00030@000080020102H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V028V00<T9RH00P4i00811@005B@192@V92H19R@V90018P4R0B818P090@03
8P4R02010P000`40000D00030@0002X000<100008`0002X000<100000`000`40000=00030@0000<0
00<100000`000`40000Y00030@000080048102H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H0
9P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P020CP00`48000M0B@V92H102@0900T8P4R0B81
8P4R0B818P4R0B803`403B818P4R0B818P4R0B801`4J00030@0002T000<100008`0002X000<10000
0`000`40000=00030@0000<000<100000`000`40000Z000W0@00004V0BH09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H001`V0`4h00812`020@0X92HT0B@V92HT8P4R0B818P4R0B818P4R0B818P4R
0B818P4R0B818QH100<S00005`000`40000Y00030@0002<0000[00050@0000404@001@40000100H0
0P4/00100@00004V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P81>0030@h000X19R@192HT02@00`409B818P4R0B818P4R0B818P4R0B818P4R0B81
8P4R0B818P4R0B801@400b818P060@038`0001L000<10000:0000`40000S0000;0030A<00`470003
0@0002`002T100000BH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V000C9P<1>0020A40
00/T0BH19R@V92HT8P070@0I8P4R0B818P4R0B818P4R0B818P4R0B818P0F0AT000<10000:0000`40
000S0000C0000`40000]000i0@00004V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H00081>0030A@000lV904T9R@090018P4R0B801@406B818P4R0B818P4R0B81
8P4R0B818P4R0B800`402b818P4R0B818P4R00@16`000`40000W00030@0002<0001<00030@0002h0
02d100000BH09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P002BH20CL00`4G0003
0B@1008T00dV92HT8P4R0B818P4R00T100TR0B818P4R0B806`4L00030@0002H000<100008`0004`0
00<10000;`00<`4000019P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P01
00020CL00P4K0009904T9R@0900100DR00X18P4R0B818P4R2`404B818P4R0B818P4R0B818P4R00<1
00<R0B80104L00030@0002H000<100008`0004`000<10000<000;`4000019P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H000<1=P030Ad000LT0B@T9R@V00<101dR0B818P4R0B81
8P4R0B818P4R0B818P4R0B818P0E0@03004001`000<100009@000`40000S0000C0000`40000`0003
0@00008002L19P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P0V02H09P000`4g00818@001`4T
9R@090400R820@@R02018P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P<100DR0B818P040Al0
00<100009@000`40000S0000C0000`40000a00030@00008002@19P0V02H09P0V02H09P0V02H09P0V
02H09P0V02H09P0V02H20CL00`4S00050B@19R@00P400b818P050@0G8P4R0B818P4R0B818P4R0B81
8P4R0B8050400`01000N00030@0002@000<100008`0004`000<10000<P000`400002000P0BH09P0V
02H09P0V02H09P0V02H09P0V02H09P0V02H30CL00P4V00060B@V904T22820@0I8P4R0B818P4R0B81
8P4R0B818P4R0B818P050@098P4R0B818P4R00D17`000`40000T00030@0002<0001;00030@0103@0
00<100000P00704V02H09P0V02H09P0V02H09P0V02H09P0V02H30CL00`4X000@0B@19P4V0B818P4R
0B818PD100dR0B818P4R0B818P4R00T100<R0@403@4Q00030@0002<000<100008`0004/00P4f0003
0@00008001T19P0V02H09P0V02H09P0V02H09P0V02H00081>0020B/00P4012HT0@4=8P030B8100h1
00lR0B818P4R0B818P4R0B801`4S00030@00028000<100008`0004/000<10000=P000`400002000E
0BH09P0V02H09P0V02H09P0V02H000<1=`030BX00`401P010@4T9P<1024R0B818P4R0B818P4R0B81
8P4R0B818P4R0B818P4R0B80404S00030@00028000<100008`0004/000<10000=`000`400002000B
0BH09P0V02H09P0V02H09P0V0P4h0081:P030@D000@T0BHT0R820@dR01h18P4R0B818P4R0B818P4R
0B818P4R0B818P4R0B850B@000<100008@000`40000S0000B`000`40000h00030@00008000l19P0V
02H09P0V02H09P000P4g00<1:P020@P00P401R@V0B818P<101dR0B818P4R0B818P4R0B818P4R0B81
8P4R0B818P0@0B@000<100008@000`40000S0000B`000`40000i00030@00008000/19P0V02H09P0V
00030CL00P4Z00<12`000`4V90068P<12R804P4R0B818P4R0B818P4R0B818P<100LR0B818P4R00@1
9P000`40000P00030@0002<0001;00030@0003X000h100000@0V02H09P0V0081=P030BX00P4>0081
00TV0B818P4R0B800`405B818P4R0B818P4R0B818P4R0B818P070@038P4100T19`000`40000O0003
0@0002<0001;00030@0003/000X100000@0V02H00`4e00<1:P030A0000@19R@10b800bDR8P048PD1
0b802P4R0B818P4R0B830@0?8P4R0B818P4R0B818P4R00@19`000`40000O00030@0002<0000Y00@1
00<00@406`000`40000l00070@0000409P020CH00P4[008150000`4V0@020@0=8P4R0B818P4R0B81
8P0=0@0;8P4R0B818P4R0B802P4Y00030@0001h000<100008`0002P000<100000P020A`000<10000
?0000`4000020081=@030BX00`4F00050B@R8P404280604R0B818P4R0B818P4R0BL18P4R0B818P<1
00<R0@400P4X00030@0001h000<100008`0002P000<100000`000`40000J00<1?@000`40000f0081
:P030AT000@1904R0`408b818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R00X1:P000`40
000N00030@000280000X00030@0000<000<100006P000`40000n00030@0003800`4Z00817@000b@R
8P038P813280504R0B818P4R0B819`4W0BL18P4R0`401B818P4R00<1:P000`40000N00030@000280
000X00030@0000<000<100006P000`40000o00030@0002l00P4Z00<17`020@058P4R0B801@406b81
8P4R0B818P4R0B818P4R0B818P4R0B818P0;0B`000<100007@000`40000R0000:0000`4000030003
0@0001X000<10000@0000`40000[00<1:P020B8000<19R8012801BDR9B8000@10R804P4R0B818P4W
0BL19`4W0BL18P<100LR0B818P4R00@1;@000`40000L00030@000280000X00030@0000<000<10000
6P000`40001100030@0002P00P4Z00<19@030@0;8P4R0B818P4R0B801@402B818P4R0B818P070@03
8P4R00X1;P000`40000L00030@000280000W008110020A`000<10000@P000`40000T00<1:@030BP0
00<18P401B802RDR028U8P0R028<0@0?9`4W0B818P4R0B819`4R00<1;`000`40000K00030@000280
001;00030@0004<000<100008@020BX00P4[00030B81008102DR0B818P4R0B818P4R0B818P4R0B81
8P4R0B818P4R0B818P4R00H1;`000`40000K00030@000280001;00030@0004@000<100007@030BT0
0`4]00040B8U8P810R808BDR028U8P0R02808P0R004W0BL19`4W0BL18P4R0BL19`030@038P410340
00<100006P000`40000R0000B`000`40001500030@0001X00P4Z0081<0001P4R0B818P<101lR0B81
8P4R0B818P4R0B818P4R0B818P4R0B818P4R00L1<@000`40000J00030@000280000[00<120020@X0
104500<1AP000`40000F00<1:@030C<000L19B8U8RDR008101dU8RDR9B808P0R028W0BL19`4W0BL1
9`4R0BL19`030@048P410C<000<100006@000`40000R0000:P001@40000100L00P4<00030@0000@0
00<10000A`000`40000C0081:@030CH00P401b818P4R0B800`405B818P4R0B818P4R0B818P4R0B81
8P050@038P4100@1=0000`40000H00030@000280000Y00030@0000<000<100004P000`4000040003
0@0004L000<1000040030BT00P4j000:0B8U8RDR02808P8100lU8P0R02808RL19`4W0BL01`401bL1
9`4W0B800P4e00030@0001P000<100008P0002T000<100000`000`40000=00L11@000`4000180003
0@0000`00`4Y00<1?000304R0B818P4R0B818Q4100TR0B818P4R0B80104g00030@0001L000<10000
8P0002T000<100000`000`40000=00030@00008000<1000010000`40001900030@0000T00P4Z0081
@000:04U8RDR9B808P0R9B808P0R004W0BL100400@019`4W0BL19`4R0B8h00030@0001L000<10000
8P0002T000<100000`000`40000>00050@0000401P000`40001:00030@0000D00`4Y00<1@P009B81
8P4R0B818P4R0B818P4R0B818P4R0B818P4R0B818P4R0B800P4j00030@0001H000<100008P0002T0
00<100000`000`40000>00050@0000401P000`40001;00030@0000800P4Y00<1AP009BD10B8U8P0R
02808P0R02808P0R02800@01004W0BL19`4W0B80?0000`40000E00030@000280000Y00030@0000<0
00<100003`001040004600030@0004`0104Y0081B@000`4R0@020@0M8P4R0B818P4R0B818P4R0B81
8P4R0B818P4R0B800P4m00030@0001D000<100008P0002T000<100000`000`40000?00040@000@H0
00<10000B`030BP00`5<00048RDR9@8101XR02808P0R02808P0R02800@01004W0BL19`81?`000`40
000D00030@000280000Y00030@0000<000<1000040000`400@0600030@0004P00`4300030@0002<0
0P5?00058P4R0B800`405B818P4R0B818P4R0B818P4R0B818P030D4000<1000050000`40000R0000
:P001@40000101<000<100401P030DH00P4700030@0001l00`5B00079B8U8RDR00050@0=02808P0R
00400@0100050D<000<100004`000`40000R0000:`030AD00P4600030@0004<00`4:00030@0001`0
0P5F000;8P4R0B818P4R0B803`400b81001300030@0001<000<100008P0004/000<10000@@020@h0
00<1000060030ET001XU8RDR02808P0R02808P0R00400@019`4W0DH000<100004P000`40000R0000
B`000`40000n00<14@000`40000D00<1H@0052818P4R0B818P4R0B818P4R0B81B0000`40000A0003
0@000280001;00030@0003`00P4D00030@0001800P5V000@8P0R02808P0R00400@019dX000<10000
4@000`40000R0000B`000`40000i00<15`000`40000>00<1J0001b818P4R0B80E0000`40000@0003
0@000280001;00030@0003L00P4K00030@0000/00P5]00038P0R05H000<1000040000`40000R0000
B`000`40000d00<17P000`40000700<1b@000`40000?00030@000280001;00030@0003400`4R0003
0@0000<00`7=00030@0000h000<100008P0004/000<10000;`020BH000<100000P7@00030@0000h0
00<100008P0004/00`4/00<1:@020M<000<100003@000`40000R0000B`000`40000Z0081;0000`40
003B00030@0000d000<100008P0004/000<100009`030Bh000<10000d`000`40000<00030@000280
001;00030@0002D00P4a00030@000=<000<1000030000`40000R0000B`000`40000R00<1<`000`40
003D00030@0000/000<100008P0004/000<1000080020CH000<10000e@000`40000:00030@000280
001;00030@0001d00`4h00030@000=D000<100002P000`40000R0000B`000`40000K0081>`000`40
003F00030@0000T000<100008P0004/000<1000060030Cd000<10000eP000`40000900030@000280
001;00815`020D0000<10000e`000`40000800030@000280001;00030@0001<00`5200030@000=P0
00<100001`000`40000R0000B`000`40000A0081A@000`40003H00030@0000L000<100008P0004/0
00<100003P030DL000<10000f@000`40000600030@000280001;00030@0000/00`5:00030@000=T0
00<100001P000`40000R0000B`00104000480081C@000`40003J00030@0000D000<100008P0004/0
00@100011@030Dl000<10000fP000`40000500030@000280001;00040@000@<00P5B00030@000=/0
00<1000010000`40000R0000B`000`4000040E@000<10000g0000`40000300030@000280000[00<1
20020@T00`4700@1E`000`40003L00030@0000<000<100008P0002X000D100000@07008120001@40
000100H00P400`01001F00030@000=d000<100000P000`40000R0000:@000`40000300030@0000d0
00<100000`000`40000400030@0105L000<10000g@000`40000200030@000280000Y00030@0000<0
00<100003@000`40000300030@0000D00P5G00030@000=h000D100000@0T0000:@000`4000030003
0@0000d000<100000`000`40000600030@0005D000<10000g`001040004T0000:@000`4000030003
0@0000d00P4300030@0000P000<100000P000`40001?00030@000=l000@10001900002T000<10000
0`000`40000=00030@01008130020@8000<10000C`000`40003P00030@0102@0000Y00030@0000<0
00<100003@000`40000@00030@01054000<10000h0000`400@0T0000:@000`40000300030@0000h0
00<1000040020E4000<10000h@020B@0000Y00030@0000<000<100003P000`40000A00030@0004l0
00<10000h@020B@0000Z00050@0000404P000`40000@0081D0000`40003Q0081900002/00`4D00@1
40000`40000200030@0004P000<10000gP030BH0001G00050@000040BP000`40003K00<1:@0005P0
00@10001BP000`40003H00<1;00005T000<10040BP000`40003E00<1;`0005X00P5:00030@000=40
104b0000F`000`40001800030@000<h00`4f0000G0020@<000<10000@P000`40003;00<1>@0005h0
00@10001A0000`40003800<1?00005l000<10040A0000`40003500<1?`0006000P5400030@000<80
0`520000H@000`40001200030@000;l00`550000C0030A<000<100000P000`40000l00030@000;`0
0`580000B`001@40000101<000D100000@0n00030@000;T00`5;0000BP000`40000300030@000140
00@10001?P000`40002f00<1CP0004X000<100000`000`40000B00030@0100D000<10000=P000`40
002c00<1D@0004X000<100000`000`40000C00811@000`40000f00030@000:l0105D0000BP000`40
000300030@0001@00P4400030@0003H000<10000[0030EP0001:00030@0000<000<100005P001@40
000103P000<10000Z@030E/0001:00030@0000<000<100005`001040004h00030@000:H00`5N0000
BP000`40000300030@0001P000<10040>0000`40002S00<1H@0004X000<100000`000`40000I0081
>0000`40002P00<1I00004/000D100000@0M00030@0003H000<10000W@030FL0001<00<17`000`40
000200030@00030000<10000VP030FX0001_00050@000040<P000`40002G00<1K@00070000@10001
<P000`40002D00<1L000074000<10040<P000`40002@00@1L`0007800P4b00030@0008d00`5g0000
L`020C4000<10000RP030GX0001e00050@000040;0000`40002700<1O@0007H000@10001;0000`40
002400<1P00007L000<10040;0000`40002100<1P`0007P00P4/00030@0007h00`660000N@000`40
000Z00030@0007/00`690000NP001@40000102L000<10000N0030H`0001k00040@000BL000<10000
M@030Hl0001l00030@0102L000<10000LP030I80001m00819`000`40001^00@1U@0007h00P4V0003
0@0006/00`6I0000P0001@400001024000<10000J0030I`0002100040@000B4000<10000I@030Il0
001Y00L14P000`400@0500030@0001T000<10000HP030J80001/00030@0001@00P4500030@0001T0
00<10000G`030JD0001/00030@0001D000<100000`000`40000I00030@0005`00`6X0000K0000`40
000F00030@00008000<100006@000`40001I00<1Z`0006`000<100005`001@40000101/000<10000
EP030Jh0001/00030@0001P000@100016`000`40001C00<1/@0006`000<100006@000`400@0K0003
0@0005000`6d0000K0000`40000J00816`000`40001<00@1]`0006`000<1000070001@40000101H0
00<10000B@030K/0001/00030@0001d000@100015P000`40001600<1_P0006T00P400`01000O0003
0@0101H000<10000@`030L40001[00818@020AH000<10000@0030L@0002?00030@0001@000<10000
?@030LL0002@00050@0000404@000`40000j00<1bP00094000@100014@000`40000g00<1c@000980
00<100404@000`40000d00<1d00009<00P4A00030@0003400`7C0000U0020A0000<10000;@040MH0
002F00050@0000402`000`40000Z00<1fP0009L000@100012`000`40000W00<1g@0009P000<10040
2`000`40000T00<1h00009T00P4;00030@0002400`7S0000VP000`40000200030@0000@000<10000
7P030NH0002K00050@0000401P000`40000K00<1j@0009`000@1000110000`400@0J00<1k00009d0
00<1004010000`400@0G00<1k`0008D0204A008110000`400@0D00<1lP0008H000<100000`000`40
000@00030@00008000<100404@030OD0002700030@0001H00P4200030@0100d0107h0000R0000`40
000G00050@0100402P030O`0002900030@0001L00P400`01000600<1o`0008X000<100005`000`40
0@0400<1o`030000R`000`40000G008100@00@41o`060000S0000`40000G0081o`090000QP000`40
000300030@000?l08P0008H000<100000`000`40003o0280002700050@000040o`0U0000R0030Ol0
9P000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000:40107o00`0002P00030@00008000<10000o`090000YP000`40003o00P0
002V00030@000?l020000:H000<10000o`080000Y@000`40003o00T0002R00<1o`0<0000Y@000`40
003o00T0002V00030@000?l020000:H000<10000o`080000X0000`40000200030@000?l02@000:40
107o00`0003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40
003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40
003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40001g00D1o`0e0000NP000`40003o03@0
001j00030@000?l0=00007/000<10000o`0c0000N`000`40003o03<0001j00030@010?l0=00007X0
00<10040o`0d0000N@001@4000010?l0<`0007T000D100000@3o03<0001h00030@0000<000<10000
o`0`0000N0000`40000300030@000?l0<00007L00`4300<1o`0a0000o`2a0000o`2a0000o`2a0000
\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.234686, 0.31168, 0.00465498, 
  0.00125278}}]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Internal Documentation", "Section"],

Cell[CellGroupData[{

Cell["Global variables and functions", "Subsection"],

Cell[TextData[{
  "The code in the package is using the global variables ",
  StyleBox["nodes",
    FontWeight->"Bold"],
  ", ",
  StyleBox["elements",
    FontWeight->"Bold"],
  ", ",
  StyleBox["segments",
    FontWeight->"Bold"],
  " and ",
  StyleBox["Node2Degree",
    FontWeight->"Bold"],
  ".\nModifying those may cause problems and lead to wrong results."
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Internal variables and functions", "Subsection"],

Cell["\<\
{dof , Node2Degree} = FindDOF[nodes]
   dof is the number of degrees of freedom
   Node2Degree contains the mapping of the nodes to the DOF, an entry 0 \
corresponds to a fixed value
   
 {elMat, elVec} =  ElementContribution[corners, aCoeff, bCoeff, fCoeff]
    find the  element stiffness matrix and vector for an element with given \
corners. 
    aCoeff, bCoeff, fCoeff are the functions describing the PDE 
   
segVec=ElementContributionEdge[corners, gCoeff]
    find the contribution of one segment, Nemann condition with function \
gCoeff   
   
insertElement[element, aCoeff, bCoeff, fCoeff, gCoeff]
    insert the conribution of the element with the given nodes into Aglobal \
and bglobal
    
insertElementEdge[sdge, gCoeff]
    insert the conribution of the element with the given nodes into  bglobal
   
 Aglobal  the global stiffness matrix
 bGlobal  the global vector\
\>", "Text"]
}, Open  ]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
WindowSize->{842, 905},
WindowMargins->{{Automatic, 47}, {-15, Automatic}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{597.562, 842.375},
"PaperOrientation"->"Portrait",
"Magnification"->1},
Magnification->1.5
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1731, 51, 59, 1, 150, "Title"],
Cell[1793, 54, 139, 5, 96, "Text"],

Cell[CellGroupData[{
Cell[1957, 63, 30, 0, 77, "Section"],
Cell[1990, 65, 2795, 55, 1096, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4822, 125, 34, 0, 77, "Section"],

Cell[CellGroupData[{
Cell[4881, 129, 83, 1, 64, "Subsection"],
Cell[4967, 132, 42, 1, 39, "Input"],

Cell[CellGroupData[{
Cell[5034, 137, 46, 1, 39, "Input"],
Cell[5083, 140, 230, 3, 79, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5350, 148, 50, 1, 39, "Input"],
Cell[5403, 151, 136, 2, 56, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5576, 158, 46, 1, 39, "Input"],
Cell[5625, 161, 154, 3, 56, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5816, 169, 48, 1, 39, "Input"],
Cell[5867, 172, 126, 2, 56, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6030, 179, 54, 1, 39, "Input"],
Cell[6087, 182, 156, 2, 79, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6280, 189, 53, 1, 39, "Input"],
Cell[6336, 192, 149, 2, 56, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6522, 199, 53, 1, 39, "Input"],
Cell[6578, 202, 149, 2, 56, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[6776, 210, 55, 0, 64, "Subsection"],
Cell[6834, 212, 98, 2, 39, "Input"],
Cell[6935, 216, 76, 1, 39, "Input"],

Cell[CellGroupData[{
Cell[7036, 221, 108, 2, 85, "Input"],
Cell[7147, 225, 144393, 6049, 272, 66130, 5078, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[151589, 6280, 34, 0, 38, "Subsection"],
Cell[151626, 6282, 494, 11, 200, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[152157, 6298, 32, 0, 38, "Subsection"],

Cell[CellGroupData[{
Cell[152214, 6302, 300, 6, 108, "Input"],
Cell[152517, 6310, 157008, 5397, 444, 95645, 4635, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[309574, 11713, 34, 0, 38, "Subsection"],

Cell[CellGroupData[{
Cell[309633, 11717, 147, 3, 85, "Input"],
Cell[309783, 11722, 70250, 3616, 279, 35790, 3186, "GraphicsData", 
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[380070, 15343, 153, 3, 78, "Input"],
Cell[380226, 15348, 58887, 3333, 265, 33881, 3020, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[439150, 18686, 318, 6, 131, "Input"],

Cell[CellGroupData[{
Cell[439493, 18696, 128076, 3230, 444, 75675, 2578, "GraphicsData", 
"PostScript", "Graphics"],
Cell[567572, 21928, 166684, 6756, 444, 111367, 6068, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{
Cell[734329, 28692, 41, 0, 45, "Section"],

Cell[CellGroupData[{
Cell[734395, 28696, 52, 0, 64, "Subsection"],
Cell[734450, 28698, 375, 14, 71, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[734862, 28717, 54, 0, 64, "Subsection"],
Cell[734919, 28719, 905, 24, 496, "Text"]
}, Open  ]]
}, Closed]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

